package charite.christo;
import charite.christo.interfaces.*;
import java.awt.Container;
import java.io.File;
import java.lang.ref.*;
import java.net.URL;
import java.util.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP

   <h3>Introduction</h3>

   For the default sequence and 3D alignment programs,
   WIKI:executables are included in STRAP - ready
   to use for Windows, Macintosh and Linux.  The versions of
   these pre-compiled executables may not be up-to-date.

   STRAP provides a more general installation method for all external
   programs embedded in STRAP: Installation from WIKI:Source_code.
   This method is applicable to all embedded programs and ensures that
   the most recent version is installed.  It may result in higher
   execution speed.

   <br>
   <br>

   Two program packages are required on your PC to enable installation
   from source code: WIKI:compilers and WIKI:make_(software).  If you
   are an expert you may enhance execution speed by specifying
   optimized compiler options.

   <br><br>

   <h3>Installing compilers</h3>

   XXX <br><br>

   Adding support for C/C++ and Fortran to your system does not improve already
   installed programs unless they are deinstalled: Simply  delete the respective
   folders in the directory ~/.StrapAlign/bin/.

   <h3>Testing, whether compilers are already installed</h3>
   Usually, the C-compiler is started with the command "cc" and the C++ compiler with the command "c++".
   Fortran compilers are started with "gfortran" or "f77" or "f95".
   If  C++ is installed then typing into the command shell
   "c++"  brings the output
   <pre>c++: no input files</pre>
   But the message
   <pre>c++: command not found</pre>
   indicates that either C++ is not installed yet or that the PATH-variable is not set correctly.

   @author Christoph Gille
   The abstract class <i>AbstractExecutable</i> is used to make a Java wrapper for a
   program that is written in  C or C++ or any other compatible computer language other than Java.
   <br>
   Conventions:
   <ul>
   <li>There is only one executable.</li>
   <li>The filename of the executable is getName() +  ".exe".</li>
   <li>
   </ul>
*/
public class AbstractExecutable implements CanBeStopped,Disposable,ChRunnable,HasPrgParas  {
    public final static String
        DEFAULT_BINARY="*",
        REPLACEME_INLCUDE_STRING[][]={{"#include <string>","#include <string.h>"}};
    /* >>> Instance >>> */
    private String name;
    public void setName(String n) { name=n;}
    public AbstractExecutable(String n) { setName(n); }
    public AbstractExecutable() {}
    public final String getName() {return name;}
    public final void iniT() {
        if (!Insecure.EXEC_ALLOWED) return;
        final ChExec ex=myExec;
        if(ex!=null) ex.dispose();
        installPackage();
        delFile(fileOut());
        delFile(fileIn());
        CanBeStopped.vALIGNMENTS.add(weakRef(this));
    }
    @Override public String toString() {
        return TabItemTipIcon.item(this) +
            (sze(getPackageURLs(false))>0 ? " The program file will be installed automatically." : "" )+
            (this.srcInstScript!=null ?  " It will be installed from source. " : "");
    }
    public void dispose() {
        if (myExec!=null) try { myExec.dispose();} catch(Throwable e){}
        myExec=null;
    }
    /*
    @Override protected void finalize() throws Throwable {
        logFinalize(ANSI_BLACK+ANSI_FG_WHITE+"AbstractExecutable",getClass());
        dispose();
        super.finalize();
    }
    */
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Files and Directories >>> */
    private File dirBinaries_,fileExecutable_,dirTmp_,fileIn_,fileOut_;
    /** A directory for temporary files. */
    private static int countInstances;
    /** temporary directory for transient data. Will be deleted after the program session. */
    public final File dirTemp() {
        if (dirTmp_==null) {
            mkdrs(dirTmp_=file(ChUtils.dirTmp()+"/"+getName()+"/"+ countInstances++ ));
            delFileOnExit(dirTmp_);
        }
        return dirTmp_;
    }
    /** You can use this as your input file. */
    public final File fileIn() { if (fileIn_==null) fileIn_=file(dirTemp(),"input"); return fileIn_;}
    /** You can use this as your output file. */
    public final File fileOut() {
        if (fileOut_==null) fileOut_=file(dirTemp(),"output");
        return fileOut_;
    }
    /** The directory of the installed program. */
    public final File dirBinaries() {
        final String n=getName();
        if (n!=null) {
            if (dirBinaries_==null) mkdrs(dirBinaries_=file(dirBin(),n));
            if (!dirBinaries_.isDirectory()) delFile(dirBinaries_);
        }
        return dirBinaries_;
    }
    /** The File of the executable.  is name() plus  dot "exe"  */
    public File fileExecutable() {
        if (fileExecutable_==null) fileExecutable_=file(dirBinaries(),getName()+".exe");
        return fileExecutable_;
    }
    /* <<< Files and Directories <<< */
    /* ---------------------------------------- */
    /* >>> Installation >>> */
    private boolean sentMsgThatInstallationFailed, version;
    private String binaryPackageURLs[], sourcePackageURLs[], srcInstScript;
    /**
       Set URL of the zip-archive for the program binary.
       An array of alternative URLs can be provided.
       Each String may contain a list of required zip files.
       Each URL may be followed by a file path denoting a file inside the zip archive that is the exe file.
    */
    public void setBinaryPackageURLs(String...urlPipeFileExecutable) {
        if (sze(urlPipeFileExecutable)==0) this.binaryPackageURLs=null;
        else {
            for(int i=urlPipeFileExecutable.length; --i>=0; ) {
                if (urlPipeFileExecutable[i]==DEFAULT_BINARY) {
                    urlPipeFileExecutable[i]=urlOfBinariesForThisOS()+getName()+".zip.gz";
                }
            }
        }
        this.binaryPackageURLs=urlPipeFileExecutable;
    }
    // See SubcellularLocationSherLoc
    /**
       Set URL of the zip-archive for the program source.
       An array of alternative URLs can be provided.
       Each String may contain a list of required zip files.
    */
    public final void setSourcePackageURLs(String...urls) { this.sourcePackageURLs=urls; }
    private final String[] getPackageURLs(boolean isSource) {
        if (isSource) {
            String[] uu=this.sourcePackageURLs;
            if (uu==null && srcInstScript!=null) uu=this.sourcePackageURLs=new String[]{URL_STRAP_SRC+getName()+".zip.gz"};
            return uu;
        } else return binaryPackageURLs;
    }

    /** @return a shell script, usually a compile command. Type the string "C_COMPILER" instead of "gcc" and "C_PLUS_PLUS_COMPILER" instead of "g++". */
    public String setSourceInstallationScript(String shell_script) {
        return srcInstScript=shell_script;
    }

    private List vSrcRplc=new ArrayList();
    public void replaceInSourceFile(String[][] replacements, File f) {
        (vSrcRplc==null ? vSrcRplc=new ArrayList() : vSrcRplc).add(new Object[]{replacements,f});
    }

    private boolean extract(String urlsPipeExe[], boolean isSource) {
        if (!Insecure.EXEC_ALLOWED) return false;
        boolean success=false;
        for(int iTry=0; iTry<sze(urlsPipeExe); iTry++) {
            final String urls[]=splitTokns(0L, urlsPipeExe[iTry]);
            for(int iU=0; iU<urls.length; iU++) {
                final File fZip=zipFile(urls[iU],isSource);
                if (fZip==null) continue;
                mkParentDrs(fZip);
                if (!looks(LIKE_EXTURL,urls[iU])) {
                    try {
                        cpy(CLASSLOADER.getResourceAsStream(urls[iU]),fZip);
                    } catch(Exception e){putln("caught in AbstractExecutable#extract ",e);}
                } else {
                    final URL url=url(withoutLastComponent(urls[iU],'|'));
                    InteractiveDownload.downloadFiles(new URL[]{url(url)},new File[]{fZip});
                    if (sze(fZip)==0) { error("Did not download <br>"+fZip); continue; }
                }
                new ChTarPROXY().extractAll(fZip,dirBinaries());
                final String exeInZip=lastComponent(urls[iU],'|');
                if (sze(fileExecutable())==0 && sze(exeInZip)>0) {
                    final File fE=file(dirBinaries()+"/"+exeInZip);
                    if (sze(fE)==0) putln(ANSI_RED+"Error in AbstractExecutable "+ANSI_RESET+" no such file ",fE);
                    else cpy(fE,fileExecutable());
                }
                final String files[]=dirBinaries().list();
                if (sze(files)>0) {
                    for(String f : files) if (f.endsWith(".exe")) makeExecutable(file(dirBinaries(),f));
                    success=true;
                } else if (sze(fZip)>0) error("Unable to extract "+fZip);
            }
            if (success) break;
        }
        return success;
    }
    private static File zipFile(String u, boolean isSource) {
        if (!Insecure.EXEC_ALLOWED) return null;
        final File dir=file(dirBin()+ (isSource ? "/downloadedSourcePackages" : "/downloadedBinaryPackages"));
        final String s=u.substring(u.lastIndexOf('/')+1);
        return file(dir,delSfx(".bz2",delSfx(".gz",s)));
    }
    public void installPackage() {
        if (!Insecure.EXEC_ALLOWED) {
            Insecure.tellNoPermission('X',"");
            return;
        }
        final String n=getName();
        if (n==null) return;
        if (!isEDT()) ChDelay.invokeAndWait(this,"installPackage",null);
        else if (sze(fileExecutable())==0) {
            if (!version) version=currVers(getName());
            extract(getPackageURLs(true), true);
            for(Object[] rplc_f : toArry(vSrcRplc,Object[].class)) rplcInTextFile( (String[][])rplc_f[0], (File)rplc_f[1] );
            putln(ANSI_BLUE,vSrcRplc,ANSI_RESET);
            make(this.srcInstScript);

            if (sze(fileExecutable())==0) extract(getPackageURLs(false),false);

        }
    }
    /* <<< Installation <<< */
    /* ---------------------------------------- */
    /* >>> Execute >>> */
    /** Show a message telling that the program could not be started */
    public final void failedToLaunch() {
        if (myExec==null || !myExec.couldNotLaunch()||sentMsgThatInstallationFailed) return;
        final String urlHome=TabItemTipIcon.homePage(this);
        final JComponent msg=
            vBhB(
                 pnl("<h2>Failed to run </h2>",getName()),
                 " ",
                 getCtrlPanel(myExec),
                 hB("Please install ",fileExecutable(),"#"),
                 " ",
                 urlHome!=null ? hB("Web:",urlHome) : null
                 );
        error(0L,null,msg);
    }
    /** For internal use. If override please make a call super.run(...) */
    public Object run(String id,Object arg) {
        if (!Insecure.EXEC_ALLOWED) return null;
        if (id=="installPackage") installPackage();
        if (id=="AFTER_COMPILATION") {
            final JComponent pSouth=(JComponent)arg;
            final boolean success=sze(fileExecutable())>0;
            pSouth.add(pnl( C(success ? 0x00b200:0xFF0000), success ? "compilation finished successfully" : "compilation failed"));
            if (!success) {
                if (sze(getPackageURLs(false))>0) pSouth.add(hB("As a workaround I am going to install the pre-compiled package","#"));
                pSouth.add(hB("For Installing compilers click ", ChButton.actionViewMessage("Install compiler ...",msgInstallCompilers()),"#"));
            }
        }
        if (id=="COMPILATION") {
            final Object oo[]=(Object[])arg;
            final ChExec exec=(ChExec)oo[1];
            final JComponent pSouth=(JComponent)oo[0];
            exec.run();
            ChDelay.invokeAndWait(AbstractExecutable.this,"AFTER_COMPILATION",pSouth);
            final java.awt.Window d=parentWndw(myComponent);
            if (d instanceof JDialog) d.dispose();
        }
        return null;
    }
    private static JComponent myComponent;
    private ChExec execMake;
    private void make(String commands) {
        if (!Insecure.EXEC_ALLOWED || commands==null) return;
        String cmd=rplcToStrg("C_PLUS_PLUS_COMPILER",  getCustomSetting(Customize.CplusPlus_compiler), commands);
        cmd=rplcToStrg("C_COMPILER",getCustomSetting(Customize.C_compiler),cmd);
        cmd=rplcToStrg("FORTRAN_COMPILER", getCustomSetting(Customize.fortran_compiler),cmd);
        cmd=rplcToStrg("MNO_CYGWIN_FLAG",isWindows() ? "-mno-cygwin" : "",cmd);
        if (isWindows()) cmd=ChExec.WINDOWS_addBinToPath+cmd;
        final  File fScript=file(dirBinaries(),"makeByStrap.sh");
        wrte(fScript,cmd);

        execMake=new ChExec(ChExec.STDOUT | ChExec.STDERR | ChExec.IGNORE_ERROR).dir(dirBinaries());
        final Container
            ctrl=getCtrlPanel(execMake),
            log=vBhB(
                     hB("Installing ",getName()," from source code. This may take some minutes","#"),
                     fScript,
                     ctrl,
                     AbstractExecutable.myComponent=new JSeparator()
                     );
        execMake.setCommandLine(((isWindows() ? dirCygwin()+"\\bin\\sh.exe " : "sh ")+fScript.getName()));
        ChDelay.newThread(this,"COMPILATION",new Object[]{log,execMake}).start();
        InteractiveDownload.addTab(null,false, ChFrame.AT_CLICK);
        ChMsg.option(log, new Object[]{"abort installation"});
        InteractiveDownload.addTab(log, sze(fileExecutable())>0, ChFrame.AT_CLICK);
    }
    /* <<< Execution <<< */
    /* ---------------------------------------- */
    /* >>> Perl >>> */
    /** @return usually "perl". If Windows and cygwin installed then returen the path to perl.exe */
    public static String perlCommand() {
        if (isWindows()) {
            final File cygwinPerl=file(dirCygwin()+"/bin/perl.exe");
            if (sze(cygwinPerl)>0) return cygwinPerl.getAbsolutePath();
        }
        return "perl";
    }
    private static boolean isPerlInstalled;
    public static boolean isPerlInstalled() {
        if (!isPerlInstalled) {
            final ChExec ex=new ChExec(ChExec.IGNORE_ERROR).setCommandLineV(perlCommand(),"-version");
            ex.run();
            isPerlInstalled=ex.exitValue()==0;
        }
        return isPerlInstalled;
    }
    private static boolean isBioPerlInstalled;
    public static boolean isBioPerlInstalled() {
        if (!isBioPerlInstalled) {
            final Object cmd[]={perlCommand(),"-e","use Bio::Tools::Analysis::Protein::Sopma;"};
            final ChExec ex=new ChExec(ChExec.IGNORE_ERROR).setCommandLineV(cmd);
            ex.run();
            isBioPerlInstalled=ex.exitValue()==0;
        }
        return isBioPerlInstalled;
    }
    /** If BioPerl is not yet installed then install it or tell the user to do so. */
    public static void installBioPerl() {
        if (!Insecure.EXEC_ALLOWED) return;
        if (isBioPerlInstalled()) return;
        String msg="Install BioPerl.<br>See http://www.bioperl.org for details.<br>";
        if (isSystProprty(IS_LINUX) && systProprty(SYSP_OS_VERSION).indexOf("gentoo")>=0) msg+="Simply emerge BioPerl; <pre class=\"terminal\">emerge bioperl</pre><br>";
        new ChTextViewer(addHtmlTags(msg),0).shw(0);
    }
    /* <<< Perl <<< */
    /* ---------------------------------------- */
    /* >>> Command line options >>> */
    private final PrgParas prgPara=new PrgParas();
    public final PrgParas getPrgParas() { return prgPara;}
    /* <<<  Command line options <<< */
    /* ---------------------------------------- */
    /* >>> Process >>> */
    private boolean isStopped;
    private ChExec myExec;
    public ChExec exec() {
        if (myExec==null && Insecure.EXEC_ALLOWED) myExec=new ChExec(ChExec.IGNORE_ERROR);
        return myExec;
    }
    public boolean execNZ() {  return myExec!=null;}
    /** What the program wrote on stdout stream */
    public final byte[] getStdout() { ChExec e=myExec; return e==null ? null : e.getStdout();}
    /** What the program wrote on stderr stream */
    public final byte[] getStderr() { ChExec e=myExec; return e==null ? null : e.getStderr();}
    /** Download zip file and directory name and executable file name are derived from name()  */
    public boolean isStopped() { return isStopped;}
    public void stop() {
        isStopped=true;
        final ChExec ex=myExec;
        if (ex!=null) ex.kill();
    }

    /* <<< Process <<< */
    /* ---------------------------------------- */
    /* >>> Static tools >>> */
    public static Class[] checkForUpdates(Class classes[]) {
        if (!Insecure.EXEC_ALLOWED) return NO_CLASS;
        final List<Class> v=new ArrayList();
        for(Class c : classes) {
            if (isAssignblFrm(AbstractExecutable.class, c)) {
                final AbstractExecutable ex= HotswapUtil.instance(c, AbstractExecutable.class);
                for(boolean src : FALSE_TRUE) {
                    final String[] urlsPipeExe=ex!=null ?  ex.getPackageURLs(src) : null;
                    for(int iTry=sze(urlsPipeExe); --iTry>=0;) {
                        final String urls[]=splitTokns(0L, urlsPipeExe[iTry]);
                        for(int iU=0; iU<urls.length; iU++) {
                            final String url=withoutLastComponent(urls[iU],'|');
                            final File fZip=zipFile(url,src);
                            if (sze(fZip)>0 &&  0==isUpToDate(fZip,url(url))) adUniq(c,v);
                        }
                    }
                }
            }
        }
        return toArry(v,Class.class);
    }
    public static String msgInstallCompilers() {
        if (!Insecure.EXEC_ALLOWED) return "";
        final boolean isDebian=sze(file("/usr/bin/apt-get"))>0, isSuse=sze(file("/sbin/yast2"))>0;
        final String
            MSG=
            isDebian ?
            "To install the compilers on a Debian system run the following in a terminal: "+
            "<pre class=\"terminal\">\n"+
            "sudo apt-get install g++ gfortran make \n"+
            "</pre>" :
            isSuse ?
            "To install the compilers on a Suse system run the following in a terminal: "+
            "<pre class=\"terminal\">\n"+
            "sudo /sbin/yast2 --install gcc-c++ gcc-fortran make \n"+
            "</pre>" :
            isMacintosh() ?
            "A C++-Compiler is included in WIKI:Xcode. "+
            "Xcode is found on one of the CDs which came with your Macintosh. " +
            "With X-Code all external programs included in STRAP can be run on Macintosh. "+
            "The alignment program ClustalW will run much faster. "+
            "Unfortunately, Fortran is not included in Xcode. Tm-align is currently the only Fortran program in STRAP. " :
            isWindows() ?
            "The compilers are provided by WIKI:Cygwin. <i>BUTTON:charite.christo.Microsoft#newCygwinButton(\"CC\")!</i>" :
            "To install the compilers on your system open the software manager and activate the packages "+
            "<ul>"+
            "<li>C++-compiler</li>"+
            "<li>Fortran compiler</li>"+
            "<li>make</li>"+
            "</ul>";
        return rplcToStrg("XXX",MSG,getHlp(AbstractExecutable.class));
    }

}

