package charite.christo;
import java.net.URL;
import charite.christo.interfaces.*;
import java.awt.Container;
import java.io.File;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/*
AbstractProxy is used by classes that depend on jar files which are download when the class is used.
These classes cannot be loaded with the boot class loader because the boot classpath does not yet contain these jar files.
Instead they are loaded with a class loader which is part of STRAP.
*/
public abstract class AbstractProxy {
    private final static Object sync_po=new Object();
    private final static Map mapProxy=new WeakHashMap();
    private final static Map<String,Class> MAP=new HashMap();
    private final static Set<String>neverTryAgain=new HashSet(), alreadyFailed=new HashSet();
    private Object po;
    /**
        A problem occurs when a package is loaded with
        <I>ChClassLoader</i>
        that internally uses
        <i>JAVADOC:Class.forName(String)</i>.
        Class instances loaded by different class loaders are not compatible and this situation must be avoided.
        This is the case e.g. when dom4j.jar is in the classpath.
        Workaround: Test whether a certain class can be loaded with Class.forName(String).
        If yes then take boot class loader otherwise proceed as normal with the
        <i>JAVADOC:ChClassLoader</i>.
    */
    //private String takeBootclassloaderIfThisClassCanBeLoaded;
    //public void takeBootclassloaderIfThisClassCanBeLoaded(String className) { takeBootclassloaderIfThisClassCanBeLoaded=className;}
    public abstract String getRequiredJars();
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Native DLLs  >>> */
    public String jarsContainingNativeLibraries() { return null;}
    /** Space separated list of obsolete jar-files, which need to be deleted. Regular expressions are allowed. */
    private static void extractDlls(File dir, String files, String extractOrRemove[]) {
        if (dir==null || files==null) return;
        final StringTokenizer st=new StringTokenizer(files);
        while(st.hasMoreTokens()) {
            final java.util.regex.Pattern pattern=java.util.regex.Pattern.compile(st.nextToken());
            for(String file : dir.list()) {
                if (pattern.matcher(file).matches()) {
                    final File f=new File(dir,file);
                    if (extractOrRemove==NO_STRING) {
                        delFile(f);
                        if (f.length()>0) {
                            final String fn=f.getAbsolutePath();
                            final String del=getPrpty(AbstractProxy.class,"DELETE_FILES", "");
                            if (del==null || del.indexOf(" "+fn)<0)
                                setPrpty(AbstractProxy.class,"DELETE_FILES",del+" "+fn);
                        }
                    } else  {
                        try {
                            ChZip.extractAllFilesWithEnding( new java.util.zip.ZipFile(f),null,extractOrRemove,dir,false);
                        } catch(Throwable e){stckTrc(e);}
                    }
                }
            }
        }
    }
    private static boolean delet_DLL_done;
    {
        if (!delet_DLL_done) {
            delet_DLL_done=true;
            final String del=getPrpty(AbstractProxy.class,"DELETE_FILES", "");
            if (del!=null) {
                final StringTokenizer st=new StringTokenizer(del);
                while(st.hasMoreTokens()) delFile(new File(st.nextToken()));
                setPrpty(AbstractProxy.class,"DELETE_FILES","");
            }
        }
    }
    /* <<< Native DLLs <<< */
    /* ---------------------------------------- */
    /* >>> proxyObject >>> */
    public Object proxyObject() {
        synchronized(sync_po) {
            if (po==null) {
                final String cn=delSfx("PROXY",getClass().getName());
                if (neverTryAgain.contains(cn)) return null;
                else if (!Insecure.CLASSLOADING_ALLOWED) {
                    final Class c=Insecure.findClassByName(cn);
                    try {
                        this.po=c!=null ? c.newInstance() : null;
                    } catch(Exception e) {
                        putln(" Caught in AbstractProxy cn=",cn,"\n",e);
                    }
                    if (this.po==null) neverTryAgain.add(cn);
                } else if (loadDirectly) {
                    try {
                        this.po=Class.forName(cn).newInstance();
                    } catch(Exception e) {
                        putln(" Caught in AbstractProxy cn=",cn,"\n",e);
                        neverTryAgain.add(cn);
                    }
                } else {
                    if (alreadyFailed.contains(cn) &&  !ChMsg.yesNo("I already tried to load<br>"+clazzName(this)+"<br>Do you want to try again?")) {
                        neverTryAgain.add(cn);
                        return null;
                    }
                    /*if (this instanceof FindUtils.Interface) FindUtils.register((FindUtils.Interface)this);*/
                    final File dirJars=dirJars();
                    final URL u=url(ChConstants.URL_STRAP_JARS), uuRequired[]=Web.urls(u,getRequiredJars());
                    final boolean p200=ChZip.supportForPack200();
                    if (!p200 || !InteractiveDownload.downloadFilesAndUnzip(uuRequired,dirJars)){
                        if (!InteractiveDownload.downloadFilesAndUnzip(Web.urls(u,getRequiredJars().replaceAll(".pack.gz","")),dirJars)) return null;
                    }
                    if (p200) ChZip.unpack200(dirJars,ChZip.PACK_IF_NEWER);
                    final String obsolete=
                        " jortho.jar "+
                        " neobio.jar "+
                        " edu_sdsc.jar  edu_sdsc_[1-8].jar "+
                        " biojava-1.[45].* "+
                        " biojava.jar "; //!!!!!
                    extractDlls(dirJars,obsolete,NO_STRING);
                    { /* Protein-Workshop */
                        final String s=jarsContainingNativeLibraries();
                        if (s!=null) {
                            final String ln=System.mapLibraryName("libname"),  SUFFIX_DLL[]={ln.substring(ln.lastIndexOf('.'))};
                            extractDlls(dirJars,s.substring(s.lastIndexOf('/')+1),SUFFIX_DLL);
                        }
                    }
                    try {
                        ChClassLoader.addUrls();
                        ChClassLoader.newInstanceIfNewJar();
                        Class c=MAP.get(cn);
                        /*
                          if (c==null && takeBootclassloaderIfThisClassCanBeLoaded!=null)
                          try {
                          Class.forName(takeBootclassloaderIfThisClassCanBeLoaded);
                          c=Class.forName(cn);
                          } catch(Exception e) {putln("Caught in AbstractProxy takeBootclassloaderIfThisClassCanBeLoaded ", e);}
                        */
                        if (c==null) c=Insecure.findClassByName(cn);
                        MAP.put(cn,c);
                        try {
                            if (c!=null) po=c.newInstance();
                        } catch(Exception er){
                            stckTrc(er);
                        }
                    } catch(Exception e){
                        stckTrc(e);
                    }
                    if (po==null) {
                        putln("Error in AbstractProxy.proxyObject "+getClass());
                        if (isMyComputer()) ChClassLoader.log().showInFrame("AbstractProxy: ChClassLoader",0);
                    }
                }
                if (po!=null) mapProxy.put(po,new java.lang.ref.WeakReference(this));
                else alreadyFailed.add(cn);
            }
        }
        return po;
    }
    /* <<< proxyObject <<< */
    /* ---------------------------------------- */
    /* >>> Various Methods >>> */
    public float getScore() {
        final Object o=proxyObject();
        return isInstncOf(HasScore.class,o) ? ((HasScore)o).getScore() : Float.NaN;
    }
    public Container getControlPanel() {
        final Object o=proxyObject();
        return o instanceof HasControlPanel ? ((HasControlPanel)o).getControlPanel() : null;
    }
    public Container getSharedControlPanel() {
        final Object o=proxyObject();
        return o instanceof HasSharedControlPanel ? ((HasSharedControlPanel)o).getSharedControlPanel() : null;
    }
    public PrgParas getPrgParas() {
        final Object o=proxyObject();
        return isInstncOf(HasPrgParas.class,o) ? ((HasPrgParas)o).getPrgParas() : null;
    }
    private boolean _disposed;
    public void dispose() {
        if (!_disposed) {
            _disposed=true;
            final Object o=proxyObject();
            if (isInstncOf(Disposable.class,o)) ((Disposable)proxyObject()).dispose();
        }
    }
    public void interpret(String p1, String p2, long options) {
        final Object o=proxyObject();
        if (isInstncOf(CommandInterpreter.class,o))  ((CommandInterpreter)o).interpret(p1,p2,options);
    }
    /* <<<  Various Methods <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */

    // http://jakarta.apache.org/site/downloads/downloads_commons-collections.cgi
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Static  Utils >>> */
    public final static short COMMONS_COLLECTIONS=1, JFREECHART=2, JCOMMON=3, LOG4J=4, EDTFTPJ=5, TAR=6, ACTIVATION=7, BIOJAVA=9,
        JOGL=10, GLUGEN=11, BIOJAVA_3D=12;
    /**
       In the method getRequiredJars() frequently used jar-libraries should be refered to by a constant.
       For example jarFile(COMMONS_COLLECTIONS) returns "commons-collections-3.2.jar.pack.gz "
    */
    public static String jarFile(short id) {
        if (id==JOGL || id==GLUGEN) {
            final boolean isLinux= isSystProprty(IS_LINUX), isMac= isSystProprty(IS_MAC), isWin= isSystProprty(IS_WINDOWS);
            final String
                OS_ARCH=systProprty(SYSP_OS_ARCH),
                OS_NAME=systProprty(SYSP_OS_NAME),
                osSpecificFile=
                isWin && (OS_ARCH=="amd64"||OS_ARCH=="x86_64")  ? "windows-amd64" :
                isWin  ? "windows-i586" :
                OS_NAME=="sunos" && OS_ARCH=="sparc" ? "solaris-sparc" :
                OS_NAME=="sunos" && OS_ARCH=="sparcv9" ? "solaris-sparcv9" :
                OS_NAME=="sunos" && OS_ARCH=="x86" ? "solaris-i586" :
                OS_NAME=="sunos" && (OS_ARCH=="amd64"||OS_ARCH=="x86_64") ? "solaris-amd64" :
                isLinux && (OS_ARCH=="amd64"||OS_ARCH=="x86_64") ? "linux-amd64" :
                isLinux && isSystProprty(IS_86)  ? "linux-i586" :
                isMac && OS_ARCH=="ppc" ? "macosx-ppc" :
                isMac ? "macosx-universal" : "";
            return
                id==JOGL ? "http://download.java.net/media/jogl/builds/archive/jsr-231-webstart-current/jogl-natives-"+osSpecificFile+".jar "+
                "http://download.java.net/media/jogl/builds/archive/jsr-231-webstart-current/jogl.jar.pack.gz " :
                id==GLUGEN ?     "http://download.java.net/media/gluegen/webstart/gluegen-rt-natives-"+osSpecificFile+".jar "+
                "http://download.java.net/media/gluegen/webstart/gluegen-rt.jar " :
                null;
        }
        return
            id==COMMONS_COLLECTIONS ? "commons-collections-3.2.jar.pack.gz " :
            id==JFREECHART ? "jfreechart-0.9.18.jar.pack.gz " :
            id==JCOMMON ? "jcommon-0.9.3.jar.pack.gz " :
            id==LOG4J ? "log4j-1.2.8.jar.pack.gz " :
            id==EDTFTPJ ? "edtftpj-1.5.2.jar.pack.gz " :
            id==TAR ? "tar.jar.gz " :
            id==ACTIVATION ? "activation.jar " :
            id==BIOJAVA ? "http://www.biojava.org/download/bj16/bin/biojava.jar " :
            id==BIOJAVA_3D ? "Structure_biojava.jar.pack.gz" :
            null;
    }
    public static <T extends Object> T getProxy(Object o, Class<T> clas) {
        return o!=null ? (T)deref(mapProxy.get(o), clas) : null;
    }
    public static void disposeProxy(Object o) {
        final Disposable d=getProxy(o,Disposable.class);
        if (d!=null) d.dispose();
    }
    /* <<< Static Utils <<< */
    /* ---------------------------------------- */
    /* >>> Static settings >>> */
    public static void setSystemClassLoader(boolean b) { loadDirectly=b;}
    private static boolean loadDirectly;

}

