package charite.christo;
import charite.christo.interfaces.*;
import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import java.lang.ref.*;
import java.util.zip.*;

import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP
Some procedures are computationally expensive such as BLAST searches or 3D superpositions and results are stored on HD
<br><b>Implementation</b>: Several results are stored within one
file.
*/
public class CacheResult implements ChRunnable {
    private Reference<Archive> archive_;
    private final File file,fileHeader;
    private Map<String,ByteArray> mapStringBytesGet=new HashMap();
    private CacheResult(Class cacheClass,String section) {
        file=getFile(cacheClass,section);
        fileHeader=file!=null ? new File(file.getAbsolutePath()+".entries") : null;
    }
    private Archive archive() {
        Archive archive=archive_!=null?archive_.get() : null;
        if (archive==null) {
            archive_=new SoftReference(archive=new Archive(file,fileHeader));
        }
        return archive;
    }
   @Override protected  void finalize() throws Throwable {
        logFinalize(ANSI_RED+ANSI_FG_WHITE+"CacheResult","");
        final Archive archive=archive_!=null?archive_.get() : null;
        if (archive!=null) archive.dispose();
        archive_=null;
        super.finalize();
    }

    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Save >>> */
    public static void save() {
        debug(ANSI_GREEN+"CacheResult.save"+ANSI_RESET+"\n");
        for (Map<String,Reference<CacheResult>> mmm : toArry(MAP.values(),Map.class)) {
            for(Object ref : valueArry(mmm)) {
                final CacheResult c=deref(ref,CacheResult.class);
                final Archive archive=c!=null && c.archive_!=null ? c.archive_.get() : null;
                if (archive!=null) {
                    archive.dispose();
                    archive.close();
                }
            }
        }
    }
    public static void saveAndClear() { save(); MAP.clear(); }
    static { addShutdownHook1( ChDelay.newThread(new CacheResult(null,null),"EXIT",null)); }
    private static boolean alreadyExited;
    public Object run(String id, Object arg) {
        if (id=="EXIT") {
            if (!alreadyExited) {
            alreadyExited=true;
            save();
            }
        }
        return null;
    }
    private void save(String key, ByteArray ba) {
        synchronized(CacheResult.class) {
            final Archive archive=archive();
            archive.write(key,ba);
        }
    }
    /* <<< Save <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */

    private static Map<String,Reference<CacheResult>> map(Class c) {
        Map<String,Reference<CacheResult>>  m=MAP.get(c);
        if (m==null) MAP.put(c,m=new TreeMap());
        return m;
    }
    private final static Map<Class,Map<String,Reference<CacheResult>>> MAP=new HashMap();
    private static File dir;
    public static synchronized File dir() {
        if (dir==null) mkdrsErr(dir=file("~/@/cached/"),"In this directory computational results are cached.");
        return dir;
    }

    private final static Set<String> vMkDir=new HashSet();
    private final static StringBuffer file_sb=new StringBuffer(333);
    protected static File getFile(Class clas,String section) {
        if (clas==null) return null;
        final String cn=clas.getName();
        file_sb.setLength(0);
        file_sb.append(dir()).append('/').append(cn.substring(1+cn.lastIndexOf('.'))).append('/').append(section);
        final File f= file(file_sb.toString());
        final String parent=f.getParent();
        if (!vMkDir.contains(parent)) {
            mkdrsErr(file(parent));
            vMkDir.add(parent);
        }
        return f;
    }
    private static CacheResult getCacheResult(Class clas,String section) {
        if (clas==null||section==null) return null;
        final Map<String,Reference<CacheResult>> m=map(clas);
        if (m==null || section==null) return null;
        CacheResult prop=getV(section,m,CacheResult.class);
        if (prop==null) {
            final File f=CacheResult.getFile(clas,section);
            if (sze(f)>0) {
                m.put(section,new SoftReference(prop=new CacheResult(clas,section)));
            }
        }
        return prop;
    }

    public static void deleteCacheResult(Class clas,String section) {
        if (clas==null||section==null) return;
        final Map<String,Reference<CacheResult>> m=map(clas);
        if (m!=null) {
            final CacheResult c=getV(section,m,CacheResult.class);
            m.remove(section);
            if (c!=null) {
                Reference<Archive> refA=c.archive_;
                Archive a=refA!=null ? refA.get() : null;
                if (a!=null) a.deleteBecauseError();
            }
        }

    }

    /** Get the cached text */

    public static ByteArray getValue(Class clas, String section, String key, ByteArray buffer) {
        return (ByteArray)getCacheResult(clas,section,key, buffer, BYTES);

    }
    public static InputStream getValue(Class clas, String section, String key) {
        return (InputStream)getCacheResult(clas,section,key, (ByteArray)null, STREAM);

    }

    private final static int BYTES=3, STREAM=4;
    private synchronized static Object getCacheResult(Class clas, String section, String key, ByteArray buffer, int mode) {
        if (key==null || section==null) return null;
        final CacheResult prop=getCacheResult(clas, section);
        if (prop==null) return null;
        ByteArray ba=prop.mapStringBytesGet.get(key);
        if (ba==null) ba=prop.archive().get(key,buffer);
        if (ba==null) return null;
        try {
            return mode==BYTES ? ba : new ByteArrayInputStream(ba.getBytes(),0,ba.getEnd());
        } catch(Exception ex){ putln("caught in CacheResult#getValue ",ex);}
        return null;
    }
    public static synchronized void putValue(Class clas, String section, String key, ByteArray value,boolean overwrite) {
        if (!storing || clas==null || section==null || key==null) return;
        CacheResult prop=getCacheResult(clas,section);
        Map<String,Reference<CacheResult>> mmm=map(clas);
        if (prop==null) {
            prop=new CacheResult(clas,section);
            mmm.put(section,new SoftReference(prop));
        }
        prop.mapStringBytesGet.put(key,value);
        if (overwrite || !prop.archive().contains(key)) prop.save(key,value);
    }

    private static boolean isEnabled=true;
    public static boolean isEnabled() { return isEnabled;}
    public static void setEnabled(boolean b) { isEnabled=b; }
    private static boolean storing=true;
    //public static void setDoStore(boolean b) { storing=true;}

    /**
       When was the Result that had been retrieved recently modified?
       This is not thread save !
    */
    public static synchronized long getLastModified(Class clas, String section, String key) {
        if (key==null || section==null) return 0;
        final CacheResult prop=getCacheResult(clas, section);
        return prop!=null ? prop.archive().getLastModified(key) : 0;
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Static methods  >>> */
    public static String keyForUngappedSeqences(byte ss[][]) {
        if (ss==null) return null;
        final StringBuffer sb=new StringBuffer(ss.length*30);
        for(byte seq[]:ss) {
            for(int i=0;i<8 && i<seq.length; i++) sb.append((char)seq[i]);
            sb.append(hashCd(seq,0,MAX_INT,true)).append('_');
        }
        return sb.toString();
    }
    private static String keyForUngappedSeqences(String ss[]) {
        if (ss==null) return null;
        final StringBuffer sb=new StringBuffer(99);
        for(String seq:ss) {
            seq=seq.toUpperCase();
            final String start=sze(seq)>=8 ? seq.substring(0,8) :seq;
            sb.append(start).append(seq.hashCode()).append("_");
        }
        return sb.toString();
    }
    public static String sectionForSequence(String s) {
        if (s==null) return null;
        final int l=sze(s);
        return l<10 ? s : s.substring(0,10)+"_"+s.substring(maxi(l-10,10),l);
    }

   public static StringBuffer sectionForSequence(byte bb[], StringBuffer sb) {
        if (bb==null) return null;
        if (sb==null) sb=new StringBuffer();
        for(int i=0; i<bb.length &&  i<16; i++) sb.append((char)bb[i]);
        sb.append('-');
        for(int i=maxi(bb.length-16,16); i<bb.length; i++) sb.append((char)bb[i]);
        sb.append('-').append(hashCd(bb,0,99999,true));
        return sb;
    }
    public static long keyForCalphaAtoms(float xyz[]) {
        if (xyz==null) return 0;
        long sum=0;
        final int n=xyz.length/3;
        for(int i=0,j=0,count=0;i<n;i++,j+=3) {
            sum+=(long)( (xyz[j]+xyz[j+1]*2+xyz[j+2]*3)*100)*count++;
        }
        return sum;
    }

}
