package charite.christo;
import java.net.URL;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
   @author Christoph Gille
*/
public class ChClassLoader extends java.net.URLClassLoader {
    private java.lang.ref.Reference<Map<String,Class>> MAP_CLASSES;
    private Map<String,Class> mapClasses() {
        Map m=MAP_CLASSES!=null ? MAP_CLASSES.get() : null;
        if (m==null) MAP_CLASSES=new java.lang.ref.SoftReference(m=new HashMap());
        return m;
    }
    public ChClassLoader(URL[] urls,ClassLoader parent) { super(urls,parent); }
    private static AsciiView log_;
    public static AsciiView log(String... oo) {
        if (isGraphicsAvailable()) {
            if (log_==null) log_=new AsciiView(9999);
            log_.appendSS(oo);
        }
        return log_;
    }

    @Override public String findLibrary(String libname) {
        if (!Insecure.CLASSLOADING_ALLOWED) { assert false; return null; }
        for(boolean lib : new boolean[]{true,false}) {
            final java.io.File f=file(dirJars(),System.mapLibraryName(libname));
            if (f.exists()) {
                log(ANSI_CYAN+"ChClassLoader.findLibrary("+ANSI_WHITE,libname,") ==> "+f+"\n");
                return f.toString();
            }
        }
        return super.findLibrary(libname);
    }
    private final static UniqueList<String> vNO_PARENT_DELEGATION=new UniqueList(String.class);
    private static String logLastClasspath;
    public static Class staticFindClass(final String cn0) {
        if (!Insecure.CLASSLOADING_ALLOWED) { assert false; return null; }
        final String cn=maybeChangeClassName(cn0);
        newInstanceIfNewJar();
        vNO_PARENT_DELEGATION.add(cn);
        final ChClassLoader ucl=getInstance();
        Class c=ucl.mapClasses().get(cn);
        if (c!=null) return c;
         final String logClasspath="["+joinAsStrg(instance.getURLs(),",")+"]";
         if (!logClasspath.equals(logLastClasspath)) log(ANSI_GREEN+"CLASSPATH="+ANSI_RESET,logLastClasspath=logClasspath,"\n");
        try {
            c=ucl.loadClass(cn);
            log(ANSI_GREEN+"staticFindClass "+ANSI_RESET,toStrg(c),"\n");
            return c;
        } catch(Throwable e){ log(ANSI_RED+"caught in ChClassLoader .myFindClass "+ANSI_RESET+cn+" :"+e+"\nCLASSPATH:"+classPathAtStart()+"\n"); }
        return null;
    }
    // ----------------------------------------
    private static String maybeChangeClassName(String cn0) {
        return cn0!=null && cn0.startsWith("com.sun.tools.javac.resources.compiler_") ? "com.sun.tools.javac.resources.compiler" : cn0;
    }
    // ----------------------------------------
    private final static long durationScroll[]=new long[ARRAY_DURATION];
    @Override public synchronized Class findClass(String cn0) throws ClassNotFoundException{
        if (!Insecure.CLASSLOADING_ALLOWED) { assert false; return null; }
        //final boolean debug=cn0.startsWith("com.inet.jortho.DictionaryFactory");
        final String cn=maybeChangeClassName(cn0);
        Class c=mapClasses().get(cn);
        if (c!=null) return c;
        if (cn.startsWith("org.openscience.jmol.Properties")) throw(new ClassNotFoundException());
        log(cn.indexOf('$')<0 ? ANSI_CYAN : ANSI_RESET,"ChClassLoader#findClass "+ANSI_RESET,cn);
        try {
            try {
                //if (log_!=null && log_.shouldAppend()) ChDelay.scrollToEnd(log_,999,durationScroll);
                c=super.findClass(cn);
                if (c!=null) mapClasses().put(cn,c);
                log(ANSI_GREEN+" OK"+ANSI_RESET+"\n");
                return c;
            } catch(ClassNotFoundException ex) {
                /*
                  in Java 1.6 : findClass java.lang.ClassNotFoundException: charite.christo.strap.Jmol2Strap
                */
                log(ANSI_RED+" findClass "+ANSI_RESET+ex+"\n");
                //throw(new ClassNotFoundException());
                throw(ex);
            }
        } catch(NoClassDefFoundError er) {
            log(ANSI_RED+"findClass "+ANSI_RESET+er+"\n");
            throw(new ClassNotFoundException());
        }
    }
    //
    static {
        vNO_PARENT_DELEGATION.add(_CCP+"ChJmol");
    }
        /*
          Oherwise one gets:
          java.lang.IllegalAccessError: tried to access class charite.christo.strap.Jmol2Strap$1 from class charite.christo.strap.Jmol2Strap
        */
    @Override public synchronized Class loadClass(String cn) throws ClassNotFoundException {
        if (!Insecure.CLASSLOADING_ALLOWED) { assert false; return null; }
        final char c0=cn.charAt(0);
        boolean findClass=
            !(c0=='c'&&cn.startsWith("charite.")   || c0=='j'&&cn.startsWith("java.")       ||
              c0=='j'&&cn.startsWith("javax.")     || c0=='c'&&cn.startsWith("com.sun.")    ||
              c0=='a'&&cn.startsWith("apple.awt.") || c0=='s'&&cn.startsWith("sun.")        ||
              c0=='o'&&cn.startsWith("org.xml.sax")|| c0=='o'&&cn.startsWith("org.w3c.dom") ||
              c0=='o'&&cn.startsWith("org.ietf")   || c0=='o'&&cn.startsWith("org.omg"));
        //&& !cn.startsWith("com.sun.opengl.") ;
        if (!findClass && cn.indexOf("PROXY")<0)
            for(String s : vNO_PARENT_DELEGATION.asArray())
                if (cn.startsWith(s)) {
                    findClass=true;
                    break;
                }
        //if (cn.startsWith("astex.")) findClass=true;
        if (cn.startsWith(_CCS) && cn.indexOf("Biojava")>=0) findClass=true;
        log(ANSI_RESET,findClass+" loadClass  ",cn ,findClass ? "==>  findClass" : " ==> super.loadClass","\n");
        Class cRet=null;
        try {
            cRet=findClass ? findClass(cn): super.loadClass(cn);
        } catch(Throwable t){ if(!findClass) cRet=findClass(cn); }
        return cRet;
    }
    private static ChClassLoader instance;
    public static void newInstance() {
        if (!Insecure.CLASSLOADING_ALLOWED) return;
        if (instance!=null) invokeMthd("close",instance);
        instance=null;
        log(ANSI_GREEN+"newInstance"+ANSI_RESET+"\n");
    }
    private static long lastFileHash;
    public static void newInstanceIfNewJar() {
        if (!Insecure.CLASSLOADING_ALLOWED) return;
        ChZip.jarFilesJF();
        final int modi=ChZip.jarFiles_modificationCount();
        if (lastFileHash!=modi) newInstance();
        lastFileHash=modi;
    }
    public static ChClassLoader newInstance(URL[] url,ClassLoader parent) {
        if (!Insecure.CLASSLOADING_ALLOWED) return null;
        log(ANSI_GREEN+"newInstance "+ANSI_RESET+"[",joinAsStrg(url,", "),"]\n");
        return new ChClassLoader(url,parent);
    }
    public static ChClassLoader getInstance() {
        if (instance==null) {
            instance=new ChClassLoader(NO_URL, CLASSLOADER);
            log("new ChClassLoader done\n");
            instance.addURL(url(dirPlugins()));
            /* The following line is important since Java 1.6 */
            // final URL uJar=url(thisJarFile());
            final URL uJar=urlThisJarFile();
            if (uJar!=null) instance.addURL(uJar);
            addUrls();
        }
        return instance;
    }
    // ----------------------------------------
    public static void addUrls() {
        if (!Insecure.CLASSLOADING_ALLOWED) return;
        if (instance==null) { getInstance(); return;}
        final ChClassLoader cl=instance;
        final URL uu[]=cl.getURLs();
        final Set setUU=new HashSet();
        for(URL u:uu) setUU.add(u.toString());
        //cl.addURLs(setUU,url(dirSettings()));
        cl.addURLs(setUU,classPathAtStart());
        cl.addURLs(setUU,ChZip.jarFilesURL());
        if (toolsDotJar()!=null)
        cl.addURLs(setUU,url(toolsDotJar()));
    }
    private void addURLs(Collection<URL> oldUrls, URL... newUrls) {
        if (!Insecure.CLASSLOADING_ALLOWED) return;
        for(Object url: newUrls) {
            if (url==null) continue;
            String sUrl=url.toString();
            if (!oldUrls.contains(sUrl))
                try {
                    addURL(url instanceof URL ? (URL)url : new URL(sUrl));
                } catch(Throwable e){}
        }
    }
    // ----------------------------------------
    static URL resource(String s) {
        if (!Insecure.CLASSLOADING_ALLOWED) return null;
        return instance!=null ? instance.getResource(s) : null;
    }
    static Class name2c(String s){
        if (!Insecure.CLASSLOADING_ALLOWED) return null;
        if (instance!=null)
            try{try{ return instance.loadClass(s);} catch(ClassNotFoundException e){} } catch(NoClassDefFoundError er){}
        return null;
    }
    public static java.io.InputStream resourceAsStream(String s) { return instance!=null ? instance.getResourceAsStream(s) : null;}
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Security  >>> */
    static {
        if (Insecure.CLASSLOADING_ALLOWED) try { setSecurityPolicy(); } catch(Throwable t){}
    }
    private static java.security.Policy policy;
    public static void setSecurityPolicy() {
       if (Insecure.CLASSLOADING_ALLOWED && policy==null) {
            policy=new java.security.Policy() {
                    @Override public java.security.PermissionCollection getPermissions(java.security.CodeSource codesource) {
                        java.security.Permissions perms = new java.security.Permissions();
                        perms.add(new java.security.AllPermission());
                        return perms;
                    }
                    @Override public void refresh(){}
                };
            try {  java.security.Policy.setPolicy(policy); } catch(Throwable e) {
              putln(RED_CAUGHT_IN+"ChClassLoader ",e);
            }
        }
    }
}
