package charite.christo;
import static charite.christo.ChUtils.*;
import charite.christo.interfaces.*;
import java.util.*;
import java.awt.*;
import javax.swing.*;
import javax.swing.event.ListDataEvent;
import javax.swing.table.TableModel;
import static java.lang.System.currentTimeMillis;
import static java.lang.Boolean.TRUE;
import static java.lang.Boolean.FALSE;
import static charite.christo.ChConstants.*;
/**
This utility class manages threads.
It fullfills same task as SwingWorker in JRE6
Methods in  Swing are not thread safe.  Exceptions: #repaint() #revalidate()
   @author Christoph Gille
*/
public class ChDelay extends Thread implements ChRunnable{
    {  assertInstAllowed(); }

    public final static byte
        ENABLED=1, REPAINT=2, SCROLL_DOWN=3,
        CONTENTS_CHANGED=4,
        SET_PREF_SIZE=5,
        RESTORE_BACKGROUND=8,
        PACK=10, SET_JLIST_II=11, SET_TEXT=12,
        FG_COLOR=14, BG_COLOR=15,
        VALIDATE=16, UPDATEUI=17, DOCLICK=18, ADDTAB=19, REVALIDATE=20,RUN=21,
        RUNNABLE_ARG=28, ADD_COMPONENT=29,
        SCROLL_UP=32, REMOVE_FROM_P=33, MAX_VARIABLE=34,
        SPLIT_DIVIDER_LOCATION=35;
    public final static int DONT_SET_DURATION=4;
    static { assert EDT!=DONT_SET_DURATION && PAUSE!=DONT_SET_DURATION;}
     /* ---------------------------------------- */
    /* >>> The Job instance >>> */
    public final long LAST[];
    private final Object object, data;
    private final byte actionID;
    private long options;
    private long  nextInvoke;
    private static String[] names;
    { if (names==null) names=getFinalStaticFields(ChDelay.class, MAX_VARIABLE); }
    public ChDelay(long options, int id, Object o, Object data) { this(options,id,o,0,data,(long[])null);}
    public ChDelay(long options, int id, Object o, long delay, Object data, long[] duration) {
        this.actionID=(byte)id;
        this.object=o;
        this.nextInvoke=currentTimeMillis()+delay;
        this.data=data;
        this.LAST=duration;
        this.options=options;
        setName("ChDelay");
    }
    @Override public String toString() {
        final Object runID=actionID!=RUNNABLE_ARG ? "" : data instanceof String ? data : get(0,data);
        return shortClassName(object)+"."+get(actionID, ChDelay.names)+" "+runID;
    }
    private static String shortClassName(Object o0) {
        final Object o=o0 instanceof ChDelay ? ((ChDelay)o0).object : o0;
        if (o==null) return "";
        final Class c= o instanceof Class ? (Class)o : o.getClass();
        final String cn=c.getName();
        return  cn.substring(cn.lastIndexOf('.')+1);
    }
    /* <<< The Job instance <<< */
    /* ---------------------------------------- */
    /* >>> Threading >>> */
    //static int countDebug;
    @Override public void run() {
        final long time=currentTimeMillis();
        try {
            ChDelay.runJob(this.options, this.actionID, this.object, this.data);
        } catch(Throwable ex) {
            putln("ChDelay actionID="+actionID+"  options="+options+(data instanceof String ? "data="+data : ""));
            stckTrcErr(ex);
        }
        final long[] d=this.LAST;
        if ((options&DONT_SET_DURATION)==0 && d!=null && d.length>0) d[DURATION]=(d[WHEN]=currentTimeMillis())-time;
        this.stckTrc=null;
    }
    private final static UniqueList<ChDelay> vQueue=new UniqueList(ChDelay.class);
    private static ChDelay inst;
    private static ChDelay instance() {if (inst==null) inst=new ChDelay(0,0,null,0,null,null); return inst;}
    private static boolean isRunning;
    private static void runJob(long options, int actionID,final Object object0,Object data) {
        final Object object=deref(object0), d0, d1, d2;
        if (object==null) return;
        if (data instanceof Object[]) {
            final Object dd[]= (Object[])data;
            d0=dd.length>0?dd[0]:null;
            d1=dd.length>1?dd[1]:null;
        }  else { d0=d1=null; }
        switch(actionID) {
        case ADD_COMPONENT:
            Component o=(Component)object;
            final Container c=(Container)d0;
            final Object constraint=d1;
            if (c==null || o==null) break;
            if (constraint!=null) c.add(o,constraint); else c.add(o);
            break;
        case RUN:((Runnable)object).run(); break;
        case RUNNABLE_ARG:
            ((ChRunnable)object).run((String) (d0!=null?d0: data), d1);
            break;
        case PACK: if (object instanceof Window) ((Window)object).pack(); break;
        }
        if (object instanceof Component) runJob(actionID,(Component)object,data);
        assert !(object instanceof Object[]);
        if (object instanceof Object[]) {
            for(Object o : ((Object[])object)) {
                o=deref(o);
                if (o instanceof Component) runJob(actionID,(Component)o,data);
            }
        }
    }
    private static void revali(JComponent c) {
        if (c!=null) {
            c.revalidate();
            pcp(KEY_NEEDS_REVALIDATE,null,c);
        }
    }

    private static void runJob(final int actionID, final Component c,final Object data) {
        final JComponent jc=c instanceof JComponent ? (JComponent)c : null;
        switch(actionID) {
        case CONTENTS_CHANGED:
            updateListTableCombo(jc);
            break;
        case REMOVE_FROM_P:
            ChUtils.rmFromParent(c);
            break;
        case SET_PREF_SIZE:
            final Dimension d=data instanceof Dimension ? (Dimension)data : null;
            if (d!=null) jc.setPreferredSize(d);
            break;
        case ENABLED:
            setEnabld(c,data==TRUE?true : data==FALSE?false: c.isEnabled());
            if (gcp(KEY_PREF_SIZE_IF_DISABLED,c)!=null) revali(jc);
            break;
        case REVALIDATE: revali(jc); c.repaint(); break;
        case REPAINT:
            c.repaint();
            break;
        case RESTORE_BACKGROUND:
            Color col=(Color)gcp(KEY_BACKGROUND,c);
            if (col!=null) c.setBackground(col);
            break;
        case UPDATEUI:
            jc.updateUI();
            break;
        case VALIDATE:  c.validate(); break;
        case DOCLICK: ((AbstractButton)c).doClick(); break;
        case SCROLL_DOWN:
        case SCROLL_UP:
            final JScrollPane sp=(JScrollPane)findParent(jc,JScrollPane.class);
            final JScrollBar sb=sp!=null ? sp.getVerticalScrollBar() : null;
            if (sb!=null) sb.setValue(actionID==SCROLL_DOWN ? sb.getMaximum() : 0);
             break;
        case ADDTAB:
            final JTabbedPane tp=(JTabbedPane)data;
            if (tp!=null) {
                final String tab=TabItemTipIcon.tab(c);
                adTab(tab,c, tp);
                if (tp.getSelectedIndex()==tp.getTabCount()-2 || tp.getSelectedIndex()==0) tp.setSelectedComponent(c);
                pcp(JTabbedPane.class,null,c);
            }
            break;
        case SPLIT_DIVIDER_LOCATION:
            final double val=atof(data);
            if (c instanceof JSplitPane) ((JSplitPane)c).setDividerLocation(val);
            break;
        case SET_TEXT:
            final String text=toStrg(data);
            invokeMthd("setText",c,new Object[]{text},0L);
            break;
        case SET_JLIST_II:
            final int ii[]=(int[])data;
            ((ChJList)c).setSelII(ii);
            break;
        case FG_COLOR: c.setForeground((Color)data);  break;
        case BG_COLOR: if (data!=null) c.setBackground((Color)data); else setDefaultBG(c); break;
        }
    }
    /* <<< Threading <<< */
    /* ---------------------------------------- */
    /* >>> Public API  >>> */
    public static void afterMS(int options, int id, Object object, long delay,Object data) {
        if (object==null) return;
        if (delay==0 && isEDT()) {
            runJob(options, id, object, data);
        } else if (id==RUN && object instanceof ChDelay) {
            final ChDelay j=(ChDelay)object;
            j.options=options;
            j.nextInvoke=Long.MAX_VALUE;
            enqueue(j);
            final long time=currentTimeMillis();
            final long[] d=j.LAST;
            if ((options&PAUSE)!=0 && d!=null && d.length>0 && time-d[WHEN]>delay) {
                if ((options&EDT)!=0) inEDT(j); else j.run();
                vQueue.remove(j);
            } else {
                j.nextInvoke=time+delay;
            }
        } else  {
            afterMS(options, id, object, delay, data, null);
        }
    }
    public static void afterMS(int options, int id, final Object object, long delay, Object data, long duration[]) {
        if (object==null)  return;
        final ChDelay j=getJob(object,id, data);
        final long time=currentTimeMillis();
        if (duration!=null && duration.length>0 && time-duration[WHEN]>delay) {
            if (j!=null) j.nextInvoke=Long.MAX_VALUE;
            if (isEDT()|| (options&EDT)==0)  runJob(options, id, object, data);  else inEDT(new ChDelay(options, id,object,delay,data,duration));
            if (j!=null) vQueue.remove(j);
        } else {
            if (j!=null  && !(id==RUNNABLE_ARG && data==j.data) ) {
                if (j.nextInvoke>time+delay) j.nextInvoke=time+delay;
            } else {
                enqueue(new ChDelay(options, id,object,delay,data,duration));
            }
        }
    }
    private static void enqueue(ChDelay j) {
        if (j==null) return;
        if (!isRunning) {
            isRunning=true;
            ChDelay.newThread(instance(),"LOOP",null).start();
        }
        if (!vQueue.contains(j)) vQueue.add(j);
    }
    public static void addComponentToContainer(Component o,Container c,Object constraint,long delay) { afterMS(EDT,ADD_COMPONENT, o, delay,new Object[]{c,constraint},(long[])null); }
    public static void runAfterMS(int options,Runnable r, long delay) { afterMS(options,RUN,r,delay, (Object)null);}
    public static void runAfterMS(int options,ChRunnable c, String id, long delay) { afterMS(options, RUNNABLE_ARG,c,delay,id,(long[])null); }
    public static void set(JProgressBar b, int value, int min, int max, String txt) {
        if (b!=null) invokeAndWait(instance(),"PROGRESS_BAR",new Object[]{b,intObjct(value),intObjct(min),intObjct(max),txt});
    }
    public static void revalidate(Component c, long delay) { revalidate(c,delay, (long[])null); }
    public static void revalidate(Component c, long delay, long[] duration) {if (c instanceof JComponent) afterMS(DONT_SET_DURATION,REVALIDATE,c,delay,null, duration);}
    public static void repaint(Object c,long delay) { afterMS(DONT_SET_DURATION,REPAINT,c,delay,null,(long[])null);}
    public static void repaint(Object c,long delay, long duration[]) { afterMS(0,REPAINT,c,delay,null, duration);}
    public static void updateUI(Object c,long delay, long duration[]) { afterMS(EDT,UPDATEUI,c,delay,null,duration); }
    public static void scrollToEnd(Object c,long delay, long[] duration)  { afterMS(EDT,SCROLL_DOWN,c,delay,null,duration); }
    public static void setText(Object c,CharSequence text,long delay) { afterMS(EDT,SET_TEXT,c,delay,text,(long[])null);}
    public static void setSelII(ChJList c, int ii[],long delay) { afterMS(EDT,SET_JLIST_II,c,delay,ii,(long[])null);}
    public static void bg(Object c,Color col,long delay) { afterMS(EDT,BG_COLOR,c,delay,col,(long[])null);}
    public static void fg(Object c,Color col,long delay) { afterMS(EDT,FG_COLOR,c,delay,col,(long[])null);}

public static void setDividerLocation(float val, JSplitPane c, long delay) { afterMS(EDT,SPLIT_DIVIDER_LOCATION,c,delay, new Float(val),(long[])null);}

    public static void addTab(JTabbedPane tp,JComponent c,long delay) { afterMS(EDT,ADDTAB,c,delay,tp,(long[])null);}

    public static void contentsChanged(JComponent c, long delay) {  afterMS(EDT,CONTENTS_CHANGED,c,delay,null,(long[])null);}

    public static void setEnabled(Component c, boolean b,long delay) { if (c!=null) afterMS(EDT,ENABLED,c,delay, b?TRUE:FALSE,(long[])null);}
    public static void removeFromParent(JComponent c,long delay) { if (c!=null) afterMS(EDT,REMOVE_FROM_P,c,delay, null,(long[])null);}

    public static void parentWindowToFront(Component c,int afterMs){  newThread(instance(),"TO_FRONT",new Object[]{intObjct(afterMs),c}).start(); }
    public Object run(String id,Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[])arg : null;
        if (id=="PROGRESS_BAR") {
            final JProgressBar b=(JProgressBar)argv[0];
            if (b!=null) {
            final int value=atoi(argv[1]);
            final int min=atoi(argv[2]);
            final int max=atoi(argv[3]);
            final String txt=(String)argv[4];
            b.setStringPainted(txt!=null);
            if (min<=value && value<=max && min <max) {
                b.setIndeterminate(false);
                b.setValue(value);
                b.setMinimum(min);
                b.setMaximum(max);
                if (txt!=null) b.setString(txt);
            } else b.setIndeterminate(true);
            }
        }
        if (id=="TO_FRONT") {
            try {
                sleep((Long)(argv[0]));
                final Window d=argv[1] instanceof Window ? (Window)argv[1] : parentWndw(argv[1]);
                if (d!=null && d.isVisible()) {d.toFront(); d.setVisible(true);}
            } catch(Exception e){}
        }
        if (id=="LOOP") {
            while(true) {
                final long ct=currentTimeMillis();
                for(int i=vQueue.size(); --i>=0;) {
                    ChDelay job=null;
                    try {
                        job=vQueue.get(i);
                        if (job==null||job.nextInvoke>ct) continue;
                        if ( (job.options&EDT)!=0) SwingUtilities.invokeAndWait(job); else job.run();
                    } catch(Throwable ex){
                        stckTrc(ex);
                    }
                    vQueue.remove(job);
                }
                ChUtils.sleep(55);
            }
        }
        return null;
    }
    // ++++++++++++++++++++++++++++++++++++++++
    public static ChDelay newThread(ChRunnable r,String id,Object arg) { return newThread(r,id,arg,(long[])null);}
    public static ChDelay newThread(ChRunnable r,String id,Object arg, long[] duration) {
        assert !(arg instanceof long[]);
        return new ChDelay(EDT, RUNNABLE_ARG, r,0, arg!=null ? new Object[]{id,arg} : id, duration);
    }
    public static void invokeAndWait(ChRunnable r,String id,Object o) {
        if (!isGraphicsAvailable() || isEDT()) r.run(id,o);
        else try {SwingUtilities.invokeAndWait(newThread(r,id,o)); } catch(Throwable e) { stckTrc(e);}
    }
    public static void invokeLater(ChRunnable r,String id,Object o) {
        if (!isEDT()) try { SwingUtilities.invokeLater(newThread(r,id,o));} catch(Exception e) { stckTrc(e);}
        else r.run(id,o);
    }
    /* <<< Public API <<< */
    /* ---------------------------------------- */
    /* >>> Static Utils >>> */
    private static ChDelay getJob(Object object,int id, Object data) {
        try {
            for(ChDelay j:vQueue.asArray())
                if (object==j.object && (id<1 || id==j.actionID && data==null || data==j.data)) {
                    return j;
                }
        } catch(Exception e) {}
        return null;
    }
    public static boolean contains(Object c) { return getJob(c,0,null)!=null;}
    public static void remove(Object o,int id) {
        while(true) {
            final ChDelay j=getJob(o,id,null);
            if (j==null) return;
            vQueue.remove(j);
        }
    }
    private static ListDataEvent listDataEvent;
    private static void updateListTableCombo(JComponent c) {
        if (c==null) return;
        final ListModel lm=c instanceof JList ? ((JList)c).getModel() : c instanceof JComboBox ? ((JComboBox)c).getModel() : null;
        if (lm instanceof AbstractListModel) {
            final AbstractListModel alm=(AbstractListModel)lm;
            if (listDataEvent==null) listDataEvent=new ListDataEvent(ChDelay.class, ListDataEvent.CONTENTS_CHANGED,0, MAX_INT);
            for(javax.swing.event.ListDataListener l : alm.getListDataListeners()) {
                l.contentsChanged(listDataEvent);
            }
        }
        if (c instanceof JTable) {
            final JTable jt=(JTable)c;
            final javax.swing.table.TableModel tm=jt.getModel();
            if (tm!=null) jt.tableChanged(new javax.swing.event.TableModelEvent(tm));
        }
        c.revalidate();
        pcp(KEY_NEEDS_REVALIDATE,null,c);
        c.repaint();
    }
    private static boolean debug=false;

    String stckTrc;
    public void setDebug(boolean b) { debug=b;}
    @Override public void start() {
        if (debug) this.stckTrc=stackTrceAsStrg(new Exception("ChDelay "+this));
        super.start();
    }
}
