package charite.christo;
import javax.swing.*;
import java.awt.*;
import java.util.List;
import java.util.*;
import java.awt.event.*;
import charite.christo.interfaces.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static java.awt.event.MouseEvent.*;
/**HELP
   This list supports<ul>
<li>WIKI:Drag_and_drop  (Java 1.5 or higher, Watch movie
  <i>STRING:ChConstants#MOVIE_Export_Proteins</i>
  <i>STRING:ChConstants#MOVIE_Drag_to_another_STRAP</i>
  <i>STRING:ChConstants#MOVIE_Drag_protein_to_3D</i>
  <i>STRING:ChConstants#MOVIE_Sequence_Features_in_3D</i>)
<li>Reordering</li>
<li>WIKI:Context_menu (Watch movie <i>STRING:ChConstants#MOVIE_Context_Menu</i>)</li>
</ul>
   <b>Keep window always visible: </b>You may want to  set the window "Floating on top". See  Strap tool-bar==&gt;options.<br><br>
   <b>Frequently used menu-items:</b> of the context-menu can be dragged out of the menu.
   These extracted menu items are found at the bottom and can be used directly.<br><br>
   <b>Sequence groups:</b> If the list is displaying proteins, two buttons at the bottom provide "Sequence group" functionality.
   Exactly the proteins in the list are displayed in the alignment, all others are hidden.<br>
   Hidden proteins can be reintroduced into the alignment by dragging the tree node "Hidden proteins" of the object tree.
   The object tree is usually not visible because the vertical divider is at the very left. It becomes visible by opening the divider.<br><br>
   @author Christoph Gille
*/
public class ChJList extends JList implements IsEnabled, ActionListener, HasModiCount {
    public final static int OPTIONS_FILES=ChJTable.ICON_ROW_HEIGHT|ChJTable.DEFAULT_RENDERER|ChJTable.FILE_POPUP|ChJTable.RTT|ChJTable.DRAG_ENABLED;
    public final static String KEY_NO_HELP_ON_RIGHT_CLICK="CJL$$NH";
    private int selectedValuesMC;
    private long options;
    private boolean isInitialized;
    /* ---------------------------------------- */
    /* >>> Instance >>> */
    public void setData(Object data) {
        if (data instanceof ListModel) setModel((ListModel)data);
        else if (data instanceof List) setModel(new JListModel(list=(List)data));
        else if (data instanceof Object[]) {
            final DefaultListModel m=new DefaultListModel();
            for(Object o : (Object[])data) m.addElement(o);
            setModel(m);
        }
    }
    public long getOptions() { return this.options;}
    public void addOptions(long options) { this.options|=options;}
    public ChJList(Object data, long options) {
        this.options=options;
        setData(data);
        int fixed=3*EX/2;
        if ( (options&ChJTable.EX_ROW_HEIGHT)!=0) fixed=EX+2;
        if ( (options&ChJTable.ICON_ROW_HEIGHT)!=0) fixed=ICON_HEIGHT;
        if ( (options&ChJTable.DEFAULT_RENDERER)!=0) setCellRenderer(new ChRenderer());
        if ( (options&ChJTable.CLASS_RENDERER)!=0) setCellRenderer(new ChRenderer().options(ChRenderer.CLASS_NAMES));
        if ( (options&ChJTable.RTT)!=0) rtt(this);
        if ( (options&ChJTable.FILE_POPUP)!=0) { setProcessEvt(FilePopup.instance(), this); addTooltip(FilePopup.instance(),this);}
        if ( (options&ChJTable.SINGLE_SELECTION)!=0) setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        setFixedCellHeight(fixed);
        if ( (options&ChJTable.DRAG_ENABLED)!=0) setDragEnabled(true);
        setForeground(Color.BLACK);
        scrollByWheel(this);
    }
    @Override public String toString() {
        Object oo[]=getSelectedValues();
        return oo==null ? "" : oo.length==1 ? toStrg(oo[0]) : joinAsStrg(oo,"\t");
    }
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Model >>> */
    private List list;
    public List getList() { return list; }
    public boolean isEnabled(Object o) {
        final Object
            enabled0=gcp(KEY_ENABLED,this),
            enabled=enabled0 instanceof ChCombo ?  ((ChCombo)enabled0).instanceOfSelectedClass(IsEnabled.class) : enabled0;
        final IsEnabled en= isInstncOf(IsEnabled.class,enabled) ?(IsEnabled) enabled: null;
        return en!=null ? en.isEnabled(o) : true;
    }
    public ChJList setSelII(int[] ii) {
        pcp(KEY_IGNORE_SELECTION_EVT,"",this);
        setSelectedIndices(ii);
        pcp(KEY_IGNORE_SELECTION_EVT,null,this);
        selectedValuesMC++;
        return this;
    }
  public ChJList setSelI(int i) {
        pcp(KEY_IGNORE_SELECTION_EVT,"",this);
        setSelectedIndex(i);
        pcp(KEY_IGNORE_SELECTION_EVT,null,this);
        selectedValuesMC++;
        return this;
    }
    public ChJList setSelO(Object o) {
        final ListModel m=getModel();
        if (m!=null) {
            for(int i=m.getSize(); --i>=0;) {
                if (m.getElementAt(i)==o) setSelI(i);
            }
        }
        return this;
    }
    public ChJList setSelOO(Object[] oo) { setSelII(selIndices(oo,getModel())); return this; }
    public int getModificationCount() { return selectedValuesMC;}
    /* <<< Model <<< */
    /* ---------------------------------------- */
    /* >>> Persistence >>> */
    private Class clas;
    private String id;
    public ChJList property(Class clas,String id_str) {
        this.clas=clas; this.id="list_"+id_str;
        int  savedIdx=getPrpty(clas,id,-1);
        if (savedIdx>=0 && getModel().getSize()>savedIdx) setSelectedIndex(savedIdx);
        return this;
    }
    private void storeInProperties() {
        if (clas!=null && id!=null)
            setPrpty(clas,id,ChJList.this.getSelectedIndex());
    }
    /* <<< Persistence <<< */
    /* ---------------------------------------- */
    /* >>> preferredSize >>> */
    @Override public Dimension getPreferredSize() {
        final Dimension d=getPrefSize(this);
        try {
            return d!=null ? d : super.getPreferredSize();
        } catch(Exception e) { return dim(EM,EX);}
    }
    /* <<< preferredSize <<< */
    /* ---------------------------------------- */
    /* >>> Paint >>> */
    {pcp(KEY_NOT_YET_PAINTED,"",this);}
    @Override public void paint(Graphics g) { try {super.paint(g);} catch(Throwable e) {} };
    @Override public void paintComponent(Graphics g) {
        if (!isInitialized){
            isInitialized=true;
            pcp(KEY_NOT_YET_PAINTED,null,this);
            pcp(KEY_IGNORE_SELECTION_EVT,null,this);
            adSelectLisener(this);
        }
        if (paintHooks(this,  g,false)) {
            if (!isEnabled()) g.clearRect(0,0,Short.MAX_VALUE,Short.MAX_VALUE);
            else {
                super.paintComponent(g);
                if (!paintHooks(this,  g,true)) return;
            }
        }
    }
    /* <<<  Paint <<< */
    /* ---------------------------------------- */
    /* >>> Event >>> */
    public ChJList li(ActionListener l) {
        addActLi(l,this);
        actionEventWhenClicked(this,l);
        return this;
    }
    { this.enableEvents(ENABLE_EVT_MASK);}
    @Override public void processEvent(AWTEvent ev) {
        if (processEvt(this,ev)) {
            final int kcode=keyCode(ev), id=ev.getID(), modi=modifiers(ev);
            String cmd=null;
            if (0==(options&ChJTable.CHJTABLE_BACKSPACE_DEACTIVATED) && (kcode==KeyEvent.VK_BACK_SPACE || kcode==KeyEvent.VK_DELETE)) cmd=ACTION_DELETE_OR_BACKSPACE;
            if (kcode==KeyEvent.VK_A && modi==shortCutMask()) selectAllItems(this);
            final boolean isPopup=isPopupTrigger(ev);
            final Point point=point(ev);
            if (point!=null) {
                final int idx=locationToIndex(point);
                if ( 0!=(options&ChJTable.FILE_TRANSFER_HANDLER) && jListDoDnD(ev)) return;
                if (isPopup && gcp(KEY_NO_HELP_ON_RIGHT_CLICK,this)==null) {
                    if (popup==null) {
                        miCtrlPanel=new ChButton("Shared control panel").li(this);
                        miHelp=new ChButton("Help").li(this);
                        popup=jPopupMenu("list/context-menu",new Object[]{miHelp,miCtrlPanel});
                    }
                    boolean showPopup=false;
                    final Object o=get(idx,this);
                    if (o instanceof Class || o instanceof charite.christo.hotswap.ProxyClass) {
                        final Container cp=getSharedCtrlPan(o,true);
                        if (cp!=null) {
                            pcp("O",cp, miCtrlPanel.enabled(showPopup=true));
                            pcp("C",o,miCtrlPanel);
                        }
                    } else miCtrlPanel.enabled(false);
                    if (getHlp(o)!=null) pcp("C",o,miHelp.enabled(showPopup=true));
                    else miHelp.enabled(false);
                    if (showPopup) {
                        popup.show(this, x(ev), y(ev));
                    }
                }
                if (id==MouseEvent.MOUSE_CLICKED) pcp(KEY_LAST_MOUSE_EVENT,ev, this);
            }
            if (!isPopup) {
                handleActEvt(this,cmd,0);
                //final boolean hack=id==MOUSE_PRESSED && 0!=(options&ChJTable.DRAG_ENABLED) &&  0==(modi&(SHIFT_MASK|CTRL_MASK|ALT_MASK));
                //final boolean de=getDragEnabled();
                //if (!de && hack) setDragEnabled(true);
                super.processEvent(ev);
                //if (!de && hack) setDragEnabled(false);
            }
        }
    }
    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final Object para=gcp("O",q), clazz=gcp("C",q);
        if (q==miHelp) showHelp(clazz);
        if (q==miCtrlPanel) ChFrame.frame(shrtClasNam(clazz),para,ChFrame.SCROLLPANE|ChFrame.PACK|ChFrame.AT_CLICK).shw(0);
    }
    private ChButton miHelp, miCtrlPanel;
    private JPopupMenu popup;
    /* <<< Event <<< */
    /* ---------------------------------------- */
    /* >>> Tip >>> */
    @Override public String getToolTipText(MouseEvent ev) {
        final String tip=ballonMsg(this, ev);
        return  tip!=null ? addHtmlTagsAsStrg(tip) : super.getToolTipText();
    }
    /* <<< Tip  <<< */
    /* ---------------------------------------- */
    /* >>> Frame >>> */
    public ChFrame showInFrame(long option, String title) {
        JComponent pan=gcp(KEY_PANEL, this, JComponent.class);
        return ChFrame.frame(title,pan!=null ? pan : this,0L).shw(option);

    }
    /* <<< Frame <<< */
    /* ---------------------------------------- */
    /* >>> Static utils >>> */
    public static long getOptions(Object o) {
        return o instanceof ChJTable ? ((ChJTable)o).getOptions() :
            o instanceof ChJList ? ((ChJList)o).options :
            0;
    }
    public List getDndDateien() { return getDndFilesV(this);}

    /*
    @Override public int[] getSelectedIndices() {
        try {
        return super.getSelectedIndices();
        } catch(Exception ex) {
            putln(" ex="+ex+"  min="+getSelectionModel().getMinSelectionIndex()+" max="+getSelectionModel().getMaxSelectionIndex());
        }
        return NO_INT;
    }
    */
}
