package charite.christo;
import charite.christo.interfaces.*;
import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.util.List;
import java.util.*;
import javax.swing.table.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static java.awt.event.MouseEvent.*;
/**
   @author Christoph Gille
*/
public class ChJTable extends JTable implements ChProcessEvent, Comparator{
    {  assertInstAllowed(); }
    public final static String COLUMN_MOVED="ChJTable_COLUMN_MOVED", COLUMN_MARGIN_CHANGED="ChJTable_COLUMN_MARGIN_CHANGED";
    public final static int DEFAULT_RENDERER=1, ICON_ROW_HEIGHT=1<<1, EX_ROW_HEIGHT=1<<2, CLASS_RENDERER=1<<3, RTT=1<<4, FILE_POPUP=1<<5,  EXX_ROW_HEIGHT=1<<5, TABLE_ROW_SORTER=1<<6;
    public final static int FILE_TRANSFER_HANDLER=1<<6, SINGLE_SELECTION=1<<7,  SET_VALUE_SELECTED_ROWS=1<<9;
  public final static int DRAG_ENABLED=1<<10, NO_REORDER=1<<11, FILE_NAME=1<<12, CHJTABLE_FOCUSED_BLACK=1<<13, CHJTABLE_SHOW_BUT_MAXIM=1<<14,
      CHJTABLE_BACKSPACE_DEACTIVATED=1<<15, CHJTABLE_CLEAR=1<<16;
    private final ChEventAdapter LI=new ChEventAdapter(this);
    private String[] headerTT;
    private boolean isInitialized;
    private long options;
    public long getOptions() { return options;}
    public ChJTable() {}
    public ChJTable(TableModel m, long options) {
        this.options=options;
        if (m!=null) {
            setModel(m);
            if ((TABLE_ROW_SORTER&options)!=0) setRowSorter(null);
        }
        if ((DEFAULT_RENDERER&options)!=0) {
            final ChRenderer r=new ChRenderer();
            this.renderer=r;
            setDefaultEditor(Object.class, r);
        }
        if ((ICON_ROW_HEIGHT&options)!=0) setRowHeight(32);
        if ((EX_ROW_HEIGHT&options)!=0) setRowHeight(EX);
        if ((EXX_ROW_HEIGHT&options)!=0) setRowHeight(EX*3/2);
        getTableHeader().addMouseMotionListener(LI);
        if ( (options&NO_REORDER)!=0) getTableHeader().setReorderingAllowed(false);
        adSelectLisener(this);
    }

     @Override public void columnMoved(javax.swing.event.TableColumnModelEvent ev) {
         super.columnMoved(ev);
         handleActEvt(this,COLUMN_MOVED,0);
     }
    @Override public void columnMarginChanged(javax.swing.event.ChangeEvent ev) {
        super.columnMarginChanged(ev);
        handleActEvt(this,COLUMN_MARGIN_CHANGED,0);

    }
    { pcp(SCROLLABLE_TRACKS_VIEWPORT_WIDTH,"",this);}
    @Override public boolean getScrollableTracksViewportWidth() {  return gcp(SCROLLABLE_TRACKS_VIEWPORT_WIDTH,this)!=null; }

    /* <<< Paint <<< */
    /* ---------------------------------------- */
    /* >>> Renderer >>> */

    {pcp(KEY_NOT_YET_PAINTED,"",this);}
    @Override public void paintComponent(Graphics g) {
        try{
            super.paintComponent(g);
            if (!isInitialized) {
                isInitialized=true;
                pcp(KEY_NOT_YET_PAINTED,null,this);
            }
        } catch(Throwable e){}
    }

    /* <<< Paint <<< */
    /* ---------------------------------------- */
    /* >>> Renderer >>> */
    private TableCellRenderer renderer;
    public ChJTable renderer(TableCellRenderer r) { renderer=r; return this;}
    public ChRenderer getChRenderer() { return renderer instanceof ChRenderer ? (ChRenderer)renderer : null; }
    public ChJTable headerTip(String...tt) { headerTT=tt; return this;}
    @Override public TableCellRenderer getCellRenderer(int row, int column) {
        return (renderer!=null) ? renderer : super.getCellRenderer(row,column);
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent  >>> */
    //    private static long button3When;
    { this.enableEvents(ENABLE_EVT_MASK);}
    @Override public void processEvent(AWTEvent ev) {
        final AWTEvent ev2=processEvent(this,ev);
        if (ev2!=null) {
            try {
                /*
                if (MouseEvent.BUTTON3==buttn(ev)) {
                    final long time=System.currentTimeMillis(), diff=time-button3When;
                    button3When=time;
                    if (diff<999) {
                    }
                }
                */
                super.processEvent(ev2);
            } catch(Throwable e){}
        }
    }
    public AWTEvent processEvent(JComponent component, AWTEvent ev) {
        final int id=ev.getID(),  modi=modifiers(ev);
        final Point p=point(ev);
        if (p!=null && component instanceof JTableHeader) {
            final JTableHeader h=(JTableHeader)component;
            final int col=columnAtPoint(p);
            h.setToolTipText(col>=0 && col<sze(headerTT) ? headerTT[col]:null);
        }
        if (component==this && (id==MOUSE_CLICKED || id==MOUSE_PRESSED)) {
            boolean clearSelection=false;
            final int row=rowAtPoint(p), col=columnAtPoint(p);
            if (col>=0 && row>=0 && isCellEditable(row,col)) { /* Workaround: under windows difficult to get the cursor into a cell */
                editCellAt(row,col);
                final TableCellEditor tce=getCellEditor(row,col);
                final Component comp=tce!=null ? tce.getTableCellEditorComponent(this, getValueAt(row,col),false, row, col) : null;
                final ChTextField tf=comp instanceof ChTextField ? (ChTextField)comp : null;
                if (tf!=null) {
                    final Rectangle r=getCellRect(row,col,false);
                    if (r!=null) tf.processEvent(new MouseEvent( comp, id,0,0, p.x-r.x, p.y-r.y, 1,false));
                    tf.requestFocus();
                }
                if (comp instanceof AbstractButton && id==MOUSE_CLICKED) ((AbstractButton)comp).doClick();
            } else if (id==MOUSE_PRESSED) {
                final int[] selected=getSelectedRows();
                if (sze(selected)==1 && selected[0]==row) clearSelection=true;
                if (clearSelection) { clearSelection(); return null;}
            }
        }
        if (processEvt(this,ev))  return ev;
        return ev;
    }

    @Override public String getToolTipText(MouseEvent ev) {
        final StringBuffer sb=appendTooltips(this,ev,null);
        if (sb!=null) return sb.toString();
        return super.getToolTipText(ev);
    }

    public TableColumn getColumn(int col) {
        if (col<0 || col>=getColumnCount()) return null;
        try { return getColumnModel().getColumn(col); } catch(Exception e) {return null;}
    }
    public void setColumnWidth(int col,int w,boolean isResizable) {
        final TableColumn c=getColumn(col);
        if (c==null) return;
        if (w>0) {
            c.setPreferredWidth(w);
            if (!isResizable) { c.setMaxWidth(w); c.setMinWidth(w); }
        }
        doLayout();
    }
    public static int[] selectedRowsConverted(JTable t) {
        final int[] rr0=t!=null ? t.getSelectedRows() : null, rr;
        if (rr0==null) return NO_INT;

        if (javaVsn()>15) {
            rr=rr0.clone();
            //TableRowSorter
            for(int i=rr.length; --i>=0;) rr[i]=rowIndexToModel(rr[i],t);
        } else rr=rr0;
        return rr;
    }

    private final static int[] rowIndexRet={0};
    private final static Object[] rowIndexPar={null,rowIndexRet};
    public static int rowIndexToModel(int row, JTable t) {
        if (t instanceof ChJTable && javaVsn()>15) {
            rowIndexRet[0]=row;
            rowIndexPar[0]=t;
            forJavaVersionHigher(16).run("convertRowIndexToModel",rowIndexPar);
            return rowIndexRet[0];
        }
        return row;
    }

    public int compare(Object o1, Object o2) {
        if (o1==o2) return 0;
        if (o1 instanceof Boolean && o2 instanceof Boolean) return ((Boolean)o1).booleanValue() ? -1 : 1;
        if (o1 instanceof Object[] && get(0,o1)==ChRenderer.KEY_TOGGLE_SELECTED && o2 instanceof Object[] && get(0,o2)==ChRenderer.KEY_TOGGLE_SELECTED) {
            final Color c1=get(1,o1,Color.class);
            final Color c2=get(1,o2,Color.class);
            if (c1!=null && c2!=null) {
                final int rgb1=c1.getRGB(),rgb2=c2.getRGB();
                return rgb1<rgb2?-1 : rgb1==rgb2?0 : 1;
            }
        }
        if (o1 instanceof Comparable)  return ((Comparable)o1).compareTo(o2);
        return 0;
    }
    public static void setRowSorter(Comparator cc[], JTable t) {
        if (javaVsn()>15) {
            forJavaVersionHigher(16).run("setRowSorter",new Object[]{t,cc});
        }
    }
public List getDndDateien() { return getDndFilesV(this);}
}
