package charite.christo;
import charite.christo.interfaces.*;
import javax.swing.*;
import java.util.*;
import java.util.List;
import javax.swing.tree.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static java.awt.event.MouseEvent.*;
/**
   @author Christoph Gille
*/
public class ChJTree extends JTree implements  HasWeakRef  {
    public final static String KEY_NUM_EXPAND="CT$$NE", KEY_EXPAND="CT$$KE";
    private boolean alreadyPainted;

    public ChJTree(long options, TreeModel m) {
        this.options=(int)options;
        if (m!=null) setModel(m);
        if ((ChJTable.DEFAULT_RENDERER&options)!=0) setCellRenderer(new ChRenderer());
        pcp(SCROLLABLE_TRACKS_VIEWPORT_WIDTH,"",this);
        adSelectLisener(this);
        if ( (options&ChJTable.DRAG_ENABLED)!=0) setDragEnabled(true);
    }
    private int options;

    @Override public boolean getScrollableTracksViewportWidth() {  return gcp(SCROLLABLE_TRACKS_VIEWPORT_WIDTH,this)!=null; }
    private final java.lang.ref.Reference WEAK_REF=new java.lang.ref.WeakReference(this);
    public java.lang.ref.Reference getWeakRef() { return WEAK_REF; }

    /* ---------------------------------------- */
    /* >>> Event >>> */
    { this.enableEvents(ENABLE_EVT_MASK);}
    @Override public void processEvent(java.awt.AWTEvent ev) {
        final int id=ev.getID(), modi=modifiers(ev);
        if (0!=(options&ChJTable.DRAG_ENABLED) && jListDoDnD(ev)) return;
        if (this.needsUpdate || id==MOUSE_ENTERED) updateKeepState(0L);

        if (processEvt(this,ev)) {
            //final boolean hack=id==MOUSE_PRESSED && 0!=(options&ChJTable.DRAG_ENABLED) &&  0==(modi&(SHIFT_MASK|CTRL_MASK|ALT_MASK));
            //final boolean de=getDragEnabled();
            //if (!de && hack) setDragEnabled(true);
            try {super.processEvent(ev);} catch(Throwable e){}
            //if (!de && hack) setDragEnabled(false);
        }
    }
    /* <<< Event <<< */
    /* ---------------------------------------- */
    /* >>> Paint >>> */
    {pcp(KEY_NOT_YET_PAINTED,"",this);}
    @Override public void paintComponent(java.awt.Graphics g) {
        if (!alreadyPainted) {
            alreadyPainted=true;
            pcp(KEY_NOT_YET_PAINTED,null,this);
            for(int i=atoi(gcp(KEY_NUM_EXPAND,this)); --i>=0;) expandAllNodes(null,this);
        }

        if (this.needsUpdate) updateKeepState(0L);
        if (paintHooks(this, g, false)) try {super.paintComponent(g);}catch(Throwable t){stckTrc(t);}
        paintHooks(this, g, true);
    }
    @Override public void updateUI() { try { super.updateUI();} catch(Exception e){stckTrc(e);}  }
    /* <<< Paint <<< */
    /* ---------------------------------------- */
    /* >>> Tip >>> */
    @Override public String getToolTipText(java.awt.event.MouseEvent ev) {
        final String tip=ballonMsg(this, ev);
        return  tip!=null ? tip : super.getToolTipText();
    }
    { rtt(this);}
    /* <<< Tip <<< */
    /* ---------------------------------------- */
    /* >>> Save State  >>> */
    private final Set<String> vExpanded=new HashSet();
    private final static StringBuffer SB=new StringBuffer();
    private TreePath[] selectionPaths;
    private void saveState() {
        vExpanded.clear();
        for(int row=getRowCount(); --row>=0;) {
            final TreePath path=getPathForRow(row);
            if (isExpanded(path)) vExpanded.add(path2String(path));
        }
        this.selectionPaths=getSelectionPaths();
    }
    private void applySavedState() {
        final long time=System.currentTimeMillis();
        for(int depth=4; --depth>=0;) {
            boolean changed=false;
            for(int row=getRowCount(); --row>=0;) {
                final TreePath path=getPathForRow(row);
                if (isExpanded(path)) continue;
                if (vExpanded.contains(path2String(path))) {
                    changed=true;
                    setExpandedState(path,true);
                }
            }
            if (!changed)break;
        }
        if (sze(selectionPaths)>0) setSelectionPaths(selectionPaths);
        //putln("applySavedStates time="+(System.currentTimeMillis()-time));
    }
    private final WeakHashMap<Object,String> mapPath=new WeakHashMap();
    private String path2String(TreePath path) {
        String o=mapPath.get(path);
        if (o==null) {
            SB.setLength(0);
            path2String(path,SB);
            o=SB.toString();
            mapPath.put(path,o);
        }
       return o;
    }
    private static void path2String(TreePath path, StringBuffer SB) {

        for(TreePath p=path; p!=null; p=p.getParentPath()) {
            final Object o=p.getLastPathComponent();
            final Object nodeId= o instanceof ChRunnable ? ((ChRunnable)o).run(PROVIDE_JTREE_NODE_ID,null) : null;
            if (nodeId!=null) SB.append(nodeId);
            else SB.append(ChRenderer.hashC(o));
            SB.append('/');
        }
    }
    /* <<< Save State <<< */
    /* ---------------------------------------- */
    /* >>> Update JTree >>> */
    private boolean needsUpdate;
    private Thread _threadUpdate;
    public void updateKeepStateLater(long options, int ms) {
        this.needsUpdate=true;
        ChDelay.runAfterMS(EDT,thread_updateKeepState(options),ms);
    }
    public Thread thread_updateKeepState(long options) {
        if (_threadUpdate==null) _threadUpdate=threadM("updateKeepState", this, new Object[]{ChUtils.longObjct(options)});
        return _threadUpdate;
    }
    private long sumChildsHC;

    public void updateKeepState(long options) {
        final long sumChildsHC=sumOfChildsHashValues();

        final boolean hashChanged=this.sumChildsHC!=sumChildsHC;

        this.sumChildsHC=sumChildsHC;

        if (hashChanged) {
            //if (isMyComputer()) putln(ANSI_FG_YELLOW+"ChJTree needsUpdate"+ANSI_RESET);
            final TreeModel m=getModel();
            this.needsUpdate=false;
            final long time=System.currentTimeMillis();
            saveState();
            if (m instanceof ChTreeModel)  {
                ((ChTreeModel)m).dispatch(ChTreeModel.DISPATCH_CHANGED,new javax.swing.event.TreeModelEvent(m, new TreePath(new Object[]{m.getRoot()})));
            }
            revalidate();
            applySavedState();
        } else {
            if (ListOrTreeNeedsRepaint.treeNeedsRepaint(this)) {
                if (isMyComputer()) putln(ANSI_FG_YELLOW+"ChJTree needsRepaint"+ANSI_RESET);
                ChDelay.repaint(this,222);
            }
        }

        final Object toBeExpanded=gcp(KEY_EXPAND,this);
        if (toBeExpanded!=null) expandAllNodes(toBeExpanded,this);

        pcp(KEY_EXPAND,null,this);
        //putln("updateKeepState time="+(System.currentTimeMillis()-time));
    }
    public static void updateTreeLater(long options, JTree t, int ms) {
        if (t instanceof ChJTree) ((ChJTree)t).updateKeepStateLater(options, ms);
    }
    /* <<< update tree <<< */
    /* ---------------------------------------- */
    /* >>> Private long hash >>> */
    private long sumOfChildsHashValues() {
        long sum=0;
        int toLen=0, shift=0;
        final TreeModel m=getModel();
        boolean rootDone=false;
        for(int i=getRowCount(); --i>= -1;) {
            final TreePath tp=i>=0 ? getPathForRow(i) :  null;
            final Object parent;
            if (tp==null) {
                if (rootDone) continue;
                rootDone=true;
                parent=m.getRoot();
            } else parent=tp.getLastPathComponent();
            final int pathCount=tp==null ? 1 : tp.getPathCount();
            if (pathCount<toLen || rootDone) {
                final Object childs[]=parent instanceof UniqueList ? ((UniqueList)parent).asArray() : null;
                for(int iChild=childs!=null? childs.length : m.getChildCount(parent); --iChild>=0;) {
                    final Object child=childs!=null ? childs[iChild] : m.getChild(parent,iChild);
                    sum+=(ChRenderer.hashC(child)<<(shift++%32));
                }
            }
            toLen=pathCount;
        }
        return sum;
    }

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Util >>> */
    public static void expandAllNodes(Object object, JTree t) {
        if (t==null) return;
        final Object objects[]= object instanceof Object[] ? (Object[])object:null;
        while(true) {
            boolean changed=false;
            for(int i=t.getRowCount(); --i>=0;) {
                final TreePath tp=t.getPathForRow(i);
                if (tp==null || t.isExpanded(tp)) continue;
                final Object pc=object==null ? null : tp.getLastPathComponent();
                if (object==pc || objects!=null && cntains(pc,objects)) {
                    t.expandPath(tp);
                    changed=true;
                }
            }
            //if (!changed) break;
            break;
        }

    }
    public void sort(int type) {
        final TreeModel m=getModel();
        final int selRows[]=getSelectionRows();
        for(int onlySelected=2; --onlySelected>=0;) {
            int count=0;
            for(int i=getRowCount(); --i>=0;) {
                final TreePath tp=getPathForRow(i);
                if (!isExpanded(tp)) continue;
                if (onlySelected>0 && idxOf(i,selRows,0, MAX_INT)<0) continue;
                final Object o0=tp.getLastPathComponent();
                ChComparator.sortTreeNodes(ChTreeModel.rplc(o0), type);
                count++;
            }
            if (count>0) break;
        }
        updateKeepState(0L);
    }

    /* <<< util <<< */
    /* ---------------------------------------- */
    /* >>> DnD  >>> */
    public List getDndDateien() { return getDndFilesV(this);}

}
