package charite.christo;
import java.net.URL;
import java.io.File;
import java.io.PrintWriter;
import java.io.ByteArrayOutputStream;
import java.util.Collection;
import java.util.ArrayList;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;

/**
   @author Christoph Gille
**/
public class ChJavacCompiler  {
    public final ChTextArea LOG=new ChTextArea("");
    {
        pcp(KEY_IF_EMPTY,"This panel will contain \nthe compiler messages",LOG);
        pcp(KEY_SCROLL_TO_END,"",LOG);
    }

    public boolean compile(File sourceFile) {
        final boolean isLogging=true;
        final File classFile=file(delSfx(".java",sourceFile)+".class");
        final ArrayList<String> argl=new ArrayList();
        adAllUniq(toStrgArry(OPTIONS), argl);
        CLASSPATH.add(thisJarFile());
        final StringBuffer sbPath=new StringBuffer(444);
        for(Object o:CLASSPATH.toArray()) {
            if (o!=null) sbPath.append(file(o)).append(systProprty(SYSP_PATH_SEPARATOR));
        }
        argl.add("-classpath");
        argl.add(sbPath.toString());
        argl.add(sourceFile.toString());
        final ByteArrayOutputStream baos=new ByteArrayOutputStream();
        boolean wasSuccessful=false;
        try {
            final long ageClassFile=classFile.lastModified();
            invokeCompiler(new PrintWriter(baos), toArry(argl,String.class));
            wasSuccessful=ageClassFile!=classFile.lastModified();
        } catch (Exception ioex) {
            stckTrc(ioex);
            if (isLogging) LOG.appnd(stackTrceAsStrg(ioex),"\n");
            wasSuccessful=false;
        }
        if (isLogging) {
            final String sArg=argl.toString();
            if (!sArg.equals(sLastArg) || LOG.getDocument().getLength()==0) {
                sLastArg=sArg;
                LOG.appnd(fTools,"\n"+ANSI_WHITE+"Compiler arguments: "+ANSI_YELLOW+sArg+ANSI_RESET+"\n");
            }
            LOG.appnd(sourceFile+"\n"+baos.toString().replace('\t',' ')+"\n");
            if (!wasSuccessful) LOG.appnd("Compilation "+ANSI_RED+"failed"+ANSI_RESET+"\n\n");
            else LOG.appnd(ANSI_GREEN+" success: "+ANSI_RESET+classFile+" "+classFile.length()+" bytes\n\n");
        }
        return wasSuccessful;
    }
    private String sLastArg;
    private final ArrayList<String> OPTIONS=new ArrayList();
    public void addOptions(String... options) { adAllUniq(options,OPTIONS);}
    private final ArrayList<Object> CLASSPATH=new ArrayList();
    public void addURLsToClasspath(Object... oo) {
        if (oo==null) return;
        for(Object o : oo) {
            if (o instanceof Collection) addURLsToClasspath(((Collection)o).toArray());
            else if (o instanceof Object[]) addURLsToClasspath((Object[])o);
            else adUniq(toStrg(url(o)),CLASSPATH);
        }
    }
    // ================================
    // Static Fields
    // ================================
    private static java.lang.reflect.Method compile;
    private static File fTools;
    private static int alreadyDownloaded;
    public static void downloadCompiler() {
        fTools=toolsDotJar();
        if (fTools==null) {
            final char c2=systProprty(SYSP_JAVA_VERSION).charAt(2);
            final String TOOLS_JAR="tools_"+systProprty(SYSP_JAVA_VERSION).substring(0,2)+ (c2>'6'?'6':c2) +".jar";
            fTools=file(dirJars(),TOOLS_JAR);
            if ( fTools.length()==0) renamFile(file(dirJars(),TOOLS_JAR+"_"), fTools);

            if ( fTools.length()==0 && alreadyDownloaded++<5) {
                final String suffix=javaVsn()==14 ? "" : ".pack.gz";
                InteractiveDownload.downloadFilesAndUnzip(new URL[]{url(URL_STRAP_JARS+TOOLS_JAR+suffix)},dirJars());
                if (ChZip.supportForPack200()) ChZip.unpack200(dirJars(),ChZip.PACK_IF_NEWER);
                //                if (fTools.length()==0) error("The file "+fTools+" Does not exist.<br>if you have installed jdk and not only jre than you have tools.jar on your hard disk.<br>Please copy or link it.<br>Subsequently restart the hotswap dialog");
            }
        }
    }

    private static Class main;
    private boolean invokeCompiler(PrintWriter printWriter,String argv[]) {
        if (compile==null) {
            if (fTools==null) {
                downloadCompiler();
            }
            if (main==null) {
                ChClassLoader.newInstanceIfNewJar();
                main=Insecure.findClassByName("com.sun.tools.javac.Main");
            }
            try {
                if (compile==null) compile=main.getMethod("compile",new Class[]{ String[].class ,PrintWriter.class} );
            } catch(Exception e){ putln("caught ChJavacCompiler#invokeCompiler ",e);}
        }
        try {
            compile.invoke(null, new Object[]{argv,printWriter});
            return true;
        } catch(Exception ex){
            putln("caught ChJavacCompiler#invokeCompiler ", ex);
            return false;
        }
    }
}
/*
http://today.java.net/pub/a/today/2008/04/10/source-code-analysis-using-java-6-compiler-apis.html
*/