package charite.christo;
import java.util.*;
import java.lang.ref.*;
import javax.swing.Icon;
import charite.christo.interfaces.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
public class ChMap<K,V>  implements Map<K,V>, ChRunnable, HasModiCount, TabItemTipIcon.Interface {
    public final static int TYPE_TreeMap=1<<1, TYPE_WeakHashMap=1<<2, WORD_COMPLETION_K=1<<3, WORD_COMPLETION_V=1<<4;

    private transient volatile int mc;
    private final Class clazzK,clazzV;
    private final int options, initialCapacity;
    public ChMap() { this(0L, Object.class,Object.class,99); }
    public ChMap(long options, Class cK, Class cV, int capacity) {
        this.clazzK=cK;
        this.clazzV=cV;
        this.options=(int)options;
        this.initialCapacity=capacity;
    }

    private Map _map;
    private Map<K,V> m() {
        if (_map==null) {
            _map=
                0!=(options&TYPE_TreeMap) ? new TreeMap() :
                0!=(options&TYPE_WeakHashMap) ? new WeakHashMap(initialCapacity) :
                new HashMap(initialCapacity);

        }
        return _map;
    }
    public  int size(){ return m().size();}
    public  boolean isEmpty(){ return m().isEmpty();}
    public  boolean containsKey(Object key){ return m().containsKey(key);}
    public  boolean containsValue(Object value){ return m().containsValue(value);}
    public  V get(Object key){ return m().get(key); }
    public  Set<K> keySet(){ return m().keySet();}
    public  Collection<V> values(){ return m().values();}
    public  Set<Map.Entry<K, V>> entrySet(){ return m().entrySet();}

    public V put(K key, V value) {
        final V old=m().put(key,value);
        if (old!=value) { mc++; refCached=null; }
        return old;
    }
     public void clear() {
        mc++;
         refCached=null;
        m().clear();
    }

     public void putAll(Map<? extends K,? extends V> m) {
        mc++;
        refCached=null;
        m().putAll(m);
    }
     public V remove(Object o) {
        mc++;
        refCached=null;
        return m().remove(o);
    }

    public int getModificationCount() { return mc;}

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Arrays >>> */

    private Reference<Object[]> refCached;
    private int lastSize;
    Object[] cached() {
        final Reference<Object[]> r=refCached;
        Object oo[]=r!=null ? r.get() : null;
        final int L=size();
        if (oo==null) refCached=new SoftReference(oo=new Object[9]);
        else if (this.lastSize!=L) for(int i=oo.length; --i>=0;)oo[i]=null;
        this.lastSize=L;
        return oo;
    }

    public V[] vArray(ChRunnable run, String runID) { return (V[]) toArrayKV(false, run, runID); }
    public K[] kArray(ChRunnable run, String runID) { return (K[]) toArrayKV(true,  run, runID); }
    public V[] vArray() { return (V[]) toArrayKV(false, null,""); }
    public K[] kArray() { return (K[]) toArrayKV(true,  null,""); }
    Object toArrayKV(boolean isKey, ChRunnable run, String runID) {
        final Object cached[]=cached();
        Object[] vv=(Object[])cached[isKey ? C_ARRAYk : C_ARRAYv];
        if (vv==null) {
            vv=(isKey ? keySet() : new HashSet(values())).toArray(emptyArray(isKey ? clazzK : clazzV));
            vv=rmNull(vv,isKey ? clazzK : clazzV);
            cached[isKey ? C_ARRAYk : C_ARRAYv]=vv;
            if (run!=null) run.run(runID,vv);
        }
        return vv;
    }

    //    public Map<String,HasID> mapID2Value()
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Derived >>> */
    public Map<String,V[]>  getMapName2objects(int reserve) {
        final Object cached[]=cached();
        Map<String, V[]> m=(Map)cached[C_MAP_name2o];
        if (m==null) {
            cached[C_MAP_name2o]=m=new HashMap();
            for(V v : vArray()) mapOneToMany(v, getNam(v), m, 2, clazzV);
        }
        return m;
    }

    @Override public  boolean equals(Object o){ return m().equals(o);}
    @Override public  int hashCode(){return m().hashCode(); }
    //@Override public  String toString() { return super.toString()+"#"+size();}
    @Override public  String toString() { return m().toString();}

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> As List >>> */

    private int vList_mc, kList_mc;
    private List vList,kList;
    public List list(boolean isKey, ChRunnable run ,String runID) {
        List v= isKey ? kList : vList;
        if ( (isKey? kList_mc : vList_mc)!=mc || v==null) {
            vList_mc=mc;
            if (v==null) {
                v=new ArrayList();
                if (isKey) kList=v; else vList=v;
            } else v.clear();
            v.addAll(isKey ? keySet() : values());
            while(v.remove(null)); /* values kann null enthalten */
            if (run!=null) run.run(runID,vList);
        }
        return v;
    }
    public List<V> vList(ChRunnable run ,String runID) { return list(false, run, runID); }
    public List<K> kList(ChRunnable run ,String runID) { return list(true, run, runID); }
    public List<V> vList() { return list(false, null, ""); }
    public List<K> kList() { return list(true, null, ""); }

    /* <<< As List <<< */
    /* ---------------------------------------- */
    /* >>> Renderer >>> */
    private String labelText;
    private Object ic;

    public ChMap t(String title) { this.labelText=title; return this;}
    public ChMap i(String icon)  { this.ic=icon; return this;}
    public TabItemTipIcon getTabItemTipIcon(java.awt.AWTEvent ev) {
        final Icon icon=ic instanceof Icon ? (Icon)ic : ic instanceof String ? iicon((String)ic) : null;
        final String t=labelText;
        return TabItemTipIcon.shared(null,t,null, icon);
    }

   /* <<< Renderer  <<< */
    /* ---------------------------------------- */
    /* >>> Word completion >>> */
    private Object[] wordCompl;
    private int mc_treeChilds;
    public  Object run(String id,Object arg) {
        if (id==PROVIDE_WORD_COMPLETION_LIST) {
            final boolean k=0!=(options&WORD_COMPLETION_K), v=0!=(options&WORD_COMPLETION_V);
            if (!v && !k) return null;
            if (v && k) {
                if (wordCompl==null) wordCompl=new Object[2];
                wordCompl[0]=vArray();
                wordCompl[1]=kArray();
                return wordCompl;
            }
            return v ? vArray() : kArray();
        }
        if (id==DialogStringMatch.PROVIDE_HAYSTACK || id==PROVIDE_JTREE_CHILDS) {
            return vArray();
        }

        return null;
    }
}