package charite.christo;
import charite.christo.interfaces.ChRunnable;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.lang.ref.Reference;
import java.net.URL;
import java.net.URLConnection;
import java.util.*;
import javax.swing.JComponent;
import javax.swing.JPopupMenu;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static java.awt.event.MouseEvent.*;
/**HELP
<h2>Showing Abstracts</h2>

When the mouse hovers over a Pubmed reference of the form
\"PMID:19762348\", the abstract is downloaded automatically in the
background and shown in a certain text area.  The same works for
Uniprot identifiers of the form UNIPROT:Q9H936 or human Ensembl
references of the form ENSG00000103064.

<h2>Showing Full-text</h2>

The user might associate a full PDF text with a Pubmed ID. Those
Pubmed links are indicated by red background. The PDF can be looked at
by using the context menu (right click).  If the ALT-key is hold down
while the mouse is moving over a Pubmed ID, then not the abstract, but
the full text (without images) is shown.

<h2>Context menu</h2>

Right click on a underlined PubMed link.
<h2>The download directory</h2>
PDF-files are downloaded with the Web-browser into a certain directory.
The program needs to know this directory.
The  PDF file that has been created at last in this directory
is identified by the program and associated to the respective PMID.

<h2>Customizing the Browser</h2>

Many Web-browsers display PDF documents within the
Browser's frame.  With this behavior, the mechanism is not working.
The browser  must be configured such that
PDF files are downloaded into a specific directory and
displayed in an external PDF viewer such as WIKI:Evince or WIKI:Xpdf
or WIKI:Acroread.
<br><br>
In Firefox browser go to Menu-bar&gt;Edit&gt;Preferences.  The
preferences frame will open.  Click the Icon "Applications". A list
with file types appear.  Find the line "PDF document" and specify an
action. A good choice which gives maximum control is "Always ask".
<br><br>
In other Web browsers customization works in a similar way.

@author Christoph Gille

 */
public final class ChPubmed implements ChRunnable, ActionListener {
    public final static String ACTION_FULL_TEXT_ASSOCIATED="CPM$$FTA";
    public final static int LIST=0, PDF=1, ABSTRACT=2, UNIPROT=3, PDF2TXT=4, GO=5, AUTO=6, PDF_TITLE=7;
    private final static ByteArray baTmp=new ByteArray(999);
    private static Map<String,Thread> mapThread=new HashMap();
    private static Map<String,Vector<String> > mapQueue=new HashMap();
    private static Map<String,HashSet<String> > mapRequests=new HashMap();
    private static ChButton bPdf2txt, togShared;
    private final static Map<String,File>[] mapPMID_Pdf=new Map[GO+1];
    private static File[] dirL=new File[GO+1];
    public static void queueForDownload(String id0,  boolean isListPdf) {
        final String id;
        if (isListPdf) {
            final int digit=nxt(DIGT,id0);
            if (digit<0) return;
            id=id0.substring(digit);
            if (sze(getFile(id,LIST))>0) return;
        } else id=id0;
        final String T=isListPdf ? "LIST" : "ABSTRACT";
        Thread t=mapThread.get(T);
        if (t==null) {
            mapThread.put(T,t=ChDelay.newThread(instance(),T,null));
            mapRequests.put(T, new HashSet());
            mapQueue.put(T, new Vector());
            t.start();
        }
        if (!mapRequests.get(T).contains(id)) {
            mapRequests.get(T).add(id);
            mapQueue.get(T).add(id);
        }
    }
    /* <<< queue <<< */
    /* ---------------------------------------- */
    /* >>> Threading >>> */
    public Object run(String id, Object arg ) {
        final boolean bySfx=id=="ABSTRACT";
        if (bySfx || id=="LIST") {
            while(true) {
                final Vector<String> vQ=mapQueue.get(id);
                if (sze(vQ)==0) { sleep(500); continue;}
                final String db_id=get(0,vQ,String.class);
                if (db_id==null) continue;
                vQ.remove(db_id);
                final File fDest=getFile(db_id, bySfx ? AUTO : LIST);
                if (sze(fDest)==0) {
                    if (bySfx) {
                        final String sURL=Hyperrefs.toUrlString(db_id,Hyperrefs.PLAIN_TEXT|Hyperrefs.PROTEIN_FILE);
                        //putln(db_id+" sURL="+sURL);
                        final File fSrc=file(sURL);
                        if (fSrc!=null && sze(fSrc)==0)  {
                            //putln("ChPubmed.unattendedDownload ",url(sURL),"  ",fSrc);
                            unattendedDownload(url(sURL));
                            handleActEvt(ChPubmed.class, ACTION_FULL_TEXT_ASSOCIATED, 0);
                        }
                        if (sze(fSrc)>0 && !fSrc.equals(fDest)) renamFileOrCpy(fSrc,fDest);
                    } else  collectPdfs(atoi(db_id,nxt(DIGT,db_id)));
                }
            }
        }
        if (id=="SHARED") {
            listSharedPdf=true;
            if (setSharedPdf==null) setSharedPdf=new HashSet();
            final File dir=dirSharedPdf();
            final String ff[]=dir!=null && dir.isDirectory() ? dir.list() : null;
            synchronized(setSharedPdf) {setSharedPdf.clear();}
            if (ff!=null) {
                for(String f : ff) {
                    if (f!=null && f.endsWith(".pdf")) {
                        synchronized(setSharedPdf) { setSharedPdf.add(delSfx(".pdf",f));}
                    }

                }
            }
            handleActEvt(ChPubmed.class, ACTION_FULL_TEXT_ASSOCIATED, 0);
            listSharedPdf=false;
        }
        if (id=="CPY_PDF") {
            final File fPdf=(File)arg;
            if (dirSharedPdf().isDirectory()) {
                final File fCopy=file(dirSharedPdf()+"/"+fPdf.getName()), fTmp=file(fCopy+"TMP");
                //putln("Going to copy "+fPdf+" to "+fCopy);
                cpy(fPdf,fTmp);
                renamFileOrCpy(fTmp, fCopy);
                if (!isSystProprty(IS_WINDOWS)) new ChExec(0).setCommandLine("chmod a+r "+fCopy).run();
                if (sze(fCopy)==0) error("could not copy "+fPdf+" to "+fCopy);
                //putln(" fCopy "+sze(fCopy)+" Bytes");
            }
        }
        if (id=="PDF_TITLE") {
        }
        if (id=="PDF2TXT") {
           setEnabld(bPdf2txt,false);
           viewFile(getDir(PDF2TXT), 0);
           for(File dir : new File[]{getDir(PDF),dirSharedPdf()}) {
               final String ss[]=dir.list();
               for(int i=sze(ss); --i>=0;) {
                   if(!ss[i].endsWith(".pdf")) continue;
                   final File fPdf=new File(dir,ss[i]);
                   if (sze(fPdf)==0) continue;
                   final File fTxt=getFile(withoutDotSuffix(ss[i]),PDF2TXT);
                   if (sze(fTxt)==0 || fTxt.lastModified()<fPdf.lastModified()) {
                       new ChPdfUtilsPROXY().pdf2txt(fPdf,fTxt);
                       ChUtils.sleep(111);
                   }
               }
           }
           setEnabld(bPdf2txt,true);
        }
        return null;
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */

    private static Thread threadPdfRenameUseTitle(File dirP, String fileSuffix) {
        final Object[] arg={dirP, fileSuffix};
        return threadM("pdfRenameUseTitle", ChPubmed.class,arg);
    }
    public static void pdfRenameUseTitle(File dirP, String fileSuffix) {
        //putln("pdfRenameUseTitle "+dirP+" "+fileSuffix);
        final String ff[]=dirP.list();
        ByteArray sbSkip=null, sbSuccess=new ByteArray(999);
        for(int i=sze(ff); --i>=0;) {
            final String fn=ff[i];
            final File fP=file(dirP,fn);
            final int iPmid=atoi(fn);
            if (sze(fP)==0 || !fn.endsWith(fileSuffix) || iPmid==0) continue;
            final String title=pubmedTitle(iPmid);
            if (sze(title)==0) {
                if (sbSkip==null) sbSkip=new ByteArray(999).append("Skipped because abstract not yet downloaded: \n");
                sbSkip.append("PUBMED:").append(iPmid).append('\n');
                continue;
            }
            String t=rplcNoFilenameChars(title);
            t=rplcToStrg(".","_",t);
            t=rplcToStrg("__","_",t);
            t=rplcToStrg("--","-",t);
            t=rplcToStrg("_-","-",t);
            t=rplcToStrg("-_","-",t);
            while (sze(t)>0 && !is(LETTR_DIGT,t,0)) t=t.substring(1);
            while (sze(t)>0 && !is(LETTR_DIGT,t,sze(t)-1)) t=t.substring(0,sze(t)-1);
            if (sze(t)==0) continue;
            final File fNew=file(dirP+"Title/"+t+fileSuffix);
            if (!fNew.exists()) {
                hardLinkOrCopy(fP,fNew);
                sbSuccess.append(fP).append(" ==> ").append(fNew).append(fNew.exists() ? " success\n\n":" failed\n\n");
            }
        }
        if (sbSkip!=null) sbSuccess.append('\n').append(sbSkip);
        new AsciiView(sbSuccess).showInFrame("Rename using Pubmed title",ChFrame.AT_CLICK);

    }

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> File >>> */
    public static File getFile(String w0, int type0) {
        final int type;
        if (type0==AUTO) {
            type=
                w0==null || w0.indexOf('{')>=0 ? -1 :
                w0.startsWith("PMID") || w0.startsWith("PUBMED") || w0.startsWith("pmid") ? ABSTRACT :
                w0.startsWith("GO:") ? GO :
                w0.startsWith("UNIPROT:") || w0.startsWith("SWISS:")? UNIPROT : -1;
        } else type=type0;
        if (type<0) return file(Hyperrefs.toUrlString(w0,Hyperrefs.PLAIN_TEXT|Hyperrefs.PROTEIN_FILE));
        Map<String,File> m=mapPMID_Pdf[type];
        if (m==null) m=mapPMID_Pdf[type]=new HashMap();
        File f=m.get(w0);
        if (f==null) {
            final String w;
            if (type==ABSTRACT || type==PDF2TXT || type==PDF ||  Hyperrefs.hasNumericID(w0)) {
                final int num=nxt(DIGT,w0);
                if (num<0) return null;
                w=w0.substring(num, nxtE(NO_DIGT,w0,num, MAX_INT));
            } else w=w0.replace(':','_');
            m.put(w0,f=file(getDir(type)+"/"+w+ getFileSuffix(type)));
        }
        return f;
    }
    private static String getFileSuffix(int type) {
        return type==LIST ? ".list" : type==PDF ? ".pdf" : type==UNIPROT ? ".swiss" : type==PDF2TXT ? ".txt" : type==ABSTRACT ? ".M" : "";
    }
    private static boolean listSharedPdf;
    private static HashSet setSharedPdf;
    public static boolean hasPdf(String pmid) {
        if (setSharedPdf==null) mayUpdateShared(true);
        return pmid!=null && setSharedPdf!=null && (setSharedPdf.contains(pmid) || sze(getFile(pmid,PDF))>0);
    }
    /* <<< Files <<< */
    /* ---------------------------------------- */
    /* >>> Directories >>> */
    public static File getDir(int type) {
        if (dirL[type]==null){
            final String s=type==LIST?"listPdfs": type==PDF ? "pdfs" : type==PDF_TITLE ? "pdfTitle" : type==UNIPROT ? "uniprot" : type==PDF2TXT ? "pdf2txt" : type==GO ? "GO" : "abstracts";
            mkdrs(dirL[type]=file(dirSettings()+"/pubmed/"+ s));
        }
        return dirL[type];
    }
    private static ChTextField tfDirDownloadedPdf[], _tfDirSharedPdf;
    private static ChTextField tfDirSharedPdf() {
        if (_tfDirSharedPdf==null) {
            final String s=(isSystProprty(IS_WINDOWS) ? "//olli.charite.de" : "/nfs")+"/sysbio/Forschung/pdf4pmid";
            _tfDirSharedPdf=new ChTextField(s).li(instance()).saveInFile("dirSharedPdf");
            pcp(ACTION_FOCUS_LOST,"SHARED",_tfDirSharedPdf);
        }
        return _tfDirSharedPdf;
    }
    private static File _dirSharedPdf;
    public static File dirSharedPdf() {
        File d=_dirSharedPdf;
        if (d==null) d=_dirSharedPdf=file(tfDirSharedPdf());
        return d;
    }
    private static ChTextField[] tfDirDownloadedPdf() {
        ChTextField tf[]=tfDirDownloadedPdf;
        if (tf==null) {
            final String ss[]={"~/","~/Desktop/","~/Downloads/",ChEnvironment.get("TEMP"),"C:/temp/","D:/temp/","/tmp/"};
            tf=tfDirDownloadedPdf=new ChTextField[ss.length];
            for(int i=ss.length; --i>=0;) {
                final File f= ss[i]==null || ss[i].indexOf(':')>0 && isSystProprty(IS_WINDOWS) ? null : file(ss[i]);
                if (f==null || !f.isDirectory()) continue;
                tf[i]=new ChTextField(rplcToStrg(" ","%20",toStrg(f)), COMPLETION_DIRECTORY_PATH).saveInFile("ChPubmed_dirDownloadedPdf"+i);
            }
        }
        return tf;
    }
    /* <<< Directories <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */
    private static void mayUpdateShared(boolean always) {
        if (always || !listSharedPdf) ChDelay.newThread(instance(),"SHARED",null).start();
    }
    private final static String[] EE={".pdf",".PDF","fulltext","PDFSTART"};
    protected static boolean maybeAssociatePdf(int iPmid, boolean longDialog) {
        if (iPmid<=0) return false;
        if (togShared==null) togShared=toggl("Save a copy of the PDF in the shared directory").s(true);
        final Object
            tff[]=tfDirDownloadedPdf(),
            title=monospc(pubmedTitle(iPmid)),
            msg=vB(
                   "<h2>Associating PDF with the Pubmed abstract</h2>",
                   title,
                   " ",
                   "Customize the browser such that PDFs are not shown within the browser.<br>Specify the directories where the Web browser is downloading PDFs to.",
                   tff,
                   " ",
                   togShared.cb(),
                   hB("Shared directory: ",tfDirSharedPdf()),
                   " ",
                   longDialog ? pnl("First visit Pubmed by clicking ",monospc("PUBMED:"+iPmid)," and download PDF.") : null,
                   "After downloading confirm this dialog by pressing \"Yes\"",
                   " "
                   );
        if (ChMsg.yesNo(msg)) {
            final File dirs[]=new File[tff.length];
            for(int i=dirs.length; --i>=0;) dirs[i]=file(tff[i]);
            final File pdfLastModi=mostRecentFile(dirs,EE);
            if (sze(pdfLastModi)>0) {
                final File fFullText=createTempFile(".txt");
                new ChPdfUtilsPROXY().pdf2txt(pdfLastModi,fFullText);
                final Object msg2=
                    newCNSEW(sze(fFullText)>0?scrllpn(0L,new AsciiView(fFullText),dim(99,444)) : "Unable to convert pdf to txt",
                             vBpnl(hB("The PDF download directories are "),
                                   "<pre>"+joinAsStrg(dirs,"\n")+"</pre>",
                                   " ",
                                   hB("Associate PDF ",pdfLastModi," ",day(pdfLastModi.lastModified()))),
                             vBpnl(" ",hB("with ",monospc("PMID"+iPmid),title)),
                             null,null);
                if (ChMsg.yesNo(msg2)) {
                    final File fPdf=ChPubmed.getFile(toStrg(iPmid),PDF);
                    cpy(pdfLastModi,fPdf);
                    if(sze(fPdf)>0) {
                        if (isSelctd(togShared)) ChDelay.newThread(instance(),"CPY_PDF",fPdf).start();
                        handleActEvt(ChPubmed.class, ACTION_FULL_TEXT_ASSOCIATED, 0);
                        if (Insecure.canModify(pdfLastModi) && ChMsg.yesNo("Successfully associated.<br>Delete the downloaded file?<pre>"+pdfLastModi+"</pre>")) {
                            delFile(pdfLastModi);
                        }
                        mayUpdateShared(true);
                    }
                    return true;
                }
            } else error("Could  not find PDF in <pre>"+joinAsStrg(dirs,"\n")+"</pre>");
        }
        return false;
    }
    public static File maybeDownload(String w0) {
        final String w=delPfx("ALIGN:",delPfx("EXPR:",delPfx("LOC:",w0)));
        final File f=getFile(w,AUTO);
        if (f!=null && sze(f)==0) {
            final String sURL=Hyperrefs.toUrlString(w, w.startsWith("ENSG") ? Hyperrefs.WEB_LINK : Hyperrefs.PLAIN_TEXT|Hyperrefs.PROTEIN_FILE);
            final File fSrc=file(sURL);
            if (sze(fSrc)>0 && !fSrc.equals(f)) { renamFileOrCpy(fSrc,f);  }
            if (sze(f)==0) queueForDownload(w,false);
        }
        return sze(f)>0?f:null;
    }
    private static ChPubmed _inst;
    private static ChPubmed instance() { return _inst==null ? _inst=new ChPubmed() : _inst;}
    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        //if (cmd=="TEST") handleActEvt(ChPubmed.class, ACTION_FULL_TEXT_ASSOCIATED, 0);
        if (gcp(ACTION_FOCUS_LOST,q)=="SHARED" && (cmd==ACTION_ENTER || cmd==ACTION_FOCUS_LOST)) {
            if (file(_dirSharedPdf)!=file(_tfDirSharedPdf)) {
                _dirSharedPdf=null;
                mayUpdateShared(false);
            }
        }
        if (cmd!=null && cmd.startsWith("TITLE_")) {
            final int type= cmd=="TITLE_ABSTRACT" ? ABSTRACT : PDF;
            final File dir= cmd=="TITLE_PDF_SHARED" ?  dirSharedPdf() : getDir(type);
            threadPdfRenameUseTitle(dir,getFileSuffix(type)).start();
        }
        if (cmd=="DIR") {
            final String
                msg="<ul>"+
                "<li><b>WIKI:Pubmed abstracts</b> "+filepathUnixStyle(getDir(ABSTRACT))+" </li>"+
                "<li><b>WIKI:Gene_ontology</b> "+filepathUnixStyle(getDir(GO))+" </li>"+
                "<li><b>WIKI:Plain_text  Pdf2Txt</b> "+filepathUnixStyle(getDir(PDF2TXT))+" </li></ul>",
                m="<html><body> "+msg+" <ul>"+
                "<li><b>WIKI:PDF</b> "+filepathUnixStyle(getDir(PDF))+" </li>"+
                "<li><b>Shared PDF directory</b> "+dirSharedPdf()+" </i></li>"+
                "</ul></body></html>";
            new ChTextViewer(m).shw(0);
        }
        if (cmd=="PDF2TXT") {
            Thread t=mapThread.get(cmd);
            if (t==null) mapThread.put(cmd,t=ChDelay.newThread(instance(),cmd,q));
            t.start();
        }
        final File fPdf=pmidPopup!=null ? getFile(pmidPopup,PDF) : null;
        if (fPdf!=null) {
            if (cmd=="PDF_A") maybeAssociatePdf( atoi(fPdf.getName()),true);
            if (cmd=="PDF_0") {
                final File fBackup=file(fPdf+"_");
                delFile(fBackup);
                Insecure.renameToIgnoreSecurity(fPdf, fBackup);
                delFile(getFile(pmidPopup,PDF2TXT));
                handleActEvt(ChPubmed.class, ACTION_FULL_TEXT_ASSOCIATED, 0);
                final File fShared=file(dirSharedPdf()+"/"+fPdf.getName());
                if (sze(fShared)>0 && Insecure.canModify(fShared) && ChMsg.yesNo("Also delete <pre>"+fShared+"</pre>")) delFile(fShared);
                mayUpdateShared(false);
            }
            if (cmd=="PDF_V") {
                viewFile(sze(fPdf)>0 ? fPdf : file(dirSharedPdf(),fPdf.getName()), modifiers(ev));
            }
            pmidPopup=null;
        }
    }    /* <<< ActionListener <<< */
    /* ---------------------------------------- */
    /* >>> JPopupMenu >>> */
    private static JPopupMenu pmidMenu;
    private static String pmidPopup;
    private static ActionListener li;
    public static void showPubmedPopup(String pmid, JComponent location, int x, int y) {
        ChPubmed.pmidPopup=pmid;
        final ChTextField tfShared=new ChTextField(tfDirSharedPdf().getDocument()).cols(80,true,true);
        tfShared.li(instance());
        pcp(ACTION_FOCUS_LOST,"SHARED",tfShared);
        if (pmidMenu==null) {
            li=instance();
            final Object[] items={
                "A recently (1h) downloaded PDF can be associated to the Pubmed ID.",
                "The program needs to know the directory of downloaded  PDFs.",
                "-",
                new ChButton("PDF_A").li(li).t("Associate PDF ...").i(IC_PDF),
                new ChButton("PDF_V").li(li).t("View PDF ( Shift-Left-Click )"),
                new ChButton("PDF_0").li(li).t("Delete PDF").i(IC_KILL),
                "-",
                new Object[]{
                    "Utils",
                    bPdf2txt=new ChButton("PDF2TXT").t("Extract plain texts from PDFs ").li(li),
                    new ChButton("DIR").li(li).t("List important folders ...").i(systProprty(SYSP_ICON_FILE_BROWSER)),
                    "-",
                    new ChButton("TITLE_ABSTRACT")   .t("Abstract files   - Create nice file names using Pubmed title").li(li),
                    new ChButton("TITLE_PDF")        .t("PDF files        - Create nice file names using Pubmed title").li(li),
                    new ChButton("TITLE_PDF_SHARED") .t("Shared PDF files - Create nice file names using Pubmed title").li(li)
                },
                new Object[]{
                    "Customize",
                    ChButton.actionViewMessage("Change shared PDF directory...",tfShared),
                    Customize.newButton(Customize.databases, Customize.proteinDatabases, Customize.fileBrowsers, Customize.pdfViewer, Customize.webBrowser).t("Customize urls...")
                },
                new Object[]{
                    "Source code and licenses",
                    ChButton.actionOpenURL(URL_STRAP_JARS+"src").t("Java source codes"),
                    "For source code of STRAP download strap.jar from home page"
                },
                "-",
                helpBut(ChPubmed.class).t("Help")
            };
            pmidMenu=jPopupMenu("",  items);
        }
        pmidMenu.show(location, x,y);
    }
    /* <<< JPopupMenu <<< */
    /* ---------------------------------------- */
    /* >>> Collect PDF Links >>> */
    public static String viaTest(URL u) {
        final String via=getHeaderFld("Via",u);
        if (via==null) return null;
        final int b=nxt(LETTR,via);
        final int e=b>=0? via.indexOf(',',b) : -1;
        if  (e<0) return null;
        final String ret=via.substring(b,e);
        if (isMyComputer()) {
            wrte(file("~/var/log/http_field_via/"+rplcNoFilenameChars(u)),u+"\t"+via);
            //wrte(file("~/var/log/http_field_via/"+u.toString().replaceAll(RPLC_IN_FILENAME,"_")),u+"\t"+via);
        }
        return ret.indexOf(' ')<0 && ret.indexOf(':')<0 ? ret : null;
    }
    private static File collectPdfs(int pmid) {
        final File f=getFile(toStrg(pmid),LIST);
        final String oId=toStrg(pmid);
        final String urlPubmed=Hyperrefs.toUrlString("PMID:"+pmid,Hyperrefs.WEB_LINK);
        final ByteArray ba=sze(file(oId))>0?   readBytes(file(oId),null) :  readBytes(inStream(urlPubmed,0L));
        if (ba==null) { putln("Error in ChPubmed: cannot read "+urlPubmed); return null;}
        final byte[] T=ba.getBytes();
        final int E=ba.getEnd();
        int linkout=0;
        final HashSet listPDF=new HashSet();
        while(  (linkout=strstr("\"linkout-icon",T,linkout+1,E))>0) {
            int href=linkout;
            while(href>0 && !equls("href=\"", T,href)) href--;
            if (href<=0) continue;
            final int urlB=href+6;
            final int urlE=strchr('"',T,urlB+1,E);
            baTmp.setEnd(0);
            if (T[urlB]=='/') baTmp.append("http://www.ncbi.nlm.nih.gov");
            baTmp.append(T,urlB,urlE);
            try {
                final URL u=url(htmlDecod(baTmp));
                final ByteArray txtWithPdf=readBytes(u.openStream());
                if (txtWithPdf!=null) {
                    collectPdfs(txtWithPdf,u,listPDF,0,new boolean[1]);
                } else putln(" ChPubmed# cannot read "+u);
            } catch(Exception ex){putln(ex);}
        }
        collectPdfs(ba,url(urlPubmed),listPDF,999,new boolean[1]);
        baTmp.setEnd(0).append('\n');
        final Object ss[]= listPDF.toArray();
        for(Object o : ss) baTmp.append(o).append(ChTextComponents.PDF4PMID).append(pmid).append('\n');
        wrte(f,baTmp);
        return f;
    }
    private static void collectPdfs(ByteArray ba, URL url, HashSet result,int depth,boolean foundSdarticle[]) {
       final byte[] T=ba.getBytes();
        final int E=ba.getEnd();
        final  boolean isSpringer=strstr("title>SpringerLink",ba)>0;
        final boolean isPnas=strstr("content=\"http://www.pnas.org/",ba)>0;
        if (DEBUG && isMyComputer()) {
            //putln( " collectPdfs "+url+" isSpringer="+isSpringer);
            ByteArray baV=new ByteArray(E+9999).appendln(url).appendln("depth="+depth);
            try{
                final URLConnection c=url!=null ? url.openConnection() : null;
                if (c!=null) {
                    baV.appendln(c.getHeaderFields());
                    for(int i=0;i<1000;i++) {
                        String f=c.getHeaderField(i);
                        if (f==null && i>0) break;
                        baV.append(i).append(' ').appendln(f);
                    }
                }
            } catch(Exception ex) { }
            baV.append('\n').append('=',80).append('\n').append(ba);
            new AsciiView(baV).showInFrame(toStrg(url),0);
        }
        for(int search=2; --search>=0;) {
            int urlPdf=0;
            while(true) {
                urlPdf=search==0 ? strstr(STRSTR_AFTER,"href=",T,urlPdf+1,E) : strstr("http://",T,urlPdf+1,E);
                if (urlPdf<0) break;
                if (T[urlPdf]=='"') urlPdf++;
                final int gt=strchr('>',T,urlPdf,E);
                final int quote=strchr('"',T,urlPdf,gt);
                final int endAddr=quote>0 ? quote : gt;
                if (endAddr<0) continue;
                final int  debug=  strstr(STRSTR_IC,"pdf",T,0,E);
                if (DEBUG && debug>0) putln("urlPdf="+urlPdf+" quote="+quote+"  debug="+debug+"  "+strstr(STRSTR_IC,"pdf",T,urlPdf,endAddr+4));
                final int end=(T[endAddr-1]=='"' ? 1:0)+endAddr;
                final String lnk=htmlDecodAsStrg(ba.newString(urlPdf,end));
                if (  ( strstr(STRSTR_IC,"pdf",T,urlPdf,endAddr+4)+strstr(STRSTR_IC,"PDF",T,urlPdf,endAddr+4)>0 )  &&
                      lnk.indexOf("jpg")+lnk.indexOf("jpeg")+lnk.indexOf("png")+lnk.indexOf("gif")<-2) {
                    if (lnk.indexOf("sdarticle")>0) {
                        if (foundSdarticle[0])  break;
                        foundSdarticle[0]=true;
                    }
                    final String absoluteLink;
                    if (T[urlPdf]=='/') {
                        String host=viaTest(url);
                        if (isSpringer) host="http://www.springerlink.com/";
                        else if (isPnas) host="http://www.pnas.org/";
                        if (host==null) url.getHost();
                        absoluteLink="http://"+host+lnk;
                    } else absoluteLink=lnk;
                    if (DEBUG) putln(" lnk="+lnk);
                    if (absoluteLink!=null && absoluteLink.indexOf("/null/")<0) result.add(absoluteLink);
                } else  if (!foundSdarticle[0] && depth==0 &&  lnk.indexOf(".html")+lnk.indexOf(".htm")>0) {
                    if (lnk.startsWith("http://www.sciencedirect.com") ||
                        lnk.startsWith("http://journals.elsevierhealth.com/")
                        ) {
                        ByteArray ba2=readBytes(inStream(lnk,0L));
                        collectPdfs(ba2, url(lnk),result,depth+1,foundSdarticle);
                    }
                }
            }
       }
    }
    /* <<< Collect PDF Links <<< */
    /* ---------------------------------------- */
    /* >>> Title/Abstract >>> */
    private static ByteArray baTmp1, baTmp2;
    public final static long SHOWABSTRACT_FULL_TEXT=1<<1;
    public static String pubmedTitle(int pmid) {
        final ByteArray ba=readBytes(getFile(toStrg(pmid),ABSTRACT) ,baTmp.setEnd(0));
        if (ba==null) return null;
        ba.insidePreTags();
        //final byte[] T=ba.getBytes();
        //final int E=ba.getEnd();
        final int b=strstr(STRSTR_AFTER,"\nTI ", ba), e=b>=0 ? strchr('\n',ba) : -1;
        return e>0 ? ba.newString(b,e) : null;
    }
    public static void showAbstract(File f, ChTextComponents dest, long options) {
        final String noExt=withoutDotSuffix(f.getName());
        if (baTmp1==null) {
            baTmp1=new ByteArray(999);
            baTmp2=new ByteArray(999);
        } else {
            baTmp2.setEnd(0);
            baTmp1.setEnd(0);
        }
        if ( (options&SHOWABSTRACT_FULL_TEXT)!=0) {
            final File fPdf=ChPubmed.getFile(noExt,ChPubmed.PDF), fTxt=ChPubmed.getFile(noExt,ChPubmed.PDF2TXT);
            //putln("fPdf="+fPdf);
            if (sze(fPdf)>0 && (sze(fTxt)==0 || fPdf.lastModified()>fTxt.lastModified())) {
                //putln("SHOWABSTRACT_FULL_TEXT neu laden");
                new ChPdfUtilsPROXY().pdf2txt(fPdf,fTxt);
            }
            readBytes(fTxt,baTmp1);
        } else {
            if (noExt.startsWith("Summary_g_")) {
                readBytes(f ,baTmp1);
                final int from=strstr("<h2>Gene:",baTmp1);
                final int to=strstr("</p>",baTmp1,from,MAX_INT);
                final byte[] T=baTmp1.getBytes();
                int lt=0;
                if (from>0 && to>from) {
                    for(int i=from; i<to; i++) {
                        if (T[i]=='<') lt++;
                        else if (T[i]=='>') lt--;
                        else if (lt==0) baTmp2.append((char)T[i]);
                    }
                } else if (debug(" from="+from+" to="+to+"\n")) putln(baTmp1);
                baTmp2.replace("Source: UniProtKB/Swiss-Prot ","\nUNIPROT:")
                    .replace("\n      ","\n").replace("(EC ","\n").replace(")(", " ").replace("\n\n","\n").append('\n');
                final int uniprot=strstr("UNIPROT:",baTmp2);
                baTmp1.setEnd(0);
                if (uniprot>0) {
                    final int uniprotE=nxtE(SPC,baTmp2,uniprot,MAX_INT);
                    final String unip=baTmp2.newString(uniprot,uniprotE);
                    final File fU=ChPubmed.getFile(unip,ChPubmed.UNIPROT);
                    if (sze(fU)==0) ChPubmed.maybeDownload(unip);
                    readBytes( fU ,baTmp1);
                }
            } else {
                if (f.getName().endsWith(".M")) baTmp2.append(readBytes(ChPubmed.getFile(noExt, ChPubmed.LIST) ,baTmp1));
                readBytes(f,baTmp1.setEnd(0)).insidePreTags();
                if(strstr("not available</title>",baTmp1)>0) delFile(f);
                final int iTi=strstr("\nTI ", baTmp1);
                if (iTi>0) baTmp1.setBegin(iTi);
            }
        }
        if (!equls("<?xml ",baTmp1.getBytes(),baTmp1.getBegin())) {
            dest.setTextTS(baTmp2.append(baTmp1));
            final int ti=strstr(STRSTR_LF_L, "TI ",baTmp2);
            //putln(" ti="+ti+"  "+isEDT());
            ChTextComponents.setScrollPositionV(ti>0?ti:0,dest.jComponent());
        }
    }
    /* <<< Title/Abstract <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */
    private static boolean DEBUG;
    /*
    public static void main(String argv[]) throws Exception {
        Hyperrefs.setPubmedOption("holding=idecuhmlib_fft&");
        DEBUG=true;
        for(String pmid : argv)  {
            final File f=file(pmid);
            if (sze(f)>0)  collectPdfs(readBytes(f),url("http://proxy.charite.de"), new HashSet(),0,new boolean[1]);
            else {
                delFile(getFile(pmid,LIST));
                collectPdfs(atoi(pmid));
            }
        }
    }
    */
}
/*
http://eutils.ncbi.nlm.nih.gov/entrez/eutils/elink.fcgi?holdingXX=idecuhmlib_fft&dbfrom=pubmed&id=9298984&cmd=llinks
Dear Colleague,
You will want to use the &cmd=llinks or &cmd=llinkslib options in your
elink call. Set &dbfrom=pubmed and &id=pmid of the article.
See
http://eutils.ncbi.nlm.nih.gov/entrez/query/static/elink_help.html#LinkC
ommandValues
Example:
http://eutils.ncbi.nlm.nih.gov/entrez/eutils/elink.fcgi?dbfrom=pubmed&id
=18716402&cmd=llinks
Regards,
Eric
*/
