package charite.christo;
import java.util.*;
import java.lang.ref.*;
import charite.christo.interfaces.*;
import static charite.christo.ChUtils.*;
public class ChSet<E> implements Set<E>, HasModiCount {

    public final static long  WEAK_REFERENCE=1<<0, IDENTITY=1<<1;
    private int modiCount=0;
    public int getModificationCount() { return modiCount;}
    private Set<E> set;
    private Map<E,String> map;
    private final Class clazz;
    private final long options;
    public ChSet(long options, Class clazz) {
        this.options=options;
        this.clazz=clazz;
    }

    private Set<E> set() { return set==null ? set=new HashSet() : set; }

    private Set<E> theSet() { return 0!=((IDENTITY|WEAK_REFERENCE)&options) ? map().keySet() : set(); }

    private Map map() {
        if (map==null) map=0!=(options&IDENTITY) ? new IdentityHashMap() : new WeakHashMap();
        return map;
    }
    public Iterator iterator() {
        return theSet().iterator();
    }
    public int size() {
        return map==null && set==null ? 0 : theSet().size();
    }
    public boolean isEmpty() { return size()==0;}
    public boolean contains(Object o) {
        return map!=null ? map.containsKey(o) : set!=null ?set.contains(o) : false;
    }
    public boolean add(E o) {
        if (o==null) return false;
        final boolean changed= 0!=(options&WEAK_REFERENCE) ? null==map().put(o, "") : set().add(o);

        if (changed) modiCount++;
        return changed;
    }
    public boolean remove(Object o) {
        final boolean changed=map!=null ?  map.remove(o) == "" : set!=null ? set.remove(o) : false;
        if (changed) modiCount++;
        return changed;
    }
    public void clear() {
        boolean changed=false;
        if (map!=null) { changed=map.size()>0; map.clear(); }
        if (set!=null) { changed=set.size()>0; set.clear(); }
        if (changed) modiCount++;
    }
    public boolean removeAll(Collection<?> c) {
       final boolean changed=map!=null ? map.keySet().removeAll(c) : set!=null?set.removeAll(c) : false;
       if (changed) modiCount++;
       return changed;
    }
    public boolean retainAll(Collection<?> c) {
        boolean changed=map!=null ? map.keySet().retainAll(c) : set!=null ? set.retainAll(set) : false;
        if (changed) modiCount++;
        return changed;
    }
    public boolean addAll(Collection<? extends E> c) {
        if (c==null) return false;
        boolean changed=false;
        if (0!=(options&WEAK_REFERENCE)) {
            for(Object o : c.toArray()) {
                changed|=map().put((E)o,"")!="";
            }
        } else changed=set.addAll(c);
        if (changed) modiCount++;
        return changed;
    }
    public boolean containsAll(Collection<?> c) {
        return map!=null ? map.keySet().containsAll(c) : set!=null ? set.containsAll(c) : false;
    }
    public Object[] toArray() {
        return map!=null ? map.keySet().toArray() : set!=null ? set.toArray() : NO_OBJECT;
    }
    public <T> T[] toArray(T[] a) {
        return map!=null ? map.keySet().toArray(a) :  set().toArray(a);
    }

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Arrays >>> */
    private final static int C_ARRAYv=1;
    private Reference<Object[]> refCached;
    private int lastSize;
    private Object[] cached() {
        final Reference<Object[]> r=refCached;
        Object oo[]=r!=null ? r.get() : null;
        final int L=size();
        if (oo==null) refCached=new SoftReference(oo=new Object[2]);
        else if (this.lastSize!=L) for(int i=oo.length; --i>=0;) oo[i]=null;
        this.lastSize=L;
        return oo;
    }

    public E[] asArray() { return asArray(null,null);}
    public E[] asArray(ChRunnable run, String runID) {
        final Object cached[]=cached();
        E[] vv=(E[])cached[C_ARRAYv];
        if (vv==null) {
            cached[C_ARRAYv]=vv=(E[])toArray(emptyArray(clazz));
            if (run!=null) run.run(runID,vv);
        }
        return vv;
    }

}

