package charite.christo;
import charite.christo.interfaces.*;
import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.lang.ref.Reference;
import java.util.*;
import java.util.List;
import javax.swing.*;
import javax.swing.border.LineBorder;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;

/**
   A reimplementation of tabbed pane since JTabbedPane does not work very well
   @author Christoph Gille
*/
public class ChTabPane extends JPanel implements ActionListener,PaintHook,PreferredSize,Comparator  {
    public final static String ACTION_REMOVED="TP_A_RM", ACTION_ADDED="TP_A_ADD", ACTION_SELECTION_CHANGED="TP_A_SC",
        KEY_SORT="CTB$$S";
    private final Object KEY_WHEN=new Object(), KEY_TAB=new Object();

    private CollapseComponent collapseComponents;
    public ChTabPane enableCollapse() {
        collapseComponents=new CollapseComponent(completePanTabs,"EW").parent(ChTabPane.this).child(panTabs);
        return this;
    }
    private final ArrayList<JComponent> componentsV=new ArrayList();
    private final JComponent panTabs,pTabs;
    private JComponent completePanTabs;
    private transient Reference<JComponent>ref_pCenter;
    private String orientation;
				private final Container panUserButtons=new ChPanel();
				public void addUserButton(JComponent c) { panUserButtons.add(c);}
				public void removeUserButtons() { panUserButtons.removeAll();}
    private final List<Component> focusedV=new ArrayList();
    private JList jListHistory;
    private final ChButton butPrevious=new ChButton(null,ChButton.NO_FILL|ChButton.NO_BORDER).li(this).magnifyUnderMouse("<").tt("selects the previously focused dialog");

    public int compare(Object o1,Object o2) {
        final String k1=gcp(KEY_SORT,o1,String.class), k2=gcp(KEY_SORT,o2,String.class);
        final int w1=(atoi(gcp(KEY_WHEN,o1))<<16) | componentsV.indexOf(o1), w2=(atoi(gcp(KEY_WHEN,o2))<<16) | componentsV.indexOf(o2);
        final int comp=k1!=null && k2!=null  ? k1.compareTo(k2) : 0;
        return comp!=0 ? comp : w1<w2 ? -1 : 1;
    }
    public ChTabPane(String orientation) {
        this.orientation=orientation;
        setPrefSize(this, panTabs=new ChPanel());
        pTabs=new ChPanel();
        pTabs.setMinimumSize(dim(0,0));
        pTabs.add(panTabs);
        completePanTabs=pTabs;
        if (orientation=="W") {
            pTabs.setOpaque(false);
            //butPrevious.setOpaque(false);
            butPrevious.paintIfEnabled();
            butPrevious.setEnabled(false);
            completePanTabs=new ChPanel();
            addPaintHook(this,completePanTabs);
            enlargeFont(completePanTabs,2);
            completePanTabs.setLayout(new BorderLayout());
            completePanTabs.add(pTabs,BorderLayout.CENTER);
            setMinimumSize(dim(3,3));
        }
        JComponent east=null,west=null,north=null;
        if (orientation=="E") {
            east=completePanTabs;
        }
        if (orientation=="W") {
            west=completePanTabs;
            completePanTabs.add(panUserButtons,BorderLayout.NORTH);
            panUserButtons.add(butPrevious);
        }
        cnsew(this,null,north,null,east,west);
    }
    public void setSelectedComponent(JComponent o) {
        addTab(o);
    }
    public JComponent getSelectedComponent() {
        return (JComponent)deref(ref_pCenter);
    }
    public JComponent[] getTabComponents() { return toArry(componentsV,JComponent.class);}
    private String message;
    public ChTabPane message(String msg) { message=msg; return this;}
    public boolean paintHook(JComponent toBePainted,Graphics g1, boolean after) {
        final Graphics2D g=(Graphics2D)g1;
        if (after && toBePainted==completePanTabs) {
            final JComponent pan=(ChPanel)toBePainted;
            final Rectangle charBounds=charBounds(g.getFont());
            g.setColor(new Color(0,0,0,0.5f));
            final int h=pan.getHeight(),w=pan.getWidth();
            final int x=w/2-charBounds.height/2+charBounds.y,y=h;
            {
                final Paint gp=g.getPaint();
                g.setPaint(new GradientPaint(0,0,pan.getBackground(),w,h,pan.getBackground().darker().darker()));
                g.fillRect(0,0,w-2,h);
                g.setPaint(gp);
            }
            if (message!=null)
            {
                g.rotate(-Math.PI/2,x,y);
                g.drawString(message,x,y);
                g.rotate(Math.PI/2,x,y);
            }
        }
        return true;
    }
    public void removeAllTabs() {
        componentsV.clear();
        final Component pC=deref(ref_pCenter,Component.class);
        if (pC!=null && pC.getParent()==this) remove(pC);
        updateGUI();
        revalidate();
        repaint();
        handleActEvt(this,ACTION_REMOVED,0);

    }

   private void updateGUI() {
        panTabs.removeAll();
        panTabs.setLayout(new GridLayout(sze(componentsV),1));
        for(JComponent jc: getTabComponents())  {
            final ChButton tog=gcp(KEY_TAB,jc,ChButton.class);
            if (orientation=="E") {
                AbstractButton r=gcp("R",tog,ChButton.class);
                if (r==null) pcp("R",r=tog.radio(),tog);
                panTabs.add(r);
            } else  panTabs.add(tog);
        }
        ChDelay.revalidate(jListHistory,222,(long[])null);
        for(int i=sze(focusedV);--i>=0;) {
            if (!componentsV.contains(focusedV.get(i))) focusedV.remove(i);
        }
        for(JComponent c:getTabComponents()) {
            final ChButton tog=gcp(KEY_TAB,c, ChButton.class);
            if (tog==null) continue;
            tog.s(getSelectedComponent()==c);
            if (orientation!="E")tog.setBorder(new LineBorder(tog.s() ? Color.GREEN : Color.WHITE/*, isSelected ? 5 :2*/));
        }
        try {
            butPrevious.setEnabled(sze(focusedV)>3);
        } catch(Throwable ex){}
    }
    public void removeTab(JComponent... cc) {
        JComponent c0=null;
        for(JComponent c:cc) {
            if (c==null || !componentsV.contains(c)) continue;
            pcp(ChTabPane.class,null,c);
            pcp(KEY_TAB,null,c);
            c0=c;
        }
        previousTab(c0,true);
        handleActEvt(this,ACTION_REMOVED,0);
    }
    private void previousTab(Object c,boolean shouldRemove) {
        if (shouldRemove) componentsV.remove(c);
        if (sze(focusedV)>0 && (c==null || focusedV.get(0)==c)) {
            focusedV.remove(0);
        }
        if (c==null || c==getSelectedComponent()) {
            final JComponent newC=get(0, sze(focusedV)>0 ? focusedV : componentsV,JComponent.class);
            privateSelectedComponent(newC);
        }
        updateGUI();
        revalidate();
        repaint();
        butPrevious.repaint();
    }

    public void addTab(Component c) {
        if (c==null || c==getSelectedComponent()) {
            handleActEvt(this,ACTION_SELECTION_CHANGED,0);
            return;
        }
        if (gcp(KEY_WHEN,c)==null) pcp(KEY_WHEN,intObjct((int)(System.currentTimeMillis()&0x7fFFffFF)),c);
        final Object noDuplicates=gcp("NO_DUPLICATES",c);
        if (noDuplicates!=null) {
            for(int i=sze(componentsV); --i>=0;) {
                final JComponent jc=get(i,componentsV,JComponent.class);
                if (noDuplicates.equals(gcp("NO_DUPLICATES",jc))) {
                    privateSelectedComponent(jc);
                    return;
                }
            }
        }
        handleActEvt(this,ACTION_ADDED,0);
        adUniq(c,componentsV);
        if (gcp(KEY_TAB,c)==null) pcp(KEY_TAB,newTabButton(c),c);
        pcp(ChTabPane.class,weakRef(this),c);
        privateSelectedComponent(c);
        addMoliRefreshTip(c);
        final JComponent sorted[]=getTabComponents();
        sortArry(sorted,this);
        componentsV.clear();
        adAllUniq(sorted,componentsV);
        updateGUI();
        doLayout();
        revalidate();
        repaint();
        ChDelay.updateUI(this,2222,durationUpdateUI);
    }
    private long durationUpdateUI[]=new long[ARRAY_DURATION];
    private void privateSelectedComponent(Component c) {
        final JComponent center=getSelectedComponent();
        if (center!=null) remove(center);
        ref_pCenter=weakRef(c);
        if (c==null) return;
        add(c,BorderLayout.CENTER);
        if (center!=c)  handleActEvt(this,ACTION_SELECTION_CHANGED, 0);

        /*
           IllegalArgumentException:
           adding a container to a container on a different GraphicsDevicecause: nulljava.lang.IllegalArgumentException: adding a container to a container on a different GraphicsDevice
        */
        final ChButton tog=gcp(KEY_TAB,c,ChButton.class);
        if (tog!=null) tog.s(true);
        focusedV.remove(c);
        focusedV.add(0,c);
        revalidate();
        repaint();
        butPrevious.repaint();
    }
    private ChButton newTabButton(Component comp) {
        final String tab=TabItemTipIcon.tab(comp);
        final ChButton tog=toggl("TAB",ChButton.NO_MARGIN).li(this)
            .t(tab).i(TabItemTipIcon.icn(comp))
            .tt(TabItemTipIcon.tip(comp));
        tog.setHorizontalAlignment(AbstractButton.LEFT);
        pcp(JComponent.class,weakRef(comp),tog);
        pcp("IF_SELECTED_GREEN","",tog);
        if (collapseComponents!=null) collapseComponents.preventedBy(this);
        if (orientation=="W") setTabUndocksPanel(tog,comp);
        return tog;
    }

    public void updateTabLabels() {
        for(JComponent c:getTabComponents()) {
            final ChButton tog=gcp(KEY_TAB,c, ChButton.class);
            if (tog!=null) {
                tog.t(TabItemTipIcon.tab(c));
                // if (c instanceof ChTableLayout) tog.setEnabled(((ChTableLayout)c).getColumnCount()>0);
            }
        }
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> ActionListener >>> */
    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        if (cmd=="TAB") {
            {
                final JComponent pc=getSelectedComponent();
                if (pc!=null) remove(pc);
                final ChButton tog=gcp(KEY_TAB,pc,ChButton.class);
                if (tog!=null && tog!=q) tog.s(false);
            }
            addTab(gcp(JComponent.class,q,JComponent.class));
        }
        if (q==butPrevious) {
            if (isCtrl(ev)) {
                if (jListHistory==null) jListHistory=new ChJList(focusedV,ChJTable.ICON_ROW_HEIGHT|ChJTable.CLASS_RENDERER);
                ChFrame.frame("History",jListHistory,ChFrame.SCROLLPANE).shw(0);
            } else previousTab(null,false);
            ChDelay.revalidate(jListHistory,333,(long[])null);
        }
    }
    /* <<< ActionListener <<< */
    /* ---------------------------------------- */
    /* >>> getPreferredSize >>> */

    public Dimension getPreferredSize(JComponent component) {
        int width=ICON_HEIGHT, height=0;
        final Component cc[]=component.getComponents();
        for(Component c: cc) width=maxi(c.getPreferredSize().width,width);
        if (component==panTabs) height=mini(pTabs.getHeight()-ICON_HEIGHT,sze(componentsV)*ICON_HEIGHT);
        if (component==pTabs)   height=mini(cc.length*ICON_HEIGHT,getHeight()-butPrevious.getHeight());
        return dim(width,height);
    }

}
