package charite.christo;
import static charite.christo.ChUtils.*;
import charite.christo.interfaces.*;
import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableColumn;
import static charite.christo.ChConstants.*;
import static java.awt.event.MouseEvent.*;
/**
   Container that arranges components horizontally.
   A header allows dragging of columns and resizing like in a JTable.
   @author Christoph Gille
   // BasicTableHeaderUI
   */
public final class ChTableLayout extends ChPanel implements ChRunnable, ChProcessEvent, javax.swing.table.TableCellRenderer, TooltipProvider, PaintHook {
    public final static String COLUMN_ADDED="CTL$$CA", COLUMN_REMOVED="CTL$$CRM",
        KEY_OFFSET_COMPONENT="CTL$$OC", KEY_HEIGHT_SOUTH_PANEL="$CTL$$HSP";
    private final static Object KEY_ITEM=new Object();
    private final static int MIN_SIZE=EM, DIVIDER=EM, SQUEEZE=2*EM;
    private final ChEventAdapter LI=new ChEventAdapter(this);

    private final ChJTable table;
    private final long options;
    public ChTableLayout(long options) {
        this.options=options;
        setLayout(null);
        pcp(ChTableLayout.class,weakRef(this), this);
        addActLi(LI,table=new ChJTable(tm,options));
        table.setPreferredSize(dim(0,0));
        final JTableHeader tHeader=getTableHeader();
        if (0!=(options&ChJTable.CHJTABLE_FOCUSED_BLACK)) tHeader.setDefaultRenderer(this);
        tHeader.setPreferredSize(dim(9999,EX));
        add(tHeader);
        add(table);
        addComponentListener(LI);
        monospc(tHeader);
        tHeader.addMouseListener(LI);
        tHeader.addMouseMotionListener(LI);
        addTooltip(this, LABEL);
        tHeader.setResizingAllowed(true);
    }
    public ChJTable getJTable() { return table;}
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Event  >>> */
    private final javax.swing.border.Border H_BORDER=hBorder(), BORDER=hBorder();
    private javax.swing.border.Border hBorder() {
        try { return new javax.swing.plaf.metal.MetalBorders.TableHeaderBorder(); } catch(Throwable t){}
        return null;
    }
    private void recursiveAddMoli(Component c, java.lang.ref.Reference item) {
        if (c!=null) {
            c.addMouseListener(LI);
            c.addFocusListener(LI);
            pcp(KEY_ITEM,item,c);
        }
        if (c instanceof Container) {
            for(Component child : ((Container)c).getComponents()) recursiveAddMoli(child,item);
        }
    }
    private static int xButton(int type, int rw) {
        final int
            xClose=-EM-(rw>4*EM?EM:EM/2),
            xMaxim=xClose-EM-(rw>3*EM?EM/2:2);
        return type==ChRenderer.SMALL_CLOSE ? xClose : type==ChRenderer.SMALL_MAXIM ? xMaxim : 0;
    }

    public int getSelectedIndex() { return selectedColumn; }
    private boolean  isDragged, likeDivider, likeDividerWhenMoved, isResizingCol;
    private int resizingCol=-1;
    private int colUnderMouse=-1;
    public AWTEvent processEvent(JComponent c, AWTEvent ev) {
        if (needs_setPrefAsIs && !isDragged)setPrefAsIs();
        setMinMax();
        paintedOrEvent=true;
        final int id=ev.getID(), modi=modifiers(ev), x=x(ev), y=y(ev);
        final int nCol=sze(vComponents);
        final Object q=ev.getSource();
        final Component comp=q instanceof Component ? (Component)q : null;
        final String cmd=actionCommand(ev);
        final boolean ownRenderer=0!=(options&ChJTable.CHJTABLE_FOCUSED_BLACK);
        if (id==ActionEvent.ACTION_PERFORMED) {
            if (cmd==ChJTable.COLUMN_MARGIN_CHANGED ||cmd==ChJTable.COLUMN_MOVED) {
                ChDelay.runAfterMS(EDT,THREAD_LOCATION,99);
                THREAD_LOCATION.run();
            }
        }
        final JTableHeader tHeader=getTableHeader();
        final Object item;
        { Object i=gcp(KEY_ITEM,q);item=i!=null ? i:q;}
        if (q==tHeader) {
            final boolean headerCanResizeVertically=gcp(KEY_LOOKS_LIKE_DIVIDER, tHeader)!=null;
            Object removeColumn=null;
            Disposable[] toBeDisposed=null;
            final int hw=tHeader.getWidth(), h=tHeader.getHeight();
            if (id==MOUSE_EXITED || id==MOUSE_RELEASED) {
                tHeader.repaint();
                setPrefAsIs();
            }

            for(int col=getColumnCount(); --col>=0;) {
                final Object component=get(col,vComponents);
                final boolean disposable=hasDisposabl(component);
                final Rectangle rectCol=rectangleAt(col);
                if (rectCol==RECT_ERROR) continue;
                final boolean isInColumn=rectCol.x<x && x<rectCol.x+rectCol.width;
                if (isInColumn && (id==MOUSE_MOVED && col!=columnUnderMouse || id==MOUSE_ENTERED || id==MOUSE_CLICKED)) {
                    if (headerCanResizeVertically) looksLikeDivider(tHeader,null,0,0,hw-2*EM, EM);
                    for(int iType=2; --iType>=0;) {
                        int type=0;
                        switch(iType) {
                        case 0:
                            type=ChRenderer.SMALL_CLOSE;
                            if (!disposable) continue;
                            break;
                        case 1:
                            if (0==(options&ChJTable.CHJTABLE_SHOW_BUT_MAXIM) || nCol<2) continue;
                            type=ChRenderer.SMALL_MAXIM;
                            break;
                        }
                        final int xBut=rectCol.x+rectCol.width+xButton(type,rectCol.width);
                        if (xBut<0) continue;
                        if (!ownRenderer) {
                            ChRenderer.setSmallButtonX(type,xBut, q);
                            ChRenderer.drawSmallButtons(tHeader, null);
                        }
                        if (id==MOUSE_CLICKED && xBut<x && x<xBut+EM) {
                            if (type==ChRenderer.SMALL_CLOSE) {
                                toBeDisposed=disposabls(component);
                                removeColumn=component;
                            }
                            else narrowAllOthers(component, 3*EM);
                        }
                    }
                }
                if (id==MOUSE_EXITED && (headerCanResizeVertically || disposable)) tHeader.repaint();

                if (id==MOUSE_DRAGGED && q==tHeader && undockEnabld()==null && gcp(KEY_UNDOCKABLE,component)!=null ) {
                    final Component pnl=getPnl(component);
                    msgUndock[0]=ANSI_W_ON_B+"Hint: Panels can be dragged out, so-called undocking";
                    if (!(pnl instanceof ChPanel))  drawMsg(msgUndock,pnl,(Graphics)null,0);
                }
                if (isDragged
                    && id==MOUSE_RELEASED
                    && gcp(KEY_UNDOCKABLE,component)!=null
                    && col==columnUnderMouse
                    && !isResizingCol
                    && null==tHeader.getResizingColumn()
                    && (y<-EX || y>EX+h || x<-EX || x>EX+hw)) {
                    ChDelay.runAfterMS(EDT, thread_panelUndock(component, xScreen(ev)-EX, yScreen(ev)-EX/2), 333);
                }
                if (id==MOUSE_MOVED) {

                    if (rectCol.x+EM<x && x<rectCol.x+rectCol.width-3*EM && headerCanResizeVertically) tHeader.setCursor(getCursr('n'));
                    if (isInColumn && columnUnderMouse!=col) {
                        columnUnderMouse=col;
                        tHeader.repaint();
                    }
                }
            }
            if (toBeDisposed!=null) {
                ChDelay.runAfterMS(EDT,ChDelay.newThread(this,"DISPOSE",new Object[]{toBeDisposed, removeColumn, (modi&CTRL_MASK)==0?"":null}),99);
            }
        }
        /* >>> Focussed gets a Border >>> */
        final int col=item==null ? -1 : vComponents.indexOf(item);
        if (id==MOUSE_PRESSED || id==FocusEvent.FOCUS_GAINED) {
            if (col>=0 && this.selectedColumn!=col) {
                this.selectedColumn=col;
                if (ownRenderer) {
                    if (BORDER!=null) {
                        for(int i=nCol; --i>=0;) {
                            final JComponent pnl=deref(getPnl(get(i,vComponents)),JComponent.class);
                            if (pnl!=null) pnl.setBorder(i==selectedColumn?BORDER:null);
                        }
                    }
                }
                final JComponent pnlOffset=gcp(KEY_OFFSET_COMPONENT,q,JComponent.class);
                if (pnlOffset!=null && col>0) {
                    pnlOffset.setPreferredSize(dim(  (int)(EX*Math.log((col*8+1)))    ,0));
                    pnlOffset.revalidate();
                }
                tHeader.repaint();
            }
        }
        /* >>> Begin Divider >>> */
        final Component parent=comp!=null ? comp.getParent() : null;
        final JViewport viewPort=parent instanceof JViewport ? (JViewport)parent : null;
        int yMax=atoi(gcp(KEY_HEIGHT_SOUTH_PANEL,q));
        final int colConv=table.convertColumnIndexToView(col);
        if ( (yMax==0 || yMax>y) &&
             viewPort==null &&
             (!(comp instanceof JScrollPane)) &&
             (!(comp instanceof JViewport)) &&
              colConv>=0 && colConv<nCol-1 && comp!=null) {

            final Rectangle rectCol=rectangleAt(col);
            final int w=rectCol.width;
            if (yMax==0) yMax=comp.getHeight();
            final boolean resizeCursorOld=likeDivider || resizingCol==col;
            final int vpX=0; //viewPort!=null ? viewPort.getViewPosition().x : 0;
            likeDivider=w-DIVIDER<=x-vpX && x-vpX<=w;
            if (id==MOUSE_DRAGGED) {
                if (!isDragged && (likeDividerWhenMoved || likeDivider)) resizingCol=col;
            }
            if (id==MOUSE_MOVED) { resizingCol=-1; likeDividerWhenMoved=likeDivider;}
            if (resizingCol==col) {
                pcp(KEY_IS_RESIZING,"",comp);
                pcp(KEY_IS_RESIZING,"",q);
                chColSize(col,x ,w);
                table.validate();
            } else if (id==MOUSE_MOVED) {
                for(int iCol=getColumnCount(); --iCol>=0;) {
                    final Object component=get(iCol,vComponents);
                    pcp(KEY_IS_RESIZING,null,component);
                    pcp(KEY_IS_RESIZING,null,q);
                }
            }
            Rectangle rDiv=gcp(KEY_LOOKS_LIKE_DIVIDER,q,Rectangle.class);
            if (likeDivider && !isDragged && id!=MOUSE_EXITED) {
                if (rDiv==null) { rDiv=new Rectangle(); comp.repaint(); }
                rDiv.x=w-EM; rDiv.width=EM; rDiv.height=yMax;
                pcp(KEY_LOOKS_LIKE_DIVIDER,rDiv,comp);
                pcp(KEY_LOOKS_LIKE_DIVIDER,rDiv,item);
            } else {
                pcp(KEY_LOOKS_LIKE_DIVIDER,null,comp);
                pcp(KEY_LOOKS_LIKE_DIVIDER,null,item);
                if (rDiv!=null)comp.repaint();
            }
            if ( likeDivider || resizingCol>=0 ) {
                if (!(comp instanceof ChPanel)) looksLikeDivider(comp,null,w-EM,0,EM,yMax);
                comp.setCursor(getCursr('e'));
            } else if (resizeCursorOld ) comp.setCursor(null);
        } /* End Divider */

        if (id==ComponentEvent.COMPONENT_RESIZED) ChDelay.runAfterMS(EDT,THREAD_LOCATION,333);
        if (id==MOUSE_DRAGGED) isDragged=true;
        if (id==MOUSE_MOVED || id==MOUSE_RELEASED) {
            isDragged=isResizingCol=false;
            msgUndock[0]=null;
        }

        return ev;
    }

    /* <<< Event  <<< */
    /* ---------------------------------------- */
    /* >>> Header >>> */
    private final String[] msgUndock={null};
    private final static Rectangle RECT_ERROR=new Rectangle();
    public int columnWidth(int i) { return rectangleAt(i).width;}
    private final Rectangle rectangleAt(int i) {
        try {
            if (i>=0) return getTableHeader().getHeaderRect(table.convertColumnIndexToView(i));
        } catch(Throwable t){}
        return RECT_ERROR;
    }
    private int selectedColumn=-1;
    private final ChLabel LABEL=new ChLabel();
    {
        addPaintHook(this,LABEL);
        LABEL.setMinimumSize(dim(0,0));
    }
    private int labelCol,columnUnderMouse=-1;

    public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
        final boolean sel=column==this.selectedColumn;
        LABEL.bg(C(sel ? 0xffFFff : DEFAULT_BACKGROUND)).t(getTitleForObject(get(column,vComponents))).setBorder(H_BORDER);

        labelCol=column;
        return LABEL;
    }
    public String getToolTipText(JComponent c, AWTEvent ev) {
        if (c==LABEL) {
            final String t=LABEL.getText();
            if (strgWidth(LABEL,t)>columnWidth(labelCol)) return t;
        }
        return null;
    }
    public boolean paintHook(JComponent c, Graphics g, boolean after) {
        paintedOrEvent=true;
        if (after) {
            if (c==LABEL) {
                if ( table.convertColumnIndexToModel(labelCol)==columnUnderMouse && null==getTableHeader().getResizingColumn()) {
                    for(int iType=2; --iType>=0;) {
                        int type=0;
                        switch(iType) {
                        case 0:
                            type=ChRenderer.SMALL_CLOSE;
                            if (!hasDisposabl(get(labelCol,vComponents))) continue;
                            break;
                        case 1:
                            if (0==(options&ChJTable.CHJTABLE_SHOW_BUT_MAXIM) || sze(vComponents)<2) continue;
                            type=ChRenderer.SMALL_MAXIM;
                            break;
                        }
                        final int xBut=xButton(type,c.getWidth());
                        if (xBut<0) {
                            ChRenderer.setSmallButtonX(type,xBut,c);
                        }
                    }
                    ChRenderer.drawSmallButtons(c, g);
                }
            } else {
            final Rectangle divider=gcp(KEY_LOOKS_LIKE_DIVIDER,c,Rectangle.class);
            if (divider!=null && !isDragged) looksLikeDivider(c,g,divider.x, divider.y, divider.width, divider.height);
            if (msgUndock[0]!=null && c==getPnl(get(columnUnderMouse, vComponents))) drawMsg(msgUndock,c,g,0);
            }
        }
        return true;
    }
    public final static String RUN_COLUMN_TITLE="$CTL$$RCL";
    public static String getTitleForObject(Object o) {
        for(int pass=2; --pass>=0;) {
            final Object run= pass==0 ? o  : gcp(RUN_COLUMN_TITLE,o);
            if (run instanceof ChRunnable) {
                final String name= (String)((ChRunnable)run).run(RUN_COLUMN_TITLE,null);
                if (name!=null) return name;
            }
        }
        return TabItemTipIcon.item(o);
    }
    public JTableHeader getTableHeader() { return table.getTableHeader();}
    public void setTitle(int col, String title) {
        final TableColumn c=getColumn(col);
        if (c!=null) {
            c.setHeaderValue(title!=null ? title : "");
            getTableHeader().repaint();
        }
    }
    private TableColumn getColumn(int col0) {
        table.convertColumnIndexToView(-1);
        table.convertColumnIndexToView(100);
        if (col0<0 || col0>=getColumnCount()) return null;
        final int col=table.convertColumnIndexToView(col0);
        try { return table.getColumnModel().getColumn(col); } catch(Exception e) {return null;}
    }
    /* <<< Header  <<< */
    /* ---------------------------------------- */
    /* >>> add remove get Columns >>> */
    public void addColumn(Object p) {
        assertEDT();
        pcp(ChTableLayout.class,weakRef(this),p);
        final Component panel=getPnl(p);
        if (panel!=null) {
            pcp(ChTableLayout.class,weakRef(this),panel);
            if (0!=(options&ChJTable.CHJTABLE_FOCUSED_BLACK)) {
                panel.removeMouseListener(LI);
                panel.removeMouseMotionListener(LI);
                panel.addMouseListener(LI);
                panel.addMouseMotionListener(LI);
                addPaintHook(this,deref(panel,JComponent.class));
            }
            recursiveAddMoli(panel,weakRef(p));
        }
        if (adUniq(p,vComponents)) {
            final int n=sze(vComponents);
            if (RESIZABLE.length<=n) RESIZABLE=chSze(RESIZABLE,n+10);
            fixColumnModel();
            handleActEvt(this,COLUMN_ADDED,0);
        }
    }
    private void fixColumnModel() {
        table.createDefaultColumnsFromModel();
        layoutColumns();
        repaint();
        getJTable().tableChanged(new javax.swing.event.TableModelEvent(tm));

    }
    public void removeColumn(Object... pp) {
        assertEDT();
        if (pp==null) return;
        boolean changed=false;
        for(Object p: pp) {
            if (vComponents.remove(p)) {
                changed=true;
                if (p instanceof Component) ((Component)p).removeMouseListener(LI);
            }
        }
        if (changed) {
            fixColumnModel();
            handleActEvt(this,COLUMN_REMOVED,0);
        }

    }
    public final Object[] getAllComponents() { return vComponents.toArray();}
    private final ArrayList<Object> vComponents=new ArrayList();
    public void layoutColumns() {
        for(Component c : getComponents()) {
            if (c!=table && c!=getTableHeader()) remove(c);
        }
        for(Object o : vComponents.toArray()) {
            final Component c=getPnl(o);
            add(c!=null ? c :  LABEL.fg(C(0x0)).bg(C(0x0FFffFF)).t("error"));
        }
    }
    /* <<< add remove get Columns <<< */
    /* ---------------------------------------- */
    /* >>> Thread  >>> */
    private final Runnable THREAD_LOCATION=ChDelay.newThread(this,"SET_LOCATION",null);
    public Object run(String id, Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[])arg : null;
        if (id=="SET_LOCATION") {
            final int n=sze(vComponents);
            int iDragged=-1;
            final TableColumn col=getTableHeader().getDraggedColumn();
            if (col!=null)  iDragged=col.getModelIndex();
            final JTableHeader header=getTableHeader();
            final int height=getHeight(), width=getWidth();
            final int headerH=getTableHeader().getPreferredSize().height;
            header.setBounds(0,0,width,headerH);
            table.setBounds(0,headerH, width,1);
            for(int iC=0;iC<n;iC++) {
                final Component c=getPnl(get(iC,vComponents));
                if (c==null) continue;
                final Rectangle r=rectangleAt(iC);
                if (r==RECT_ERROR) continue;
                final int x= r.x + ( iC==iDragged ? header.getDraggedDistance() : 0 );
                c.setBounds(x,r.y+headerH, maxi(0,r.width-1),height-headerH);
            }
            try { table.revalidate(); revalidate();} catch(Exception e){}
            {
                Runnable[] run=derefArray(gcpClear("setColWidth",this,Object[].class),Runnable.class);
                if (sze(run)>0) ChDelay.runAfterMS(EDT, newThrd(run),333);
            }
        }
        if (id=="DISPOSE") {
            if (dispos(argv[2]!=null ? OPTION_DISPOSE_ASK_CTRL : 0L, argv[0]))
                removeColumn(argv[1]);
        }
        return null;
    }
    /* <<< Thread <<< */
    /* ---------------------------------------- */
    /* >>> TableModel >>> */
    private final javax.swing.table.TableModel tm=new javax.swing.table.DefaultTableModel() {
            @Override public int getRowCount(){ return 0;}
            @Override public int getColumnCount() { return ChTableLayout.this.getColumnCount(); }
        };
    public int getColumnCount() { return sze(vComponents); }
    /* <<< TableModel <<< */
    /* ---------------------------------------- */
    /* >>> Layout >>> */
    public void narrowAllOthers(Object c, int narrowWidth) {
        final int column=vComponents.indexOf(c), N=getColumnCount();
        final TableColumn tCol=getColumn(column);
        if (column<0 || column>=N || tCol==null || N<2) return;
        final int prefW=maxi(narrowWidth, getTableHeader().getWidth()-narrowWidth*(N-1));
        for(int col=N; --col>=0;) {
            final TableColumn tc=getColumn(col);
            if (tc==null) continue;
            tc.setMaxWidth(30000);
            tc.setMinWidth(MIN_SIZE);
            tc.setResizable(true);
            tc.setPreferredWidth(tc==tCol ? prefW : narrowWidth);
        }
    }
    private void chColSize(int column,int newWidth0, int oldWidth) {
        final TableColumn tCol=getColumn(column);
        final int N=getColumnCount();
        if (N<2 || tCol==null) return;
        final int newWidth=mini(newWidth0, getTableHeader().getWidth()-(N-1)*SQUEEZE);
        setMinMax();
        final int nW=maxi(newWidth,SQUEEZE);
        int reduce=nW-oldWidth;
        for(int col=0; col<N; col++) {
            final TableColumn tc=getColumn(col);
            if (tc==null) continue;
            tc.setMaxWidth(30000);
            tc.setMinWidth(MIN_SIZE);
            final int oldW=tc.getWidth(), prefW;
            if (col>column) {
                prefW=maxi(SQUEEZE,oldW-reduce);
                reduce-=(oldW-prefW);
            } else prefW=oldW;
            tc.setPreferredWidth(prefW);
        }
        tCol.setPreferredWidth(nW-reduce);
        if (0!= (reduce+=nW-newWidth)) {
            tCol.setPreferredWidth(nW);
            for(int col=column; --col>=0 && reduce!=0;) {
                final TableColumn tc=getColumn(col);
                if (tc==null) continue;
                final int oldW=tc.getWidth(), newW=maxi(SQUEEZE,oldW-reduce);
                reduce-=(oldW-newW);
                tc.setPreferredWidth(newW);
            }
        }
        needs_setPrefAsIs=true;
    }
    private boolean paintedOrEvent;
    public Thread thread_setColWidth(boolean hard, int col,int w0) {
        return threadM("setColWidth", this, new Object[]{boolObjct(hard), intObjct(col), intObjct(w0)});
    }
    public void setColWidth(boolean hard, int col,int w0) {
        if (gcp(KEY_NOT_YET_PAINTED,this)!=null) {
            addCP("setColWidth",thread_setColWidth(hard,col,w0),  this, 3);
            return;
        }
        assertEDT();
        final TableColumn tc=getColumn(col);
        final int width=getTableHeader().getWidth();
        final int w=paintedOrEvent ? maxi(MIN_SIZE,mini(width-EM,w0)) : w0;
        if (tc!=null && col<RESIZABLE.length) {
            setMinMax();
            RESIZABLE[col]=true;
            if (w>0) {
                if (hard) tc.setWidth(w);
                tc.setPreferredWidth(w);
                tc.setMaxWidth(w);
                tc.setMinWidth(w);
            }
            needs_setPrefAsIs=true;
        }
    }

    private boolean needs_setPrefAsIs;
    private void setPrefAsIs() {
        for(int col=getColumnCount(); --col>=0;) {
            final TableColumn tc=getColumn(col);
            if (tc==null) continue;
            tc.setPreferredWidth(tc.getWidth());
            RESIZABLE[col]=true;
        }
        needs_setPrefAsIs=false;
    }

    private void setMinMax() {
        final int N=getColumnCount();
        for(int col=N; --col>=0;) {
            final TableColumn tc=getColumn(col);
            if (tc==null) continue;
            tc.setMaxWidth(30000);
            tc.setMinWidth(MIN_SIZE);
            tc.setResizable(true);
        }
    }
    private boolean[] RESIZABLE=new boolean[33];

}
