package charite.christo;
import static charite.christo.ChUtils.*;
import java.util.*;
/**
   A fast Tokenizer
   @author Christoph Gille
*/
public class ChTokenizer  {
    {  assertInstAllowed(); }
    private byte[] text;
    private int end,pos, from,to;
    private boolean delimiters[];
    private String  asString;
    public int countTokens() {
        final int oldPos=pos;
        int count=0;
        while(nextToken()) count++;
        pos=oldPos;
        return count;
    }

    public ChTokenizer setText(ByteArray ba) { if (ba!=null) setText(ba.getBytes(), ba.getBegin(), ba.getEnd()); return this;}
    public ChTokenizer setText(byte bb[],int begin, int endPosition) { asString=null; from=to=-1; text=bb; pos=begin; end=endPosition; if (text!=null && text.length<end) end=text.length; return this; }
    public ChTokenizer setDelimiters(boolean[] charClass) {delimiters=charClass; return this;}

    public ChTokenizer trim() {
        if (from>=0 && to>=0 && text!=null) {
            to=prevE(NO_SPC,text,to-1,from-1)+1;
            from=nxtE(NO_SPC,text,from,to);
        }
        return this;
    }

    public boolean containsOnly(int charClass) {
        final int f=getFrom(), t=getTo();
        if (f<0 || t-f<1) return false;
        return cntainsOnly(charClass, this.text, f,t);
    }

    public boolean eq(String s) {
       final int f=getFrom(), t=getTo();
       return s!=null && f>=0 && t-f==s.length() &&  equls(s,text,f);
    }
    public int getFrom(){ return from;}
    public int getTo(){ return to;}
    public boolean nextToken() {
        if (delimiters==null) delimiters=charClass(SPC);
        asString=null;
        final byte[] T=this.text;
        if (T==null || pos<0) { this.from=this.to=0; return false; }
        final int end=mini(T.length, this.end);
        while(pos<end && T[pos]>=0 && delimiters[T[pos]]) pos++;
        if (pos>=end) { this.from=this.to=0; return false; }
        this.from=pos++;
        while(pos<end && !(T[pos]>=0 && delimiters[T[pos]]) ) pos++;
        this.to=pos;
        return this.from>=0 && this.from<pos && pos<=end;
    }
    public String nextAsString() { return nextToken() ? asString() : null;}
    public String nextAsStringTrim() {
        if (!nextToken()) return null;
        final byte[] T=this.text;
        final int F=getFrom(), E=getTo();
        final int f=nxtE(NO_SPC,T,F,E), e=prev(NO_SPC,T, E-1,f-1)+1;
        return T!=null && f>=0 && f<=e && e<=end ?  new String(T,f,e-f) : null;
    }
    public String asString() {
        String s=this.asString;
        if (s==null) s=this.asString=toStrg(this.text, getFrom(), getTo());
        return s;
    }
    public String asStringIP() { return toStrgIP(this.text, getFrom(), getTo()); }
    public byte[] asBytearray(byte[] buffer) {
        final byte[] t=this.text;
        final int f=getFrom(), e=getTo();
        if (t==null || f<0 || e<f) return NO_BYTE;
        final byte[] buf=buffer!=null ? buffer : new byte[e-f];
        Arrays.fill(buf,(byte)0);
        System.arraycopy(t,f,buf,0,mini(e-f,buf.length));
        return buf;
    }
    public int asInt() {
        final byte[] T=this.text;
        int f=nxt(NO_SPC,T,getFrom(),getTo()), t=prev(NO_SPC, T, getTo()-1, f-1)+1;
        if (f<0 || f>=t) return ChConstants.INT_NAN;
        if (T[f]=='+') f++;
        if ( f>=t || !is(DIGT_DASH, T, f) || f+1<t && !(cntainsOnly(DIGT,T, f+1, t))) return ChConstants.INT_NAN;
        return atoi(T,f);
    }

    public double asFloat() {
        final byte[] T=this.text;
        final int f=nxt(NO_SPC,T,getFrom(),getTo()), t=prev(NO_SPC, T, getTo()-1, f-1)+1;
        return f>=0 && isFloat(T,f,t) ? atof(T,f) : Double.NaN;
    }

    public byte asChar() {
        final int f=getFrom();
        final byte[] t=this.text;
        return t!=null && f>=0 ? t[f] : -1;
    }

}
