package charite.christo;
import java.io.*;
import java.util.*;
import java.awt.datatransfer.*;
import static charite.christo.ChUtils.*;
/**HELP

<div class="floatR">
    <table summary="Dag-and-Drop actions"    rules="groups" border="1"  >
        <CAPTION>Actions which are performed when objects are dropped on certain targets.</CAPTION>
      <colgroup>
        <col width="1*">
          <col width="1*">
            <col width="1*">
              <col width="1*">
                <col width="1*">
                  <col width="1*">
      </colgroup>
      <thead style="border-bottom: 3px solid black;" >
        <tr><th style="border-right: 3px solid black;"></th><th colspan=4 >D r o p p e d &nbsp;  O b j e c t s</th></tr>
        <tr>
          <th style="border-right: 3px solid black;"> Drop target</th>
          <th style="border-right: 2px solid gray;"> Proteins</th>
          <th style="border-right: 2px solid gray;"> Residue selections</th>
          <th style="border-right: 2px solid gray;"> Hetero groups</th>
          <th> Images</th>
        </tr>
      </thead>

      <tbody>
        <tr style="border-bottom: 2px solid gray;">
          <th style="border-right: 3px solid black;"> Proteins<br>&nbsp;</th>
          <td style="border-right: 2px solid gray;"></td>
          <td style="border-right: 2px solid gray;" nowrap> Copy selection to protein </td>
          <td style="border-right: 2px solid gray;" nowrap> Add heteros to protein </td>
          <td nowrap> Set icon image</td>
        </tr>

        <tr style="border-bottom: 2px solid gray;">
          <th style="border-right: 3px solid black;"> Residue selections <br>&nbsp;</th>
          <td style="border-right: 2px solid gray;"></td>
          <td style="border-right: 2px solid gray;"></td>
          <td style="border-right: 2px solid gray;"></td>
          <td nowrap> Set background image </td>
        </tr>

        <tr style="border-bottom: 2px solid gray;">
          <th style="border-right: 3px solid black;"> 3D-View<br>&nbsp;</th>
          <td style="border-right: 2px solid gray;" nowrap> Add 3D-structure to view</td>
          <td style="border-right: 2px solid gray;">Highlight atoms/amino acids in 3D</td>
          <td style="border-right: 2px solid gray;">Add  hetero structure to view</td>
          <td></td>
        </tr>

        <tr style="border-bottom: 2px solid gray;">
          <th style="border-right: 3px solid black;">Alignment panel<br>&nbsp;</th>
          <td style="border-right: 2px solid gray;" nowrap>Load or un-hide  protein</td>
          <td style="border-right: 2px solid gray;" nowrap>Copy selection to protein</td>
          <td style="border-right: 2px solid gray;" nowrap>Add heteros to protein</td>
          <td></td>
        </tr>

        <tr style="border-bottom: 2px solid gray;">
          <th style="border-right: 3px solid black;" nowrap>Tree-node "Hidden"<br>&nbsp;</th>
          <td style="border-right: 2px solid gray;" nowrap>Hide proteins</td>
          <td style="border-right: 2px solid gray;"></td>
          <td style="border-right: 2px solid gray;"></td>
          <td></td>
        </tr>

        <tr>
          <th style="border-right: 3px solid black;" nowrap> Desktop or file browser <br>&nbsp;</th>
          <td style="border-right: 2px solid gray;" nowrap> Copy protein file </td>
          <td style="border-right: 2px solid gray;"></td>
          <td style="border-right: 2px solid gray;" nowrap> Write PDB-file </td>
          <td></td>
        </tr>
      </tbody>
    </table>
        <br>
</div>
Objects are dragged with the mouse and dropped on a target (see WIKI:Drag_and_drop). The effect depends on the objects that are dragged and the
destination where they are dragged onto. On Windows and Linux the presumed operation is written at the bottom of the Strap frame.
The page   <a href="http://3d-alignment.eu/dragProteinLink.html">Dragging Web-links</a> contains examples of protein and alignment Web links that can be dragged.
*/

public class ChTransferable implements Transferable {
    public final static int FILE=1, OBJECT=2;
    private final Object[] data;
    private List<File> vFiles;
    private final String transferDataS;

    private final int options;
    public ChTransferable(Object[] alist, int options) {
        this.options=options;
        this.data = alist;
        final StringBuffer sb=new StringBuffer();
        for(Object o : data) {
            if (o instanceof File)
                sb.append(((File)o).toURI()).append("\r\n");
        }
        this.transferDataS=sb.toString();
        if ((options&FILE)!=0) {
            vFiles=new ArrayList();
            for(Object o : data) if (o instanceof File)  vFiles.add((File)o);
        }
    }

    public Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException {
        return
            flavor==DataFlavor.stringFlavor || flavor==flavor('U') ? transferDataS :
            flavor==DataFlavor.javaFileListFlavor ? vFiles :
            data;
    }

    private static DataFlavor _flavors[];
    private static DataFlavor flavor(char t) {
        if (_flavors==null) {
            _flavors=new DataFlavor['U'+1];
            try {_flavors['U']=new DataFlavor("text/uri-list;class=java.lang.String"); } catch(Exception e){ assert false; }
            try {_flavors['O']=new DataFlavor(Object.class,"Any"); } catch(Exception e){ assert false;}
        }
        return _flavors[t];
    }

    public DataFlavor[] getTransferDataFlavors() {
        return  (options&FILE)!=0 ? new DataFlavor[]{ DataFlavor.javaFileListFlavor, DataFlavor.stringFlavor, flavor('U'), flavor('O')} :
        new DataFlavor[]{flavor('O')};
    }
    public boolean isDataFlavorSupported(DataFlavor flavor) {
        for(DataFlavor f : getTransferDataFlavors()) if (f==flavor) return true;
        return false;
    }

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Static Utils  >>> */
    public static Object getData(String type, Transferable t) {
        final DataFlavor df=
            type=="S"  ? DataFlavor.stringFlavor:
            type=="URI"? flavor('U'):
            type=="FF" ? DataFlavor.javaFileListFlavor:
            type=="OO" ? flavor('O'):
            null;
        assertNotNull(df);
        if (t.isDataFlavorSupported(df)) {
            Object o=null;
            try { o=t.getTransferData(df); } catch(Exception e){ }
            if (o instanceof String && ((String)o).startsWith("Unsupported ")) {
                putln("ChTransferable  ",o);
                putln(" type=",type);
                o=null;
            }
            return o;
        }
        return null;
    }

    protected static File[] getFiles(Transferable t, int options) {
        List<File> vF=null;
        final List dataFileLst=(List)ChTransferable.getData("FF",t);
        final String dataUriLst=(String)ChTransferable.getData("URI",t);
        final String dataString=(String)ChTransferable.getData("S",t);
        final Object dataObject=ChTransferable.getData("OO",t);
        putln(" type=",t);
        putln(" dataFileLst=", dataFileLst);
        putln(" dataUriLst=", dataUriLst);
        putln(" dataString=", dataString);
        putln(" dataObject=", dataObject);
        for(String lines : new String[]{dataString , dataUriLst}) {
            if (lines==null) continue;
            for(String line :  lines.split("\n")) {
                String sData=line.trim();
                final File f=nxt(LETTR_DIGT,sData)>=0 ? file(sData) : null;
                if (vF!=null && vF.contains(f)) continue;
                (vF==null ? vF=new ArrayList() : vF).add(f);
            }
        }
        for(int i=sze(dataFileLst); --i>=0;) {
            final File f=get(i,dataFileLst,File.class);
            if (sze(f)==0 || f.isDirectory()) continue;
            if (vF!=null && vF.contains(f)) continue;
            (vF==null ? vF=new ArrayList() : vF).add(f);
        }
        putln(" ChTransferable#getFiles ",vF);
        return toArry(vF,File.class);
    }
    final static HashSet setAlreadyReported=new HashSet();
    public static String reportDataFlavors(Transferable t) {
        if (t==null) return null;
        final String key=t+" "+t.getTransferDataFlavors();
        putln("ChTransferable.reportDataFlavors ",key);
        if (setAlreadyReported.contains(key)) return null;
        setAlreadyReported.add(key);
        try {
            final StringBuffer sb=new StringBuffer();
            sb.append("\nDataFlavors of ").append(t);
            for(DataFlavor f : t.getTransferDataFlavors()) {
                sb.append("\n    ").append(shrtClasNam(f));
                sb.append("\n    getHumanPresentableName()=").append(f.getHumanPresentableName());
                sb.append("\n    toString()=").append(f);
                sb.append("\n    getTransferData()=");
                try { sb.append(shrtClasNam(t.getTransferData(f))).append(' ').append(t.getTransferData(f)); } catch(Throwable ex) {sb.append(ex);}
                sb.append('\n');

            }
            return sb.toString();
        } catch(Throwable ex) {
            stckTrc(ex);
            return stackTrceAsStrg(ex);
        }
    }
}
// http://java.sun.com/docs/books/tutorial/uiswing/misc/dnd.html

