package charite.christo;
import java.util.*;
import java.io.*;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;
import java.net.URL;
import java.util.jar.JarFile;
import java.util.jar.JarEntry;
import java.util.jar.JarOutputStream;
import java.util.jar.Pack200;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
   to extract jar packages
   @author Christoph Gille
*/
public class ChZip  {
    public final static long PACK_IF_NEWER=1<<1;
    private ChZip(){}
    public final static int X=0;
    public static boolean extractAll(File jf,File destDirectory) {
        try {
            return extractAll(new ZipFile(jf),destDirectory);
        } catch(Exception e) { putln(RED_CAUGHT_IN+"ChZip#extractAll ",e); return false;}
    }
    public static boolean extractAll(ZipFile jf,File destDirectory) {

        if (destDirectory==null || !destDirectory.isDirectory()) {
            error("Error in ChZip.extractAll: "+destDirectory+" is not a directory");
            return false;
        }
        try {
            final List<ZipEntry> vE=new ArrayList(1000);
            final Enumeration entries=jf.entries();
            while(entries.hasMoreElements()) vE.add((ZipEntry)entries.nextElement());
            String lastDir=null;
            long lastTime=0;
            int count=0;
            final int nE=sze(vE);
            for(int iE=0; iE<nE; iE++) {
                final ZipEntry e=vE.get(iE);
                // putln("e="+e);
                final File f=file(destDirectory,e.getName());
                if (e.isDirectory()) {
                    mkdrs(f);
                } else {
                    final File dir=f.getParentFile();
                    final String dirS=dir.toString();
                    if (lastDir==null || !lastDir.equals(dirS)) mkdrs(dir);
                    lastDir=dirS;
                    delFile(f);
                    cpy(jf.getInputStream(e),f);
                    final long time=System.currentTimeMillis();
                    if (time-lastTime>999) {
                        sayDownloading(null,ANSI_MAGENTA+"Extracted "+jf.getName()+" #"+count+"/"+nE);
                        lastTime=time;
                    }
                    count++;
                }
            }
            sayDownloading(null,ANSI_GREEN+"Extracted "+jf+" #"+count+" done");
        } catch(Exception e) {  putln("caught  ChZip#extractAll ", e); return false;}
        return true;
    }
    public static String[] extractAllFilesWithEnding(ZipFile zf, String sRoot, String fileExtension[], File dirDest,boolean createSubdirs) {
        final ArrayList<String> vEntries=new ArrayList();
        final Enumeration entries=zf.entries();
        while(entries.hasMoreElements()) {
            final ZipEntry e=(ZipEntry)entries.nextElement();
            final String sE=e.getName();
            if (sze(sRoot)==0 || sE.startsWith(sRoot)) {
                for(String ext:fileExtension) {
                    if(ext==null || sE.endsWith(ext)) {
                        adUniq(sE,vEntries);
                        final File f= file(dirDest+"/"+(createSubdirs ? sE : sE.substring(sE.lastIndexOf('/')+1)));
                        mkdrs(f.getParentFile());
                        if (f.lastModified()<e.getTime() || f.length()!=e.getSize()) {
                            InputStream is=null;
                            try {
                                cpy(zf.getInputStream(e),f);
                            }catch(Exception ex){ putln("caught#extractAllFilesWithEnding ", ex);}
                            closeStrm(is);
                        }
                    }
                }
            }
        }
        mkdrsErr(dirDest);
        return toArry(vEntries,String.class);
    }

    public static boolean checkIntegrityOfZipFile(File f) {
        final byte[] buffer=new byte[4096];
        try {
            final ZipFile zf=new ZipFile(f);
            final Enumeration entries=zf.entries();
            while(entries.hasMoreElements()) {
                final ZipEntry e=(ZipEntry)entries.nextElement();
                BufferedInputStream is=null;
                try {
                    is=new BufferedInputStream(zf.getInputStream(e));
                    while(is.read(buffer)!=-1);
                } catch(IOException ioex) { putln("ChZip.checkIntegrity ", e); return false;}
                closeStrm(is);
            }
        } catch(Exception e){putln("ChZip.checkIntegrity ",e); return false;}
        return true;
    }

    // pack200 -Hfalse tt.pack.gz S.jar; time unpack200 tt.pack.gz t.jar; ls -lh t.jar
    private static Set setAlreadyTried=new HashSet();
    public static File unpack200(File fPack, long options) {
        if (fPack==null) return null;
        if (fPack.isDirectory()) {
            StringBuffer sbError=null;
            for(File f:fPack.listFiles()) {
                final String fn=f.getAbsolutePath();
                if (!fn.endsWith(".jar.pack.gz") && !fn.endsWith(".jar.pack.bz2")) continue;
                if (new File(delSfx(".pack.gz",fn)).length()>0) continue;
                final String key=f.getAbsolutePath()+f.lastModified();
                if (setAlreadyTried.contains(key)) continue;
                setAlreadyTried.add(key);
                final File fJar=unpack200(f, options);
                if (fJar==null) {
                    if (sbError==null) sbError=new StringBuffer(999);
                    sbError.append("unpack200 "+f+" failed\n");
                }
            }
            if (sbError!=null) error(sbError);
            return null;
        }
        final String fnPack=fPack.getAbsolutePath();
        final File fJar=new File(delSfx(".pack.gz",fnPack));
        try {
            JarOutputStream jostream = new JarOutputStream(fileOutStream(fJar));
            Pack200.newUnpacker().unpack(deflaterAccordingToExtension(new FileInputStream(fPack),fnPack), jostream);
            // Must explicitly close the output.
            jostream.close();
            return fJar;
        } catch (IOException ioe) {
            stckTrc(ioe);
            return null;
        }
    }

    private JarFile jarFile;
    private String classFileEntry;
    private static List<String> vCN;

    private  static ChZip instance;

    protected static String fullCN(String shortOrLongClassName, String packages[]) {
        final ChZip ju=ChZip.classFile(shortOrLongClassName,packages);
        return ju!=null && ju.classFileEntry!=null ? delSfx(".class",ju.classFileEntry).replace('/','.') : null;
    }

    private static ChZip classFile(String shortOrLongClassName, String packages[]) {
        if (shortOrLongClassName==null) return null;
        final String classFile=shortOrLongClassName.replace('.','/')+".class";
        if (vCN==null) vCN=new ArrayList(); else vCN.clear();
        vCN.add(classFile);
        if (CLASSLOADER.getResource(classFile)!=null) {
            ChZip ret=new ChZip();
            ret.classFileEntry=classFile;
            return ret;
        }
        vCN.add(classFile);
        if (packages!=null) {
            for(String p : packages) {
                if (p==null) continue;
                if (p.endsWith(shortOrLongClassName) && p.endsWith("."+shortOrLongClassName)) {
                    ChZip ret=new ChZip();
                    ret.classFileEntry=p;
                    return ret;
                }
                p=p.replace('.','/');
                final String full = p+classFile;
                vCN.add( full);
                if (CLASSLOADER.getResource(full)!=null) {
                    ChZip ret=new ChZip();
                    ret.classFileEntry=full;
                    return ret;
                }
            }
        }
        for(JarFile jf : jarFilesJF()) {
            for(Object o:vCN.toArray()) {
                if (o==null) continue;
                String classFileEntry=o.toString();
                final JarEntry je=jf.getJarEntry(classFileEntry);
                if (je!=null) {
                    final ChZip ret=new ChZip();
                    ret.classFileEntry=classFileEntry;
                    ret.jarFile=jf;
                    return ret;
                }
            }
        }
        vCN.clear();
        return null;
    }

    protected static Object javaSource(String shortOrLongClassName, String packages[],boolean returnStream) {
        final ChZip ju=ChZip.classFile(shortOrLongClassName,packages);
        if (ju==null) return null;
        final String javaFile=delSfx(".class",ju.classFileEntry)+".java";
        if (ju.jarFile==null) return returnStream ? CLASSLOADER.getResourceAsStream(javaFile) :
            ((CLASSLOADER.getResource(javaFile)!=null) ? Boolean.TRUE:null);
        final JarEntry je=ju.jarFile.getJarEntry(javaFile);
        if (je!=null) {
            try {
                return ju.jarFile.getInputStream(je);
            } catch(Exception e){}
        }
        return null;
    }

    private static long lastFileHash;
    private static File[] jarFiles_;
    private static int jarFiles_modificationCount;
    public static int jarFiles_modificationCount() {return jarFiles_modificationCount;}
    public static File[] jarFiles() {
        long sumModi=0;
        for(int i=0;i<2;i++) {
            final File dir=i==0 ? dirJars() : dirPlugins();
            sumModi+=dir.lastModified();
            if (isWindows()) {
                final File ff[]=dir.listFiles();
                if (ff!=null) {
                    for(File f : ff) {
                        final String s=f.toString();
                        if (s.endsWith(".jar")) {
                            sumModi+=f.lastModified()+ ( (long)s.hashCode() << 32);
                        }
                    }
                }
            }
        }
        if (lastFileHash!=sumModi) {
            lastFileHash=sumModi;
            final List v=new ArrayList();
            for(int i=0; i<2; i++) {
                final File dir=i==0 ? dirJars() : dirPlugins();
                final File ff[]=dir.listFiles();
                if (ff!=null) for(File f : ff) if (f.toString().endsWith(".jar")) v.add(f);
            }
            jarFiles_=toArry(v,File.class);
            jarFiles_modificationCount++;
            jarFilesJF_=null;
            jarFilesURL_=null;
            for(int i=classesDownloadDelete.length; --i>=0;) classesDownloadDelete[i]=null;
        }
        if (jarFiles_!=null) {
            if (instance==null) instance=new ChZip();
            sortArry(jarFiles_,comparator(COMPARE_NAME));
        }
        return jarFiles_!=null ? jarFiles_ : NO_FILE;
    }

    private static URL[] jarFilesURL_;
    public static URL[] jarFilesURL() {
        final File ff[]=jarFiles();
        if (jarFilesURL_==null) {
            final URL uu[]=new URL[ff.length];
            for(int i=ff.length; --i>=0;) {
                try{
                    uu[i]=ff[i].toURL();
                } catch(Exception e){}
            }
            jarFilesURL_=uu;
        }
        return jarFilesURL_;
    }

    private static JarFile[] jarFilesJF_;
    public static JarFile[] jarFilesJF() {
        final File ff[]=jarFiles();
        if (jarFilesJF_==null) {
            ArrayList<JarFile> v=new ArrayList();
            for(int i=ff.length; --i>=0;) {
                try{
                    v.add(new JarFile(ff[i]));
                } catch(Exception e){}
            }
            jarFilesJF_=toArry(v, JarFile.class);
        }
        return jarFilesJF_;
    }

    private final static String[][] classesDownloadDelete=new String['Z'][];
    public static String[][] classesDownloadDelete() {
        final JarFile jarFiles[]=jarFilesJF();
        for(char i=0;i<128;i++) {
            final String fn=
                i=='C' ? "classes.txt" :
                i=='L' ? "download.txt" :
                i=='D' ? "delete.txt" :
                i=='A' ? "author.txt" :
                null;

            if (fn==null) continue;
            if (classesDownloadDelete[i]==null) {
                final List<String> v=new ArrayList();
                for(JarFile jf:jarFiles) {
                    final JarEntry je=jf.getJarEntry(fn);
                    if (je==null) continue;
                    try {
                        for(String s : readLines(jf.getInputStream(je))) {
                            final String line=withoutLastComponent(s,'#').trim();
                            if (i=='D'||i=='L') {
                                for(String t:splitTokns(0L, line)) {
                                    if (i=='L') {
                                        final String  n=withoutLastComponent(jf.getName(),'.');
                                        v.add(t+"?"+n.substring(maxi(n.lastIndexOf('/'),n.lastIndexOf('\\'))+1));
                                    } else v.add(t);
                                }
                            } else if (!"".equals(line)) v.add(line);
                        }
                    } catch(Exception e){}
                }
                classesDownloadDelete[i]=toStrgArry(v);
            }
        }
        return classesDownloadDelete;
    }
    public static boolean copyZip(File fFrom, File fTo,Collection<String> skip,boolean compressed) {
        final java.util.zip.CRC32 crc32 =new java.util.zip.CRC32();
        final ZipFile zf;
        try{
            zf=new ZipFile(fFrom);
        } catch(Exception e){putln(" caught in copyZip "+fFrom); return false;}
        final ZipOutputStream zOut;
        try{
            zOut=new ZipOutputStream(fileOutStream(fTo));
        } catch(Exception e){ stckTrc(e); return false;}

        final Enumeration<? extends ZipEntry> entries=zf.entries();
        ByteArray BUFFER=new ByteArray(0);
        while(entries.hasMoreElements()) {
            final ZipEntry eIn=entries.nextElement();
            final String name=eIn.getName();
            if (skip==null || !skip.contains(name)) {
                try {
                    ByteArray ba=readBytes(new BufferedInputStream(zf.getInputStream(eIn)),BUFFER);
                    final ZipEntry e=new ZipEntry(name);
                    if (!compressed) {
                        crc32.reset();
                        crc32.update(ba.getBytes(),0,ba.getEnd());
                        e.setMethod(ZipEntry.STORED);
                        e.setCrc(crc32.getValue());
                        e.setSize(ba.getEnd());
                    }
                    zOut.putNextEntry(e);
                    wrte(zOut,ba);
                } catch(Exception ex){stckTrc(ex);}
            }
        }
        closeStrm(zOut);
        return true;
    }

    public static boolean supportForPack200() { return javaVsn()>14;}
    // java charite.christo.ChZip  -unpack200 file.jar.pack.gz
    /* *********** */
    /*  unpack200  */
    /* *********** */
    public static void main(String argv[])  {
        if ("-unpack200".equals(get(0,argv))) {
            for(int i=1; i<argv.length; i++) {
                final File f=file(argv[i]);
                if (argv[i].endsWith(".pack.gz")) {
                    ChZip.unpack200(f,PACK_IF_NEWER);
                }
            }
        }
    }

}
