package charite.christo;
import charite.christo.protein.DNA_Util;
import charite.christo.interfaces.*;
import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.util.*;
import javax.swing.*;
import java.lang.ref.Reference;
import static charite.christo.ChUtils.*;
import static charite.christo.ChTextComponents.Highlight;
import static charite.christo.ChConstants.*;
/**
   views html and plain text
   @author Christoph Gille
*/
public class DialogFindInText implements Disposable,ActionListener{
    //  protected final static Color COLORS[]={CYAN,YELLOW,PINK,RED,GREEN,LIGHT_GRAY,MAGENTA,BLUE};
    protected final static Color COLORS[]={
        new Color(0x8000ffff,true),
        new Color(0x80ffff00,true),
        new Color(0x80ff0000,true),
        new Color(0x8000ff00,true),
        new Color(0x80c0c0c0,true),
        new Color(0x800000ff,true),
        new Color(0x80ff00ff,true)
    };
    private boolean noAutoActivation[]=new boolean[COLORS.length];
    public final static Object KEY_ID="DialogFindInText";
    private final static int N=COLORS.length;
    private final Reference<JComponent> myComponent;
    private ChTextField[] textField;
    private final ChTextField[] labCount=new ChTextField[N],tfDelim={new ChTextField(" \\t\\r\\n"),new ChTextField(" \\t\\r\\n")};
    private ChButton[] bNext=new ChButton[N], bPrev=new ChButton[N], bFrst=new ChButton[N], bLast=new ChButton[N];
    private int posMarchingAnts, antsWidth=-1;
    private ChFrame frame;
    private final Highlight highlight[][]=new Highlight[N][7];

    private final ChButton
        CB_SKIP_BLANKS=toggl("Ignore space and dash").li(this).tt("i.e. \"website\" is found in \"web site\" because the space is ignored"),
        CB_IGNORE_LEN_OF_BLANK=toggl("Ignore length of white space").li(this).s(true),
        CB_DOT=toggl("Dots match any").li(this).s(true),
        CB_CASE=toggl("Case sensitive").li(this),
        CB_RevCompl=toggl("DNA-RevCompl").li(this),
        CB_FWD=toggl("DNA-Forward").li(this),
        CB_Delim=toggl("").li(this);

    {
        final String tt="Entering a nucleotide sequence as search String also searches for the translated  AA sequences of the 3 reading frames";
        CB_FWD.tt(tt);
        CB_RevCompl.tt(tt);
    }
    public DialogFindInText(JComponent tc) {  myComponent=weakRef(tc);  }
    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final Object tc=deref(myComponent);
        final ChTextComponents tools=ChTextComponents.tools(tc);
        final boolean isDelim=CB_Delim.s();
        tfDelim[0].setEnabled(isDelim);
        tfDelim[1].setEnabled(isDelim);
        boolean enableDisable=false;
        for (int row=N; --row>=0;) {
            final ChButton tog=butSearch[row];
            if (q==textField[row] && !noAutoActivation[row] && !tog.s() && 1<sze(textField[row].getText())) { tog.s(noAutoActivation[row]=true);}
            if (q==tog) noAutoActivation[row]=true;
            if (q==tog || tog.s() && ( q==textField[row] || q==CB_DOT || q==CB_CASE || q==CB_SKIP_BLANKS || q==CB_IGNORE_LEN_OF_BLANK || q==CB_FWD || q==CB_RevCompl || q==CB_Delim || q==CB_IGNORE_LEN_OF_BLANK || q==tfDelim[0]||q==tfDelim[1] )) {
                highlight(row);
            }
            if (q==bLast[row] || q==bFrst[row] || q==bNext[row] || q==bPrev[row]) {
                final int pos=q==bLast[row] ? MAX_INT : q==bFrst[row] ? -1 :  posMarchingAnts;
                int thisFrom=-1, thisTo=-1, thisIdx=-1;
                for(Highlight  h : highlight[row]) {
                    if (h==null)  continue;
                    final boolean fwd=q==bFrst[row] || q==bNext[row];
                    final int fromTo[]=h.fromTo();
                    /* otherwise, (-(insertion point) - 1) */
                    final int idx0=Arrays.binarySearch(fromTo, pos + (fwd?2*maxi(antsWidth, 1):-2));
                    final int idx= (idx0<0 ? maxi(0, -idx0-(fwd?1:2)) : idx0)&~1;
                    if (idx<fromTo.length) {
                        thisFrom=fromTo[idx];
                        thisTo=fromTo[idx|1];
                        thisIdx=idx;
                    }
                }
                if (thisFrom>=0 && thisTo>=0) {
                    labCount[row].t((1+thisIdx)+"/"+count(row));
                    tools.marchingAnts(ANTS_SEARCH, posMarchingAnts=thisFrom, thisTo);
                    this.antsWidth=thisTo-thisFrom;
                }
                enableDisable=false;
            }
        }
        if (enableDisable) enableDisable();
    }

    protected void setPosition(int pos) { posMarchingAnts=pos; antsWidth=-1; }
    private void highlight(int row) {
        final Object tc=deref(myComponent);
        final ChTextComponents tools=ChTextComponents.tools(tc);
        if (tools==null) return;
        final Highlight[] hh=highlight[row];
        for(int j=hh.length; --j>=0;) {
            tools.removeHighlight(hh[j]);
            hh[j]=null;
        }
        if (butSearch[row].s()) {
            final String needle=textField[row].getText();
            if (needle!=null && needle.length()==0) {
                for(int i=0;i<2;i++) {
                    final Object jc=i==0? tc : tools.verticalScrollBar();
                    if (jc instanceof JComponent) ((JComponent)jc).repaint();
                }
            } else if (needle!=null) {
                final long mode=(CB_CASE.s()?0:HIGHLIGHT_IGNORE_CASE) | (CB_SKIP_BLANKS.s()?HIGHLIGHT_SKIP_BLANKS:0) | (CB_IGNORE_LEN_OF_BLANK.s()?HIGHLIGHT_IGNORE_LEN_OF_BLANK:0) | (CB_DOT.s()?HIGHLIGHT_DOT_MATCHES_ANY:0);
                final String dL=CB_Delim.s() ? rplcToStrg("\\t","\t", rplcToStrg("\\r","\r", rplcToStrg("\\n","\n", tfDelim[0].getText()))) : "";
                final String dR=CB_Delim.s() ? rplcToStrg("\\t","\t", rplcToStrg("\\r","\r", rplcToStrg("\\n","\n", tfDelim[1].getText()))) : "";
                hh[0]=tools.highlightOccurrence(needle, charClass(dL) , charClass(dR),mode|HIGHLIGHT_REPAINT|HIGHLIGHT_UPDATE_IF_TEXT_CHANGES, COLORS[row]);
                if (CB_DOT.s() || CB_SKIP_BLANKS.s()) {
                    final String trim=needle.trim();
                    if (nxt(NO_ACTGUN,trim)<0) {
                        final byte[] seq=trim.getBytes();
                        final boolean isTranslated[]=new boolean[seq.length];
                        final long m=HIGHLIGHT_IGNORE_CASE|HIGHLIGHT_SKIP_BLANKS|HIGHLIGHT_UPDATE_IF_TEXT_CHANGES;
                        for(int frame=3; --frame>=0;) {
                            for(int j=seq.length; --j>=0;) isTranslated[j]= (j>=frame);
                            if (CB_FWD.s()) hh[1+frame]=tools.highlightOccurrence(DNA_Util.translateTriplet2aaCompact(seq,isTranslated), null,null,m, COLORS[row]);
                            if (CB_RevCompl.s()) {
                                final byte[] revCompl=DNA_Util.reverse(DNA_Util.complement(seq,MAX_INT),MAX_INT);
                                hh[1+3+frame]=tools.highlightOccurrence(DNA_Util.translateTriplet2aaCompact(revCompl,isTranslated), null,null,m, COLORS[row]);
                            }
                        }
                    }
                }
            }
            enableDisable();
        }

        tools.repaintLater(0);
    }
    private ChButton butSearch[]=new ChButton[N];
    public void showInFrame() {
        final Object tc=deref(myComponent);
        if (tc==null) return;
        final String title=(String) gcp(KEY_ID,tc);
        final boolean isAA=gcp(KEY_CONTAINS_AA,tc)!=null;
        if (gcp(AsciiView.KEY_BLAST_RESULT,tc)!=null) CB_SKIP_BLANKS.s(true);

        if (frame==null) {
            textField=new ChTextField[N];
            final JComponent
                panel=pnl(new GridLayout(N,1)),
                pSouth=vB(
                                 hB(CB_SKIP_BLANKS.cb(), CB_IGNORE_LEN_OF_BLANK.cb()),
                                 hB(CB_CASE.cb(),CB_DOT.cb(),
                                    isAA?pnl(CB_FWD.cb(),CB_RevCompl.cb()):hB("#TB Left and right delimiter characters", CB_Delim.cb(), tfDelim[0], tfDelim[1]))
                                 );

            frame=new ChFrame("Find").ad(newCNSEW(panel,null,pSouth,null,null));
            for(int i=0; i<N;i++) {
                ChTextField tf=textField[i]=new ChTextField("", UNDERLINE_REFS_DISABLE).cols(30,false,false).li(this);
                final ChTextComponents tools=tf.tools();
                tools.saveInFile("DialogFindInText_"+title +"_"+i);
                if (i==N-1) pcp(KEY_IF_EMPTY,"Enter search String here",tf);
                closeOnKeystroke(tf,new Object[]{frame}, CLOSE_ON_CtrlW);
                butSearch[i]=toggl("highlight", ChButton.NO_MARGIN).li(this).s("".equals(tf.toString()));
                (labCount[i]=new ChTextField().cols(7,false,false)).setEditable(false);
                bNext[i]=new ChButton(">",  ChButton.NO_MARGIN|ChButton.DISABLED).li(this);
                bPrev[i]=new ChButton("<",  ChButton.NO_MARGIN|ChButton.DISABLED).li(this);
                bFrst[i]=new ChButton("|<", ChButton.NO_MARGIN|ChButton.DISABLED).li(this);
                bLast[i]=new ChButton(">|", ChButton.NO_MARGIN|ChButton.DISABLED).li(this);
                panel.add(newCNSEW(tf,null,null,hB(bFrst[i],bPrev[i],monospc(labCount[i]),bNext[i],bLast[i]),butSearch[i].cb()));
                tf.tools().highlightBlanks();
            }
            for(ChTextField tf:tfDelim) {
                tf.tools().highlightBlanks();
                tf.li(this).tt("Enter character sets   left / right of the match.<br>The \\t denotes  the tabulator and  \\r and \\n denote different line separators.");
                tf.setEnabled(false);
                noMargn(tf);
            }
        }
        enableDisable();
        frame.shw(ChFrame.PACK|ChFrame.AT_CLICK);
        textField[COLORS.length-1].requestFocus();

    }
    private int count(int row) {
        int count=0;
        for(Highlight h : highlight[row]) if (h!=null) count+=h.fromTo().length;
        return count/2;
    }
    public void enableDisable() {
        int sum=0;
        for(int row=0;row<N;row++) {
            if (textField[row]==null) continue;
            final int c=count(row);
            labCount[row].t(c==0 ? "" : format10(c,8,true,strBuf(SB_TIP))).setBackground(COLORS[row]);
            bNext[row].enabled(c>0);
            bPrev[row].enabled(c>0);
            bFrst[row].enabled(c>0);
            bLast[row].enabled(c>0);
            sum+=c;
        }
        if (sum==0) setMarchingAnt(ANTS_SEARCH, deref(myComponent, JComponent.class),(Rectangle)null,(Color)null,(Color)null);
    }
    public void dispose() {
        final ChTextComponents tools=ChTextComponents.tools(myComponent);
        if (tools!=null) {
            boolean doRepaint=false;
            for(int i=0;i<N;i++)
                for(int j=highlight[i].length; --j>=0;)
                    doRepaint|=tools.removeHighlight(highlight[i][j]);
            if (doRepaint) tools.repaintLater(0);
        }
        if (frame!=null) frame.dispose();
        butSearch=null;
    }

}
