package charite.christo;
import charite.christo.interfaces.*;
import java.util.regex.*;
import java.awt.AWTEvent;
import java.util.*;
import javax.swing.JComponent;
import static java.awt.event.MouseEvent.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
/**HELP

   The user can chose from a number of shell scripts or Web addresses
   which can be applied to selected text regions.

   For convenience, they are arranged in a tree structure.

   The goal is to start other programs and tools within Metannogen or Strap.
   The customize dialog allows to register further shell-scripts at run-time.

    <div class="figure" >
    <table style="caption-side: bottom">
    <CAPTION><b>Figure</b> Dialog to apply shell scripts or  Web addresses to selected text.
    Each  node in the tree represents a shell script which can be started by left click.
    The text in the text box "Argument for shell scripts" replaces the placehoder in the script (asterisk). Users can define own scripts.
    </CAPTION>
    <tr><td><i>JCOMPONENT:ExecByRegex#screenshot()!</i></td></tr>
    </table>
    </div>
    
   <br><br>  
   Alternatively, the program option "-customizeAddScriptByRegex" can be used to specify a file with shell-scripts.
   Example:
   <pre class="data">
   demo\Example google%09.*%09http://www.google.com/search?q=*
   demo\Example: Matches starting with M%09^M.*$%09 echo starts with letter M *
   </pre>
   Each line associates a shell script (3rd column) to regular expression (2nd column).
   The asterisk is replaced by the text.
   The shell script is presented to the user as a menu item.
   The menu item path is given in the 1st column with back-slash being the separator.
   If the value in the 3rd column starts with http://, then it is not taken as a shell script but as a Web address.

   @author Christoph Gille
*/
public final class ExecByRegex implements ChProcessEvent, ChRunnable {

    private JComponent panel,  cbCygwin, cbDebug;
    private ChTextArea ta;
    private ChJTree jTree;
    private List vLines;
    private final Map<String,String> vSCRIPT=new HashMap();
    private ChLabel labError;
    private static ExecByRegex instances[];
    public static ExecByRegex getInstance(int i) {
        if (sze(instances)<=i) instances=ChUtils.chSze(instances, i+5,ExecByRegex.class);
        if (instances[i]==null) instances[i]=new ExecByRegex();
        return instances[i];
    }
    public ExecByRegex setText(String s) {
        getPanel();
        ta.t(s!=null ? s:"");
        if (jTree!=null) run("U",null);
        return this;
    }

    public JComponent getPanel() {
        if (panel==null) {
            addActLi(LI,ta=new ChTextArea(3,3).saveInFile("ta_ExecByRegex"));

            vLines=new UniqueList(String.class);
            final ChTreeModel tm=new ChTreeModel(0L,new Strings2Tree(Strings2Tree.LEAFS_CONTAIN_ENTIRE_PATH, vLines, '\\'));
            jTree=new ChJTree(0L,tm);
            jTree.setCellRenderer(new ChRenderer().options(ChRenderer.LAST_PATH_COMPONENT_SLASH));
            jTree.setRootVisible(false);
            jTree.setShowsRootHandles(true);
            jTree.addMouseListener(LI);
            pcp(ChJTree.KEY_NUM_EXPAND, "4",jTree);
            labError=new ChLabel().fg(C(0xFF0000));
            cbCygwin=isWindows() ? toggl("Cygwin-Shell").save(ExecByRegex.class,"cygwin").cb() : null;
            cbDebug=toggl("Debug").cb();
            final Object
                pNorthS=newCNSEW(ta,"\nArgument for shell script  ","Click on item to run shell scripts"," ", " "),
                pNorthEast=pnl(smallSourceBut(ExecByRegex.class),Customize.newButton(Customize.scriptByRegex).magnifyUnderMouse(IC_CUSTOM)),
                pNorth= newCNSEW("<h2>Execute shell script</h2>",null,pNorthS,pNorthEast,smallHelpBut(ExecByRegex.class)),
                pSouth=newCNSEW(labError, null, null, pnl(cbCygwin,cbDebug),null);
            this.panel=newCNSEW(scrllpn(0L,jTree,dim(1,4*EX)),pNorth,pSouth,null,null);
            addActLi(LI,Customize.customize(Customize.scriptByRegex));
            inEDT(ChDelay.newThread(this,"U",null));
            panel.setPreferredSize(dim(300,300));

        }
        return panel;
    }
    public Object run(String id, Object arg) {
        final Object argv=arg instanceof Object[] ? (Object[])arg: null;
        if (id=="U") {
            final String txt=ta.toString();
            final List v=this.vLines;
            v.clear();
            vSCRIPT.clear();
            String error="";
            for(String line : getCustomSettings(Customize.scriptByRegex)) {
                if (chrAt(0,line)=='#') continue;
                final String cols[]=line.split(line.indexOf('\t')>0?"\t":"%09");
                if (cols.length<3 || cols[0].length()*cols[2].length()==0 ) continue;
                if (cols[1].length()>0) {
                    Pattern p=null;
                    try { p=ChRegex.pattern(cols[1]); } catch(Exception ex) {}
                    if (p==null) { error="Error regular expr: "+cols[1]; continue; }
                    if (!p.matcher(txt).matches()) continue;
                }
                vSCRIPT.put(cols[0],cols[2]);
                TabItemTipIcon.set("", null ,cols[2],iicon(cols[2].startsWith("http://")?IC_WWW:IC_BATCH), cols[0]);
                v.add(cols[0]);
            }
            labError.t(error);
            jTree.updateKeepStateLater(0L, 99);
        }
        if (id=="EXE") {
            new ChExec((isSelctd(cbDebug) ? ChExec.SHOW_STREAMS : 0L)|ChExec.STDOUT_IN_TEXT_BOX).setCommandLineV(argv)
                .run();
        }
        return null;
    }
    /* <<< ChRunnable <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */

    final ChEventAdapter LI=new ChEventAdapter(this);
    public AWTEvent processEvent(JComponent c, AWTEvent ev) {
        final Object q=ev.getSource();
        final String cmd=actionCommand(ev);
        final int modi=modifiers(ev); //setEventLocationOnScreen(ev);
        final int id=ev.getID();
        final int x=x(ev), y=y(ev);
        final int kcode=keyCode(ev);

        final Object item=objectAt(ev);
        final String sel=item!=null ? vSCRIPT.get(item) : null;
        if (q==ta || q instanceof Customize ) {
            ChDelay.runAfterMS(EDT, this, "U", 99);
        }
        if (id==MOUSE_CLICKED && q==jTree && sel!=null) {                       
            final String txt=ta.toString();
            final String command0=rplcToStrg("*",txt,sel);
            if (sel.trim().startsWith("http://")) {
                visitURL(command0.trim(), modi|(isSelctd(cbDebug)? CTRL_MASK:0));
            } else {
                final String command[]={isSelctd(cbCygwin) ? "CYGWINSH" : "SH", sel!=command0 ? command0 : sel+txt};
                ChDelay.newThread(this,"EXE",command).start();
            }
        }
        return ev;
    }

    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> html >>> */
    public static JComponent screenshot() {
        ExecByRegex ex=new ExecByRegex();
        final JComponent jc=ex.getPanel();
        ex.jTree.expandAllNodes(null,ex.jTree);
        ex.jTree.expandAllNodes(null,ex.jTree);
        return jc;
    }
    /* <<< html <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */

   /*
      Java6 charite.christo.ExecByRegex
     */
    public static void main(String argv[]) throws Exception {
        ExecByRegex e=ExecByRegex.getInstance(0);
        new ChFrame("").ad(e.getPanel()).size(EM*80,EX*40).shw(ChFrame.TO_FRONT);
    }
}