package charite.christo;
import charite.christo.interfaces.*;

import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.lang.ref.*;
import java.text.DateFormat;
import java.util.*;
import java.net.URL;

import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static java.awt.event.MouseEvent.*;
import static java.awt.event.KeyEvent.*;
/**HELP

This list shows  currently downloaded files.<br>

Stalled downloads are recognized by a file size that does not change for a long time.<br>

They can block the program flow.

This view provides the possibility to cancel running downloads.
<br>
<br>
Sometimes it happens that a very large file is downloaded unintendedly by clicking the wrong hyper-link.
In this case cancel the download here.

@author Christoph Gille
*/
public class FilePopup  implements ActionListener,TooltipProvider/*,ChRunnable*/, ChProcessEvent {
    private  static FilePopup inst;
    public static FilePopup instance() {if (inst==null) inst=new FilePopup(); return inst;}
    public final static String DOUBLE_CLICK="DOUBLE_CLICK";
    private File file;
    private JPopupMenu popFile;
    private Reference<JList> ref;

    private final ChLabel LAB_TITLE=new ChLabel();
    //public void setFile(File f) { this.file=f;}
    public File getFile() { return file;}
    public JPopupMenu menu() {
        if (popFile==null) {
            final boolean my=false;//isMyComputer();
            Object oo[]={
                "? context-menu of files",null,
                LAB_TITLE,
                " "," ",
                new ChButton("C").li(this).t("Copy file path to clipboard"),
                new ChButton("X").li(this).t("Open with ..."),
                new ChButton("V").li(this).t("View text"),
                new ChButton("FB").li(this).t("File browser").i(systProprty(SYSP_ICON_FILE_BROWSER)),
                new ChButton("MV").li(this).t("Rename ..."),
                new ChButton("ED").li(this).t("Into text-editor").i(IC_EDIT),
                " ",
                new ChButton("CHAINS").li(this).t("Split multi chain PDB file ..."),
                new ChButton("NCBI").li(this).t("Split multipart NCBI file ...").i(IC_NCBI),
                " ",
                new ChButton("RM").li(this).t("File to ./trash/ ...").i(IC_KILL),
                " ",
                my? new ChButton("REV").li(this).t("revalidate") : null,
                my? new ChButton("FIRE").li(this).t("fireContentsChanged") : null,
                my? new ChButton("UI").li(this).t("updateUI") : null
            };
            popFile=jPopupMenu("file-contextmenu",oo);
        }
        return popFile;
    }
    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        final File ff[]=getFiles();
        JList jList=ref!=null ? ref.get():null;
        if (jList==null) return;
        if (cmd=="C") {
            final StringBuffer sb=new StringBuffer(999);
            for(int i=0;i<ff.length;i++) if (ff[i]!=null) sb.append(ff[i].toString()).append('\n');
            toClipbd(sb);
        }

        if (cmd=="X") {
            int count=0;
            String nn[]=new String[ff.length];
            for(int i=0;i<ff.length;i++)
                if (ff[i]!=null)
                    nn[count++]=ff[i].toString();
            execFile.show(nn);
        }
        if (cmd=="RM"){
            for(File f :ff) {
                File trashed=file("trash/"+f.getName());
                mkdrsErr(file("trash/"),"DELETED FILES");
                if (trashed.exists()) delFile(trashed);
                renamFile(f, trashed);
                jList.revalidate();
                ChDelay.repaint(jList,99);
            }
        }
        if (cmd=="MV") {
            for(File f :ff) {
                if (f==null) return;
                final String newName=input("Rename file","Enter new name for "+f.getName(),f.getName());
                final File fNew=file(newName!=null ? newName.trim() : null);
                if (fNew==null) return;
                if (fNew.exists()) { error("The file "+fNew+" already exist "); return; }
                if (newName.indexOf(' ')>=0) { error("The file name "+fNew+"  contains blanks "); return; }
                renamFile(f, newName);
                jList.revalidate();
            }
        }
        if (cmd=="REV") { jList.revalidate(); }
        if (cmd=="FIRE") { jList.revalidate(); }
        if (cmd=="UI")  {jList.updateUI();  }
        if (cmd=="CHAINS") new PDB_separateChains().interactiveProcess(ff);
        if (cmd=="NCBI") new NCBI_separateEntries().interactiveProcess(ff);
        if (q instanceof AbstractInteractiveFileProcessor) {
            final File ffCreated[]=((AbstractInteractiveFileProcessor)q).getCreatedFiles();
            if (ffCreated.length>0) new ChFrame("created files").ad(new ChJList(ffCreated,ChJList.OPTIONS_FILES)).shw(0);
        }
        if (cmd=="FB") {
            ArrayList<String> v=new ArrayList();
            for(File f :ff) adUniq(f.getParent(), v);
            for(String fn: toArry(v, String.class)) openInFilebrowser(file(fn), modifiers(ev));
        }
        if (cmd=="V") {
            for(File f :ff) {
                final AsciiView v=new AsciiView(f).showInFrame(toStrg(f),ChFrame.STAGGER);
                v.tools().underlineRefs(0);
                pcp(DialogFindInText.KEY_ID, "ProteinFile",v);
            }
        }
        if (cmd=="ED") {
            for(File f :ff) edFile(-1, f, modifiers(ev));
        }
        fileUnderMouse=null;
    }
    private static DialogExecuteOnFile execFile;
    private File[] selectedFiles(JList jList) {
        final Object oo[]=jList.getSelectedValues();
        if (oo==null) return NO_FILE;
        int count=0;
        final File ff[]=new File[oo.length];
        //for(int i=0;i<oo.length;i++)  if (oo[i] instanceof File) ff[count++]=(File)oo[i];

        for(Object o : oo) if (o instanceof File || o instanceof URL) ff[count++]=file(o);
        final File files[]=new File[count];
        System.arraycopy(ff,0,files,0,count);
        return files;
    }
    private File fileUnderMouse;
    public File[] getFiles() {
        final JList jList=ref!=null ? ref.get():null;
        File ff[]=jList!=null ? selectedFiles(jList) : NO_FILE;
        return fileUnderMouse!=null && idxOf(fileUnderMouse,ff)<0 ? new File[]{fileUnderMouse} : ff;
    }

    public AWTEvent processEvent(JComponent component, AWTEvent ev) {
        final Object q=ev.getSource();
        final boolean ctrl=isCtrl(ev), shift=isShift(ev);
        final int id=ev.getID();
        final int kcode=keyCode(ev);
        final File fMouse=toFile(ev);
        final JList jList=q instanceof JList ? (JList)q : null;
        if (id==MOUSE_CLICKED && fMouse!=null) {
            if (!shift && !ctrl && fMouse.isDirectory()) openInFilebrowser(fMouse, modifiers(ev));
            else if (clckCount(ev)>1) {
                this.file=fMouse;
                handleActEvt(this,DOUBLE_CLICK,0);
            }
        }
        if (isPopupTrigger(ev) && fMouse!=null) {
            this.fileUnderMouse=fMouse;
            final int n=getFiles().length;
            ref=weakRef(q);
            LAB_TITLE.setText("Menu for "+n+" file"+(n==1 ? "" : "s"));
            if (execFile==null) execFile=new DialogExecuteOnFile(Customize.customize(Customize.openFileWith));
            menu().show(jList,x(ev)+5,y(ev)+5);
            setPrpty(File.class,"popupShown","");
        }
        if (id==KEY_PRESSED && (kcode==VK_BACK_SPACE || kcode==VK_DELETE)) {
            final File[] ff=selectedFiles(jList);
            if (sze(ff)>0 && ChMsg.yesNo("Delete "+sze(ff)+" files?")) {
                for(File f:ff) delFile(f);
                ChDelay.repaint(q,99);
            }
        }
        return ev;
    }

    private String input(String title,String msg,String old) {
        final JList jList=ref!=null ? ref.get():null;
        if (jList==null) return null;
        return (String)JOptionPane.showInputDialog(jList.getTopLevelAncestor(),msg,title,JOptionPane.QUESTION_MESSAGE,null,null,old);
    }

    private static File lastF;
    private static String tip;

    public String getToolTipText(JComponent component, AWTEvent ev) {
        File f=toFile(ev);
        if (tip==null || f!=lastF) {
            lastF=f;
            if (f==null)  tip="";
            else if (f.isDirectory()) tip="Directory";
            else tip=format10(sze(f),10,true,strBuf(SB_TIP).append("<html><body>"))
                     .append(" bytes <br>changed: ")
                     .append(day(f.lastModified()))
                     .append(getPrpty(File.class,"popupShown",null)==null?"<br><font color=#FF4455><b>Right-click  to open context menu</b></font>":"")
                     .append("</body></html>").toString();
        }
        return tip;
    }

}

