package charite.christo;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
   Selections of Residues (sors) can refere to URLs or database entries.
   Database IDs are translated into URLs.
   @author Christoph Gille
*/
//http://www.expasy.org/expasy_urls.html
//http://www.chem.qmul.ac.uk/iubmb/enzyme/EC1/1/1/3.html

public class Hyperrefs  {
    private Hyperrefs(){}
    public final static String PAIR_ALIGNMENT="PAIR_ALIGNMENT:";
    public final static int  WEB_LINK=1, PLAIN_TEXT=2, PROTEIN_FILE=4;
    private final static int COUNT=3, modiCount[]=new int[(1<<COUNT)+1];
    private final static String[][][] databases=new String[(1<<COUNT)+1][][];

    private static String[] keysWithBrace,keysToHighlight;
    private static String pubmedOption;

    private final static List<String> vTmpK=new ArrayList(333),vTmpU=new ArrayList(333);
    public static void setPubmedOption(String s) { pubmedOption=s;}     // holding=idecuhmlib_fft&

    public static boolean hasNumericID(String w) {
        final char c=chrAt(0,w);
        return ( ((c|32)=='p' && w.startsWith("PMID") || w.startsWith("PUBMED") || w.startsWith("pmid")) || c=='G' && w.startsWith("GO:")) && w.indexOf('{')<0;
    }

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Databases >>> */
    public static int getModificationCount() {
        int mc=0;
        for(int i=0;i<COUNT; i++) {
            final Customize c=Customize.customize(1<<i);
            c.getSettings();
            mc+=c.getModificationCount();
        }
        return mc;
    }
    public static synchronized String[] getKeysWithBrace() {
        final String[][] key_url=getDatabases(WEB_LINK);
        if (keysWithBrace==null) {
            vTmpK.clear();
            for(String key:key_url[0]) {
                if (key!=null && key.endsWith("{")) vTmpK.add(key);
            }
            keysWithBrace=toArry(vTmpK,String.class);
        }
        return keysWithBrace;
    }
    public static synchronized String[] getKeysToHighlight() {
        final String[][] key_url=getDatabases(WEB_LINK|PLAIN_TEXT|PROTEIN_FILE);
        if (keysToHighlight==null) {
            vTmpK.clear();
            for(String key: key_url[0]) {
                if (!vTmpK.contains(key)) {
                    vTmpK.add(key);
                    if ("EMBL:".equals(key) || "PIR:".equals(key) || "DIP:".equals(key) ||  "PDB:".equals(key) ) vTmpK.add(" "+delSfx(":",key)+"; ");
                }
            }
            vTmpK.add("ftp://");
            vTmpK.add("http://");
            vTmpK.add(PAIR_ALIGNMENT);
            keysToHighlight=toArry(vTmpK,String.class);
        }
        return keysToHighlight;
    }
    public static synchronized String[][] getDatabases(int id) {
        boolean isSingle=false;
        for(int i=0;i<COUNT; i++) isSingle|= (id==1<<i);
        if (isSingle) {
            final Customize cust=Customize.customize(id);
            final String lines[]=cust.getSettings();
            final int mc=cust.getModificationCount();
            if (databases[id]==null || modiCount[id]!=mc) {
                modiCount[id]=mc;
                keysToHighlight=keysWithBrace=keysToHighlight=null;
                for(int i=databases.length; --i>=0;) if ( (i&id)!=0) databases[i]=null;
                vTmpK.clear();
                vTmpU.clear();
                for(String line:lines) {
                    if (chrAt(0,line)=='#') continue;
                    final String[] tokens= splitTokns(0L, line);
                    final int nT=sze(tokens);
                    if (nT>1) {
                        for(int i=0;i<nT-1;i++) {
                            vTmpK.add(tokens[i]);
                            vTmpU.add(tokens[nT-1]);
                        }
                    } //else if (nT!=0) putln("Hyperrefs#getDatabases Error: line has only one token ",line);
                }
                final int n=sze(vTmpK);
                final String[][] key_url=new String[2][n];
                for(int i=0;i<n;i++) {
                    key_url[0][i]=vTmpK.get(i);
                    key_url[1][i]=vTmpU.get(i);
                }
                databases[id]=key_url;
            }
        } else {
            for(int i=0;i<COUNT; i++) getDatabases(1<<i);
            if (databases[id]==null) {
                vTmpK.clear();
                vTmpU.clear();
                for(int i=0;i<COUNT; i++) {
                    if ( ( (1<<i) & id )==0) continue;
                    adAll(getDatabases(1<<i)[0],vTmpK);
                    adAll(getDatabases(1<<i)[1],vTmpU);
                }
                databases[id]=new String[][]{toArry(vTmpK,String.class),toArry(vTmpU,String.class)};

            }
        }
        return databases[id];
    }
    /* <<< getDatabases  <<< */
    /* ---------------------------------------- */
    /* >>> toUrlString >>> */

    public static boolean isHyperref(byte[] T, int pos,int end0) {
        final int end=mini(sze(T),end0),  L=end-pos;
        if (end==0 || L<4) return false;
        final byte t0=T[pos];
        if (L>10 && (t0|32)=='u' && equls(STRSTR_IC,"uniref",T,pos) && get(nxt(NO_DIGT,T,pos+6,end),T)==':') return true;
        nextKey:
        for(String key : getKeysToHighlight()) {
            final int len=sze(key);
            if (len==0) continue;
            final char k0=key.charAt(0), kz=key.charAt(len-1);
            if (k0!=t0 || pos+len>=T.length) continue;
            if (L<len+4 && kz!='{' && !(kz=='e' && key.endsWith("database"))) continue;
            for(int k=len;--k>=0;)  if (key.charAt(k)!=T[pos+k]) continue nextKey;
            if (T.length>pos+len && t0=='P' && T[pos+1]=='D' && T[pos+2]=='B') {
                if (!is(LETTR_DIGT,T[pos+len])) continue;
            }
            //if (t0=='P' && k0=='P' && "PMID".equals(key)) ChTextComponents.pubmedHeadline(new String(T,pos,end-pos));
            return true;
        }
        return false;
    }
    public static String toUrlString(String url_or_DBreference, int mode) {
        if (url_or_DBreference==null) return null;
        //putln(" toUrlString ",url_or_DBreference);
        final String trim=url_or_DBreference.trim();
        final int l_trim=trim.length();
        if (l_trim<5) return null;
        final char c0=trim.charAt(0);
        final String txt;
        if (c0=='P' && url_or_DBreference.startsWith(PAIR_ALIGNMENT))
            return toStrg(url(file(dirTmp()+"/sequFeature/align/"+delPfx(STRAPTMP,url_or_DBreference.replace(':','_'))+".txt")));
        if (trim.length()>10 && (c0|32)=='u' && equls(STRSTR_IC,"uniref",trim)) {
            final int colon=nxt(NO_DIGT,trim,6,MAX_INT);
            if (chrAt(colon,trim)!=':') return null;
            txt= (equls(STRSTR_IC,"UPI",trim,colon+1) ? "UNIPARC" : "UNIPROT") + url_or_DBreference.substring(colon);
        } else txt=trim;
        if ((c0=='h' || c0=='f') && looks(LIKE_EXTURL,txt) ) {
            final URL u=url(txt);
            if (u!=null) return u.toString();
        }
        //final boolean debug="SpiceBrowser".equals(url_or_DBreference);
        final int L=txt.length();
        final String[] dd[]=getDatabases(mode), keys=dd[0], urls=dd[1];
        for(int iK=0; iK<keys.length; iK++) {
            final String key=keys[iK];
            if (key==null) continue;
            String url=urls[iK];
            final int lKey=key.length();
            if (lKey==0) continue;
            final char k0=key.charAt(0), kz=key.charAt(lKey-1);
            //if (debug) putln(iK+" "+key+"  k0="+k0+"  kz="+kz+" "+(c0==k0 && txt.startsWith(key)));
            if (c0==k0 && txt.startsWith(key) && ( L>=lKey+1  || kz=='*' ||  /*L>=lKey &&*/ kz=='{' || (kz=='e' && key.endsWith("database"))) ){
                final char cz=txt.charAt(L-1);
                final String v0=delSfx(",",txt.substring(lKey,cz=='}'?L-1:L));
                String v= kz=='{'  || v0.indexOf(':')>0 ? new ByteArray(99).urlEncode(v0.trim()).newString() : v0;
                if (k0=='P' && "PDB:".equals(key)){
                    if (v.length()>4)  v= delPfx("pdb",v);
                    if (v.length()>4)  v= v.substring(0,4);
                }
                if (k0=='H' && key.equals("HSSP:") || k0=='F' && key.equals("FSSP:")) new AnonymousFTPPROXY().proxyObject();

                if (pubmedOption!=null)  url=rplcToStrg("entrez/query.fcgi?","entrez/query.fcgi?"+pubmedOption,url);
                if (hasNumericID(txt)) v=toStrg(  atoi(txt, nxt(DIGT,txt)));
                {
                    final int iRplc=url.indexOf("REPLACE_CHAR./");
                    if (iRplc>0) { url=url.substring(0,iRplc); v=v.replace('.','/'); }
                }
                final String rplc=rplcToStrg("*",v,url);
                String sUrl=rplc!=url?rplc:url+v;
                final String OPT=";output=embl;strand=feature;st=similarity;st=repeat;st=genscan;st=contig;st=variation;st=marker;st=gene;st=vegagene;st=estgene;_format=Text;";
                sUrl=rplcToStrg("---ENSGOPTIONS---", mode==PROTEIN_FILE ? "/Gene/Export?db=core"+OPT : "/Gene/Summary?",sUrl);
                sUrl=rplcToStrg("---VEGAOPTIONS---", mode==PROTEIN_FILE ? "/Gene/Export?db=vega"+OPT : "/Gene/Summary?",sUrl);
                sUrl=delSfx(",",sUrl);
                if (k0=='W' && "WIKI:".equals(key) && txt.indexOf('(')<0) sUrl=delSfx(")",sUrl);
                //putln(" toUrlString=",sUrl);
                return sUrl;
            }
        }
        if (url_or_DBreference.startsWith("ENSEMBL:")) return  toUrlString(url_or_DBreference.substring(8),mode);
        return null;
    }
    /* <<< toUrlString <<< */
    /* ---------------------------------------- */
    /* >>> toDbColonID >>> */
     public static String toDbColonID(byte[] txt, int from, int to) {
         final String url=new String(txt,from, mini(txt.length,to)-from);
         return toDbColonID(url);
     }

   public final static int idxAW(String haystack, String... needles) {
        for(String s : needles) {
            final int idx=iaw(s,haystack);
            if (idx>0) return idx;
        }
        return -1;
    }
    public static String toDbColonID(String mayBeUrl) {
        final String
            u=(mayBeUrl),
            lc=u.toLowerCase();
        final int
            equals1=u.indexOf('='),
            equals=u.lastIndexOf('='),
            slash=u.lastIndexOf('/'), quest=u.indexOf('?',slash+1), L=u.length(),
            id=maxi(ia("list_uids=",u),ia("&id=",u),ia("?id=",u));
        String pf=null;
        int idx, iUniprot=-1;

        if (u.indexOf("strap")>0 && u.indexOf("load=")>0) {
            for(String key : getDatabases(PROTEIN_FILE)[0]) {
                if ((idx=iaw("load="+key,u))>0) {
                    pf=key+wordAt(idx,u,LETTR_DIGT_US);
                    break;
                }
            }
        }
        if (pf!=null) {}
        else if (u.indexOf(".php?align=")>0) pf=urlDecode(u.substring(equals+1,e(equals,u)));
        else if (u.indexOf(".ensembl.org")>0 && u.indexOf("/Align")<0) {
            for(int i=u.length()-5; --i>=0;) {
                if (u.charAt(i)=='=' && is(UPPR,u,i+1)) {
                    pf=u.substring(i+1,e(i,u));
                    break;
                }
            }
        } else if (u.indexOf("genome.jp/dbget-bin/")>0 || u.indexOf("genome.ad.jp/dbget-bin/")>0 ) {
            int colon=-1;
            {
                int i=quest+2;
                for(; i<L; i++) if (u.charAt(i)=='+' && is(LETTR_DIGT_US,u,i-1) && is(LETTR_DIGT_US, u,i-2)) { colon=i; break;}
            }
            if (colon<0) colon=u.indexOf(':',quest);
            if (colon>0)  pf="KEGG_AA:"+u.substring(prevE(NO_LETTR_DIGT_US, u,colon-1,-1)+1,colon)+"+"+u.substring(colon+1,nxtE(NO_LETTR_DIGT_US, u,colon+1,L));
        }
        else if (
                 (idx=idxAW(lc,"http://www.rcsb.org/pdb/files/","entryid=pdb:","pdbid=","db=structure&orig_db=structure&term=",
                            "pdb=","pdb_code=","pdbcode=","get-pdb.pl?","/pdbsum/",
                            "pfam.sanger.ac.uk/structure?id=","macmol.pl?filename="))>0
                 || iaw("PDBsend",u)>0 &&  (idx=iaw("code=",u))>0
                 || iaw("pdbid_search",lc)>0 && (idx=iaw("query=",lc))>0
                 || iaw("entry_pdb.sh",lc)>0 && (idx=iaw("entry=",lc))>0
                 || u.indexOf("dx.doi.org/")>0 && (idx=ia("/pdb",u))>0 && idx+4<L
                 || u.indexOf(".ebi.ac.uk")>0 && u.indexOf("/msd")>0 &&  ( (idx=ia("/summary/",u))>0 || (idx=id)>0)
                 || (idx=ia("/pdbsum/",u))>0
                 || (idx=ia("/pdb/files/",u))>0
                 || (idx=iaw("load=PDB:",u))>0
                 ) {
            pf="PDB:"+ wordAt(idx,u,LETTR_DIGT);
        }
        else if ( (idx=ia("/pdb.cgi?sid=d",u))>0 && u.indexOf("scop")>0 && 5+idx<L) {
            pf="PDB:"+u.substring(idx,idx+4)+":"+u.substring(idx+4,idx+5).toUpperCase();
        } else if ((u.indexOf(".rcsb.org/")>0 || u.startsWith("http://www.pdb.org/") || u.startsWith("http://www.fli-leibniz.de/cgi-bin/ImgLib.pl?CODE=")) && equals>0) {
            final int a=ia("structureId=",u);
            if (a>0) {
                pf="PDB:"+u.substring(a,e(a,u));
                final int c=ia("chainId=",u);
                if (c>0) pf+=":"+u.charAt(c);
            } else  pf="PDB:"+u.substring(equals+1,e(equals,u));
        } else if ( u.indexOf("/pdbe-srv/")>0 && (idx=ia("/entry/",u))>0) { pf="PDB:"+wordAt(idx,u,LETTR_DIGT);
        } else if (
                   (u.indexOf("UniProt")>0 && u.indexOf("biomyn.de")>0 && (idx=ia("mid=",u))>0)
                   || (u.indexOf("http://www.tcdb.org/")>=0 && (idx=ia("anum=",u))>0)
                   || (lc.indexOf(".expasy.")>0 &&  (idx=ia("/niceprot.pl?",u))>0)
                   || (idx=idxAW(lc,"uniprot+",
                                 "uniprot:", "uniprot.org/entry/", "uniprot.org/uniref/",
                                 "[swiss_prot-id:'", "[trembl-id:'", "[uniprot-acc:(", "[uniprot-proteinid:",
                                 "sptr_ac=", "uniprot.org/blast/?about=",  "swissprotrelease-accnumber:%27"))>0
                   || (idx=ia("/uniprot/",u))>0
                   || (idx=ia("[UNIPROT-acc:",u))>0
                 ) iUniprot=idx;
        else if (lc.indexOf("/get-sprot")>0) iUniprot=quest+1;

        else if ((idx=idxAW(u,"[EMBL:'","[emblidacc-id:","emblfetch?id=","EMBL_features-id:","www.ebi.ac.uk/cgi-bin/expasyfetch?"))>0) pf="EMBL:"+wordAt(idx,u,LETTR_DIGT_US);

        else if (u.indexOf("howdy.jst.go.jp")>0 && u.indexOf("GenBank")>0 && (idx=iaw("Val=",u))>0)  pf="NCBI_NT:"+u.substring(idx,e(idx,u));

        else if ( lc.indexOf(".ncbi.")>0) {
            String db=lc.indexOf("/protein/")+iaw("db=protein",lc)+u.indexOf("=NM_")>0 ? "NCBI_AA:" :
                lc.indexOf("/nuccore/")+iaw("db=n",lc)>0 ? "NCBI_NT:" :
                    null;
            if ( (idx=ia(".gov/nuccore/",u))>0 || (idx=ia(".gov/protein/",u))>0 ) pf=db+atoi(u,idx);
            else if ((u.indexOf("/vast")>0  || u.indexOf("Op=VastSum")>0) && equals>0) pf="NCBI_NEIGHBOR:"+u.substring(equals+1);
            else if (u.indexOf("/mmdb")>0 && equals>0) pf="NCBI_3D:"+u.substring(equals+1);
            else if (u.indexOf("cddsrv.cn3")>0 && (idx=ia("uid=",u))>0)  pf="NCBI_3D:"+wordAt(idx,u,DIGT);
            else if ((idx=ia("/nucleotide/",u))>0) pf="NCBI_NT:"+wordAt(idx,u,DIGT);
            else {
                int idOrVal=id>0 ? id : iaw("val=",lc);
                if (idOrVal<0) idOrVal= iaw("term=",lc);
                if (idOrVal>0) {
                    if (db==null && L>idOrVal && is(LETTR_DIGT,u, idOrVal)) db="NCBI_NT:";
                    if (db!=null) pf= db+u.substring(idOrVal,e(idOrVal,u));
                }
            }
        }
        else if (u.indexOf("pfam.sanger.")>0  && ( (idx=ia("acc=",u))>0 || (idx=ia("entry=",u))>0 )
                 || (idx=ia("&fam=PF",u)-2)>0 )
            pf= (u.indexOf("=full")>0? "PFAM_FULL:":"PFAM:")+u.substring(idx,e(idx,u));
        else if (u.indexOf("prodom")>0 && u.indexOf(".fr/")>0 && (idx=u.indexOf("=PD"))>0) pf="PRODOM:"+u.substring(idx+1,e(idx,u));
        else if ( u.indexOf("viperdb.")>0 && (idx=ia("VDB=",u))>0) pf="VIPERDB:"+u.substring(idx,e(idx,u));
        else if ((idx=equals1)>0 &&  u.startsWith("http://www.ebi.ac.uk/interpro/") || (idx=ia("[interpro-AccNumber:",u))>0)  pf="INTERPRO:"+wordAt(idx,u,LETTR_DIGT);
        else if (u.indexOf("systers.molgen.mpg.de")>0 && ( (idx=ia("CLNR=",u))>0 || (idx=iaw("nr=",u))>0 ))  pf="SYSTERS_CLUSTER:"+u.substring(idx,e(idx,u));
        else if ( (idx=ia("UNIPARC-refdbsv:",u))>0)  pf="UNIPARC:"+wordAt(idx,u,LETTR_DIGT);
        else if ( (idx=ia("[emblcds-id:",u))>0)  pf="EMBLCDS:"+wordAt(idx,u,LETTR_DIGT);

        else if ( strstr("ce.php?",u)>0 && ( (idx=ia("align=",u))>0  || (idx=ia("alignAndRearange=",u))>0)) {
            final int bar=mini( strchr(STRSTR_E,'|',u,idx, MAX_INT), strchr(STRSTR_E,'&',u,idx, MAX_INT), strchr(STRSTR_E,'%',u,idx, MAX_INT));
            if (bar>0) pf=u.substring(idx,bar);
        }
        else if ((idx=iaw("hgnc_id=",u))>0) pf="HGNC:"+wordAt(idx,u,LETTR_DIGT);
        if (pf==null && iUniprot>0) pf="UNIPROT:"+wordAt(iUniprot,u,LETTR_DIGT_US);
        if (pf==null &&  u.startsWith("javascript")) {
            int i=0;
            while( (i=u.indexOf('\'',i)+1)>1) {
                final int noLDUS=nxt(NO_LETTR_DIGT_US,u,i,L);
                if (chrAt(noLDUS,u)!='\'') continue;
                final int noLD=nxt(NO_LETTR_DIGT,u,i,L);
                final int lttr=nxt(LETTR,u,i,noLDUS) , digt=nxt(DIGT,u,i,noLDUS);
                String db=null;
                if (noLD==noLDUS) {
                    if (noLD==i+LEN_PDB_ID) db="PDB:";
                    if (noLD==noLDUS && noLDUS-i==LEN_UNIPROT  && digt>0 && lttr>0) db="SWISS:";
                } else if (lttr>0) {
                    int countUS=0;
                    boolean goodPos=false;
                    for(int k=i; k<noLDUS; k++) {
                        if (u.charAt(k)=='_') {
                            countUS++;
                            goodPos=k-i>2 && k-i<5 && L-i>2;
                        }
                    }
                    if (goodPos && countUS==1) db="SWISS:";
                }
                if (db!=null) {
                    pf=db+u.substring(i,noLDUS);
                    break;
                }
            }
        }
        return rplcToStrg("%20","",pf);
    }
    /* <<< toDbColonID <<< */
    /* ---------------------------------------- */
    /* >>> References in HTML >>> */
    public static char isHtmlOrXml(ByteArray ba) {
        final byte T[]=ba.getBytes();
        final int E=ba.getEnd();
        int  b=nxt(NO_SPC, ba);
        if (b>=0 && equls("<!-- ",T, b) &&  (b=ia(" -->",T,b,E))>0) b=nxtE(NO_SPC, T,b,E);
        return
            equls(STRSTR_IC,"<html>",T,b)||equls("<!DOCTYPE HTML",T,b) ? 'H' :
            equls(STRSTR_IC,"<?xml",T,b) ? 'X' :
            ' ';
    }
    public static CharSequence getProteinReferencesInHtmlFile(ByteArray ba) {
        final byte T[]=ba.getBytes();
        final int ends[]=ba.getLineEnds(),  B=ba.getBegin(), E=ba.getEnd();
        final char htmlOrXml=isHtmlOrXml(ba);
        final ByteArray ret=new ByteArray(99);
        int idx=0;
        if (htmlOrXml=='H') {
            if (strstr(STRSTR_IC,"<title>SCOP: ",T,B,E)>0) {
                final List<String> v=new ArrayList();
                for(int pdb=0;  (pdb=ia(".cgi?pd=",T,pdb,E))>0; ) {
                    final int gt=strchr('>',T,pdb,E);
                    if (gt>0) {
                        final int chain=ia(";pc=",T,pdb,gt);
                        if (chain>0) {
                            String s="PDB:"+ba.newString(pdb,pdb+4)+"_"+(char)(get(chain,T)&~32);
                            if (!v.contains(s)) v.add(s);
                        }
                    }
                }
                return joinAsStrg(v.toArray()," ");
            } else if (strstr(STRSTR_IC,"title=\"InterPro",T,B,E)>0) {
                for(int li=B;  (li=ia("<li>",T,li,E))>0; ) {
                    int liEnd=strstr(0L, "</li>",T,li,E);
                    if (strchr(' ',T,li,liEnd)<0) ret.append(" UNIPROT:").append(T,li,liEnd);
                }
            } else if (strstr(STRSTR_IC,"<title>Vast",T,B,E)>0) {
                final int iUid=ia("uid=",T,B,E), iLF=strchr('\n',T,iUid+1,E), iQuotes=strchr('"',T,iUid+1,iLF);
                if (iUid>0 && iQuotes>0 && iLF>0) {
                    final int iChain=ia(" sequence ",T,iQuotes,iLF);
                    ret.append("PDB:"+new String(T,iUid,iQuotes-iUid)+(iChain>0?":"+(char)T[iChain] : "").toLowerCase());
                }
            } else if ( (idx=ia("<title>Structure Summary, ",T,B,E))>0) {
                final int iComma=strchr(',',T,idx,E);
                if (iComma>0) ret.append("PDB:"+new String(T,idx,iComma-idx).trim().toLowerCase());
            } else if(ia("define=\"title\">InterPro:",T,B,E)>0) {
                int pos=0;
                while( (pos=ia("/uniprot/",T,pos,E))>0) {
                    final int e=nxt(NO_LETTR_DIGT,T,pos,E);
                    if (get(e,T)=='"')  ret.append(" UNIPROT:").append(ba.newString(pos,e));
                }
            }
        }
        if (htmlOrXml!='H') {
            boolean isUniprotList=ends.length>0;
            for(int iL=0; iL<ends.length; iL++) {
                final int b=iL==0?B:ends[iL-1]+1, e=prev(NO_SPC,T,ends[iL]-1,b-1)+1;
                if (e-b==0) break;
                if (e-b!=6 || !is(LETTR,T[b]) || nxt(NO_LETTR_DIGT,T,b,e)!=-1 || nxt(DIGT,T,b,e)==-1) {
                    isUniprotList=false; break;
                }
            }
            if (isUniprotList) {
                final ByteArray sb=new ByteArray(16*ends.length);
                for(int iL=0; iL<ends.length; iL++) {
                    final int b=iL==0?B:ends[iL-1]+1;
                    sb.append("UNIPROT:").append(T,b,ends[iL]).append(' ');
                }
            }
        }

        if ((htmlOrXml=='H' || htmlOrXml=='X') && sze(ret)==0) {
            String ids[]=collectWordsAfter("http://www.uniprot.org/uniprot/", ba);
            if (ids.length==0) ids=collectWordsAfter("http://www.uniprot.org/entry/", ba);
            for (String id : ids) {
                if (is(UPPR,id,0)) ret.append("UNIPROT:"+id+" ");
            }
            ids=collectWordsAfter("http://www.ncbi.nlm.nih.gov/entrez/viewer.fcgi?val=", ba);
            for (String id : ids) {
                if (is(UPPR,id,0)) ret.append("NCBI:"+id+" ");
            }
        }
        return ret;
    }
    public final static String[] collectWordsAfter(String prefix, CharSequence haystack) {
        List <String>v =null;
        for( int i=0; (i=strstr(STRSTR_AFTER,prefix,haystack,i,MAX_INT))>0;) {
            final String word=wordAt(i,haystack,LETTR_DIGT_US);
            if (sze(word)>2) adUniq( word, v==null ? v=new ArrayList() : v);
        }
        return toStrgArry(v);
    }

    private final static int e(int i, String u) {
        final int amp=u.indexOf('&',i);
        return amp>0 ? amp : u.length();
    }

    private final static int ia(String needle, Object haystack) {  return strstr(STRSTR_AFTER, needle,haystack); }
    private final static int iaw(String needle, Object haystack) { return strstr( STRSTR_w_L|STRSTR_AFTER, needle,haystack); }
    private final static int ia(String needle, Object haystack, int haystackFrom, int haystackTo) {  return strstr(STRSTR_AFTER, needle,haystack, haystackFrom,haystackTo); }

    public static void toSeqvista(java.io.File ff[]) {
        final java.io.File fJar=InteractiveDownload.downloadFileIfNewer("http://zlab.bu.edu/SeqVISTA/ws/SeqVISTA.jar");
        if (sze(fJar)>0) {
            ChExec ex=new ChExec(0).setCommandLineV(file(systProprty(SYSP_BIN_JAVA)),"-jar", fJar, ff);
            new Thread(ex).start();
        }
    }
}
