package charite.christo;
import charite.christo.interfaces.*;
import java.awt.event.*;
import java.util.*;
import java.net.URL;
import java.io.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP
WIKI:Windows32 is the world's most widely applied operation system.

<br>
Windows differs to conventional OS with respect to
<ul>
<li>Command line syntax</li>
<li>File separator character "\". On normal systems backslash is an WIKI:Escape_character and not a file separator character.</li>
<li>Path separator character ";"</li>
<li>File system:
<ul>
<li>File and folder names: Case insensitive, some names not possible because reserved for MS_DOS</li>
<li>File system: Language dependent directory tree</li>
<li>Missing hard links</li>
<li>Limitation of file path length to 260 characters. See WIKI:Path_(computing).</li>
<li>Unusual characters in file and directory names like space and Umlaute</li>
</ul>
</li>
</ul>

<b>Always-On-Top attribut of a window:</b> This is one of the
essential features the Windows desktop is still lacking.  Fortunately
there are free and shareware programs to install the Stay-on-Top
attribut of a windows on the desktop: AlwaysOnTopMaker (recommended, because it is very small) or Acer-GridVista (free) or WindowSpace.

<br><br>

<b>Some important Windows tips for STRAP users</b><br>

<ul>

<li>If STRAP cannot download files because of  Web-proxy problems it might help to webstart STRAP with IE.</li>

<li>The Explorer option "<I>suppress known file extensions</I>" should be turned off for the entire system.
With the default Explorer setting some problems remain unnoticed:
Windows sometimes appends the suffix ".txt" to file names which have already a file extension
resulting in  endings like  ".txt.txt" or ".pdb.txt".
Files with the ending ".jar" are sometimes arbitrary renamed and get the ending ".zip".
</li>

<li>The editor WIKI:Notepad does not cope with file ends of some protein files. Use WIKI:Notepad_plus_plus or WIKI:Wordpad instead.</li>

<li>
Certain file-names are reserved by the system such as "aux.pdb", "con.swissprot", "prn.fasta", "icon" and "icons".
Their use leads to loss of data.
</li>

<li>
Some Bioinformatics  programs run only on English but
not on German Windows installations because the file directory tree differs.
Sometimes it helps to set the english language settings for the decimal separator.
</li>

<li>
Some bioinformatics programs expect slash as a file separator
rather than the back-slash which is the native separator on MS-Windows.
</li>

</ul>

<br>

<br><br><b>Cygwin:</b>

STRAP uses the Gnu-compiler in WIKI:Cygwin to
install Bioinformatics programs if no other is specified.

<br><br><b>The directory for program settings:</b>

WIKI:Dot_files are used for files that store settings of application programs.

Since there had been problems on some Windows versions  with file names starting with a dot, the dot is omitted.

Consequently the  settings directory is "%HOMEPATH%\StrapAlign\" rather than ~/.StrapAlign/".

But if %HOMEPATH%  contains white space then "C:\StrapAlign\" is used instead.

An alternative directory can be provided in the file %HOMEPATH%\.location_of_StrapAlign.ini.

@author Christoph Gille

But file:///C:/ is fine but failes file:/C:/...
Do not use quotes. rundll32.exe shell32.dll,Control_RunDLL "file:///C:/....html" does not work. Without quotes it works.

*/
public class Microsoft implements ActionListener,ChRunnable {
    private Microsoft(){}
    private static Microsoft inst;
    private static Microsoft instance() {if (inst==null) inst=new Microsoft(); return inst;}
    public static String cmd() { return isSystProprty(IS_WINDOWS_95_FAMILY)  ? "command.com" : "cmd.exe";}

    private static void installCygwin(String packages,boolean always) {
        putln("cccccc installCygwin ",packages);
        if ( appendPackageToInstalledDb(dirCygwin(),packages) | always) {
            if (ChMsg.yesNo("Going to install/update the cygwin packages<pre>"+packages+"</pre>This will take some time."))
                ChDelay.newThread(instance(),"INSTALL_CYGWIN",packages).start();
        }
    }
    private static ChFrame fCygwin;
    public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        if (cmd=="CW") {
            final String packages=(String)gcp("PACKAGES",ev.getSource());
            putln(" cccccccccccccc  ",packages);
            installCygwin(packages,true);
            //            if (sze(packages)==0) installCygwin("",true);
 //            else {
//                 if (fCygwin==null) (fCygwin=new ChFrame("Install Cygwin").ad(panInstallCygwin(null))).pack();
//                 fCygwin.shw(0);
//             }
        }
    }
    private static ChButton cbCygwinUnattended_;
    private static ChButton cbCygwinUnattended() {
        if (cbCygwinUnattended_==null) cbCygwinUnattended_=toggl("Run Cygwin setup.exe unattended and automatically ").s(true);
        return cbCygwinUnattended_;
    }

    public static ChButton newCygwinButton(String packages) {
        ChButton b=new ChButton("CW").t("Install cygwin...").li(instance());
        if (packages!=null) {
            b.t("Install the Cygwin packages  "+packages);
            pcp("PACKAGES",packages.replaceAll("CC","gcc-g++ gcc-g77 make python"),b);
        }
        return b;
    }
    public static JComponent panInstallCygwin(String packages) {
        final JComponent
            panExist=vBhB(
                          hB("If you already have","WIKI:Cygwin"," then enter the file path where it is installed","#"),
                          ChTextField.newTextField(KEY_CYGWIN),
                          hB("the default path is  ", "?:/cygwin/","#")
                          ),
            panProblem=vB("#TB In case of problems with Cygwin installation",
                          " Unselect the following check-box and try installation again. ",
                          cbCygwinUnattended().cb()
                          ),
            pan=vBhB(
                     hB("WIKI:Cygwin"," is necessary for many Bioinformatics programs embedded in STRAP.","#"),

                     hB("Press the following button if you already have ","WIKI:cygwin"," ", ChButton.actionViewMessage("Existing cygwin installation ... ",panExist),"#"),
                     " ",
                     hB("Otherwise select a drive with sufficient free capacity (500M) ", comboRoots().cln(),"#"),
                     " ",
                     packages==null ? hB("Install a minimum set of cygwin packages: ",  newCygwinButton("CC grep coreutils"),"#") : newCygwinButton(packages),
                     " ",
                     packages==null ? hB( "or extended set including  PDF tools ",  newCygwinButton("CC grep coreutils tetex tetex-extra gv wget"),"#") :null,
                     " ",
                     pnl(" ",panProblem),
                     " ",
                     hB("The file ", file(dirCygwin()+"/etc/setup/installed.db")," lists all installed packages","#")
                     );
        return pan;
    }
    private static boolean setupFailed;
    public Object run(String id,Object arg) {
        if (id=="INSTALL_CYGWIN") {
            final String packages=arg!=null ? arg.toString() : "gcc-g++ gcc-g77 findutils wget gcc-g++ make ";
            if (777==777) {
            ChMsg.msgDialog(
                                "The Cygwin Setup has changed. Therefore you need to install Cygwin yourself. Visit <br>"+
                                "http://www.cygwin.org/<br>"+
                                "and run the setup.<br>"+
                                "Select the  following packages for installation:<br>"+
                                packages
                                );
            return null;
            }

            final File f=file(dirBin()+"/cygwinSetup/setup.exe");
            mkParentDrs(f);
            String drive=toStrg(comboRoots());
            final File dirCygwin=file(comboRoots()+"/cygwin");
            final ChTextField tf=ChTextField.newTextField(KEY_CYGWIN);
            if (tf!=null) tf.setText(dirCygwin.toString());
            setPrpty(ChUtils.class,KEY_CYGWIN,dirCygwin.toString());
            final long age=(System.currentTimeMillis()-f.lastModified())/1000;
            if (age> 60*60*24* 7) delFile(f);
            InteractiveDownload.downloadFiles(new URL[]{url("http://www.cygwin.com/setup.exe")},new File[]{f});
            appendPackageToInstalledDb(dirCygwin, packages);
            final File fBash=file(dirCygwin()+"/bin/bash.exe");
            final File fProxy=file(dirCygwin()+"/etc/setup/last-connection");
            if (sze(fProxy)==0) { // Web-Proxy
                final java.net.InetSocketAddress proxy=proxyForUrl(url("http://www.cygwin.org/"));
                if (proxy!=null) wrte(fProxy,"Proxy\n"+proxy.getHostName()+"\n"+proxy.getPort()+"\n\n");
            }
            final File fAction=file(dirCygwin()+"/etc/setup/last-action");
            if (sze(fAction)==0) wrte(fAction,"Download, Install\n");
            new ChExec(setupFailed ? ChExec.SHOW_STREAMS : 0).setCommandLineV(cmd(),"/c",f, cbCygwinUnattended().s()? "-q":null, "-R",drive+"cygwin","-l",drive+"cygwin").run();
            if (fBash.length()==0)  { setupFailed=true; error("Can not find <br>"+fBash); }
            else {
                File ff[]=new File[999];
                int count=0;
                for(File f0 : dirCygwin().listFiles()) {
                    if (f0.toString().indexOf("%3a%2f%2f")>0 && count<ff.length)
                        ff[count++]=f0;
                }
                if (count>0 && ChMsg.yesNo("Save memory and delete temp files?"))
                    for(int i=count; --i>=0;) deleteTree(ff[i]);
            }
        }
        return null;
    }
private static boolean appendPackageToInstalledDb(File dirCygwin, String packageNames) {
        if (packageNames==null) return false;
        boolean needRunSetup=false;
        final File installedDb=file(dirCygwin+"/etc/setup/installed.db");
        mkParentDrs(installedDb);
        final StringBuffer inst=new StringBuffer(9999).append(readStrg(installedDb));
        final StringTokenizer st=new StringTokenizer(packageNames);
        while (st.hasMoreTokens()) {
            final String packageName=st.nextToken();
            putln("PACKAGE packageName= "+packageName);
            boolean contains=false, alreadyInstalled=false;
            final String tar=" "+packageName+"-0-0.tar.bz2 0";
            int  idx=0;
            while( (idx=inst.indexOf(" "+packageName+"-",idx+1))>0 ) {
                contains=true;
                alreadyInstalled |=  (!inst.substring(idx).startsWith(tar));
            }
            if (!contains) {
                inst.append("\r\n"+packageName+tar+"\r\n");
                putln("PACKAGE missing "+packageName);
            }
            needRunSetup |= !alreadyInstalled;
        }
        if (needRunSetup) {
            wrte(installedDb,inst);
            putln("Microsoft.java: "+installedDb+"\n\n"+inst+"\n\n");
        }
        return needRunSetup;
    }
    private static ChCombo comboRoots_;
    public static ChCombo comboRoots() {
        if (comboRoots_==null) {
            final File ff[]=listExistingFileRoots();
            comboRoots_=new ChCombo((Object[])ff,0);
            for(File f:ff)
                if ((32|chrAt(0,toStrg(f)))=='c')
                    comboRoots_.s(f);
        }
        return comboRoots_;
    }
    public static String toCygwinPath(Object f) {
        final String p=toStrg(f);
        if (p==null) return null;
        return isWindows() ? "/cygdrive/"+p.charAt(0)+"/"+p.substring(2).replace('\\','/') : p;
    }

    public static String regQuery(String query) {
        File fTmp=createTempFile(".reg");
        final ChExec exec=new ChExec(0).setCommandLineV("c:\\winxp\\i386\\regedit.exe","/e",fTmp,"\"HKEY_CLASSES_ROOT\\Software\\DeLano Scientific\\PyMOL\\PYMOL_PATH\"");
        exec.run();
        new ChFrame("").ad(getCtrlPanel(exec)).shw(0);
        final ChExec exec2=new ChExec(ChExec.STDOUT).setCommandLineV("type",fTmp);
        new ChFrame("").ad(getCtrlPanel(exec2)).shw(0);
        final String stdout[]=splitStrg(getCtrlPanel(exec2),'\n');
        if (stdout==null) return null;
        new ChTextViewer(joinLines(stdout)).shw(0);
        for(String s:stdout) {
            if (s.startsWith("@="))
                return s.substring(2);
        }
        return null;
    }

}
/*

which curl

lt-8 -> {
alt-5 -> [
alt-shift-7 -> \
alt-n -> ~

http://cygwin.com/setup-1.7.exe
http://alumnit.ca/~apenwarr/netselect/
http://sourceware.org/cygwin-apps/package-server.html
Or you could modify the attached Windows batch file (I've made it
a TXT file so it passes through most filters) and change the
mirror name to one that's closer.

You might have to make one batch file per country :-)

Ralph

*/
