package charite.christo;
import java.net.URL;
import java.net.HttpURLConnection;
import java.io.*;
import java.util.Set;
import java.util.Map;
import java.util.HashSet;
import java.util.HashMap;
import java.lang.ref.Reference;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.PROTEIN_MIN_FILE_SZE;
public final class SOAP_WSDbfetch {
    public final static int SKIP_IF_ANY_EXISTS=1;
    public final static String
        CMD_WSDBFETCH_SEND_QUERY="CMD_WSDBFETCH_SEND_QUERY",
        CMD_WSDBFETCH_ERROR="CMD_WSDBFETCH_ERROR",
        CMD_WSDBFETCH_SUCCESS="CMD_WSDBFETCH_SUCCESS",
        DB_EMBL="EMBL",DB_UNIPARC="UNIPARC",
        DB_UNIPROT="UNIPROT",DB_UNIPROTKB="UNIPROTKB",
        DATABASES[]={DB_UNIPROTKB,DB_UNIPROT, DB_EMBL,DB_UNIPARC};
    private SOAP_WSDbfetch(){}
    private static String preferredFormat(String db) {
        if (DB_UNIPROTKB.equalsIgnoreCase(db) || DB_UNIPROT.equalsIgnoreCase(db) ) return "uniprot";
        if (DB_EMBL.equals(db)) return "embl";
        return "fasta";
    }
    private static Reference<Map<String,File>> mapEntryFile;
    public static File dbColonID2file(String entry) {
        Map<String,File> map=mapEntryFile!=null ? mapEntryFile.get() : null;
        if (map==null) mapEntryFile=new java.lang.ref.SoftReference(map=new HashMap());
        if (entry==null || sze(entry)>66) return null;
        File f=map.get(entry);
        if (f==null) {
            final int colon=entry.indexOf(':');
            if (colon<0) return null;
            final String database=( entry.startsWith("EXPASY:") || entry.startsWith("SWISS:") ) ? DB_UNIPROT : entry.substring(0,colon);
            map.put(entry,f=id2file(database,entry.substring(colon+1), null));
        }
        return f;
    }
    private static File dir_;
    private static File dir() {
        if (dir_==null) {
            dir_=file(dirSettings() +"/wsdbfetch/");
            final String README=
                "You may remove the files in the directory to save memory\n"+
                "This directory contains protein files which are downloaded\n"+
                "via SOAP WSDbfetch from EBI\n";
            mkdrsErr(dir_,README);
        }
        return dir_;
    }
private static File id2file(String database,String id,String format0) {
        final String format=sze(format0)>0 ? format0 : preferredFormat(database);
        if (id==null||sze(id)>33 || database==null||sze(database)>33) return null;
        return file(dir()+"/"+(database+"_"+rplcChars(charClass(NO_FILENM),'_',id)).toUpperCase()+"."+format);
    }

    public static void download(String tokens[],int options) {
        final Set<String> vQuery=new HashSet();
        for(String db :DATABASES) {
            vQuery.clear();
            final String dbColon=db+":";
            if (tokens!=null) {
                for(String dbColonId:tokens) {
                    if (dbColonId!=null && dbColonId.startsWith(dbColon) ||
                        dbColonId!=null && DB_UNIPROT==db && (dbColonId.startsWith("EXPASY:")||dbColonId.startsWith("SWISS:"))) {
                        final int colon=dbColonId.indexOf(':'), pipe=dbColonId.indexOf('|');
                        final String s=pipe>0 ? dbColonId.substring(0,pipe) : dbColonId;
                        final String id=s.substring(colon+1);
                        if ( (options&SKIP_IF_ANY_EXISTS)!=0 &&
                             (sze(id2file(db,id,preferredFormat(db)))>=PROTEIN_MIN_FILE_SZE || sze(file(Hyperrefs.toUrlString(s,Hyperrefs.PROTEIN_FILE)))>=PROTEIN_MIN_FILE_SZE)) {
                            //if (isMyComputer()) putln("SOAP_WSDbfetch: skipping ",s);
                            continue;
                        }
                        vQuery.add(id);
                    }
                }
            }
            SOAP_WSDbfetch.download(db,toArry(vQuery,String.class),SOAP_WSDbfetch.preferredFormat(db));
        }
    }
    private final static int MAXNUM=100;
    public static void download(final String database,final String[] IDs,final String format0)  {
        int fromIdx=0;
        while(fromIdx<IDs.length) {
            final String format=format0!=null && format0!="" ? format0 : preferredFormat(database);
            final StringBuffer sb=new StringBuffer();
            for(int i=fromIdx, count=0; i<IDs.length && count<MAXNUM; i++) {
                final File f=id2file(database,IDs[i],format);
                if (f!=null && sze(f)==0) {
                    if (sze(sb)>0) sb.append(' ');
                    sb.append(IDs[i]);
                    count++;
                }
                fromIdx=i+1;
            }
            final String queries=sb.toString();
            if (sze(queries)>0) {
                final String info="database="+database+"\nformat="+format+"\nqueries="+queries+"\n";
                try {
                    final URL u=new URL("http://www.ebi.ac.uk/ws/services/WSDbfetch");
                    final HttpURLConnection con=(HttpURLConnection) u.openConnection();
                    con.setDoOutput(true);
                    con.setDoInput(true);
                    con.setRequestMethod("POST");
                    con.setRequestProperty("SOAPAction", "http://www.ebi.ac.uk/ws/services/WSDbfetch#fetchBatch");
                    con.setRequestProperty("Accept","application/soap+xml, application/dime, multipart/related, text/*");
                    con.setRequestProperty("Content-Type", "text/xml; charset=\"utf-8\"");
                    con.setRequestProperty("User-Agent","STRAP");
                    con.setRequestProperty("Cache-Control","no-cache");
                    final OutputStream wout=con.getOutputStream();
                    mkdrsErr(file(dir()+"/log/"),"Log messages for debugging. You can delete these files.");
                    final File fLog=file(dir()+"/log/"+System.currentTimeMillis()+".txt");
                    final OutputStream wLog=fileOutStream(fLog);
                    final String query= "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"+
                        "<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\" xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">"+
                        "<soapenv:Body>"+
                        "<ns1:fetchBatch soapenv:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\" xmlns:ns1=\"http://wsdbfetch.ws.jdbfetch.ebi.ac.uk\">"+
                        "<db xsi:type=\"xsd:string\">"+
                        database+
                        ("</db>"+
                         "<ids xsi:type=\"xsd:string\">")+
                        queries+
                        ("</ids>"+
                         "<format xsi:type=\"xsd:string\">")+
                        format+
                        ("</format>"+
                         "<style xsi:type=\"xsd:string\">"+
                         "raw</style>"+
                         "</ns1:fetchBatch>"+
                         "</soapenv:Body>"+
                         "</soapenv:Envelope>"+
                         "\n\n");
                    final ByteArray sSend=new ByteArray(query);
                    {
                        sSend.write(wLog);
                        sSend.write(wout);
                        wLog.flush();
                        wout.flush();
                        closeStrm(wout);
                        final InputStream in=con.getInputStream();
                        final ByteArray ba=readBytes(new BufferedInputStream(in),getByteArray("SOAP_WSDbfetch",0));
                        if (ba!=null) {
                            ba.write(wLog);
                            closeStrm(wLog);
                            final byte txt[]=ba.getBytes();
                            final int end0=ba.getEnd();
                            final int beginO=strstr("<fetchBatchReturn" ,txt,0,end0);
                            final int end=strstr("</fetchBatchReturn",txt,beginO,end0);
                            final int begin=1+strchr('>',txt,beginO,end0);
                            if (begin>0) {
                                ba.setBegin(begin);
                                ba.setEnd(end);
                                final int ends[]=ba.getLineEnds();
                                int from=0;
                                for(int iL=0; iL<ends.length; iL++) {
                                    final int b=iL==0 ? begin : ends[iL-1]+1, e=ends[iL];
                                    if (e-b<8) continue;
                                    final boolean entryStarts=txt[b]=='I' && txt[b+1]=='D' && txt[b+2]==' ';
                                    if (entryStarts || txt[b]=='A' && txt[b+1]=='C' && txt[b+2]==' ') {
                                        from=b;
                                        for(int iID=IDs.length; --iID>=0;) {
                                            final String id=IDs[iID];
                                            final int idxID=strstr(STRSTR_IC,id,txt,b,e);
                                            if (idxID>0 && !is(LETTR_DIGT,txt[idxID-1]) && !is(LETTR_DIGT,txt,idxID+sze(id))) {
                                                final int slashSlash=strstr("\n//",txt,b,end);
                                                if (slashSlash>0) {
                                                    final File f=id2file(database,id,format);
                                                    wrte(f, txt, from,slashSlash+1);
                                                } else stckTrc();
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                } catch(IOException ex){
                    final File fErr=file("~/@/wsdbfetch_failure/"+rplcNoFilenameChars(queries.substring(0, mini(40,sze(queries)))+info.hashCode()));
                    wrte(fErr,info+"\n"+stackTrceAsStrg(ex));
                    stckTrc(ex);
                }
            }
        }
    }
}
/*
  echo rm ~/.StrapAlign/wsdbfetch/*
  echo rm -r ~/.StrapAlign/wsdbfetch/* ~/.StrapAlign/web/*
*/

