package charite.christo.blast;
import charite.christo.*;
import charite.christo.protein.*;
import java.util.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;

public class BlastAlignment {
    private byte[] matchSeq,querySeq,matchUngapped;

    private int queryStart,queryEnd, matchStart,matchEnd, matchHashcode;
    private int identity,positives,score,number;
    private double expect,bits;
    private int querySeqFrom,querySeqTo, matchSeqFrom, matchSeqTo;
    public void setMatchSeq(byte[] s,int from,int to) {this.matchSeq=s; matchSeqFrom=from; matchSeqTo=mini(s.length,to); }
    public void setQuerySeq(byte[] s,int from,int to) {this.querySeq=s; querySeqFrom=from; querySeqTo=mini(s.length,to); }

    public void setQueryStartEnd(int s,int e) {queryStart=s;queryEnd=e;}
    public void setMatchStartEnd(int s,int e) {matchStart=s; matchEnd=e;}
    public void setIdentity(int i) {identity=i;}
    public void setPositives(int i) {positives=i;}
    public void setScore(int i) {score=i;}
    public void setNumber(int i) {number=i;}
    public void setExpect(double d) {expect=d;}
    public void setBits(double d) {bits=d;}

    public byte[] getMatchSeq() {
        final int f=matchSeqFrom, t=matchSeqTo;
        if (this.matchSeq!=null && (f>0 || t!=matchSeq.length)) {
            final byte bb[]=new byte[t-f];
            System.arraycopy(this.matchSeq,f,bb,0, t-f);
            this.matchSeqFrom=0;
            this.matchSeqTo=bb.length;
            this.matchSeq=bb;
        }
        return this.matchSeq;
    }

    public int getMatchSeqHashCode() {
        if (matchHashcode==0) matchHashcode=hashCd(getMatchSeq(),0,MAX_INT,true);
        return matchHashcode;
    }

    public byte[] getQuerySeq() {
        final int f=querySeqFrom, t=querySeqTo;
        if (this.querySeq!=null && (f!=0 || t!=querySeq.length) ) {
            final byte bb[]=new byte[t-f];
            System.arraycopy(this.querySeq,f,bb,0, t-f);
            this.querySeqFrom=0;
            this.querySeqTo=bb.length;
            this.querySeq=bb;
        }
        return this.querySeq;
    }

    public byte[] getMatchUngapped() {
        if (matchUngapped==null) matchUngapped=onlyLetters(getMatchSeq());
        return matchUngapped;
    }
    public int getQueryStart() { return queryStart;}
    public int getQueryEnd() { return queryEnd;}
    public int getMatchStart() { return matchStart;}
    public int getMatchEnd() { return matchEnd;}
    public int getIdentity() { return identity;}
    public int getPositives() { return positives;}
    public int getScore() { return score;}
    public int getNumber() { return number;}
    public double getExpect() { return expect;}
    public double getBits() { return bits;}
    public String getDatabase(){ return blastHit.getDatabase();}
    public String getID(){ return blastHit.getID();}
    private BlastHit blastHit;
    public BlastHit getBlastHit() {
        assertNotNull(blastHit);
        return blastHit;}
    public void setBlastHit(BlastHit b) { blastHit=b;}
    private boolean isNovel;
    public void setIsNew(boolean b){isNovel=b;}
    public final void toText(ByteArray sb,int outputWidth) {
        final BlastHit bh=getBlastHit();
        final String db=bh.getDatabase(), id=bh.getID();
        sb.append("> ").append(db);
        if (db!=null && db.startsWith("EM")) {
            sb.append("EMCDS".equals(db) ? " EMBLCDS" : " EMBL");
        }
        sb.append(':').append(id).append("  ");
        sb.append(bh.description,bh.descriptionFrom,bh.descriptionTo);
        sb.append( isNovel ? " (NEW)\n\n" : "\n\n");
        sb.append(" Score = ").append(getScore()).append(" (").append((float)getBits()).append("), Expect =  ");
        if (getExpect()<1E-44) sb.append(getExpect()); else sb.append((float)getExpect());
        final byte[] mSeq=this.matchSeq;
        sb.append("\nIdentities = ").append(getIdentity()).append('/').append(matchSeqTo-matchSeqFrom);
        sb.append(", Positives = ").append(getPositives()).append('/').append(matchSeqTo-matchSeqFrom).append("\n\n");
        final byte[] query=this.querySeq;
        final int len=maxi(matchSeqTo-matchSeqFrom, querySeqTo-querySeqFrom, matchSeqTo-matchSeqFrom);
        int posQuery=getQueryStart(), posMatch=getMatchStart();
        for(int col=0; col<len; col+=outputWidth) {
            final int colE=mini(col+outputWidth,len);
            sb.append('Q').append(posQuery+1,8).append(' ');
            sb.append(query,col+querySeqFrom,colE+querySeqFrom).append(' ').append(posQuery+=countLetters(query,col+querySeqFrom,colE+querySeqFrom)).append('\n');
            sb.append("          ");
            midline(query,col+querySeqFrom,colE+querySeqFrom,mSeq,col+matchSeqFrom,colE+matchSeqFrom,sb);
            sb.append("\nH").append(posMatch+1,8).append(' ');
            sb.append(mSeq,col+matchSeqFrom,colE+matchSeqFrom).append(' ').append(posMatch+=countLetters(mSeq,col+matchSeqFrom,colE+matchSeqFrom)).append("\n\n");
        }
    }

    public static void midline(byte[] q,int qFrom, int qTo, byte[] m, int mFrom, int mTo, ByteArray sb) {
        final byte[][] M=Blosum.BLASTMIDLINE;
        for(int iQ=qFrom, iM=mFrom; iQ<qTo && iM<mTo; iQ++,iM++) {
            sb.append( (char) M[q[iQ]&127][m[iM]&127]);
        }
    }
    public final static BlastAlignment[] NONE={};
}
