package charite.christo.blast;
import java.util.*;
import java.io.*;
import charite.christo.*;
import static java.lang.System.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
You probably do not need to use this class directly.
Better use class BlastResult.
*/
public class BlastParser {
    private boolean  writeNew;
    private int countNovelHits, countHits;
    public int countNovelHits() { return countNovelHits;}
    public int countHits() { return countHits;}
    private ByteArray baWhen;
    public void setDay(ByteArray ba,boolean isNew) { baWhen=ba; this.writeNew=isNew;}

    private final static byte[][] RPLC={
        "&apos;".getBytes(), {'\''},
        "&gt;".getBytes(), {'>'},
        "&lt;".getBytes(), {'<'},
        "&quot;".getBytes(), {'"'}
    };

    private static String s(byte txt[],int from,int endText,char term) {
        if (from<0) return "error";
        final int to=strchr(term,txt,from,endText);
        if (from<0 || from>=to || to>txt.length) return "error";
        return new String(txt,from,to-from);
    }

    public boolean parseXML(ByteArray ba,int width, BlastResult blastResult) { return parseXML(ba,width,blastResult,null); }
    public boolean parseXML(ByteArray ba,int width, OutputStream fOut) { return parseXML(ba,width,new BlastResult(null),fOut); }
    private boolean parseXML(ByteArray ba,final int width, BlastResult blastResult, OutputStream fOut) {
        final byte[] T=ba.getBytes();
        final int B=ba.getBegin();
        final boolean isNcbi=strstr("BlastOutput_iterations",ba)>0;
        final int E=strstr(STRSTR_AFTER,isNcbi ? "</Iteration_hits" : "</hits",T, B, ba.getEnd());
        final BlastDatabase database=new BlastDatabase();
        int headerE=0;
        final String databaseName;
        if (isNcbi) {
            final int prgS=strstr(STRSTR_AFTER,"<BlastOutput_program>",T, B, E);
            if (prgS>0) blastResult.setProgramName(s(T,prgS, E,'<'));
            final int dbS=strstr(STRSTR_AFTER,"<BlastOutput_db>",T, B, E);
            databaseName=dbS>0 ? s(T,dbS, E,'<') : "error";
        } else {
            final int headerS=strstr(STRSTR_AFTER,"<Header",T,B,E);
            headerE=strstr(STRSTR_AFTER,"</Header",T,headerS,E);
            final int prgS=strstr(STRSTR_AFTER,"<program ",T,headerS,headerE);

            blastResult.setProgramName(s(T,strstr(STRSTR_AFTER,"name=\"",T,prgS,headerE),headerE,'"'));
            blastResult.setProgramVersion(s(T,strstr(STRSTR_AFTER,"version=\"",T,prgS,headerE), headerE,'"'));

            final int dbS=strstr(STRSTR_AFTER,"<database ",T,headerS,headerE);
            databaseName=s(T,strstr(STRSTR_AFTER,"name=\"",T,dbS,headerE) ,headerE,'"');
            database.setWhenCreated(s(T,strstr(STRSTR_AFTER,"created=\"",T,dbS,headerE),headerE,'"'));
        }
        database.setName(databaseName);
        blastResult.addDatabase(database);
        int hitE=headerE;
        final ByteArray sbOut=fOut!=null ? new ByteArray("") : null;
        try {
            if (fOut!=null) blastResult.writeFile(fOut,width);
        } catch(IOException ioex) {stckTrc(ioex); }
        final int currentDay=(int)(currentTimeMillis()/(1000*24*60*60)), whenEnds[];
        final byte[] whenBytes;
        {
            ByteArray a=this.baWhen;
            if (a!=null) {
                whenEnds=a.getLineEnds();
                whenBytes=a.getBytes();
            } else { whenEnds=null; whenBytes=null;}
        }
        while(true) {
            final int hitS=strstr(STRSTR_AFTER,isNcbi ? "<Hit>": "<hit ",T,hitE,E);
            if (hitS<0) break;
            hitE=strstr(STRSTR_AFTER,isNcbi ? "</Hit>": "</hit>",T,hitS,E);
            if (hitE<0) break;
            final int descriptionS=strstr(STRSTR_AFTER,isNcbi ? "<Hit_def>" : "description=\"", T,hitS,hitE);
            final int descriptionE=strstr(isNcbi ? "</Hit_def>" : "\">",T,descriptionS+1,MAX_INT);
            final int databaseS=strstr(STRSTR_AFTER,isNcbi ? "database=\"" : "database=\"", T,hitS,hitE); // !!!!!
            final BlastHit hit=new BlastHit();
            if (descriptionS>0 && descriptionE>descriptionS) {
                if (strchr('&',T,descriptionS,descriptionE)>=0) {
                    byte[] description=null;
                    for(int i=0; i<RPLC.length; i+=2) {
                        final byte dd[]= rplc(0L, RPLC[i],RPLC[i+1], i==0 ? T : description, i==0 ? descriptionS : 0,  descriptionE);
                        if (description==dd)  break;
                        description=dd;
                    }
                    hit.setDescription(description,0,MAX_INT);
                } else hit.setDescription(T,descriptionS,descriptionE);
            }
            hit.setNumber(atoi(T,strstr(STRSTR_AFTER,isNcbi ? "<Hit_num>" : "number=\"",T,hitS,hitE)));
            final String db=isNcbi ? databaseName : uCase(s(T,databaseS,hitE,'"'));
            hit.setDatabase(db);

            int idS= strstr(STRSTR_AFTER,isNcbi ? "<Hit_id>" : "id=\"",T,hitS,hitE);
            if ( equls(STRSTR_IC,db,T,idS)) idS+=db.length();
            if (T[idS]=='_') idS++;
            final String hitId;
            if (isNcbi && strstr("|BL_ORD_ID|",T,idS, strchr('<',T,idS,hitE))>0) {
                /* --- NCBI-blast writes the following: <Hit_id>gnl|BL_ORD_ID|55894</Hit_id> --- */
                /* --- Other NCBI output: <Hit_id>gi|2624719|pdb|1RHO|A</Hit_id> --- */
                idS=descriptionS;
                int idE=strchr('<',T,idS,hitE);
                final int idxSpc=strchr(' ',T,idS,idE);
                if (idxSpc>0) idE=idxSpc;
                final int idxPipe=strchr('|',T,idS,idE);
                if (idxPipe>0) idE=idxPipe;
                hitId=new String(T,idS,idE-idS);
            } else  hitId=s(T,idS,hitE, isNcbi ? '<' : '"');

            final int hitIdL=sze(hitId);
            if (hitIdL==0) continue;

            hit.setId(hitId);
            final int hitId0=hitId.charAt(0);
            boolean isNovel=false;
            if (whenBytes!=null) {
                int day=0;
                for(int iL=0; iL<whenEnds.length; iL++) {
                    final int b= iL==0 ? 0 : whenEnds[iL-1]+1;
                    final int e=whenEnds[iL];
                    if (e-b<hitIdL+2 || hitId0!=whenBytes[b] || whenBytes[b+hitIdL]!='\t' || strstr(hitId,whenBytes,b,b+hitIdL)<0) continue;
                    day=(int)atol(whenBytes,b+hitIdL+1,e);
                    if (writeNew) whenBytes[e-1]=' ';  else isNovel=whenBytes[e-1]=='N';
                    break;
                }
                if (day==0) {
                    isNovel=true;
                    baWhen.append(hitId).append('\t').append(currentDay).append('N').append('\n');
                    countNovelHits++;
                }
            }
            countHits++;
            blastResult.addHit(hit);

            int alignmentE=hitS;
            blastResult.addHit(hit);
            while(true) {
                final int alignmentS=strstr(STRSTR_AFTER,isNcbi ? "<Hsp>" : "<alignment ", T,alignmentE,hitE);
                if (alignmentS<0) break;
                alignmentE=strstr(STRSTR_AFTER,isNcbi ? "</Hsp>" : "</alignment",T,alignmentS,hitE);
                if (alignmentE<0) break;
                final BlastAlignment alignment=new BlastAlignment();
                alignment.setIsNew(isNovel);
                for(boolean isHit : FALSE_TRUE) {
                    final String pattern=!isHit ? (isNcbi ? "<Hsp_qseq>" : "<querySeq") : (isNcbi ? "<Hsp_hseq>" : "<matchSeq");
                    final int sequenceS=strstr(STRSTR_AFTER,pattern,T,alignmentS,alignmentE);
                    final int sequenceE=strchr('<',T,sequenceS+1,alignmentE);
                    final int start=atoi(T, isNcbi ? strstr(STRSTR_AFTER,!isHit ? "<Hsp_query-from>" : "<Hsp_hit-from>",T, alignmentS,alignmentE) : strstr(STRSTR_AFTER,"start=\"", T,sequenceS,sequenceE));
                    final int   end=atoi(T, isNcbi ? strstr(STRSTR_AFTER,!isHit ? "<Hsp_query-to>" :   "<Hsp_hit-to>",  T, alignmentS,alignmentE) : strstr(STRSTR_AFTER,"end=\"",   T,sequenceS,sequenceE));
                    final int seqS=isNcbi ? sequenceS : strstr(STRSTR_AFTER,"\">",T,sequenceS,sequenceE);
                    if (!isHit) {
                        alignment.setQueryStartEnd(start-1,end);
                        alignment.setQuerySeq(T,seqS,sequenceE);
                    } else {
                        alignment.setMatchStartEnd(start-1,end);
                        alignment.setMatchSeq(T,seqS,sequenceE);
                    }
                }
                alignment.setScore(atoi(T,strstr(STRSTR_AFTER,isNcbi ? "<Hsp_score>" : "<score>",T,alignmentS,alignmentE)));
                alignment.setExpect(atof(T,strstr(STRSTR_AFTER,isNcbi ? "<Hsp_evalue>" : "<expectation>",T,alignmentS,alignmentE)));
                alignment.setBits(atof(T,strstr(STRSTR_AFTER,isNcbi ? "<Hsp_bit-score>" : "<bits>",T,alignmentS,alignmentE)));
                alignment.setIdentity(atoi(T,strstr(STRSTR_AFTER,isNcbi ? "<Hsp_identity>" : "<identity>",T,alignmentS,alignmentE)));
                alignment.setPositives(atoi(T,strstr(STRSTR_AFTER,isNcbi ? "Hsp_positive>" : "positives>",T,alignmentS,alignmentE)));
                alignment.setBlastHit(hit);
                if (fOut!=null) {
                    sbOut.setEnd(0).append(countHits);
                    alignment.toText(sbOut,width);
                    try { wrte(fOut,sbOut);}catch(IOException ex){}
                } else hit.addAlignment(alignment);
            }
            if (fOut==null) blastResult.addHit(hit);
        }
        return true;
    }

}
