package charite.christo.protein;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.awt.*;
import java.lang.ref.*;
import javax.swing.*;
import java.awt.event.*;
import java.io.File;
import java.util.*;
import java.util.List;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
public class ChAstex implements ProteinViewer, ChProcessEvent, astex.MoleculeRendererListener, ChRunnable {
    /*
      http://www.mattryall.net/blog/2007/03/composition-stream-concatenation#comment935
      InputStream FileInputStream

GUI:

Select Ligands-menu: Tell the name of the Ligand
Save View: Tell that run script. And define default file extension like .astexScript.

     */
    /* ---------------------------------------- */
    /* >>> Conversions >>> */
    public astex.MoleculeRenderer getMoleculeRenderer() {
        final astex.MoleculeViewer mv=getAstex();
        return mv!=null ? mv.getMoleculeRenderer() : null;
    }
    private astex.Molecule getMolecule() {
        final astex.MoleculeRenderer mr=getMoleculeRenderer();
        return mr!=null ? mr.getMolecule(this.proteinName) : null;
    }
    private astex.Residue getResidue(int iA) {
        final astex.Molecule mol=getMolecule();
        final Protein p=getProtein();
        if (mol==null || p==null) return null;
        final int  rN=p.getResidueNumber(iA);
        final byte rC=p.getResidueChain(iA);
        final byte rI=p.getResidueInsertionCode(iA);
        for(int iC=mol.getChainCount(); --iC>=0;) {
            final astex.Chain chain=mol.getChain(iC);
            if (rC==0||chrAt(0,chain.getName())==rC) {
                for(int iR=chain.getResidueCount(); --iR>=0;) {
                    final astex.Residue res=chain.getResidue(iR);
                    if (res.getNumber()==rN && (rI==0||res.getInsertionCode()==rI)) return res;
                }
            }
        }
        return null;
    }
    private int toResidueIndex(astex.Atom atom) {
        final astex.Molecule molecule=getMolecule();
        final Protein p=getProtein();
        if (atom==null || p==null || atom.getMolecule()!=molecule) return INT_NAN;
        final astex.Residue residue=atom.getResidue();
        if (residue!=null && residue.isStandardAminoAcid()) {
            final astex.Chain chain = residue.getParent();
            final char chainChar=chrAt(0,chain!=null ? chain.getName() : null);
            return p.residueNumberAndChain2residueIdx(residue.getNumber(),chainChar, (char)residue.getInsertionCode());
        }
        return INT_NAN;
    }
    /* <<< Conversions <<< */
    /* ---------------------------------------- */
    /* >>> Protein >>> */
    private Protein protein;
    private String proteinName;
    public Protein getProtein() { return protein;}
    public boolean setProtein(Protein p, ProteinViewer inSameView, Runnable runAfterInstallation, long options) {
        //putln(ANSI_YELLOW+"ChAstex.setProtein "+ANSI_RESET,p);
        if (protein!=null) return false;
        final File file=p.getFileMayBeWithSideChains();
        if (file==null) return false;
        this.protein=p;
        this.proteinName=p.getName();
        final ChAstex inSameViewPO= (ChAstex) ( inSameView instanceof ChAstex ? (ChAstex)inSameView : inSameView instanceof ChAstexPROXY ? ((ChAstexPROXY)inSameView).pv() : null);
        astex.MoleculeViewer moleculeViewer=inSameViewPO!=null ? inSameViewPO.getAstex() : null;
        if (moleculeViewer==null) {
            moleculeViewer=new astex.MoleculeViewer();
            moleculeViewer.setApplication(true);
            final String rplc[][]={
                {"Exit"},
                {"Save Molecule"},
                {"Save All"},
                {"Run Script...","Load view / Run Script..."},
                {"Write BMP...", "Write image in BMP format"}
            };
            final JMenuBar mb=this.share_menubar=menuBar2j(moleculeViewer.createMenuBar(), rplc);
            mb.add(pnl());
            mb.add(new ChButton("SHW_OBJECT",ChButton.NO_MARGIN).t("Surfaces and ribbons ...").li(LI));
            moleculeViewer.setUsePopupMenu(true);
            this.share_refPanel=weakRef(moleculeViewer);
            this.share_vView=new UniqueList(ChAstex.class);
            this.share_ref_astexMV=weakRef(moleculeViewer);
            this.share_vObjects=new ArrayList();
            this.share_mapSelection=new HashMap();
            this.share_mapSelectionExpression=new HashMap();
            this.share_mapCartoonsCmd=new HashMap();
            this.share_surface_color=new String[1];
        } else {
            this.share_menubar=inSameViewPO.share_menubar;
            this.share_ref_astexMV=inSameViewPO.share_ref_astexMV;
            this.share_refPanel=inSameViewPO.share_refPanel;
            this.share_vView=inSameViewPO.share_vView;
            this.share_vObjects=inSameViewPO.share_vObjects;
            this.share_mapSelection=inSameViewPO.share_mapSelection;
            this.share_mapSelectionExpression=inSameViewPO.share_mapSelectionExpression;
            this.share_mapCartoonsCmd=inSameViewPO.share_mapCartoonsCmd;
            this.share_surface_color=inSameViewPO.share_surface_color;
        }
        getMoleculeRenderer().addMoleculeRendererListener(this);
        adUniq(this, share_vView);
        // getMoleculeRenderer().addMolecule(file.toString(), p.getName());
        astex.Molecule molecule=null;

        final File ff[];
        {
            final List<File>vH=new ArrayList();
            vH.add(file);
            for(HeteroCompound h : p.getDnaAndRnaAndHeteros()) adUniq(h.getFile(), vH);
            ff=toArry(vH, File.class);
        }
        final File fileLoad;
        if (ff.length>1) {
            int size=0;
            for(File f : ff) size+=sze(f)+1;
            final ByteArray ba=new ByteArray(size);
            for(File f : ff) ba.append(readBytes(f)).append('\n');
            wrte(fileLoad=createTempFile(".pdb"), ba);
        } else fileLoad=file;
        try {
            molecule=astex.MoleculeIO.readPDB(new astex.FILE(ChUtils.inStream(fileLoad.toString(), STREAM_UNZIP)));
        } catch(Exception ex) {stckTrc(ex); return false; }
        if (molecule==null) return false;
        molecule.setSymmetry(null);
        molecule.setFilename(file.toString());
        molecule.setName(p.getName());
        getMoleculeRenderer().addMolecule(molecule);
        getViewerPanel().addMouseListener(LI);
        getViewerPanel().addMouseWheelListener(LI);
        return true;
    }
    /* <<< Protein <<< */
    /* ---------------------------------------- */
    /* >>> Disposable >>> */
    private boolean _disposed;
    public void dispose() {
        if (!_disposed) {
            _disposed=true;
            final UniqueList v=this.share_vView;
            interpret("molecule remove '"+this.proteinName+"';", (String)null,0L);
            v.remove(this);
            AbstractProxy.disposeProxy(this);
            for (String surface : toStrgArry(vObjects)) addRemoveObject(false,surface,true);
            this.share_ref_astexMV=null;
            this.share_menubar=null;
        }
    }
    /* <<< Disposable <<< */
    /* ---------------------------------------- */
    /* >>> CommandInterpreter >>> */
    private final static String supported[]={
        COMMANDupdate_color_surfaces,
        COMMANDzoom,COMMANDrotate,COMMANDcenter,
        COMMANDspheres,COMMANDlines,COMMANDsticks,COMMANDribbons,COMMANDcolor,COMMANDsurface, COMMANDsurface_color, COMMANDcartoon, COMMANDlabel };
    public boolean getSupportsCommand(String cmd) { return idxOfStrg(cmd,supported)>=0; }
    public void interpret(String command0, String id, long options) {
        //putln("ChAstex#interpret  >"+command0+"<  id="+id+"<");
        final String command=command0.trim();
        final astex.MoleculeRenderer mr=getMoleculeRenderer();
        final astex.Renderer r=mr!=null ? mr.renderer:null;
        if (r==null) return;
        final String word0=toStrgIntrn(frstTkn(command));
        final int L0=sze(word0);
        if (L0==0) return;
        String send=idxOfStrg(word0,COMMANDS)>=0 ? null : rplcToStrg(RESIDUES,id, rplcToStrg(PROTEIN,this.proteinName,command));
        if (getSupportsCommand(word0)) {

            final boolean off=command.endsWith(" off");
            final String
                selExpression=id!=null ? this.share_mapSelectionExpression.get(id):null,
                group=selExpression!=null ? selExpression : id!=null ? " ( group '"+id+"' ) " : " ( molecule \""+this.proteinName+"\" and aminoacid ) ",
                objectPfx= word0==COMMANDribbons ? "ribbon_" : word0==COMMANDcartoon ? "schematic_" :  word0==COMMANDsurface ? "surface_" : null,
                objectName;
            if (objectPfx!=null) {
                final boolean[] boolSelected=id!=null?this.share_mapSelection.get(id):null;
                objectName=objectPfx==null ? null : objectPfx+this.proteinName+"_"+(boolSelected!=null?boolToTxt(boolSelected).replace(',','-'):id);
                if (!off) {

                    addRemoveObject(true,objectName,false);
                    //if (share_jListObjects!=null)    share_jListObjects.revalidate(); share_jListObjects.repaint();

                }
            } else objectName=null;

            final String color;
            if (word0==COMMANDcolor || word0==COMMANDsurface_color) {
              String c=delPfx(word0,command).trim();
              final int nxt=nxt(NO_SPC,c);
              if (chrAt(nxt,c)=='#') c="0x"+c.substring(nxt+1);
              color=c;
            } else color=null;
            if (objectName!=null && off) addRemoveObject(false,objectName,true);
            else if (word0==COMMANDzoom) r.applyZoom((100.0-atoi(command,L0))/100);
            else if (word0==COMMANDsurface_color) { share_surface_color[0]=color; send=null;  }
            else if (word0==COMMANDribbons) send="schematic -name '"+objectName+"' -alltube  true  -ribbonwidth 20 "+group+";\n";
            else if (word0==COMMANDsurface) {
              String c=share_surface_color[0];
              if (c==null) c="0xff00ff";
              send="surface -solid true '"+objectName+"' "+c+" "+group+";\n";
            }
            else if (word0==COMMANDcartoon) send="schematic -name '"+objectName+"' "+group+";\n";
            else if (word0==COMMANDcenter)  send="center  ( sphere 20 around "+group+" ); ";
            else if (word0==COMMANDlabel) {
                final String txt=delPfx(word0,command);
                send="label "+(off?"clear":txt) +  " ( atom CA  and "+group.substring(2)+";\n";
            } else if (word0==COMMANDcolor)  {
                send="color "+color+group+";\n";
            } else if (word0==COMMANDrotate) {
                final int noSpc=nxt(NO_SPC,command,L0,999);
                final char axis=chrAt(noSpc,command);
                final double f=atof(command,noSpc+1);
                if (axis=='x') r.rotateX(f);
                if (axis=='y') r.rotateY(f);
                if (axis=='z') r.rotateZ(f);
            } else if (word0==COMMANDupdate_color_surfaces) {
                for(String o : toStrgArry(vObjects))  updateCartoons(o);
            } else {
                final String s=
                    word0==COMMANDspheres?"spheres":
                    word0==COMMANDlines?"lines":
                    word0==COMMANDsticks?"sticks":
                    null;
                if (s!=null) send="display "+s+(off?" off":" on")+group+";\n";
            }
            if (objectName!=null) share_mapCartoonsCmd.put(objectName,send);
        }
        if (sze(send)>0) {
            putln(ANSI_RED+"Astex>"+ANSI_RESET+send+ANSI_RED+"<"+ANSI_RESET);
            mr.executeInternal(send);
            if (word0==COMMANDcolor) mr.executeInternal("select ( none  );");

        }
        ChDelay.runAfterMS(EDT,threadRepaint,555);
    }
    /* <<< CommandInterpreter <<< */
    /* ---------------------------------------- */
    /* >>> ResidueSelection >>> */
    private final StringBuffer SB=new StringBuffer(99);

    private Thread thread_highlightCalpha(boolean[] bb,String type) {
        final Object[] arg={bb,type};
        return threadM("highlightCalpha",this,arg);
    }
    public void highlightCalpha(boolean[] bb, String type) {
        for(int iA=sze(bb); --iA>=0;) {
            if (!bb[iA]) continue;
            final astex.Residue res=getResidue(iA);
            if (res!=null) {
                for(int i=res.getAtomCount(); --i>=0;){
                    final astex.Atom atom=res.getAtom(i);
                    if ("CA".equals(atom.getAtomLabel())) sphere(atom,type);
                }
            }
        }
    }
    private void sphere(astex.Atom atom, String type) {
        final Component canvas=getViewerPanel();
        final Graphics g=canvas!=null ? canvas.getGraphics() : null;

        if (atom!=null && g!=null) {
            final astex.MoleculeRenderer mr=getMoleculeRenderer();
            final astex.Renderer r=mr!=null ? mr.renderer :null;
            if (r!=null) {
                g.setColor(Color.PINK);
                final int x=atom.xs>>astex.Renderer.FixedBits;
                final int y=atom.ys>>astex.Renderer.FixedBits;
                final int R=EX;
                if(type=="D") g.fillOval(x-R,y-R,2*R,2*R);
                if(type=="B") ChIcon.drawCursor(g, x-R,y-R,2*R,2*R, getViewerPanel());

            }
        }
    }
    public void setSelectedResidues(long options,final boolean[] bb,final int atomType,final String representation_id) {
        if (0!=(options&SET_SELECTED_RESIDUES_HIGHLIGHT)) {
            highlightCalpha(bb,"D");
            ChDelay.runAfterMS(EDT, thread_highlightCalpha(bb,"B"),222);
        }
        {
            final String id=sze(representation_id)>0 ? representation_id : "selection";
            final astex.MoleculeRenderer mr=getMoleculeRenderer();
            final Protein p=getProtein();
            if (mr==null || p==null) return;
            final StringBuffer sb=SB;
            sb.setLength(0);
            sb.append("( molecule ").append('"').append(this.proteinName).append('"');
            final int rn[]=p.getResidueNumber();
            final byte chains[]=p.getResidueChain();
            char chain=0;
            String expression="( )";
            if (firstTrue(bb)>=0) {
                if (rn!=null) {
                    sb.append(" and residue ");
                    for(int i=mini(rn.length, bb.length); --i>=0;) {
                        if (!bb[i]) continue;
                        sb.append(rn[i]).append(' ');
                        if (chains!=null && chains.length>i) chain=(char)chains[i];
                    }
                    if (chain!=0) sb.append(" and chain '").append(chain).append('\'');
                }
                sb.append(" )");
                expression=sb.toString();
                sb.insert(0,"select ").append(";\n");
                sb.append("define ").append(id).append(" current;");
                interpret(sb.toString() , id,0L);
            }
            share_mapSelectionExpression.put(id,expression);
            share_mapSelection.put(id,bb);
            mr.repaint();
        }
    }
    /* <<< ResidueSelection <<< */
    /* ---------------------------------------- */
    /* >>> Ribbons Cartoons Schematic >>> */
    private String share_surface_color[];
    private Map<String,boolean[]> share_mapSelection;
    private Map<String,String> share_mapSelectionExpression;
    private Map<String,String> share_mapCartoonsCmd;
    private List<String> share_vObjects;
    private final List<String> vObjects=new ArrayList();
    private JComponent share_panObjects;
    private ChJList share_jListObjects;
    private void addRemoveObject(boolean add, String ribbon, boolean rmFromList) {
        if (add) {
            adUniq(ribbon,share_vObjects);
            adUniq(ribbon,vObjects);
        } else {
            interpret("object remove '"+ribbon+"';", (String)null,0L);
            if (rmFromList) {
                share_vObjects.remove(ribbon);
                vObjects.remove(ribbon);
            }
        }
        final JList jl=share_jListObjects;
        if (jl!=null) {
            jl.revalidate();
            jl.repaint();
        }
    }

    private JComponent panObjects() {
        if (share_panObjects==null) {
            share_jListObjects=new ChJList(share_vObjects,0L);
            final JComponent panButtons=vBhB(
                                             new ChButton("NEW_RIB").t("Create Ribbon/surface ...").i(IC_BLANK).li(LI),
                                             new ChButton("DEL_OBJECT").t("Delete ribbon or surface").i(IC_KILL).li(LI),
                                             new ChButton("UPDATE_OBJECT").t("Update the color of ribbons").i(IC_COLOR).li(LI)
                                             );
            share_panObjects=newCNSEW(scrllpn(0L,share_jListObjects),"This is a list of ribbons and surfaces",panButtons,null,null);
        }
        return share_panObjects;
    }

    public void updateCartoons(String cartoon) {
        final astex.MoleculeRenderer mr=getMoleculeRenderer();
        if (mr==null || cartoon==null) return;
        final String cmd=share_mapCartoonsCmd.get(cartoon);
        if (cmd!=null) {
            addRemoveObject(false,cartoon,false);
            putln(ANSI_YELLOW+"Astex>"+ANSI_RESET+"  ",cmd);
            mr.executeInternal(cmd);
        }
    }
    /* <<< Objects <<< */
    /* ---------------------------------------- */
    /* >>> Viewer Panel >>> */
    private Reference<Component> share_refPanel;
    private Reference<astex.MoleculeViewer> share_ref_astexMV;
    private UniqueList<ChAstex> share_vView;
    private JMenuBar share_menubar;
    private astex.MoleculeViewer getAstex() { return deref(share_ref_astexMV, astex.MoleculeViewer.class); }
    public ChAstex[] getViewersSharingView() {
        final UniqueList<ChAstex> v=this.share_vView;
        return v!=null ?  v.asArray() : null;
    }
    public Component getViewerPanel() { return share_refPanel!=null ? share_refPanel.get() : null; }
    /* <<< Viewer Panel <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */

    private final ChEventAdapter LI=new ChEventAdapter(this);
    public AWTEvent processEvent(JComponent component, AWTEvent ev) {
        final int id=ev.getID();
        final String cmd=actionCommand(ev);
        final Object q=ev.getSource();
        final astex.MoleculeRenderer mr=getMoleculeRenderer();
        final Protein p=getProtein();
        if (mr==null) return null;
        final int x=x(ev), y=y(ev);
        if (id==MouseEvent.MOUSE_CLICKED) {
            final int iA=toResidueIndex(mr.getNearestAtom(x, y));
            if (iA>=0) new StrapEvent(this,StrapEvent.PROTEIN_VIEWER_PICKED).setAminoAcid(iA,p).run();
        }
        if (cmd=="NEW_RIB") {
            final JMenu m=child(this.share_menubar,JMenu.class, "Style.*");
            //if (m!=null) m.getPopupMenu().show(component, 10,10);
            if (m!=null) m.doClick();
            else ChMsg.msgDialog("Please use the menu \"Style for ... residues\" <br>to create ribbons and surfaces");
        }
        if (cmd=="DEL_OBJECT" || cmd=="UPDATE_OBJECT") {
            final Object oo[]=share_jListObjects.getSelectedValues();
            if (sze(oo)==0) error("You must select at least one ribbon or surface object");
            else {
                for(Object o : oo) {
                    final String ribbon=(String)o;
                    addRemoveObject(false,ribbon, cmd=="DEL_OBJECT");
                    if (cmd=="UPDATE_OBJECT") updateCartoons(ribbon);
                }
            }
        }
        if (cmd=="SHW_OBJECT") {
            ChFrame.frame(getTxt(q), panObjects(),0L).shw(ChFrame.AT_CLICK_R);
        }
        mr.repaint();
        return ev;
    }
    public JMenuBar getMenuBar() { return share_menubar; }
    public long getPropertyFlags() {
        return PROPERTY_NAMED_SELECTIONS|PROPERTY_MULTI_MOLECULE | PROPERTY_MULTI_VIEW | PROPERTY_HEAVY_WEIGHT | PROPERTY_COMMAND_INTERPRETER|
          PROPERTY_RIBBON_COLOR_NO_CHANGE;
    }
    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> Matrix3D >>> */
    private final static WeakHashMap<astex.Molecule, float[]> mapMolecule2xyz=new WeakHashMap();
    public void transform(Matrix3D m3d){
        final astex.MoleculeRenderer mr=getMoleculeRenderer();
        final astex.Molecule molecule=mr!=null ? mr.getMolecule(this.proteinName) : null;
        if (molecule==null) return;
        float xyzORIG[]=mapMolecule2xyz.get(molecule);
        final int N=molecule.getAtomCount();
        if (xyzORIG==null) {
            if (m3d==null || m3d.isUnit()) return;
            mapMolecule2xyz.put(molecule,xyzORIG=new float[N*3]);
            for(int i=mini(N,xyzORIG.length/3); --i>=0;) {
                final astex.Atom atom=molecule.getAtom(i);
                for(int j=3;--j>=0;) xyzORIG[i*3+j]=(float)atom.get(j);
            }
        }
        assert xyzORIG.length==3*N;
        final float[] xyz;
        if (m3d==null || m3d.isUnit()) xyz=xyzORIG;
        else {
            xyz=new float[3*N];
            m3d.transformPoints(xyzORIG, xyz, N);
        }
        for(int i=mini(N,xyz.length/3); --i>=0; ) {
            final astex.Atom atom=molecule.getAtom(i);
            final int i3=3*i;
            atom.set(xyz[i3],xyz[i3+1],xyz[i3+2]);
        }
        for(int i=sze(share_vObjects); --i>=0;) updateCartoons(share_vObjects.get(i));
        interpret("center  ( sphere 20 around ( molecule '"+this.proteinName+"' ) ); ", (String)null,0L);
        mr.repaint();
    }
    public void moleculeAdded(astex.MoleculeRenderer renderer, astex.Molecule molecule){}
    public void genericAdded(astex.MoleculeRenderer renderer, astex.generic.Generic generic){}
    public void genericRemoved(astex.MoleculeRenderer renderer, astex.generic.Generic generic){}
    public void mapAdded(astex.MoleculeRenderer renderer, astex.Map map){}
    public void mapRemoved(astex.MoleculeRenderer renderer, astex.Map map){}
    public void atomSelected(astex.MoleculeRenderer renderer, astex.Atom atom){}
    public void moleculeRemoved(astex.MoleculeRenderer renderer, astex.Molecule molecule){
        //putln("MoleculeRendererListener moleculeRemoved");
        if (molecule!=null && ChAstex.this.proteinName.equals(molecule.getName())) ChAstex.this.dispose();
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */
    private final Thread threadRepaint=ChDelay.newThread(this,"REPAINT",null);
    public Object run(String id, Object arg) {
        if (id=="REPAINT") {
            final astex.MoleculeRenderer mr=getMoleculeRenderer();
            if (mr!=null)  mr.repaint();
            final Component c=getViewerPanel();
            if (c!=null) c.repaint();
        }
        return null;
    }
}
/*
MoleculeRendererListener#atomSelected should be atomPicked
Atom#x  should be private
schematic should move with protein
remove group
how to list groups
how to draw cursor

java.lang.NegativeArraySizeException
	at astex.Molecule.connect2(Molecule.java:836)
	at astex.MoleculeIO.readPDB(MoleculeIO.java:1019)

*/
