package charite.christo.protein;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.awt.*;
import java.io.File;
import javax.swing.*;
import java.util.List;
import java.util.*;
import org.openscience.jmol.viewer.JmolStatusListener;
import org.openscience.jmol.viewer.JmolViewer;
import org.openscience.jmol.app.Jmol;
import javax.vecmath.*;
import static java.lang.System.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static java.awt.event.MouseEvent.*;
/**
   Protein 3D viewer JMol.
   <b>Home:</b> http://jmol.sourceforge.net/<br>
   <b>License:</b> WIKI:LGPL
   @author Christoph Gille
*/
public class ChJmol  implements ProteinViewer, JmolStatusListener, ChRunnable {
    private final ChFrame FRAME=new ChFrame("").i(IC_JMOL).size(20,20);
    /* --- Interface ObservableProteinCursor --- */
    private Jmol jmol;
    private JComponent panel;
    private JMenuBar menuBar;
    public JComponent getViewerPanel() { return panel;}
    public JMenuBar getMenuBar() { return menuBar;}

    /* <<< Fields  <<< */
    /* ---------------------------------------- */
    /* >>> ResidueSelection >>> */
    private static void setSelection(Protein p,boolean selected[],Jmol jmol) {
        final JmolViewer viewer=jmol.viewer;
        final BitSet bs=viewer.getSelectionSet();
        final int atomNo[]=p.getResidueAtomNumber();
        final int n=bs.size();
        for(int i=0;i<n;i++) {
            if (bs.get(i)) {
                final int an=viewer.getAtomNumber(i);
                final int iAa=idxOf(an,atomNo, 0,MAX_INT);
                if (iAa>=0 && iAa<selected.length) { selected[iAa]=true; }
            }
        }
    }
    /* <<< ResidueSelection <<< */
     /* ---------------------------------------- */
    /* >>> JmolStatusListener  >>> */
    public void notifyAtomPicked(int atomIndex, String strInfo){
        if (protein==null) return;
        final int atomNumber=jmol.viewer.getAtomNumber(atomIndex);
        final int iAa=protein.atomNumber2residueIdx(atomNumber);
        if (iAa>=0) new StrapEvent(this, StrapEvent.PROTEIN_VIEWER_PICKED).setAminoAcid(iAa,getProtein()).run();
    }
    public void handlePopupMenu(int x, int y){};
    public void notifyFileNotLoaded(String fullPathName, String errorMsg){};
    public void notifyFileLoaded(String fullPathName, String fileName, String modelName, Object clientFile){};
    public void notifyFrameChanged(int frameNo){};
    public void notifyMeasurementsChanged(){};
    public void notifyScriptTermination(String statusMessage, int msWalltime){}
    public void scriptEcho(String strEcho){ putln("scriptEcho:strEcho=", strEcho); };
    public void scriptStatus(String strStatus){
        if (!isEDT()) {
            ChDelay.invokeAndWait(this,"scriptStatus",strStatus);
            return;
        }
        if (strStatus.indexOf("atoms selected")>0) {
            boolean sel[]=new boolean[protein.countResidues()];
            setSelection(protein,sel,jmol);
            //protein.addResidueSelection(this);
            new StrapEvent(this,StrapEvent.RESIDUE_SELECTION_CHANGED).run();
        }
    }
    public void setStatusMessage(String statusMessage){ putln("setStatusMessage statusMessage=", statusMessage); };
    /* <<< JmolStatusListener <<< */
    /* ---------------------------------------- */
    /* >>> ResidueSelection >>> */
    public void setSelectedResidues(long options,final boolean[] bb,final int atomType,final String representation_id) {
        final Protein protein=getProtein();
        if (protein==null) return;
        if (representation_id==SELECTION_CURSOR) {
            final int iAa=firstTrue(bb);
            if (iAa<0) return;
            final int atomIndex=jmol.viewer.getAtomIndexFromAtomNumber(protein.getResidueAtomNumber(iAa));
            final Point3f p3f=jmol.viewer.getAtomPoint3f(atomIndex);
            final Point3i p3i=p3f!=null ? jmol.viewer.transformPoint(p3f) : null;
            if (p3i==null) return;
            int d=16;
            final JComponent pan=getViewerPanel();
            ChIcon.drawCursor(pan.getGraphics(), p3i.x-d/2,p3i.y-d/2,d,d, pan);
        } else {
            String select=Rasmol.selectCommand(bb,protein,atomType);
            if (select==null) return;
            select=select.replaceAll(": ",":?");
            interpret(select,null,0L);
            jmol.viewer.setSelectionHaloEnabled(true);
        }
    }
    /* <<< ResidueSelection <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Protein >>> */
    private Protein  protein;
    public Protein getProtein() { return protein;}
    public boolean setProtein(Protein p, ProteinViewer inSameView, Runnable runAfterInstallation, long options) {
        if (protein!=null) return false;
        this.protein=p;
        final File pdb=p!=null? p.getFileMayBeWithSideChains() : null;
        if (pdb==null) return false;
        final ChFrame f=FRAME;

        f.setTitle(p.toString());
        mayBeLaunch();
        interpret("load "+pdb,null,0L);
        sleep(999);
        for(String s : getCustomSettings(ChJmolPROXY.class, Customize.CLASS_InitCommands)) interpret(s,null,0L);
        TabItemTipIcon.set(null,p.toString(),null,null,this);
        removeMenuItems(new String[]{"help"}, new String[]{"Jmol Console"},menuBar);
        if (f!=null) {
            f.setVisible(false);
            f.dispose();
        }
        return true;
    }
    private void mayBeLaunch() {
        if (jmol==null) {
            jmol=Jmol.getJmol(FRAME);
            jmol.remove(jmol.getComponents()[2]);
            panel=(JComponent)jmol.getComponents()[1];
            panel=(JComponent)panel.getComponents()[1];
            panel=(JComponent)panel.getComponents()[0];
            if (protein!=null) FRAME.setTitle(protein.toString());
            jmol.viewer.setJmolStatusListener(this);
            jmol.viewer.setColorSelection(Color.PINK);
            try {
                final JMenuBar mb=this.menuBar=child(jmol,JMenuBar.class);
                mb.remove(child(this.menuBar,JMenu.class,"Plugins"));
                Container parent=null;
                JComponent tb=child(jmol,JToolBar.class);
                parent=tb.getParent();
                parent.remove(tb);
                child(this.menuBar,JMenu.class,"Tools").add(tb);
            }catch(Exception ex){stckTrc(ex);}
            //panel.addMouseListener(LI);
            FRAME.setVisible(false);
            ChThread.callEvery(100,threadSend,"ChJmol",EDT);
        }
    }

    /* <<< Protein <<< */
    /* ---------------------------------------- */
    /* >>> ug workaround  >>> */
    private final List jobs=new Vector();
    public Object run(String id,Object arg) {
        if (id=="scriptStatus") scriptStatus((String)arg);
        else if (id=="SEND") {
            if (sze(jobs)>0) ChDelay.invokeAndWait(this,"SEND_EDT",null);
        }
        else if (id=="SEND_EDT") {
            final StringBuffer sJobs=joinLines(jobs.toArray()).append('\n');
            jobs.clear();
            jmol.viewer.evalString(sJobs.toString());
        }
        return null;
    }
    private final Runnable threadSend=ChDelay.newThread(this,"SEND",null);
    /* <<< bug workaround <<< */
    /* ---------------------------------------- */
    /* >>> CommandInterpreter  >>> */
    public boolean getSupportsCommand(String cmd) { return false; }
    public synchronized void interpret(String command,String id,long options) {
        final String s=command+";\n";
        jobs.add(s);
    }
    /* <<< CommandInterpreter <<< */
    /* ---------------------------------------- */
    /* >>> Disposable >>> */
    private boolean _disposed;
    public void dispose() {
        if (!_disposed) {
            _disposed=true;
            ChThread.callEvery(-1,threadSend,"",0);
            AbstractProxy.disposeProxy(this);
        }
    }
    /* <<< Disposable <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */

    public void transform(Matrix3D m){};
    public ProteinViewer[] getViewersSharingView() { return null; }
    public long getPropertyFlags() { return PROPERTY_COMMAND_INTERPRETER; }
}
