package charite.christo.protein;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.awt.*;
import java.lang.ref.*;
import javax.swing.*;
import java.awt.event.*;
import java.util.*;
import static java.lang.System.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import org.pdbj.viewer.gui.ViewerFrame;
import org.pdbj.viewer.molecules.MoleculeFile;
import org.pdbj.viewer.controller.FileStatus;
import static java.awt.event.MouseEvent.*;
import javax.media.opengl.*;
//import
//import org.pdbj.viewer.gui.FileLocator;

/**
   http://en.wikipedia.org/wiki/Java_OpenGL
   http://download.java.net/media/jogl/jogl-2.x-docs/
   @see ChJvPROXY
   @author Christoph Gille
   http://ef-site.hgc.jp/wiki/jV/index.php/Command_List

   Each instance refers to only one protein.
   Several instances can share the pdbj.viewer instance such that several proteins appear in the same view.
   JScrollPane JSplitPane
*/
public class ChJv  implements ProteinViewer, ChRunnable, GLEventListener {
    /* ---------------------------------------- */
    /* >>> Conversions >>> */
    public MoleculeFile[] getMoleculeFiles() {
        final ViewerFrame vf=getViewerFrame();
        return toArry(vf!=null ? vf.getViewerController().getMoleculeFiles():null, MoleculeFile.class);
    }

    private MoleculeFile getMoleculeFile() {
        for(MoleculeFile mf : getMoleculeFiles()) {
             if (this.proteinName.equals(lastPathCompnt(mf.getFilePath()))) return mf;
        }
        return null;
    }

    private MoleculeFile getMoleculeFileForId(int id) {
        for(MoleculeFile mf : getMoleculeFiles())  if (mf.getFileId()==id) return mf;
        return null;
    }
    private Protein getProteinForId(int id) {
        final MoleculeFile mf=getMoleculeFileForId(id);
        final Protein protein=getProtein();
        final ProteinAlignment align=protein!=null ? protein.getAlignment() : null;
        if (align!=null && mf!=null) {
            final String n=lastPathCompnt( mf.getFilePath());
            final int nP=align.countRows();
            for(int i=align.countRows(); --i>=0;) {
                Protein p=align.getProtein(i);
                if (p!=null && p.getName().equals(n)) return p;
            }
        }
        return null;
    }
    public void hideFrame() {
        final JFrame f=getViewerFrame();
        f.setVisible(false);
        f.setExtendedState(JFrame.ICONIFIED);
        f.setDefaultCloseOperation(JFrame.HIDE_ON_CLOSE);
        f.setSize(dim(1,1));
        //f.setUndecorated(true);
    }
    /* <<< Conversions <<< */
    /* ---------------------------------------- */
    /* >>> Protein >>> */
    private Protein protein;
    private String proteinName;

    public Protein getProtein() { return protein;}
    public boolean setProtein(Protein p, ProteinViewer inSameView, Runnable runAfterInstallation, long options) {
        ChMsg.savedMsg("Please read the jV license<br>http://ef-site.hgc.jp/eF-site/servlet/DownloadViewerBinary","jV_license");
        if (protein!=null) return false;
        this.protein=p;
        this.proteinName=p.getName();
        final ChJv inSameViewPO= (ChJv) ( inSameView instanceof ChJv ? (ChJv)inSameView : inSameView instanceof ChJvPROXY ? ((ChJvPROXY)inSameView).pv() : null);
        org.pdbj.viewer.gui.ViewerFrame vf=inSameViewPO!=null ? inSameViewPO.getViewerFrame() : null;
        if (vf==null) {
            vf=new ViewerFrame(new String[0], false) {
                    @Override public void receiveMousePick( org.pdbj.viewer.mouse.PickResult pick) {
                        super.receiveMousePick(pick);
                        ChJv.this.receiveMousePick(pick);
                        hideFrame();
                    }
                    @Override public void fileAdded(ArrayList fileStatusList) {
                        super.fileAdded(fileStatusList);
                        //putln("fileStatusList="+fileStatusList);
                        hideFrame();
                    }
                    @Override public void fileRemoved(ArrayList fileStatusList)  {
                        super.fileRemoved(fileStatusList);
                        ChJv.this.fileRemoved(fileStatusList);
                    }
                    @Override public void statusChanged(ArrayList fileStatusList)  {
                        //putln("fileStatusList="+fileStatusList);
                        super.statusChanged(fileStatusList);
                        hideFrame();
                    }
                };
            this.share_menubar=menuBar2j(vf.getMenuBar(), new String[][]{{"Exit"}});
            {
                share_refCommandArea=weakRef(child(vf.getContentPane(), org.pdbj.viewer.gui.CommandArea.class));
                final Component canvas=vf.getViewerController().getCanvasL();
                this.share_refPanel=weakRef(canvas);
                ((GLCanvas)canvas).addGLEventListener(this);
            }
            this.share_vView=new UniqueList(ChJv.class);
            this.share_refJV=weakRef(vf);
            this.share_isInitialized=new boolean[1];
        } else {
            this.share_menubar=inSameViewPO.share_menubar;
            this.share_refJV=inSameViewPO.share_refJV;
            this.share_refPanel=inSameViewPO.share_refPanel;
            this.share_vView=inSameViewPO.share_vView;
            this.share_isInitialized=inSameViewPO.share_isInitialized;
        }
        adUniq(this, share_vView);
        vf.show();
        if (vf.graphicsReady()) {
            final String cmd="load pdb "+toStrg(p.getFileMayBeWithSideChains());
            vf.executeSingleCommand(cmd);
            //vf.loadPDBFile(toStrg(p.getFileMayBeWithSideChains()), (FileStatus)null);
            //viewerFrame.loadFile(FileType.PDB, FileLocator.fromPath(toStrg(p.getFileMayBeWithSideChains())), !viewerFrame.isScriptMode(), null);
        }
        //for(String s : getCustomSettings(ChJvPROXY.class, Customize.CLASS_InitCommands)) interpret(s,null);
        //ChDelay.runAfterMS(EDT, this,"HIDE", 999);
        hideFrame();
        return true;
    }
    public Object run(String id, Object arg) {
        final ViewerFrame vf=getViewerFrame();
        if (vf==null) return null;
        if (id=="HIDE") hideFrame();
        if (id=="TRANSFORM") {
            putln("TRANSFORM"+isEDT());
            final Matrix3D m3d=(Matrix3D)arg;
            if (isEDT()) transform(m3d);
            else {
                sleep(2222);
                for(int i=100; --i>=0 && !share_isInitialized[0]; ) { putln("sleep"); sleep(333); }
                inEDT(ChDelay.newThread(this,"TRANSFORM",m3d));
            }
        }
        return null;
    }
    /* <<< Protein <<< */
    /* ---------------------------------------- */
    /* >>> Disposable >>> */
    private boolean _disposed;
    public void dispose() {
        if (!_disposed) {
            _disposed=true;
            final UniqueList v=this.share_vView;
            final ViewerFrame vf=getViewerFrame();
            final MoleculeFile mf=getMoleculeFile();
            if (vf!=null && mf!=null) vf.getViewerController().closeFile(new int[]{mf.getFileId()});
            if (v!=null) v.remove(this);
            AbstractProxy.disposeProxy(this);
        }
    }
    /* <<< Disposable <<< */
    /* ---------------------------------------- */
    /* >>> CommandInterpreter >>> */
    private final static String supported[]={
        COMMANDshowScriptPanel,
        COMMANDspheres,COMMANDlines ,COMMANDribbons,COMMANDcolor,  // COMMANDsticks COMMANDsurface
        COMMANDrotate,COMMANDzoom,COMMANDcenter,
        COMMANDshowScriptPanel
    };
    public boolean getSupportsCommand(String cmd) { return idxOfStrg(cmd,supported)>=0; }
    public void interpret(String command,String id, long options) {
        final ViewerFrame vf=getViewerFrame();
        if (vf==null) return;
        final String word0=toStrgIntrn(toStrgTrim(frstTkn(command)));
        final int L0=sze(word0);
        if (L0==0) return;
        String send=idxOfStrg(word0,COMMANDS)>=0 ? null : rplcToStrg(RESIDUES,id, rplcToStrg(PROTEIN,this.proteinName,command));
        if (getSupportsCommand(word0)) {
            final boolean off=command.endsWith(" off");
            final String onOff=off?" off":" on";
            if (word0==COMMANDrotate) send="rotate "+delPfx(word0,command);
            else if (word0==COMMANDzoom)   send="zoom "+delPfx(word0,command);
            else if (word0==COMMANDcenter) send="center "+id;
            else if (word0==COMMANDspheres) send="spacefill "+onOff;
            else if (word0==COMMANDlines) send="wireframe "+onOff;
            else if (word0==COMMANDribbons) send="ribbons "+onOff;
            else if (word0==COMMANDcolor) send="color "+command.substring(L0);
            else if (word0==COMMANDshowScriptPanel) ChFrame.frame("jV",deref(share_refCommandArea,Component.class),ChFrame.SCROLLPANE).shw(ChFrame.AT_CLICK);

        }
        if (command!=null && command.equals("test1") ) {
            GLCanvas canvas=(GLCanvas)getViewerPanel();
            putln("test1 canvas="+canvas);
            putln("test1 Context="+canvas.getContext());
            return;
        }
        if (sze(send)>0) {
            putln(ANSI_RED+"jV>"+ANSI_RESET,send);
            vf.executeSingleCommand(send);
        }
    }
    /* <<< CommandInterpreter <<< */
    /* ---------------------------------------- */
    /* >>> ResidueSelection >>> */
    private String lastSelection;
    public void setSelectedResidues(long options,final boolean[] bb,final int atomType,final String representation_id) {
        final ViewerFrame vf=getViewerFrame();
        final Protein p=getProtein();
        if (vf!=null) {
            final StringBuffer sb=new StringBuffer()/*.append(p).append(',')*/;
            selectionString(bb,p,sb);
            lastSelection=sb.toString();
            sb.insert(0,"select ");
            interpret(sb.toString() ,null,0L);
            final Component canvas=vf.getViewerController().getCanvasL();
            final Graphics g=canvas!=null ? canvas.getGraphics() : null;
            if (g!=null) {
                g.setColor(Color.WHITE);
                g.fillRect(0,0,10,10);
            }
        }
    }
    public static void selectionString(boolean[] bb,Protein protein, StringBuffer sb) {
        final byte[] cn=protein.getResidueChain();
        final int rn[]=protein.getResidueNumber(), iMax=mini(sze(bb), sze(rn));
        boolean isFirst=true;
        for(int i=0;i<iMax; i++)
            if (bb[i]) {
                 if (!isFirst) sb.append(", "); isFirst=false;
                 sb.append(rn[i]);
                 final byte chain=cn!=null && i<cn.length ? cn[i] : 0;
                 if (chain!=0 && chain!=' ') sb.append(':').append((char)chain);
            }
    }
    /* <<< ResidueSelection <<< */
    /* ---------------------------------------- */
    /* >>> Viewer Panel >>> */
    /* vars with prefix "share_" are shared among all ChJv objects that share the same view */
    private Reference<Component> share_refCommandArea;
    private Reference<Component> share_refPanel;
    private Reference<ViewerFrame> share_refJV;
    private UniqueList<ChJv> share_vView;
    private JMenuBar share_menubar;
    private boolean[] share_isInitialized;
    //private static ArrayList avoidGC=new ArrayList();
    private ViewerFrame getViewerFrame() { return share_refJV!=null ? share_refJV.get() : null; }
    public ChJv[] getViewersSharingView() { final UniqueList<ChJv> v=this.share_vView; return v!=null ?  v.asArray() : null; }
    public Component getViewerPanel() { return share_refPanel!=null ? share_refPanel.get() : null; }
    /* <<< Viewer Panel <<< */
    /* ---------------------------------------- */
    /* >>> FileStatusListener  >>> */
    public void fileRemoved(ArrayList fileStatusList)  {
        nextJV:
        for(ChJv jv : getViewersSharingView()) {
            for(FileStatus fs : toArry(fileStatusList, FileStatus.class)) {
                final String fp=fs.getMoleculeFile().getFilePath();
                if (jv.proteinName.equals(lastPathCompnt(fp))) continue nextJV;
            }
            jv.dispose();
        }
    }
    /* <<< FileStatusListener  <<< */
    /* ---------------------------------------- */
    /* >>> MousePickListener >>> */
    public void receiveMousePick(org.pdbj.viewer.mouse.PickResult pick) {
        for(ChJv v:getViewersSharingView()) {
            final Protein p=v.getProteinForId(pick.getFileId());
            if (p!=null && p==v.getProtein()) {
                final int idx=p.residueNumberAndChain2residueIdx(pick.getResidueId(), chrAt(0,pick.getChainId()), pick.getInsCode());
                if (idx>=0) new StrapEvent(this,StrapEvent.PROTEIN_VIEWER_PICKED).setAminoAcid(idx,p).run();
            }
        }
    }
    /* <<< MousePickListener  <<< */
    /* ---------------------------------------- */
    /* >>> AWT >>> */
    public JMenuBar getMenuBar() { return share_menubar; }
    public long getPropertyFlags() {
        return PROPERTY_NAMED_SELECTIONS|PROPERTY_MULTI_MOLECULE | PROPERTY_MULTI_VIEW | PROPERTY_OPENGL | PROPERTY_COMMAND_INTERPRETER;
    }
    /* <<< AWT  <<< */
    /* ---------------------------------------- */
    /* >>> Transform >>> */
    private static void convertMatrix(Matrix3D m3d, org.pdbj.viewer.utils.Matrix4f Y) {
        Y.setIdentity();
        final double[] r[]=m3d.getRotation(), t=m3d.getTranslation();
        Y.m11=(float)r[0][0]; Y.m12=(float)r[0][1]; Y.m13=(float)r[0][2];  Y.m14=(float)t[0];
        Y.m21=(float)r[1][0]; Y.m22=(float)r[1][1]; Y.m23=(float)r[1][2];  Y.m24=(float)t[1];
        Y.m31=(float)r[2][0]; Y.m32=(float)r[2][1]; Y.m33=(float)r[2][2];  Y.m34=(float)t[2];
    }

    public void transform(Matrix3D m3d) {
        putln("\n m3d="+(m3d!=null)+ANSI_GREEN+"  isInitialized="+ANSI_RESET+share_isInitialized[0]);
        if (!share_isInitialized[0]) {
            ChDelay.newThread(this,"TRANSFORM",m3d).start();
            return;
        }
        if (m3d!=null) putln(m3d.toPlainText());
        final MoleculeFile mf=getMoleculeFile();
        if (mf==null) return;
        for(int i=mf.numMolecules();--i>=0;) {
            final org.pdbj.viewer.molecules.Molecule m=mf.getMolecule(i);
            org.pdbj.viewer.images.MoleculeImage mi=m.getImage();
            org.pdbj.viewer.utils.Matrix4f mx4=mi.getTransform();
            if (m3d!=null) convertMatrix(m3d, mx4);
            else mx4.setIdentity();
        }
        getViewerPanel().repaint();

    }
    /*
    public static void printMatrix(org.pdbj.viewer.utils.Matrix4f Y) {
        putln("\nMatrix4f");
        putln(Y.m11+"  "+ Y.m12 +"  "+ Y.m13  +"  "+ Y.m14);
        putln(Y.m21+"  "+ Y.m22 +"  "+ Y.m23  +"  "+ Y.m24);
        putln(Y.m31+"  "+ Y.m32 +"  "+ Y.m33  +"  "+ Y.m34);
        putln(Y.m41+"  "+ Y.m42 +"  "+ Y.m43  +"  "+ Y.m44);
    }
    */
    /* <<< Transform <<< */
    /* ---------------------------------------- */
    /* >>> JOGL see CanvasRenderer >>> */
    //private javax.media.opengl.glu.GLU _glu;
    //private com.sun.opengl.util.GLUT _glut;
    public void display(GLAutoDrawable drawable) {

    }
    public void 	displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) { /* putln("displayChanged "); */ }
    public void 	init(GLAutoDrawable drawable) {
        interpret("zoom 100.01",null,0L);
        share_isInitialized[0]=true;
        ChDelay.revalidate(getViewerPanel(),10);
    }
    public void 	reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {  /* putln("reshape "); */ }
}
/*

http://ef-site.hgc.jp//wiki/jV/images/jV.png

Wishlist
Differentiate between "Invalid command syntax" and "Unknown command"

org.pdbj.viewer.gui.ViewerFrame should not be a JFrame but should have
methods to obtain the JFrame or to obtain a panel.

Exception in thread "AWT-EventQueue-0" java.lang.NullPointerException
	at org.pdbj.viewer.gui.ViewerFrame.init(ViewerFrame.java:237)
	at org.pdbj.viewer.gui.ViewerFrame.access$000(ViewerFrame.java:66)
	at org.pdbj.viewer.gui.ViewerFrame$1.windowOpened(ViewerFrame.java:160)

 */

