package charite.christo.protein;
import charite.christo.interfaces.*;
import java.util.*;

/**
   A protein viewer can display proteins three-dimensionally.
   @author Christoph Gille
*/
public interface ProteinViewer extends Disposable,HasProtein, CommandInterpreter {

    String PROTEIN="PROTEIN", RESIDUES="RESIDUES", SELECTION_CURSOR="cursor", SELECTION_PICKED="picked";
    String KEY_PROTEIN_VIEWER="PV$$KPV", KEY_FRAME_TITLE="PV$$FT";
    long PROPERTY_MULTI_MOLECULE=1<<0, PROPERTY_MULTI_VIEW=1<<2, PROPERTY_EXTERNAL_PROCESS=1<<3, PROPERTY_TK=1<<4;
    long PROPERTY_HEAVY_WEIGHT=1<<5, PROPERTY_OPENGL=1<<6;
    long PROPERTY_COMMAND_INTERPRETER=1<<7, PROPERTY_NOTIFY_ALIGNMENT_CURSOR_DELAY=1<<8;
    long PROPERTY_CAN_ZOOM=1<<20, PROPERTY_CAN_ROTATE=1<<21;
    long PROPERTY_NAMED_SELECTIONS=1<<30, PROPERTY_RIBBON_COLOR_NO_CHANGE=1<<31;

    long SET_SELECTED_RESIDUES_HIGHLIGHT=1;

    long INTERPRET_NO_MSG_DIALOGS=1<<1;
    ProteinViewer[] NONE={};

    String
        COMMANDopen="3D_open", COMMANDclose="3D_close", COMMANDtoFront="3D_to_front", COMMANDbiomolecule="3D_biomolecule",
        COMMANDzoom="3D_zoom", COMMANDrotate="3D_rotate", COMMANDcenter="3D_center",
        COMMANDribbons="3D_ribbons", COMMANDspheres="3D_spheres", COMMANDsticks="3D_sticks", COMMANDlines="3D_lines",
      COMMANDsurface="3D_surface", COMMANDsurface_color="3D_surface_color",
        COMMANDmesh="3D_mesh", COMMANDdots="3D_dots", COMMANDcartoon="3d_cartoon",
        COMMANDcolor="3D_color",
        COMMANDupdate_color_surfaces="3D_update_color_surfaces",
        COMMANDlabel="3D_label",
        COMMANDscript="3D_script_define", COMMANDsendScript="3D_script_send", COMMANDshowScriptPanel="3D_script_panel",
        STYLE_COMMANDS[]={COMMANDribbons,COMMANDspheres,COMMANDsticks,COMMANDlines, COMMANDsurface,    COMMANDdots, COMMANDmesh,COMMANDcartoon},
        COMMANDS[]={COMMANDopen,COMMANDclose,COMMANDtoFront, COMMANDbiomolecule,
                    COMMANDzoom, COMMANDrotate,COMMANDcenter,
                    COMMANDribbons,COMMANDspheres,COMMANDsticks,COMMANDlines, COMMANDsurface,
                    COMMANDdots, COMMANDmesh,COMMANDcartoon,
                    COMMANDcolor, COMMANDupdate_color_surfaces,
                    COMMANDscript, COMMANDsendScript,COMMANDshowScriptPanel,
                    COMMANDlabel};

    WeakHashMap<Object,java.lang.ref.Reference<ProteinViewer>> mapViewer=new WeakHashMap();
    HashSet<Class> SHOW_CURSOR_IF_OTHER_PROTEIN=new HashSet();

/**
       @param bb {false,true,true} means select exactly the snd and 3rd
       @param representation_id Vmd for instance requires identifiers for each representation. It is a String starting with a letter and containing only letters and numbers. Rasmol for Example ignors this String
       @param atomType ATOM_ANY or ATOM_CA
*/

    void setSelectedResidues(long options,boolean[] bb,int atomType,String representation_id);
    /**
       The only use of the setProtein and getProtein methods are to
       to keep a reference to the protein object.
       Otherwise both methods are not important.
    */
    boolean setProtein(Protein p, ProteinViewer inSameView,  Runnable runAfterInstallation, long options);

    void transform(Matrix3D m);
    /**
        If non null then the viewer will be displayed in the STRAP Frame
    */
    java.awt.Component getViewerPanel();

    /**
        Each instance is connected with only one protein.
        But several instances can share the view such that several proteins are displayed in the same
        panel.
    */
    ProteinViewer[] getViewersSharingView();
    /** Either null or a javax.swing.JMenuBar or a java.awt.MenuBar */

    javax.swing.JMenuBar getMenuBar();

    long getPropertyFlags();

    boolean getSupportsCommand(String cmd);
}

