package charite.christo.protein;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.util.*;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
/**HELP
WIKI:Rasmol is a fast  three-dimensional protein structure viewer which is  easy to use
<br><br><b>Compatibility:</b>
This wrapper class should work on all UNIX/LINUX computers, but does not support  Windows.
<br><br><b>Requirements:</b>
The two required programs are usually present on UNIX-machines
<ul>
  <li>xterm</li>
  <li>wish (TCL-TK)-interpreter</li>
</ul>

   @author Christoph Gille
*/
public class Rasmol extends AbstractExecutable implements ProteinViewer, ActionListener {
    {
        setName("rasmol_"+Toolkit.getDefaultToolkit().getColorModel().getPixelSize()+"BIT");
        if (isSystProprty(IS_LINUX386)) setBinaryPackageURLs(DEFAULT_BINARY);
    }
    //  getCustomizableInitCommands
    public static String getCustomizableInitCommands() {
        return "cartoon\n"+"select hetero and not water\n"+"cpk\n"+"select all\n"+"wireframe off";
    }
    private String   tkWindow=" ";
    public Protein getProtein() {return protein;}
    //end of source text which can be customized
    private Protein protein;
    public Rasmol() {
        ProteinUtils.initCustomizableSettings(this);
    }
    // ----------------------------------------
    public void setSelectedResidues(long options,boolean[] b,final int atomType, String representation_id_notNeeded) {
        //selected=b;
        interpret( selectCommand(b,protein,atomType),null,0L);
    }
    public static void selectionString(boolean[] bb,Protein protein, StringBuffer sb) {
        final int rn[]=protein.getResidueNumber();
        final byte[] cn=protein.getResidueChain();
        if (bb==null || rn==null) return;
        int i,iMax=bb.length;
        if (iMax>rn.length) iMax=rn.length;
        boolean isFirst=true;
        for(i=0;i<iMax; i++)
            if (bb[i]) {
                int from=i;
                char cnFrom=0,cnTo=0;
                while(i<iMax && bb[i]) {
                    cnFrom=cnTo=' ';
                    try {
                        cnFrom=(char)cn[from];
                        cnTo=(char)cn[i];
                    } catch(Exception e){}
                    if (cnFrom!=cnTo) break;
                    i++;
                }
                if (!isFirst) sb.append(','); isFirst=false;
                sb.append(rn[from]);
                if (from<i-1) sb.append("-").append(rn[i-1]);
                if (cnFrom!=0) sb.append(":").append(cnFrom);
            }
    }
    public static  String selectCommand(boolean[] bb,Protein protein, final int atomType) {
        final StringBuffer sb=new StringBuffer(99).append("select ");
        selectionString(bb,protein, sb);
        if (sb.length()>7) {
            if ( (atomType & ATOM_CA) !=0) sb.append(" and alpha ");
            return sb.toString();
        } else return "select 0";
    }
    // ---
    @Override public void dispose() {
        interpret("quit",null,0L);
        super.dispose();
    }
    // ----------------------------------------
    private JComponent panel, shared;
    private static ChTextField tfDisplay;
    public Container getSharedControlPanel() {
        if (shared==null) {
            tfDisplay=new ChTextField().cols(60,true,true);
            shared=vB("Optionally enter an X-display (Example 141.42.56.114:0.0)", tfDisplay);
        }
        return shared;
    }

    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        if (q instanceof ChTextField) interpret(((ChTextField)q).getText(),null,0L);
        if (cmd==ChExec.ACTION_TERMINATED) {
            new StrapEvent(this,StrapEvent.PROTEIN_VIEWER_CLOSED).run();
        }
    }
    // ----------------------------------------
    private boolean launch() {
        final String display=tfDisplay!=null?toStrg(tfDisplay) : "";
        if (isMacintosh() || isWindows()) {error("Rasmol is not integrated in STRAP yet.\nUse Pymol or Jmol");  return  false;}

        installPackage();
        //        if (!fileExecutable().exists()) ChMsg.msgDialog(FindUtils.infoBox(new String[]{getName()+".jar"},dirBinaries()));
        String oldWish= Wish.getWinfo(display);
        iniT();
        final ChExec exec=exec().printlnStdout((ChRunnable)this).setCommandLine(fileExecutable()).setOptions(ChExec.STDOUT|ChExec.STDERR);
        addActLi(this,exec);
        new Thread(exec,"exec Rasmol").start();
        exec.waitFor();
        final String error=
            exec.hasFinished() ? "could launch Rasmol but the prosess  already terminated" :
            !exec.isRunning() ? "could not launch Rasmol" : null;
        if (error!=null) {
            error(0L,"",vBpnl("Error in Rasmol.java: <br>"+error, getCtrlPanel(exec)));
            return false;
        }
        final String newWish= Wish.getWinfo(display);
        putln("Rasmol: newWish="+newWish+" oldWish="+oldWish+ " display="+display);
        if (newWish!=null) {
            tkWindow=newWish.substring(oldWish.length());
            tkWindow=rplcToStrg("{","",tkWindow);
            tkWindow=rplcToStrg("}","",tkWindow).trim();
            return true;
        } else return false;
    }
    // ----------------------------------------
    //    private File pdb;
    public long getPropertyFlags() { return PROPERTY_EXTERNAL_PROCESS|PROPERTY_TK|PROPERTY_NOTIFY_ALIGNMENT_CURSOR_DELAY; }
    public boolean setProtein(Protein p, ProteinViewer inSameView, Runnable runAfterInstallation, long options) {
        protein=p;
        final File pdb=p!=null?p.getFileMayBeWithSideChains() : null;
        if (pdb==null) return false;
        if (!launch()) return false;
        interpret("load "+pdb,null,0L);
        //for(String s : getCustomSettings(getClass(), Customize.CLASS_InitCommands)) interpret(s,null);
        return true;
    }
    // ----------------------------------------+-----------
    public boolean getSupportsCommand(String cmd) { return false; }
    public void interpret(String s,String id, long options) {
        final String display=tfDisplay!=null?toStrg(tfDisplay) : "";
        Wish.wish(display,tkWindow,s);
    }

    public static String getCustomizableExamples() { return ProteinUtils.RASMOL_EXAMPLES; }

    @Override public Object run(final String id, Object arg) {
        if (id==ChRunnable.INTERPRET_LINE && arg!=null){
            final String s=toStrg(arg).trim();
            final int l=s.length();
            final int num=atoi(s,s.indexOf("Group: ")+6);
            final int iChain=s.indexOf("Chain: ");
            final char chain=iChain>0 && iChain+7<l ? s.charAt(iChain+7) : 0;
            residueIndexCursor=protein.residueNumberAndChain2residueIdx(num,chain,' ');
            new StrapEvent(this,StrapEvent.PROTEIN_VIEWER_PICKED).setAminoAcid(residueIndexCursor,getProtein()).run();
        }
        super.run(id,arg);
        return null;
    }
    private int residueIndexCursor=-1;

    public void transform(Matrix3D m){}
    public Component getViewerPanel() {return null;}
    public ProteinViewer[] getViewersSharingView() { return null; }
    public JMenuBar getMenuBar() { return null; }
}
/*
  Testers are needed for a new windows release of RasMol.
  Herbert J. Bernstein" <yaya@bernstein-plus-sons.com
*/
