package charite.christo.protein;

/**HELP
<i>PACKAGE:charite.christo.strap.</i>

Residue selections are objects attached to proteins which select one or more amino acid positions from the amino acid sequence.

They usually have color, style attributes and a balloon message. They may be shown in
the alignment (<b>1D</b>) and in the built in 3D-wire-frame
(<b>3D</b>) and the scrollbar view-port (<b>VP</b>).  Selections are created in different ways:

<ul>

<li><b>Mouse selection:</b> Dragging the mouse in the alignment pane selects a continuous chain of residues which  may be pasted e.g. into a BLAST web form.
</li>

<li>The dialog <i>DIALOG:charite.christo.strap.DialogHighlightPattern</i>.
All occurrences of the patterns will be highlighted in the selected color.
</li>

<li><b>3D-Visualizations:</b> Clicking  a 3D-backbone atom creates selections.</li>

<li>Some dialogs create selection objects:
<ol>
<li><i>DIALOG:DialogDotPlot</i>.</li>
<li><i>DIALOG:DialogSelectionOfResiduesMain</i>.</li>
</ol>

<li><b>User extension:</b> Users familiar with
Java-Programming may implement their own algorithms. See <i>DIALOG:charite.christo.strap.StrapHotswap</i>.</li>

<li><b>Residue annotations:</b>
Annotations are residue selections that can contain text attributes. They are saved to HD.
Simple selections can easily be transfered into residue annotations
using
<i>ITEM:ResidueSelectionPopup#ACTION_convertToResidueAnnotation</i>

</li>

<li><b>Sequence features:</b>
Sequence features like glycosylation are residue annotations obtained from remote servers.
</li>

</ul>

@author Christoph Gille
*/
public interface ResidueSelection {
    String NAME_MOUSE_OVER="mouse over", NAME_BACKBONE="Backbone",
        CURSOR="Cursor";
/**
       Usually, an instance variable is set with the given protein.
       @param p a reference to the protein object.
*/
    boolean setProtein(Protein p);
/**
       Usually, the protein reference which is stored in an instance variable is returned.
       @return a reference to the protein object.
*/
    Protein getProtein();
/**
       @return  true for each selected residue.
*/
    boolean[] getSelectedAminoacids();
    int getSelectedAminoacidsOffset();
/**
       The selected residues might be initialized.
       Usually the method body can be kept empty.
*/
    void setSelectedAminoacids(boolean bb[],int offset);
       final static ResidueSelection[] NONE={};
}
