package charite.christo.protein;
import java.util.*;
import charite.christo.*;
import java.io.*;
import static charite.christo.ChUtils.*;
/**
   @author Christoph Gille
*/
public interface SimilarPdbProvider {
    Result[] NOT_FOUND={};
    int MIN_FILE_SIZE=10;
    void get(String[] proteinID, Map<String,Result[]> m, Map<String, Result[]> vNew);
    Class CLASSES[]={SimilarPdbProviderSSMap.class};
    
    public static class Result {
        public float identity;
        public String id, pdbID;
        public int from, to;
    }
    public static class Utils {
        public static void parse(ByteArray txt, Map<String,Result[]> m, Map<String,Result[]> mNew) {
            if (txt==null || chrAt(0,txt)=='<')  return;
            final byte[] T=txt.getBytes();
            final int ends[]=txt.getLineEnds();
            final int[] TABS=new int[7];
            for(int iL=0; iL<ends.length; iL++) {
                final int b= iL==0 ? 0: ends[iL-1]+1, e=ends[iL];
                final int nTab=tabulatrs('\t', T, b, e, TABS);
                if (nTab<2 || TABS[0]-b<5) continue;
                for(int i=TABS.length; --i>=0; ) if (TABS[i]<0) TABS[i]=e;
                final String id=txt.newString(b,TABS[0]);
                //putln("id="+id+" nTab="+nTab);
                final Result r=new Result();
                r.id=id;
                r.pdbID=txt.newString(TABS[0]+1,TABS[1]);
                r.identity=(float)atof(txt, TABS[1]+1);
                final int dash=strchr('-',T,TABS[2],TABS[3]);
                final int from=dash>0 ? atoi(T,TABS[2]+1) : 0;
                final int to=dash>0 ? atoi(T,dash+1) : 0;
                r.from=from;
                r.to=to-from>0?to : Integer.MAX_VALUE;
                if (mNew!=null) mNew.put(id, ad(r, mNew.get(id),10,Result.class ));
                m.put(id, ad(r,m.get(id),10,Result.class ));
            }
        }
        public static ByteArray toText(Result rr[], ByteArray ba0) {
            ByteArray ba=ba0;
            if(rr!=null) for(Result r : rr) ba=toText(r,ba);
            return ba;
        }

        public static ByteArray toText(Result r, ByteArray ba0) {
            if (r==null) return ba0;
            final ByteArray ba=ba0!=null ? ba0 : new ByteArray(99);
            ba.append(r.id).append('\t').append(r.pdbID).append('\t').append(r.identity).append('\t')
                .append(r.from).append('-').append(r.to).append('\n');

            return ba;
        }
        public static File[] getDirectories() {
            if (dirs_==null) {
                dirs_=new File[CLASSES.length];
                mFile=new Map[CLASSES.length];
                for(int iC=CLASSES.length; --iC>=0;) {
                    dirs_[iC]=file(dirWeb()+"/"+lastComponent(CLASSES[iC].getName(),'.'));
                    mFile[iC]=new HashMap();
                }
            }
            return dirs_;
        }

        private static Map<String,File> mFile[];
        private  static File dirs_[];
        public static File getFile(String id, Class clazz, boolean isWin) {
            final File[] dirs=getDirectories();
            final int iClazz=idxOf(clazz,CLASSES);

            File f=mFile[iClazz].get(id);
            assertNotNull(id);
            if (f==null) mFile[iClazz].put(id,f=new File(dirs[iClazz], id.replace(':','_')));
            return f;
        }

        public static void notFound(String ids[], Map m, Map mNew, SimilarPdbProvider inst) {
            final Class clazz=inst.getClass();
            final boolean isWin=isWindows();
            for(String id : ids) {
                if (id==null) continue;
                if (m.get(id)==null) {
                    m.put(id,NOT_FOUND);
                    if (mNew!=null) mNew.put(id,NOT_FOUND);
                    wrte( getFile(id, clazz, isWin),"nothing"); // must be shorter than MIN_FILE_SIZE
                }
            }
        }

    }

}