package charite.christo.protein;
import charite.christo.*;
import charite.christo.interfaces.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;

/**HELP

 The fasta format is characterized by a <b>greater than</b> character followed by the header ( WIKI:FASTA_format).
*/
public class SingleFastaParser implements ProteinParser {
    public boolean parse(Protein prot, long options, ByteArray ba) {
        final byte[] txt=ba.getBytes();
        final int E=ba.getEnd(), B=ba.getBegin();
        if (B>=E || txt[B]!='>') return false;
        int nl=B+1, nameLength=0;
        boolean id=true;
        while(nl<E) {
            final byte c=txt[nl];
            if (id && !( 'A'<=c && c<='Z' || '0'<=c && c<='9' || c=='_')) {
                id=false;
                if (nl-B>6) prot.addSequenceRef(new String(txt,B+1,nl));
            }
            if (c=='\n') {
                final int firstIdx=strstr(STRSTR_AFTER,"FIRSTINDEX=",txt,B,nl);
                if (firstIdx>0) prot.setFirstResidueIndexOffset(atoi(txt,firstIdx));
                int idxPDB=B;
                while(  (idxPDB=strstr(STRSTR_AFTER,"PDB:",txt,idxPDB,nl))>0) {
                    int endPdb=idxPDB;
                    while(endPdb<nl && (is(LETTR_DIGT,txt[endPdb])||txt[endPdb]==':')) endPdb++;
                    prot.addSequenceRef(new String(txt,idxPDB-4,endPdb-idxPDB+4));
                }
                final int idxIDs=strstr(STRSTR_AFTER,"ALL_IDS=",txt,B,nl);
                if (idxIDs>0) {
                    int to=idxIDs;
                    while(to<nl && txt[to]!=' ') to++;
                    for (String s : new String(txt,idxIDs,to-idxIDs).split(",")) prot.addSequenceRef(s);
                }
                break;
            }
            if (is(LETTR_DIGT,c)) nameLength++;
            nl++;
        }

        if (nl>=E || nameLength==0) return false;
        final byte[] aa=onlyLetters(txt, nl,E);

        final int keggOrganis=strstr(STRSTR_AFTER,"<title>KEGG ",txt,0,E);
        if (keggOrganis>0) {
            prot.setOrganism(ba.newString(keggOrganis, strchr(':',txt,keggOrganis,keggOrganis+99)));
            final int spc=nxtE(SPC,txt,B+1,nl);
            //if (spc>0) prot.setAccessionID(( onlyACTGUN(aa,0,MAX_INT) ? "KEGG_NT:" : "KEGG_AA:") + ba.newString(B+1,spc));
            if (spc>0) prot.setAccessionID(( nxt(NO_ACTGUN,aa)<0 ? "KEGG_NT:" : "KEGG_AA:") + ba.newString(B+1,spc));

            final int semicolon1=strchr(';',txt,spc,nl);
            prot.setCompound(ba.newString(nxt(NO_SPC,txt,semicolon1+1,nl), strchr(STRSTR_E,';',txt,semicolon1+1,nl)));
            final int ec=strstr(STRSTR_AFTER,"(EC:",txt,B,nl);
            if (ec>0) prot.setEC(ba.newString(ec, strchr(')',txt,ec,nl)));
        }

        if (aa.length>0) { prot.setResidueType(aa); return true;}

        return false;
    }
}
