package charite.christo.protein;
import charite.christo.*;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
// java charite.christo.protein.SwissHeaderParser
/**
   The header of swissprot files contains the definition, organism ... .

Asscssion Number:
AC   P04585; O09777; Q9WJC5;

EC-Number
DE   (EC 2.7.7.7) (EC 3.1.26.4)

CDS
FT   CDS             complement(join(1787209..1787388,1787427..1788041,
FT   CDS             1819165..1821225
     CDS             join(5721..5921,6225..6368,6657..6813,6945..6972,
                     7581..7689,7772..7864,7968..8031,8450..8548,8787..8890,
...
                     26572..26589)
                     /gene="MYH7"
                     /note="cardiac myosin"
                     /codon_start=1
                     /product="beta-myosin heavy chain"
                     /protein_id="AAA51837.1"
                     /db_xref="GI:179508"
                     /db_xref="GDB:G00-120-215"

in ENSMUSG00000014704 it  looks  CDS             join(complement(2261..2639),complement(881..1620))

<br><br>
<b>Format Specification:</b> http://www.expasy.org/sprot/userman.html http://www.uniprot.org/manual
   @author Christoph Gille
*/
public class SwissHeaderParser {
    private SwissHeaderParser(){}
    private static boolean[] delim1;
    private final static List vTemp=new ArrayList();
    private final static ByteArray SB=new ByteArray(0);
    public static boolean parse(Protein p, ByteArray ba) {
        ba.delBlanksL();
        final byte[] T=ba.getBytes();
        final int  begin=ba.getBegin(), ends[]=ba.getLineEnds(), end=ba.getEnd();
        final boolean isDNA=p.containsOnlyACTGNX() || p instanceof ProteinNT && ((ProteinNT)p).getNucleotides()!=null;
        ChTokenizer tokenizer=null;
        byte ss[]=p.getResidueSecStrType();
        if (isDNA) parseCDS(p,ba);
        {
            final int b=begin;
            if (end-b<20) return false;
            if (T[b]=='L' && equls("LOCUS ",T,b)) {  return parseNCBI(p,isDNA,ba); }
        }
        collectEntries('D','E',SB.setEnd(0),ba);
        p.setCompound(SB.newString());
        collectEntries('O','S',SB.setEnd(0),ba);
        if (SB.getEnd()>0) {
            final int parOpen=strstr(" (",SB);
            int parClose=-1;
            if (parOpen>1) {
                parClose=strstr( ") (",SB,parOpen, MAX_INT); /*OS   Danio rerio (Zebrafish) (Brachydanio rerio).*/
                if (parClose<0) parClose=strstr(").",SB,parOpen, MAX_INT);
            }
            final String os=parClose<0?SB.delSuffix(".").newString() : SB.newString(0,parOpen);
            p.setOrganismScientific(os);
            if (parClose>0) p.setOrganism(SB.newString(parOpen+2,parClose));
            SB.setEnd(0);
        }
        findEC_Number(p,ba);
        boolean isEnsemble=false;
        String accession=null;
        for(int pass=0, countCDS=sze(p.getCDS());  pass<2;  pass++) {
            for(int iL=0; iL<ends.length; iL++) {
                final int b= iL==0 ? begin : ends[iL-1]+1,  e=prev(NO_SPC,T,ends[iL]-1,b-1)+1;
                if (e-b<6 || T[b+2]!=' ') continue;
                final byte c0=T[b], c1=T[b+1];
                isEnsemble=isEnsemble || c0=='D' && c1=='E' && equls("DE   annotated by Ensembl",T,b);
                /* --- FT                   /db_xref="PDB:1IRU" --- */
                /* --- FT                   /db_xref="UniProtKB/TrEMBL:Q9FDF2" */
                if (c0=='F' && c1=='T' && countCDS<2) { /* --- No refs for DNA-files more than one CDS --- */
                    final int dbref=strstr(STRSTR_AFTER,"/db_xref=",T,b,e);
                    if (dbref>0) {
                        int idx=-1;
                        final String DB=
                            (idx=strstr(STRSTR_AFTER,"Swiss-Prot:",T,dbref,e))>0 ? "UNIPROT:" :
                            (idx=strstr(STRSTR_AFTER,"TrEMBL:",T,dbref,e))>0 ? "TREMBL:" :
                            (idx=strstr(STRSTR_AFTER,"PDB:",T,dbref,e))>0 ? "PDB:" :
                            null;
                        if (DB!=null) {
                            final int quote=strchr('"',T,idx,e);
                            if (quote>0) p.addSequenceRef(DB+ba.newString(idx,quote));
                        }
                    }
                }
                if (pass==0) continue;

                if (c0=='F' && c1=='T') {
                    final char c=equls("FT   HELIX ",T,b) ? 'H' :equls("FT   STRAND ",T,b) ? 'E' : equls("FT   TURN ",T,b) ? 'T' : ' ';
                    if (c!=' ' && ss==null) p.setResidueSecStrType(ss=new byte[p.countResidues()]);
                    if (ss!=null) {
                        final int i1=mini(ss.length, atoi(T,b+14)), i2=mini(ss.length,atoi(T,b+21));
                        if (i1>0 && i2>0 && i2>i1) Arrays.fill(ss, i1-1, i2,(byte)c);
                    }
                }
                if (c0=='I' && c1=='D' || c0=='A' && c1=='C') {
                    final String db=isEnsemble ? "ENSEMBL:" : isDNA ? "EMBL:" : "UNIPROT:"; // !!!!!!!!!
                    final String acc=db+tokn(T,b+3,e," ;,.");
                    if (accession==null && c0=='A') accession=acc;
                    p.addSequenceRef(acc);
                    /* AC   Q86UR5; A7MBN6; O15048; Q5JY25; Q5SZK1; Q8TDY9; Q8TDZ5; Q9HBA1; */
                    if (T[e-1]==';') {
                        final int semikolon=strchr(';',T,b+3,e-3);
                        if (semikolon>0) {
                            if (delim1==null) delim1=charClass(" ;,.");
                            if (tokenizer==null)  tokenizer=new ChTokenizer().setDelimiters(delim1);
                            tokenizer.setText(T,b+3,e-1);
                            while(tokenizer.nextToken()) {
                                p.addSequenceRef(db+tokenizer.asString());
                            }
                        }
                    }
                }
                if (c0=='D' && c1=='R') {
                    {
                        final int idx=strstr(STRSTR_AFTER,"PDB; ",T,b+3,e);
                        if (idx>0) {
                            final int semi=strchr(';',T,idx,e);
                            if (semi>0) p.addSequenceRef("PDB:"+ba.newString(idx,semi));
                        }
                    }
                    for(int i=0;i<4;i++) {
                        final String DB,db;
                        switch(i) {
                        case 0: db="Pfam; "; DB="PFAM"; break;
                        case 1: db="PRINTS; "; DB="PRINTS"; break;
                        case 2: db="ProDom; "; DB="PRODOM"; break;
                        case 3: db="PROSITE; "; DB="PROSITE"; break;
                        default: DB=db="PROSITE; ";
                        }
                        final int idx=strstr(STRSTR_AFTER,db,T,b+3,e);
                        if (idx>0) {
                        final int semi=strchr(';',T,idx,e);
                        if (semi>0) p.addDatabaseRef(DB+":"+ba.newString(idx,semi));
                        }
                    }
                }
            }
        }
        if (accession!=null) p.setAccessionID(accession);
        return true;
    }
    private static void findEC_Number(Protein prot, ByteArray ba) {
        final int ends[]=ba.getLineEnds();
        final byte[] T=ba.getBytes();
        vTemp.clear();
        for(int iL=0;iL<ends.length; iL++) {
            final int b= iL==0 ? ba.getBegin() : ends[iL-1]+1;
            final int e=ends[iL];
            if (e-b<10 || T[b]!='D' || T[b+1]!='E') continue;
            int pos=b;
            while( (pos=strstr(STRSTR_AFTER,"(EC ",T,pos+1,e))>0) {
                while(pos<e && T[pos]==' ') pos++;
                final int close=strchr(')',T,pos,e);
                if (close>pos) vTemp.add(new String(T,pos,close-pos));
            }
        }
        final int nEC=sze(vTemp);
        if (nEC>0) prot.setEC(toArry(vTemp, String.class));
    }

    private static void collectEntries(char k0, char k1,ByteArray sb, ByteArray ba) {
        final int  ends[]=ba.getLineEnds();
        final byte[] T=ba.getBytes();
        loop_lines:
        for(int iL=0; iL<ends.length; iL++) {
            final int b= iL==0 ? ba.getBegin() : ends[iL-1]+1,  e=prev(NO_SPC,T,ends[iL]-1, b-1)+1;
            if (e-b<5 || k0!=T[b] || k1!=T[b+1] || ' '!=T[b+2]) continue;
            final int nxtNoSpc=nxt(NO_SPC,T,b+2,e);
            if (nxtNoSpc>0) {
                if (sb.getEnd()>0) sb.append(' ');
                sb.append(T,nxtNoSpc, e);
            }
        }
    }
    private static boolean parseNCBI(Protein p, boolean isDNA, ByteArray ba) {
        final byte[] T=ba.getBytes();
        final int begin=ba.getBegin(), ends[]=ba.getLineEnds();
        for(int iL=0; iL<ends.length; iL++) {
            final int b0= iL==0 ? begin : ends[iL-1]+1,  e=prev(NO_SPC,T,ends[iL]-1,b0-1)+1;
            final int b=nxt(NO_SPC,T,b0,e);
            if (b<0 || e-b<9) continue;
            final byte c0=T[b];
            if (c0=='O' && equls("ORGANISM ",T,b))  p.setOrganismScientific(strgTrim(T,nxt(NO_SPC,T,b+8,e),e));
            if (c0=='A' && equls("ACCESSION ",T,b))  p.setAccessionID("ENTREZ:"+tokn(T,nxt(NO_SPC,T,b+9,e),e));
            if (c0=='D' && equls("DEFINITION ",T,b)) p.setCompound(strgTrim(T,b+10,e));
            if (c0=='V' && equls("VERSION ",T,b)) { /* --- VERSION     M57965.2  GI:24429600 --- */
                final int gi=strstr("GI:",T,b+7,e);
                if (gi>0) p.addSequenceRef(tokn(T,gi,e));
            }
            if (c0=='D' && equls("DBSOURCE ",T,b)) {  /* --- DBSOURCE    pdb: molecule 1CF1, chain 68,  --- */
                final int pdb=strstr(STRSTR_AFTER,"pdb: molecule",T,b+7,e);
                if (pdb>0) {
                    final int iChain=strstr(STRSTR_AFTER,", chain ",T,pdb,e);
                    final int chain=iChain>0 ? atoi(T,iChain) :0;
                    p.setPdbID("PDB:"+tokn(T,pdb,e,", ").toLowerCase()+ (chain>='0' ? ":"+(char)chain:""));
                }
            }
        }
        return true;
    }
    private final static ByteArray baCDS=new ByteArray(0);
    private static void parseCDS(Protein p, ByteArray ba) {
        final byte T[]=ba.getBytes();
        final int[] ends=ba.getLineEnds();
        boolean found=false;
        final StringBuffer sbXref=new StringBuffer(333);
        for(int source=0; source< (found?1:2); source++) {
            for(int iL=0; iL<ends.length; iL++) {
                final int b= iL==0 ? ba.getBegin() : ends[iL-1]+1, e=prev(NO_SPC,T,ends[iL]-1,b-1)+1;
                if (e-b<6 || !(T[b]=='F'&& T[b+1]=='T' || T[b]==' '&& T[b+1]==' ')) continue;
                final int nonSpc=nxt(NO_SPC,T, b+2,e);
                if (nonSpc<0) continue;
                if (source>0) { /* FT   source          1..754 */
                    final int idxSource=equls("source ",T,nonSpc) ? nxt(NO_SPC,T,nonSpc+7,e) : -1;
                    if (idxSource>0 && strstr("..",T,idxSource,e)>0) p.addCDS(ba.newString(idxSource,e),new String[]{"All","","",""});
                    break;
                }
                final int idxCDS=equls("CDS ",T,nonSpc) ? nxt(NO_SPC,T,nonSpc+4,e) : -1;
                if (idxCDS>0) {
                    sbXref.setLength(0);
                    boolean alreadyProteinID=false;
                    String sCDS=null, ss[]=null;
                    baCDS.setEnd(0).append(T,idxCDS,e);
                    final int indent=idxCDS-b;
                    for(int jL=iL+1; jL<ends.length; jL++) {
                        final int b1=ends[jL-1]+1, e1=prev(NO_SPC,T,ends[jL]-1,b1-1)+1, nxt=b1+indent;
                        if (nxt(NO_SPC,T,b1+2,e1)!=nxt) break;
                        if (sCDS==null) {
                            if (T[nxt]=='/') sCDS=baCDS.newString();
                            else  baCDS.append(T,nxt,e1);
                        }
                        if (sCDS!=null) {
                            for(int i=4;  --i>=0;) {
                                final String s= i==Protein.CDS_GENE ? "/gene=" : i==Protein.CDS_PRODUCT ? "/product=" : i==Protein.CDS_PROTEIN ? "/protein_id=" : "/note=" ;
                                final int quote=nxt+sze(s);
                                if (equls(s,T,nxt) && quote<e1) {
                                    if (ss==null) ss=new String[5];
                                    ss[i]=ba.newString(quote+ (T[quote]=='"'?1:0), e1- (T[e1-1]=='"'?1:0));
                                    break;
                                }
                            }
                            if (equls("/db_xref=\"",T,nxt)) {
                                final int n10=nxt+10;
                                final boolean isProtId=equls("protein_id",T,n10);
                                if (!equls("AFFY_",T,n10) && !equls("Illumina_",T,n10) && !equls("GO:GO:",T,n10) && !equls("Vega_transcript",T,n10) && !equls("Agilent",T,n10) && !(isProtId && alreadyProteinID)) {
                                    alreadyProteinID|=isProtId;
                                    String x=toStrg(T,n10, strchr(STRSTR_E,'"',T,n10,e1));
                                    final byte c0=T[n10];
                                    //if (c0=='p') x=rplc("protein_id","NCBI",x);
                                    if (c0=='U') x=rplcToStrg("Uniprot/SPTREMBL:","UNIPROT:",x);
                                    if (c0=='U') x=rplcToStrg("Uniprot/SWISSPROT:","UNIPROT:",x);
                                    if (c0=='E') x=rplcToStrg("EntrezGene:","NCBI:",x);
                                    if (c0=='R') x=rplcToStrg("RefSeq_dna:","NCBI:",x);
                                    if (c0=='R') x=rplcToStrg("RefSeq_peptide:","NCBI:",x);
                                    sbXref.append(x).append('\t');
                                }

                            }
                        }
                    }

                    if (sCDS!=null) {
                        while (p.containsCDS(sCDS)) sCDS+=" ";
                        if (sCDS.indexOf('(')>0) assert sCDS.indexOf(')')>0;
                        ss[Protein.CDS_XREFS]=sbXref.toString();
                        p.addCDS(sCDS, ss);
                        found=true;
                    }
                }
            }
        }
    }

}
