package charite.christo.protein;
import charite.christo.*;

import static charite.christo.ChUtils.*;
/**HELP
<?xml version='1.0' encoding='UTF-8'?>
<uniparc ... >
<entry accession="UPI0000EB038F">
<dbReference db="ENSEMBL" id="ENSCAFP00000017783" ... />
<sequence   ..>
PSDKLERLYQR...NGGPPLS
</sequence>
</entry>
</uniparc>

   @author Christoph Gille
*/
public class XML_SequenceParser implements ProteinParser {
    public boolean parse(Protein p, long options, ByteArray ba) {
        final byte[] T=ba.getBytes();
        final int B=ba.getBegin();
        final int E=ba.getEnd();
        if (E-B<9 || !(T[B]=='<' && T[B+1]=='?' && T[B+2]=='x' && T[B+3]=='m' && T[B+4]=='l')) return false;
        final int dasSequence=strstr(STRSTR_AFTER,"<DASSEQUENCE",T,B,E);
        final int entry=dasSequence>0 ? dasSequence : strstr(STRSTR_AFTER,"<entry",T,B,E);
        final int entryGT=strchr('>', T,entry,E);
        final int entryE=entry>0? strstr(dasSequence>0? "</DASSEQUENCE" : "</entry", T, entry, E):-1;
        if (entryE<0) return false;
        final int sequence=strstr(STRSTR_AFTER,dasSequence>0?"<SEQUENCE":"<sequence",T,entry,entryE);
        if (sequence<0) return false;
        final int sequenceGT=strchr('>', T, sequence,E);
        final int sequenceE=strstr("</",T,sequence,entryE);
        if (sequenceE<0) return false;
        byte aa[]=null;
        while(true) {
            int count=0;
            for(int i=sequenceGT+1; i<sequenceE; i++) {
                if (is(LETTR,T[i])) {
                    if (aa!=null) aa[count]=(byte)T[i];
                    count++;
                }
            }
            if (aa==null) aa=new byte[count]; else break;
        }
        p.setResidueType(aa);

        for(int dbRef=entry; (dbRef=strstr(STRSTR_AFTER,"<dbReference",T,dbRef,entryE))>0;) {
            int dbRefE=strstr("</dbReference",T,dbRef,entryE);
            if (dbRefE<0) dbRefE=strstr("/>",T,dbRef,entryE);
            //putln(" dbRef="+dbRef+"  "+dbRefE);
            if (dbRefE>0) {
                final String db=xmlAttribute("db",T,dbRef,dbRefE);
                final String id=xmlAttribute("id",T,dbRef,dbRefE);
                //putln(" db="+db+" id="+id);
                if (db!=null && id!=null) p.addDatabaseRef(db+":"+id);
            }
        }
        final String id=dasSequence>0 ? xmlAttribute("id", T, sequence, sequenceGT) : xmlAttribute("accession", T, entry, entryGT);
        if (id!=null) {
            final String db=
                id.indexOf(':')>0 ? null :
                id.startsWith("UPI") ? "UNIPARC:" :
                null;
            p.setAccessionID(db!=null ? db+id : id);
        }
        return aa.length>0;
    }

}
