package charite.christo.strap;
import charite.christo.interfaces.*;
import charite.christo.*;
import charite.christo.protein.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static java.lang.System.*;
import static java.awt.Color.*;
/**
   The user can predict secondary structure or TM helices in selected proteins.
   @author Christoph Gille
*/
public abstract class AbstractDialogPredict extends AbstractDialogJTabbedPane implements ActionListener  {
    private ChCombo comboClass;
    protected final static Object RUN_AFTER_COMPUTATION=new Object();
    private final AbstractButton cbClone=cbox("in new row");
    private Class dialogClass, interfaceClass;
    private String predictWhat;
    private JTabbedPane tabbedPane;
    private ProteinList proteinList;
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Constructor >>> */
    protected final PredictionFromAminoacidSequence newPredictor(StrapProtein[] proteins,Object clazz) {
        final PredictionFromAminoacidSequence predictor=HotswapUtil.instance(clazz,PredictionFromAminoacidSequence.class,true);
        if (predictor==null)  { error("Could not instantiate <br>"+clazz);return null;}
        if (proteins==null) return null;
        final String seq[]=new String[proteins.length];
        for(int i=0;i<proteins.length; i++) {
            seq[i]=proteins[i]==null ? null : proteins[i].getGappedSequenceAsStringUC();
        }
        predictor.setGappedSequences(seq);
        return predictor;
    }
    public void init(AbstractDialogPredict dialog, Class interfaceClass, String predictWhat) {
        dialogClass=dialog.getClass();
        this.interfaceClass=interfaceClass;
        this.predictWhat=predictWhat;
        getPanel();
    }
    private ChCombo comboClass(){
        if (comboClass==null) {
            comboClass=StrapAlign.classChoice(interfaceClass).save(dialogClass,"PredictionFromAminoacidSequence");
            pcp(KEY_CHOICE_CLASS,comboClass,this);
        }
        return comboClass;
    }
    public JTabbedPane getPanel() {
        if (tabbedPane!=null) return tabbedPane;
        proteinList=new ProteinList(align,0);
        pcp(StrapAlign.KEY_CHOICE_PROTEIN,proteinList,this); 
        tabbedPane=this;
        final boolean isTM=interfaceClass==TransmembraneHelix_Predictor.class;
        final Container pCenter=vBhB(
                                      newCNSEW(null,vB(hB("Prediction Method:","#"),comboClass().panel()),null,null,null),
                                      proteinList.scrollPane(),
                                      "",
                                      pnl(cbCache(),cbClone," ", new ChButton("GO",ChButton.NO_MARGIN).t(ChButton.GO).li(this)),
                                      " ",
                                      isTM?"#JS": null,
                                      isTM?pnl("In preparation:",StrapAlign.b(CLASS_DialogPredictTmHelicesMulti),"#") : null
                                      ),
            pMain=newCNSEW(pCenter,align.headLine(this),null,null,null);
        adMainTab(pMain,tabbedPane,getClass());
        return tabbedPane;
    }
    // ----------------------------------------
    private int resultNumber;
    private PredictionFromAminoacidSequence predictor;
    public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        if (cmd=="GO") {
            final StrapProtein proteins[]=proteinList.getSelectedOrAllProteins();
            if (proteins.length==0) return;
            predictor=newPredictor(proteins,comboClass());
            final Result r=new Result(predictor,proteins, getClass());
            adTab("result "+resultNumber++,r, tabbedPane);
        }
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Class Result >>> */
    protected final class Result extends JPanel implements Disposable,Runnable,ActionListener,ChRunnable {
        private final Thread thread;
        private final StrapProtein proteins[];
        protected final UniqueList<AbstractResidueSelection> vSelections=new UniqueList(AbstractResidueSelection.class);
        protected StrapProtein[] getProteins() { return proteins;}
        protected final PredictionFromAminoacidSequence predictor;
        private final ChLabel labIsCalculating=new ChLabel("Is calculating please wait ...");
        protected final ChButton togVisibleSB=toggl("SB", ChButton.SHOW_HIDE|ChButton.NO_MARGIN).t("in alignment").tt("show result in scroll pane").li(this);
        private final ChButton togVisible3D=toggl("3D", ChButton.SHOW_HIDE|ChButton.NO_MARGIN).t("in 3D").tt("show result in scroll pane").li(this);
        protected boolean isFinished,isFailed;
        public Class clazz() { return HotswapUtil.getClass(predictor);}
        /* <<<  <<< */
        /* ---------------------------------------- */
        /* >>> Constructor  >>> */
        public Result(PredictionFromAminoacidSequence al,StrapProtein pp[], Class clazz) {
            predictor=al;
            if (cbClone.isSelected()) {
                if (pp.length>1) error("The option <pre>"+cbClone.getText()+"</pre> is only available for one protein");
                else {
                    final StrapProtein p0=pp[0],p=new StrapProtein(align);
                    p.setTransient(true);
                    p.setResidueType(p0.getResidueType());
                    p.setResidueGap(p0.getResidueGap());
                    p.setFile(null);
                    p.setName(shrtClasNam(predictor));
                    align.addProteins(-1,new StrapProtein[]{p});
                    pp=new StrapProtein[]{p};
                }
            }
            proteins=pp;
            final String help_to_model=
                "The computed predictions are only attached in form of residue selections to proteins.<br>"+
                "Thus, the result of the prediction is not directly part of the protein models ( method  Protein#getResidueSecStrType() )<br>"+
                "<br>"+
                "By clicking this button, the prediction is stored in the protein object as it would be if the protein was loaded from a PDB file. <br>";
            final JComponent
                panDetails= clazz!=DialogPredictCoiledCoil.class ? hB(new ChButton("TO_MODEL").li(this).t("Set result to protein model"),smallHelpBut(help_to_model),"#") : null,
                pMain=vBpnl(
                            hB(shrtClasNam(predictor.getClass()),"#",ChButton.actionClose(this)),
                            labIsCalculating,
                            pnl(ChButton.actionCtrl(predictor),togVisibleSB.s(true),togVisible3D),
                            panDetails!=null ? hB(toggl().actionCollapse(false,panDetails) ,"More options","#") : null,
                            panDetails
                            );
            togVisible3D.s(false);
            togVisibleSB.s(false);
            cnsew(this,null,pMain,null,null,null);
            (thread=new Thread(this,"AbstractDialogPredict")).start();
        }
        public void dispose() {
            if (thread!=null && thread.isAlive()) thread.stop();
            try { ((Disposable)predictor).dispose();}catch(Exception e){}
            //try { ((Disposable)lastRunning).dispose();}catch(Exception e){}
            for (AbstractResidueSelection s : vSelections.asArray()) {
                Protein p=s.getProtein();
                if (p!=null) p.removeResidueSelection(s);
            }
            togVisibleSB.s(false);
            togVisible3D.s(false);
            run("UPDATE_SELECTIONS",null);
        }
    /* <<< Constructor/dispose <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */
        public void actionPerformed(ActionEvent ev) {
            final String cmd=ev.getActionCommand();
            if (cmd=="3D" || cmd=="SB") run("UPDATE_SELECTIONS",null);
            if (cmd=="TO_MODEL") {
                final char prediction[][]=predictor.getPrediction();
                for (int iP=proteins.length; --iP>=0;) {
                    final StrapProtein p=proteins[iP];
                    byte[] ss=p.getResidueSecStrType();
                    final String KEEP="ADP$$KEFP";
                    if (ss==null || gcp(KEEP,p)!=null) {
                        pcp(KEEP,"",p);
                        ss=new byte[p.countResidues()];
                        final char pred[]=prediction[iP];
                        if (pred==null) continue;
                        for(int i=mini(ss.length, pred.length); --i>=0;) {
                            final int c=pred[i]|32;
                            if (c=='h' || c=='t' || c=='e' || c=='c') ss[i]=(byte)pred[i];
                        }
                        p.setResidueSecStrType(ss);
                    }
                }
                StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,10);
            }
        }
    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> Threading >>> */
        public Object run(String id,Object arg) {
            if (id=="SET_ENABLED") {
                togVisible3D.s(true);
                togVisibleSB.s(true);
            }
            if (id=="UPDATE_SELECTIONS" || id=="SET_ENABLED") {
                    if (sze(vSelections)>0) {
                        final boolean is3D=togVisible3D.s(),isAlign=togVisibleSB.s();
                        final int visible=(is3D ? VisibleIn123.STRUCTURE : 0) | (isAlign ? (VisibleIn123.SEQUENCE | VisibleIn123.SB) : 0);
                        for(AbstractResidueSelection s:vSelections.asArray()){
                            final Protein p=s.getProtein();
                            p.removeResidueSelection(s);
                            if(isAlign || is3D) p.addResidueSelection(s);
                            s.setVisibleWhere(visible);
                        }
                        StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
                    }
            }
            return null;
        }
        public void run(){
            predictor.compute();
            isFinished=true;
            final Object hook=gcp(RUN_AFTER_COMPUTATION,this);
            if (hook instanceof Runnable) ((Runnable)hook).run();
            final char prediction[][]=predictor.getPrediction();
            isFailed=prediction==null;
            boolean success=false;
            for(int iP=proteins.length;--iP>=0;){
                if (prediction!=null && prediction[iP]!=null) {
                    makeSelection(proteins[iP],prediction[iP]);
                    success=true;
                }
            }
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
            String msg=null;
            if (!success) msg="The method did not produce a result. <br>Chose another one.<br> Press <I>details</I> for more information.";
            else {
                ChDelay.invokeAndWait(this,"SET_ENABLED",null);
                msg="Computation finished: ";
                if (sze(vSelections)>0) {
                    add(new DialogPredictTexshade(prediction,proteins,shrtClasNam(predictor)),BorderLayout.CENTER);
                    msg+=predictWhat+" are highlighted in the alignment";
                } else {
                    rmFromParent(togVisible3D);
                    rmFromParent(togVisibleSB);
                    msg+="<font color=\"#FF3333\">no</font> "+predictWhat+" <font color=\"#FF3333\">found</font><br>"+
                        "<sup>If you think that this is due to a temporary error (of a server) <br>"+
                        "and you want to compute it again switch of the cache.</sup>";
                }
            }
            labIsCalculating.t(msg);
            ChDelay.invokeAndWait(this,"UPDATE_SELECTIONS",null);
        }
        /* <<< Threading <<< */
        /* ---------------------------------------- */
        /* >>> ResidueSelection >>> */
        private void makeSelection(StrapProtein p,char[] predi) {
            if (predi==null) return;
            for(char c='0';c<='z';c++) {
                if (!is(LETTR_DIGT,c)) continue;
                final Color col=
                    c=='X' ? PINK :
                    c=='H' || c=='h' ? RED :
                    c=='E' || c=='e' ? YELLOW :
                    GREEN;
                if (col==null) continue;
                int lastIdx=-1;
                for(int i=predi.length;--i>=0;) if (predi[i]==c) { lastIdx=i; break;}
                if (lastIdx<0) continue;
                final boolean bb[]=new boolean[lastIdx+1];
                for(int i=0;i<=lastIdx;i++) bb[i]=(predi[i]==c);
                final Image im=char2image(c);
                if (im==null) continue;
                final AbstractResidueSelection sel=new AbstractResidueSelection();
                final String name=(shrtClasNam(predictor.getClass()))+'_'+c;
                sel.setStyle(VisibleIn123.IMAGE_NO_FILL);
                sel.setProtein(p);
                sel.setSelectedAminoacids(bb,0);
                sel.setImage(im);
                sel.setColor(col);
                sel.setName(name);
                ImageIcon icon=TabItemTipIcon.imageIcon(predictor);
                if (icon==null)
                    icon=iicon(
                              isAssignblFrm(TransmembraneHelix_Predictor.class,predictor) ? IC_HELIX :
                              isAssignblFrm(CoiledCoil_Predictor.class,predictor) ? IC_COILEDCOIL :
                              isAssignblFrm(SecondaryStructure_Predictor.class,predictor) ? IC_STRUCTURE :
                              null);
                TabItemTipIcon.set(null,name,shrtClasNam(predictor),icon,sel);
                vSelections.add(sel);
            }
        }
        private Image char2image(char c){
            ImageIcon i=iicon(c=='s' ?  IC_SCISSOR : c=='h' ? "prediction_H|" : "prediction_"+c);
            if (i==null) i=iicon("prediction_"+(char)(c&~32));
            return i!=null ? i.getImage() : null;
        }
    }
}
