package charite.christo.strap;
import charite.christo.*;
import charite.christo.protein.MergeAlignments;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
public class AcceptAlignment2 {

    final Row rows[];
    final byte[][] alignment;
    final int[]  countInserts, starts; //lastChangedA
    final StrapProtein[] proteins, proteinsTmp;
    public AcceptAlignment2(StrapProtein pp[], byte[][] alignment) {
        final int nP=sze(pp);
        this.proteins=pp;
        this.rows=new Row[nP];
        this.starts=new int[nP];
        //this.lastChangedA=new int[nP];
        this.countInserts=new int[nP];
        this.proteinsTmp=new StrapProtein[nP];
        this.alignment=alignment;
        if (alignment==null) return;
        for(int r=0; r<nP; r++) {
            final byte gg[]=pp[r].getGappedSequenceExactLength();
            starts[r]=idxOfLetters(alignment[r], gg);
            if (starts[r]<0) {
                putln(RED_ERROR+"Not a subsequence \n",alignment[r],"\n",gg,"\n");
                starts[r]=0;
            }
            StrapProtein[] ppRow=StrapProtein.pp(pp[r].getProteinsSameRow().clone());
            if (ppRow.length==0) ppRow=new StrapProtein[]{pp[r]};
            for(int i=ppRow.length; --i>=0;) ppRow[i]=newP(ppRow[i]);
            rows[r]=new Row(proteinsTmp[r]=newP(pp[r]), ppRow,alignment[r]);

        }
        eliminateCommonGaps();
        accept();
        for (Row r : rows) {
            for(StrapProtein p : r.all) gcp(KEYorig,p,StrapProtein.class).setResidueGap(p.getResidueGap());
        }
    }
    public int[] getCountInserts() { return countInserts;}
    //public int[] getLastChangedA() { return lastChangedA;}

    private void eliminateCommonGaps() {
        final StrapProtein pp[]=this.proteinsTmp;
        int col1=MAX_INT, col2=0;
        for(int r=0; r<rows.length; r++) {
            col1=mini(col1,starts[r]);
            col2=maxi(col2,starts[r]+countLetters(alignment[r]));
        }
        final byte[][] gapped=new byte[pp.length][];
        for(int r=pp.length; --r>=0;) gapped[r]=pp[r].getGappedSequenceExactLength();
        //for(byte bb[] : gapped) putln("gapped ",bb);
        int count=0;
        for(int c=col2; --c>=col1; ) {
            if (!ChAlignUtils.atLeastNumLettersInColumn(1,c,gapped)) {
                for(StrapProtein p : pp) {
                    final int a=p.column2nextIndex(c);
                    if (a<0) continue;
                    final int gg[]=p.getResidueGap();
                    assert gg[a]>0;
                    gg[a]--;
                }
                count++;
            }
        }
        for(int r=pp.length; --r>=0;) {
            pp[r].setResidueGap(pp[r].getResidueGap());
            countInserts[r]-=count;
        }
        putln(" eliminateCommonGapsCount="+count);
    }
    private void accept() {
        final Row[] stacks=this.rows;
        final byte[][] aa=new byte[stacks.length][], ali1d=new byte[stacks.length][];
        for(int r=0; r<stacks.length; r++)  {
            aa[r]=stacks[r].gapped;
            ali1d[r]=stacks[r].top.getGappedSequence(); // !! auffullen ? ne, gerade neu berechnet
        }
        final MergeAlignments merge=new MergeAlignments(MergeAlignments.START_IDX_AUTO, aa,ali1d);
        final int inserts[][]=merge.getInsertsAtCol();
        for(int r=0; r<stacks.length; r++) {
            for(StrapProtein p : stacks[r].all) {
                addGapAtColumn(r, inserts[r],p.getResidueGap(),  p.columns2nextIndices(),p.getMaxColumn());
            }
        }
    }
    private void addGapAtColumn(final int row, int insert[], int resGap[], int column2nextIndex[],int maxCol) {
        for(int iCol=mini(mini(maxCol,column2nextIndex.length-1),insert.length);--iCol>=0;) {
            final int ias=column2nextIndex[iCol];
            if (ias>=resGap.length || insert[iCol]==0) continue;
            resGap[ias]+=insert[iCol];
            countInserts[row]+=insert[iCol];
            //lastChangedA[row]=maxi(ias,lastChangedA[row]);
        }
    }

    private final static Object KEYorig=new Object();
    private static StrapProtein newP(StrapProtein p) {
        final StrapProtein neu=new StrapProtein();
        //neu.setName("x"+p.getName());
        neu.setResidueType(p.getResidueTypeExactLength());
        neu.setResidueGap(p.getResidueGap());
        pcp(KEYorig,p,neu);
        return neu;
    }

    /* --- Instance --- */
    private static class Row {
        final StrapProtein top, all[];
        final byte gapped[];
        public Row(StrapProtein top,StrapProtein[] all, byte[] gapped) {
            this.top=top;
            this.all=all;
            this.gapped=gapped;
        }
    }

}

