package charite.christo.strap;
import charite.christo.*;
import charite.christo.protein.*;
import java.awt.Component;
import java.lang.ref.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import static java.awt.Color.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static charite.christo.protein.ProteinViewer.*;
/**HELP

   With this form annotations can be added to residue annotations.

   <i>PACKAGE:charite.christo.protein.</i>

   The type of entry such as
   <i>CLASSICON_CLASSNAME:Pymol</i>, <i>CLASSICON_CLASSNAME:Texshade</i>
   and the text is chosen.

   For example choosing  <i>CLASSICON_CLASSNAME:Pymol</i> as entry type gives a
   choice of several Pymol commands, which alter the rendering style of
   the selected residues.

   After inserting a new entry the text can be modified in the table of
   annotation entries.

   <i>SEE_CLASS:ResidueAnnotation</i>
   <i>SEE_CLASS:ResidueAnnotationView</i>
   <i>SEE_CLASS:ActivateDeactivate</i>
   <i>SEE_CLASS:DialogCopyAnnotations</i>
   <i>SEE_DIALOG:DialogResidueAnnotationChanges</i>
   <i>SEE_DIALOG:DialogResidueAnnotationList</i>

   @author Christoph Gille
*/
public final class AddAnnotation extends JPanel implements ActionListener, ListCellRenderer {
    public final ChButton CB_VARIABLES=toggl(ResidueAnnotationView.BUT_LAB_VARIABLES).li(this);
    private final JComponent CB_SAME=cbox("Allow identical entries to occur several times");
        private final ChButton butInsert=new ChButton("Insert").li(this).tt("Add this annotation");
    private final ChLabel LABEL=new ChLabel();
    private final ChTextField jtf=new ChTextField().li(this);
    private final Reference<ResidueAnnotation> vResidueAnnotation[];
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Constructor >>> */
    public AddAnnotation(ResidueAnnotation... ss) {
        vResidueAnnotation=new WeakReference[sze(ss)];
        for(int i=sze(ss); --i>=0;) vResidueAnnotation[i]=weakRef(ss[i]);
        comboKeys=new ChJList(new ComboKeys(false),ChJTable.CLASS_RENDERER|ChJTable.ICON_ROW_HEIGHT|ChJTable.SINGLE_SELECTION);
        actionEventWhenClicked(comboKeys,this);
        final char plural=ss.length>1 ? 's' : ' ';
        butInsert.setEnabled(false);
        pcp(KEY_IF_EMPTY,"Annotation text to be added",jtf);
        final JList listValues=new ChJList(vValues,ChJTable.SINGLE_SELECTION).li(this);
        pcp("V","",listValues);
        listValues.setCellRenderer(this);
        final JComponent
            pTop=hB(panHS(this), new ChButton(IC_CUSTOM).li(this).magnifyUnderMouse(IC_CUSTOM),"#",
                    "For "+(ss.length>1 ? String.valueOf(ss.length) : "one")+
                    " residue selection"+plural+
                    ":<br>First select a key at the left, then select <br>an example text line to be added to the residue selection"+plural),
            pAdvanced=vB(CB_VARIABLES.cb(), CB_SAME),
            pTF=newCNSEW(jtf , null, null, butInsert,null),
            pSouth=vB(pTF, hB("More options ",toggl().actionCollapse(false,pAdvanced),"#"), pAdvanced),
            scrollValues=new OverlayScrollPane(listValues).addButton(new ChButton("ED",ChButton.NO_MARGIN).t("Edit this list").li(this));
        cnsew(this,scrollValues,pTop,pSouth,null,comboKeys);
    }
    /* <<< Constructor <<< */
    /* ---------------------------------------- */
    /* >>> Keys/Values >>> */
    private final DefaultListModel vValues=new DefaultListModel();
    private final ChJList comboKeys;
    private static ChCombo choiceViewer3D;
    private static Customize customizeKeys;
    public static Customize customizeKeys() {
        if (customizeKeys==null) customizeKeys=new Customize("keys",new String[]{"Remark","Note","Balloon"},"Here you can define new keys for your set entries.",Customize.LIST);
        return customizeKeys;
    }

    /* <<< Keys/Values <<< */
    /* ---------------------------------------- */
    /* >>> Rendering >>> */
    public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
        String str=toStrg(value);
        if (isSelctd(CB_VARIABLES)) {
            for(Reference r:vResidueAnnotation) {
                final ResidueAnnotation a=(ResidueAnnotation)deref(r);
                if (a!=null) {
                    str=ResidueAnnotation.replaceVariables(a,str);
                    break;
                }
            }
        }
        return LABEL.t(str).fg( isSelected ? BLUE : BLACK);
    }
    /* <<< Rendering  <<< */
    /* ---------------------------------------- */
    /* >>> Insert  >>> */
    private void insertThis(Object value) {
        if (sze(toStrg(value))==0) { StrapAlign.errorMsg("You must enter a text into the text field"); return;}
        final Object key0=comboKeys!=null ? comboKeys.getSelectedValue() : null;
        final Object key= key0==ResidueAnnotation.SPECIFIC_VIEW3D && choiceViewer3D!=null ? choiceViewer3D.getSelectedItem() : key0;
        if (key==null) { StrapAlign.errorMsg("You must select a key"); return;}
        final String sKey=key instanceof Class ? ((Class)key).getName() : toStrg(key);
        nextA:
        for(ResidueAnnotation a: derefArray(vResidueAnnotation, ResidueAnnotation.class)) {
            if (!isSelctd(CB_SAME)) {
                for(ResidueAnnotation.Entry e : a.getEntries()) {
                    if (sKey.equals(e.getKey()) && value.equals(e.getValue()))
                        continue nextA;
                }
            }
            a.add(sKey, toStrg(value),true);
        }
        StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
    }
    /* <<< Insert <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */
    public void actionPerformed(ActionEvent ev) {
        StrapAlign.errorMsg("");
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();

        if (gcp("V",q)!=null && (cmd==ACTION_CLICKED || cmd==ACTION_CLICKED2 || cmd==ACTION_SELECTION_CHANGED)) {
            final String v=toStrg(q);
            if (sze(v)>0){
                jtf.t(q);
                butInsert.setBackground(GREEN);
                butInsert.setEnabled(true);
                ChDelay.bg(butInsert,  null ,2222);
                butInsert.revalidate();
                ChDelay.revalidate(butInsert,99);
                if (cmd==ACTION_CLICKED2) insertThis(v);
            }
        }
        if (cmd==IC_CUSTOM) {
            Customize.addDialog(customizeKeys());
            ComboKeys c=new ComboKeys(false);
            for(int i=c.getSize(); --i>=0;)  Customize.addDialog(examples(c.getElementAt(i)));
        }
        if (q==butInsert) insertThis(toStrg(jtf));
        if (cmd=="ED") Customize.addDialog(examples(comboKeys.getSelectedValue()));
        if (q==jtf) { butInsert.setEnabled(nxt(NO_SPC,jtf.tools().byteArray())>=0); butInsert.requestFocus();}
        if (q==comboKeys && cmd==ACTION_CLICKED) {
            vValues.clear();
            final Object key0=comboKeys.getSelectedValue(), key;
            if (key0==ResidueAnnotation.SPECIFIC_VIEW3D) {
                if (choiceViewer3D==null) {
                    choiceViewer3D=StrapAlign.classChoice(ProteinViewer.class);
                    final Object pnl=vB("Please chose a specific 3D-viewer to add an entry for",
                                        choiceViewer3D,
                                        "Note: it is recommended, not to use specific 3D-viewer commands",
                                        "It is recommended to use general 3D-viewer commands with entries of the type \""+ResidueAnnotation.VIEW3D+"\".");
                    pcp("pnl",pnl,choiceViewer3D);
                }
                key=ChMsg.yesNo(gcp("pnl",choiceViewer3D)) ? choiceViewer3D.getSelectedItem() : null;
            } else key=key0;
            if (key==ResidueAnnotation.BG_IMAGE) vValues.add(0,URL_STRAP+"knoppix.png");
            else {
                final Customize c=examples(key);
                if (c!=null) {
                    for(String s:c.getSettings()) vValues.add(0,s);
                    CB_VARIABLES.setEnabled(true);
                }
            }
            repaint();

        }
        if (q==CB_VARIABLES) {
            repaint();
        }
    }
    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> Examples_ >>> */
    private final static Map<String,Customize> mapExamples=new HashMap();
    public static Customize examples(Object key) {
        Customize cust=null;
        final Class c=HotswapUtil.getClass(key);
        if (c!=null && c!=String.class) cust=Customize.getCustomizableForClass(c, Customize.CLASS_Example);
        else {
            final String k=toStrgIntrn(key);
            cust=mapExamples.get(k);
            if (cust==null) {
                String
                    examples3D=COMMANDspheres+"\n"+COMMANDribbons+"\n"+COMMANDsticks+"\n"+COMMANDlines+"\n"+COMMANDsurface+"\n"+COMMANDcolor+" red\n"+COMMANDcolor+" #FF00AA\n",
                    lines=
                    "Note"==k ?  "Experimantal confirmation awaited\n"+"Is a cleaved\n"+
                    "Is modified\n"+"Is important for catalysis\n"+"PUBMED:10547838 PDB:1RYP EMBL:X59417" :
                    "Remark"==k ? "Important note\n"+"Experimantal confirmation awaited" :
                    "Balloon"==k ? "A Balloon text that appears when the mouse is over" :
                    ResidueAnnotation.VIEW3D==k ? examples3D :
                    "Some text";
                cust=new Customize("Examples_"+k,splitStrg(lines,'\n') , "Examples", Customize.LIST);
                mapExamples.put(k,cust);
            }
        }
        return cust;
    }
    /* <<< Examples <<< */
    /* ---------------------------------------- */
    /* >>> Html  >>> */
   public static AddAnnotation docuSnapshotWithPymol() {
        final StrapProtein p=new StrapProtein();
        p.setName("test_protein");
        final ResidueAnnotation s=new ResidueAnnotation(p);
        p.addResidueSelection(s);
        s.setValue(ResidueAnnotation.GROUP,"new");
        s.setValue(ResidueAnnotation.POS,"6-8");
        final AddAnnotation al=new AddAnnotation(s);
        al.comboKeys.setSelI(7);
        return al;
    }

}
