package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import org.biojava.bio.seq.*;
import org.biojava.bio.symbol.*;
import org.biojava.utils.*;
import org.biojava.bio.*;
/**
Converting a Biojava sequence objects into a StrapProtein object.
This class extends StrapProtein and is instantiated with a biojava sequence object.
The gapped sequence as well as the annotations and features are converted.
*/

public class BiojavaSequence2StrapProtein extends StrapProtein {
	//				private StrapProtein protein;
				public BiojavaSequence2StrapProtein(Sequence  seq) {
								super(null);
								final String seqStr=seq.seqString();
								setResidueType(seqStr.replaceAll("-","").getBytes());
								if (seq instanceof GappedSequence) inferGapsFromGappedSequence(seqStr.replace('-',' ').getBytes());
								setAnnotation(seq);
								setFeatures(seq);
								setName(seq.getName());
				}

				private void setAnnotation(Sequence seq) {
								final Annotation a=seq.getAnnotation();
        for(Object k : a.keys().toArray()) {
												final String v=a.getProperty(k).toString();
												if ("ORGANISM".equals(k) || "organism".equals(k) || "OS".equals(k)) setOrganism(v);
												if ("COMPOUND".equals(k) || "compound".equals(k) || "DE".equals(k)) setCompound(v);
								}
				}

				private void setFeatures(Sequence seq) {
								final Iterator it=seq.features();
								while(it.hasNext()) {
            final Object o=it.next();
            if (!(o instanceof Feature)) continue;
												final Feature f=(Feature) o;
												final Location l=f.getLocation();
												final Annotation a=f.getAnnotation();
												final ResidueAnnotation resAnno=new ResidueAnnotation(this);
												addResidueSelection(resAnno);
            for(Object key : a.keys().toArray()) {
																resAnno.setValue(key.toString(),a.getProperty(key).toString());
												}
												final StringBuffer sb=new StringBuffer();
												final Iterator itBlock=l.blockIterator();
												while(itBlock.hasNext()) {
																final Location lBlock=(Location)itBlock.next();
																final int min=lBlock.getMin();
																final int max=lBlock.getMax();
																sb.append(min);
																if (min!=max) sb.append('-').append(max);
																sb.append(", ");
												}
												resAnno.setValue(ResidueAnnotation.NAME,"biojava");
												resAnno.setValue(ResidueAnnotation.POS,sb.toString());
								}
				}

}
