package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;

/**HELP
    <i>PACKAGE:charite.christo.strap.extensions.</i>
    <div class="floatR">
      <table border="1" summary="3D-viewers">
        <caption>Comparing different 3D-viewers</caption>
        <tr>
          <th></th>
          <th><i>CLASS_REF:ChJmolPROXY</i>JMOL</th>
          <th><i>CLASS_REF:Rasmol</i>RASMOL</th>
          <th><i>CLASS_REF:Pymol</i>PYMOL</th>
          <th><i>CLASS_REF:Vmd</i>VMD</th>
        </tr>
        <tr>
          <th>Installed by STRAP</th>
          <td>allways</td>
          <td>Linux</td>
          <td>Linux, Windows, Mac</td>
          <td>never</td>
        </tr>
        <tr>
          <th>Operation system</th>
          <td>all</td>
          <td>Linux, Unix</td>
          <td>all</td>
          <td>Linux, Unix</td>
        </tr>
        <tr>
          <th>Performance</th>
          <td>medium</td>
          <td>high</td>
          <td>high with accelerated video cards</td>
          <td>very high</td>
        </tr>
        <tr>
          <th>Memory consumption</th>
          <td>high</td>
          <td>low</td>
          <td>medium</td>
          <td>medium</td>
        </tr>
        <tr>
          <th>#  models per view</th>
          <td>1</td>
          <td>1</td>
          <td>many</td>
          <td>many</td>
        </tr>
        <tr>
          <th>Scripting language</th>
          <td>Rasmol scripting language</td>
          <td>Rasmol scripting language</td>
          <td>Python</td>
          <td>TCL, Python</td>
        </tr>
        <tr>
          <th>Number of atom selections</th>
          <td>1</td>
          <td>1</td>
          <td>many</td>
          <td>many</td>
        </tr>
        <tr>
          <th>Shows superimposed proteins</th>
          <td>no</td>
          <td>no</td>
          <td>yes</td>
          <td>no</td>
        </tr>
        <tr>
          <th>Move cursor to picked residue</th>
          <td>yes</td>
          <td>yes</td>
          <td>yes</td>
          <td>no</td>
        </tr>
      </table>
      <br>
    </div>

      Protein PDB structure viewers are
      used for detailed 3D-visualization.

      Depending on the protein viewer the alignment pane and the 3D-view are
      interoperable.

    <br><br><b>Communication objects:</b> Loading a protein into a protein structure viewer
    creates a communication object which is found under the respective protein node in the object
    tree.  A context menu can be opened.

    <br><br><b>Simple residue selection:</b> When the mouse is dragged
    over the sequence in the multiple alignment view the resulting selection
    will also be shown in the 3D-viewer.

    In the 3D-viewer the  rendering style of the selected amino acids can be changed.

    <br><br><b>Interpreting commands:</b> A major advantage of running a
    protein structure viewer from STRAP is that 3D-rendering commands can
    be stored in <i>DIALOG:ResidueAnnotationView</i>.

    <br><br><b>Atom/Residue picking:</b>
    When users pick an Atom in the 3D-view the alignment cursor jumps to the respective alignment position.
    This feature is implemented in most 3D-viewers.

    Related links:  http://biodesign.asu.edu/labs/assets/lindsay/molvis/MOL.html

    <i>SEE_CLASS:ResidueAnnotation</i>
    <i>SEE_CLASS:ResidueAnnotationView</i>
    <i>SEE_CLASS:charite.christo.protein.ChJmolPROXY</i>
    <i>SEE_CLASS:Protein3d</i>

    @author Christoph Gille

*/
public class Dialog3DViewer extends AbstractDialogJPanel implements  ActionListener{
    private ChButton cbSend=toggl("Send all rendering commands when started").tt("The residue selections may contain<br>rendering commands for this protein viewer.<br>If this is the case they are send after start").s(true);
    private ChCombo comboClass;
    private ProteinCombo choiceProtein;

    public Dialog3DViewer() {
        comboClass=StrapAlign.classChoice(ProteinViewer.class).save(Dialog3DViewer.class,"ProteinViewer");
        pcp(KEY_CHOICE_CLASS,comboClass,this);
        choiceProtein=new ProteinCombo(align,ProteinList.CALPHA);
        final JComponent
            pCenter=vBhB(pTop(), " ",
                         hB("Protein:",choiceProtein,"#")," ",
                         comboClass.panel(),
                         pnl(new ChButton("GO").t(ChButton.GO).li(this)),
                         cbSend.cb(),
                         WATCH_MOVIE+MOVIE_Drag_protein_to_3D
                         );
        add(pCenter);
        pcp(StrapAlign.KEY_CHOICE_PROTEIN, choiceProtein,this);
    }
    public void actionPerformed(ActionEvent ev)  {
        if ("GO"==ev.getActionCommand()) {
            final StrapProtein p=choiceProtein.getStrapProtein();
            if (p!=null && (p.hasCalpha() || ChMsg.yesNo("The protein "+p+" does not have 3D-coordinates.<br>Continue anyway?"))) {
                if (p.getName().endsWith(".dssp") && !ChMsg.yesNo("The dssp file format is usually not supported.<br>Continue anyway?")) return;
                final int opt=(cbSend.s()?ProteinViewerUtils.OPEN_SEND_ALL_ANNOTATIONS:0)|(isCtrl(ev)?ProteinViewerUtils.OPEN_VERBOSE:0);              
                ProteinViewerUtils.open3dViewer(new Protein[]{p},comboClass, ProteinViewerUtils.OPEN_INIT_SCRIPT|opt);
                setNextProcessHasLogPanel(false);
            }
        }
    }

}
/* NOC */
