package charite.christo.strap;
import charite.christo.interfaces.*;
import charite.christo.*;
import charite.christo.protein.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;

/**HELP
   <h2>Cutting out the matches</h2>

   The following lines can be used to cut out the match in the amino acid sequence.
   After the exclamation mark the first and the last matching residue index is written.
   These lines must be applied to the rename/copy dialog which is in the protein context menu.
   After renaming or copying the resulting proteins are shortened according to the match.
   The proteines can be reverted to the original length with the rename/copy dialog.
   <br>
*/
class DialogAlignOneToAllAlignment extends AbstractResidueSelection implements Disposable,ActionListener,Runnable,Comparable,ChRunnable,PaintHook,ChProcessEvent {
    private final static String NAME="MATCHING_";
    private final Point ORG=new Point();
    private final int idx;

    private final SequenceAligner aligner;
    private final JComponent pMain;
    private float score;
    private int fontSize=12, sliderValue=MIN_INT;

    private JComponent labScore;
    private final JComponent
        alignment=new ChPanel(),
        butUndo=new ChButton("U",ChButton.NO_MARGIN|ChButton.DISABLED).t("Undo").li(this);
    private final JScrollPane sp=scrllpn(0L,alignment);
    private final StrapProtein proteins[];
    private final DialogAlignOneToAllResult resultPane;

    public DialogAlignOneToAllAlignment(StrapProtein[] pp, SequenceAligner al, int idx, DialogAlignOneToAllResult result) {
        this.idx=idx;
        this.resultPane=result;
        this.proteins=pp;

        setProtein(pp[1]);
        setName(NAME+shrtClasNam(al));
        aligner=al;
        pcp(SCROLLABLE_TRACKS_VIEWPORT_HEIGHT,"",alignment);
        StrapProteinUtils.setSeqs(aligner, pp);
        if (aligner instanceof NeedsProteins)   ((NeedsProteins)aligner).setProteins(pp);
        pMain=hB("Calculating, please wait");
        pcp(ChRenderer.KEY_RENDERER_COMPONENT,pMain,this);
        setColor(C(0xFF00FF));
    }

    static String RUN_ENABLE="RE", RUN_SLIDER="SLIDER";

    public Object run(String id, Object arg) {
        final float[] minMaxScore=resultPane.minMaxScore;
        if (id==RUN_ENABLE && labScore!=null && minMaxScore[1]!=minMaxScore[0]) labScore.setForeground(float2blueRed((score-minMaxScore[0])/(minMaxScore[1]-minMaxScore[0])));
        if (id==RUN_SLIDER) this.sliderValue=atoi(arg);
        return null;
    }
    public int compareTo(Object o) {
        if (!(o instanceof DialogAlignOneToAllAlignment)) return 0;
        DialogAlignOneToAllAlignment al=(DialogAlignOneToAllAlignment)o;
        final float v1,v2;
        if (isSelctd(resultPane.TOG_SORT_SCORE)) { v1=al.score; v2=score; } else {v1=al.idx; v2=idx;}
        return v1<v2 ? -1 : v1>v2 ? 1 :0;
    }
    /* <<< Comparable  <<< */
    /* ---------------------------------------- */
    /* >>> Dispose >>> */
    public void dispose() {
        resultPane.vALignments.remove(this);
        resultPane.jTable.tableChanged(new TableModelEvent(resultPane));
    }
    /* <<< Dispose  <<< */
    /* ---------------------------------------- */
    /* >>> paint >>> */
    private boolean alreadyPainted;
    public boolean paintHook(JComponent component, Graphics g, boolean after) {
        final StrapProtein pp[]=this.proteins;
        resultPane.setFont(getFnt(fontSize,true,0));
        final byte[][] seqs=aligner.getAlignedSequences();
        if (sze(seqs)<2 ||  0==sze(seqs[0])*sze(seqs[1])) return true;
        final byte[] s0=seqs[0], s1=seqs[1];
        final Color bg=C(isWhiteBG() ? 0xFFffFF : 0);
        final int h=component.getHeight(), w=component.getWidth();
        if (component instanceof ChScrollBar && after) {
            final byte[][] BLOSUM=Blosum.BLOSUM62;
            final ChScrollBar sb=(ChScrollBar)component;
            final Rectangle track=sb.getTrack();
            final Rectangle thumb=sb.getThumb();
            final int L=maxi(s0.length, s1.length);
            final int[] rgb=sharedRGB(L);
            final java.awt.image.BufferedImage image=sharedImage('H',L);
            g.setColor(C(0x8080ff,0x80));
            for(int i=L; --i>=0;) {
                final int c1=s1.length>i? s1[i]|32:0;
                final int c0=s0.length>i? s0[i]|32:0;
                rgb[i]=c0==c1?0xffFFffFF : BLOSUM[c0&255][c1&255]>0 ? 0x777777 : 0;
            }
            image.setRGB(0,0,L,1,rgb,0,L);
            g.drawImage(image,track.x,0,  track.x+track.width,h, 0,0,L,1,alignment);
            g.fillRect(thumb.x,0, thumb.width,thumb.height);
            g.setColor(C(0xff));
            g.drawRect(thumb.x,0, thumb.width,thumb.height);
            g.setColor(C(DEFAULT_BACKGROUND));
            g.fillRect(0,0,track.x,99);
            g.fillRect(track.x+track.width,0,9999,99);
        }
        if (component==alignment && after) {
            final Rectangle cb=charBounds(g.getFont());
            final int charW=cb.width, charH=cb.height;
            g.setColor(bg);
            g.fillRect(0,0,w,99999);
            final int mode=isSelctd(resultPane.cbShadeStru) ? DrawGappedSequence.SHADE_SEC_STRU:0;
            int firstL=0,lastL=MAX_INT;
            for(int iS=2; --iS>=0;) {
                final byte[] s=seqs[iS];
                final int startsAt=idxOfLetters(s,pp[iS].getResidueTypeExactLength());
                if (startsAt>=0) {
                    ORG.y=(charH+1)*iS*2;
                    resultPane.dgs.draw(alignment,g,ORG, maxi(0, startsAt),s,0,MAX_INT, pp[iS], resultPane.shadingAA.getColors(),mode);
                    g.setColor(C(DEFAULT_BACKGROUND));
                    final int fL=nxt(LETTR,s),lL=prev(LETTR,s,MAX_INT,-1);
                    if (firstL<fL) firstL=fL;
                    if (lastL>lL) lastL=lL;
                    g.fillRect(0,iS*charH, fL*charW, charH*2+2);
                    g.fillRect((lL+1)*charW, iS*charH, w, charH*2+2);
                }
            }
            if (!alreadyPainted) { alignment.scrollRectToVisible(new Rectangle(firstL*charW,0,(lastL-firstL)*charW,1)); alreadyPainted=true;}
            ORG.y=charH+1;
            g.setColor(C(0xFFffFF));
            resultPane.dgs.drawBlastMidline(s0,s1, 0, MAX_INT, Blosum.BLASTMIDLINE, g, 0, ORG.y);
        }
        return true;
    }
    /* <<< paint <<< */
    /* ---------------------------------------- */
    /* >>> ChRunnable >>> */
        public void run() {
        final StrapProtein pp[]=this.proteins;
        setSharedParas(resultPane.sharedInstance,aligner);
        aligner.compute();
        final byte[][] seqs=this.aligner.getAlignedSequences();
        if (seqs==null || seqs.length<2) return;

        final String label,tip;
            this.score=getScor(aligner);
            if (!Float.isNaN(this.score)) {
                label="score="+this.score;
                tip=null;
            }
        else {
            final Ratio r=CrudeSequenceSimilarity.getIdentity(seqs[0],seqs[1],0,9999999);
            score= r.getDenominator()==0 ? 0:r.getNumerator()/(float)r.getDenominator();
            label="ident="+r;
            tip="<br>identical amino acid "+r.getNumerator()+"<br>in "+r.getDenominator()+" positions";
        }
        final float[] minMaxScore=resultPane.minMaxScore;
        if (minMaxScore[0]>score) minMaxScore[0]=score;
        if (minMaxScore[1]<score) minMaxScore[1]=score;
        labScore=ChButton.actionCtrl(aligner);
        if (labScore!=null) ((ChButton)labScore).t(label); else labScore=new ChLabel(label);
        noMargn(labScore);

        final ProteinLabel proteinLabel= new ProteinLabel(pp[1],0L);
        proteinLabel.setBackground(C(DEFAULT_BACKGROUND));
        rtt(proteinLabel);
        final ChTextField tf=new ChTextField().cols(10,true,true).li(this);
        pcp(KEY_IF_EMPTY,"notes",tf);
        final JComponent
            butApply=new ChButton("A",ChButton.NO_MARGIN).t("Infer alignment").li(this)
            .tt("<b>Accept & apply to the alignment</b><br>Only the protein "+pp[0]+
                " is changed.<br>The protein "+pp[1]+" is not changed "+tip),
            pWest=pnl(tf, labScore," ", butApply,butUndo),
            rowHeader=newCNSEW(pWest,null,null,ChButton.actionClose(this),proteinLabel);

        noInsets(sp);
        ChScrollBar sb=(ChScrollBar)sp.getHorizontalScrollBar();
        sb.setUnitIncrement(fontSize);
        addPaintHook(this,sb);
        generateMiddleForLeftButton(sb);
        cnsew(pMain,sp,rowHeader,null,null,null);
        setDragScrolls(alignment,resultPane.scrollPane);
        setDragScrolls(alignment,sp);
        final Rectangle cb=charBounds(resultPane.getFont());
        alignment.setPreferredSize(dim(cb.width*maxi(seqs[0].length,seqs[1].length), 3*cb.height+3));
        addPaintHook(this,alignment);
        final int prefHeight=3*cb.height+3+ICON_HEIGHT+EX+4;
        final JTable jT=resultPane.jTable;
        if (jT.getRowHeight()!=prefHeight) { jT.setRowHeight(prefHeight);}
    }
    /* <<< ChRunnable <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */

    public AWTEvent processEvent(JComponent c, AWTEvent ev) {
        if (c==alignment) resultPane.jTable.processEvent(ev);
        return ev;
    }

    public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        if (cmd=="A" || cmd=="U") {
            final StrapProtein pp[]=this.proteins;
            final int[] gaps, gapsSaved=gcp(butUndo, pp[0], int[].class);
            if (cmd=="A") {
                final byte[][] seqs=aligner.getAlignedSequences();
                final byte[] mGapped=seqs[0], mResidues=pp[0].getResidueTypeExactLength();
                final byte[] tGapped=seqs[1], tResidues=pp[1].getResidueTypeExactLength();
                final int[] tGaps=pp[1].getResidueGap(), mGaps=pp[0].getResidueGap();
                if (gapsSaved==null) pcp(butUndo, pp[0].getResidueGap().clone(),pp[0]);
                gaps=ChAlignUtils.useGapsFromExternalAlignment(tGapped, mGapped,tResidues,  mResidues,  tGaps, mGaps);
            } else gaps=gapsSaved;
            if (gaps!=null) pp[0].setResidueGap(gaps.clone());
            StrapAlign.getInstance().dispatchLater(StrapEvent.ALIGNMENT_CHANGED,111);
            setEnabld(butUndo,cmd!="U");
        }

    }

    private boolean selAA[];
    private int sel_lastSlider, sel_mc;
    private int[] sel_blosums;
    private final static int WINDOW=4;

    @Override public boolean[] getSelectedAminoacids() {
        final int slider=this.sliderValue;
        boolean bb[]=this.selAA;
        final int nR=proteins[1].countResidues();
        final byte[] rt=proteins[1].getResidueTypeExactLength();
        int mc=proteins[1].getModificationCount(ProteinModiCount.MC_RESIDUE_TYPE);
        if (bb==null || sel_lastSlider!=slider || sel_mc!=mc) {
            sel_lastSlider=slider;
            sel_mc=mc;
            Protein.incrementModificationCount(ProteinModiCount.MC_RESIDUE_SELECTIONS, proteins[1]);
            final byte[][] BLOSUM=Blosum.BLOSUM62;
            final byte[][] seqs=aligner.getAlignedSequences();
            if (sze(seqs)<2 ||  0==sze(seqs[0])*sze(seqs[1])) return null;
            final byte[] s0=seqs[0], s1=seqs[1];
            final int L=mini(s0.length, s1.length);
            bb=this.selAA=redim(bb, rt.length, 0);
            if (this.sel_blosums==null) {

                this.sel_blosums=new int[bb.length];
                final int startsAt=maxi(0,idxOfLetters(s1,rt));
                if (startsAt<0) return null;
                for(int col=0, a=startsAt; col<L; col++) {
                    int sum=0;
                    window:
                    for(int k=-WINDOW+col; k<=WINDOW+col; k++) {
                        if (k<0 || k>=L) continue;
                        if (!is(LETTR,s0,k) || !is(LETTR,s1,k)) {
                            sum=MIN_INT;
                            break window;
                        }
                        sum+=BLOSUM[s0[k]&127][s1[k]&127];
                    }
                    if (a<sel_blosums.length) sel_blosums[a]=sum;
                    if (is(LETTR,s1,col)) a++;
                }
            }
            for(int i=0; i<bb.length; i++)  bb[i]=i<sel_blosums.length && sel_blosums[i]>slider;
        }
        return bb;
    }
}

