package charite.christo.strap;
import charite.christo.interfaces.*;
import charite.christo.*;
import charite.christo.protein.*;
import charite.christo.strap.interfaces.*;
import java.awt.*;
import java.awt.event.*;
import java.util.List;
import java.util.*;
import javax.swing.*;
import javax.swing.table.TableModel;
import javax.swing.event.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP
<h2>Cutting out the matches</h2>

The following lines can be used to cut out the match in the amino acid sequence.
After the exclamation mark the first and the last matching residue index is written.
These lines must be applied to the rename/copy dialog which is in the protein context menu.
After renaming or copying the resulting proteins are shortened according to the match.
The proteines can be reverted to the original length with the rename/copy dialog.
<br>
*/
public class DialogAlignOneToAllResult extends JPanel implements Runnable,Disposable,TableModel,ActionListener {
    private SequenceAligner aligner;
    private final StrapProtein protein,proteins[];
    private final JProgressBar progress;
    private final Object clazz;
    private boolean hasScore=false;
    private boolean isDisposed;
    final ChButton TOG_SORT_SCORE=toggl("sort score").li(this);
    final ChButton cbShadeStru=toggl("STR").t("structure").li(this).tt("shade secondary structure");
    final ShadingAA shadingAA=new ShadingAA(); {shadingAA.li(this).s(2);}
    final UniqueList<DialogAlignOneToAllAlignment> vALignments=new UniqueList(DialogAlignOneToAllAlignment.class);
    final ChJTable jTable=new ChJTable(this,ChJTable.DEFAULT_RENDERER);
    final JScrollPane scrollPane;
    final float minMaxScore[]={Float.NaN,Float.NaN};
    Object sharedInstance;
    final DrawGappedSequence dgs=new DrawGappedSequence();
    public DialogAlignOneToAllResult(StrapProtein pp[],StrapProtein protein,ChCombo comboClass) {
        pcp(SCROLLABLE_TRACKS_VIEWPORT_WIDTH,"",jTable);
        this.proteins=pp;
        this.protein=protein;
        progress=new JProgressBar(0,proteins.length);
        progress.setStringPainted(true);
        UIupdateList.add(jTable,UIupdateList.PROTEIN_AT_CURSOR);
        addActLi(this,jTable);
        jTable.setShowVerticalLines(false);
        //jTable.setBackground(C(DEFAULT_BACKGROUND));
        jTable.setColumnWidth(0,ICON_HEIGHT,false);
        this.clazz=comboClass.getSelectedItem();
        this.sharedInstance=comboClass.instanceOfSelectedClass(null);
        boolean anyWithSecStr=protein.getResidueSecStrType()!=null;
        for(StrapProtein p:proteins) if ( p.getResidueSecStrType()!=null) { anyWithSecStr=true; break;}
        final JComponent
            cb=anyWithSecStr  ?cbShadeStru.cb() : null,
            pCentr=pnl(
                       shadingAA,cb,TOG_SORT_SCORE.cb(),
                       new ChButton("SEL").li(this).t("Select high matching regions ...")
                       ),
            bClose=javaVsn()<16 ? ChButton.actionClose(this, OPTION_DISPOSE_RM_CHILDS) : null,
            pTop=newCNSEW(pCentr,null,null,bClose,smallHelpBut(new Object[]{DialogAlignOneToAll.class,HotswapUtil.getClass(comboClass)}));
        cnsew(this,scrollPane=scrllpn(0L,jTable),pTop,progress,null,null);
        new Thread(this,"DialogAlignOneToAllResult").start();

    }

    /* <<< Constructor  <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent  >>> */
    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        final StrapAlign align=StrapAlign.getInstance();
        boolean de=false;
        if (cmd==ACTION_SELECTION_CHANGED) { addResidueSelections(false); de=true; }
        if (q==shadingAA) {
            cbShadeStru.s(false);
            jTable.repaint();}
        if (q==TOG_SORT_SCORE) {
            sortArry((List)vALignments);
            jTable.repaint();
        }
        if (cmd=="STR") jTable.repaint();
        if (cmd=="NARROW") {
            final ResidueSelection aa[]=getSelectedAlignments();
            final String msg=
                "The amino acid sequence from the first to the last match will remain.<br>"+
                "Non-selected termini are cut off.<br>"+
                "This is implemented by appending an suffix to the protein name containing an exclamation mark and a position range.<br>"+
                "The original sequences can be reverted by removing the suffix from the protein names.<br><br>"+
                "Narrow sequences?";
            if (aa.length==0 || !ChMsg.yesNo(msg)) return;
            final StrapProtein pp[]=new StrapProtein[aa.length];
            final Object key=getClass();
            for(int i=0; i<aa.length; i++) {
                final boolean bb[]=aa[i].getSelectedAminoacids();
                pp[i]=(StrapProtein)aa[i].getProtein();
                pcp(key, new int[]{firstTrue(bb), lastTrue(bb)},  pp[i]);
            }
            DialogRenameProteins.instance(getClass()).rename(0L, pp, key);
        }

        if (cmd=="REALIGN") ResidueSelectionUtils.openDialogRealign(vALignments.asArray());
        if (cmd=="SEL") {
            if (panSelection==null) jTable.addRowSelectionInterval(0, getRowCount()-1);
            addResidueSelections(true);
            showSelectionPane(ChFrame.AT_CLICK);
            de=true;
        }

        if (cmd=="SLIDER") {
            final int v=((ChSlider)q).getValue();
            final Object intObjct=intObjct(v);
            for(DialogAlignOneToAllAlignment a:vALignments.asArray()) {
                a.run(a.RUN_SLIDER,intObjct);
                de=true;
            }
        }
        if (de) StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
    }
    public DialogAlignOneToAllAlignment[] getSelectedAlignments() {
        final int ii[]=jTable.getSelectedRows();
        final DialogAlignOneToAllAlignment[] sel=new DialogAlignOneToAllAlignment[sze(ii)];
        for(int i=0; i<sze(ii);i++) {
            sel[i]=vALignments.get(ii[i]);
            if (sel[i]!=null && sel[i].getProtein()==this.protein) sel[i]=null;
        }
        return rmNull(sel,DialogAlignOneToAllAlignment.class);
    }

    private void addResidueSelections(boolean all) {
        final DialogAlignOneToAllAlignment[] aa=vALignments.asArray(), aaSel=getSelectedAlignments();
        for(DialogAlignOneToAllAlignment a:aa) {
            final Protein p=a.getProtein();
            if (all || cntains(a,aaSel)) {
                p.addResidueSelection(a);
            } else p.removeResidueSelection(a);
        }
        if (aaSel.length>0) showSelectionPane(0L);
    }
    private JComponent panSelection;
    private void showSelectionPane(long options) {
        if (this.panSelection==null) {
            Object
                pN="Create Residue selections from high matching regions",
                pA=vB(
                      "#TB The created residue selection objects of high matching regions may be applied:",
                      hBL("Realign all matching regions ",new ChButton("REALIGN").t("Realign").li(this)),
                      hBL("Narrow proteins to matching regions ",new ChButton("NARROW").t("Narrow").li(this))
                      ),
                pS=vB(" ",pA," ");
            final ChSlider sliderMatch=new ChSlider("SLIDER",-100,100,0).endLabels("All","High matching");
            addActLi(this,sliderMatch);
            this.panSelection=newCNSEW(newCNSEW(sliderMatch,null,null,null,"Threshold "),pN,pS,null,null);
        }
        ChFrame.frame("Matching regions",panSelection,ChFrame.PACK).shw(options|ChFrame.TO_FRONT);
    }
    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> ChRunnable >>> */
    public void run() {
        final Class c=HotswapUtil.getClass(clazz);
        if (c==null) return;
        hasScore=HasScore.class.isAssignableFrom(c);
        minMaxScore[0]=Float.MAX_VALUE;
        minMaxScore[1]=Float.MIN_VALUE;
        final Object shared_sa=StrapPlugins.instanceSA(clazz,2,true);
        if (isInstncOf(SequenceAligner.class,shared_sa)) {
            for(int i=0;i<proteins.length && !isDisposed;i++) {
                this.aligner= StrapPlugins.instanceSA(c,2,true);
                setSharedParas(shared_sa,this.aligner);
                final DialogAlignOneToAllAlignment a;
                a=new DialogAlignOneToAllAlignment(
                                                   new StrapProtein[]{this.protein,proteins[i]},
                                                   this.aligner, sze(vALignments)+1, this);
                vALignments.add(a);
                inEDT(a);
                ChDelay.runAfterMS(EDT,a,a.RUN_ENABLE,444);
                progress.setValue(progress.getValue()+1);
                jTable.revalidate();
                jTable.repaint();
                sleep(5);
            }
            ChDelay.repaint(this,333);
            rmFromParent(progress);
        }
    }
    /* <<< ChRunnable <<< */
    /* ---------------------------------------- */
    /* >>> Disposable >>> */
    public void dispose() {
        for(DialogAlignOneToAllAlignment a:vALignments.asArray())  a.getProtein().removeResidueSelection(a);
        isDisposed=true;
        StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,1);
    }
    /* <<< Disposable <<< */
    /* ---------------------------------------- */
    /* >>> TableModel >>> */
    public String getColumnName(int c) {return   "";}
    public boolean isCellEditable(int row,int column) {return column>0;}
    public void setValueAt(Object o,int row,int col) {}
    public Class getColumnClass(int c) {return Object.class;}
    public int getColumnCount() {return 2;}
    public int getRowCount() {return sze(vALignments);}
    public Object getValueAt(int row,int col) {
        final ResidueSelection a=vALignments.get(row);

        return a==null || col==1 ? a : a.getProtein()==this.protein ? "" :  ChRenderer.KEY_CB_SELECTED;
    }
    public void addTableModelListener(TableModelListener l){}
    public void removeTableModelListener(TableModelListener l){}

}
