package charite.christo.strap;
import charite.christo.interfaces.*;
import charite.christo.*;
import charite.christo.strap.extensions.*;
import charite.christo.blast.*;
import charite.christo.protein.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import java.lang.ref.*;
import java.awt.event.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static java.awt.Color.*;
/**HELP

STRAP allows to search public and private sequence databases to
identify sequences that are similar to a query sequences.

Depending on the type of sequences, there are four search methods:
<ol>
<li>The query is an amino acid (aa) sequence and the database contains nucleotide (nt) sequences.</li>
<li>  ...   nt   ...  nt. ...  </li>
<li>  ...   aa   ...  aa  ...  </li>
<li>  ...   nt   ...  aa  ...  </li>
</ol>

Most users are already familiar with WIKI:BLAST Web services which are
accessible with a Web browser.

STRAP communicates with the Web services and presents the results.
It provides functionality that is not available in Web-browsers:

<ul>

<li>Many jobs can be started simultaneously.</li>

<li>The Blast result is saved in the cache. If  the same query is used a
second time the result is loaded from the cache and appears immediately.
</li>

<li>
Blast alert:

The sequence databases are permanently
growing.

With the cache disabled   STRAP conducts a new Blast
and compares the new result with the previous result stored in the cache.

It reports if there are new hits since last run and  highlights all new hits.
</li>

<li>In STRAP several patterns can be highlighted in different colors.
STRAP identifies patterns even if they are disrupted  by a gap.
</li>

<li>Sequences from the hit-list can be loaded  with one click.</li>
</ul>
   See UBIC:blastall UBIC:blat UBIC:wu_blast
   @author Christoph Gille
*/
public class DialogBlast extends AbstractDialogJTabbedPane implements ActionListener,ChRunnable  {
    private final static String SELECTION_NAME="blast_match";
    private final ProteinList proteinList;
    private final Vector vDatabases=new Vector();

    private final ChCombo
        comboDatabases=new ChCombo(vDatabases,0).save(DialogBlast.class,"db"),
        comboClass=StrapAlign.classChoice(SequenceBlaster.class).li(this).save(DialogBlast.class,"SequenceBlaster"),
        comboNT=new ChCombo("Aminos or nucleotides: Strap's guess","Query is nucl sequ with introns","Query is the coding nucleotide sequence").li(this);
    private final JLabel labFirst=new JLabel("first residue "), labLast=new JLabel(" last residue ");
    private transient SequenceBlaster lastRunning;
    private final Object[] radioDatabase;
    public DialogBlast() {
        pcp(KEY_CHOICE_CLASS,comboClass,this);
        proteinList=new ProteinList(align,0);
        pcp(StrapAlign.KEY_CHOICE_PROTEIN, weakRef(proteinList), this);
        proteinList.li(this);
        comboClass.setPreferredSize(dim(comboClass.getPreferredSize().width,butGo.getPreferredSize().height));
        updateDatabases();
        radioDatabase=radioGrp(new String[]{"Proteins with 3D-structure", "All proteins", "Non-redundant set of proteins"},0,this);
        final JComponent
            pSimple= vBhB(
                        "Searching for similar sequences in:",
                        hBL(" ",vB(radioDatabase)),
                        " "
                        ),
            pSettings=vBhB(
                         pnl(new GridLayout(2,2),hB("#",labFirst),tfFrom,hB("#",labLast),tfTo),
                         "Database: <font size=\"1\">(Most often used: pdb, uniref100, embl)</font>",
                         hBL(" ",comboDatabases),
                         "Blast service:",
                         hBL(" ",comboClass.panel()),

                         hBL(" ",comboNT),
                         cbCache()
                         ),
            pSide=vBpnl(
                        SCROLLABLE_TRACKS_VIEWPORT_WIDTH,
                        pSettings,
                        pSimple,
                        hBL("Details ",toggl().actionCollapse(false,pSettings).actionCollapse(true, pSimple),"#",pnl(butGo))
                        ),
            pLeftRight=pnl(new GridLayout(1,2),proteinList.scrollPane(),scrllpn(0L,pSide)),
            pMain=newCNSEW(pLeftRight,align.headLine(this),null,null,null);
								adMainTab(pMain,this,null);
    }

    private final ChTextField
        tfOutputWidth=new ChTextField("999999").cols(7,true,true),
        tfFrom=new ChTextField("1").cols(6,true,true),
        tfTo=new ChTextField("9999").cols(6,true,true)/*, tfMaxLen=new ChTextField("  3000")*/;
    private void updateDatabases(){
        vDatabases.clear();
        final SequenceBlaster b=HotswapUtil.instance(comboClass,SequenceBlaster.class,true);
        if (b==null) return;
        final String s;
        if (comboNT.i()==0) {
            b.setAAQuerySequence("A");
            s=" residue";
        }else {
            b.setNTQuerySequence("A");
            s=" nucleotide";
        }
        final String dd[]=b.getAvailableDatabases();
        if (sze(dd)>0) {
            adAll(dd,vDatabases);
            butGo.enabled(true);
            comboDatabases.s(dd[0]);
        } else {
            vDatabases.add("Error: no Database");
            butGo.enabled(false);
        }
        labFirst.setText("First "+s);
        labLast.setText("Last "+s);
        proteinList.repaint();
    }
				private Runnable todo[];
    private void compute() {
        StrapProtein pp[]=proteinList.getSelectedProteins();
        if (pp.length==0 && align.getVisibleProteins().length<4) pp=align.getVisibleProteins();
        if (pp==null||pp.length==0) { StrapAlign.errorMsg("Select a protein!"); return;}
								int count=0;
								this.todo=new Runnable[pp.length];
        final int mode=comboNT.i();
        for(StrapProtein p : pp) {
            if (!proteinList.isEnabled(p)) continue;
            final SequenceBlaster blaster=lastRunning=HotswapUtil.instance(comboClass,SequenceBlaster.class,true);
            if (blaster==null)  { StrapAlign.errorMsg("Could not instantiate"+comboClass);return;}
            final String seq;
            if (mode==0)  seq=p.getResidueTypeAsStringUC();
            else {
                final byte nts[]=p.getNucleotidesCurrentStrand();
                if (nts==null  && p.containsOnlyACTGNX()) seq=p.getResidueTypeAsStringUC();
                else if(nts!=null) {
                    seq=toStrg(mode==1 ? nts:DNA_Util.onlyTranslated(nts,p.areNucleotidesCurrentStrandTranslated()));
                } else seq=null;
            }
            if (seq==null) return;
            final int from=mini(maxi(1,pp.length>1 ? 1 : atoi(tfFrom)),seq.length());
            final int to=mini(maxi(from,pp.length>1 ? MAX_INT : atoi(tfTo)),seq.length());
            final String query=seq.substring(from-1,to);
            if (mode==0 && !(p.countResidues()>50 && p.containsOnlyACTGNX())) blaster.setAAQuerySequence(query);
            else blaster.setNTQuerySequence(query);

            blaster.setDatabase(comboDatabases.toString());
            final Result r=new Result(blaster,p,mode==0,query);
            todo[count++]=ChDelay.newThread(r,"RUN",null);
            adTab(" ... ",r, (JTabbedPane)this);
        }
        ChDelay.newThread(this,"RUN",null).start();
    }
    public Object run(final String id,final  Object arg) {
        if (id=="RUN") {
            final Runnable rr[]=todo;
            if (rr!=null)
                for(Runnable r : rr)
                    if (r!=null && !isDisposed())
                        r.run();

        }
        return null;
    }
    private final ChButton butGo=new ChButton("GO").t(ChButton.GO).li(this).tt("Invokes BLAST computation.");
    public void actionPerformed(ActionEvent ev) {
        StrapAlign.errorMsg("");
        final Object q=ev.getSource();
        if (q==comboNT || q==comboClass) {
            final int mode=comboNT.i();
            proteinList.setFilter(mode==1 ? ProteinList.NT_or_ACTGN : mode==2 ? ProteinList.NT : 0);
            proteinList.repaint();
            updateDatabases();
        }
        if (q==butGo) compute();
        if (q==proteinList) {
            final int nP=proteinList.getSelectedProteins().length;
            tfFrom.setEnabled(nP==1);
            tfTo.setEnabled(nP==1);
        }
        final int iDB=radioGrpIdx(q,radioDatabase);
        if (iDB>=0) {
            comboDatabases.s(iDB==0?Blaster_ebi_ac_uk.DB_PDB : iDB==1?Blaster_ebi_ac_uk.DB_UNIPROT :Blaster_ebi_ac_uk.DB_UNIREF90);
        }

    }
    private class Result extends JPanel implements Disposable,ChRunnable, ActionListener {
        private final SequenceBlaster blaster;
        private final AsciiView tpResult=new AsciiView((File)null);
        private final ChButton BUT_applyFormat=new ChButton("Apply",ChButton.DISABLED);

        private StrapProtein protein;
        private JLabel labInfo=new JLabel();
        private String query;
        private String database;
        private boolean isComputed;

        public void dispose() {
            try { ((Disposable)blaster).dispose();}catch(Exception e){}
            try { ((Disposable)lastRunning).dispose();}catch(Exception e){}
        }

        Result(SequenceBlaster bl,StrapProtein p, boolean isAA,String query) {
            this.query=query;
            protein=p;
            blaster=bl;
            final String txt=
                "query="+blaster.getQuerySequence()+
                "\nprogram="+blaster.getClass().getName()+
                "\ndatabase="+blaster.getDatabase()+
                "\nis calculating please wait ...\n";
            pcp(KEY_CONTAINS_AA,"",tpResult);
            tpResult.setText(new ByteArray(txt));
            closeOnKeystroke(tpResult,new Object[]{Result.this}, DISPOSE_ON_CtrlW);
            pcp(ACTION_DATABASE_CLICKED,"load",tpResult);
            addActLi(StrapAlign.li(),tpResult);
            pcp(DialogFindInText.KEY_ID, (isAA ? "BLAST_AA" : "BLAST_NT"),tpResult);
            final String msgFetch=
                "<h2>Downloading proteins in the blast result</h2>"+
                "<i>PACKAGE:charite.christo.strap.</i>"+
                " Each entry in the  blast-result starts with a greater than sign "+
                "followed by a datatbase identifier and the name of a sequence entry. "+
                "The datatbase identifier which is <b>left from the colon</b> acts as a hyperlink and opens the entry in the Web-browser. "+
                "The sequence identifier which is <b>right from the colon</b> can be clicked to load the sequence into the alignment pane."+
                "<br><br>"+
                "<i>SEE_DIALOG:DialogFetchSRS</i>"+
                "<i>SEE_DIALOG:DialogFetchPdb</i>";
            final Runnable run=ChDelay.newThread(this,"RUN",null);
            final JComponent changeFormat=vBhB(
                                               "The format of the BLAST output can be changed.",
                                               "For searching (ctrl F) long lines are better.",
                                               hB("Residues per text line ", tfOutputWidth),
                                               pnl(BUT_applyFormat.r(run))
                                               ),

                pTools=vB(
                          hBL(new ChButton("RES_SEL",ChButton.NO_MARGIN).t("Create residue selection of matched regions").li(this)),
                          hBL(ChButton.actionViewMessage("Change Output format",changeFormat).options(ChButton.NO_MARGIN))
                          ),
                pNorth=vB(
                          hBL(
                              toggl().actionCollapse(false,pTools),"Tools   ",
                              ChButton.actionViewMessage("Download proteins...",msgFetch).options(ChButton.NO_MARGIN),
                              noMargn(ChButton.actionCtrl(blaster)),
                              labInfo," ",
                              new ScrollLabel(ChTextViewer.KEYBINDINGS)
                              ),
                          pTools
                          );
            cnsew(
                  this,
                  tpResult.tools().scrollPane(),
                  newCNSEW(pNorth,null,null,ChButton.actionClose(this, OPTION_DISPOSE_RM_CHILDS),null),
                  null,	null,null
                  );
        }
        public Object run(String id, Object arg) {
            if (id=="RUN") {
                final String db_=comboDatabases.toString();
                database=db_.substring(db_.lastIndexOf(' ')+1);
                ByteArray xml=CacheResult.isEnabled() ? blaster.getResultXml() : null;
                if (xml==null) { blaster.compute();  this.isComputed=true; xml=blaster.getResultXml();}
                if (xml!=null) ChDelay.invokeAndWait(this,"RUN_EDT",xml);
            }
            if (id=="RUN_EDT") {
                ByteArray xml=(ByteArray)arg;
                final File fResult=createTempFile(".blast");
                final OutputStream fOut;
                try {  fOut=fileOutStream(fResult); } catch(IOException e){ putln(" caught in DialogBlast ",e); return null;}
                final BlastParser parser=new BlastParser();
                ByteArray when=null;
                String whenSection=null;
                final AbstractBlaster abstrB=deref(blaster, AbstractBlaster.class);
                if (abstrB!=null) when=CacheResult.getValue(abstrB.getClass(), whenSection="/day/"+abstrB.cacheSK('S'), abstrB.cacheSK('K'), null);
                if (when==null) when=new ByteArray(0);
                parser.setDay(when, this.isComputed);
                try {
                    fOut.write((fResult+"\n"+shrtClasNam(blaster)+" "+database+"\nQuery="+query+"\n").getBytes());
                } catch(IOException ex){}
                parser.parseXML(xml,maxi(10,atoi(tfOutputWidth)),fOut);
                closeStrm(fOut);
                final int count=parser.countHits(), countN=parser.countNovelHits();
                {
                    final String n=protein.getName();
                    final int l2=n.length()/2;
                   setTtleAt(this,"<font size=\"2\">"+(l2>10?n.substring(0,l2)+"<br>"+n.substring(l2):n)+ (countN>0 ? " *":"")+"</font>", DialogBlast.this);
                }
                //if (isComputed) CacheResult.put(blaster,"/day/",when,true); !!!!!!!!!!!!!!!!!
                if (abstrB!=null) CacheResult.putValue(abstrB.getClass(), whenSection, abstrB.cacheSK('K'), when,true);

                xml=null;
                tpResult.setFile(fResult);
                pcp(AsciiView.KEY_BLAST_RESULT,"",tpResult);

                final ChTextComponents tools=tpResult.tools();
                tools.underlineRefs(0);
                labInfo.setText(countN>0 ? countN+" Novel Blast hits." : "#hits = "+count);
                if (count>countN && countN>0) tools.highlightOccurrence("(NEW)",null,null,HIGHLIGHT_NOT_IN_SCROLLBAR, PINK);
                BUT_applyFormat.setEnabled(true);
                tpResult.revalidate();

            }
            return null;
        }
        private ArrayList<ResidueAnnotation> vSelections=new ArrayList();
        public void actionPerformed(ActionEvent ev) {
            final Object q=ev.getSource();
            final String cmd=ev.getActionCommand();
            if (cmd=="RES_SEL") {
                StrapProtein pp[]=StrapAlign.getInstance().getSelectedProteins();
                if (sze(pp)==0) pp=StrapAlign.getInstance().getVisibleProteins();
                final ByteArray xml=blaster.getResultXml();
                if (sze(xml)==0) { error("Blast not finished yet"); }
                else {
                    BlastResult bResult=new BlastResult(xml);
                    final ArrayList<ResidueAnnotation> v=this.vSelections;
                    v.clear();
                    for(StrapProtein p : pp) {
                        final boolean[] selected=bResult.getMatchingAminoacids(p.getResidueTypeExactLength());
                        if (countTrue(selected)>0) {
                            ResidueAnnotation a=null;
                            for(ResidueAnnotation ra : p.getResidueAnnotations()) if (SELECTION_NAME.equals(ra.getName())) a=ra;
                            if (a==null) a=new ResidueAnnotation(p);
                            a.setValue(ResidueAnnotation.POS, boolToTxt(selected,1,"-",(ByteArray)null));
                            a.setValue(ResidueAnnotation.NAME,SELECTION_NAME);
                            a.setValue(ResidueAnnotation.GROUP, "blast");
                            p.addResidueSelection(a);
                            v.add(a);
                        }
                    }
                    if (sze(v)>0)  {
                        final String T="Residue selections created by Blast";
                        final Object
                            pApply=vBpnl(
                                         "#TB Apply created residue selections",
                                         hBL("Compute multiple sequence alignment using only these matching parts ",new ChButton("REALIGN").t("Realign ...").li(this))
                                         ),
                            pNorth=vBpnl(T,hB(" ",pApply," "));
                        StrapAlign.showInJList(0L,null, v.toArray(), pNorth, null).showInFrame(ChFrame.AT_CLICK,T);
                        StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,9);
                    }
                    else ChMsg.msgDialog("I analyzed "+pp.length+" proteins but could not create residue selections from blast result");
                }
            }
            if (cmd=="REALIGN") ResidueSelectionUtils.openDialogRealign(toArry(vSelections,ResidueAnnotation.class));
        }
    }

}
