package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.strap.interfaces.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;

import static java.lang.System.*;
/**HELP

This dialog allows the pairwise comparison of a set of proteins.
<i>PACKAGE:charite.christo.strap.extensions.</i>

Numeric (float) values are computed for pairs of proteins using the method selected by the user.

As an example  "sequence dissimilarity" yields the proportion of amino acids that are different
in both sequences.

<br>
These values can be visualized as:
<ul>
<li>A table with each row and column representing a protein.</li>

<li>A graph of proteins. The  proteins are the graph nodes.
The length of the  graph edges which connect  the  protein nodes   represent the  distance.</li>

</ul>

<i>SEE_DIALOG:DialogBarChart</i>
<i>SEE_CLASS:SequenceDisSimilarity</i>

@author Christoph Gille

*/
public class DialogCompareProteins extends AbstractDialogJTabbedPane implements ActionListener {
    private final ChTextField tfFrom,tfTo;
    private final ProteinList selOfProtRow,selOfProtCol;
    private final ChCombo comboClass;
    private final JComponent pMain,listProtRow,listProtCol,pBoth=pnl(new GridLayout(1,2));
    private void layoutPCenter() {
        pMain.remove(listProtRow);
        pMain.remove(pBoth);
        pBoth.removeAll();
        switch(RADIO_GRAPH_TABLE.i()) {
        case 0: pMain.add(listProtRow,BorderLayout.CENTER); break;
        case 1:
            pBoth.add(listProtRow);
            pBoth.add(listProtCol);
            pMain.add(pBoth,BorderLayout.CENTER);
            break;
        case 2:
            pMain.add(pBoth,BorderLayout.CENTER);
            break;
        }
        pMain.revalidate();
        pMain.repaint();
    }

    private final ChCombo RADIO_GRAPH_TABLE=new ChCombo("graph","table (n x m matrix)","list of protein pairs").li(this).save(getClass(),"mode");

    public void actionPerformed(ActionEvent ev) {
        StrapAlign.errorMsg("");
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        if (q==RADIO_GRAPH_TABLE) { comboClass.revalidate(); layoutPCenter();}
        if (q==comboClass) { selOfProtRow.repaint(); selOfProtCol.repaint();}

        if (cmd=="GO") {
            StrapProtein
                ppRow[]=selOfProtRow.getSelectedProteins(),
                ppCol[]=selOfProtCol.getSelectedProteins();
            int from=atoi(tfFrom)-1, to=atoi(tfTo);
            if (to==99999) to=MAX_INT;
            if (from<0) from=0;
            Container result=null;
            String tab=null;
            if (ppRow.length==0) ppRow=align.getProteins();
            final Object clas=comboClass.getSelectedItem();
            if (clas==null) return;
            final ProteinProteinValue sharedInstance=(ProteinProteinValue)comboClass.instanceOfSelectedClass(null);
            switch(RADIO_GRAPH_TABLE.i()) {
            case 0:
                if (ppRow.length>2) {
                    if (isAssignblFrm(ProteinProteinSimilarity.class,clas))
                        error("The class <pre>"+clas+"</pre>can probably not be applied for drawing graphs<br> because the values that are computed from pairs of proteins are not \"distances\".<br>A distance has small values for similar and large values for dissimilar proteins.");
                    result=new ResultCompareProteins(clas,sharedInstance,ppRow,null,from,to).getPanel();
                    tab="graph representation";
                }
                else  StrapAlign.errorMsg("Select at least 3 proteins");
                break;
            case 1:
                if (ppCol.length==0) {
                    ppCol=ppRow.length<2 ? align.getProteins() :  ppRow;
                }
                tab="table";
                result=new ResultCompareProteins(clas,sharedInstance,ppRow,ppCol,from,to).getPanel();
                break;
            case 2:
                tab="pairs";
                final String cn=clas.toString();
                if ( (!cn.endsWith(CLASS_SequenceDisSimilarityAsAligned))|| ChMsg.yesNo("Are you sure you want to use "+cn.substring(cn.lastIndexOf('.')+1)+"<br>and not AlignmentScore?") )
                    result=new ResultCompareProteinPairs(clas,sharedInstance,from,to);
                break;
            }
            adTab(tab,result, (JTabbedPane)this);
        }
    }
    public DialogCompareProteins() {
        selOfProtRow=new ProteinList(align,0);
        selOfProtCol=new ProteinList(align,0);
        listProtRow=selOfProtRow.scrollPane();
        listProtCol=selOfProtCol.scrollPane();
        tfFrom=new ChTextField("1").cols(6,true,true);
        tfTo=new ChTextField("99999").cols(6,true,true);
        comboClass=StrapAlign.classChoice(ProteinProteinValue.class).save(DialogCompareProteins.class,"clazz");
        pcp(KEY_CHOICE_CLASS,comboClass,this);
        pcp(KEY_ENABLED,comboClass,selOfProtCol);
        pcp(KEY_ENABLED,comboClass,selOfProtRow);
        comboClass.li(this);
        final JComponent b=vB(
                              comboClass.panel(),
                              hB("Alignment range  from: ",tfFrom,"  to: ",tfTo,"(inclusive) ","#"),
                              pnl("View option: ",RADIO_GRAPH_TABLE," ",new ChButton("GO").t(ChButton.GO).li(this).tt("show as table or graph"), cbCache())
                              );
        pMain=newCNSEW(null,align.headLine(this),b,null,null);
        layoutPCenter();
        adMainTab(pMain,this,null);
        pcp(StrapAlign.KEY_CHOICE_PROTEIN,selOfProtRow,this);
    }

}
