package charite.christo.strap;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import charite.christo.protein.*;
import charite.christo.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.*;

import java.util.*;

/**HELP
   Based on the current sequence alignment two aligned proteins are
   compared residue by residue and all different residues are collected
   in a table.

   The table can be used in different ways:

   <ol>
   <li>
   Homology modeling: Some protein viewers such as the Swiss-viewer
   (http://www.expasy.org/spdbv/  can substitute residues in a
   3D-structure.

   With this dialog the commands  to replace
   the residues of the 3D-structure by the residues of a raw
   sequence can be generated.
   Alignment gaps are not considered.

   <br>
   <b>Customization: </b> These commands can be can be modified  using the customize dialog.

   The contents of the n-th column  is referred to by backslash n.
   </li>
   <li>
   Differing positions may be used to generate residue annotations.
   </li>
   </ol>

   <b>Related resources:</b>    PUBMED:14983078

   <i>SEE_CLASS:ResidueAnnotation</i>
   <i>SEE_CLASS:ResidueAnnotationView</i>
   @author Christoph Gille

*/
public class DialogDifferentResidues extends AbstractDialogJTabbedPane implements ActionListener  {

    public void actionPerformed(ActionEvent ev) {
        StrapAlign.errorMsg("");
        StrapProtein pSeq=choiceProt.getStrapProtein(),pPdb=choiceProt3D.getStrapProtein();
        if (pSeq==null || pPdb==null || pSeq==pPdb) {StrapAlign.errorMsg("Select two different proteins "); return;}
        adTab("<sup>"+pSeq+"<br>"+pPdb+"</sup>", new Result(pSeq,pPdb),  (JTabbedPane)this);
    }
    public final static String SWISS_VIEWER="SwissViewer";
    private ProteinCombo choiceProt3D,choiceProt;

    public DialogDifferentResidues() {
        setSettings(this, replaceResiduesLine());
        final JComponent pCenter=vBpnl(
                                      pnl("#TB Select target protein (e.g. a pdb-file)",choiceProt3D=new ProteinCombo(align,0)),
                                      pnl("#TB Select source protein e.g. a sequence file)",choiceProt=new ProteinCombo(align,0)),
                                      new ChButton("GO").li(this).t(ChButton.GO)
                                      ),
            pMain=newCNSEW(pCenter,align.headLine(this),null,null,null);
        adMainTab(pMain,this,null);
    }
    private static Customize replaceResiduesLine;
    private static Customize replaceResiduesLine() {
        if (replaceResiduesLine==null) {
            replaceResiduesLine=new Customize("replaceResiduesLine",new String[] {
                                                  SWISS_VIEWER+" $X = select in 0 num \\10; mutate $X to \"\\05\";",
                                                  SWISS_VIEWER+" $X = select in 0 pos \\02+1; mutate $X to \"\\05\";",
                                                  "Pymol select strapMut, ( ///\\12/\\10 ); mutate PROTEIN, strapMut, \\07; delete strapMut;"
                                              },
                                              "Generate a text from the table rows.\nThe contents of the table cells are addressed by backslash followed by to digits indicating the row  index",Customize.LIST);
        }
        return replaceResiduesLine;
    }
    // ----------------------------------------
    private static String formatChain(byte b) {char c=(char)b; return is(LETTR_DIGT,c) ? String.valueOf(c) : ""; }
    private class Result extends JPanel implements ActionListener {
        public void actionPerformed(ActionEvent ev) {
            StrapAlign.errorMsg("");
            final String cmd=ev.getActionCommand();
            if (cmd=="SAVE") {
                java.io.File f=file(STRAPOUT+"/replaceResidues");
                wrte(f,tpReplace.getText());
                edFile(-1,f,modifiers(ev));
            }
            if (cmd=="RPLC") replacementCommands();
            if (cmd=="SEL") {
                final int ii[]=table.getRowCount()==1 ? new int[]{0} : ChJTable.selectedRowsConverted(table);
                if (ii.length==0) { StrapAlign.errorMsg("Error: No table rows selected");return;}
                final ResidueAnnotation sors[]=new ResidueAnnotation[ii.length];
                final String group="compared to "+pSeq;
                for(int i=0;i<ii.length;i++) {
                    int row=ii[i];
                    final ResidueAnnotation s=sors[i]=new ResidueAnnotation(pPdb);
                    s.add(ResidueAnnotation.NAME,table.getValueAt(row,8)+"=>"+table.getValueAt(row,7),true);
                    s.add(ResidueAnnotation.GROUP,group,true);
                    s.add(ResidueAnnotation.POS,String.valueOf(atoi(table.getValueAt(row,2))+1),true);
                    s.add(ResidueAnnotation.COLOR,"255,255,128",true);
                    pPdb.addResidueSelection(s);
                }
                StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_ADDED,111);
            }
        }
        private ChJTable table;
        private StrapProtein pSeq,pPdb;
        private JTextPane  tpReplace;
        public Result(StrapProtein pSeq, StrapProtein pPdb) {
            this.pSeq=pSeq; this.pPdb=pPdb;
            int n=pSeq.countResidues();
            Object[][] data=null;
            while(true){
                int iRow=0;
                for(int iA=0,count=0;iA<n;iA++) {
                    int
                        posFrom=pSeq.getResidueColumn(iA),
                        iATo=pPdb.column2nextIndex(posFrom),
                        posTo=pPdb.getResidueColumn(iATo);
                    if (posFrom<0 || posTo<0) continue;
                    if (iATo>=0 && iA<=pSeq.countResidues() && posFrom==posTo && pSeq.getResidueType(iA)!=pPdb.getResidueType(iATo)) {
                        final char asFrom=(char)pSeq.getResidueType(iA);
                        final char asTo=(char)pPdb.getResidueType(iATo);
                        final Object[] row={
                            intObjct(count++),
                            intObjct(iA),
                            intObjct(iATo),
                            intObjct(posFrom),
                            intObjct(posTo),
                            String.valueOf(asFrom),
                            String.valueOf(asTo),
                            StrapProtein.toThreeLetterCode(asFrom),
                            StrapProtein.toThreeLetterCode(asTo),
                            intObjct(pSeq.getResidueNumber(iA)),
                            intObjct(pPdb.getResidueNumber(iATo)),
                            formatChain(pSeq.getResidueChain(iA)),
                            formatChain(pPdb.getResidueChain(iATo))};
                        if (iRow<sze(data)) data[iRow]=row;
                        iRow++;
                    }
                }
                if (data==null) data=new Object[iRow][]; else break;
            }
            final String SP=" of the source protein", TP=" of the target protein",
                C1="One letter code ", C3="Three letter code ", RI="Residue index ", AC="Alignment column ", RN="residue number ", RC="residue chain ";
            table=new ChJTable(new ChTableModel("Idx","Idx from","Idx to","Col from","Col to","AA from","AA to","AA from","AA to","Num from","Num to","Chain from","Chain to").setData(data),0)
                .headerTip("Index of the line", RI+SP, RI+TP, AC+SP, AC+TP, C1+SP, C1+TP, C3+SP, C3+TP, RN+SP, RN+TP, RC+SP, RC+TP);
            comboReplace=replaceResiduesLine().newComboBox();

            final JComponent
                butSubst=new ChButton("RPLC").t("Generate replacement commands").li(this),
                butResidueSelection=new ChButton("SEL").t("Create residue selections").li(this).tt("create  annotations for protein "+pPdb+"<br>for each different position"),
                panSubst=ChUtils.vB(
                                    "#ETB",
                                    enlargeFont(comboReplace,0.8),
                                    pnl(butSubst, " ",replaceResiduesLine.newButton().options(ChButton.NO_MARGIN))
                                    ),
                panSor=pnl("#ETB",butResidueSelection,pPdb),
                pSouth=vB(panSubst,panSor),
                pPdbp=hB(smallHelpBut(DialogDifferentResidues.this),"#","Different Residues","#",ChButton.actionClose(this, OPTION_DISPOSE_RM_CHILDS));
            cnsew(this,scrllpn(0L,table),pPdbp,pSouth,null,null);
        }
        private JComboBox comboReplace;
        public void replacementCommands() {
            StringBuffer sb=new StringBuffer();
            int rows[]=ChJTable.selectedRowsConverted(table);
            if (rows.length==0) rows=count01234(table.getRowCount());
            int nCol=table.getColumnCount();
            String line=(String)comboReplace.getSelectedItem();
            boolean isSwissViewer=line.startsWith(SWISS_VIEWER);
            if (isSwissViewer) sb.append("please do \nopen from \"disk\" \"").append(pPdb.getFile()).append("\";\n");
            for(int iRow=0;iRow<rows.length;iRow++) {
                if (line==null || line.indexOf(' ')<0) return;
                String l=rplcToStrg("\\\\","@",line.substring(line.indexOf(' ')));
                for(int iCol=0;iCol<nCol;iCol++) {
                    l= rplcToStrg("\\"+iCol/10+iCol%10, ""+table.getValueAt(rows[iRow],iCol),l);
                }
                l=rplcToStrg("@","\\\\",l);
                sb.append(l).append('\n');
            }
            if (isSwissViewer) sb.append("\nthank you");
            monospc(tpReplace=new JTextPane());
            tpReplace.setText(sb.toString());
            ChButton bSave=new ChButton("SAVE").t("save as file \"replaceResidues\"").li(this);
            new ChFrame("replaceResidues").ad(newCNSEW(scrllpn(0L,tpReplace),null,pnl(bSave),null,null)).show();
        }
    }
}
