package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.strap.interfaces.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import static java.lang.System.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP

This dialog aims at loading many PDB-files from file servers.  The user enters a text containing
pdb-codes such as <I>1prp</I>, <I>pdb1ryp</I> or <I>1SBC.</I>

Other parts of the text are ignored.

The optional chain identifier must be preceeded by colon,vertical bar or underscore.

   @author Christoph Gille
*/
public class DialogFetchPdb extends AbstractDialogJTabbedPane implements ActionListener {

    public final static String BUT_LAB_Highlight="Highlight pdb-codes";
    private ChCombo comboPdbSite;
    private ChTextArea ta;
    private ChTextComponents.Highlight highlight;
    public void actionPerformed(ActionEvent ev) {
        StrapAlign.errorMsg("");
        final String cmd=ev.getActionCommand();
        final Object q=ev.getSource();
        final ChTextComponents tools=ta.tools();
        if (q==ta) { tools.removeHighlight(highlight); highlight=null;}
        if (cmd==BUT_LAB_Highlight || cmd=="GO") {
            final Color color=new Color(0x80FF4400,true);
            final ByteArray ba=tools.byteArray();
            final int selStart=ta.getSelectionStart(), selEnd=ta.getSelectionEnd();
            final int[] fromTo=findIds(ba, selStart<selEnd?selStart:0, selStart<selEnd?selEnd:MAX_INT);
            final int n=fromTo.length/2;
            if (n==0) {  StrapAlign.errorMsg("No PDB codes found in the text!"); return;}
            if (cmd==BUT_LAB_Highlight) {
                tools.removeHighlight(highlight);
                tools.addHighlight(highlight=ChTextComponents.newHighlight(fromTo, color,HIGHLIGHT_NOT_IN_SCROLLBAR));
                repaint();
            } else if (cmd=="GO" && fromTo!=null) {
                final String pdbBase=comboPdbSite.toString();
                final String urls[]=new String[n];
                final String fileNames[]=new String[n];
                final byte bb[]=ba.getBytes();
                for(int i=n; --i>=0; ) {
                    final int f=fromTo[2*i],t=fromTo[2*i+1];
                    if (t-f<4) continue;
                    final String code4=new String(bb,f, t-f).toLowerCase();
                    String chain="";
                    if (t+1<bb.length) {
                        final byte b0=bb[t], b1=bb[t+1], b2=bb.length>t+2 ? bb[t+2] : 0;
                        if ( (b0==':' || b0=='_' || b0=='|') && w(b1) && !w(b2)) {
                            chain="_"+ (char)b1;
                        }
                    }
                    fileNames[i]="pdb"+code4+chain+".ent";
                    urls[i]=rplcToStrg("??",code4.substring(1,3),rplcToStrg("????",code4,pdbBase));
                }
                adTab("Files fetched",new FilesFetchedFromServer(urls,fileNames), (JTabbedPane)this);
            }
        }
        if (cmd=="EXAMPLE") {
            ta.setText(
                       "Here come two pdb codes: 1sbc 1sbt \n"+
                       "They are highlighted when the  \"highlight\" button is pressed.\n"+
                       "They are download when the \"GO\" button is pressed.\n"+
                       "Words consisting of  four letters or digits  are assumed to be PDB identifiers.\n"+
                       "\n"+
                       "Optionally a chain letter can be specified after a colon such as 1ryp:L\n"+
                       ta.getText());
        }
    }
    public DialogFetchPdb() {
        setSettings(this,Customize.pdbSite);
        comboPdbSite=Customize.customize(Customize.pdbSite).newComboBox().save(DialogFetchPdb.class,"ftp-site");
        //final UrlListRenderer renderer=new UrlListRenderer(comboPdbSite);
        final JComponent  pButtons1=pnl(new ChButton(BUT_LAB_Highlight).li(this),new ChButton("GO").t(ChButton.GO).li(this)),
            pButtons=vB(pButtons1,WATCH_MOVIE+MOVIE_Load_Proteins+" "+MOVIE_Load_Proteins_web_pdb);
        this.ta=new ChTextArea("").saveInFile("DialogFetchPdb");
        pcp(KEY_IF_EMPTY, "Enter here a text containing  PDB IDs",this.ta);
        addActLi(this,this.ta);
        ta.tools().enableUndo(true).enableWordCompletion(dirWorking());
        final JComponent pNorth=vBpnl(
                                      align.headLine(this),
                                      pnl(
                                          new ChButton("EXAMPLE",ChButton.NO_MARGIN).t("For newbies: show example").li(this),
                                          "  Alternatively:",StrapAlign.b(CLASS_DialogFetchSRS).options(ChButton.NO_MARGIN)
                                          )
                                      ),
            butLog=null,
            pMain=newCNSEW( scrllpn(0L,ta),pNorth,vB(newCNSEW(comboPdbSite,null,null,butLog,newButTestProxy()),pButtons),null,null);
        adMainTab(pMain,this,null);
    }
    private static boolean w(byte b) { return  '0'<=b && b<='9' || ('a'<=b && b<='z' || 'A'<=b && b<='Z');}

    private static int[] findIds(final ByteArray ba, final int selBeg, final int selEnd) {
        final byte bb[]=ba.getBytes();
        final int n=mini(ba.getEnd(),selEnd);
        int[] fromTo=null;
        for(;;) {
            int count=0;
            for(int i=selBeg; i<n-6; i++) {
                if ((bb[i]|32)=='p' && (bb[i+1]|32)=='d' && (bb[i+2]|32)=='b' && w(bb[i+3]) && w(bb[i+4]) && w(bb[i+5]) && w(bb[i+6])) {
                    if (fromTo!=null) { fromTo[2*count]=i+3; fromTo[2*count+1]=i+7; }
                    count++;
                }
            }
            for(int i=selBeg; i<n-3; i++) {
                if ( (i==0||!w(bb[i-1])) && '0'<=bb[i] && bb[i]<='9' && w(bb[i+1]) && w(bb[i+2]) && w(bb[i+3]) && ( i+4==n || !w(bb[i+4])) ){
                    if (fromTo!=null) { fromTo[2*count]=i; fromTo[2*count+1]=i+4;}
                    count++;
                }
            }
            if (fromTo==null) fromTo=new int[2*count]; else break;
        }
        return fromTo;
    }
}
