package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import javax.swing.*;
import javax.swing.text.JTextComponent;
import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.util.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
/**HELP

With this dialog a number of sequence files can be loaded from public servers.

A text containing  sequence IDs is pasted into the text area.

The program automatically identifies and highlights certain types of IDs within the text
on pressing the respective highlight button.

Subsequently, they can be  fetched from the HTTP-server.

Pressing the button  <i>BUTTON:"UNIPROT"</i> loads the highlighted entries from the SRS-server.
In parallel a batch request is sent to EBI which delivers large amounts of UNIPROT files efficiently.

It is important to chose the appropriate server.

<br>
<br><br><b>List of loaded files:</b>
The  files are downloaded into the project directory and listed.
<br>
<i>JCOMPONENT:FilesFetchedFromServer#docuSnapshot()</i>
<br><b>Figure:</b> Downloaded files appear in a list.
<br>
<br>

<br><br><b>Adding servers:</b> New servers can be added in the
customize dialog <i>ICON:IC_CUSTOM</i>.  STRAP forms an
URL using a base and an ID.  The URL-mask should contain an asterisk
"*" which is replaced by the ID.

<br><br><b>Related Web-sites:</b> http://www.ensembl.org/Multi/martview

<i>SEE_DIALOG:DialogFetchPdb</i>

   @author Christoph Gille
*/
public class DialogFetchSRS extends AbstractDialogJTabbedPane implements ActionListener {
    private final static String sTYPE[]={"HSLV_ECOLI\" (Swissprot)","Q8ICC3\" (letter + letters/digits)",
                                         "M57965\" (letter + digits)","gi|20259105 \" (NCBI)","NM_002726\" (NCBI)"};
    private final static int HSLV_ECOLI=0,Q8ICC3=1,M57965=2,GI_20259105=3,NM_002726=4;
    private int fromTo[];
    private final ChTextArea ta;
    private final JComponent  pLoad=new ChPanel();
    private static Customize customize;
    private void layoutLoadPanel() {
        pLoad.removeAll();
        final String dbs[]=customize.getSettings();
        pLoad.setLayout(new GridLayout(1,dbs.length));
        for(String db: dbs) {
            ChButton b=new ChButton("L").li(this).t(delSfx(":",db));
            pcp("L",db,b);
            pLoad.add(b);
        }
    }
    private ChTextComponents.Highlight highlight;
    public void actionPerformed(ActionEvent ev) {
        StrapAlign.errorMsg("");
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        final ChTextComponents tools=ta.tools();
        final ByteArray ba=tools.byteArray();
        if (q instanceof Customize) { layoutLoadPanel(); revalidate();}
        if (cmd=="EXAMPLE")
            ta.setText(
                       "HSLV_ECOLI or HSLV_BACSU are swissprot id's and are loaded by pressing \nthe respective [highlight]-button and pressing the button \"Uniprot\".\n"+
                       "Q8ICC3  can also be loaded from Uniprot\n\n"+
                       "M57965 can  be loaded from embl.\n\n"+
                       "UPI0000672BF8 ore UPI0000685C48 are loaded from UNIPARC\n"+
                       "You can specify further databases.\n"+
                       ta.getText());
        final Color color=new Color(0x80FF4400,true);
        for(int i=sTYPE.length;--i>=0;) {
            if (cmd.endsWith(sTYPE[i])) {
                tools.removeHighlight(highlight);
                this.fromTo=findIds(ta,ba,i);
                tools.addHighlight(highlight=ChTextComponents.newHighlight(this.fromTo, color,HIGHLIGHT_NOT_IN_SCROLLBAR));
                repaint();
            }
        }

        if (q==ta) { tools.removeHighlight(highlight); this.fromTo=null; this.highlight=null;}
        if (cmd=="L") {
            final String DB=toStrg(gcp("L",q));
            if (ba.onlyWhiteSpace()) { StrapAlign.errorMsg("Type/paste a text containing sequence IDs ");return;}

            if (this.fromTo==null) StrapAlign.errorMsg("Highlight the IDs by pressing one of the 'highlight..'-buttons");
            else {
                final int n=sze(fromTo)/2;
                if (n>0) {
                    final String urls[]=new String[n];
                    final String fileNames[]=new String[n];
                    final int len=ba.getEnd();
                    for(int i=0;i<n;i++) {
                        final int f=fromTo[2*i],t=fromTo[2*i+1];
                        if (len>=t) {
                            final String id=ba.newString(f,t).trim();
                            fileNames[i]=id;
                            urls[i]=DB+id;
                        }
                    }
                    adTab(delSfx(":",DB),new FilesFetchedFromServer(urls,fileNames), (JTabbedPane)this);
                }
            }
        }
    }
    public DialogFetchSRS() {
        if (DialogFetchSRS.customize==null) {
            final String ss[]={"EXPASY:","UNIPROT:","EMBL:","NCBI:","KEGG_AA:","KEGG_NT:", "ENSEMBL:"};
            DialogFetchSRS.customize=new Customize("fetchProteinFile",ss,"URLs to get protein files from. Asterisk $* is replaced by the ID.",Customize.LIST);
        }
        addActLi(this,DialogFetchSRS.customize);
        setSettings(this, DialogFetchSRS.customize);
        final Container pHighlight=new JPanel(new GridLayout(sTYPE.length,1));
        for(String t : sTYPE) {
            ChButton b=new ChButton("like \""+t,ChButton.NO_MARGIN).li(this);
            b.setHorizontalAlignment(LEFT);
            pHighlight.add(b);
        }
        layoutLoadPanel();
        ta=new ChTextArea("").saveInFile("DialogFetchSRS");
        final Object
            pWeb=pnl( newButTestProxy().options(ChButton.NO_MARGIN),new ChButton("EXAMPLE",ChButton.NO_MARGIN).t("For newbies: show example").li(this)),
            pNorth=vBpnl(align.headLine(this),pWeb),
            pSouth=vB(hB("Last step: Chose the appropriate data sources/servers for downloading:","#"),pnl(pLoad),WATCH_MOVIE+MOVIE_Load_Proteins+" "+MOVIE_Load_Proteins_web_uniprot),
            pHL=newCNSEW(pnl(pHighlight),"Second step: <br>Highlight all IDs in the text",null,null,null),
            pCenter=newCNSEW(ta.tools().scrollPane(),null,null,pHL,null);
        pcp(KEY_IF_EMPTY,"First step: Enter here a text with  protein IDs \n",ta);
        addActLi(this,ta.tools().enableWordCompletion(dirWorking()).enableUndo(true));
        adMainTab(newCNSEW(pCenter,pNorth,pSouth,null,null), this,null);
    }
    // ----------------------------------------
    private static boolean t(char type, byte b) {
        switch(type) {
        case 'n': return '0'<=b && b<='9';
        case 'L': return 'A'<=b && b<='Z';
        case 'W': return '0'<=b && b<='9' || 'A'<=b && b<='Z';
        case 'w': return '0'<=b && b<='9' || 'a'<=b && b<='z' || 'A'<=b && b<='Z';
        }
        return false;
    }
    private static int[] findIds(JTextComponent ta,ByteArray ba,int type) {
        final int iStart,iEnd, length=ba.getEnd();
        {
            int iE=ta.getSelectionEnd(), iS=ta.getSelectionStart();
            if (iS<1 || iS>=iE-1) { iS=0; iE=length;}
            iStart=iS;
            iEnd=iE;
        }
        final byte bb[]=ba.getBytes();
        int[] fromTo=null;
        for(;;) {
            int count=0;
            for(int i=iStart; i<iEnd-2; i++) {
                switch(type) {
                case NM_002726:
                case HSLV_ECOLI:
                    if (bb[i]=='_') {
                        final int nLeft=count("LEFT",bb,i-1), nRight=count("DIGIT_OR_LETTER",bb,i+1);
                        if (nRight>2 && nLeft>1) {
                            if (fromTo!=null) { fromTo[2*count]=i-nLeft; fromTo[2*count+1]=i+nRight+1;}
                            count++;
                        }
                    }
                    break;
                case Q8ICC3:
                    if ((i==0 || !t('w',bb[i-1])) && t('L',bb[i]) && t('W',bb[i+1]) ) {
                        final int nDigitsOrCapitals=count("DIGIT_OR_CAPITAL",bb,i+1);
                        int nDigits=0;
                        for(int j=0;j<nDigitsOrCapitals;j++) if (t('n',bb[i+1+j])) nDigits++;
                        if (nDigitsOrCapitals>4 && nDigits>0) {
                            if (fromTo!=null) { fromTo[2*count]=i; fromTo[2*count+1]=i+nDigitsOrCapitals+1;}
                            count++;
                        }
                    }
                    break;
                case GI_20259105:
                    if (i>=3 && (i==3 || !t('w',bb[i-4]) && (bb[i-3]|32)=='g' && (bb[i-2]|32)=='i'  && (bb[i-1]=='|' || bb[i-1]==':')) ) {
                        final int nDigits=count("DIGIT",bb,i+1);
                        if (nDigits>0) {
                            if (fromTo!=null) { fromTo[2*count]=i; fromTo[2*count+1]=i+nDigits+1;}
                            count++;
                        }
                    }
                    break;
                case M57965:
                    if ((i==0 || !t('w',bb[i-1])) && t('L',bb[i]) && t('n',bb[i+1]) ) {
                        final int nDigits=count("DIGIT",bb,i+1);
                        if (nDigits>4 && (bb.length<=i+nDigits+1 || !t('w',bb[i+nDigits+1]))) {
                            if (fromTo!=null) {  fromTo[2*count]=i; fromTo[2*count+1]=i+nDigits+1;}
                            count++;
                        }
                    }
                    break;
                }
            }
            if (fromTo==null)  fromTo=new int[2*count]; else break;
        }
        return fromTo;
    }
    // ----------------------------------------
    private static int count(String what,byte[] bb,int i) {
        if (bb==null) return 0;
        int count=0;
        final char t=what=="DIGIT" ? 'n' : what=="DIGIT_OR_CAPITAL" ? 'W' : what=="DIGIT_OR_LETTER" ? 'w' : 0;
        if (t!=0)  while (i<bb.length && t(t,bb[i])) { count++; i++;}
        else if (what=="LEFT") while (i>=0 && t('w',bb[i])) { count++; i--;}
        return count;
    }
}
