package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.*;
import java.io.File;
import static java.lang.Boolean.*;
import java.awt.Color;
import java.awt.Rectangle;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP

This dialog is used to underline all occurrences of sequence
patterns.

The table contains three check-boxes to toggle visualization on and
off for the sequence alignment window, the horizontal scrollbar, and
the 3D-views.

A 4th check-box indicates whether the search string is applied to
the <b>amino acid</b> sequence or <b>nucleotide</b> sequence.

The latter option is only applied if the amino acid sequence is
translated from nucleotide sequence.

The search string complies with  UNIX regular expressions.
We explain briefly the notation with some useful
examples.

Many search queries can be entered in one line and must be separated by white space.
<ul>
<li>"FSP" is a simple search string identifying all
occurrences of "FSP" in the amino acid sequence.  Each position is
highlighted in the specified color.</li>

<li>"F.P" leaves the second residue undefined since dots match any
character.</li>

<li>"IFSP|TFSP" is equivalent to "[IT]FSP" or "(IF|TF)SP" or "IFSP&nbsp;&nbsp;&nbsp;TFSP" and
highlights all occurrences of IFSP and TFSP. Vertical bars separate
alternatives. </li>

<li>"(.....)\1" selects all penta repeats. A backslash-number is a
reference to a parenthesized group.</li>

<li>"[FYW][FYW]" selects adjacent aromatic amino acids. Square
brackets enclose groups of letters at one single sequence
position. </li>

<li>"CGT|CGC|CGA|CGG|AGA|AGG" highlights all codons for
Arg. Searching the underlying nucleotide sequence requires the
check-box "n" not to be checked</li>

<li>"[ke]{3,99}"highlights all occurrences of 3 to 99 continuous
positions containing either Lys or Glu.</li>

</ul>

The Check-boxes in the first  table columns determine, where the matches are shown.
<ol>
<li><b>1D</b> Highlight matches in alignment panes</li>
<li><b>3D</b> Highlight matches in 3d-Backbone views</li>
<li><b>SB</b> Highlight matches in the scroll-bar</li>
<li><b>NT</b> Highlight patterns in nucleotide sequence. The aa-sequence of the protein must be translated from nt-sequence</li>
</ol>

<li>list of blank separated reg-expressions<br>case insensitive<br>E.g. '[AG]..[KR]' matchas all occurrence of<br> Ala of Gly in one position <br>and Lys or Arg at the position n+3<br>The two dotes match any.");

See  WIKI:Regular_expression for a more detailed description.
<br>

One table row may contain several expressions separated by white space.

Each expression will create a  <i>CLASS_REF:ResidueSelection</i>residue selection object
for each protein.

<i>SEE_DIALOG:DialogSelectionOfResiduesMain</i>
<i>SEE_CLASS:ResidueSelection</i>
@author Christoph Gille
*/
public class DialogHighlightPattern extends AbstractDialogJPanel implements  Disposable, StrapListener, javax.swing.table.TableModel, ActionListener, ChRunnable {
    private String labError="";
    private final ChLabel LAB_ERROR=new ChLabel();
    protected void setError(String msg) {
        if (msg!=labError) {
            LAB_ERROR.t(msg).fg(C(0xFF0000));
            if ((msg.length()==0) != (labError.length()==0)) { LAB_ERROR.revalidate();}
            labError=msg;
        }
    }
    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        final HighlightPattern hp=vPattern.get(jTable.getChRenderer().editedRow());
        if (q instanceof JComboBox) {
            for(HighlightPattern h : vPattern.asArray()) {
                h.setStyle(((JComboBox)q).getSelectedIndex());
            }
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
        }
        if (hp!=null) {
            if (q==jTable.getChRenderer().textField() && cmd==ACTION_TEXT_CHANGED) {
                final String s=q.toString();
                final boolean addRemove= (sze(s)==0)!=(sze(hp.getText())==0);
                hp.setText(s);
                if (addRemove)  hp.addToProteins(getProteins());
                StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
            }
            for(int i=bbNext.length; --i>=0;) {
                if (q==bbNext[i] && i!=2) {
                  hp.nextAndPrevSelection.goNextMatch(cmd,hp.getResidueSelections(align.getProteins()));
                  return;
                }
            }
        }
        if (q==cbDnaForward && !cbDnaForward.s()) cbDnaRevCompl.s(true);
        if (q==cbDnaRevCompl && !cbDnaRevCompl.s()) cbDnaForward.s(true);
        StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
    }
    /* -----------------------*/
    /* ---   TableModel   --- */
    /* -----------------------*/
    private final ChButton [] bbNext=new ChButton[5];
    public int getRowCount(){ return vPattern.size();};
    public int getColumnCount(){return 10;};
    public String getColumnName(int col){
        return
            col==0 ? "1D":
            col==1 ? "3D":
            col==2 ? "SB":
            col==3 ? "NT":
            col==4 ? "One or several search strings. Separated by white space.":
            "";
    }
    public Class getColumnClass(int col){ return Object.class; }
    public boolean isCellEditable(int row, int col){return true;}
    public Object getValueAt(int row, int col){
        HighlightPattern hp=vPattern.get(row);
        if (hp==null) return "error";
        final String txt=hp.getText();
        if (col==3) return hp.isNucleotide() ? TRUE:FALSE;
        if (col==4) return txt;
        if (5<=col && col<=9) {
            if (sze(txt)==0) return blankLabel();
            if (col==5+2) return hp.butColor();
            final ChButton b= bbNext[col-5];
            return b;
        }
        return ((hp.getVisibleWhere() & (col==0 ? VisibleIn123.SEQUENCE : col==1 ? VisibleIn123.STRUCTURE : VisibleIn123.SB))!=0) ? TRUE:Boolean.FALSE;
    }
    private void addToProtein() {
        final StrapProtein pp[]=getProteins();
        for(HighlightPattern hp:vPattern.asArray()) hp.addToProteins(pp);
        StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
    }
    private void newTableRow(){
        for(HighlightPattern p:vPattern.asArray())
            if (p.getText().length()==0)
                return;
        vPattern.add(new HighlightPattern("",this));
        jTable.revalidate();
    }
    public final void setValueAt(Object o, int row, int col){
        final HighlightPattern hp=vPattern.get(row);
        if (hp==null || o==null) return;
        final int visible0=hp.getVisibleWhere();
        int mask=0;
        boolean changed=false, addRemove=false;
        switch(col) {
        case 0:  mask=VisibleIn123.SEQUENCE; break;
        case 1:  mask=VisibleIn123.STRUCTURE; break;
        case 2:  mask=VisibleIn123.SB; break;
        case 3:  changed|=hp.setNucleotide(o==TRUE);
            setError(hp.isNucleotide() && !ProteinList.isEnabled(getProteins(),ProteinList.NT)  ? "Error selecting \"nt\" : No nucleotide sequence present." : "");
            break;
        case 4:
            final String s=o.toString();
            addRemove|= (sze(s)==0)!=(sze(hp.getText())==0);
            changed|=hp.setText(s);
            newTableRow();
            jTable.repaint();
            break;
        default:
        }
        if (mask!=0) {
            final int visible= o==TRUE ? (visible0|mask) :  (visible0&~mask);
            addRemove|= (visible0==0) != (visible!=0);
            if (visible!=visible0) {
                changed=true;
                hp.setVisibleWhere(visible);
            }
        }
        if (addRemove) {
            hp.addToProteins(getProteins());
            newTableRow();
        }
        if (changed) {
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,99);
            jTable.repaint();
        }
    }
    public final void addTableModelListener(javax.swing.event.TableModelListener li){}
    public final void removeTableModelListener(javax.swing.event.TableModelListener li){}

    // ----------------------------------------
    /* --- VisibleIn123 --- */
    private final ChJTable jTable;
    private final UniqueList<HighlightPattern> vPattern=new UniqueList(HighlightPattern.class);
    private final ChButton
        cbDnaForward=toggl("fwd").li(this).s(true),
        cbDnaRevCompl=toggl("rev-compl").li(this).s(true);
    private final AbstractButton  cbOnlySelectedPP=cbox("Search only in selected proteins");
    public DialogHighlightPattern() {
        vPattern.add(new HighlightPattern("",this));
        final File file=file(".strap_sequence_patterns.dat");
        final String[] txt=readLines(file);
        addShutdownHook1( ChDelay.newThread(this,"SAVE",file));
        if (txt!=null)
            for(String t:txt)
                vPattern.add(new HighlightPattern(t,this));
        jTable=new ChJTable(this,ChJTable.EXX_ROW_HEIGHT|ChJTable.DEFAULT_RENDERER|ChJTable.NO_REORDER);
        jTable.setRowSelectionAllowed(false);
        jTable.getChRenderer().textField().li(this);
        rtt(jTable);
        for(int i=bbNext.length; --i>=0;) {
            final String
                t=i==0 ? "|<" : i==1 ? "<" : i==2 ? "" : i==3 ? ">" : ">|",
                tt=i==2 ?  "Set Color" : "Go to "+ (i==0 ? "first": i==1 ? "previous" : i==3 ? "next" : "last")+" match";
            if (i!=2) {
                final ChButton b=bbNext[i]=new ChButton(t,ChButton.NO_MARGIN|(i!=2?ChButton.PAINT_IN_TABLE:0)).li(this).tt(tt);
                pcp(ChRenderer.KEY_KEEP_COLOR,"",b);
                pcp(ChRenderer.KEY_EDITOR_COMPONENT, b.cln().options(ChButton.NO_MARGIN), b);
            }
        }
        for(int i=getColumnCount();--i>=0;) {
            final int w= i==4 ? 0 : i==5+2 ? 4*EX : 5*EX/2;
            if (w!=0)  jTable.setColumnWidth(i,w,false);
        }
        jTable.setCellSelectionEnabled(false);
        final JComponent pSouth=pnl("Search in nucleotides: ",cbDnaForward.cb(),cbDnaRevCompl.cb(),cbOnlySelectedPP, AbstractVisibleIn123.newComboStyle().li(this));
        cnsew(this,scrllpn(0L,jTable),pTop(),vBpnl(pSouth,monospc(LAB_ERROR)),null,null);
        addToProtein();
    }
    // ----------------------------------------
    protected int getOrientations() {
        int r=0;
        if (cbDnaRevCompl.isSelected()) r|=2;
        if (cbDnaForward.isSelected()) r|=1;
        return r;
    }
    // ----------------------------------------
    public final Object run(String id, Object arg) {
        if (id=="SAVE") {
            final StringBuffer sb=new StringBuffer();
            for(HighlightPattern hp:vPattern.asArray()) hp.save(sb);
            wrte((File)arg,sb);
        }
        return null;
    }
    // ----------------------------------------
    @Override public final void dispose() {
        for(StrapProtein p: align.getProteins())
            p.removeResidueSelection(HighlightPattern.Selector.class);
        StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
        clearMarchingAnts();
    }
    // ----------------------------------------
    public final void handleEvent(StrapEvent ev) {
        final int type=ev.getType();
        if (type==StrapEvent.PROTEINS_ADDED || type==StrapEvent.PROTEINS_KILLED) addToProtein();
        else if ( (type&StrapEvent.FLAG_RESIDUE_TYPES_CHANGED)!=0) StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,333);
    }
    private final void clearMarchingAnts() {
        final StrapView v=align.getAlignmentPanelWithFocus();
        if (v!=null) addMarchingAnt(ANTS_SEARCH, v.getAlignmentPane(), (Rectangle)null,(Color)null, (Color)null );
    }

    private StrapProtein[] getProteins() {
        if (cbOnlySelectedPP.isSelected()) return align.getSelectedProteins();
        final StrapView view=align.getAlignmentPanelWithFocus();
        return view!=null ? view.getProteins(): new StrapProtein[0];
    }

}
