package charite.christo.strap;
import charite.christo.interfaces.*;
import charite.christo.*;
import charite.christo.protein.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP

With this Dialog one or more nucleotide sequences can be translated
such that they optimally match a given amino acid sequence. The tool
determines the reading frame and and the untranslated regions and
introns. The goal is to achieve an optimal alignment of the translated
sequence and the reference amino acid sequence.
<br><br>
The tool requires that the program <b>bl2seq</b> is installed and in the executable path.
The bl2seq is part of the NCBI-toolkit.

@author Christoph Gille
*/
public class DialogInferCDS extends AbstractDialogJPanel implements ActionListener {

    private ProteinCombo comboProtein;
    private ProteinList proteinList=new ProteinList(StrapAlign.getInstance(), ProteinList.NT_or_ACTGN).selectAll(0);
    private ChButton butUndo=new ChButton("Undo").enabled(false).li(this), butLog=new ChButton("Log").enabled(false).li(this);

  private final ByteArray baLog=new ByteArray(9999);
  public DialogInferCDS() {
    comboProtein=new ProteinCombo(align,ProteinList.NOT_ACTGN);
    final Container
      pWestNorth=vBpnl(
                       "Reference amino acid sequence", comboProtein,
                       " ",
                       pnl(new ChButton("GO").t(ChButton.GO).li(this), butUndo, butLog)
                       ),
      pWest=newCNSEW(null,pWestNorth,null,null,SCROLLABLE_TRACKS_VIEWPORT_WIDTH),
      pEast=newCNSEW(proteinList.scrollPane(),"List of nucleotide sequence to be translated",null,null,null),
      pCenter=pnl(new GridLayout(1,2),scrllpn(0L,pWest),pEast);
      cnsew(this,pCenter,align.headLine(this),null,null,null);
  }

  public void actionPerformed(ActionEvent ev) {
        StrapAlign.errorMsg("");
        final String cmd=ev.getActionCommand();
        final Object q=ev.getSource();
        final ChLog log=new ChLog(this.baLog.setEnd(0));
        if (cmd=="GO" || q==butUndo) {
            final StrapProtein pp[]=proteinList.getSelectedOrAllProteins();
            int count=0;
            if (cmd=="GO") {
                final StringBuffer sbError=new StringBuffer(), sbWarning=new StringBuffer();
                final StrapProtein pRef=(StrapProtein)comboProtein.getProtein();
                if (pp.length==0) sbError.append("Select at least one protein with nt-sequence.\n");
                if (pRef==null) sbError.append("No reference amino acid sequence selected.\n");
                for(StrapProtein p : pp) {
                    if (p==pRef) continue;
                    if (!p.containsOnlyACTGNX() && p.getNucleotides()==null) {
                        sbWarning.append(p+" does not contain nucleotides (Letters A, C, T and G)\n");
                    }
                }

                if (sze(sbWarning)>0) error(sbWarning);
                if (sze(sbError)>0) { error(sbError); return; }
                final byte aa[]=pRef.getResidueTypeExactLength();
                for(StrapProtein p : pp) {
                    if (p==pRef) continue;
                    this.baLog.append('=',99).append('\n').append("Sequence ").append(p).append('\n');
                    byte[] ntSeq=p.getNucleotides();
                    if (ntSeq==null && p.containsOnlyACTGNX()) ntSeq=p.getResidueType();
                    if (ntSeq==null) continue;
                    final Bl2seq bl2seq=new Bl2seq(ntSeq, aa, Bl2seq.BLASTX, log, 0L);
                    bl2seq.compute();
                    final Bl2seq.Hit hh[]=bl2seq.getHits();
                    if (sze(hh)>0) {
                        boolean[] trans=Bl2seq.getTranslatedNucleotides(ntSeq, aa, hh[0],  hh,0L);
                        putln("trans="+boolToTxt(trans));
                        if (countTrue(trans)>20) {
                            final boolean reverse=hh[0].getFrom(0)>hh[0].getTo(0);
                            if (reverse) {
                                reverseArray(trans,0,MAX_INT);
                            }
                            pcp(butUndo,new Object[]{p.getOrientationOfReadingFrame(), p.areNucleotidesCurrentStrandTranslated()},p);
                            p.setTranslatedStrand(reverse?StrapProtein.REVERSE_COMPLEMENT : StrapProtein.FORWARD);
                            p.setNucleotidesCurrentStrandTranslated(trans);
                            count++;
                        }
                    }
                }
                if (count>0) {
                    butUndo.enabled(true);
                    butLog.enabled(true);
                }
            }
            if (q==butLog) new AsciiView(this.baLog).showInFrame("DialogInferCDS log",ChFrame.AT_CLICK);

            if (q==butUndo) {
                for(StrapProtein p : pp) {
                    final Object oo[]=gcp(butUndo,p,Object[].class);
                    if (oo!=null) {
                        p.setTranslatedStrand((String)oo[0]);
                        p.setNucleotidesCurrentStrandTranslated((boolean[])oo[1]);
                        count++;
                    }
                }
            }
            if (count>0) StrapAlign.dispatchLater(StrapEvent.NUCL_TRANSLATION_CHANGED,33);
        }
    }

}
