package charite.christo.strap;
import charite.christo.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP
   Makes texshade commands for the structure prediction
   Text lines starting with a percent sign are ignored since "%" preceeds comments in WIKI:LaTeX.
   <br><br>
   The line valid for helices is preceeded by "Hh" and the line valid for extended (i.e. beta sheet) is preceeded by "Ee".
   Generally speaking, each line is started with a  character set followed by  TeXshade code.

   @author Christoph Gille
*/
public class DialogPredictTexshade extends AbstractDialogJPanel implements java.awt.event.ActionListener {
    private final StrapProtein  proteins[];
    private final char prediction[][];
    private final ChTextArea ta=new ChTextArea(5,80);
    private final String name;
    private final ChButton TOG=toggl("Add to PDF ");
    public DialogPredictTexshade(char prediction[][], StrapProtein pp[],String name) {
        proteins=pp;
        this.name=name;
        this.prediction=prediction;
        if (prediction==null) return;
        final Object
            //butTex=new ChButton("TEX",ChButton.NO_MARGIN).t("Test PDF for one protein").li(this).i(IC_PDF).tt("For the first protein the PDF will be generated and displayed"),
            butTexshade=StrapAlign.b(CLASS_Texshade).t("Open PDF Dialog").options(ChButton.NO_MARGIN),
            panDetails=pnl(/*butTex,*/ butTexshade,smallHelpBut(this)),
            pnl=newCNSEW((ta), panDetails,null,null,"#TB TeXshade -- generating PDF -- only the first selected protein");
        setDefaultText();
        cnsew(this,null,hB(TOG.actionCollapse(false,pnl).cb(),"#"),pnl,null,null);
    }
    private void setDefaultText() {
        final String
            S0="\\shaderegion{"+ResidueAnnotation.PROTEIN_NO+"}{"+ResidueAnnotation.TEXSHADE_RESIDUES+"}{",
            T=
            "\\allmatchresidues{Black}{White}{upper}{}\n"+
            "\\nomatchresidues{Black}{White}{upper}{}\n"+
            "\\conservedresidues{Black}{White}{upper}{}\n"+
            "\\similarresidues{Black}{White}{upper}{}\n"+
            "\\shadingmode{none}\n"+
            "Hh "+S0+"White}{Red}\n"+
            "Ee "+S0+"Black}{Yellow}\n"+
            "%X  \\feature{bbottom}{$TEX_PROTEIN}{"+ResidueAnnotation.TEXSHADE_RESIDUES+"}{fill:X}{}\n";
        ta.t(T);
    }
    private void getText(StringBuffer sb,StrapProtein texProteins[], int iTab) {
        if (prediction==null) return;
        int maxLen=0;
        for(char cc[] : prediction) if (maxLen<sze(cc)) maxLen=cc.length;
        final ByteArray ba=ta.tools().byteArray().append('\n');
        final byte T[]=ba.getBytes();
        final int ends[]=ba.getLineEnds();
        final boolean[] selA=new boolean[maxLen];
        sb.append("%   ").append(name).append('\n');
        for(int iP=-1; iP<proteins.length; iP++) {
            if (iP>=0 && !cntains(proteins[iP],texProteins)) continue;
            for(int iL=0; iL<ends.length; iL++) {
                final int E=ends[iL];
                final int B=nxtE(NO_SPC, T, iL==0?0:ends[iL-1], E);
                if (E-B<4 || T[B]=='%') continue;
                if (strstr(ResidueAnnotation.PROTEIN_NO,T,B,E)>0) {
                    final int spc=nxt(SPC,T,B,E);
                    if (iP<0 || spc<0) continue;
                    for(int iA=selA.length; --iA>=0;) selA[iA]= iA<prediction[iP].length && strchr(prediction[iP][iA], T,B,spc)>=0;
                    if (countTrue(selA)>0)
                        sb.append(ResidueAnnotation.replaceVariables(selA, 0,proteins[iP],ba.newString(spc,E+1), this.name, texProteins));
                } else  if (iP==0) {
                    final String ins=ba.newString(B,E+1);
                    if (iTab==0 || sb.indexOf(ins)<0) sb.append(ins);
                }
            }
        }
    }
    private static DialogPredictTexshade instance;
    public static void getTexts(StringBuffer sb,StrapProtein proteins[]) {
        if (instance!=null) { instance.getText(sb,proteins,0);return;}
        final DialogPredictTexshade dd[]=childs(StrapAlign.getInstance().getDialogPanel(),DialogPredictTexshade.class);
        for(int i=0;i<dd.length;i++) if (dd[i].TOG.s()) dd[i].getText(sb,proteins,i);

    }
    public void actionPerformed(java.awt.event.ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        if (cmd=="TEX")  {
            this.instance=this;
            final StringBuffer sb=new StringBuffer();
            StrapProtein p=proteins[0];
            for(int i=proteins.length; --i>=0;) {
                if (prediction[i]==null) continue;
                for(char c:prediction[i]) {
                    if (c!=' ' && c!=0) p=proteins[i];
                }
            }
            getText(sb, new StrapProtein[]{p},0);
            ResidueSelectionUtils.texshade(sb.toString(),p);
            DialogPredictTexshade.this.instance=null;
        }
        if (cmd=="RST" && (sze(ta.toString())==0 || ChMsg.yesNo("Really discard text and set the default text?"))) setDefaultText();
    }
}
