package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import charite.christo.strap.extensions.Hydrophobicity;
import charite.christo.strap.interfaces.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;
import java.util.*;
import java.util.List;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;
import javax.swing.table.TableModel;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
/**
   @author Christoph Gille
*/
public class DialogPredictTmHelicesMulti extends DialogPredictTmHelices2 implements StrapListener {
    private ProteinList proteinList;
    private ChJList classList;
    private int iResult;
    @Override public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        if (cmd=="GO") {
            final StrapProtein pp[]=proteinList.getSelectedProteinsAtLeast(1);
            if (sze(pp)==0) return;
            final Object classes[]=classList.getSelectedValues();
            if (sze(classes)==0) StrapAlign.errorMsg("Select at least one method");
            else adTab( String.valueOf(iResult++),new ResultPane(pp,classes), (JTabbedPane)this);
        } else super.actionPerformed(ev);
    }
    private ChCombo comboSettings_;
    private ChCombo comboSettings() {
        if (comboSettings_==null) comboSettings_=StrapAlign.classChoice(TransmembraneHelix_Predictor.class).classRenderer(HasSharedControlPanel.class);
        return comboSettings_;
    }
    private final ChButton butSettings() {
        return ChButton.actionViewMessage(
                                          "more details ...",
                                          vBpnl(
                                                "method",
                                                comboSettings(),
                                                " ",
                                                ChButton.actionSharedCtrl(comboSettings()).cln().t("Settings"),
                                                helpBut(comboSettings()).t("Documentation"),
                                                smallSourceBut(comboSettings()),
                                                newButTestProxy()
                                                )
                                          );
    }
    @Override public JTabbedPane getPanel() {
        proteinList=new ProteinList(align,0);
        classList=new ChJList(StrapPlugins.allClassesV(TransmembraneHelix_Predictor.class),ChJTable.CLASS_RENDERER);
        Container
            pCenter=pnl(new GridLayout(1,2),proteinList.scrollPane(),new JScrollPane(classList)),
            pSouth=pnl(new GridLayout(1,3),
                               new ChButton("GO").t(ChButton.GO).li(this),
                               butSettings(),cbCache()
                               );

        adMainTab(newCNSEW(pCenter,align.headLine(this),pSouth,null,null), this,null);
        align.addListener(this);
        return this;
    }

    private class ResultPane extends JPanel implements TableModel,CellEditorListener,Disposable,TableCellRenderer,ActionListener,Runnable,PaintHook,MouseMotionListener {
        private JComponent panOverview=new ChPanel();
        {addPaintHook(this,panOverview).addMouseMotionListener(this);}

        public void mouseDragged(MouseEvent ev) {
            final StrapProtein p=proteins[0];
            final StrapView v=align.getAlignmentPanelWithFocus();
            if (p!=null && v!=null) {
                int col=p.getResidueColumn(maxi(0,ev.getX())*p.countResidues()/panOverview.getWidth());
                if (col<0) col=p.getMaxColumn();
                v.setHorizontalScrollPosition(col*v.charB().width);
                panOverview.repaint();
            }
        }
        public void mouseMoved(MouseEvent ev) {}

        public boolean  paintHook(JComponent c,Graphics g, boolean after) {
            if (after && c==panOverview && sze(proteins)==1) {
                final StrapProtein p=proteins[0];
                final int width=c.getWidth(),height=c.getHeight(),nRes=p.countResidues();
                if (nRes==0) return false;
                final int rgb[]=new int[nRes];
                final BufferedImage imageLineT=new BufferedImage(nRes,1,BufferedImage.TYPE_INT_ARGB);
                final int nR=results.length;
                for(int iR=nR; --iR>=0;) {
                    final AbstractResidueSelection ss[]=results[iR].vSelections.asArray();
                    for(int iS=0; iS<ss.length; iS++) {
                        if (ss[iS].getProtein()!=p) continue;
                        ss[iS].getRGB(rgb);
                        imageLineT.setRGB(0,0,nRes,1, rgb,0,nRes);
                        final int y=height*iR/nR;
                        final int h=height*(iR+1)/nR-y-1;
                        try {
                            g.drawImage(imageLineT,0,y,width,y+(h>0 ? h : 1), 0,0,nRes,1,this);
                        } catch(Exception e){}
                    }
                }
                final StrapView sv=align.getAlignmentPanelWithFocus();
                if (sv!=null) {
                    final Rectangle r=sv.getVisibleRowsAndCols();
                    final int from=p.column2nextIndex(r.x)*width/nRes;
                    final int to=p.column2thisOrPreviousIndex(r.x+r.width)*width/nRes;
                    g.setColor(C(0x8888FF,128));
                    if (from<to) g.fillRect(from,0,to-from,height);
                }
            }
            return true;
        }
        public void actionPerformed(ActionEvent ev) {
            final String cmd=ev.getActionCommand();
            final Object q=ev.getSource();
            revalidate();
            showHide();
            if (cmd=="TXT" && proteins!=null && results!=null) {
                final StrapProtein ppAli[]=new StrapProtein[(1+results.length)*proteins.length];
                for(int iR=0;iR<results.length;iR++) {
                    final StrapProtein pp[]=results[iR].getProteins();
                    for(int iP=0; iP<pp.length; iP++) {
                        ppAli[iP*(results.length+1)]=pp[iP];
                        final StrapProtein pPredict=new StrapProtein();
                        final char[][] predictions=results[iR].predictor.getPrediction();
                        if (predictions!=null && predictions.length>0) {
                            final char[] pr=predictions[0];
                            if (pr!=null) {
                                pPredict.setResidueType(new String(pr).replace((char)0,'x').replace(' ','x'));
                                final StrapProtein p=pp[iP];
                                pPredict.inferGapsFromGappedSequence(p.getGappedSequence());
                                pPredict.setGappedSequence(pPredict.getGappedSequenceAsString().replace('x',' '));
                            }
                        }
                        final String name=shrtClasNam(results[iR].predictor);
                        pPredict.setName(name);
                        ppAli[iP*(results.length+1)+iR+1]=pPredict;
                    }
                }
                final int residuesPerLine=atoi(ChMsg.input(ChMsg.AT_CLICK,"The result will be shown as a multiple alignment.<br>Long sequence lince will be broken.<br>How many residues per line ?",7,"60"));
                final StringBuffer sb=new StringBuffer();
                ExportAlignment.export(ppAli,ExportAlignment.CLUSTALW|ExportAlignment.FILE_SUFFIX,null,residuesPerLine,null,0f,sb);
                new ChTextViewer(sb).shw(0);
            }
            if (q==togSingle || q==togConsensus) {
                final StrapView sv=align.getAlignmentPanelWithFocus();
                if (sv!=null) sv.getPanel().revalidate();

            }
        }
        public void dispose() {
            for(Result r:results) r.dispose();
            for(AbstractResidueSelection c:consensus) {
                if (c==null) continue;
                Protein p=c.getProtein();
                if (p!=null) { p.removeResidueSelection(c); }
            }
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
            for(StrapView sv:align.getAlignmentPanels()) {
                if (sv==null) continue;
                for(ValueOfResidue vv[] : hydrophobicity)
                    for(ValueOfResidue v : vv)
                        sv.getPlottersV().remove(v);
            }
            StrapAlign.dispatchLater(StrapEvent.VALUE_OF_RESIDUE_CHANGED,111);
        }
        private final StrapProtein proteins[];
        private final AbstractResidueSelection consensus[];
        private final Object classes[];
        private final ChJTable table;
        private Result results[];
        private final ChButton butCtrl[];
        private final ChButton
            togProteins[],
            togProteinsH[],
            togMethods[],
            togHydro[],
            togConsensus=toggl("show<br>Consensus",ChButton.SHOW_HIDE|ChButton.HEIGHT_I).li(ResultPane.this),
            togSingle=toggl("show<br>single",ChButton.SHOW_HIDE|ChButton.HEIGHT_I).li(ResultPane.this);
        private final ChCombo comboConsensusMethod;

        ResultPane(StrapProtein[] pp,Object[] cc) {
            {
                final List v=new ArrayList();
                for(int i=0;i<cc.length;i++) v.add("> "+i);
                comboConsensusMethod=new ChCombo(v.toArray(),0).li(this);
            }
            classes=cc;
            proteins=pp;
            hydrophobicity=new ValueOfResidue[proteins.length][Hydrophobicity.SCALE_NAMES.length];
            consensus=new AbstractResidueSelection[proteins.length];
            table=new ChJTable(this,ChJTable.ICON_ROW_HEIGHT|ChJTable.NO_REORDER).renderer(this);
            table.setCellSelectionEnabled(false);
            table.setDefaultEditor(Object.class,chRen);
            table.getTableHeader().setDefaultRenderer(this);
            table.getTableHeader().setCursor(getCursr('M'));
            togMethods=new ChButton[classes.length];
            togHydro=new ChButton[Hydrophobicity.SCALE_NAMES.length];
            for(int i=togMethods.length; --i>=0;){
                Object o=togMethods[i]=toggl(null,ChButton.SHOW_HIDE|ChButton.PAINT_IN_TABLE).s(true).li(this);
                pcp(ChRenderer.KEY_RENDERER_COMPONENT,weakRef(o),o);
            }
            for(int i=togHydro.length; --i>=0;){
                Object o=togHydro[i]=toggl(null,ChButton.SHOW_HIDE|ChButton.PAINT_IN_TABLE).li(this);
                pcp(ChRenderer.KEY_RENDERER_COMPONENT,weakRef(o),o);
            }

            togProteins=new ChButton[proteins.length];
            for(int i=togProteins.length; --i>=0;) {
                Object o=togProteins[i]=toggl(null,ChButton.SHOW_HIDE|ChButton.PAINT_IN_TABLE).s(true).li(this);
                pcp(ChRenderer.KEY_RENDERER_COMPONENT,weakRef(o),o);
            }

            togProteinsH=new ChButton[proteins.length];
            for(int i=togProteinsH.length; --i>=0;) {
                Object o=togProteinsH[i]=toggl(null,ChButton.SHOW_HIDE|ChButton.PAINT_IN_TABLE).s(true).li(this);
                pcp(ChRenderer.KEY_RENDERER_COMPONENT,weakRef(o),o);
            }

            final JComponent pButtons=pnl(
                 pnl("#TB consensus from all predictions",comboConsensusMethod,togConsensus),
                 togSingle,
                 new ChButton("TXT").li(this).t("export<br>as text")
                 );
            final Container pNorth=newCNSEW(null,null,null,ChButton.actionClose(this, OPTION_DISPOSE_RM_CHILDS),null);
            final JScrollPane sp=new JScrollPane(table);
            JComponent pCenter=sp;
            if (pp.length==1) {
                pCenter=new JSplitPane(JSplitPane.VERTICAL_SPLIT,panOverview,sp);
                panOverview.setPreferredSize(dim(32,32));
                sp.setPreferredSize(dim(32,32));
                panOverview.setBackground(C(0));
            }
            cnsew(this,pCenter,pNorth,pButtons,null,null);
            butCtrl=new ChButton[classes.length];
            results=new Result[classes.length];
            for(int i=0;i<classes.length;i++) {
                final PredictionFromAminoacidSequence predictor=newPredictor(proteins,classes[i]);

                pcp(RUN_AFTER_COMPUTATION,this,results[i]=new Result(predictor,proteins,getClass()));
                results[i].togVisibleSB.s(false);
                if ((butCtrl[i]=ChButton.actionCtrl(predictor))!=null) butCtrl[i].t("details").i("").options(ChButton.NO_MARGIN|ChButton.PAINT_IN_TABLE);
            }
            for(int i=getColumnCount(); --i>=0;) {
                if (i==1) continue;
                final int width=i==0 ?  7*EM : ICON_HEIGHT+1;
                final TableColumn tc=table.getColumnModel().getColumn(i);
                tc.setMaxWidth(width);
                tc.setMinWidth(width);
            }
            chRen.addCellEditorListener(this);
            showHide();
        }
        public void run() { table.revalidate(); table.repaint(); showHide();}
        final ValueOfResidue hydrophobicity[][];
        private void showHide() {
            final StrapView sv=align.getAlignmentPanelWithFocus();
            final boolean isSingle=togSingle.s();
            if (sv==null) return;
            for(int iR=0; iR<results.length; iR++) {
                if (results[iR]==null) continue;
                final AbstractResidueSelection ss[]=results[iR].vSelections.asArray();
                for(int iS=0;iS<ss.length; iS++) {
                    final AbstractResidueSelection s=ss[iS];
                    s.setImage("");
                    s.setStyle(isSingle ? ((results.length-iR-1)<<VisibleIn123.BIT_SHIFT_LINE) + VisibleIn123.LINES : VisibleIn123.IMAGE_NO_FILL);
                    s.setVisibleWhere(isSingle ? VisibleIn123.SEQUENCE : VisibleIn123.IMAGE);
                    final int iMethod=idxOf(results[iR].clazz(),classes);
                    if (iMethod<0) continue;
                    final Protein p=s.getProtein();
                    final int iProtein=idxOf(p,proteins);
                    if (iProtein<0) continue;
                    p.removeResidueSelection(s);
                    if (isSingle && togMethods[iMethod].s() && togProteins[iProtein].s()) p.addResidueSelection(s);
                    s.setColor(method2color(iMethod));
                }
            }
            for(int iP=proteins.length; --iP>=0;) {
                final StrapProtein p=proteins[iP];
                AbstractResidueSelection c=consensus[iP];
                p.removeResidueSelection(c);
                if (!togConsensus.s()) continue;
                if (c==null) {
                    c=consensus[iP]=new AbstractResidueSelection();
                    c.setStyle(VisibleIn123.FULL_BACKGROUND);
                    c.setName("consensus TM-prediction");
                    c.setColor(C(0xFF0000));
                }
                final UniqueList<boolean[]> vPred=new UniqueList(boolean[].class);
                for(Result r: results)
                    for(AbstractResidueSelection s : r.vSelections.asArray()) {
                        if (s==null) continue;
                        final int iMethod=idxOf(r.clazz(),classes);
                        if (s.getProtein()==p && togMethods[iMethod].s()) {
                            vPred.add(s.getSelectedAminoacids());
                        }
                    }
                setPredictions(vPred.asArray(),c);
                p.addResidueSelection(c);
            }
            int countHydro=0;
            for(int iP=proteins.length; --iP>=0;) {
                for(int iH=Hydrophobicity.SCALE_NAMES.length; --iH>=0;) {
                    final boolean shown=togHydro[iH].s() && togProteinsH[iP].s();
                    if (shown) {
                        if (hydrophobicity[iP][iH]==null) {
                            Hydrophobicity h=new Hydrophobicity();
                            h.setProtein(proteins[iP]);
                            h.setScale(Hydrophobicity.SCALE_NAMES[iH]);
                            h.setColor(method2color(iH));
                            hydrophobicity[iP][iH]=(h);
                        }
                        countHydro++;
                        sv.getPlottersV().add(hydrophobicity[iP][iH]);
                    } else sv.getPlottersV().remove(hydrophobicity[iP][iH]);
                }
            }
            if (countHydro>=0) StrapView.button(StrapView.TOG_PLOT_ENTIRE_PANEL).s(false);

            sv.setSpaceBetweenRows(maxi(isSingle ? results.length*3 : 0,countHydro==0 ? 0 : sv.getPanel().getFont().getSize()));
            panOverview.repaint();
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
            StrapAlign.dispatchLater(StrapEvent.VALUE_OF_RESIDUE_CHANGED,111);
        }
        public String getColumnName(int c) {return "";}
        public boolean isCellEditable(int row,int col) {
            final int rowHydrophobFrom=classes.length+2,rowHydrophobTo=rowHydrophobFrom+Hydrophobicity.SCALE_NAMES.length;
            if (row<rowHydrophobFrom) {
            return
                ((col==0 || col==proteins.length+2) && row<classes.length) ||
                (row==classes.length && (col>1 && col<proteins.length+2));
            }
            return  rowHydrophobFrom<=row && row<rowHydrophobTo && col==2+proteins.length ||
                row==rowHydrophobTo && col>1 && col<2+proteins.length;
        }
        public void setValueAt(Object o,int row,int col) {}
        public Class getColumnClass(int c) {return c==0 ? Class.class : JComponent.class;}
        public int getColumnCount() {return proteins.length+3;}
        public int getRowCount() { return classes.length+1  + 1+ Hydrophobicity.SCALE_NAMES.length+1;}
        public Object getValueAt(int row,int col) {
            final int rowHydrophobFrom=classes.length+2,rowHydrophobTo=rowHydrophobFrom+Hydrophobicity.SCALE_NAMES.length;
            final int iProt=col-2;
            if (row<rowHydrophobFrom-1) {
                final int iMethod=row;
                if (iMethod==classes.length) return (iProt>=0 && iProt <proteins.length) ? togProteins[iProt] : "";
                if (col==0) return  butCtrl[iMethod];
                if (col==1) return  classes[iMethod];
                if (iProt==proteins.length) return togMethods[iMethod];
                if (iProt>=0 && iProt<proteins.length && iMethod>=0 && iMethod<classes.length) {
                    if (results[iMethod].isFailed) return iicon(IC_BUG);
                    if (!results[iMethod].isFinished) return iicon(IC_HOURGLASS);
                    return togMethods[iMethod].s() && togProteins[iProt].s() ? method2color(iMethod) : "";
                }
            }
            if (row==rowHydrophobFrom-1) return "";
            if (row>=rowHydrophobFrom && row<rowHydrophobTo) {
                final int iHydro=row-rowHydrophobFrom;
                return
                    col==0 ? "":
                    col==1 ? Hydrophobicity.SCALE_NAMES[iHydro]:
                    col==2+proteins.length ? togHydro[iHydro] :
                    (togHydro[iHydro].s() && togProteinsH[iProt].s() ? method2color(iHydro) : "");
            }
            if (row==rowHydrophobTo) return iProt>=0 && iProt<proteins.length ?  togProteinsH[iProt]: "";
            return "error";
        }
        private Color method2color(int iMethod) {
            final int i=iMethod%7;
            return C(i==0 ? 0xff0000 : i==1 ? 0xffafaf : i==2 ? 0xFF00FF : i==3 ? 0x00FF00 : i==4 ? 0Xff : i==5 ? 0xffFF : 0xffFf00);
        }
        public void addTableModelListener(TableModelListener l){}
        public void removeTableModelListener(TableModelListener l){}
        private ChRenderer chRen=new ChRenderer();
        public Component getTableCellRendererComponent(JTable t,Object o,boolean isSel,boolean foc,int row,int col) {
            if (row==-1) return  col<2 || col>=proteins.length+2 ? blankLabel() : proteins[col-2].getVerticalRendererComponent();
            if (o instanceof JComponent) return (JComponent)o;
            return  chRen.getTableCellRendererComponent(t,o,isSel,foc,row,col);
        }
        public void editingStopped(ChangeEvent ev) {}
        public void editingCanceled(ChangeEvent ev) {}

        private  void setPredictions(boolean[][] predictions, AbstractResidueSelection cons) {
            int len=0;
            final String sMethod=comboConsensusMethod.toString();
            if (chrAt(0,sMethod)=='>') {
                final int num=atoi(sMethod,1);
                for(boolean[] p:predictions) if (p!=null && p.length>len) len=p.length;
                final boolean bb[]=new boolean[len];
                for(int iA=len;--iA>=0;) {
                    int count=0;
                    for(boolean[] p:predictions) {
                        if (p!=null && p.length>iA && p[iA]) count++;
                    }
                    bb[iA]=count>num;
                }
                cons.setSelectedAminoacids(bb,0);

            }
        }

    }
    public void handleEvent(StrapEvent ev) {
        final int type=ev.getType();
        if (type==StrapEvent.ALIGNMENT_SCROLLED) {
            final Component c=getSelectedComponent();
            if (c instanceof ResultPane) ChDelay.repaint(((ResultPane)c).panOverview,222);
        }
    }
}
/*

1Q3I_A.pdb
1MHS_A.pdb
1NRW_A.pdb
1Y8A_A.pdb
1SU4_A.pdb
1WR8_B.pdb

*/
