package charite.christo.strap;
import charite.christo.interfaces.*;
import charite.christo.*;
import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.io.File;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static java.awt.Color.*;
import static java.awt.event.MouseEvent.*;
/**HELP

Each alignment project is stored in one folder.
This file directory is specified by the
user at the beginning of the session and cannot be changed during
the session.  The complete directory path appears in the title bar
of the STRAP application frame.

<br><br><b>Loading protein files:</b> Protein files are loaded into STRAP
by dragging one or more files from the desktop or any other location
into the STRAP-application.
<br><br>
Web-pages may contain links to protein databases.
These links can be dragged from the Web-browser into STRAP
( http://www.bioinformatics.org/strap/dragProteinLink.html for details).
<br><br>
Several STRAP-sessions can be opened at the same time and proteins can be dragged from one to the other.

   @author Christoph Gille
*/
public class DialogProjectFiles extends JPanel implements Disposable,ActionListener,ChRunnable  {
    public final static String BUT_LAB_Load="Load selected";
    private String fileNames[]=NO_STRING;
    private File files[]=NO_FILE;
    private byte fileType[]={};
    private static String description[], sExtensions[];
    private static byte extensions_[][][];
    private Tab tabs_[];
    private int idxSelected=1;
    private final ChButton butLoad=new ChButton(BUT_LAB_Load).li(this);
    private final ChSlider sliderHeight=new ChSlider("sliderH",8,64,32)
        .tt("A fixed cell height  accelerates the drawing of the file view.<br>"+
            "However, an individual cell height (left slider position) looks better");
    {
        addActLi(this,sliderHeight);
        sliderHeight.setPreferredSize(dim(32,32));
        pcp(BOTTOM_TEXT,"Row height",sliderHeight);
    }

    private Tab tab(int i) {
        if (tabs_==null) tabs_=new Tab[extensions().length];
        if (tabs_[i]==null) {
            final Tab t=tabs_[i]=new Tab(i);
            pcp("SP",new OverlayScrollPane(t).addStringMatchButton(null,0L,"ProteinFiles"),t);
            setProcessEvt(FilePopup.instance(),t);
            if (renderer!=null) t.setCellRenderer(renderer);
            addTooltip(FilePopup.instance(),t);
            if (renderer instanceof KeyListener) t.addKeyListener((KeyListener)renderer);
            StrapAlign.getInstance().newDropTarget(t,false);
            pcp(KEY_DRAG_TARGET_DIRCTORY,this.dir,t);

        }
        return tabs_[i];

    }
    private void set(int idx) {
        idxSelected=idx;
        if (pCenter!=null) remove(pCenter);
        add(pCenter=(OverlayScrollPane)(gcp("SP",tab(idx))),BorderLayout.CENTER);
        revalidate();
        repaint();
    }
    private byte[][][] extensions() {
        if (extensions_==null) {
            final  String[] ZIP_EXT={"",".bz2",".Z",".gz"};
            final String ss[]=getCustomSettings(Customize.proteinFileExtensions);
            final int n=ss.length;
            description=new String[n];
            sExtensions=new String[n];
            extensions_=new byte[n][][];
            final StringBuffer sb=new StringBuffer();
            for(int j=0;j<n;j++) {
                final String line=ss[j];
                final int iSharp=line.indexOf('#');
                description[j]=iSharp>0 ? line.substring(iSharp+1) : "";
                final StringTokenizer t=new StringTokenizer(iSharp>0 ? line.substring(0,iSharp) : line);
                final int nE=t.countTokens();
                extensions_[j]=new byte[nE*ZIP_EXT.length][];
                int count=0;
                sb.setLength(0);
                for(int iE=nE;--iE>=0;) {
                    final String sT=t.nextToken();
                    for(String zip : ZIP_EXT) {
                        final String s=(sT.startsWith(".") ? sT: "."+sT)+zip;
                        extensions_[j][count++]=s.getBytes();
                        if (zip=="") sb.append(s).append(" ");
                    }
                }
                sExtensions[j]=sb.toString();
            }
        }
        return extensions_;
    }
    private Container pCenter;
    private ListCellRenderer renderer;
    public DialogProjectFiles(ListCellRenderer r,File directory) {
        dir=directory;
        renderer=r;
        myUpdate();
    }

    private boolean alreadyPainted;
    private ChThread LOOP;
    @Override public void paintComponent(Graphics g) {
        try {super.paintComponent(g);} catch(Exception e){}
        if (!alreadyPainted) {
            LOOP=ChThread.callEvery(999,ChDelay.newThread(this,"LOOP",null),"DialogProjectFiles:"+dir(),0);
            alreadyPainted=true;
        }
    }
    private void enableDisable() {
        int count=0;
        if (tabs_!=null)
            for(Tab t :tabs_) {
                if (t==null) continue;
                for(Object o : t.getSelectedValues()) {
                    if (StrapFileCellRenderer.isEnabled((File)o))
                        count++;
                }
            }
        butLoad.t("Load "+ (count==0 ?  ""  : count==1 ? "one" : String.valueOf(count))+" selected");
        butLoad.bg(count>0 ? GREEN : WHITE).setEnabled(count>0);
    }

    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        final Tab t=tab(idxSelected);
        t.setForeground(BLACK);
        t.setBackground(WHITE);
        if (q instanceof Tab) enableDisable();
        if (cmd=="sliderH") {
            t.setFixedCellHeight(sliderHeight.getValue());
            revalidate();
        }
        final int compareType=cmd=="A" ? COMPARE_ALPHABET : cmd=="M" ? COMPARE_WHEN_MODIFIED : cmd=="S" ? COMPARE_SIZE : 0;
        if (compareType!=0) {
            t.sortFiles(comparator(compareType));
        }
         if (q==butLoad) {
            final StringBuffer sb=new StringBuffer(999);
            for(int i=extensions().length; --i>=0;) {
                for(Object o: tab(i).getSelectedValues()) {
                    sb.append(o).append('\n');
                }
            }
            StrapAlign.getInstance().loadTheProteinsInList(sb,StrapAlign.OPTION_PROCESS_LOADED_PROTS|StrapAlign.OPTION_EVENT_PROTEINS_ADDED);
            putln(sb);
            for(Tab tab:DialogProjectFiles.this.tabs_) if (tab!=null) tab.clearSelection();
        }
        if (radioGrpIdx(q)>=0) set(radioGrpIdx(q));
    }

    public void myUpdate() {
        removeAll();
        readFiles();
        extensions();
        final int n=sExtensions.length;
        final JComponent  pTabs=new ChPanel(new GridLayout(n,1));
        final Object[] cbExtension=radioGrp(sExtensions,0,this);
        for(int i=0;i<n;i++) {
            final JComponent cb=get(i,cbExtension,JComponent.class);
            pTabs.add(cb);
            cb.setToolTipText(description[i]);
        }
        final Object oo[]={
             new GridLayout(5,1),
             sliderHeight,
             new ChButton("A").t("Sort alpha").li(this),
             new ChButton("M").t("Sort time").li(this),
             new ChButton("S").t("Sort size").li(this),
             butLoad
             };
        for(Object c:oo)
            if (c instanceof ChButton)
                ((ChButton)c).setHorizontalAlignment(SwingUtilities.LEFT);
        cnsew(this,pCenter=null,null, null,vB(pnl(oo),"##"), pTabs);
        set(0);
        enableDisable();

    }
    private class Tab extends ChJList  {
        boolean valid;
        File[] ff=NO_FILE;
        final int iType;
        @Override public void processEvent(AWTEvent ev){
            if (isPopupTrigger(ev)) FilePopup.instance().processEvent(this,ev);
            else super.processEvent(ev);
            //final boolean isMetal=UIManager.getLookAndFeel().getClass().getName().indexOf(".metal.")>0;
            //if (!(ev instanceof MouseEvent) || !(isPopupTrigger(ev) || isMetal && ev.getID()==MOUSE_RELEASED))  super.processEvent(ev);
        }
        //@Override public String getToolTipText(MouseEvent e) { return DialogProjectFiles.this.getToolTipText(e);}
        @Override public void paintComponent(Graphics g) {
            if (!valid) {
                final byte[] ft=DialogProjectFiles.this.fileType;
                final String ffn[]=DialogProjectFiles.this.fileNames;
                final File allFiles[]=DialogProjectFiles.this.files;
                int count=0;
                for(int i=0;i<ffn.length;i++) if (ft[i]==iType || iType==0) count++;
                if (ff.length!=count) this.ff=new File[count];
                count=0;
                for(int i=0;i<ffn.length;i++) {
                    if (ft[i]==iType || iType==0) {
                        if (count>=ff.length) break;
                        if (allFiles[i]==null) allFiles[i]=new File(dir(),ffn[i]);
                        this.ff[count++]=allFiles[i];
                    }
                }
                setListData(this.ff);
                valid=true;
            }
            try { super.paintComponent(g);}catch(Exception e){}
            final int n=getModel().getSize();
            final String msg= n==0 ? "no file with the given suffix" : "Note: right-click for file context menu";
            g.setColor(GRAY);
            g.drawString(msg,0,getFixedCellHeight()*n+EX);
        }
        Tab(int i) {
            super(null,ChJTable.RTT|ChJTable.FILE_TRANSFER_HANDLER|ChJTable.ICON_ROW_HEIGHT|ChJTable.DRAG_ENABLED);
            li(DialogProjectFiles.this);
            setFixedCellWidth(300);
            iType=i;
        }
        public void sortFiles( Comparator c) {
            //putln(" c="+c);
            final Set<String> selectedBefore=new HashSet();
            final int selectedIndices[]=new int[selectedBefore.size()];
            sortArry(ff,c);
            setListData(ff);
            repaint();
            int count=0;
            for(Object o : selectedBefore.toArray()) selectedIndices[count++]=idxOf(o,ff);
            setSelII(selectedIndices);
        }
    }
    public void dispose() {
        if(LOOP!=null) LOOP.setEveryMs(-1);
        fileNames=NO_STRING;
        if (tabs_!=null)
            for(Tab tab :tabs_) if (tab!=null) tab.ff=NO_FILE;
    }
    private File dir;
    private File dir() { if (dir==null) dir=file("./"); return dir;}
    public void readFiles() {
        fileNames=dir().list();
        if (fileNames==null) {  fileNames=NO_STRING; StrapAlign.errorMsg("Could not read directory "+dir);}
        //putln(" files.length="+files.length);
        if (files.length<fileNames.length) files=new File[fileNames.length]; else Arrays.fill(files,null);
        if (fileType.length<fileNames.length) fileType=new byte[fileNames.length+99];
        sortArry(fileNames,comparator(COMPARE_ALPHABET));
        final int n=extensions().length;
        final byte extensions[][][]=extensions();
        for(int iF=fileNames.length; --iF>=0;) {
            final String fn=fileNames[iF];
            final int len=fn.length();
            if (len==0) continue;
            final int cLast=fn.charAt(len-1)|32;
            fileType[iF]=(byte)(n-1);
            for(int iType=n; --iType>=0;) {
                final byte[][] ee=extensions[iType];
                for(int iE=ee.length; --iE>=0;) {
                    final byte[] e=ee[iE];
                    if (cLast== e[e.length-1] && /*endsWith(e,fn,len,true)*/ endWith(STRSTR_IC, e,fn)) {
                        fileType[iF]=(byte)iType;
                        break;
                    }
                }
            }
        }
        for(int i=n; --i>=0;) tab(i).valid=false;
        final Tab t=tab(idxSelected);
        //putln(" idxSelected ="+idxSelected+(t!=null));
        if (t!=null) { t.revalidate(); t.repaint(); }
    }

    public Object run(String id,Object arg) {
        if(id=="LOOP") {
            if(DialogProjectFiles.this.isVisible()) {
            final long time=System.currentTimeMillis();
            if (hasChanged(isWindows())) ChDelay.invokeAndWait(this,"REREAD",tab(idxSelected));
            LOOP.setEveryMs(999+9*(System.currentTimeMillis()-time));
            }
        }
        if(id=="REREAD") {
            Tab t=(Tab)arg;
            t.setBackground(RED);
            readFiles();
            ChDelay.afterMS(EDT,ChDelay.BG_COLOR,t,111,WHITE);
        }
        return null;
    }
    private long lastModified;
    private boolean hasChanged(boolean isWin) {
        long lastModi=0;
        if (isWin) {
            final String[] ss=dir().list();
            if (ss==null) return false;
            for(int i=ss.length; --i>=0;) lastModi+=ss[i].hashCode();
        } else lastModi=dir().lastModified();
        final boolean changed=lastModified!=0 && lastModified!=lastModi;
        lastModified=lastModi;
        return changed;
    }
}
