package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.strap.interfaces.*;
import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP

A list of protein files is entered into the text area.

All names of  protein files in the text area are detected and loaded int STRAP.

The files can be specified with relative or absolute file paths.

Use tab-key for WIKI:Word_completion.

<br><br><b>White space:</b> Blanks within file names must be written as "%20".

<br><br><b>Grouping:</b>
Protein files can be grouped by enclosing them in curled brackets.
These proteins will appear in one single row of the multiple sequence alignment
(See <i>DIALOG:DialogManyInOneRow</i>).

<br><br><b>Chain identifier:</b>
The expression  "pdb1ryp_B.ent" loads only chain B of the pdb file  pdb1ryp.ent.

<br>
<br><br><b>Advanced Users: Loading sequence regions:</b><br>
To load not the entire sequence but only a certain
sequence region a residue subset expression  can be written at the end of  the protein name.
The expression must be preceded by an exclamation mark.
For example if a PDB  file with the name "1ITB.pdb!1:B-120:B" is loaded, the protein object will
contain only residues 1 to 120 of chain B.

<br><br>
At th UNIX command the line the expression must be quoted. Example:
<pre class="terminal">
 java -jar strap.jar  '1ITB_B.pdb!1:B-120:B'
</pre>

<i>SEE_DIALOG:DialogFetchSRS</i>
<i>SEE_DIALOG:DialogFetchPdb</i>
   @author Christoph Gille
*/
public class DialogProteinList extends AbstractDialogJPanel implements ActionListener {
    private final ChFileChooser fs;
    private final ChTextArea ta=new ChTextArea("");
    {pcp(KEY_IF_EMPTY,"Enter names of protein files.\nTabulator for word expansion.\nSee docu for how to specify PDB  chains.",ta);}
    public DialogProteinList() {
        fs=new ChFileChooser(getClass()).setVisibleFalseParentFrame();
        addActLi(this,fs);
        fs.filter().acceptSuffix(true,".list",".text",".txt");
        fs.setMultiSelectionEnabled(false);
        final JComponent
            pSouth=pnl(new ChButton("L").t("Load protein files").li(this) ,ChButton.actionViewMessage("From hard disk",fs)),
            pTop=align.headLine(DialogProteinList.this);
        cnsew(this,ta.tools().enableUndo(true).enableWordCompletion(new Object[]{dirWorking(),align.vProteins}).scrollPane(),pTop,pSouth,null,null);
    }
    public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        if (cmd==JFileChooser.APPROVE_SELECTION)  ta.t(readStrg(fs.getSelectedFile()));
        if (cmd=="L") {
            String txt=ta.getSelectedText();
            if (txt==null) txt=ta.getText();
            align.loadTheProteinsInList(txt,StrapAlign.OPTION_PROCESS_LOADED_PROTS|StrapAlign.OPTION_EVENT_PROTEINS_ADDED);
        }

    }
}
