package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.ChProcessEvent;
import charite.christo.protein.Protein;
import charite.christo.protein.DNA_Util;
import javax.swing.*;
import java.awt.AWTEvent;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP

This dialog allows to create residue annotations from  lists of the  form
<pre class="data">
E815R  G4711A  C4444Q
</pre>
or
<pre class="data">
Glu815Arg  Gly4711Ala  Cys4444Gln
</pre>

These items may be mutations or SNPs or residue lists from crystallographic papers.

Collections of SNPs and mutations can be found in mutation
and SNP databases in the web for many diseases and proteins.

Each item should contain a number which is interpreted as a sequence position or a PDB-residue number.

For example, for the entry <I>E815R</I> the residue position 815 will
be assumed and a selection at amino acid number 815 will be created.
<br>
<br>
<u>Related resources</u> http://www.mutdb.org http://gila.bioengr.uic.edu/snp/toposnp/
<br>
<i>SEE_CLASS:ResidueAnnotation</i>
<i>SEE_CLASS:AddAnnotation</i>
<i>SEE_DIALOG:DialogResidueAnnotationChanges</i>

   @author Christoph Gille
*/
public class DialogResidueAnnotationList extends JPanel implements ChProcessEvent {
    public final static String
        NT_INDEX="Indices of nucleotide",
        AA_INDEX="Indices of amino acids";
    private final static String
        RES_NUMBER="Residue numbers as recorded in the pdb file";
    private final ChEventAdapter LI=new ChEventAdapter(this);
    private final JComboBox combo_AA_NT=new JComboBox(new String[]{AA_INDEX,NT_INDEX,RES_NUMBER});
    private final ProteinCombo comboProtein;
    private final ChTextField
        tfOffset=new ChTextField("    0").cols(4,true,true),
        tfGroupOfAnnotations=new ChTextField("mutations").cols(30,false,true);
    public DialogResidueAnnotationList(StrapAlign align) {
        JComponent
            pSouth=pnl("#",
                       new ChButton("E",ChButton.NO_MARGIN).t("For newbies load example").li(LI),
                      "  ",
                       new ChButton("GO",ChButton.NO_MARGIN).t(ChButton.GO).li(LI).tt("Creates a selection for each token in the text area"),
                      "#"
                      ),
            pTop=align.headLine(this),
            pNorth=vBhB(
                         pTop,
                         hB("Protein:  ",comboProtein=new ProteinCombo(align,0),"  ",butColor),
                         combo_AA_NT,
                         hB("Index offset:  ",tfOffset," (Negative values shift the residue selections to the left, positive values to the right.)","#"),
                         hB("Group:  ",tfGroupOfAnnotations,"#")
                         );
        final ChTextComponents tools=taList.tools();
        tools.enableUndo(true).enableWordCompletion(align.vProteins);
        cnsew(this,tools.scrollPane(),pNorth,pSouth,null,null);
    }
    private final ChTextArea taList=new ChTextArea("\n\n\n");
    {pcp(KEY_IF_EMPTY,"Enter a list of mutations such as E33A Lys144Glu ...",taList);}
    public AWTEvent processEvent(JComponent component, AWTEvent ev) {
        StrapAlign.errorMsg("");
        final String cmd=actionCommand(ev);
        if (cmd=="E") taList.setText("E33Q, L44I, M55C \n\n\n\n"+taList);
        if (cmd=="GO") {
            final StrapProtein protein=comboProtein.getStrapProtein();
            if (protein==null) { StrapAlign.errorMsg("No protein selected"); return ev;}
            final String iType=(String)combo_AA_NT.getSelectedItem();
            if (iType==RES_NUMBER && protein.getResidueNumber()==null) {
                StrapAlign.errorMsg("No PDB residue numbers found.");
                return ev;
            }
            if (iType==NT_INDEX && protein.getNucleotides()==null) {
                StrapAlign.errorMsg("No nucleotides found");
                return ev;
            }
            final String[] tokens=splitTokns(0L, taList);
            if (sze(tokens)==0) { StrapAlign.errorMsg("Enter some tokens like E33Q or L44I");  return ev;  }
            final UniqueList<ResidueAnnotation> vNewResidueAnnotations=new UniqueList(ResidueAnnotation.class);
            for(String t:tokens) {
                final ResidueAnnotation s=makeResidueAnnotation(t,protein,iType);
                if (s!=null) {
                    s.setColor(butColor.getColor());
                    vNewResidueAnnotations.add(s);
                }
            }
            final ResidueAnnotation aa[]=ResidueSelectionUtils.mayBeAddResidueAnnotations(vNewResidueAnnotations.asArray(),DialogResidueAnnotationList.this);
            final StringBuffer sb=new StringBuffer();
            doesTheNameAndResidueTypeMatch(aa,sb);
            final JComponent pSouth=newCNSEW("Note: this list has context menus",null,null,sze(sb)>0?ChButton.actionViewMessage("Warnings",sb):null,null);
            final String title="Created annotations";
            StrapAlign.showInJList(0L,null, aa, title, pSouth).showInFrame(ChFrame.AT_CLICK, title);
        }
        return ev;
    }
    // ----------------------------------------
    private ResidueAnnotation makeResidueAnnotation(String txt,StrapProtein p,String iType) {
        if (txt==null || p==null) return null;
        ResidueAnnotation s=null;
        final int dgt=nxt(DIGT,txt);
        if (dgt>=0) {
            final String idx=String.valueOf(atoi(txt,dgt)+atoi(tfOffset));
            s=new ResidueAnnotation(p);
            s.add(ResidueAnnotation.NAME,txt,true);
            s.add(ResidueAnnotation.POS, iType==RES_NUMBER ? idx+":":idx,  combo_AA_NT.getSelectedItem()!=NT_INDEX);
        }
        if (s==null) {
            final int open=txt.indexOf('['), slash=txt.indexOf('/'), close=txt.indexOf(']');
            if (open>0 && open+2==slash && slash<close) {
                final byte[] seq=(txt.substring(0,open)+txt.charAt(open+1)+txt.substring(close+1)).getBytes();
                final int iNT=strstr(STRSTR_IC,seq,p.getNucleotides());
                final int iTrans=iNT<0 ? strstr(STRSTR_IC,seq,DNA_Util.onlyTranslated(p.getNucleotidesCurrentStrand(),p.areNucleotidesCurrentStrandTranslated())) : -1;
                final int iAa=strstr(STRSTR_IC,seq,p.getResidueTypeExactLength());
                if (iNT>=0||iTrans>=0||iAa>=0) {
                    s=new ResidueAnnotation(p);
                    final int pos=iNT>=0 ? iNT : iTrans>=0 ? iTrans/3 : iAa;
                    s.add(ResidueAnnotation.NAME,(pos+1)+txt.substring(open+1,close),true);
                    s.add(ResidueAnnotation.POS,String.valueOf(pos+open+1), iNT<0);
                    s.add("Note",txt,true);
                }
            }
        }
        if (s!=null) {
            s.add(ResidueAnnotation.GROUP,tfGroupOfAnnotations.getText(),true);
            s.add(CLASS_Texshade,"\\feature{bottom}{$TEX_PROTEIN}{$TEX_RESIDUES}{fill:$\\uparrow$}{"+ResidueAnnotation.SORE_NAME+"}",true);
            s.add(CLASS_Pymol,"hide everything, RESIDUES; show spheres,RESIDUES",true);
            if (p.getResidueCalpha()!=null) {
                if (!isMacintosh() && !isWindows()) {
                    s.add(CLASS_Rasmol,"spacefill on",true);
                    s.add(CLASS_Rasmol,"ATOM_TYPE_CA LABEL %n",true);
                }
            }
            if (iType==NT_INDEX) s.setEnabled(false,ResidueAnnotation.POS);
        }
        return s;
    }
    // --------------------------------------

    private final ButColor butColor=new ButColor(0L,C(0xffff00),null);
    private void doesTheNameAndResidueTypeMatch(ResidueAnnotation[] vAdded,StringBuffer sb) {
        boolean isFirst=true;
        final String msg="There are possibly discrepancies between the residues in the sequence and the designations of the mutations.\nPlease check:\n";
        for(ResidueAnnotation s : vAdded) {
            final String name=s.getName();
            final Protein p=s.getProtein();
            final int iDigit=nxt(DIGT,name);
            if (NT_INDEX!=combo_AA_NT.getSelectedItem()) {
                final int iAa=firstTrue(s.getSelectedAminoacids())+s.getSelectedAminoacidsOffset();
                int aa=p.getResidueType(iAa);
                if (iDigit==1 && aa>0) {
                    aa&=~32;
                    final int c0=name.charAt(0)&~32;
                    if (aa!=c0) {
                        if (isFirst) sb.append(msg);
                        isFirst=false;
                        sb.append(name).append(": residue at ").append(1+iAa).append(" is ").append((char)aa).append('\n');
                    }
                }
                if (iDigit==3) {
                    int c=Protein.toOneLetterCode(name.charAt(0),name.charAt(1),name.charAt(2));
                    if (c!=0 && c!='X' && c!=aa) {
                        if (isFirst) sb.append(msg);
                        isFirst=false;
                        sb.append(name).append(": residue at ").append(iAa).append(" is ").append((char)aa).append(' ').append(Protein.toThreeLetterCode(aa)).append('\n');
                    }
                }
            } else {
                final int iNt=firstTrue(s.getSelectedNucleotides())+s.getSelectedNucleotidesOffset();
                int nt=((StrapProtein)p).getNucleotide(iNt);
                if (nt>0 && iDigit==1) {
                    nt&=~32;
                    final int c0=name.charAt(0)&~32;
                    if (c0!=nt) {
                        if (isFirst) sb.append(msg);
                        isFirst=false;
                        sb.append(name).append(": nucleotide at ").append(1+iNt).append(" is ").append((char)nt).append('\n');
                    }
                }
            }
        }
    }

}
