package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.io.File;
import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.*;
import java.util.Arrays;
import java.util.Collections;
import java.util.Vector;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;

/**HELP

A backup of the multiple sequence alignment data can be written to the
directory "./backup/.

The user can chose a backup-directory to restore data.

The files in the backup-directory contain the gaps, the annotations to
amino acid and nucleotide positions, the nucleotide frame-shift and
the 3D-rotation and 3D-translation.

The saved information is loaded read by pressing <i>LABEL:ChButton#BUTTON_GO</i>.

Usually, all alignment data is loaded but the user may deselect the check-boxes:

<i>CB:DialogRestoreFromBackup#CB_LAB_Gaps</i>,
<i>CB:DialogRestoreFromBackup#CB_LAB_Sores</i>,
<i>CB:DialogRestoreFromBackup#CB_LAB_Frameshift</i>
and

<i>CB:DialogRestoreFromBackup#CB_LAB_Transformation3D</i>.

<br><br><b>Warning:</b> by restoring the multiple sequence alignment data the actual  data is overwritten.
   @author Christoph Gille
*/
public class DialogRestoreFromBackup extends AbstractDialogJPanel implements StrapListener,ActionListener {
    public final static String
        CB_LAB_Frameshift="restore nt-<br>frame shift",
        CB_LAB_Transformation3D="restore 3D-<br>transformation",
        CB_LAB_Sores="restore resi-<br>due annotations",
        CB_LAB_Gaps="restore align<br>ment gaps";
     private static ChButton
         togFrameshift=toggl(CB_LAB_Frameshift).s(true),
         togTransformation3D=toggl(CB_LAB_Transformation3D).s(true),
         togSores=toggl(CB_LAB_Sores).s(true),
         togGaps=toggl(CB_LAB_Gaps).s(true);
    public void actionPerformed(ActionEvent ev) {
        StrapAlign.errorMsg("");
        final String cmd=ev.getActionCommand();
        if (cmd=="GO") {
            int i=jList.getSelectedIndex();
            if (i<0) { StrapAlign.errorMsg("You must select an archive"); return;}
            final File dir=file("backup/"+listing.get(i));
            if (dir==null) {
                StrapAlign.errorMsg("A backup folder must be selected");
                return;
            }
            StrapProtein pp[]=align.getSelectedProteins();
            if (pp==null||pp.length<1)
                if (ChMsg.yesNo("No protein is selected. Do you want to restore all proteins?"))
                    pp=align.getProteins(); else return;
            for(StrapProtein p:pp){
                String what="";
                if (togFrameshift.isSelected()) what+="N";
                if (togGaps.isSelected()) what+="G";
                if (togTransformation3D.isSelected()) what+="3";
                if (togSores.isSelected()) what+="A";
                StrapAlign.readAttributes(what,p,dir);
            }
            align.setChanged();
            StrapAlign.dispatchLater(StrapEvent.NUCL_TRANSLATION_CHANGED,111);
            StrapAlign.dispatchLater(StrapEvent.ALIGNMENT_CHANGED,111);
            align.gap2col().setInvalid();
        }
    }
    public DialogRestoreFromBackup() {
        setSettings(this, Customize.fileBrowsers);
        Container pNorth=vBpnl(pTop(),"Select the proteins to be restored and a backup directory<br>The date format is Year_Month_Day_Hour_Minute");
        cnsew(this,new JScrollPane(jList),pNorth,pnl("Restore alignment data of selected proteins", new ChButton("GO").like(ChButton.BUTTON_GO).li(this),fDirectory()),
              vB(togGaps.cb(),togSores.cb(),togTransformation3D.cb(),togFrameshift.cb()),null);
        myUpdate();
    }
    public void handleEvent(StrapEvent ev) {
        if (ev.getType()==StrapEvent.BACKUP_WRITTEN) myUpdate();
    }
    private final Vector<String> listing=new Vector();
    private final JList jList=new JList(listing);
    private File fDirectory_;
    private File fDirectory(){
        if (fDirectory_==null) fDirectory_=file("backup");
        mkdrs(fDirectory_);
        return fDirectory_;
    }
    private void myUpdate() {
        listing.clear();
        listing.addAll(Arrays.asList(fDirectory().list()));
        sortArry(listing);
        ChDelay.afterMS(EDT,ChDelay.CONTENTS_CHANGED,jList,111,null);
    }
}
