package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import charite.christo.strap.interfaces.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.lang.ref.*;
import static java.lang.System.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
   @author Christoph Gille
*/
public class DialogSelectionOfResidues extends AbstractDialogJPanel implements ActionListener,IsEnabled {
    private final Object classSor;
    private final ResidueSelection sharedInstance;
    private final ButColor bColor;
    private final ChButton
        cb1D=toggl("1d",ChButton.SHOW_HIDE).li(this).s(true).tt("mark selected residues in alignment").s(true),
        cb3D=toggl("3d",ChButton.SHOW_HIDE).li(this).s(true).tt("mark selected residues in 3d-wire representation"),
        cbSP=toggl("view port",ChButton.SHOW_HIDE).li(this).s(true).tt("mark selected residues in horizontal scroll-bar").s(true);
    private final ChCombo comboStyle;
    private final ProteinList proteinList;

    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        boolean dispatch=false;
        if (q==proteinList) {
            myUpdate();
            dispatch=true;
        }
        if (q==cb1D || q==cb3D || q==cbSP || q==bColor || q==comboStyle)  {
            if (q==comboStyle || q==bColor) cb1D.s(true);
            myUpdate();
            dispatch=true;
        }
        if (dispatch) StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
        if (cmd=="ALL_HETEROS") {
            final StrapProtein pp[]= proteinList().getSelectedProteins();
            if (sze(pp)==0) StrapAlign.errorMsg("No protein selected in List");
            else StrapProteinUtils.showChildObjects(HeteroCompound.class,pp);
        }
    }
    // ---
      protected ProteinList proteinList() { return proteinList;}

    public boolean isEnabled(Object p) {
        return sharedInstance instanceof IsEnabled ? ((IsEnabled)sharedInstance).isEnabled(p) : true;
    }
    public DialogSelectionOfResidues(ResidueSelection sharedInstance, Object classSor) {
        this.sharedInstance=sharedInstance;
        this.classSor=classSor;
        pcp(KEY_ENABLED, weakRef(this),  (proteinList=new ProteinList(align,0)).li(this));
        bColor=new ButColor(0L,Color.PINK,this);
        final Container
            ctrlPanel=getSharedCtrlPan(sharedInstance, true),
            pSouth=pnl(
                       cb1D,cb3D,cbSP, bColor.options(ChButton.NO_FILL|ChButton.NO_BORDER|ChButton.ICON_SIZE),
                       comboStyle=AbstractVisibleIn123.newComboStyle().s(VisibleIn123.CIRCLE).li(this)
                       ),
            pSouthAll=sharedInstance instanceof charite.christo.strap.extensions.Distance3DToHetero ?
            vB(pSouth,pnl("To obtain a list for all hetero groups press", new ChButton("ALL_HETEROS").li(this).t("list all heteros"))) : pSouth,
            pEast=ChButton.actionClose(this),
            pNorth=newCNSEW(shrtClasNam(sharedInstance),null,null,pEast,"Using the class "),
            pCenter=ctrlPanel!=null ? pnl(new GridLayout(1,2), scrllpn(0L,proteinList), scrllpn(0L,ctrlPanel)) : scrllpn(0L,proteinList);
        cnsew(this,pCenter,pNorth,pSouthAll,null,null);
        pcp(SCROLLABLE_TRACKS_VIEWPORT_WIDTH,"",ctrlPanel);
        if (isAssignblFrm(StrapListener.class,sharedInstance)) align.addListener((StrapListener)sharedInstance);
    }
    // ----------------------------------------
    private final UniqueList<AbstractResidueSelection> vSelectors=new UniqueList(AbstractResidueSelection.class);
    private void myUpdate() {
        removeAllResidueSelections();
        for(StrapProtein p:proteinList.getSelectedProteins()){
            if (p==null || !isEnabled(p)) continue;
            final ResidueSelection s=HotswapUtil.instance(classSor,ResidueSelection.class,true);
            if (s==null) continue;
            setSharedParas(this.sharedInstance,s);
            s.setProtein(p);
            final AbstractResidueSelection sel=new AbstractResidueSelection(s);
            vSelectors.add(sel);
            p.addResidueSelection(sel);
        }
        final int m=(cb1D.s() ? VisibleIn123.SEQUENCE : 0) +  (cb3D.s() ? VisibleIn123.STRUCTURE : 0) +   (cbSP.s() ? VisibleIn123.SB : 0);
        final int style=comboStyle.i();
        for(AbstractResidueSelection s : vSelectors.asArray()) {
            s.setVisibleWhere(m);
            s.setColor(bColor.getColor());
            s.setStyle(style);
        }
    }
    private void removeAllResidueSelections() {
        for(AbstractResidueSelection sel : vSelectors.asArray()) {
            vSelectors.remove(sel);
            final Protein p=sel.getProtein();
            if (p!=null) p.removeResidueSelection(sel);
        }
    }
    @Override public final void dispose() {
        removeAllResidueSelections();
        align.removeListener(sharedInstance);
        StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
    }

}
