package charite.christo.strap;
import charite.christo.interfaces.*;
import charite.christo.*;
import charite.christo.strap.interfaces.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import static java.lang.System.*;
import static charite.christo.ChUtils.*;
/**HELP
   Proteins appear in the order they are loaded from HD.
   This dialog reorders proteins so that similar proteins are in adjacent rows.
   Recommended methods: SequenceDisSimilarityAsAligned or AlignmentScore.

   @author Christoph Gille
*/
public class DialogSortProteins extends AbstractDialogJPanel implements ActionListener,ChRunnable {
    public DialogSortProteins() {
        add(vBpnl(
                  pTop(),
                  pnl(
                      "#TBem The selected proteins are reordered so that similar proteins come close together",
                      comboClass(0).panel(),
                      new ChButton("GO0").li(this).t(ChButton.GO)
                      ),
                  pnl(
                      "#TBem Other criteria",
                      comboClass(1).panel(),
                      new ChButton("GO1").li(this).t(ChButton.GO)
                      )
                  ));
    }
    public void setClass(Object cl) { comboClass(1).s(cl);}
    private final ChCombo[] comboClass=new ChCombo[2];
    private ChCombo comboClass(int n) {
        ChCombo c=comboClass[n];
        if (c==null) {
            c=comboClass[n]=StrapAlign.classChoice(n==0 ? ProteinProteinValue.class : ProteinsSorter.class);
            c.setMaximumRowCount(30);
            if (n==1) pcp(ChConstants.KEY_CHOICE_CLASS,c,this);
        }
        return c;
    }

    public Object run(String id, Object arg) {
        final StrapProtein pp[]=arg instanceof StrapProtein[] ? (StrapProtein[])arg : null;
        if (id=="GO0") {
            align.inferOrder(sortProteins(pp,comboClass(0)));
        }
        if (id=="GO1") {
            final ProteinsSorter so=HotswapUtil.instance(comboClass(1),ProteinsSorter.class,true);
            align.inferOrder(so.sortProteins(pp));
        }
        return null;
    }

    public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        StrapProtein pp[]=align.getSelectedProteinsInVisibleOrder(),ppVis[]=align.getVisibleProteins();
        if (pp.length<4 && ChMsg.yesNo("Sort all proteins that are in the alignment?")) pp=ppVis;
        if (pp.length>4) {
            if (cmd=="GO0" || cmd=="GO1") {
                ChDelay.newThread(this,cmd,pp).start();
            }
        } else if (pp.length>0) error("Too few proteins for sorting");
    }

    private  StrapProtein[] sortProteins(StrapProtein proteins[], Object classPPD) {
        final int nP=proteins.length;
        if (nP==0) return new StrapProtein[0];
        final StrapProtein pRemaining[]=proteins.clone();
        int nSorted=0;
        StrapProtein pSorted[]=new StrapProtein[nP];
        final double all_vs_all[][]=new double[nP][nP];
        final boolean isDistance=isAssignblFrm(ProteinProteinDistance.class,classPPD);
        for(int i=0;i<nP;i++) {
            for(int j=0;j<nP;j++) {
                final ProteinProteinValue ppd=HotswapUtil.instance(classPPD,ProteinProteinValue.class,true);
                if (ppd==null) { putln("Error: failed to instantiate "+classPPD); all_vs_all[i][j]=Double.NaN; continue; }
                ppd.setProtein((StrapProtein)proteins[i]);
                ppd.setOtherProtein((StrapProtein)proteins[j]);
                if (DialogSortProteins.this.isDisposed()) break;
                ppd.compute();
                all_vs_all[i][j]=ppd.getValue();
            }
        }
        int iPrev=getRepresentative(all_vs_all,isDistance);
        pSorted[nSorted++]=proteins[iPrev];
        pRemaining[iPrev]=null;

        while(true) {
            final int iClose= findClosest(iPrev, pRemaining,all_vs_all,isDistance);
            if (iClose<0) break;
            pRemaining[iClose]=null;
            pSorted[nSorted++]=proteins[iClose];
            iPrev=iClose;
        }
        return pSorted;
    }
    private static int findClosest(int i0,StrapProtein pRemaining[],double[][] mtx, boolean isDistance) {
        double min=isDistance ? Double.MAX_VALUE : Double.MIN_VALUE;
        int iBest=-1;
        for(int i=0; i<pRemaining.length; i++) {
            if (i==i0 || pRemaining[i]==null) continue;
            double v=mtx[i0][i];

            if (isDistance ? v<min : v>min ) {iBest=i; min=v;}
        }
        return iBest;
    }
    private static int getRepresentative(double[][] mtx,boolean isDistance) {
        final int n=mtx.length;
        double minSum=isDistance ? Double.MAX_VALUE : Double.MIN_VALUE;
        int idx=0;
        for(int i=0;i<n;i++) {
            double sum=0;
            for(int j=0;j<n;j++) sum+=mtx[i][j]*mtx[i][j];
            if (isDistance ? sum<minSum : sum>minSum) { minSum=sum;idx=i;}
        }
        return idx;
    }

}
