package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.awt.*;
import java.util.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP

In <i>DIALOG:DialogSelectionOfResiduesMain</i> the method   <i>COMBO:Distance3DToCursor</i> is selected.
<br>
The Euclidean distances between the cursor and  all C-alpha atom coordinates are permanently monitored
when the cursor is moved.

This way amino acids equivalent to the amino acid at the cursor position  show up in  all proteins.

Before this can be allied the proteins must be oriented in 3D-space in the same way: <i>DIALOG:DialogSuperimpose3D</i>.
<br>
<i>SEE_DIALOG:DialogSelectionOfResiduesMain</i>
<i>SEE_DIALOG:DialogSuperimpose3D</i>
   @author Christoph Gille
*/
public class Distance3DToCursor extends AbstractResidueSelection implements StrapListener,IsEnabled, HasSharedControlPanel,HasImage, javax.swing.event.ChangeListener {
    {
        setImage(IC_MEASURE);
        setStyle(VisibleIn123.IMAGE);
    }
    static {
        TabItemTipIcon.set(null,null,"distance to the c-alpha atom of the amino acid under the cursor",IC_MEASURE,Distance3DToCursor.class);
    }
    @Override public String toString() { return "3D-Distance to Cursor"; }

    public boolean isEnabled(Object o){ return o instanceof Protein &&  ((Protein)o).getResidueCalpha()!=null; }
    private static Container ctrlPanel;
    public Container getSharedControlPanel() {
        if (ctrlPanel!=null) return ctrlPanel;
        JComponent
            panNorth=pnl( "underline residues closer than ",labMax,"and more distant than",labMin,"to cursor"),
            pan=vB(minSlider,maxSlider,panNorth);
        return ctrlPanel=pan;
    }
    // ----------------------------------------
    private final static int MAX_VALUE=800;
    private static int maxValue=MAX_VALUE,minValue=0;
    private static JSlider maxSlider,minSlider;
    private JSlider slider()  {
        JSlider sl=new JSlider(0,1600,0);
        sl.setMajorTickSpacing(200);
        sl.setMajorTickSpacing(100);
        sl.setPaintTicks(true);
        sl.addChangeListener(this);
        return sl;
    }
    // ----------------------------------------
    private static ChTextField labMin,labMax;
    private static CharSequence set(ChTextField lab,long l) {
        final CharSequence s=ftoa(l/100f,5,3,null).append("\u00c5 ");
        lab.t(s);
        return s;
    }
    // ----------------------------------------
    private static float calpha[];
    private float lastCalpha[];
    private void calpha() {
        StrapView sp=StrapAlign.getInstance().getAlignmentPanelWithFocus();
        if (sp==null) return;
        StrapProtein pCursor=sp.getProteinAtCursor();
        if (pCursor==null) return;
        this.calpha=pCursor.getResidueCalpha(sp.getIndexOfAminoAcidAtCursor(),new float[3]);
    }
    // ----------------------------------------
    public final void handleEvent(StrapEvent ev) {
        final int t=ev.getType();
        if (t==StrapEvent.CURSOR_MOVED_WITHIN_PROTEIN || t==StrapEvent.CURSOR_CHANGED_PROTEIN || (t&StrapEvent.FLAG_ATOM_COORDINATES_CHANGED)!=0) {
            calpha();
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
        }
    }
    // ========================================
    public Distance3DToCursor() {
        if (maxSlider==null) {
            (labMin=new ChTextField().cols(8,true,true)).setBackground(C(DEFAULT_BACKGROUND));
            (labMax=new ChTextField().cols(8,true,true)).setBackground(C(DEFAULT_BACKGROUND));
            monospc(labMax);
            monospc(labMin);
            set(labMin,0);
            set(labMax,MAX_VALUE);

            maxSlider= slider();
            minSlider= slider();
            maxSlider.setValue(MAX_VALUE);
        }
    }
    private float lastMax,lastMin;
    private long lastVersionCalpha;
    private boolean lastSelected[];
    @Override public boolean[] getSelectedAminoacids() {
        final Protein protein=getProtein();
        if (this.calpha==null) calpha();
        final float[] calpha=this.calpha;
        if (calpha==null) return null;
        final float max=maxSlider.getValue()*0.01f, min=minSlider.getValue()*0.01f;
        if (lastVersionCalpha==protein.getModificationCount(ProteinModiCount.MC_ATOM_COORD) && lastCalpha==calpha && lastMin==min && lastMax==max) {
            return lastSelected;
        }

        final float calphas[]=protein.getResidueCalpha();
        if (calphas==null) return null;
        final int ca3=calphas.length/3, nR=protein.countResidues(), n=ca3<nR ? ca3:nR;
        final float x=calpha[0],y=calpha[1],z=calpha[2];
        final boolean selected[];
        if (this.lastSelected==null || this.lastSelected.length<n) selected=new boolean[n]; else Arrays.fill( selected=lastSelected,false);
        final float min2=min*min,max2=max*max;
        for(int i=0,i3=0; i<n; i++,i3+=3) {
            final float xx=x-calphas[i3],yy=y-calphas[i3+1],zz=z-calphas[i3+2],dd=xx*xx+yy*yy+zz*zz;
            selected[i]=dd<max2 && dd>min2;
        }
        lastVersionCalpha=protein.getModificationCount(ProteinModiCount.MC_ATOM_COORD);
        lastCalpha=calpha;
        lastMax=min;lastMax=max;
            Protein.incrementModificationCount(ProteinModiCount.MC_RESIDUE_SELECTIONS, protein);
        return lastSelected=selected;
    }
    public void stateChanged(javax.swing.event.ChangeEvent ev) {
        final Object q=ev.getSource();
        if (q==maxSlider) {
            maxValue=maxSlider.getValue();
            if(maxValue<minValue) {minSlider.setValue(minValue=maxValue);minSlider.repaint();}
            setName("distance of C-alpha Atom to cursor "+labMin+"<distance<"+set(labMax,maxValue));
        }
        if (q==minSlider) {
            minValue=minSlider.getValue();
            if(minValue>maxValue) {maxSlider.setValue(maxValue=minValue);maxSlider.repaint();}
            set(labMin,minValue);
        }
         if (q==minSlider || q==maxSlider) StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
    }
}
