package charite.christo.strap;
import charite.christo.*;
import charite.christo.protein.*;
import java.awt.*;
import static java.awt.Color.*;
import static charite.christo.ChUtils.*;
// see http://www-ab.informatik.uni-tuebingepann.de/software/welcome_en.html MACAW
/**
   A gapped sequence with colorshading is drawn.
   Not used for the main alignment view in Strap.
   It is used for the Dialoges {@link DialogAlign DialogAlign} and {@link DialogAlignOneToAll DialogAlignOneToAll}.
   @author Christoph Gille
*/
public class DrawGappedSequence  {
 public final static int SHADE_SEC_STRU=1,AA2NT=2,NUCLEOTIDES=4;
    private final static Color
        secStr2color_bright[]=ShadingAA.SECSTRU,
        secStr2color_darker[]=darker(secStr2color_bright),
        COLOR_NOT_CONSERVED=new Color(190,190,190);

    private Rectangle charBounds;
    private Font lastFont;
    private short frequenciesOfAA[][];

    private static DrawGappedSequence inst;
    public static DrawGappedSequence instance() {
        if (inst==null) inst=new DrawGappedSequence();
        return inst;
    }

/**
       Draws a gapped sequence into the graphics context.
       @param component  any component and is used for the management of images  of
       selections (ImageObserver).  May be null.
       @param origin point where to start to draw,
       @param shadeAminos An array of Color of length 256. E.g. Alanin will be
       drawn in the color shadeAminos['A'],  May be null.
*/

    public void draw(Component component,Graphics g,Point origin, StrapProtein p, Color[] shadeAminos, long mode) {
        draw(component,g,origin, 0, p.getGappedSequence(),0,p.getMaxColumn(),p,shadeAminos,mode);
    }

    public void draw(Component component, Graphics g_,Point origin, int idxOfFirstAA, byte[] seq, int fromCol, int toCol, StrapProtein p, Color[] shadeAminos, long mode) {
        final Graphics2D g=(Graphics2D)g_;
        if (seq==null && p==null) return;
        final byte[] secStru=(mode&SHADE_SEC_STRU)!=0 && p!=null ? p.getResidueSecStrType() : null;
        final Font font=g.getFont();
        if (charBounds==null || lastFont!=font) { lastFont=font;charBounds=charBounds(font);}
        final int charA=charBounds.y, charW=charBounds.width,  charH=charBounds.height;
        if (charW==0) return;
        final int n=seq.length;
        final Rectangle clip=g.getClipBounds();
        int colStart=0,colEnd=n;
        if (clip!=null) {
            if (clip.y>origin.y+charH || clip.y+clip.height<origin.y) return;
            colStart=(clip.x-origin.x)/charW;
            colEnd=mini((clip.x+clip.width-origin.x)/charW+1,n);
        }
        colEnd=mini(toCol, colEnd,seq.length);
        if (colStart<fromCol) colStart=fromCol;
        final Color colorFG=isWhiteBG() ? BLACK : WHITE;
        final ResidueSelection residueSelections[]=p!=null ? p.getAllResidueSelections() : null;
        final Rectangle RECT=new Rectangle(0,0,charW,charH);
        final byte[] selectedAA=p!=null ? p.getSelectedAminoacids() : NO_BYTE;
        final String strgLen1[]=getStringsLenght1();
        for(int iCol=0,iA=idxOfFirstAA;iCol<colEnd;iCol++) {
            final int iAmino=(mode&AA2NT)!=0 ? iA/3: (mode&NUCLEOTIDES)!=0 ? p.nucleotideIdx2translatedNucleotideIdx(iA)/3 : iA;
            final byte c=seq[iCol];
            final int lower=c|32;
            if ('a'>lower || lower>'z') continue;
            //if (idxOfFirstAA>0 && iCol<20) putln("iCol="+iCol+" iAmino="+iAmino+"  "+(char)secStru[iAmino]);
            //if (idxOfFirstAA>0) putln("iCol="+iCol+" iAmino="+iAmino);
            if (iCol>=colStart) {
                final int cLower=(c|32)-'a';
                final boolean isConserved=
                    frequenciesOfAA!=null && frequenciesOfAA.length>iCol &&
                    0<=cLower && cLower<='z'-'a' && frequenciesOfAA[iCol][cLower]>minFreq,
                    isCons=frequenciesOfAA==null || isConserved;
                final Color color=secStru!=null && secStru.length>iAmino ?
                    (isCons ? secStr2color_bright : secStr2color_darker)[secStru[iAmino]] :
                    (isCons ? shadeAminos==null ? WHITE : shadeAminos[lower] : COLOR_NOT_CONSERVED);
                RECT.x=origin.x+charW*iCol;
                RECT.y=origin.y;
                if (0<iAmino && iAmino<selectedAA.length && selectedAA[iAmino]!=0) {
                    ResidueSelectionUtils.drawResidueSelection(ProteinNT.AMINO_ACIDS,c,p,residueSelections,iAmino,g,RECT,charBounds,component);
                }
                g.setColor(WHITE==color || color==null ? colorFG : color);
                g.drawString(strgLen1[c&127],RECT.x,RECT.y+charA);
            }
            iA++;
        }
    }
    private static int minFreq=0;
    public void setConservation(short frequenciesOfAminos[][], int minFrequency) {
        frequenciesOfAA=frequenciesOfAminos;
        minFreq=minFrequency;
    }
   public void drawBlastMidline(byte b1[],byte b2[], int colFrom, int colTo, byte[][] matrix, Graphics g,int x0, int y0) {
        final byte[][] M=matrix!=null ? matrix : charite.christo.protein.Blosum.BL2SEQ_MIDLINE;
        final Rectangle cb=charBounds(g.getFont());
        final int charW=cb.width;
        final String strgLen1[]=getStringsLenght1();
        final int y=y0+cb.y, to=mini(colTo, b1.length, b2.length);
        for(int col=colFrom,x=x0+colFrom*charW; col<to; col++,x+=charW) {
            final byte mid=M[b1[col]&127][b2[col]&127];
            if (mid!=' ') g.drawString(strgLen1[mid],x,y);
        }
    }
}
