package charite.christo.strap;
import charite.christo.*;
import charite.christo.protein.*;
import charite.christo.interfaces.*;
import charite.christo.strap.interfaces.*;
import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.lang.ref.*;
import java.util.*;
import static java.lang.System.*;
import static charite.christo.ChUtils.*;
/**HELP

This Dialog allows to translate coding nucleotide sequences into amino acid sequences.
Three nucleotides are called a triplet and are translated into one amino acid (WIKI:Genetic_code).
The user specifies a reading direction and can select translated and untranslated sequence regions.
Those parts of the nucleotide sequence that are translated into amino acids are shown in black (WIKI:Coding_sequence)
and untranslated nucleotides are shown in gray (WIKI:Introns,  WIKI:Untranslated_region UTRs).

By pressing  INSERT, DELETE, SPACE-BAR and  BACKSPACE  translation of nucleotides is switched on and off.
The cursor keys, HOME, END, PAGE UP/DOWN  and the mouse allow navigation within the DNA sequence.
The SHIFT key in conjunction with these keys creates a block selection.
<br><br>
<b>ALT + Cursor left/right</b> jumps to the next interesting sequence position. These positions are intron/exon boundaries and
selected residues or selected amino acids. This is very useful when  navigation with the
scroll-bar in a <b>very long nucleotide sequence</b> is difficult.

<br><br>
Above the text field  rulers for the nucleotide positions and the translated amino acids are shown.
Residue selections  are marked in the rulers.

The horizontal scrollbar outlines the intron-exon structure of  DNA.

<br><br><b>Zoom:</b>
CTRL-<b>+</b> and CTRL+<b>-</b> or CTRL WHEEL

<br><br><b>Mouse actions:</b>
left-click  moves the alignment cursor to the clicked residue.

<i>SEE_DIALOG:DialogGenbank</i>
<i>SEE_DIALOG:DnaExonStart</i>
   @author Christoph Gille
*/
public class EditDnaParentPane extends AbstractDialogJPanel implements ActionListener {
    private final JComponent pEditors=vB(Box.createVerticalGlue(),new ChPanel()),pTables=vB();
    private final JSplitPane split;
    protected static Reference<EditDnaView> refFocused;

    protected void setFocused(EditDnaView e, boolean always) {
        if (deref(refFocused)==null ||  always) refFocused=weakRef(e);
    }
    public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        EditDnaView e=(EditDnaView)deref(refFocused);
        if (e==null) e=get(0,ee(),EditDnaView.class);
        final StrapProtein p=e!=null ? e.getProtein() :  null;
        putln("EditDnaParentPane e="+e+" p="+p," ee()=",ee());
        if (p!=null) {
            if (cmd=="ORIENTATION") setOrientation(p);
            if (cmd=="T"){
                pTables.add(new EditDnaTable(p));
                split.setDividerLocation(0.5);
                revalidate();
            }
        }
    }
    public EditDnaParentPane() {
        final JComponent
            b1=new ChButton("T",ChButton.NO_MARGIN).li(this).t("As table").tt("start and end of exons as a table"),
            b2=new ChButton("ORIENTATION",ChButton.NO_MARGIN).li(this).t("Set forward or reverse or complement"),
            pSouth=newCNSEW(pnl(b2),null,null,smallHelpBut(EditDnaParentPane.class),pnl(b1));
        split=new JSplitPane(JSplitPane.VERTICAL_SPLIT,pTables,pEditors);
        pSouth.setPreferredSize(b1.getPreferredSize());
        cnsew(this,split,null,pSouth,null,null);
    }

    // ----------------------------------------
    private EditDnaView[] ee() { return  childs(pEditors,EditDnaView.class);}
    public void addEditor(StrapProtein p,boolean newInstance) {
        if (p==null) return;
        if (!newInstance) {
            for(EditDnaView ed: ee()) {
                if (ed.getProtein()==p) return;
            }
        }
        if (p.getNucleotides()==null) p.setTranslatedStrand(ProteinNT.FORWARD);
        pEditors.add(new EditDnaView(p,this).getPanel());
        pEditors.revalidate();
    }
    // ----------------------------------------
    protected boolean setOrientation(StrapProtein p) {
        if (p==null) return false;
        final ChCombo combo=new ChCombo(ProteinNT.ORIENTATIONS);
        {
            final boolean c=p.isComplement(),r=p.isReverse();
            combo.s(
                    !r & !c ? ProteinNT.FORWARD:
                    !r &  c ? ProteinNT.FORWARD_COMPLEMENT:
                     r & !c ? ProteinNT.REVERSE :
                    ProteinNT.REVERSE_COMPLEMENT
                    );
        }
        final String
            TRANSLATE=addHtmlTagsAsStrg("Translate nuc-<br> leotide sequence"),
            NO_TRANSLATE=addHtmlTagsAsStrg("Dispose existing<br>translation"),
            GENBANK=addHtmlTagsAsStrg("Use Genbank/<br>Embl parser"),
            option[]= sze(p.getCDS())>0 ? new String[]{TRANSLATE,NO_TRANSLATE,GENBANK} : new String[]{TRANSLATE,NO_TRANSLATE},
            title="Reading frame of "+p;
        final int iChoice=JOptionPane.showOptionDialog(this,pnl("Chose translation direction",combo),title,JOptionPane.DEFAULT_OPTION,JOptionPane.QUESTION_MESSAGE,null,option,null);
        if (iChoice<0) return false;
        final String choice=option[iChoice];
        if (choice==GENBANK)  {
            DialogGenbank gb=(DialogGenbank)align.addDialog(DialogGenbank.class,false);
            gb.setProteins(p);
            gb.run();
        } else {
            p.setTranslatedStrand(choice==NO_TRANSLATE ? null:toStrg(combo));
            new StrapEvent(this,StrapEvent.NUCL_TRANSLATION_CHANGED).setParameter(p).run();
            addEditor(p,false);
        }
        return true;
    }
}
