package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import charite.christo.strap.interfaces.*;
import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.*;
import javax.swing.table.*;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static charite.christo.protein.ProteinModiCount.MC_ARE_NUCLEOTIDES_TRANSLATED;
/**
This contains many DNA views.
   In the gui "1"-based  counting.
   internally  "0"-based  counting.
   In gui start (inclusive) end(inclusive).
   Internally start (inclusive) end(exclusive).
   @author Christoph Gille
*/
public class EditDnaTable extends AbstractDialogJPanel implements ActionListener,TableModel,StrapListener,Runnable  {
    private int delay=333;
    public void run(){
        final long time=System.currentTimeMillis();
        setTextField();
        ChDelay.afterMS(EDT,ChDelay.CONTENTS_CHANGED,table,222,null);
        delay=3*(int)(System.currentTimeMillis()-time)+200;
    }
    private final StrapProtein protein;
    private final StrapAlign align;
    public StrapProtein  getProtein() { return protein;}
    private final ChJTable table;
    private ChLabel labProtName;
    private ChLabel labProtName() {
        if (labProtName==null) labProtName=new ChLabel();
        return labProtName.t(protein.getRendTxt()).i(protein.getIcon());
    }
    public EditDnaTable(StrapProtein prot) {
        protein=prot; align=StrapAlign.getInstance();
        align.addListener(this);
        table=new ChJTable(this,0);
        table.setColumnWidth(0,ICON_HEIGHT,false);
        //        table.getColumnModel().getColumn(0).setMaxWidth(ICON_HEIGHT);
        //        table.getColumnModel().getColumn(0).setWidth(ICON_HEIGHT);
        labProtName=new ChLabel(protein.toString());
        Container title=hB("Exons of ",labProtName(),"#",ChButton.actionClose(this, OPTION_DISPOSE_RM_CHILDS));
        cnsew(this, new JScrollPane(table),title,newCNSEW(tf,null,null,cbRevComplement,new ChButton("GO").t(ChButton.GO).li(this).tt("apply the exon-intron expression in the textfield")),null,null);
        tf.setToolTipText(addHtmlTagsAsStrg("an expression describing the intron-exon structure<br>e.g. <I>11..102,304..420</I> describes two exons"));
        setTextField();
        cbRevComplement.s(protein.isReverse());

    }
    // ----------------------------------------
    /* ---  TableModel --- */
    public void addTableModelListener(javax.swing.event.TableModelListener tl) {}
    public void removeTableModelListener(javax.swing.event.TableModelListener tl) {}
    public Class getColumnClass(int col){ return  Integer.class;}
    public int getRowCount(){ return countExons();}
    public int getColumnCount(){return 5;}
    public Object getValueAt(int row,int col) {
        try {
            int i=0;
            int start[]=exons()[0],end[]=exons()[1];
            switch(col) {
            case 0: i=row+1;break;
            case 1: i=start[row]+1;break;
            case 2: i=end[row];break;
            case 3: i=protein.countNucleotides()-start[row];break;
            case 4: i=protein.countNucleotides()-end[row]+1;break;
            default:
            }
            return intObjct(i);
        } catch(Exception e) {stckTrc(e);return "error";}
    }
    public boolean isCellEditable(int row,int col) { return col>0;}
    public String getColumnName(int i) {
        switch(i) {
        case 0:return "#";
        case 1:return "start";
        case 2:return "end";
        case 3:return protein.countNucleotides()+"-start+1";
        case 4:return protein.countNucleotides()+"-end+1";
        default:
        }
        return "error";
    };
    private int value, rowAtValue=-1,colAtValue=-1;
    public void setValueAt(Object o, int row, int col) {
        if (!(o instanceof Integer)) return;
        value=atoi(o);
        rowAtValue=row;
        colAtValue=col;
        setChanges(col==3||col==4);
    }
    // ----------------------------------------
    /* ---  Inner Model --- */
    private long countExons_last_isTranslated=-1;
    private int last_countExons;
    private boolean isTranslated[];
    private int countExons() {
      final int mc=protein.getModificationCount(MC_ARE_NUCLEOTIDES_TRANSLATED);
        if (countExons_last_isTranslated!=mc) {
            countExons_last_isTranslated=mc;
            isTranslated=protein.areNucleotidesCurrentStrandTranslated();
            int n=protein.countNucleotides();
            if (isTranslated==null) { isTranslated=new boolean[n];Arrays.fill(isTranslated,true);}
            int c=0;
            for(int i=mini(n,isTranslated.length);--i>=0;)
                if (isTranslated[i]&&(i==0 || !isTranslated[i-1]))
                    c++;
            last_countExons=c;
        }
        return last_countExons;
    }
    // ----------------------------------------
    private int[][] last_exons;
    private long exons_last_isTranslated=-1;
    private int[][] exons() {
      final int mc=protein.getModificationCount(MC_ARE_NUCLEOTIDES_TRANSLATED);
        if (exons_last_isTranslated!=mc || last_exons==null) {
            exons_last_isTranslated=mc;
            final int n=countExons();
            final int start[]=new int[n], end[]=new int[n], nNucl=mini(isTranslated.length,protein.countNucleotides());
            boolean isTrans=false;
            for(int i=0,count=0;i<=nNucl;i++) {
                if (!isTrans && i<nNucl && isTranslated[i]) start[count]=i;
                if (isTrans&&(i==nNucl || !isTranslated[i])) end[count++]=i;
                if (i<nNucl) isTrans=isTranslated[i];
            }
            last_exons=new int[][]{start,end};
        }
        return last_exons;
    }
    // ----------------------------------------
    /* --- The User Edits a cell ... --- */
    private int myGetValueAt(int  row, int col) {
        if (row==rowAtValue && col==colAtValue) return value;
        return atoi(getValueAt(row,col));
    }
    private int[][] readTable(boolean fromReverse) {
        int n=getRowCount(),start[]=new int[n],end[]=new int[n];
        int nNucl=protein.countNucleotides();
        for(int i=0;i<n;i++) {
            if (fromReverse) {
                start[i]=nNucl-myGetValueAt(i,3);
                end[i]=nNucl-myGetValueAt(i,4)+1;
            }
            else {
                start[i]=myGetValueAt(i,1)-1;
                end[i]=myGetValueAt(i,2);
            }
        }
        // check
        for(int i=0;i<n;i++)
            if (start[i]>=end[i] ||start[i]<0|| end[i]>nNucl)
                return null;
        return new int[][]{start,end};
    }
    // ----------------------------------------
    /* --- TextField --- */
    private final JTextField tf=new JTextField("");
    private final ChButton cbRevComplement=toggl("R",ChButton.ICON_SIZE).li(this).tt("reverse or reverse complement");
    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        if (q==cbRevComplement) setTextField();
        if (cmd=="GO"){
                final String s=rplcToStrg("..","-",tf.getText());
                boolean bb[]=parseSet(s,protein.countNucleotides(),-1);
                if (cbRevComplement.isSelected()) {
                    final int nNucl=protein.countNucleotides();
                    final boolean bbLonger[]=new boolean[nNucl];
                    System.arraycopy(bb,0,bbLonger,0,mini(nNucl,bb.length));
                    reverseArray(bbLonger,0,Integer.MAX_VALUE);
                    bb=bbLonger;
                }
                setNucleotidesCurrentStrandTranslated(bb);
                setTextField();
        }
    }
    private void setTextField() {
        final int ii[][]=exons();
        if (ii==null) {tf.setText("error");return;}
        final int start[]=ii[0], end[]=ii[1];
        final StringBuffer sb=new StringBuffer();
        final int n=start.length,nNucl=protein.countNucleotides();
        if (cbRevComplement.isSelected())
            for(int i=n;--i>=0;) {
                sb.append(nNucl-end[i]+1).append("..").append(nNucl-start[i]).append(",");
            }
        else
            for(int i=0;i<n;i++) {
                sb.append(start[i]+1).append("..").append(end[i]).append(",");
            }
        tf.setText(sb.toString());
        tf.revalidate();
    }
    // ----------------------------------------
    private void setChanges(boolean fromReverse) {
        final int ii[][]=readTable(fromReverse);
        if (ii==null) return;
        final int start[]=ii[0],end[]=ii[1];
        int nNucl=protein.countNucleotides();
        final boolean bb[]=new boolean[nNucl];
        for(int iExon=start.length;--iExon>=0;) {
            Arrays.fill(bb,start[iExon],end[iExon],true);
        }
        setNucleotidesCurrentStrandTranslated(bb);
    }
    private void setNucleotidesCurrentStrandTranslated(boolean bb[]) {
        protein.setNucleotidesCurrentStrandTranslated(bb);
        new StrapEvent(this,StrapEvent.NUCL_TRANSLATION_CHANGED).setParameter(protein).run();
        new StrapEvent(this,StrapEvent.ALIGNMENT_CHANGED).run();
        ChDelay.afterMS(EDT,ChDelay.CONTENTS_CHANGED,table,222,null);
        setTextField();
    }
    // ----------------------------------------
    public void handleEvent(StrapEvent ev){
        final int t=ev.getType();
        if (t==StrapEvent.NUCL_TRANSLATION_CHANGED) ChDelay.runAfterMS(EDT,this,delay);
        if ( (t&StrapEvent.FLAG_ROW_HEADER_CHANGED)!=0) labProtName().repaint();
    }
}
