package charite.christo.strap;
import static charite.christo.ChUtils.*;
import charite.christo.*;
import charite.christo.protein.*;
import charite.christo.interfaces.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.util.List;
import java.io.*;
import static java.lang.System.*;
import static charite.christo.ChConstants.*;
/**
   Common part of  the dialogs {@link DialogFetchSRS DialogFetchSRS} and {@link DialogFetchPdb DialogFetchPdb}.
   @author Christoph Gille
*/
public final class FilesFetchedFromServer extends AbstractDialogJPanel implements ChRunnable, ActionListener, MouseListener {
    private final ChTextArea LOG=new ChTextArea("");
    private final OverlayScrollPane scroll;
    private ChButton butDelError;
    private final List<File> vDownloaded=new UniqueList(File.class).addOptions(UniqueList.USE_HASH_SET), vAlreadyPresent=new UniqueList(File.class).addOptions(UniqueList.USE_HASH_SET);
    private final List<String>vFailure=new UniqueList(File.class).addOptions(UniqueList.USE_HASH_SET);
    private final ChJList jList,jListAlreadyPresent,jListFailure;

    public FilesFetchedFromServer(String urls[], String fileNames[]) {
        super();
        jListAlreadyPresent=new ChJList(vAlreadyPresent,ChJList.OPTIONS_FILES);
        jListFailure=new ChJList(vFailure,ChJList.OPTIONS_FILES);
        jList=new ChJList(vDownloaded,ChJList.OPTIONS_FILES);
        scroll=new OverlayScrollPane(jList);
        if (urls!=null) {
            final StrapFileCellRenderer r=new StrapFileCellRenderer(null);
            jList.setCellRenderer(r);
            jListAlreadyPresent.setCellRenderer(r);
            ChDelay.newThread(this,"F",new Object[]{urls,fileNames}).start();
            ChDelay.newThread(this,"SOAP",urls).start();
        }
        final JComponent
            pButtons=pnl(
                         new ChButton(BUT_LAB_Load).li(this).tt("The list shows the downloaded protein files.<br>Load the selected proteins into STRAP."),
                         new ChButton("log",ChButton.NO_MARGIN).actionViewMessage("log",LOG)),
            pSouth=vBpnl(pnl(labStatus,"<sup>Note: Right-click for popup-menu on files</sup>"),pButtons),
            lExist=new OverlayScrollPane(jListAlreadyPresent).addStringMatchButton(null,0L,"Fetched"),
            lError=new OverlayScrollPane(jListFailure).addStringMatchButton(null, 0L,"Fetched");
        scroll.addStringMatchButton(null, 0L, "Fetched");
        final ChTableLayout tl=new ChTableLayout(ChJTable.NO_REORDER);
        tl.addColumn(scroll);
        tl.addColumn(lExist);
        tl.addColumn(lError);
        tl.setTitle(0,"Success");
        tl.setTitle(1,"Already existing");
        tl.setTitle(2,"Not fetched");
        cnsew(this,tl,pTop(),pSouth,null,null);
        //return this;
    }

    public Object run(String id, Object arg) {
        final Object argv[]= arg instanceof Object[] ? (Object[])arg : null;
        if (id=="SOAP") SOAP_WSDbfetch.download((String[])arg,SOAP_WSDbfetch.SKIP_IF_ANY_EXISTS);
        if (id=="F") {
            final String[] urls=(String[])argv[0], fileNames=(String[])argv[1];
            final HashSet<String> vAvoidAskingTheSame=new HashSet();
            nextUrl:
            for(int i=0; i<urls.length && !this.isDisposed(); i++) {
                final String idColonDb=urls[i],  fn=fileNames[i], url=Hyperrefs.toUrlString(idColonDb,Hyperrefs.PROTEIN_FILE);
                final int lFn=sze(fn);
                if (idColonDb==null || lFn==0 || fn.indexOf(' ')>=0) continue;
                final File f=file(fn), fAllChains;
                final char chain;
                if (fn.endsWith(".ent") && chrAt(lFn-6, fn)=='_') {
                    fAllChains=file(fn.substring(0,lFn-6)+".ent");
                    chain=chrAt(lFn-5,fn);
                } else {fAllChains=f; chain=0;}
                {
                    final String s=url+chain;
                    if (vAvoidAskingTheSame.contains(s)) continue;
                    vAvoidAskingTheSame.add(s);
                }
                log("\nurl: ",url," fileNames: ",fn," allChains: ",fAllChains,"\nCache: ",file(url), "\n");
                if (sze(f)>0) {
                    log(f.getName()," already exists. skipped.\n");
                    vAlreadyPresent.add(f);
                    fileViewsUpdate();
                    continue nextUrl;
                }
                log("FilesFetchedFromServer: going to open ",url,"\n");
                final File fSoap=SOAP_WSDbfetch.dbColonID2file(idColonDb);
                final File fDownload= sze(fSoap)>0 ? fSoap : download_TS(url(url),999);
                putln(" fSoap="+fSoap);
                if (sze(fDownload)==0) {
                    log("unable to download ",url,"\n");
                    labStatus.setForeground(C(0xFF0000));
                    delFile(fDownload);
                    vFailure.add(url);
                    continue nextUrl;
                }
                final String fnDL=fDownload.getAbsolutePath();
                if (fnDL.endsWith(".Z") || fnDL.endsWith(".gz")||fnDL.endsWith(".bzip2")||fnDL.endsWith(".zip")) {
                    final InputStream is=inStream(fnDL,STREAM_UNZIP);
                    if (is!=null) cpy(is,f);
                    else {
                        delFile(fDownload);
                        continue nextUrl;
                    }
                } else {
                    if (sze(fDownload)==0) continue nextUrl;
                    if (sze(fDownload)<33333) {
                        String txt=readStrg(fDownload);
                        if (txt==null) continue;
                        txt=txt.toLowerCase().trim();
                        if (sze(txt)<7 || txt.startsWith("no entries") || txt.startsWith("<html>")) {
                            final File fError=file(f+"_error");
                            renamFileOrCpy(fDownload,fError);
                            vDownloaded.add(fError);
                            fileViewsUpdate();
                            if (butDelError==null) scroll.addButton(butDelError=new ChButton("DEL_ERROR").t("remove error files").li(this));
                            continue nextUrl;
                        }
                    }
                    if (chain!=0) {
                        if (!PDB_separateChains.writePdbOnlyChain(f, chain, readBytes(fDownload) ))
                            hardLinkOrCopy(fDownload,fAllChains);
                    } else hardLinkOrCopy(fDownload,f);
                }
                labStatus.setForeground(C(0xFF00));
                vDownloaded.add(sze(f)>0 ? f : fAllChains);
                log("success: ",fAllChains,"\n");
                fileViewsUpdate();
            }
            labStatus.t(" ALL DONE ");
            labStatus.setForeground(C(0xFF00));
            fileViewsUpdate();
        }
        return null;
    }
    private void fileViewsUpdate() {
        for(int i=3;--i>=0;) {
            final ChJList l=i==0?jList:i==1?jListAlreadyPresent:jListFailure;
            l.revalidate();
            l.repaint();
            ChDelay.afterMS(EDT,ChDelay.CONTENTS_CHANGED,jList,111,null);
        }

    }
    private final ChTextField labStatus=new ChTextField(" busy fetching ...").ro();
    public final static String BUT_LAB_Load="Load selected proteins";
    public void actionPerformed(ActionEvent ev) {
        if (align==null) return;
        final String cmd=ev.getActionCommand();
        {
            String s=null;
            Color c=null;
            final String soap_f=delPfx(SOAP_WSDbfetch.CMD_WSDBFETCH_SEND_QUERY,cmd);
            if (cmd!=soap_f) {
                s="SOAP WSDbfetch: "+soap_f;
                c=C(0xFF);
            }
            final String soap_s=delPfx(SOAP_WSDbfetch.CMD_WSDBFETCH_SUCCESS,cmd);
            if (cmd!=soap_s){
                s="SOAP WSDbfetch success\n";
                c=C(0xFF00);
            }
            final String soap_e=delPfx(SOAP_WSDbfetch.CMD_WSDBFETCH_ERROR,cmd);
            if (cmd!=soap_e) {
                s="SOAP WSDbfetch failed\n";
                final File fErr=file(soap_e);
                if (sze(fErr)==0) error(soap_e);
                c=C(0xFF0000);
            }
            if (s!=null && c!=null) {
                log(c,s,C(0xFFffFF),"\n");
                StrapAlign.drawMessage("@2"+ANSI_GREEN+s);
            }
        }
        if (cmd=="DEL_ERROR") {
            for(int i=sze(vDownloaded); --i>=0;) {
                final File f=vDownloaded.get(i);
                if (f.getName().endsWith("_error")) {
                    delFile(f);
                    vDownloaded.remove(f);
                }
            }
            fileViewsUpdate();
        }
        if (cmd==BUT_LAB_Load){
            final Object[] oo={jList.getSelectedValues(),jListAlreadyPresent.getSelectedValues()};
            if (sze(oo[0])+sze(oo[1])==0 && (sze(vDownloaded)+sze(vAlreadyPresent)<6 || ChMsg.yesNo("No protein file selected.<br>Load all?"))) {
                oo[0]=vDownloaded;
                oo[1]=vAlreadyPresent;
            }
            final StringBuffer sb=new StringBuffer();
            for(Object o : oo) {
                for(int i=sze(o); --i>=0;) {
                    final File f=get(i,o,File.class);
                    if (f!=null)  sb.append(f.getName()).append(' ');
                }
            }
            align.loadTheProteinsInList(toStrg(sb),StrapAlign.OPTION_PROCESS_LOADED_PROTS|StrapAlign.OPTION_EVENT_PROTEINS_ADDED);
            fileViewsUpdate();
        }
    }
    private void log(Object...oo) { for(Object o:oo) LOG.appnd(o); }

    public static JComponent docuSnapshot() {
        final FilesFetchedFromServer fss=new FilesFetchedFromServer( null,null);
        fss.vDownloaded.add(file("HSLV_ECOLI"));
        fss.vDownloaded.add(file("HSLV_BACSU"));
        final JComponent c=fss;
        c.setPreferredSize(dim(600,400));
        return c;
    }

}
