package charite.christo.strap;
import static charite.christo.ChUtils.*;
import charite.christo.*;
import java.util.Arrays;

/**HELP
When there are gaps occurring in all loaded sequences at the same position they are usually not displayed.
<br>
Consider the following case:

<pre class="data">
ASGATA  YTG
ATGATG  YTA
ASGGTAG FSG
       ^
</pre>
This is a common gap which is not displayed. The view shows:

<pre class="data">
ASGATA YTG
ATGATG YTA
ASGGTAGFSG
</pre>

When the sequence "ASGGTAGAFSG"
was added to the multiple sequence alignment  the gap  would not be a common gap any more and would become visible.
In the multiple sequence alignment panel is the following:

<pre class="data">
ASGATA  YTG
ATGATG  YTA
ASGGTAG FSG
ASGGTAGAFSG
</pre>

This is the reason for the following phenomenon:
When gaps are added to one sequence it might seem that gaps are erased
in all other sequences instead.

<br>

<i>SEE_CLASS:StrapKeyStroke</i>
   @author Christoph Gille
*/
public class Gaps2Columns {

    private UniqueList<StrapProtein>proteinsV;
    private boolean isNoGap[]=new boolean[666];
    private boolean isValid, isWide;
    private int w2n_mc;
    private int[] wide2narrow=new int[666], narrow2wide=new int[333];
    private int countColumns=0;
    public void setWide( boolean b){isWide=b; isValid=false; }
    public int countColumns() { wide2narrowColumn(); return countColumns;}
    public StrapProtein[] getProteins(){ return proteinsV.asArray();}
    public boolean isWide() { return isWide;}
    public int getModificationCount() { return w2n_mc;}
    public void setInvalid() {  isValid=false;}
    public void setProteinsV(UniqueList<StrapProtein>v) { proteinsV=v;}

    /**
     * Calculates the horizontal position (column) of all residues in the alignement.
     * from the number of gaps of each residue
     * Used for the default modus where  redundant gaps are suppressed.
     */
    public synchronized int[] computeWide2narrow(StrapProtein pp[]) {
        if (pp==null) return null;
        w2n_mc++;
        int maxCol=0;
        Arrays.fill(isNoGap,false);
        for(StrapProtein p :pp) {
            if (p==null) continue;
            final int gaps[]=p.getResidueGap(), nRes=p.countResidues();
            int col=-1;
            for(int i=0;i<nRes;i++) {
                assert gaps[i]>=0;
                col+=gaps[i]+1;
                if (isNoGap.length<=col) isNoGap=chSze(isNoGap,maxi(999,col*3/2));
                isNoGap[col]=true;
            }
            if (maxCol<col) maxCol=col;
        }
        int w2s[]=wide2narrow;
        if (w2s.length<=maxCol+1 || w2s.length>maxCol+666) this.wide2narrow=w2s=new int[maxCol+333];
        int col=0;
        for(int i=0;i<=maxCol;i++) {
            w2s[i]=col;
            if (isNoGap[i]) col++;
        }
        if (isNoGap.length>maxCol+666) isNoGap=new boolean[maxCol+333];
        w2s[maxCol+1]=col++;
        this.countColumns=col;
        int[] s2w=this.narrow2wide;
        if (s2w.length<=col || s2w.length>col+666) this.narrow2wide=s2w=new int[col+333];
        for(int i=maxCol+1; --i>=0;) s2w[w2s[i]]=i;
        //if (pp.length>10000 && debugTime("computeWide2narrow duration=",time);
        return w2s;
    }
    public int[] wide2narrowColumn() {
        if (!isValid || wide2narrow==null) {
            computeWide2narrow(getProteins());
            w2n_mc++;
            isValid=true;
        }
        return this.wide2narrow;
    }
    public int[] narrow2wideColumn() {
        wide2narrowColumn();
        return this.narrow2wide;
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> static utils >>> */
    protected static int[] computeColumns(int wide2narrow[],int gaps[], int nRes, int oldCols[]) {
        final int cols[]=redim(oldCols,nRes,99);
        final int nW=sze(wide2narrow);
        int  col=-1, lastCol=-1, i=0;
        for(; i<nRes; i++) {
            final int g=i<gaps.length ? gaps[i] : 0;
            assert g>=0;
            col+=g+1;
            if (col>=nW) {
                if (isMyComputer()) putln(" col="+col+" nW="+nW);
                break;
            }
            int newCol=wide2narrow[col];
            if (newCol<=lastCol) newCol=lastCol+1;
            lastCol=cols[i]=newCol;
        }
        for(; i<nRes; i++) {
            final int g=i<gaps.length ? gaps[i] : 0;
            assert g>=0;
            col+=g+1;
            if (col<=lastCol) col=lastCol+1;
            lastCol=cols[i]=col;
        }
        return cols;
    }
}
