package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.awt.*;
import java.awt.event.MouseEvent;
import java.util.Arrays;
import javax.swing.JComponent;
import static java.awt.event.MouseEvent.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
   The View of an alignment window has a horizontal scrollbar which is used as a viewport.
   @author Christoph Gille
*/
public class HScrollBar extends ChScrollBar implements ChRunnable {
    private final StrapView view;
    public HScrollBar(StrapView v) {
        super(HORIZONTAL);
        noInsets(this);
        view=v;
        CollapseComponent.preventedBy(this);
        setPreferredSize(dim(1,ICON_HEIGHT));
    }
    /* ---------------------------------------- */
    /* >>> Image >>> */
    private boolean hsb_clear;
    private transient java.lang.ref.Reference<Image> hsb_refImage;
    public HScrollBar clearImage() { hsb_clear=true; return this;}
    private  void paintImage(Graphics g,final boolean isOutline,int w,int h) {
        final int RGB_AFTER_SEQUENCE= getBackground().getRGB();
        final StrapProtein[] pVis=view.getVisibleProteinsReordered();
        final int nP=pVis.length,countCols=view.countColumns();
        if (countCols<1 || nP==0 || h<2 || w<2) { return;}
        final java.awt.image.BufferedImage imageLine=sharedImage('H',countCols);
        final int unit=countCols/w+1;
        final int[] rgb=sharedRGB(countCols+unit+1);
        /* --- gaps black --- */
        final int[] shadingColors=colors2rgb(view.getShadingColors());
        if (shadingColors==null) return;
        final int similThreshold=StrapView.sliderSimilarity().getValue();
        final boolean showSecStru=ProteinList.isEnabled(pVis,ProteinList.SECSTRU) && view.isShadingSecStru();
        boolean isNewline=true;

        final StrapProtein[] ppShading=view.ppShading();
        final short[][] frequencies=isOutline ? view.computeFrequenciesAtColumn(0,MAX_INT) :null;
        int firstCol=MAX_INT, lastCol=MIN_INT;
        for(int iP=0;iP<nP;iP++) {
            if (isNewline) {
                Arrays.fill(rgb,0,countCols, 0x999999);
                firstCol=MAX_INT;
                lastCol=MIN_INT;
                isNewline=false;
            }
            final int y1=h*iP/nP, y2=mini(h*(1+iP)/nP,h);
            final StrapProtein p=pVis[iP];
            if (p==null) continue;
            final boolean isShading=isOutline && (ppShading==pVis || cntains(p,ppShading));
            final int nRes=p.countResidues();
            final int cols[]=p.getResidueColumn(), nCols=nRes==0 ? 0:p.getMaxColumn();
            if (nRes==0||nCols==0) continue;
            if (firstCol>cols[0]) firstCol=cols[0];
            if (lastCol<cols[nRes-1]) lastCol=cols[nRes-1];
            if (!isShading) drawSelection(p,rgb,unit);
            if (y1!=y2 || iP==nP-1) {
                isNewline=true;
                final byte ss[]=showSecStru && isShading ? p.getGappedSecStru() : null;
                for(int iCol=nCols;--iCol>=0;) {
                    if (iCol>=countCols)continue;
                    final byte[] gapped=p.getGappedSequence();
                    final int rt=gapped[iCol]|32;
                    if (rt<'a' || rt>'z') rgb[iCol]=0;
                    else if (isShading) {
                        final int c=(rt|32) - 'a';
                        final boolean isConserved=similThreshold==0 ||
                            frequencies.length>iCol && frequencies[iCol] !=null && c>=0 && c<1+'z'-'a' && frequencies[iCol][c]*100>ppShading.length*similThreshold;
                        if (ss==null) {
                            if (isConserved) rgb[iCol]=shadingColors[rt];
                        } else if (iCol<ss.length) {
                            final byte s=ss[iCol];
                            if (s=='H') rgb[iCol]=isConserved ? 0xff0000 : 0xaa0000;
                            else if (s=='E') rgb[iCol]= isConserved ? 0xffff00 : 0xaaaa00;
                            else if (s==' ') rgb[iCol]=0;
                        }
                    }
                }
                Arrays.fill(rgb,0,mini(countCols,firstCol+1),RGB_AFTER_SEQUENCE);
                if (lastCol<countCols) Arrays.fill(rgb, maxi(0,lastCol+1), countCols,RGB_AFTER_SEQUENCE);

                try {
                    imageLine.setRGB(0,0,countCols,1,rgb,0,countCols);
                    g.drawImage(imageLine,0,y1,w,y2, 0,0,countCols,1,this);
                } catch(Throwable ex) { if (isMyComputer()) stckTrc(ex); }
            }
        }
    }
    /* <<< Image <<< */
    /* ---------------------------------------- */
    /* >>> paintComponent >>> */
    private boolean  hsb_likeDivider;
    private int hsb_lastThumbX;
    private int hsb_dividerWidth=EM,hsb_lastW,hsb_lastH;
    @Override public void paintChildren(Graphics g) {       }
    protected final long LAST[]=new long[ARRAY_DURATION];
    @Override public void paintComponent(Graphics g) {
        final int tw,th,tx,ty,  width,height,trx;
        {
            final Rectangle track=getTrack(), thumb=getThumb();
            tw=thumb!=null? thumb.width : 1;
            th=thumb!=null? thumb.height :10;
            tx=thumb!=null? thumb.x : 0;
            ty=thumb!=null? thumb.y : 0;
            if (track==null) { super.paintComponent(g); return; }
            width=track.width;
            height=track.height;
            trx=track.x;
        }
        final Rectangle visible=view.getVisibleXYWH();
        if (tx!=hsb_lastThumbX) {
            StrapAlign.dispatchLater(StrapEvent.ALIGNMENT_SCROLLED,99);
            if (getValue()>0 && !isAlreadyScrolled) {
                isAlreadyScrolled=true;
                repaint();
            }
            hsb_lastThumbX=tx;
        }

        final int nRows=view.countRows();
        final int nCols=view.countColumns();
        if (!isAlreadyScrolled  || nRows==0 || !view.showsSequence()){
            try {
                super.paintComponent(g);
                if (view.showsSequence() && nRows>0) {
                    final String s="This is the scroll-bar of the alignment panel. It can also scroll vertically.";
                    final Font f=new Font("",Font.PLAIN,15);
                    g.setFont(f);
                    final FontMetrics fm=getFontMetrics(f);
                    final int x=(getWidth()-javax.swing.SwingUtilities.computeStringWidth(fm,s))/2;
                    g.setColor(Color.BLACK);
                    g.drawString(s,x, (getHeight()-fm.getHeight())/2+fm.getAscent());
                }
            } catch(Throwable e){}
            return;
        }
        if (nCols<1)  return;
        final long time=System.currentTimeMillis();
        g.setColor(getBackground());
        g.fillRect(0,0,trx,height);
        g.fillRect(trx+width,0,height,999);
        final Image image;
        {
            Image i=(Image)deref(hsb_refImage);
            final int iH=i!=null ? i.getHeight(this) : MIN_INT, iW=i!=null ? i.getWidth(this) : MIN_INT;
            if (iH<height || iW<width) i=null;
            if (hsb_clear || i==null || hsb_lastW!=width || hsb_lastH!=height) {
                hsb_lastW=width;
                hsb_lastH=height;
                hsb_clear=false;
                if (i==null)  hsb_refImage=new java.lang.ref.SoftReference(i=createImage(width,height+10));
                paintImage(i.getGraphics(),view.isShadingScrollbar(),width,height);
                i=(Image)deref(hsb_refImage);
            }
            image=i;
        }
        if (image==null) return;
        g.drawImage(image,trx,0, trx+width, height,  0,0,width, height,this);
        if (nRows==0 || nCols<5) return;
        final Dimension paneSize=view.getPaneSize();
        if (visible.width<paneSize.width) {
            final int
                y1=visible.y*height/paneSize.height,
                dy=visible.height*height/paneSize.height;
            if (visible.height<paneSize.height) {
                g.setColor(C(0x8080ff,0x80));
                g.fillRect(tx,0,tw,y1);
                g.fillRect(tx, y1+dy, tw,th);
                g.setColor(Color.BLUE);
                g.drawLine(tx, y1,tx+tw,y1 );
                g.drawLine(tx, y1+dy,tx+tw,y1+dy );
            } else {
                g.setColor(C(0x8080ff,0x60));
                g.fillRect(tx,ty,tw,th);
            }
            g.setColor(Color.BLUE);
            g.drawLine(tx,0,tx,th);
            g.drawLine(tx+tw,0, tx+tw, th);
        }
        final Rectangle r=rectangleCursor();
        if (r!=null) ChIcon.drawCursor(g,r.x,r.y,r.width,r.height,this);
        {
            /* --- ValueOfAlignPosition --- */
            g.translate(trx,0);
            view.plot(g,this,width/(double)nCols,0,getHeight());
            g.translate(-trx,0);
        }
        if (hsb_likeDivider) looksLikeDivider(this,g,0,0,9999,hsb_dividerWidth);
        LAST[DURATION]=(LAST[WHEN]=System.currentTimeMillis())-time;
        paintHooks(this, g, true);
    }

    private static void drawSelection(StrapProtein p,int rgb[], int unit) {
        final int cols[]=p.getResidueColumn(), nR=p.countResidues();
        if (cols==null || cols.length<nR) return;
        for(ResidueSelection s : p.getAllResidueSelections()){
            if (!(s instanceof VisibleIn123)) continue;
            final VisibleIn123 shown= (VisibleIn123)s;
            if ((shown.getVisibleWhere()&VisibleIn123.SB)==0) continue;
            final Color c=shown.getColor();
            final int color=c!=null ? c.getRGB() : 0;
            final boolean selAA[]=ResidueAnnotation.getSelectedAminoacidsDisplayed(s);
            if (selAA==null || selAA.length==0) continue;
            final int selOffset=s.getSelectedAminoacidsOffset();
            final int iMax=mini(selAA.length+selOffset, nR);
            for(int iAa=maxi(0,selOffset); iAa<iMax; iAa++) {
                if (selAA[iAa-selOffset]) {
                    final int iCol=cols[iAa];
                    for(int i=unit+iCol; --i>=iCol;)
                        if (i<rgb.length)
                            rgb[i]=color;
                }
            }
        }
    }
    /* <<< paintComponent <<< */
    /* ---------------------------------------- */
    /* >>> Cursor >>> */
    private final static Rectangle hsb_rectCursor=new Rectangle();
    public void paintCursor() {
        repaint(hsb_rectCursor);
        final Rectangle r=rectangleCursor();
        if (r!=null) repaint(r);
    }
    private Rectangle rectangleCursor() {
        final Rectangle r=this.hsb_rectCursor;
        if (!getRectangle(view.getRowAtCursor(), view.getColumnAtCursor(), r)) return null;
        final int w=16;
        r.x+=(r.width-w)/2;
        r.y+=(r.height-w)/2;
        r.width=r.height=w;
        return r;
    }
    private boolean  getRectangle(int row, int col, Rectangle r) {
        final int nRows=view.countRows();
        final int nCols=view.countColumns();
        if (nCols<0||nRows<0|| nCols*nRows==0) return false;
        final Rectangle track=getTrack();
        r.height=track.height/nRows;
        r.width =track.height/nCols;
        r.y= track.height*row / nRows + track.y;
        r.x= track.width *col / nCols + track.x;
        return true;
    }
    /* <<< Cursor <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */
    private char hsb_Dragged;
    private int hsb_lastDraggedY;
    private int hsb_maxDragX,minDragX, hsb_draggedFromY;
    private boolean isAlreadyScrolled;
    @Override public void processEvent(AWTEvent ev) {
        final int id=ev.getID();
        boolean mouseOverDivider=false;
        if (ev instanceof MouseEvent) {
            final MouseEvent mev=(MouseEvent)ev;
            final int
                y=mev.getY(),
                x=mev.getX(),
                modi=mev.getModifiers();
            final boolean oldLooksLikeSlider=hsb_likeDivider;
            mouseOverDivider= y<=hsb_dividerWidth;
            hsb_likeDivider=mouseOverDivider && id==MOUSE_MOVED || hsb_Dragged=='Z';
            if (hsb_likeDivider!=oldLooksLikeSlider) repaint();
            if (id==MOUSE_MOVED) {
                setCursor(getCursr(y<hsb_dividerWidth ? 'n' : 'M'));
                this.hsb_Dragged=0;
            }
            if (id==MOUSE_ENTERED  && (modi&(BUTTON1_MASK|BUTTON2_MASK|BUTTON3_MASK))==0) hsb_Dragged=0;

            if (id==MOUSE_DRAGGED) {
                if (hsb_Dragged==0) {
                    hsb_Dragged= mouseOverDivider ? 'Z' :  's';
                    this.hsb_draggedFromY=y;
                    hsb_maxDragX= minDragX=x;
                }
                else if (hsb_Dragged=='Z') {
                    setPreferredSize(dim(1,maxi(EX,getHeight()-y)));
                    StrapView.correctHeight();
                    hsb_dividerWidth=EM*2/3;
                    ChDelay.revalidate(this,99,(long[])null);
                    isAlreadyScrolled=true;
                }
            }
            if (hsb_Dragged=='s') {
                final int dy=y-hsb_draggedFromY;
                if ( (hsb_maxDragX=maxi(hsb_maxDragX,x))-(minDragX=mini(minDragX,x))<(dy>0?dy:-dy)-1  || (modi&ALT_MASK)!=0)  hsb_Dragged='S';
            }
            if (hsb_Dragged=='S') view.scrollRel('v', (y-hsb_lastDraggedY)*view.getPaneSize().height/(1+getHeight()),false);
            this.hsb_lastDraggedY=y;
            if (id==MOUSE_RELEASED ||id==MOUSE_EXITED && hsb_Dragged==0) {
                hsb_likeDivider=false;
                repaint();
            }
        }
        if (hsb_Dragged!='Z' && !(mouseOverDivider&&id==MOUSE_PRESSED)) {
            try { super.processEvent(ev);} catch(Throwable e){}
        }
        final int max=view.getPaneSize().width+1;
        if (max-getMaximum()>1 || max-getMaximum()<-1) setMaximum(max);
    }
    final static String RUN_ANIM="HSB$$A";
    @Override public Object run(String id, Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[])arg : null;

        final StrapView v=this.view;
        if (id==RUN_ANIM || id==RUN_ANIM+"c") {
            final StrapProtein p=(StrapProtein)argv[0];
            final int radius_iA[]=(int[])argv[1], iA=radius_iA[1];
            final JComponent pCenter=v.getAlignmentPane();

            final int row=v.findRow(p);
            if (row<0) return null;

            if (id==RUN_ANIM) {
                for(int radius=40; --radius>=0;) {
                    radius_iA[0]=radius;
                    ChDelay.invokeAndWait(this,RUN_ANIM+"c",argv);
                    sleep(20);
                }
                pCenter.repaint();
                this.repaint();
            } else {
                final int col=p.getResidueColumn(iA);
                final int nP=view.getVisibleProteinsReordered().length;
                if (nP>0) {
                    final int radius=2*radius_iA[0], r2=radius*2;

                    final Rectangle r=new Rectangle();
                    final Graphics g=getGraphics();
                    if (g!=null && getRectangle(row,col,r)) {
                        final int x=r.x+r.width/2;
                        final int y=r.y+r.height/2;
                        g.setColor(Color.WHITE);
                        for(int i=-1;i<=1;i+=2) g.drawOval(x-radius+i,y-radius+i,r2+2*i,r2+2*i);
                        g.setColor(Color.BLUE);
                        g.drawOval(x-radius,y-radius,r2,r2);
                    }
                }
                final Graphics g=pCenter.getGraphics();
                if (g!=null) {
                    final Rectangle B=v.charB();
                    final int x=v.col2x(col)+B.width/2, y=v.row2y(row)+B.height/2;
                    final int r0=radius_iA[0], radius=r0*r0,  r2=radius*2;
                    g.setColor(Color.WHITE);
                    for(int i=-1;i<=1;i+=2) g.drawOval(x-radius+i,y-radius+i,r2+2*i,r2+2*i);
                    g.setColor(Color.BLUE);
                    g.drawOval(x-radius,y-radius,r2,r2);
                }
            }
        }
        return null;
    }

}

