package charite.christo.strap;
import charite.christo.*;
import charite.christo.blast.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.io.File;
import java.lang.ref.Reference;
import java.util.*;
import javax.swing.JComponent;
import javax.swing.JTabbedPane;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static java.awt.Color.*;
/**HELP

Identifying Blast hits that are common to two sequences helps to compare remote homologs.

The analysis requires two steps.

<ol>
<li>One  protein is selected in one, the other in the other protein list. </li>
<li>For each sequence a BLAST search is performed against a non-redundant set of sequences such as uniref50

</li>
<li>For two sequences the BLAST hits are compared and all common segments presented to the user by pressing the button <i>LABEL:ChButton#BUTTON_GO</i>.
</li>

</ol>

Alignment programs perform better when sequences are added that share similarity with the proteins
to be aligned. Using two additional steps one can improve multiple sequence alignments:

<ol>
<li>Chose hits that are similar to either sequence by activating the respective check-boxes.</li>
<li>Protein objects are created for each activated Blast segment. At this stage overlapping segments with the same sequence ID but different start and end positions are joined.</li>
<li>Finally a multiple sequence alignment is computed <i>BUTTON:DialogAlign!</i>.</li>
</ol>

<br>
<b>Related publications:</b> PUBMED:11159329 PUBMED:9367767
<br>
   @author Christoph Gille
*/
public class IntermediateSeq  extends JTabbedPane implements ActionListener,Disposable,HasControlPanel,StrapListener,ChRunnable {
    private final Map<StrapProtein,Reference<BlastResult>>mapBlastResult=new WeakHashMap();
    private final UniqueList<BlastAlignment> vAlignments=new UniqueList(BlastAlignment.class);

    private ChCombo comboDatabases;
    private ChFrame fBlast;

    private ProteinList listProt1,listProt2;
    private Vector vDatabases=new Vector();
    private boolean isRunning=true;
    private final ChButton butLoad=new ChButton("Load all selected",ChButton.DISABLED).li(this);
    private final ChLabel BLAST_labStatus=new ChLabel("blasting, be patient ...");
    private final AsciiView BLAST_LOG=new AsciiView(999);
    private final JComponent BLAST_pnl=vB();

    private void updateDatabases(){
        vDatabases.clear();
        String defaultDB=null;
        SequenceBlaster blaster=comboClass.instanceOfSelectedClass(SequenceBlaster.class);
        final String dd[]=blaster!=null ? blaster.getAvailableDatabases() : null;
        if (dd==null) return;
        for(String s:  dd) {
            vDatabases.add(s);
            if (defaultDB==null && s.indexOf("50")>0) defaultDB=s;
        }
        if (defaultDB!=null) comboDatabases.s(defaultDB);
        if (sze(vDatabases)==0) { vDatabases.add("Error: no Database"); }
        comboDatabases.revalidate();
    }
    public void actionPerformed(ActionEvent ev) {
        StrapAlign.errorMsg("");
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        if (q==comboClass || q==comboDatabases) mapBlastResult.clear();
        if (q==comboClass) updateDatabases();
        if (cmd=="GO") {
            butLoad.enabled(false);
            ChDelay.newThread(this,"GO",null).start();
            butLoad.enabled(true);
        }
        if (q==butLoad) {
            IntermediateSeq.this.vAlignments.clear();
            for(IntermediateSeqPair p: childs(this,IntermediateSeqPair.class))
                for(BlastHitIntersection a: p.vAllAlignments.asArray())
                    if (a.SELECTED[0]) {
                        BlastAlignment ba=new BlastAlignment();
                        ba.setMatchStartEnd(a.getMatchStart(),a.getMatchEnd());
                        ba.setMatchSeq(onlyLetters(a.getSeq('M',0)),0, MAX_INT);
                        ba.setBlastHit(a.getAlignment(0).getBlastHit());
                        IntermediateSeq.this.vAlignments.add(ba);
                    }
            union();
            UniqueList<StrapProtein> vProteins=new UniqueList(StrapProtein.class);
            for(BlastAlignment a:vAlignments.asArray()) {
                final String pn="xx"+a.getID()+'-'+a.getMatchStart()+'-'+a.getMatchEnd();
                if (StrapProtein.getProteinWithName(pn,null)==null){
                    final StrapProtein p=new StrapProtein(align);
                    p.setName(pn);
                    p.setResidueType(a.getMatchSeq());
                    p.setHeader(toStrg(a.getBlastHit().getDescription()));
                    vProteins.add(p);
                }
            }
            if (sze(vProteins)==0) StrapAlign.errorMsg("You must select blast hits in the result tabs");
            final StrapView v=align.getAlignmentPanelWithFocus();
            if (v!=null) v.setCursorToColumn(v.getProtein(v.countRows()-1),1);
            align.addProteins(-1,vProteins.asArray());
        }
    }
    private static ChButton TOG_LOG=toggl("log");
    public Container getControlPanel() { return TOG_LOG.cb();}
    @Override protected void finalize() throws Throwable {
        logFinalize(ANSI_CYAN+"IntermediateSeq",getClass());
        super.finalize();
    }
    private StrapAlign align;
    private final ChTextArea LOG_NO_HITS=new ChTextArea("protein pairs without  common blast hits\n\n");
    private ChCombo comboClass;

    // ----------------------------------------
    public IntermediateSeq(StrapAlign al) {
        align=al;
        comboClass();
        listProt1=new ProteinList(align,0);
        listProt1.setSelI(0);
        listProt1.li(this);
        listProt2=new ProteinList(align,0);
        listProt2.setSelI(1);
        listProt2.li(this);
        getPanel();
        updateDatabases();
    }
    // ----------------------------------------
    protected final static ChTextField tfNumAlignments=new ChTextField("55").cols(4,true,true);
    public Object run(String id, Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[])arg : null;
        if (id=="GO") {
            boolean show=false;
            int count=0;
            StrapProtein[] pp0=listProt1.getSelectedProteins(),pp1=listProt2.getSelectedProteins();
            putln("pp0=",pp0);
            putln("pp1=",pp1);
            if(pp0.length==0) pp0=pp1;
            if(pp1.length==0) pp1=pp0;
            if (pp0.length==0) {
                StrapAlign.errorMsg("Select at least one protein in the top and button protein list");
                BLAST_labStatus.t("error").fg(RED);
                return null;
            }
            final BlastResult[][] results=new BlastResult[2][];
            for(int i2=0; i2<2; i2++) {
                final StrapProtein[] pp=i2==0 ? pp0 : pp1;
                results[i2]=new BlastResult[pp.length];
                for(int iP=0; iP<pp.length; iP++) {
                    final StrapProtein p=pp[iP];
                    BlastResult result=gcp(p,mapBlastResult,BlastResult.class);
                    if (result==null) {
                        if (fBlast==null) {
                            final JComponent panMain=newCNSEW(scrllpn(0L,BLAST_pnl),"http://www.ebi.ac.uk/blast2/",pnl(BLAST_labStatus,new ChButton("log").actionViewMessage("Log",BLAST_LOG)),null,null);
                            fBlast=new ChFrame("blast").ad(panMain).shw(ChFrame.AT_CLICK);
                        }
                        show=true;
                        if (!isRunning) break;
                        BLAST_labStatus.t(count++ +" / "+(pp0.length+pp1.length));
                        final SequenceBlaster blaster=HotswapUtil.instance(comboClass(),SequenceBlaster.class,true);
                        if (blaster==null)  { BLAST_LOG.appendS("Error: could not instantiate blast service\n");  return null; }
                        blaster.setAAQuerySequence(p.getResidueTypeAsStringUC());
                        blaster.setDatabase((String)comboDatabases.getSelectedItem());
                        blaster.setNumberOfAlignments(2000);
                        final ChButton butCtrl=ChButton.actionCtrl(blaster);
                        pcp("don't gc",blaster, butCtrl);
                        if (butCtrl!=null) ChDelay.addComponentToContainer(butCtrl.t(p.getName()).i(p.getIcon()),BLAST_pnl,null,33);
                        ByteArray xml=CacheResult.isEnabled() ? blaster.getResultXml() : null;
                        if (xml==null) {blaster.compute();  xml=blaster.getResultXml();}
                        sleep(99);
                        result=new BlastResult(xml);
                        if (butCtrl!=null) butCtrl.t(butCtrl.getText()+"        "+result.countAlignments()+" alignments");
                    }
                    mapBlastResult.put(p,new java.lang.ref.SoftReference(result));
                    results[i2][iP]=result;
                }
            }
            BLAST_labStatus.t("all done").fg(GREEN);
            if (show && fBlast!=null) fBlast.shw(0);
            ChDelay.invokeAndWait(this,"COMPUTE",new Object[]{pp0,pp1,results[0],results[1]});
        }
        if (id=="COMPUTE") {
            compute((StrapProtein[])argv[0], (StrapProtein[])argv[1], (BlastResult[])argv[2], (BlastResult[])argv[3]);
        }
        return null;
    }
    public void compute(StrapProtein[] pp1,StrapProtein[] pp2, BlastResult[] rr1, BlastResult[] rr2) {
        boolean success=false;
        for(int i=0; i<pp1.length; i++) {
            for(int j=0; j<pp2.length; j++) {
                final StrapProtein p1=pp1[i], p2=pp2[j];
                if (p1==p2)  { LOG_NO_HITS.appnd("skip ",p1," with itself\n");  LOG_NO_HITS.showInFrame("Protein pairs without  common blast hits",0); continue;}
                final IntermediateSeqPair r=new IntermediateSeqPair(p1,p2,rr1[i],rr2[j]);
                if (sze(r.vHitViewShown)==0) {
                    LOG_NO_HITS.appnd(p1.getName()," and ",p2,"\n");
                    LOG_NO_HITS.showInFrame("Protein pairs without  common blast hits",0);
                    r.dispose();
                }
                else {
                    success=true;
                    adTab("<sub>"+p1+"<br>"+p2+"</sub>",r, (JTabbedPane)this);
                }
                StrapAlign.progressBar((i*pp1.length+j)*100/pp1.length/pp2.length);
            }
        }
        if (success) setSelectedIndex(getTabCount()-1);
    }
    private ChCombo comboClass(){
        if (comboClass==null) comboClass=StrapAlign.classChoice(SequenceBlaster.class).li(this);
        return comboClass;
    }
    private JTabbedPane getPanel() {
        Container pNorth=align.headLine(this);
        ChButton butTry=new ChButton("GO").t(ChButton.GO).li(this).tt("Invokes comparison of BLAST hits.");
        comboClass().setPreferredSize(dim(comboClass().getPreferredSize().width,butTry.getPreferredSize().height));
        comboDatabases=new ChCombo(vDatabases,0).li(this);
        final Color bg=C(DEFAULT_BACKGROUND);
        final JComponent
            pFirst=vB(SCROLLABLE_TRACKS_VIEWPORT_WIDTH, GRAY,
                      hB(bg,"#ETB","1st<br>step"," ",
                         vB(
                            hB(bg,"Find common blast hits. ","#"),
                            hB(bg,"database: ", comboDatabases,"#"),
                            comboClass().panel(),
                            pnl(bg,cbCache()),
                            hB(bg,enlargeFont(butTry.options(ChButton.NO_MARGIN),1.5),"#")
                            )
                         )," ",
                      hB(bg,"#ETB","2nd<br>step"," ",
                         vB(
                            hB(bg,"Load those blast hits that are<br>activated with the check-boxes \"accept\"."),
                            hB(bg,enlargeFont(butLoad.options(ChButton.NO_MARGIN),1.5))
                            )
                         )," ",
                      hB(bg,"#ETB","final<br>step"," ",
                         vB(
                            hB(bg,"Align the sequences together<br>with the loaded blast hits."),
                            hB(bg,StrapAlign.b(CLASS_DialogAlign).options(ChButton.NO_MARGIN))
                            )
                         )
                      ),
            pFirstLR=pnl(new GridLayout(1,2),scrllpn(0L,pFirst) ,pnl(new GridLayout(2,1),listProt1.scrollPane(),listProt2.scrollPane())),
            pMain=newCNSEW(pFirstLR,pNorth,null,null,null);
        adMainTab(pMain, this,null);
        return this;
    }
    // ----------------------------------------
    protected static File getFileOfHits(StrapProtein pp[],String suffix) {
        String ss[]=new String[pp.length];
        for(int i=pp.length;--i>=0;) ss[i]=pp[i].getName();
        sortArry(ss);
        final File dirTmp=file(dirTmp()+"/X-blast");
        mkdrs(dirTmp);
        return file(dirTmp,joinAsStrg(ss,"+")+suffix);
    }
//     private static File getFileOfHits(StrapProtein pp1[],StrapProtein pp2[],String suffix) {
//         UniqueList<String>  v=new UniqueList(String.class);
//         for(StrapProtein p:pp1) v.add(p.getName());
//         for(StrapProtein p:pp2) v.add(p.getName());
//         Collections.sort(v);
//         final File dirTmp=file(dirTmp()+"/X-blast");
//         dirTmp.mkdirs();
//         return file(dirTmp,joinAsStrg(v.asArray(),"+")+suffix);
//     }
    // ========================================
    public IntermediateSeq(){}
    public void run(){
        IntermediateSeq x=new IntermediateSeq(StrapAlign.getInstance());
       new ChFrame("").ad(x).size(555,555).shw(0);
    }
    // ----------------------------------------

    private void union() {
        for(int i0=sze(vAlignments);--i0>=0;) {
            BlastAlignment a0=vAlignments.get(i0);
            if (a0==null) continue;
            for(int i1=sze(vAlignments);--i1>=0;) {
                BlastAlignment a1=vAlignments.get(i1);
                if (a1==null || a1==a0) continue;
                BlastAlignment u=BlastHitIntersection.union(a0,a1);
                if (u!=null) {
                    vAlignments.set(i0,a0=u);
                    vAlignments.set(i1,null);
                }
            }
        }
        while(vAlignments.remove(null));
    }
    public void dispose() {
        for(IntermediateSeqPair p: childs(this,IntermediateSeqPair.class)) p.dispose();
        rmAllChilds(this);
        mapBlastResult.clear();
    }
    public void handleEvent(StrapEvent ev) {
        final int type=ev.getType();
        if (type==StrapEvent.AA_SHADING_CHANGED) repaint();
    }
}
// http://bibiserv.techfak.uni-bielefeld.de/altavist/
//http://www.ncbi.nlm.nih.gov/entrez/query.fcgi?cmd=Retrieve&db=pubmed&dopt=Abstract&list_uids=16076885&itool=iconabstr&query_hl=8
