package charite.christo.strap;
import charite.christo.protein.*;
import charite.christo.interfaces.*;
import charite.christo.*;
import java.awt.*;
import javax.swing.*;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
Underlining residues with the mouse.
   @author Christoph Gille
*/
public class MouseSelection extends AbstractResidueSelection implements HasRendererComponent,Runnable {
    public final static String NAME="selected";
    {
        setColor(Color.CYAN);
        setVisibleWhere(0xffffffff);
        setStyle(VisibleIn123.DOTTED);
        setName(NAME);
    }
    private int lastFrom,lastTo;
    public void set(final Protein protein,final int col1,final int col2) {
        final StrapProtein p=(StrapProtein)protein;
        final int colMin=col1<col2 ? col1 : col2;
        final int colMax=col1>col2 ? col1 : col2;
        final Protein oldP=getProtein();
        setVisibleWhere(getVisibleWhere()|VisibleIn123.SEQUENCE);
        if (oldP!=p) {
            if (oldP!=null) oldP.removeResidueSelection(this);
             StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,22);
        }
        super.setProtein(p);
        if (p==null) return;
        p.addResidueSelection(this);

        final int from=p.column2nextIndex(colMin), nR=p.countResidues();
        int to=p.column2nextIndex(colMax);
        if (to<0) to=nR;
        super.setSelectedRange(from,to);
        if (lastFrom!=from || lastTo!=to){
            lastFrom=from;
            lastTo=to;
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,22);
            if ((getVisibleWhere()&VisibleIn123.STRUCTURE)!=0) {
                if (from>=0 && from<to && to<=nR) toClipbd(toStrg(p.getResidueTypeExactLength(),from,to));
                ChDelay.runAfterMS(EDT,this,222);
            }
        }
    }
    public void run() {
      final Protein p=getProtein();
      if (p!=null) ProteinViewerUtils.residueSelectionsTo3D(ProteinViewerUtils.RSto3D_NOT_SAME_WIRE, new ResidueSelection[]{MouseSelection.this}, p.getProteinViewers(null));
    }
    private final ChLabel RENDERER_COMPONENT=new ChLabel();
    public JComponent getRendererComponent() {
        final Protein p=getProtein();
        if (p==null) RENDERER_COMPONENT.t("");
        final int from=getSelectedAminoacidsOffset();
        final int to=from+lastTrue(getSelectedAminoacids());
        final int num[]=p.getResidueNumber();
        final byte chain[]=p.getResidueChain();
        String s="Mouse selection "+((p==null || num==null) ? null : " ("+get(from,num)+":"+chrAt(from,chain)+" to "+get(to-1,num)+":"+chrAt(to-1,chain)+")");
        return RENDERER_COMPONENT.t(s);
    }
}
