package charite.christo.strap;
import charite.christo.*;
import charite.christo.protein.*;
import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.lang.ref.*;
import java.util.*;
import java.util.List;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static java.awt.event.MouseEvent.*;
/**
A label for protein objects e.g. for JTree.
   @author Christoph Gille
*/
public class ProteinLabel extends ChLabel {
    private transient Reference<StrapProtein> refProtein;
    private StrapProtein p() { return (StrapProtein)deref(refProtein);}
    private  static Font fnt;
    public final static long VERTICAL=1<<1, ALWAYS_CONTEXT_MENU=1<<2;
    private long options;
    public void setProtein(Protein p) { refProtein=p!=null ? p.WEAK_REF : null; }
    public ProteinLabel(Protein p, long options) {
        setProtein(p);
        this.options=options;
        UIupdateList.add(this,UIupdateList.PROTEIN_AT_CURSOR);
        UIupdateList.add(this,UIupdateList.PROTEIN_LABEL);
        if (fnt==null) fnt=getFont().deriveFont(Font.BOLD);
        setFont(fnt);
        bg(Color.WHITE);
    }

    @Override public Dimension getPreferredSize() {
        final StrapProtein p=p();
        if (p==null) return dim(0,0);
        final boolean isVertical=0!=(this.options&VERTICAL);
        final int iw=p.getIcon()!=null ? ICON_HEIGHT : 0;
        final CharSequence rt=getText();
        final int sw=strgWidth(this,rt);
        final int w=maxi(iw+sw+2,ICON_HEIGHT*2);
        return dim(isVertical ? ICON_HEIGHT : w, isVertical ? w: ICON_HEIGHT);
    }

    @Override public void paintComponent(Graphics g_){
        final StrapProtein p=p();
        if (p==null) return;
        final Graphics2D g=(Graphics2D) g_;
        final int height=getHeight();
        final int width=getWidth();
        final boolean isVertical=0!=(this.options&VERTICAL);
        final int h=!isVertical ? height : mini(ICON_HEIGHT,width);
        final Rectangle cb=charBounds(fnt);
        final Color bg=gcp(ChRenderer.KEY_NOT_EDITABLE_BG,this,Color.class);
        g.setColor(bg!=null ? bg: Color.WHITE);
        if (isVertical) {
            g.fillRect(width-1,0,1,height);
            g.fillRect(0,height-1,width,1);
            g.rotate(Math.PI/2,h/2f,h/2f);
        } else if (bg!=null) g.fillRect(0,0,width,height);

        antiAliasing(g,false);
        final Image image=p.getIconImage();
        if (image!=null)   g.drawImage(image,0,0,h , h,this);
        final String txt=getText();
        final int y=h/2-cb.height/2;
        final int xTxt=image!=null ? ICON_HEIGHT : 0;
        p.paintChain(Protein.PAINT_CHAIN_UNDERLINE, g, txt, xTxt,y,cb.width,cb.height);
        g.setColor(getForeground());  g.drawString(txt,xTxt,y+cb.y);
        final StrapAlign align=StrapAlign.getInstance();
        if (align!=null && p==align.getProteinAtCursor()) ChIcon.drawCursor(g,32,8,16,16,this);
        ChRenderer.paint(this,g);
        if (!p.isInAlignment()) {
                g.setColor(C(0x646464,100));
                g.fillRect(1,1,9999,999);
        }
    }
    @Override public String getText() {
        final StrapProtein p=p();
        if (p==null) return null;
        final String t=p.getRendTxt();
        return t!=null?t:"";
    }
    @Override public void processEvent(AWTEvent ev) {
        final int id=ev.getID();
        final StrapProtein p=p();
        final Object q=ev.getSource();
        if (p==null) return;
        if (id==MOUSE_PRESSED  && isCtrl(ev)) {
            final List v=StrapAlign.getSelectedObjectsV();
            if (v.contains(p)) v.remove(p);
            else v.add(p);
            new StrapEvent(this,StrapEvent.OBJECTS_SELECTED).setParameter(Protein.class).run();
        }
        if (isPopupTrigger(ev)|| id==MOUSE_CLICKED && 0!=(options&ALWAYS_CONTEXT_MENU)) {
            //StrapAlign.showContextMenu('P',oo(p),ev);
            ContextObjects.showContextMenu('P', oo(p), NO_OBJECT, (Component)q, EM, 0);
        }
        if (id==MOUSE_DRAGGED) {
          if (!doingDnd) {
              exportDrg(this , ev, TransferHandler.COPY);
              doingDnd=true;
          }
        } else doingDnd=false;
        super.processEvent(ev);
    }
    { this.enableEvents(MOUSE_EVENT_MASK|MOUSE_MOTION_EVENT_MASK);}
    @Override public String getToolTipText(MouseEvent ev) {
        if (ev.getX()>ICON_HEIGHT) return null;
        final StrapProtein p=p();
        final StringBuffer sb=p!=null ?  p.getTooltipWithoutHtmlTags(isShift(ev)) : null;
        return sb!=null ? addHtmlTagsAsStrg(sb) : null;
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
   /* >>> DnD  >>> */
  private boolean doingDnd;
  public List<File> getDndDateien() {
        final StrapProtein p=p();
        final List v=new ArrayList();
        v.clear();
        if (p!=null) adAll(p.getDndFiles(),v);
        return v;
    }

}
