package charite.christo.strap;
import charite.christo.*;

import charite.christo.protein.*;
import java.awt.*;
import javax.swing.JComponent;
import javax.swing.ListModel;

import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
   A JList to select proteins.
   @author Christoph Gille JList
*/
public class ProteinList extends ChJList {
    public final static long NT=1, CALPHA=1<<1, PDB_ID=1<<2, EC=1<<3, IN_ALI=1<<4, NOT_IN_ALI=1<<5,
        ACTGN=1<<6, NT_or_ACTGN=1<<7, NOT_ACTGN=1<<8, NO_CALPHA=1<<9,
        IN_3D=1<<10, CHAINS=1<<11, SECSTRU=1<<12,
        SOLVENT=1<<21, RES_ANNO=1<<22, RES_SEL=1<<23, SEQ_FEATURE=1<<24,
        NONE=1<<25, HETERO3D=1<<26, NUC3D=1<<27, INFERRED_3D=1<<28;

    private final StrapAlign align;
    private long filter;
    private static ProteinComboModel model;
    private final ChListSelectionModel selModel=new ChListSelectionModel(this);
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Constructor >>> */
    public ProteinList(StrapAlign a, long filter) {
        super(model=new ProteinComboModel(a,filter), ChJTable.ICON_ROW_HEIGHT|ChJTable.FILE_TRANSFER_HANDLER|ChJTable.DRAG_ENABLED);
        this.filter=filter;
        align=a;
        setMinimumSize(dim(4*EM,4*EX));
        setProcessEvt(align.getTree(),this);
        a.newDropTarget(this,false);
        setSelectionModel(this.selModel);
    }
    public void setFilter(long filter) {this.filter=filter;}
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Paint >>> */
    private boolean isInitialized;
    @Override public void paintComponent(Graphics g) {
        antiAliasing(g,false);
        if (!isInitialized) {
            isInitialized=true;
            setModel(model);
            setBackground(Color.WHITE);
            setCellRenderer(new ChRenderer().setEnabled(this));
            UIupdateList.add(this,UIupdateList.PROTEIN_LABEL);
            UIupdateList.add(this,UIupdateList.PROTEIN_AT_CURSOR);
            UIupdateList.add(this,UIupdateList.PROTEIN_ORDER);
            rtt(this);
        }
        super.paintComponent(g);
        final int cp=align.getProteins().length;
        if (cp==0) {
            String msg[]={ANSI_FG_BLUE+" This is the protein selector","which allows to select several proteins.","Currently no protein is present in STRAP."};
            drawMsg(msg,this,g,0);
        } else {
            final int y=cp*ICON_HEIGHT+EX;
            final int h=getHeight();
            if (y+3*EX<h) {
                String msg[]={ANSI_FG_BLUE+"Note: more than one proteins can be selected","by left-click with shift or ctrl or by typing ctrl-A."};
                drawMsg(msg,this,g,h-2*EX);
            }
        }
    }
    @Override public void paint(Graphics g) { try {super.paint(g);} catch(Throwable e) {} };

    /* <<< Paint <<< */
    /* ---------------------------------------- */
    /* >>> selected proteins >>> */
    final public StrapProtein[] getSelectedProteins() {
        final Set set=selModel.getSet();
        set.retainAll(align.vProteins);
        final StrapProtein pp[]=align.getVisibleProteins().clone();
        for(int i=pp.length; --i>=0;) if (!set.contains(pp[i])) pp[i]=null;
        return rmNull(pp,StrapProtein.class);
    }
    public ProteinList selectAll(int filter) {
        clearSelection();
        if (filter!=NONE) {
            final Set set=selModel.getSet();
            set.clear();
            final ListModel m=this.model;
            for(int i=sze(m);--i>=0;) {
                final StrapProtein p=(StrapProtein)model.getElementAt(i);
                if (p!=null && isEnabled(p,filter)) set.add(p);
            }
            if (isInitialized) selModel.fireValChanged();
        }
        return this;
    }
    final public StrapProtein[] getSelectedOrAllProteins() {
        final StrapProtein pp[]=getSelectedProteins(),ppVis[]=align.getVisibleProteins();
        return pp==null||pp.length<1 && ChMsg.noProtsSelectedButWantTakeAll(ppVis.length) ? ppVis : pp;
    }
    final public StrapProtein[] getSelectedProteinsAtLeast(int atLeast) {
        final StrapProtein pp[]= getSelectedOrAllProteins();
        if (pp.length<atLeast) {
            StrapAlign.errorMsg("Select at least "+atLeast+" proteins"); return null;}
        return pp;
    }

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Scrollpane, Layout >>> */

    @Override public Dimension getPreferredSize() {
        try {
            return dim(super.getPreferredSize().width,align.getProteins().length*ICON_HEIGHT);
        } catch (Exception e){return dim(4*EM,4*EX);}
    }

    public JComponent scrollPane() {
        final OverlayScrollPane sp=new OverlayScrollPane(this).addStringMatchButton(null, 0L,"Proteins");
        sp.setBorder(new javax.swing.border.EtchedBorder());
        return sp;
    }
    @Override public boolean getScrollableTracksViewportWidth() { return true;}
    /* <<< Scrollpane <<< */
    /* ---------------------------------------- */
    /* >>> static utils >>> */
    @Override public boolean isEnabled(Object o) {
        if (!(o instanceof StrapProtein)) return false;
        return super.isEnabled(o) && (filter==0 || isEnabled((StrapProtein)o,filter));
    }

    public static boolean isEnabled(final StrapProtein pp[], final long flags) {
        if (pp!=null) for(StrapProtein p:pp) if (isEnabled(p,flags)) return true;
        return false;
    }

    public static boolean isEnabled(final StrapProtein p, final long flags) {
        if (p==null) return false;
        boolean e=true;
        if ((flags & NT)!=0 && p.getNucleotides()==null) e=false;
        if ((flags & ACTGN)!=0 && !p.containsOnlyACTGNX()) e=false;
        if ((flags & NT_or_ACTGN)!=0 && (p.getNucleotides()==null && !p.containsOnlyACTGNX())) e=false;
        if ((flags & NOT_ACTGN)!=0 && p.containsOnlyACTGNX()) e=false;
        if ((flags & NO_CALPHA)!=0 && p.hasCalpha()) e=false;
        if ((flags & CALPHA)!=0 && !p.hasCalpha()) e=false;
        if ((flags & EC)!=0 && p.getEC()==null) e=false;
        if ((flags & PDB_ID)!=0 && p.getPdbID()==null) e=false;
        if ((flags & IN_ALI)!=0 && !p.isInAlignment()) e=false;
        if ((flags & CHAINS)!=0 && p.getChainsAsString().length()<2) e=false;
        if ((flags & NOT_IN_ALI)!=0 && p.isInAlignment()) e=false;
        if ((flags & IN_3D)!=0 && p.getProteinViewers((Class)null).length==0) e=false;
        if ((flags & SECSTRU)!=0 && p.getResidueSecStrType()==null) e=false;
        if ((flags & SOLVENT)!=0 && p.getResidueSolventAccessibility()==null) e=false;
        if ((flags & INFERRED_3D)!=0 && p.getInferred3dCountMatches()==null) e=false;
        if ((flags & (RES_SEL|RES_ANNO|SEQ_FEATURE))!=0) {
            final ResidueSelection ss[]=p.getAllResidueSelections();
            if ((flags & RES_SEL)!=0     && ResidueSelectionUtils.countType('S',ss)==0) e=false;
            if ((flags & RES_ANNO)!=0    && ResidueSelectionUtils.countType('A',ss)==0) e=false;
            if ((flags & SEQ_FEATURE)!=0 && ResidueSelectionUtils.countType('F',ss)==0) e=false;
        }
        if ((flags & (HETERO3D|NUC3D))!=0) {
            int countHet=0, countNuc=0;
            for(HeteroCompound h : p.getDnaAndRnaAndHeteros()) if (h.isNucleotideChain()) countNuc++; else countHet++;
            if ((flags & HETERO3D)!=0 && countHet==0) e=false;
            if ((flags & NUC3D)!=0 && countNuc==0) e=false;
        }

        return e;
    }
}
