package charite.christo.strap;
import charite.christo.*;
import charite.christo.protein.*;
import charite.christo.interfaces.*;
import javax.swing.*;
import java.awt.Component;
import java.util.*;
import java.awt.event.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
The menu for protein viewers (Pymol, JMol ... ).
   @author Christoph Gille
*/
public class ProteinViewerPopup implements ActionListener {
    private final ContextObjects objects;
    public final static String BUT_ASTEX="Full featured 3D-view";
    private final JLabel labTitle=new JLabel();
    private JPopupMenu pop;
    public ProteinViewerPopup(ContextObjects o) {objects=o;}
    public JPopupMenu menu() {
        ProteinViewer[] vv=objects.proteinViewers();
        labTitle.setText("Context menu for "+vv.length+" protein viewer"+(vv.length>1 ? "s":""));
        if (pop==null) {
            final Object oo[]={
                labTitle,
                "This Menu has been generated by STRAP",
                "-",
                m('P',null),
                m('v', null),
                m('S',IC_SUPERIMPOSE),
                new Object[]{
                    "Tools",
                    b(IC_CAMERA),
                    new ChButton("CTRL").t("Control panel").i(IC_CONTROLPANEL).li(this),
                    new ChButton("SCRIPT").t("Script panel").i(IC_EDIT).li(this),
                    new ChButton("CUST").t("Customize").i(IC_CUSTOM).li(this)
                },
                new Object[]{
                    "Residue Selection",
                    ResidueSelectionPopup.b('3')
                },
                m('H',null)
            };
            pop=jPopupMenu("3D-viewer-contextmenu",oo);
        }
        enableDisable();
        return pop;
    }

    /* >>> Menu >>> */
    /* ---------------------------------------- */
    /* <<< ActionListener  <<< */
    public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        final Object q=ev.getSource();
        if (cmd==ACTION_MENU_SELECTED) {
            enableDisable();
            return;
        }
        final int modi=modifiers(ev);
        final boolean isSelctd=isSelctd(q);
        Protein3d p3d=null;
        final List<Protein> vAllPP=new ArrayList();
        for(ProteinViewer pv : objects.proteinViewers()) {
            final Class clazz=pv.getClass();
            final Component panel=pv.getViewerPanel();
            if (panel instanceof Protein3d) p3d=(Protein3d)panel;
            final Protein pp[]=ProteinViewerUtils.getProteinsInView(pv);
            adAllUniq(pp,vAllPP);
            if (cmd=="SUPER_GO") {
                StrapAlign.dispatchLater(StrapEvent.COMPUTATION_STARTED,333);
                final String msg="It seems that a superposition job is already running.<br>Start new job anyway?";
                if (sze(pp)<2) error("Superposition requires at least \n"+
                                     "two proteins in the same 3D-view.\n\n"+
                                     "But only "+sze(pp)+
                                     (" protein  is in this view\n\n"+
                                      "Use WIKI:Drag_and_drop to load \n"+
                                      "at least one more protein.\n\n"+
                                      "Watch movie "+MOVIE_Drag_protein_to_3D));
                else if (!StrapProteinUtils.IS_SUPERIMPOSING[0] || ChMsg.yesNo(msg)) {
                    final ArrayList v=new ArrayList();
                    for(Protein p : pp) {
                        Matrix3D m3d=p.getRotationAndTranslation();
                        pcp("SUPER_UNDO",m3d!=null ? m3d : new Matrix3D(),p);
                        if (p.getResidueCalpha()!=null) v.add(p);
                    }
                    StrapProteinUtils.threadSuperimposeProteins(v,CAN_BE_STOPPED, StrapProteinUtils.SUPERIMP_EVENT).start();
                }
            }
            for(Protein p : pp) {
                if (cmd=="SUPER_UNDO" || cmd=="SUPER_HOME") {
                    p.setRotationAndTranslation(cmd=="SUPER_HOME" ? null : gcp("SUPER_UNDO",p,Matrix3D.class));
                    new StrapEvent(Protein3d.class,StrapEvent.PROTEIN_3D_MOVED).setParameters(pp).run();
                }
            }
            if (cmd=="SUPER_STOP") stopAllAlignments();
            if (cmd==CLASS_DialogNeighbors) StrapProteinUtils.structureNeighbors(pp);
            if (cmd==IC_CAMERA) {
                final java.io.File f=file(STRAPOUT+"/images/"+niceShrtClassNam(clazz)+".png");
                writePng(panel,f);
                visitURL(f, modi);
            }
            if (cmd=="INFO") showHelp(clazz==Protein3d.PView.class ? Protein3d.class : clazz);
            if (cmd=="H") showHelp(ProteinViewerUtils.class);
            if (cmd==IC_JAVA) showJavaSource(clazzName(pv),(String[])null);
            if (cmd=="CTRL") ProteinViewerUtils.showControlPanel(pv);
            if (cmd=="CUST") for(Customize c : Customize.cc(pv)) Customize.addDialog(c);
            //if (cmd==IC_CLOSE) pv.dispose();
            if (cmd=="ADD") ChMsg.msgDialog(vBpnl("To add proteins to the view please use Drag-and-Drop",
                                                "Simply drag a protein label or protein file into the view.",
                                                      WATCH_MOVIE+MOVIE_Drag_protein_to_3D));
            if (cmd=="EXPORT") {
                Object msg=gcp(cmd,this);
                if (msg==null) {
                    msg=vB(
                           "<br><b><u>Export using Drag-and-Drop</u></b><br>"+
                           "To export proteins use Drag-and-Drop:<br>"+
                           "Start dragging the mouse on a protein label and drop the protein on the desktop.<br>"+
                           "Protein labels contain the protein name and maybe a protein icon.<br>"+
                           "They are found in the alignment panel or with above menu item \"List proteins\".",
                           WATCH_MOVIE+MOVIE_Export_Proteins,
                           "<br><b><u>Export using the context menu</u></b><br>"+
                           "Alternatively, you can use the export-function in the context menu.",
                           StrapAlign.b(CLASS_DialogExportProteins),
                           WATCH_MOVIE+MOVIE_Context_Menu);
                    pcp(cmd,msg,this);
                }
                ChFrame.frame(cmd, msg, CLOSE_ON_CtrlW).shw(ChFrame.AT_CLICK|ChFrame.PACK);
                putln("nachricht");

            }
            if (cmd=="LIST_P") {
                final String t="Proteins in "+niceShrtClassNam(pv);
                StrapAlign.showInJList(ChJTable.CHJTABLE_CLEAR, mkIdObjct(t,pv), ProteinViewerUtils.getProteinsInView(pv), t, null).showInFrame(ChFrame.AT_CLICK,t);
            }

            if (cmd=="RM") {
                final ProteinViewer vv[]=ProteinViewerUtils.getViewersSharingView(pv);
                if (vv.length==1) {
                    if (ChMsg.yesNo("Close 3D-viewer "+vv[0].getProtein())) ProteinViewerUtils.disposeProteinViewer(vv[0]);
                } else if (vv.length>0) {
                    final ArrayList vP=new ArrayList();
                    for(ProteinViewer v : vv) adUniq(weakRef(v.getProtein()),vP);
                    final ChJList list=new ChJList(vP,ChJTable.DEFAULT_RENDERER);
                    if (ChMsg.yesNo(newCNSEW(list,"Please select the proteins to be removed.","Remove proteins from 3D-view?",null,null))) {
                        final int ii[]=list.getSelectedIndices();
                        for(int i=sze(ii); --i>=0;) ProteinViewerUtils.disposeProteinViewer(vv[i]);
                    }

                    //                    if (ProteinViewerUtils.getViewersSharingView(vv[0]).length==0 && ChMsg.yesNo("Close empty 3D-View?")) rmFromParent(vv[0].getViewerPanel());
                }
            }
            if (cmd=="PV" || cmd=="BB_FAVORITE_V" || cmd=="BB_ASTEX") {
                final Class c3d=HotswapUtil.getClass(cmd=="BB_ASTEX" ? CLASS_ChAstexPROXY :  gcp(ProteinViewer.class,q));
                if (sze(pp)>5 && !ChMsg.yesNo("Are you sure that you want to see "+sze(pp)+" proteins in "+niceShrtClassNam(c3d)+
                                              ".<br>Showing too many protein 3D-structures may overload the Java session.")) continue;
                final ProteinViewer pvNew=c3d!=null ?  ProteinViewerUtils.open3dViewer(pp, c3d, ProteinViewerUtils.OPEN_INIT_SCRIPT|ProteinViewerUtils.OPEN_SEND_ALL_ANNOTATIONS) : null;
                if (pv instanceof Protein3d.PView) {
                    final ProteinViewer vvNew[]=ProteinViewerUtils.getViewersSharingView(pvNew);
                    boolean isPreview=false;
                    for(int pass=0;pass<2; pass++) {
                        for(ProteinViewer v0 : ProteinViewerUtils.getViewersSharingView(pv)) {
                            final Protein p0=v0.getProtein();
                            final Matrix3D m3d=((Protein3d.PView)v0).getTransformationPreview();
                            for(ProteinViewer v : vvNew) {
                                if (v==null || v.getProtein()!=p0) continue;
                                if (pass==0) isPreview=isPreview||(m3d!=null && !m3d.isUnit());
                                else if (isPreview) {
                                    v.transform(m3d);
                                    pcp(ProteinViewerUtils.KEY_IS_PREVIEW,"",v);
                                }
                            }
                        }
                    }
                }
            }
            if(cmd=="SCRIPT") {
                ChScriptPanel script=scriptPanels.get(clazz);
                if (script==null) {
                    final String[] vCmd=ProteinViewer.COMMANDS.clone();
                    for(int i=vCmd.length; --i>=0;) if (pv!=null && !pv.getSupportsCommand(vCmd[i])) vCmd[i]=null;

                    final Object
                        wc[]={
                        vCmd,
                        dirWorking(),
                        StrapAlign.getInstance().vProteins,
                        splitTokns(0L, "color red orange brown yellow green blue cyan magenta purple white lightgrey grey darkgrey black "),
                        splitTokns(0L, (String)invokeMthd("getWordCompletionList",clazz))
                    };
                    script=new ChScriptPanel(0L, clazz.getName(), wc, url(TabItemTipIcon.homePage(clazz)), ProteinViewerUtils.INSTANCE);
                    scriptPanels.put(clazz,script);
                }
                pcp(ProteinViewer.class, weakRef(pv),script);
                script.showFrame("Script "+niceShrtClassNam(clazz));
            }
            if (cmd==ProteinViewer.COMMANDshowScriptPanel) ProteinViewerUtils.sendCommand(cmd,"",pv,0L);
            if (cmd==ProteinViewer.COMMANDcenter && panel!=null) {
                final String id=ProteinViewerUtils.mapLastSelection.get(panel);
                if (ProteinViewerUtils.sendCommand(cmd,id,pv,0L)) panel.repaint();
            }
        }

        if (cmd=="MBT" && !Insecure.EXEC_ALLOWED) {
            final List vID=new ArrayList();
            for(Protein p : toArry(vAllPP,Protein.class)) adUniq(withoutLastComponent(delPfx("PDB:",p.getPdbID()),'_'),vID);
            if (sze(vAllPP)==0) error("No PDB ID");
            else {
                final ChCombo comboPdbID=new ChCombo(vID,0);
                final Object msg=vBhB("Starting MBT viewer. see",
                                      "http://mbt.sdsc.edu/",
                                      "This requires a modern computer and fast internet.",
                                      "PCs with older video cards may crash.",
                                      "Very large proteins may cause problems.",
                                      hB("PDB ID ", comboPdbID));
                final int choice=ChMsg.option(msg,new String[]{"Simple Viewer", "Protein Workshop","Cancel"});
                final String pdbID=toStrg(comboPdbID), view=choice==0?"SV":choice==1?"PW":null;
            if (view!=null) visitURL("http://www.pdb.org/pdb/Viewers/RCSBViewers/view.jsp?viewerType="+view+"&structureId="+pdbID+"&structIdFromStrutsAction="+pdbID,0);
           }
        }
        if (p3d!=null) {
            if (cmd=="BB_DIFF_COLOR_CHAIN" || cmd=="BB_DIFF_COLOR") {
                p3d.eachProteinInDifferentColor(cmd=="BB_DIFF_COLOR_CHAIN",false);
                p3d.enableDisable();
            }
            if (cmd=="BB_RESET")  { p3d.getRotationAndTranslationOfView().unit().scale(5f); p3d.repaint(); }
            if (cmd=="BB_HIDE_DIST_HET") p3d.setHideDistantDNS(isSelctd);
        }
        menu().setVisible(false);
    }
    private final static HashMap<Class,ChScriptPanel> scriptPanels=new HashMap();
    /* <<< ActionEvent <<< */
    /* ---------------------------------------- */
    /* >>> Menu >>> */
    private Object[] m(char id, String icon) {
        if (id=='S') {
            return
                new Object[]{
                icon!=null ? "i" : null,
                icon,
                "Superimpose",
                b("SUPER_GO"), b("SUPER_STOP"), b("SUPER_UNDO"), b("SUPER_HOME"),
                b(CLASS_DialogNeighbors).i(IC_SEARCH)
            };
        }
        if (id=='P') return new Object[]{"Protein", b("ADD"), b("LIST_P"), b("EXPORT"),b("RM")};
        if (id=='H') return new Object[]{"Help",b("INFO"),b("H")};
        if (id=='V' || id=='v') {
            final Object vv[]=StrapPlugins.allClassesV(ProteinViewer.class).asArray(),
                ii[]=new Object[vv.length+4];
            if (id=='v' && vv.length==1) return null;
            ii[0]="Alternative protein Viewers";
            ii[1]="The proteins are opened in an external viewer";
            ii[2]="i";
            ii[3]=IC_JMOL;
            for(int i=vv.length; --i>=0;) {
                final Object c=vv[i];
                if (c==null) continue;
                ChButton b=new ChButton("PV").t("Open in "+niceShrtClassNam(c)).i(TabItemTipIcon.imageIcon(c)).li(this);
                pcp(ProteinViewer.class,c, b);
                ii[i+4]=b;
            }
            return ii;
        }
        return null;
    }
    /* <<< Menu <<< */
    /* ---------------------------------------- */
    /* >>> Button >>> */
    private final static Map<String,ChButton> mapButtons=new HashMap();
    protected ChButton b(String id) {
        ChButton b=mapButtons.get(id);
        if (b==null) {
            String t=null, tt=null, clazz3D=null;
            Object i=null;
            boolean toggle=false;
            if (id=="INFO") t="Information on this Protein viewer";
            if (id=="H") t="Using protein 3D viewers in STRAP";
            if (id==IC_JAVA) { t="Java source"; i=IC_JAVA; }
            if (id==IC_CAMERA) { t="Capture image (print)"; i=IC_CAMERA; }
            if (id=="RM") { t="Remove proteins / close view ...";  i=IC_CLOSE; }
            if (id=="LIST_P")  t="List proteins";
            if (id=="ADD") { t="Add proteins to the view"; i=IC_CURSOR_DRAG; }
            if (id=="EXPORT") { t="Export proteins"; i=IC_CURSOR_DRAG; }
            //if (id==IC_CLOSE) { t="Close"; i=IC_CLOSE; }
            if (id=="SUPER_GO") {
                t="Superimpose all proteins in this view";
                tt="All structures will be superimposed upon the optimal structure (OS)<br>"+
                    "The OS is the one where the sum of scores to the others is maximal.<br><br>"+
                    "To superimposed only fewer proteins,  open another 3D-view with only the proteins under consideration.<br><br>"+
                    "The Computation is interupted when the view is closed.<br>"+
                    "An 'Undo' button will appear";
                i=IC_SUPERIMPOSE;
            }
            if (id=="SUPER_STOP") { t="Stop all superposition processes"; }
            if (id=="SUPER_UNDO") { t="Reset coordinates to what it was before superposition"; }
            if (id=="SUPER_HOME") {
                t="Reset coordinates";
                tt="Reset coordinates to home the original coordinates as recorded in the protein file";
            }
            if (id=="BB_HIDE_DIST_HET") {
                t="Hide nucleotide chains and heteros that are distant from current peptide chain";
                toggle=true;
            }
            if (id=="BB_DIFF_COLOR") {
                t="Each protein in a different color";
                tt="Each protein color according to the chain ID using JMol colors.<br>SHIFT: Each in a different color indipendent of the chain ID.";
            }
            if (id=="BB_DIFF_COLOR_CHAIN") {
                t="Each protein in a different color according to chain";
            }
            if (id=="BB_RESET") {
                t="Revert original orientation";
                tt="Orientation as at the beginning";
            }
            if (id==ProteinViewer.COMMANDcenter) { t="Center"; tt="Center view to last selection/location";}
            if (id==ProteinViewer.COMMANDshowScriptPanel) { t="Script-Panel"; tt="Show Script Panel";}
            if (id=="BB_ASTEX") clazz3D=CLASS_ChAstexPROXY;
            if (id=="BB_FAVORITE_V" && getPrgParameters()!=null) {
                for(String s0 : getPrgParameters()) {
                    final String s=delPfx("-prefere",s0);
                    if (s==s0) continue;
                    for(Object c : StrapPlugins.allClassesV(ProteinViewer.class).asArray()) {
                        final String cn=shrtClasNam(c);
                        if (cn!=null && (s.equalsIgnoreCase(cn)) || s.equalsIgnoreCase(niceShrtClassNam(c))) clazz3D=cn;
                    }
                }
                if (t==null) return null;
            }

            if (clazz3D!=null) {
                final String nice=niceShrtClassNam(clazz3D);
                t="View in "+nice;
                i=TabItemTipIcon.imageIcon(clazz3D);
                tt=nice+" is an advanced 3D-viewer";
            }
            b=(toggle ? toggl(id) : new ChButton(id)).t(t).tt(tt).i(i).li(this);

            if (id==CLASS_DialogNeighbors)  b.tabItemTipIcon(CLASS_DialogNeighbors);

            if (clazz3D!=null) pcp(ProteinViewer.class, clazz3D, b);
            mapButtons.put(id, b);
        }

        return b;
    }
    private void enableDisable() {
        boolean hasMoved=false, undoSuper=false, viewRotated=false, hasHetero=false, hasNuc=false;
        boolean multiProt=false, noJogl=false, hasCtrlPnl=false, hasCust=false;
        int countP=0;
        for(ProteinViewer pv : objects.proteinViewers()) {
            final long flags=pv.getPropertyFlags();
            for(Protein p : ProteinViewerUtils.getProteinsInView(pv)) {
                final Matrix3D m3d=p.getRotationAndTranslation();
                hasMoved=hasMoved || m3d!=null && !m3d.isUnit();
                undoSuper=undoSuper || gcp("SUPER_UNDO",p)!=null;
                hasHetero=hasHetero || p.getHeteroCompounds().length>0;
                hasNuc=hasNuc || p.getDnaAndRnaStructures().length>0;
                countP++;
            }
            noJogl=noJogl ||  0==(flags&ProteinViewer.PROPERTY_OPENGL);
            hasCtrlPnl=hasCtrlPnl || getCtrlPanel(pv)!=null;
            multiProt=multiProt || 0!=(flags&ProteinViewer.PROPERTY_MULTI_MOLECULE);
            hasCust=hasCust || sze(Customize.cc(pv))>0;
            if (pv instanceof Protein3d.PView) {
                Protein3d p3d=((Protein3d.PView)pv).getViewerPanel();
                viewRotated=viewRotated|| p3d!=null && !p3d.getRotationAndTranslationOfView().isUnit();
            }
        }
        b("CTRL").setEnabled(hasCtrlPnl);
        b("CUST").setEnabled(hasCust);
        b(IC_CAMERA).setEnabled(noJogl);
        b("SUPER_UNDO").setEnabled(undoSuper);
        b("SUPER_HOME").setEnabled(hasMoved);
        b("SUPER_STOP").setEnabled(StrapProteinUtils.IS_SUPERIMPOSING[0]);
        b("RM").setEnabled(countP>0);
        b("BB_DIFF_COLOR_CHAIN").setEnabled(countP>0);
        b("BB_DIFF_COLOR").setEnabled(countP>0);
        b("BB_RESET").setEnabled(viewRotated);
        b("BB_HIDE_DIST_HET").setEnabled(hasHetero || hasNuc);
        b("ADD").setEnabled(multiProt);

    }
    /* <<< Button <<< */
    /* ---------------------------------------- */
    /* >>> ProteinViewer >>> */
    /* <<< ProteinViewer <<< */
    /* ---------------------------------------- */
    /* >>> MenuBar 3D >>> */
    private Object menusBB[][];
    private Object[][] getMenusBB() {
        if (menusBB==null) {
            final String msgToolPanel=
                "This can be done in the tool-panel.<br>"+
                "It is opened with the \"List proteins\"-toggle in the lower right corner.";
            menusBB=new Object[][]{
                m('P',null),
                new Object[]{
                    "View",
                    b("BB_ASTEX"),
                    b("BB_FAVORITE_V"),
                    m('V',null),
                    new Object[]{
                        "i", IC_COLOR,
                        "Color",
                        b("BB_DIFF_COLOR"),
                        b("BB_DIFF_COLOR_CHAIN")
                    },
                    new Object[]{
                        "i", IC_FLAVIN,
                        "Hetero compounds",
                        b("BB_HIDE_DIST_HET"),
                        ChButton.actionViewMessage("Show/Hide single hetero compounds",msgToolPanel)
                    },
                    b("BB_RESET"),

                },
                m('S',null),
                new Object[]{
                    "Tools",
                    b(IC_CAMERA),
                    b(CLASS_DialogNeighbors)
                },
                m('H',null)

            };
            undockItms(menusBB);
        }
        return menusBB;
    }

  static void tt(JMenuBar mb){
    for(Component c :mb.getComponents()) {
      if (c instanceof JComponent) ((JComponent)c).setToolTipText("This is the Menu-bar of the 3D-view which appears when you click inside.");
    }
  }
  protected JMenuBar newMenuBar(Protein3d p3d) {
      final JMenuBar mb=new JMenuBar();
    for(Object[] oo : getMenusBB()) {
      final JMenu jm=jMenu(oo,"Protein3d");
      addActLi(this,jm);
      mb.add(jm);
    }
    mb.add(new ChButton("BB_ASTEX", ChButton.NO_MARGIN).t(BUT_ASTEX).li(this));
    if (!Insecure.EXEC_ALLOWED) mb.add(new ChButton("MBT",ChButton.NO_MARGIN).t("MBT ...").li(this));
    mb.add(pnl());
    tt(mb);

    return mb;
  }

}
