package charite.christo.strap;
import charite.christo.protein.*;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.awt.*;
import java.util.*;
import java.util.List;
import java.awt.event.*;
import java.lang.ref.*;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static charite.christo.protein.ProteinViewer.*;
import static java.awt.event.MouseEvent.*;

/**HELP
   <br><br><b>Menu-bar: </b>

   The embedded 3D-views that are shown within the STRAP application
   frame come along with a menu-bar.

   This means that if the user clicks inside the 3D-view, a previous
   active menu-bar disappears and the menu-bar of the 3D-view is shown
   instead.

   The menu-bar contains of 2 parts: First it contains the native menus of this viewer.
   Second there are menus added by Strap such as a "\Style...\" menu.
   <br><br>
   To get back the menu-bar of the STRAP-alignment, the user needs to click into the alignment panel.
   In Web-mode the menu-bar is deactivated by default <i>JCOMPONENT:charite.christo.strap.StrapAlign#cbMenuBar()!</i>.

   <br><br><b>Undock menu items: </b>

   The user may need certain  menu items for changing the style rather often.
   For convenience, these menu items can be undocked by mouse-drag such that they can be directly be used without first opening the menu.

   <br><br><b>Picking Atoms in the 3D-View: </b>

   When atoms are clicked in the 3D-view, Strap will display a
   toolbar with buttons that act upon the clicked amino acid.

   One button is used to set the alignment cursor on the clicked residues.
   Two buttons are used to add the clicked amino acid to a residue selection or to remove the amino acid.

   <br><br><b>Movements of the alignment cursor: </b>

   The alignment cursor position is passed to the 3D viewer and can
   be referred to. This can be used to change the display style of a particular amino acid.

   <br><br><b>Selecting amino acids in the alignment panel: </b>
   Using the mouse or using arrow-left/ arrow-right keys together with shift, a number of amino acids can be selected
   in STRAP. This selection is passed to the 3D viewer.

   <br><br><b>Adding proteins: </b>

   Proteins can be added by WIKI:Drag_and_drop.
   See <i>STRING:ChConstants#MOVIE_Drag_to_another_STRAP</i>

   This is not yet supplorted for JMol.

   <br><br><b>Displaying Sequence Features: </b>

   Sequence Features can highlighted in the 3D-view using  WIKI:Drag_and_drop.
   See <i>STRING:ChConstants#MOVIE_Sequence_Features_in_3D</i>

   @author Christoph Gille

*/
public class ProteinViewerUtils implements ChRunnable, ChProcessEvent {
    /* ---------------------------------------- */
    /* >>> open3dViewer >>> */
    public final static int OPEN_VERBOSE=1, OPEN_ASK=1<<1, OPEN_SEND_ALL_ANNOTATIONS=1<<2, OPEN_NOT_ADD_PANEL=1<<3,
        OPEN_NARROW_OTHERS=1<<4, OPEN_ADD_TAB=1<<5, OPEN_INIT_SCRIPT=1<<6;
    public final static String KEY_IS_PREVIEW="PV$$KIP";
    private final static Object KEY_styleMenu=new Object(), REF_DEPENDS_ON=new Object();

    public static Thread thread_open3dViewer(Protein pp[],Object clazzOrViewer, long options) {
        return threadM("open3dViewer",ProteinViewerUtils.class, new Object[]{pp,clazzOrViewer,longObjct(options)});
    }
    public static ProteinViewer open3dViewer(Protein pp[],Object clazzOrViewer, long options) {
        if (contextObjects==null) contextObjects=new ContextObjects(ONE_VIEWER);
        final Class clazz=HotswapUtil.getClass(clazzOrViewer);
        final ProteinViewer inSameView=deref(clazzOrViewer,ProteinViewer.class);
        final StrapAlign align=StrapAlign.getInstance();
        if (clazz==null || sze(pp)==0 ||
            (options&OPEN_ASK)!=0 && !ChMsg.yesNo("Load "+ (pp.length==1 ? "the protein "+pp[0] : pp.length+" proteins ")+"\n into the protein viewer?")) return null;
        if ((options&OPEN_VERBOSE)!=0) setNextProcessHasLogPanel(true);
        Runnable  hook=null;
        ProteinViewer oneViewer=null;
        JMenu styleMenu=gcp(KEY_styleMenu,inSameView, JMenu.class);
        final Object[] depends={""};
    nextProtein:
        for(Protein p : pp) {
            if (p==null || (p.getResidueCalpha()==null && p.getDnaAndRnaAndHeteros().length==0)) continue;
            if (getViewerForProtein(p, inSameView)!=null) continue;
            final ProteinViewer pv=HotswapUtil.instance(clazz,ProteinViewer.class,true);
            if (pv==null) {
                error("Error in ProteinViewerUtils: cannot instantiate "+clazz);
                return null;
            }
            final long properties=pv.getPropertyFlags();
            if ((properties&ProteinViewer.PROPERTY_EXTERNAL_PROCESS)!=0) {
                StrapAlign.drawMessage("To see details of the external process,\n"+
                                       "hold CTRL when when pressing the button to start the 3D viewer.");
            }
            if ((properties&ProteinViewer.PROPERTY_COMMAND_INTERPRETER)!=0 ) {
                if (styleMenu==null) {
                    pcp(KEY_styleMenu,styleMenu=newStyleMenu(pv),pv);
                }
            }
            if (hook==null) hook=thread_open3dViewer(pp,clazzOrViewer,options);
            final boolean isWire=clazz==Protein3d.PView.class;

            if (pv.setProtein(p, oneViewer!=null ? oneViewer:inSameView, hook, options)) {
                if (0!=(options&OPEN_INIT_SCRIPT)) {
                    for(String s : getCustomSettings(clazz, Customize.CLASS_InitCommands))
                        sendCommand(s,null,pv,0L);
                }
                if (isWire || p.getFileMayBeWithSideChains()!=null)  p.addProteinViewer(pv);
                if (!isWire) {
                    final Matrix3D m3d=p.getRotationAndTranslation();
                    if (m3d!=null && !m3d.isUnit()) pv.transform(m3d);
                    if ( (options&OPEN_SEND_ALL_ANNOTATIONS)!=0) residueSelectionsTo3D(RSto3D_COMMANDS|RSto3D_COLOR,((StrapProtein)p).getResidueAnnotations(), new ProteinViewer[]{pv});
                    new StrapEvent(Dialog3DViewer.class,StrapEvent.PROTEIN_VIEWER_LAUNCHED).setParameter(pv).run();
                }
                if (oneViewer==null) oneViewer=pv;
            }
            final Component pan=pv.getViewerPanel();
            addCP(KEY_DISPOSABLE, pv, pan,5);
            final Object msg=newCNSEW(
                                      "If you came here because you intended to drag a protein with the mouse:<br>"+
                                      "This is not the right way to drag a protein.<br>"+
                                      "You can start dragging a protein  for example from the respective row header of the sequence alignment.<br>"+
                                      "The row-header displays the protein names. Here you can initiater Drag-and-Drop.",
                                      null,pnl(WATCH_MOVIE+MOVIE_Export_Proteins),null,null);
            pcp(KEY_UNDOCK_MESSAGE, msg, pv);
            pcp(KEY_UNDOCK_MESSAGE, msg, pan);
        }
        if (oneViewer!=null) {
            final long flags=oneViewer.getPropertyFlags();
            setFocusedViewer(oneViewer);
            if (0!=(options&OPEN_INIT_SCRIPT)) {
                for(String s : getCustomSettings(clazz, Customize.CLASS_LateCommands)) sendCommand(s,null,oneViewer,0L);
            }
            final Component panel=oneViewer.getViewerPanel();
            if (panel!=null) {
                pcp(KEY_UNDOCKABLE, "", panel);
                final WeakReference ref=new WeakReference(oneViewer);
                closeOnKeystroke(panel, getViewersSharingView(oneViewer), DISPOSE_ON_CtrlW);
                for(Component child : childs(panel, Component.class)) {
                    pcp(ProteinViewer.KEY_PROTEIN_VIEWER, ref, child);
                    StrapAlign.newDropTarget(child,false);
                    child.addMouseListener(LI);
                    child.addMouseMotionListener(LI);
                    child.addKeyListener(LI);
                    ProteinViewer.mapViewer.put(child,ref);
                    closeOnKeystroke(child, getViewersSharingView(oneViewer), DISPOSE_ON_CtrlW);
                }
                panel.addMouseWheelListener(LI);
                pcp(ChTableLayout.RUN_COLUMN_TITLE, ref, panel);
                final Protein3d p3d=deref(panel, Protein3d.class);
                JMenuBar mb=oneViewer.getMenuBar();
                final ProteinViewerPopup pMenu=contextObjects.proteinViewerMenu();
                //final ProteinViewerPopup pMenu=new ProteinViewerPopup(new ContextObjects(new Object[]{oneViewer}));
                if (mb==null && p3d!=null) p3d.setMenuBar(mb=pMenu.newMenuBar(p3d));
                if (mb!=null && gcp(ProteinViewerUtils.class, mb) ==null) {
                    pcp(ProteinViewerUtils.class, "",mb);
                    if (p3d==null) {
                        final JMenu[] menus=childs(mb,JMenu.class);
                        if (sze(menus)>0) {
                            final ChLabel lab=new ChLabel("Native menu ");
                            mb.add(toggl().actionCollapse(false, menus).actionCollapse(true,lab),0);
                            mb.add(lab,0);
                        }
                        if (styleMenu!=null) mb.add(styleMenu);
                        recursiveWhiteOnBlack(mb);
                        removeMenuItems(new String[]{"file"}, new String[]{"close","exit","Close","Exit"}, mb);
                        final String label="Strap";
                        if (oneViewer.getSupportsCommand(COMMANDcenter)) {
                            mb.add(pMenu.b(COMMANDcenter).cln().options(ChButton.NO_MARGIN));
                            mb.add(new JLabel(" "));
                        }
                        if (oneViewer.getSupportsCommand(COMMANDshowScriptPanel))  {
                            mb.add(pMenu.b(COMMANDshowScriptPanel).cln().options(ChButton.NO_MARGIN));
                            mb.add(new JLabel(" "));
                        }
                        final ChButton it=new ChButton(label,ChButton.NO_MARGIN).li(LI).tt("This menu is added by Strap");
                        mb.add(it);
                        mb.add(new JLabel(" "));
                        pcp("Menu",ref,it);
                        final boolean heavy= 0!=(flags&(ProteinViewer.PROPERTY_OPENGL|ProteinViewer.PROPERTY_HEAVY_WEIGHT));
                        if (heavy) {
                            setHeavy(mb);
                            setHeavy(align.getMenuBar());
                        }
                    }
                    mb.add(pnl());
                    pcp(KEY_GET_MENUBAR ,mb, panel);
                    final ImageIcon ic=TabItemTipIcon.imageIcon(oneViewer);
                    if (ic!=null) mb.add(new JLabel(ic), 0);
                    final JComponent panIndent=new JLabel();

                        for(AbstractButton item : childs(mb,AbstractButton.class)) {
                            pcp(KEY_DEPENDS_ON, depends , item);
                            pcp(KEY_DEPENDS_ON, depends , gcp(KEY_CLONED_FROM,item));
                    }
                    mb.add(panIndent,0);
                    mb.add(menuForAccelerators());
                    pcp(ChTableLayout.KEY_OFFSET_COMPONENT, panIndent, panel);
                    ProteinViewerPopup.tt(mb);
                }
                if ((options&OPEN_NOT_ADD_PANEL)==0) viewerToFront(oneViewer, options|OPEN_ADD_TAB);
            }
        }
        return oneViewer;
    }
    private static void setFocusedViewer(ProteinViewer pv) {
        ONE_VIEWER[0]=pv;
    }
    /* <<<  open3dViewer <<< */
    /* ---------------------------------------- */
    /* >>> Picked atoms >>> */
    private static ProteinLabel labPicked;
    private static JLabel labPickedAA;
    private static JComponent toolpane, butRmSel;
    private static ProteinViewer pvPicked;
    private static Protein pPicked;
    private static int aaPicked;
    public static JComponent getToolpane() { return toolpane;}
    protected static void aminoacidPicked(int iA, Protein p, ProteinViewer pv0) {
        final ProteinViewer proxy=AbstractProxy.getProxy(pv0, ProteinViewer.class);
        final ProteinViewer pv=proxy!=null ? proxy : pv0;

        if (labPicked==null) {
            labPicked=new ProteinLabel(null, ProteinLabel.ALWAYS_CONTEXT_MENU);
            //newDropTarget(labPicked
            (labPickedAA=new JLabel()).setForeground(C(0xFFffFF));
            recursiveWhiteOnBlack(labPicked);
            final Object[] bb={null,null,null};
            for(int i=3; --i>=0;) {
                final ChButton b=new ChButton(i==0?"SET_C":i==1?"ADD_S":"RM_S").t(i==0?"Set alignment cursor":i==1?"Add to selection":"Remove from selection");
                setUndockble(b.li(LI)).addMouseListener(LI);
                bb[i]=b;
                if (i==2) (butRmSel=b).setEnabled(false);
            }

            final Dimension d=StrapView.getToolbars()[0].getPreferredSize();
            toolpane=pnl( d.height>1 ? d : null, C(0), labPicked, labPickedAA, bb[0], bb[1], bb[2]);
        }
        final StringBuffer sb=new StringBuffer(40);
        final int rn=p.getResidueNumber(iA);
        if (rn==INT_NAN) sb.append(iA);
        else {
            sb.append(rn);
            final byte ins=p.getResidueInsertionCode(iA);
            if (ins!=0) sb.append((char)ins);
            sb.append(':');
            final byte chain=p.getResidueChain(iA);
            if (chain!=0) sb.append((char)chain);
        }
        labPicked.setProtein(p);
        labPickedAA.setText(sb.toString());
        labPickedAA.setOpaque(true);
        labPickedAA.setBackground(Color.WHITE);
        labPickedAA.setForeground(Color.BLACK);
        ChDelay.bg(labPickedAA,Color.BLACK,444);
        ChDelay.fg(labPickedAA,Color.WHITE,444);
        aaPicked=iA;
        pPicked=p;
        pvPicked=pv;
        if (iA>=0 && pv!=null) {
            final boolean bb[]=new boolean[iA+1];
            bb[iA]=true;
            setSelectedAminoAcidsInViewer(0L, bb, ATOM_ANY, ProteinViewer.SELECTION_PICKED,pv);

        }
        StrapAlign.getInstance().setToolpane(toolpane);
        StrapAlign.startAnimatePosition(0L, p, iA);

    }
    /* <<< Picked atoms <<< */
    /* ---------------------------------------- */
    /* >>> openOrReactivate >>> */
    final static Map<String,Reference<ProteinViewer>> mapProteinViewers=new ChMap(0L,String.class, Reference.class,9);
    /*
      private static String lastViewerID;
      private static ProteinViewer openOrFocusProteinViewer(Protein[] pp, Class clazz, String viewerID) {
      ProteinViewer pv=deref(mapProteinViewers.get(viewerID),ProteinViewer.class);
      if (isViewerActive(pv)) {
      viewerToFront(pv, OPEN_ADD_TAB);
      } else {
      pv=open3dViewer(pp, clazz, OPEN_SEND_ALL_ANNOTATIONS);
      if (pv!=null) {
      Protein p=pv.getProtein();
      if (p!=null) residueSelectionsTo3D(RSto3D_COMMANDS,  ((StrapProtein)p).getResidueAnnotations(), new ProteinViewer[]{pv});
      mapProteinViewers.put(viewerID, weakRef(pv));
      }
      }
      lastViewerID=viewerID;
      return pv;
      }
    */
    /* <<< openOrReactivate  <<< */
    /* ---------------------------------------- */
    /* >>> Disposable >>> */
    private static ArrayList vDisposed=new ArrayList();
    protected static void disposeProteinViewer(ProteinViewer v) {
        if (v!=null) {
            final Object[] depends=gcp(REF_DEPENDS_ON,v,Object[].class);
            if (depends!=null) depends[0]=null;
            final Reference ref=weakRef(v);
            if (!vDisposed.contains(ref)) {
                vDisposed.add(ref);
                assert AbstractProxy.getProxy(v,ProteinViewer.class)==null;
                final Component panel=v.getViewerPanel();
                final Protein p=v.getProtein();
                final ProteinViewer[] sameView=getViewersSharingView(v);
                v.dispose();
                removeCtrlPanel(v);
                if (sameView.length==0 || (sameView.length==1 && sameView[0]==v)) rmFromParent(panel);
                if (p!=null) p.removeProteinViewer(v);
                for(Map.Entry<String,Reference<ProteinViewer>> e : entryArry(mapProteinViewers)) {
                    if (deref(e.getValue())==v) mapProteinViewers.put(e.getKey(),null);
                }
            }
        }
    }
    /* <<< Disposable  <<< */
    /* ---------------------------------------- */
    /* >>> ChRunnable >>> */
    private static boolean cancelRmMenu;
    public Object run(String id, Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[]) arg : null;
        if (id=="RESTORE_T") {
            if (!cancelRmMenu) StrapAlign.getInstance().setMenuBar(null);
        }
        if (id=="viewerToFront") viewerToFront((ProteinViewer)argv[0], atol(argv[1]));
        if (id==ChScriptPanel.RUN_SCRIPT) {
            sendCommand(toStrg(argv[1]), current_selection_name,  (ProteinViewer)gcp(ProteinViewer.class, argv[2]),0L);
        }
        return null;
    }
    /* <<< ChRunnable <<< */
    /* ---------------------------------------- */
    /* >>> Control Panel >>> */
    private static ChTabPane tabbedViewerCtrl;

    protected static void showControlPanel(ProteinViewer pv) {
        if (tabbedViewerCtrl==null) {
            tabbedViewerCtrl=new ChTabPane("E");
            TabItemTipIcon.set("ctrl","","Control panel <br>for protein viewer",IC_JMOL,tabbedViewerCtrl);
        }
        StrapAlign.getInstance().addDialog(tabbedViewerCtrl);
        removeCtrlPanel(pv);
        final Protein p=pv.getProtein();
        final String cn=shrtClasNam(pv);
        final JComponent cp=newCNSEW(getCtrlPanel(pv),p+" in "+cn,null,null,null);
        TabItemTipIcon.set(cn,cn,"Control panel for "+p+" in "+cn,null,cp);
        tabbedViewerCtrl.addTab(cp);
    }
    private static void removeCtrlPanel(ProteinViewer v) {
        final Container ctrl=getCtrlPanel(v);
        if (ctrl!=null && tabbedViewerCtrl!=null) tabbedViewerCtrl.removeTab(parnt(ctrl,JComponent.class));
    }

    private final static Object KEY_PV_CMD=new Object(),KEY_MC_RES_SEL=new Object();
    private static ButColor butColor(Color color, String command, ProteinViewer pv) {
        ButColor bc=new ButColor(ButColor.NO_TRANSPARENCY,color,LI);
        pcp(KEY_PV_CMD,  new Object[]{pv, command}, bc);
        //pcp(KEY_DEPENDS_ON, gcp(REF_DEPENDS_ON,pv), bc);
        return bc;
    }
    private static ChButton cbUpdateSurfaceColors;
    private static JMenu newStyleMenu(ProteinViewer pv) {
        final long properties=pv.getPropertyFlags();
        final JMenu m=new JMenu("Style for all amino acids"), mm[]={m,null,null};
        int count=0;
        for(String s : ProteinViewer.STYLE_COMMANDS) {
            if (!pv.getSupportsCommand(s)) continue;
            count++;
            if (s==ProteinViewer.COMMANDsurface && pv.getSupportsCommand(COMMANDsurface_color)) {
                final String script=ProteinViewer.COMMANDsurface_color+" #COLOR\n"+
                    ProteinViewer.COMMANDsurface+" off\n"+
                    ProteinViewer.COMMANDsurface+" on\n";
                final ChButton butColor=butColor(C(0xFF00), script,pv);
                pcp(ButColor.KEY_FRAME_TITLE,ProteinViewer.COMMANDsurface,butColor);
                final JComponent mi=butColor.mi(ProteinViewer.COMMANDsurface+" on");
                m.add(mi);
                pcp(KEY_FOREGROUND, C(0x00cc00), mi);
            } else styleMenuItem(s+" on",pv,mm);
            styleMenuItem(s+" off",pv,mm);
        }
        if (pv.getSupportsCommand(ProteinViewer.COMMANDcolor)) {
            final ChButton butColor=butColor(C(0xFF00),ProteinViewer.COMMANDcolor+" #COLOR",pv).li(LI);
            final ChJList jl=new ChJList(null,ChJTable.DEFAULT_RENDERER);
            pcp(ProteinViewer.class, weakRef(pv),jl);
            jl.addMouseListener(LI);
            if (cbUpdateSurfaceColors==null) cbUpdateSurfaceColors=toggl("Also update colors of ribbons").s(true);

            if (0!=(PROPERTY_RIBBON_COLOR_NO_CHANGE&properties) && pv.getSupportsCommand(ProteinViewer.COMMANDupdate_color_surfaces)) {
                final ChButton bUpdateColor=new ChButton("STYLE").li(LI);
                pcp(KEY_PV_CMD,  new Object[]{pv, ProteinViewer.COMMANDupdate_color_surfaces}, bUpdateColor);

            }
            pcp(ChJList.class, jl,butColor);
            final JComponent
                pEntireProt=newCNSEW(
                                     scrllpn(0L,jl,dim(EX,4*ICON_HEIGHT)),
                                     hBL("To change the color of the entire protein you need to select a protein:"),
                                     null,
                                     null,null
                                     ),
                pSouth=vBhB(
                            0!=(PROPERTY_RIBBON_COLOR_NO_CHANGE&properties) ? cbUpdateSurfaceColors.cb()  : null,
                            " ",
                            pEntireProt,
                            hBL(toggl(null).actionCollapse(false,new Object[]{jl,pEntireProt}),"Change color of the entire protein"),
                            pEntireProt
                            );

            pcp(ButColor.KEY_PAN_SOUTH,pSouth,butColor);
            m.add(butColor.mi("color"));
        }
        if (mm[1]!=null) m.add(mm[1]);
        if (mm[2]!=null) m.add(mm[2]);
        JMenu mCrude=null;
        for(String s: getCustomSettings(pv.getClass(), Customize.CLASS_Example)) {
            if (mCrude==null) {

                mCrude=new JMenu("<html><body>Please avoid direct<br>commands<br>if possible!</body></html>");
                if (count==0) m.add(mCrude);
                else {
                    final JMenu notRecomended=new JMenu("Direct commands");
                    m.add(notRecomended);
                    notRecomended.add(mCrude);
                }
            }
            styleMenuItem(s,pv,new JMenu[]{mCrude});
        }

        mCrude.add(Customize.newButton(Customize.cc(pv)));
        return m;
    }
    private static void styleMenuItem(String s, ProteinViewer pv, JMenu mm[]) {
        if (s==null) return;
        final ChButton b=new ChButton("STYLE").li(LI).i("1x1");
        pcp(KEY_PV_CMD,  new Object[]{weakRef(pv), s}, b);
        //pcp(KEY_DEPENDS_ON, gcp(REF_DEPENDS_ON,pv), b);
        final int iMenu=
            mm.length<3 ? 0:
            0<strstr(STRSTR_w|STRSTR_IC, "on",s) ? 1 :
            0<strstr(STRSTR_w|STRSTR_IC,"off",s) || 0<strstr(STRSTR_w|STRSTR_IC,"remove",s) ? 2 :
            0;
        if (mm[iMenu]==null) mm[iMenu]=new JMenu(iMenu==1?"On":"Off");
        mm[iMenu].add(b.mi(s));
    }
    /* <<< Control panel <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */
    private static long whenMouseViewer;
    public static ProteinViewer[] getViewersSharingView(ProteinViewer pv) {
        if (pv==null) return ProteinViewer.NONE;
        final ProteinViewer vv[]=rmNull(pv.getViewersSharingView(),ProteinViewer.class);
        return sze(vv)>0 ? vv : new ProteinViewer[]{pv};
    }
    public static ProteinViewer getViewerForProtein(Protein p, ProteinViewer pv) {
        if (p!=null && pv!=null)
            for(ProteinViewer v : getViewersSharingView(pv))
                if (v.getProtein()==p) return v;
        return null;
    }
    public static JMenuBar getMenuBar(Object object) {
        if (object==null) return null;
        if (object instanceof ProteinViewer) return ((ProteinViewer)object).getMenuBar();
        return getMenuBar(deref(ProteinViewer.mapViewer.get(object)));
    }
    protected final static ProteinViewerUtils INSTANCE=new ProteinViewerUtils();
    protected final static ChEventAdapter LI=new ChEventAdapter(INSTANCE);
    private static ContextObjects contextObjects;
    private final static ProteinViewer[] ONE_VIEWER={null};

    public static void viewMenuBar(ProteinViewer pv) {
        final StrapAlign align=StrapAlign.getInstance();
        final JMenuBar menubar=getMenuBar(pv);
        if (menubar!=null && align!=null) {
            final Object panel=pv.getViewerPanel();
            if (align.getFrame()==parentWndw(panel)) {
                final Dimension d=align.getMenuBar().getPreferredSize();
                if (d.height>1) menubar.setPreferredSize(d);
                align.setMenuBar(menubar);
            }
        }
    }

    public AWTEvent processEvent(JComponent component, AWTEvent ev) {
        final int id=ev.getID();
        final Object q=ev.getSource();
        final String cmd=actionCommand(ev);
        final StrapAlign align=StrapAlign.getInstance();
        final boolean ctrl=isCtrl(ev), isContext=isPopupTrigger(ev);
        final int kcode=keyCode(ev), modi=modifiers(ev);
        final ProteinViewer pvMaster=get(0, gcp(KEY_PV_CMD,q), ProteinViewer.class);
        if (pvMaster!=null) {
            final ButColor butColor=q instanceof ButColor ? (ButColor)q : null;
            final Color color=butColor!=null && cmd==ACTION_COLOR_CHANGED ? butColor.getColor() : null;
            final ChJList jlProteins= gcp(ChJList.class, butColor, ChJList.class);
            if (butColor!=null && color==null) jlProteins.setData(getProteinsInView(pvMaster));

            else if (color!=null || cmd=="STYLE" ) {
                final String script=get(1, gcp(KEY_PV_CMD,q), String.class);

                final Object selectedProts[]= color==null || getPrefHeight(jlProteins)==0  ? null : jlProteins.getSelectedValues();
                final long properties=pvMaster==null ? 0L : pvMaster.getPropertyFlags();
                final boolean updateSurf=0!=(PROPERTY_RIBBON_COLOR_NO_CHANGE&properties) && (script!=null && script.indexOf(COMMANDsurface)<0);
                if (sze(selectedProts)>0) {
                    for(int i=sze(selectedProts); --i>=0 && color!=null;) {
                        final ProteinViewer pv1=getViewerForProtein(deref(selectedProts[i], Protein.class),  pvMaster);
                        if (pv1!=null) {
                            sendCommand(ProteinViewer.COMMANDcolor+"  #"+rgbHex(color.getRGB()), null,  pv1, 0L);
                            if (updateSurf) sendCommand(ProteinViewer.COMMANDupdate_color_surfaces, null,  pv1, 0L);
                        }
                    }
                } else if (pvMaster!=null) {
                    ProteinViewer pvScript=pvMaster;
                    for(ProteinViewer pv : getViewersSharingView(pvMaster)) {
                        if (pv!=null && pv.getProtein()==current_selection_protein) pvScript=pv;
                    }
                    boolean success=false;
                    final String lines=color!=null ? rplcToStrg("COLOR", rgbHex(color.getRGB()),script) : script;
                    for(String line : splitLnes(0L, lines))   {
                        success|=sendCommand(line, current_selection_name, pvScript,0L);
                        if (updateSurf && color!=null && isSelctd(cbUpdateSurfaceColors)) sendCommand(ProteinViewer.COMMANDupdate_color_surfaces, null,pvScript,0L);
                    }
                    if (success) ChDelay.repaint(pvScript.getViewerPanel(),1);
                }
            }
        }
        final ProteinViewer pvMenu=gcp("Menu",q,ProteinViewer.class);
        if (pvMenu!=null) {
            if (contextObjects==null) contextObjects=new ContextObjects(ONE_VIEWER);
            setFocusedViewer(pvMenu);
            JPopupMenu jpop=contextObjects.proteinViewerMenu().menu();
            jpop.setLightWeightPopupEnabled(false);
            shwPopupMenu(q, jpop);
        }

        if (id==MOUSE_ENTERED && q instanceof ChJList && gcp(ProteinViewer.class,q)!=null) {
            final ChJList v=(ChJList)q;
            final Protein pp[]=getProteinsInView(gcp(ProteinViewer.class,q,ProteinViewer.class));
            if (sze(v)!=pp.length || !cntainsAll(v,pp)) {
                v.setData(pp);
                v.revalidate();
            }
        }
        if ( (id==MOUSE_PRESSED || id==MOUSE_EXITED || id==MOUSE_ENTERED || id==MOUSE_MOVED) && q!=null) {
            final ProteinViewer pv=deref(ProteinViewer.mapViewer.get(q),ProteinViewer.class);
            if (pv!=null) whenMouseViewer=System.currentTimeMillis();
            final JMenuBar menubar=getMenuBar(q);
            if (menubar!=null) {
                final Object panel=pv!=null?pv.getViewerPanel() : q;
                if (align.getFrame()==parentWndw(panel)) {
                    if (id==MOUSE_PRESSED) viewMenuBar(pv);
                    if (id==MOUSE_EXITED && y(ev)>3 && (modi&(BUTTON1_MASK|BUTTON2_MASK|BUTTON3_MASK))==0) {
                        cancelRmMenu=false;
                        ChDelay.runAfterMS(EDT, INSTANCE, "RESTORE_T", 4444);
                    }
                    if (id==MOUSE_ENTERED) cancelRmMenu=true;
                }
            }
            if (id==MOUSE_PRESSED) {
                align.setToolpane(toolpane);
                if (pv!=null) setFocusedViewer(pv);
            }
        }
        {
            final ProteinViewer pv=pvPicked;
            final Protein p=pPicked;
            final int iA=aaPicked;

            if ((cmd=="ADD_S" || cmd=="RM_S" || cmd=="SET_C") && p!=null && iA>=0) {
                final String name=pv instanceof Protein3d.PView ? "3D-Backbone" : pv!=null ? niceShrtClassNam(pv) : "3D";
                AbstractResidueSelection s=findWithName(name, p.getResidueSelections(),AbstractResidueSelection.class);
                if (q instanceof ChButton && isContext) StrapAlign.showContextMenu('S',oo(s),ev);
                if (cmd=="SET_C") align.setCursorToResidue( p, iA);
                else {
                    if (s==null) {
                        s=new AbstractResidueSelection();
                        s.setName(name);
                        s.setProtein(p);
                        p.addResidueSelection(s);
                        s.setVisibleWhere(0xffffffff);
                    }
                    boolean bb[]=withoutOffset(s.getSelectedAminoacids(), s.getSelectedAminoacidsOffset());
                    if (sze(bb)<=iA) bb=chSze(bb,iA+1);
                    bb[iA]= (cmd=="ADD_S");
                    if (bb[iA]) butRmSel.setEnabled(true);

                    if (countTrue(bb)==0) { bb=null; p.removeResidueSelection(s); }
                    s.setSelectedAminoacids(bb,0);
                    setSelectedAminoAcidsInViewer(0L, bb,ATOM_CA, name,  pv);
                    p.removeResidueSelection(findWithName(ResidueSelection.CURSOR, p.getResidueSelections(),AbstractResidueSelection.class));
                    StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,10);
                }
                StrapAlign.startAnimatePosition(StrapAlign.ANIM_SET_CURS_SCROLL_TO_VISIBLE, p, iA);
            }

            if (kcode==KeyEvent.VK_W && ctrl) {
                final ProteinViewer vv[]=getViewersSharingView(gcp(KEY_PROTEIN_VIEWER,q,ProteinViewer.class));
                if (vv.length>0) {
                    for(ProteinViewer v : vv) v.dispose();
                    rmFromParent(vv[0].getViewerPanel());
                }
            }
        }
        final ProteinViewer pv=ONE_VIEWER[0];
        if (pv!=null) {
            final int wheel=wheelRotation(ev);
            int zoom=100, rotateZ=0;
            if (wheel!=0) {
                if (ctrl) zoom=100-5*wheel;
                else rotateZ=3*wheel;
            }
            if ( (kcode==KeyEvent.VK_PLUS || kcode==KeyEvent.VK_ADD) && ctrl) zoom=110;
            if (kcode==KeyEvent.VK_MINUS && ctrl) zoom=90;
            if (zoom!=100)   sendCommand(COMMANDzoom+" "+zoom,"",pv,0L);
            if (rotateZ!=0)  sendCommand(COMMANDrotate+" z"+" "+rotateZ,"",pv,0L);
        }
        return ev;
    }
    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> Script >>> */

    protected static void sendScriptToProteinViewer(ProteinViewer pViewer,  String script, ResidueAnnotation aa[], Protein ppAll[], long options) {
        ProteinViewer oneViewer=pViewer;
        if (sze(aa)==0) {
            final String txt=pViewer==null ? script : ResidueAnnotation.replaceVariables(pViewer.getProtein(), script, ppAll);
            sendCommand(txt,  (String)null, pViewer, options);
        } else {
            final ProteinViewer vv[]=getViewersSharingView(pViewer);
            for(ResidueAnnotation a :aa) {
                final Protein p=a.getProtein();
                for(ProteinViewer pv : pViewer==null ? p.getProteinViewers(null) : vv) {
                    if (p==null || p!=pv.getProtein()) continue;
                    if (pViewer==null && !pv.getClass().getName().equals(StrapAlign.getDefaultProteinViewer())) continue;
                    oneViewer=pv;
                    setSelectedAminoAcidsInViewer(0L, withoutOffset(a.getSelectedAminoacids(), a.getSelectedAminoacidsOffset()),ATOM_CA, a.getID(), pv);
                    if (script!=null) {
                        sendCommand(ResidueAnnotation.replaceVariables(a, script),  a.getID(),pv, options);
                    }
                }
            }
        }
        ChDelay.runAfterMS(EDT,runViewerToFront(oneViewer),99);
    }

    /* <<< Script <<< */
    /* ---------------------------------------- */
    /* >>> View >>> */
    public static boolean isViewerActive(ProteinViewer pv) {
        if (pv==null) return false;
        final Protein p=pv.getProtein();
        if (p==null || !cntains(pv, p.getProteinViewers((Class)null))) return false;
        return true;
    }
    private static Map<ProteinViewer,Runnable> mapV2F=new WeakHashMap();
    protected static Runnable runViewerToFront(ProteinViewer pv) {
        Runnable r=mapV2F.get(pv);
        if (r==null) mapV2F.put(pv,r=ChDelay.newThread((ChRunnable)INSTANCE, "viewerToFront",new Object[]{pv,longObjct(OPEN_ADD_TAB)}));;
        return r;
    }
    public static void viewerToFront(ProteinViewer pv, long options) {
        if (pv==null) return;
        if (!isEDT()) inEDT(runViewerToFront(pv));
        else {
            final StrapAlign align=StrapAlign.getInstance();
            final String cn=clazzName(pv);
            final Component panel=pv.getViewerPanel();
            if (panel!=null) {
                final ChTableLayout tl=align.getTablePanel(cn);
                //tl.getJTable(). setIntercellSpacing(dim(10,10));
                final Window w=parentWndw(panel);
                if (tl!=null && cntains(panel,tl.getAllComponents())) align.addDialog(tl);
                else if (gcp(KEY_dock_component,w)!=null) setWndwState('F',w);
                else if ((options&OPEN_ADD_TAB)!=0) tl.addColumn(panel);
                if ((options&OPEN_NARROW_OTHERS)!=0 && tl!=null) tl.narrowAllOthers(panel, ICON_HEIGHT);
            }
            final String frameTitle=gcp(ProteinViewer.KEY_FRAME_TITLE, pv, String.class);
            if (frameTitle!=null) {
                setWndwStateT('F',0,0, frameTitle);
                setWndwStateLaterT('F',666, frameTitle);
            }
        }
    }
    private static ArrayList<Protein> vPIV=new ArrayList();
    public static Protein[] getProteinsInView(ProteinViewer...vv) {
        if (vv==null) return StrapProtein.NONE;
        synchronized(vPIV) {
            vPIV.clear();

            for(ProteinViewer viewer : vv) {
                for(ProteinViewer v : getViewersSharingView(viewer)) {
                    adUniq(v.getProtein(), vPIV);
                }
                if (countNotNull(vv)==0) return StrapProtein.NONE;
                //final Component c=pv.getViewerPanel(); if (c instanceof Protein3d) return ((Protein3d)c).getProteins();
            }
            return toArry(vPIV,Protein.class);
        }
    }
    /* <<< View <<< */
    /* ---------------------------------------- */
    /* >>> StrapEvent >>> */
    //private final static Map<ProteinViewer, Matrix3D> mapMatrix3D=new WeakHashMap();
    private static boolean[] boolean_array;
    protected static void handleEvent(StrapEvent ev) {
        final int id=ev.getType();
        final Object[] para=ev.getParameters();
        final StrapAlign align=StrapAlign.getInstance();
        if (id==StrapEvent.PROTEIN_3D_MOVED) {
            for(int i=sze(para); --i>=0;) {
                final Protein p=deref(para[i],Protein.class);
                if (p==null) continue;
                Matrix3D m3d=p.getRotationAndTranslation();
                if (m3d!=null && m3d.isUnit()) m3d=null;
                for(ProteinViewer v : p.getProteinViewers((Class)null))  {
                    if (gcp(KEY_IS_PREVIEW,v)!=null) continue;
                    v.transform(m3d);
                }
            }
        }
        if (id==StrapEvent.CURSOR_CHANGED_PROTEIN || id==StrapEvent.CURSOR_MOVED_WITHIN_PROTEIN || id==StrapEvent.CURSOR_CLICKED || id==StrapEvent.CURSOR_MOVED_WITHIN_PROTEIN_DELAYED) {
            final Protein pCursor=align.getProteinAtCursor();
            final int iA=align.getIndexOfAminoAcidAtCursor();
            if (pCursor!=null && iA>0) {
                boolean[] bb=boolean_array;
                if (sze(bb)<iA+1) boolean_array=bb=new boolean[iA+99];
                for(int i=bb.length; --i>=0;) bb[i]=false;
                bb[iA]=true;
                for(ProteinViewer pv : pCursor.getProteinViewers((Class)null)) {
                    final boolean delay= 0!=(pv.getPropertyFlags() & PROPERTY_NOTIFY_ALIGNMENT_CURSOR_DELAY);
                    if (id==StrapEvent.CURSOR_MOVED_WITHIN_PROTEIN_DELAYED && !delay) continue;
                    if (id==StrapEvent.CURSOR_MOVED_WITHIN_PROTEIN && delay) continue;
                    setSelectedAminoAcidsInViewer(0L, bb, ATOM_ANY, ProteinViewer.SELECTION_CURSOR, pv);
                }
            }
        }
        if (id==StrapEvent.RESIDUE_SELECTION_CHANGED || id==StrapEvent.RESIDUE_SELECTION_ADDED || id==StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR ||
            (id&StrapEvent.FLAG_ATOM_COORDINATES_CHANGED)!=0 || id==StrapEvent.HETERO_OR_NUCLEOTIDE_STRUCTURE_ADDED_OR_REMOVED /*|| id==StrapEvent.BIOMOLECULES_SPECIFIED*/) {
            for(Protein p : align.getProteins()) {
                for(ProteinViewer pv : p.getProteinViewers((Class)null)) {
                  ((StrapProtein)p).getSelectedAminoacids();
                    final int mc=p.getModificationCount(ProteinModiCount.MC_SELECTED_AA);
                    //if (atoi(gcp(KEY_MC_RES_SEL,pv))!=mc) {                    pcp(KEY_MC_RES_SEL,intObjct(mc),pv);
                    /* TODO nur die geaenderten */
                    if (pv instanceof ProteinListener) ((ProteinListener)pv).handleEvent(ev);
                    if (pv instanceof Protein3d.PView) {
                        final Protein3d p3d=((Protein3d.PView)pv).getViewerPanel();
                        if (id==StrapEvent.RESIDUE_SELECTION_CHANGED)  { p3d.enableDisable(); ChDelay.repaint(p3d,200);}
                        if (id==StrapEvent.RESIDUE_SELECTION_ADDED || id==StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR ||
                            (id&StrapEvent.FLAG_ATOM_COORDINATES_CHANGED)!=0 || id==StrapEvent.HETERO_OR_NUCLEOTIDE_STRUCTURE_ADDED_OR_REMOVED)
                            ChDelay.repaint(p3d,11);
                        //if (id==StrapEvent.BIOMOLECULES_SPECIFIED)  p3d.run(Protein3d.KEY_BIO_MOLECULES,ev.getParameters());
                    }
                    // }
                }
            }
        }
    }
    /* <<< StrapEvent <<< */
    /* ---------------------------------------- */
    /* >>> ResidueSelection >>> */
    private static String current_selection_name;
    private static Protein current_selection_protein;
    protected final static Map<Component,String> mapLastSelection=new WeakHashMap();
    private static long whenLastSelected;
    protected static void setSelectedAminoAcidsInViewer(long options0,boolean[] bb,int atomType,String id, ProteinViewer pv) {
        if (pv==null) return;
        final long time=System.currentTimeMillis();
        if (id==SELECTION_CURSOR) {
            if (time-whenLastSelected<1000) return;
        } else whenLastSelected=time;
        current_selection_name=id;
        current_selection_protein=pv.getProtein();
        final long options= (time-whenMouseViewer>999) ? ProteinViewer.SET_SELECTED_RESIDUES_HIGHLIGHT :0;
        pv.setSelectedResidues(options,bb,atomType,id);
        final Component panel=pv.getViewerPanel();
        if (panel!=null) mapLastSelection.put(panel,id);
        final JMenu mStyle=gcp(KEY_styleMenu,pv, JMenu.class);
        if (mStyle!=null) mStyle.setText("Style for "+countTrue(bb)+" residues");
    }
    public static void commandToProteinViewer(final long options, final String command0, ProteinViewer v,final boolean selected[],String representation_id) {
        int atomType=ATOM_ANY;
        String command=command0;
        if (command.indexOf(sATOM_TYPE_CA)>=0) { atomType|=ATOM_CA; command=rplcToStrg(sATOM_TYPE_CA,"",command);}
        if (command.indexOf(sATOM_TYPE_CB)>=0) { atomType|=ATOM_CB; command=rplcToStrg(sATOM_TYPE_CB,"",command);}
        setSelectedAminoAcidsInViewer(0L, selected,atomType,representation_id,v);
        sendCommand(command,representation_id,v,options);
    }

    private static boolean[] selectedAAto3D_single(long options, ResidueSelection ss[],  ProteinViewer viewer, String selName) {
        final Protein p=viewer!=null ? viewer.getProtein() : null;
        final boolean selected[]=ResidueSelectionUtils.mapPositions(ss, p);
        if (firstTrue(selected)<0) return null;
        final float[] xyz=p.getResidueCalpha();
        final int countTrue=countTrue(selected);
        int count=0;
        for(int i=mini(sze(xyz)/3, sze(selected)); --i>=0;) {
            if (selected[i] && !Float.isNaN(xyz[i*3])) count++;
        }

        if (countTrue>0) {
            final ProteinViewer vv[]=viewer!=null ? new ProteinViewer[]{viewer} : getViewersOfProteinAsk(p, (Class)null);
            if (vv.length>0) {
                boolean needsName=false;
                for(ProteinViewer v : vv) needsName|= 0!=(ProteinViewer.PROPERTY_NAMED_SELECTIONS&v.getPropertyFlags());
                String name=null;
                if (needsName) {
                    final Object
                        msg=vBhB("Protein "+p+"<br>Selected aminocids: ",
                                 scrllpn(0L, monospc(Protein.selectedPositionsToText(selected,0,p)), dim(1,3*EX)),
                                 "What should the selection be named in the protein viewer?",
                                 (count==0 ? "<br>Problem: None of the selected amino acids has coordinates.<br>You may not see the selection.":null)
                                 );
                    name=
                        selName!=null ? selName :
                        0!=(ProteinViewer.PROPERTY_NAMED_SELECTIONS&viewer.getPropertyFlags()|(options&RSto3D_ASK_NAME)) ? ChMsg.input(0L,msg,20,"my_selection") :
                        "Unnamed";
                }
                if (!needsName || sze(name)>=0) {
                    for(ProteinViewer v : vv) setSelectedAminoAcidsInViewer(0L,selected, ATOM_ANY, name,v);
                }
            }
        }
        return selected;
    }
    public final static long RSto3D_COMMANDS=1<<1, RSto3D_COLOR=1<<2, RSto3D_AS_ONE_SINGLE=1<<3, RSto3D_SAME_PROTEIN=1<<4,
        RSto3D_ASK_NAME=1<<5, RSto3D_REPORT_ERROR=1<<6, RSto3D_NOT_SAME_WIRE=1<<7, RSto3D_NO_WIRE=1<<8;

    public static boolean residueSelectionsTo3D(long options, ResidueSelection ss[], ProteinViewer[] vv) {
        int count=0;;
        if (0==(options&RSto3D_AS_ONE_SINGLE)) {
            final ArrayList<String> vCommands=new ArrayList();
            for(ResidueSelection s:ss) {
                final Protein p=s!=null ? s.getProtein() : null;
                if (p==null) continue;
                for(ProteinViewer v : vv!=null ? vv : p.getProteinViewers(null)) {
                    if (0!=(RSto3D_SAME_PROTEIN&options) && p!=v.getProtein()) continue;
                    if (v instanceof Protein3d.PView) continue;
                    String commands[]=NO_STRING;
                    boolean hasColorCmd=false;
                    final String id=s instanceof ResidueAnnotation ?  ((ResidueAnnotation)s).getID() : new ByteArray(99).filter(LETTR_DIGT,s,0,MAX_INT,'_').newString();
                    final boolean bb[]=ResidueSelectionUtils.mapPositions(new ResidueSelection[]{s}, v.getProtein());
                    if ( (RSto3D_COMMANDS&options)!=0 && s instanceof ResidueAnnotation) {
                        final ResidueAnnotation a=(ResidueAnnotation)s;
                        vCommands.clear();
                        final String cn=clazzName(v), scn=shrtClasNam(v);
                        for(ResidueAnnotation.Entry e : a.getEntries()) {
                            final String k=e.getKey(), val=e.getValue();
                            hasColorCmd=hasColorCmd ||  (k.equals(ResidueAnnotation.VIEW3D) && val.startsWith(ProteinViewer.COMMANDcolor));
                            if (e.isEnabled() && (k.equals(ResidueAnnotation.VIEW3D) && v.getSupportsCommand(val) || k.equals(cn) || k.equals(scn))) {
                                vCommands.add(ResidueAnnotation.replaceVariables(a,val));
                            }
                        }
                        commands=toStrgArry(vCommands);
                    }
                    boolean selected=false;
                    if (!hasColorCmd && 0!=(RSto3D_COLOR&options)) {
                        final Color color=s instanceof Colored ? ((Colored)s).getColor() : null;
                        if (color!=null) {
                            final String hex=rgbHex(color.getRGB());
                            for(String cmd : new String[]{ProteinViewer.COMMANDcolor, ProteinViewer.COMMANDsurface_color})
                                if (v.getSupportsCommand(cmd)) commandToProteinViewer(ProteinViewer.INTERPRET_NO_MSG_DIALOGS, cmd+" #"+hex, v, bb, id);
                            selected=true;
                        }
                    }
                    for(String cmd: commands) {
                        setSelectedAminoAcidsInViewer(0L, bb,0,id,v);
                        if (cmd!=null) {
                            commandToProteinViewer(ProteinViewer.INTERPRET_NO_MSG_DIALOGS, cmd , v, bb,id);
                            selected=true;
                        }
                    }
                    if (!selected) setSelectedAminoAcidsInViewer(0L,bb,0,id,v);
                    count++;
                }
            }
        }
        if (count!=1) {
            for (ProteinViewer v  : vv!=null ? vv : getViewers(StrapProtein.getProteinsOfObjects(ss),null)) {
                if (v==null || v instanceof Protein3d.PView && 0!=(options&RSto3D_NO_WIRE)) continue;
                ResidueSelection ss2[]=ss;
                if (0!=(options&RSto3D_NOT_SAME_WIRE) && v instanceof Protein3d.PView) {
                    for(int i=sze(ss); --i>=0;) {
                        final ResidueSelection s=ss[i];
                        if (s!=null && s.getProtein()==v.getProtein()) {
                            (ss2!=ss ? ss2=ss.clone() : ss2 )[i]=null;
                            if (s instanceof VisibleIn123) {
                                final int w0=((VisibleIn123)s).getVisibleWhere(), w1=w0|VisibleIn123.STRUCTURE;
                                if (w0!=w1) { ((VisibleIn123)s).setVisibleWhere(w1); StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,99); }
                            }
                        }
                    }
                }
                count+=countTrue(selectedAAto3D_single(options, rmNull(ss2,ResidueSelection.class), v,  0!=(options&RSto3D_AS_ONE_SINGLE)?null:"Unnamed"));
            }
        }
        if (count==0 && 0!=(options&RSto3D_REPORT_ERROR)) error("No residue selection copied to any 3D-viewer.");
        return count>0;
    }

    public static boolean sendCommand(String command, String selId, ProteinViewer pv,long options) {
        if (pv==null || command==null) return false;
        final String word0=wordAt(nxt(NO_SPC,command),command);
        if (idxOfStrg(word0,ProteinViewer.COMMANDS)<0 || pv.getSupportsCommand(word0)) {

            pv.interpret(command,selId,options);
            return true;
        }
        return false;
    }
    private static Object msgDnD;
    public static void residueSelectionsTo3D_I(long option, ResidueSelection ss[], ProteinViewer vv[]) {
        AbstractButton cbOne=null, cbCmd=null, cbCol=null;
        boolean hasPv=false,hasPvNonWire=false;
        for(ResidueSelection s : ss) {
            final Protein p=s==null ? null : s.getProtein();
            if (p==null) continue;
            for(ProteinViewer v : vv!=null ? vv : p.getProteinViewers(null)) {
                hasPv=true;
                if (v instanceof Protein3d.PView) continue;
                hasPvNonWire=true;
                if (cbCol==null) cbCol=cbox("Use the color of the amino acid selection");
                if (s instanceof ResidueAnnotation) {
                    final String cn=clazzName(v), scn=shrtClasNam(v);
                    for(ResidueAnnotation.Entry e : ((ResidueAnnotation)s).getEntries()) {
                        if (!e.isEnabled()) continue;
                        final String k=e.getKey(), val=e.getValue();
                        if (cbCmd==null &&   (k.equals(ResidueAnnotation.VIEW3D) && v.getSupportsCommand(val) || k.equals(cn) || k.equals(scn)))
                            cbCmd=cbox("Send 3D-Viewer  commands contained in residue annotation");

                    }
                }
            }
        }
        if (!hasPv) return;
        cbOne=ss.length>1 && hasPvNonWire ? cbox("Combine all "+ss.length+" amino acid selection  to one single amino acid selection to be passed to the 3D-View") : null;
        if (cbOne!=null && cbCol!=null) radioGrp(0, cbOne,cbCol);
        if (cbOne!=null && cbCmd!=null) radioGrp(0, cbOne,cbCmd);
        if ( !ChMsg.yesNo(vB("Sending "+ss.length+" residue selections to 3D-View",cbOne,cbCmd, cbCol))) return;
        final long opt=option
            | RSto3D_REPORT_ERROR
            | (isSelctd(cbCmd) ? RSto3D_COMMANDS : 0)
            | (isSelctd(cbOne) ? RSto3D_AS_ONE_SINGLE|RSto3D_ASK_NAME : 0)
            | (isSelctd(cbCol) ? RSto3D_COLOR : 0);
        if (!residueSelectionsTo3D(opt,ss, vv)) {
            if (msgDnD==null) msgDnD=vB(
                                        "Conveniently,  Drag-and-drop can be used to get a residue-selection into a 3D-viewer.",
                                        "This works for all Java-based 3D-Viewers like Astex, but not for native ones like Pymol.",
                                        " ",
                                        "The protein of the residue-selection may be different from the one in the 3D-viewer.",
                                        " ",
                                        WATCH_MOVIE+MOVIE_Sequence_Features_in_3D
                                        );
            ChMsg.msgDialog(msgDnD);

        }
        if (sze(vv)==1) { setFocusedViewer(vv[0]); viewMenuBar(vv[0]); }

    }

    /* <<< ResidueSelection <<< */
    /* ---------------------------------------- */
    /* >>> Viewer of Protein >>> */
    static ProteinViewer[] getViewersOfProteinAsk(Protein p, Class clazz) {
        if (p==null) return ProteinViewer.NONE;
        final ProteinViewer vv0[]=p.getProteinViewers(clazz);
        ProteinViewer vv[]=vv0;
        for(int i=sze(vv);--i>=0;) {
            if (vv[i] instanceof Protein3d.PView) {
                if (vv==vv0) vv=vv0.clone();
                vv[i]=null;
            }
        }
        vv=rmNull(vv,ProteinViewer.class);
        if (vv.length==0) {
            final ProteinViewer v=open3dViewer(new Protein[]{p},clazz, OPEN_ASK|OPEN_SEND_ALL_ANNOTATIONS);
            if (v!=null) vv=new ProteinViewer[]{v};
        }
        return vv;
    }

    static ProteinViewer[] getViewers(Protein[] pp, Class clazz) {
        final List v=new ArrayList();
        for(Protein p : pp) adAllUniq(p.getProteinViewers(clazz),v);
        return toArry(v,ProteinViewer.class);

    }

    static ProteinViewer[] askViewersSharingView(ProteinViewer pv0, String msg) {
        final ProteinViewer[] vv=getViewersSharingView(pv0), vv2=vv.clone();
        if (vv.length<2) return vv;
        final Protein pp[]=new Protein[vv.length];
        for(int i=vv.length; --i>=0;) pp[i]=vv[i].getProtein();
        final ChJList jl=new ChJList(pp,0);
        selectAllItems(jl);
        final Object pan=newCNSEW(scrllpn(0L,jl),"The 3D-View has more than one Proteins. Please select at least one.",msg,null,null);
        if (!ChMsg.yesNo(pan)) return ProteinViewer.NONE;

        final Object selected[]=jl.getSelectedValues();
        for(int i=sze(vv2); --i>=0;) if (!cntains(vv2[i].getProtein(), selected)) vv2[i]=null;
        return rmNull(vv2,ProteinViewer.class);
    }

    public static ProteinViewer findProteinViewerWithProteins(Class clazz,Protein pp[], boolean exactly) {
        for(int iP=sze(pp); --iP>=0; ) {
            if (pp[iP]==null) continue;
            for(ProteinViewer pv : pp[iP].getProteinViewers(clazz)) {
                final Protein ppV[]=getProteinsInView(pv);
                if (!cntainsAll(pp, ppV)) continue;
                if (exactly && pp.length!=ppV.length) continue;
                return pv;
            }
        }
        return null;
    }
    /* <<< Viewer of Protein <<< */
    /* ---------------------------------------- */
   private static ChButton[] strapMenuItemsWithAccelerator;
    static JMenu menuForAccelerators() {
        final JMenu m=new ChJMenu();
        m.setText(KEY_ACCELERATOR);
        ChButton bb[]=strapMenuItemsWithAccelerator;
        if (bb==null) {
            final List<ChButton> v=new ArrayList(99);
            final Set<KeyStroke> setKS=new HashSet();
            for(JMenuItem item : childs(StrapAlign.getInstance().getMenuBar(), JMenuItem.class)) {
                final KeyStroke ks=item.getAccelerator();
                if (ks==null || setKS.contains(ks)) continue;
                setKS.add(ks);
                final ChButton b=gcp(KEY_CLONED_FROM,item,ChButton.class);
                if (b!=null) v.add(b);
            }
            strapMenuItemsWithAccelerator=bb=toArry(v,ChButton.class);
        }
        for(ChButton b : bb)  m.add(b.mi(null));
        return m;
    }
}