package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import charite.christo.strap.extensions.PairAlignerNeoBioPROXY;
import java.io.*;
import java.net.URL;
import java.util.*;
import java.lang.ref.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
public class ReferenceSequence implements ChRunnable {
    /* >>> Instance >>> */
    private final byte[] residueType;
    private final ChRunnable log;
    private final String otherId;
    private final File[] files;
    private final String linkForAli;
    private Runnable runWhenInitialized;
    private URL url;
    final WeakReference<StrapProtein> refProtein;
    public ReferenceSequence(String otherId, StrapProtein protein, ChRunnable log, boolean background, Runnable runWhenInitialized) {
        this.runWhenInitialized=runWhenInitialized;
        this.otherId=otherId;
        this.residueType=protein.getResidueTypeExactLength();
        this.log=log;
        this.linkForAli=Hyperrefs.PAIR_ALIGNMENT+otherId+"_"+protein;
        this.refProtein=new WeakReference(protein);
        this.files=new File[]{
            SOAP_WSDbfetch.dbColonID2file(otherId),
            file(Hyperrefs.toUrlString(otherId,Hyperrefs.PROTEIN_FILE))};
        if (sze(this.files[0])==0 && sze(this.files[1])==0) {
            final int MAX_AGE_DAYS=999;
            this.url=url((Hyperrefs.toUrlString(otherId,Hyperrefs.PROTEIN_FILE)));
            if (background) {
                adUniq(this,vDownload);
                if (!isDownloading) ChDelay.newThread(this,"DOWNLOAD",null).start();
            } else download_TS(url,MAX_AGE_DAYS);
        }
        if (sze(this.files[0])>0 || sze(this.files[1])>0) init();
        //debug("StrapProtein:  f=",f," "+sze(f)+"bytes "+otherId);
    }
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Init >>> */
    private short[] u2i;
    private boolean isDownloaded, success;
    private float alignmentScore=Float.NaN;
    private byte[] alignment[];
    private int aliStartsAtIdx;
    protected boolean getSuccess() { return success; }

    public StrapProtein getReferenceProtein() {
        StrapProtein p=mapP().get(this.otherId);
        if (p==null) {
            for(File f : this.files) {
                p=StrapProtein.newInstance(f);
                if (p!=null) {
                    mapP().put(this.otherId,p);
                    break;
                }
            }
        }
        return p;
    }
    public Object run(String id, Object arg) {
        if (id=="DOWNLOAD") {
            synchronized(vDownload) {
                isDownloading=true;
                for(int n; (n=sze(vDownload))>0;) {
                    for( int i=n; --i>=0;) {
                        final ReferenceSequence rs=(ReferenceSequence)get(i,vDownload);
                        if (rs!=null) {
                            download_TS(rs.url,999);
                            rs.init();
                            final StrapProtein p=refProtein.get();
                            if (p!=null) {
                                Protein.incrementModificationCount(ProteinModiCount.MC_RESIDUE_SELECTIONS,p);
                                StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_ADDED,333);
                            }
                        }
                        vDownload.remove(rs);
                    }
                    sleep(10);
                }
                isDownloading=false;
            }
        }
        return null;
    }
    private void init() {
        if (this.isDownloaded==true) return;
        final StrapProtein p=getReferenceProtein();
        final byte[] seq=p!=null ? p.getResidueTypeExactLength() : null;
        if (sze(seq)>0) {
            final PairAlignerNeoBioPROXY aligner=new PairAlignerNeoBioPROXY();
            aligner.setOptions(SequenceAligner.OPTION_CHECK_PERFECT_MATCH);
            aligner.setSequences(this.residueType,seq);
            aligner.compute();
            final byte[][] aligned=aligner.getAlignedSequences();
            final float score=((HasScore)aligner).getScore();
            if ( !(Float.isNaN(this.alignmentScore=score) || sze(aligned)!=2 || aligned[0]==null || aligned[1]==null)) {
                final byte[] a0=aligned[0];
                final byte[] a1=aligned[1];
                this.alignment=aligned;
                this.aliStartsAtIdx=maxi(0,strstr(STRSTR_IC,a0, this.residueType));
                {
                    final File fAli=file(Hyperrefs.toUrlString(this.linkForAli, Hyperrefs.PLAIN_TEXT));
                    final ByteArray sb=new ByteArray(999).append(aligned[0]).append("\r\n");
                    final int to=mini(a0.length,a1.length);
                    for(int i=0; i< to;i++) sb.append((char)Blosum.BL2SEQ_MIDLINE[a0[i]&127][a1[i]&127]);
                    sb.append("\r\n").append(aligned[1]).append("\r\n alignment score=").append(score).append("\r\n");
                    wrte(fAli,sb);
                }
                final short u2i[]=new short[seq.length];
                this.u2i=u2i;
                Arrays.fill(u2i,(short)-1);
                final int to=mini(a0.length, a1.length);
                short idxO=(short)idxOfLetters(a0, this.residueType,0,MAX_INT);
                short idxU=(short)idxOfLetters(a1,seq,0,MAX_INT);
                for(int col=0; col<to; col++) {
                    final boolean lO=is(LETTR,a0[col]), lU=is(LETTR,a1[col]);
                    if (lO && lU && 0<=idxU && idxU<u2i.length)  {
                        u2i[idxU]=idxO;
                    }
                    if (lO) idxO++;
                    if (lU) idxU++;
                }
                if (log!=null) log.run(ChRunnable.APPEND, linkForAli+" clustalW-score="+this.alignmentScore+"\n\n");
                this.success=true;
            }
        } else if (log!=null) log.run(ChRunnable.APPEND, " Protein "+this+": failed to load "+otherId+"\n\n");
        this.isDownloaded=true;
        final Runnable r=this.runWhenInitialized;
        if (this.success && r!=null) r.run();
    }
    /* <<< Init <<< */
    /* ---------------------------------------- */
    /* >>> Getters >>> */
    public String getLinkForAlignment() { return linkForAli;}
    public short[] getUniprotIdx2Idx() {  return isDownloaded ? u2i : null;}

    public boolean[] uniprotIdx2Idx(boolean[] uniprot) {
        final short[] u2i=getUniprotIdx2Idx();
        boolean bb[]=null;
        for(int i=mini(sze(u2i), uniprot.length); --i>=0; ) {
            if (uniprot[i]) {
                int idx=u2i[i];
                if (bb==null) bb=new boolean[idx+1];
                if (idx>=0 && idx<bb.length) bb[idx]=true;
            }
        }
        return bb;
    }

    public float getAlignmentScore() { return alignmentScore;}
    public ByteArray getLocalAlignment(int iAa_from, int iAa_to, ByteArray sb) {
        if (!isDownloaded) return null;
        final int[] fromToCol=new int[2];
        countMismatch(iAa_from, iAa_to, fromToCol);
        final byte[] aa[]=this.alignment, a0=aa[0],a1=aa[1];
        final int f=maxi(fromToCol[0], 0);
        final int t=mini(fromToCol[1], a0.length,a1.length);
        sb.append(' ', sze(otherId)+1);
        for(int i=f; i<t; i++) sb.append((char)a0[i]);
        sb.append('\n').append(' ',sze(otherId)+1);
        for(int i=f; i<t; i++) sb.append((char)Blosum.BL2SEQ_MIDLINE[a0[i]&127][a1[i]&127]);
        sb.append('\n');
        if (otherId!=null) sb.append(otherId).append(' ');
        for(int i=f; i<t; i++) sb.append((char)a1[i]);
        //sb.append("\n colFrom="+fromToCol[0]+" colTo="+fromToCol[1]+" aliStartsAtIdx="+aliStartsAtIdx);
        return sb.append('\n');
    }
    public int countMismatch(final int fromIdx, final int toIdx, int colFromTo[]) {
        if (!isDownloaded) return MAX_INT;
        final byte[] a0=this.alignment[0], a1=this.alignment[1];
        int iA=this.aliStartsAtIdx, mismatch=0;
        if (colFromTo!=null) colFromTo[0]=-1;
        for(int i=0; i<a0.length && i<a1.length; i++) {
            if (is(LETTR,a0[i])) {
                if (iA>=fromIdx-2) {
                    if (a0[i]!=a1[i]) mismatch++;
                    if (colFromTo!=null) {
                        if (colFromTo[0]<0) colFromTo[0]=i;
                        colFromTo[1]=i+1;
                    }
                }
                if (iA++>toIdx) break;
            }
        }
        return mismatch;
    }
    /* <<< Getters <<< */
    /* ---------------------------------------- */
    /* >>> Static Utils >>> */
    private final static Vector<ReferenceSequence> vDownload=new Vector();
    private static boolean isDownloading;

    private static Reference<HashMap<String,StrapProtein>> refMap;
    private static Map<String,StrapProtein> mapP() {
        Map<String,StrapProtein> m=(Map)deref(refMap);
        if (m==null) refMap=new SoftReference(m=new HashMap());
        return m;
    }

}
