package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.util.*;
import java.awt.event.*;
import java.awt.Insets;
import javax.swing.*;
import java.io.File;
import java.lang.ref.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static charite.christo.protein.SimilarPdbProvider.*;
import javax.swing.plaf.basic.*;
/**HELP

<h3>First option: Associating structures of exactly this protein</h3>

The button <i>JCOMPONENT:StrapView#button(BUT_INFER_3D)!</i>
associates those PDB structures that are directly referenced in the protein files.
<br>
For example the entry UNIPROT:P21243 has a reference for PDB:1FNT in the following form:
<pre>
   DR   PDB; 1FNT; X-ray; 3.20 A; A/O=1-252.
</pre>

<br>
The button acts on the proteins selected in STRAP.

<hr><h3>Second option: Associating structures of homologous proteins</h3>

The button <i>JCOMPONENT:StrapView#button(BUT_SIMIL_SEARCH)!</i>
starts a search for homologous 3D-structures based on Uniprot IDs.
<br>
The search is performed for the proteins selected in STRAP and shown in a table.
<br>
In contrast to the option above, not only PDB structures of identical but also homologous proteins are shown.

<br>

The user can select proteins in the table and associate the 3D
structure.
<br>
<br>
<u>Author</u> of SSMap-based backend: Fabrice David.<br>
<u>Cache:</u> <i>STRING_ARRAY:SimilarPdbProvider$Utils#getDirectories()</i>

<hr><h3>Third option: Blast against PDB</h3>

Press <i>BUTTON:DialogBlast!</i> to run WIKI:BLAST dialog. Chose a PDB entry from the BLAST result.
<br>
Open <i>JCOMPONENT:StrapView#button(BUT_INFER_3D_ALL)!</i> and
enter the PDB ID of your choice.
@author Christoph Gille
http://umber.sbs.man.ac.uk/cgi-bin/dbbrowser/sprint/searchprintss.cgi?display_opts=Prints&category=None&queryform=false&prints_accn=PR00401

Pruefen UNIPROT\:O56227  UNIPROT\:O56228
*/
public class SimilarPdb implements ChRunnable, ActionListener {
    private final static int SEND_NUM_IDS=33;
    public final static long INFER_3D_LOG=1<<1, INFER_3D_HOMOLOGOUS=1<<2, INFER_3D_SHOW_BACKBONE=1<<3, INFER_3D_BG=1<<4;
    private final static String EXT=".associatePDB", SKIPPED="Skipped", NOT_F="Not found";
    public final static Map<Protein,Result[]>[] MAPS_P =new Map[CLASSES.length];
    public final static Map<String,      Result[]>[] MAPS_ID=new Map[CLASSES.length];

    private static ChJTable jTable;
    private static ChFrame _frame;
    private static SimilarPdb inst;
    private static Thread threadUpdateTable;
    private static Object iconWait;
    private static List<boolean[]> vStopThread=new ArrayList();
    private final static int COL_NO=0, COL_REPORT=2;
    protected final static int COL_PROT=1;
    static {
        for(int i=CLASSES.length; --i>=0;) {
            MAPS_ID[i]=new HashMap();
            MAPS_P[i]=new WeakHashMap();
        }
    }
    private static ChFrame frame() {
        if (_frame==null) {
           final JComponent
               butGo=new ChButton("GO").t("Associate 3D-structures").li(instance()).i(IC_BLANK),
                pSouth=pnl(pnl(dim(100,1)),butGo,ProteinPopup.butRemove3D(jTable,null),pnl(dim(100,1))),
                pNorth=newCNSEW("Associating 3D-structure to proteins",null,null,smallHelpBut(SimilarPdb.class),null);
           _frame=new ChFrame("").ad(newCNSEW(scrllpn(0L,jTable,dim(1,100)),pNorth,pSouth,null,null)).shw(ChFrame.AT_CLICK|ChFrame.PACK);
        }
        return _frame;
    }
    private static SimilarPdb instance() {
        if (inst==null) {
            inst=new SimilarPdb();
            jTable=new ChJTable(null, ChJTable.DEFAULT_RENDERER|ChJTable.ICON_ROW_HEIGHT|ChJTable.NO_REORDER);
            jTable.setColumnWidth(0,2*EM,true);
            threadUpdateTable=ChDelay.newThread(inst,"UPDATE_TABLE",null);
            iconWait=iicon(IC_HOURGLASS);
        }
        return inst;
    }
    private static boolean isRunning;
    public static void startThreads(Protein[] pp, boolean showTable, Runnable whenFinished) {
        for(int i=sze(vStopThread); --i>=0;) {
            boolean[] stp=get(i,vStopThread,boolean[].class);
            stp[0]=true;
            vStopThread.remove(stp);
        }
        instance();
        final Object[][] tableData=new Object[pp.length][3];
        for(int iP=0; iP<pp.length; iP++) {
            if (pp[iP]!=null) {
                tableData[iP][COL_NO]=intObjct(iP);
                tableData[iP][COL_PROT]=pp[iP].WEAK_REF;
            }
            tableData[iP][COL_REPORT]=iconWait;
        }
        for(int iClazz=CLASSES.length; --iClazz>=0;) {
            final boolean interrupt[]=new boolean[1];
            vStopThread.add(interrupt);
            ChDelay.newThread(inst,"GO",new Object[]{pp, new Integer(iClazz),  tableData, whenFinished,interrupt}).start();
        }
        if (showTable) {
           final String RIGHT_COL="Sequence-range-graph | Sequence-identity | PDB-chain | Sequence-range";
           jTable.setModel(new ChTableModel("No","Protein",RIGHT_COL).setData(tableData));
           ChJTable.setRowSorter(new Comparator[]{jTable},jTable);
           pcp("D",tableData,jTable);
           jTable.setColumnWidth(1,16*EM,true);
           jTable.setColumnWidth(2,60*EM,true);
           jTable.revalidate();
           jTable.repaint();
           frame().shw(ChFrame.AT_CLICK);
       }
    }
    private void getSimilar(Protein[] pp, int iClazz, Object[][] tableData, boolean onlyAcc, boolean[] interrupt) {
        final Class clazz=CLASSES[iClazz];
        SimilarPdbProvider inst=null;
        inst=HotswapUtil.instance(clazz, SimilarPdbProvider.class, true);
        final Set<String> vID=new HashSet();
        final Map<String, Result[]> mapI=MAPS_ID[iClazz];
        for(int iP=0; iP<pp.length; iP++) {
            if (interrupt[0]) return;
            final Protein p=pp[iP];
            if (p==null || p.isLoadedFromStructureFile()) continue;
            for(Map<Protein, Result[]> mp : MAPS_P) {
                if (mp.get(p)==null) {
                    final String refs[]=p.getSequenceRefs();
                    for(int iID= onlyAcc ? 1 : refs.length ; --iID>=0;) {
                        final String id=onlyAcc ? p.getAccessionID() : refs[iID];
                        if (id!=null && mapI.get(id)==null) {
                            final File f=Utils.getFile(id, clazz,isWindows());
                            final int sze=sze(f);
                            if (sze>0) {
                                if (sze>=MIN_FILE_SIZE) {
                                    final int oldSze=sze(mapI);
                                    SimilarPdbProvider.Utils.parse(readBytes(f),mapI,(Map)null);
                                    //foundInFile=foundInFile || oldSze<sze(mapI);
                                }
                            } else vID.add(id);
                        }
                    }
                }
            }
        }
        final String all[]=toArry(vID,String.class);
        mapID_to_mapP(pp,iClazz, tableData);
        final Map<String,Result[]> mNew=new HashMap();
        final ByteArray baTmp=new ByteArray(99);
        if (all.length>0) {
            final String ids[]=new String[mini(SEND_NUM_IDS, all.length)];
            for(int i=0; i<all.length; i+=ids.length) {
                for(int j=ids.length; --j>=0; ) ids[j]= i+j<all.length ? all[i+j] : null;
                mNew.clear();
                inst.get(ids, mapI, mNew);
                mapID_to_mapP(pp, iClazz, tableData);
                for(Map.Entry<String,Result[]> e : entryArry(mNew)) {
                    final String k=e.getKey();
                    final File f=SimilarPdbProvider.Utils.getFile(k, clazz,isWindows());
                    if (sze(f)==0) wrte(f,Utils.toText(e.getValue(), baTmp.setEnd(0)).append('\n'));
                }
            }
        }
    }
    private static void mapID_to_mapP(Protein[] pp, int iClazz, Object[][] tableData) {
        synchronized(MAPS_P) {
            final  Map<Protein,Result[]> mp=MAPS_P[iClazz];
            for(int iP=0; iP<pp.length; iP++) {
                final Protein p=pp[iP];
                if (p==null) continue;
                if (!p.isLoadedFromStructureFile() && mp.get(p)==null) {
                    final String refs[]=p.getSequenceRefs(), acc=p.getAccessionID();
                    for(int iID=-1; iID<refs.length; iID++) {
                        final String id=iID<0 && !cntains(acc,refs) ? acc : refs[iID];
                        if (id!=null) {
                            final Result add[]=MAPS_ID[iClazz].get(id);
                            if (add!=null && add!=NOT_FOUND) {
                                Result result[]=mp.get(p);
                                nxtEntry:
                                for(Result r : add) {
                                    if (r==null) continue;
                                    if (result!=null) {
                                        for(Result rOld : result) {
                                            if (rOld==null) break;
                                            if (r.pdbID.equals(rOld.pdbID)) continue nxtEntry;
                                        }
                                    }
                                    result=ad(r, result,10, Result.class);
                                }
                                mp.put(p,result);
                            }
                        }
                    }
                }
                if (tableData!=null) insertIntoTable(iP,p, tableData);
            }
        }
    }

    private static float score(Result r, int from, int to) {
        if (r==null) return Float.MIN_VALUE;
        return r.identity;
    }

    public static String getBestPdbID(Protein p) {
        final int[] firstLast=p.getFirstAndLastResidueIndex();
        return getBestPdbID(p,firstLast[0] , firstLast[1]);
    }
    public static String getBestPdbID(Protein p, int from, int to) {
        Result rBest=null;
        float score=0;
        startThreads(new Protein[]{p}, false, (Runnable)null);
        for(int iTry=99; --iTry>=0;) {
            boolean hasNull=false;
            for(int iClazz=CLASSES.length; --iClazz>=0;) {
                sleep(333);
                puts(">");
                final Result[] rr=MAPS_P[iClazz].get(p);
                if (rr!=null && rr!=NOT_FOUND) break;
                hasNull|= rr==null;
            }
            if (!hasNull) break;
        }
        for(Map<Protein,Result[]> m : MAPS_P) {
            final Result rr[]=m.get(p);
            if (rr==null) continue;
            for(Result r : rr) {
                final float s=score(r,from,to);
                if (s>score) { score=s; rBest=r;}
            }
        }
        return rBest!=null ?rBest.pdbID : null;
    }
    public Object run(String id, Object arg) {
        final Object[] argv=arg instanceof Object[] ? (Object[])arg : null;
        if (id=="UPDATE_TABLE") {
            jTable.revalidate();
            jTable.repaint();
        }
        if (id=="GO") {
            isRunning=true;
            final Protein pp[]=(Protein[])argv[0];
            final int iClazz=atoi(argv[1]);
            final Object[][] tableData=(Object[][]) argv[2];
            final Runnable whenFinished=(Runnable) argv[3];
            final boolean[] interrupt=(boolean[]) argv[4];
            getSimilar(pp,iClazz, tableData, true, interrupt);
            getSimilar(pp,iClazz, tableData, false, interrupt);
            isRunning=false;
            inEDT(whenFinished);
            vStopThread.remove(interrupt);
        }
        return null;
    }

    final static ByteArray baTmp=new ByteArray(99),baTmpPdb=new ByteArray(99);
    final static List vTmp=new ArrayList();

    private static synchronized void insertIntoTable(int row, Protein p, Object[][] td) {
        vTmp.clear();
        if (row>=td.length || get(COL_PROT,td[row])!=p) return;
        for(int iC=CLASSES.length; --iC>=0;) {
            final Result[] result=MAPS_P[iC].get(p);
            if (result!=null) {
                for(Result r : result) {
                    if (r==null) continue;
                    baTmp.setEnd(0);
                    final boolean isRange=r.from>0 || r.to!=MAX_INT;
                    if (isRange) {
                        final int idx0=p.getFirstResidueIndex();
                        final int nR=p.countResidues();
                        if (r.from>idx0+nR-8 ||  r.to<idx0+8) continue;
                        final int W=30, f=mini(W, maxi(0, r.from-1-idx0)*W/nR), t=mini(W, maxi(0,r.to-idx0)*W/nR);
                        baTmp.append('|').append(' ',f).append('-',t-f).append(' ',W-t).append('|');
                    }
                    baTmp.append(r.identity,4,1).append(' ').append(r.pdbID);
                    if (isRange) baTmp.append(' ').append(r.from+1,3).append(" -").append(r.to,4);
                    else baTmp.append("  1 - Last");
                    vTmp.add(baTmp.newString());
                }
            }
        }
        if (sze(vTmp)>0) {
            sortArry(vTmp, Collections.reverseOrder());
            final Object gui;
            if (sze(vTmp)>1) {
                final String ss[]=toArry(vTmp,String.class);
                final ByteArray baPDB=readBytes(file(DIR_ANNOTATIONS+"/"+p+EXT),baTmpPdb);
                int selectIdx=-1;
                if (baPDB!=null)  for(int i=ss.length; --i>=0;)  if (strstr(baPDB,ss[i])>0) selectIdx=i;
                final ChCombo choice=new ChCombo(ss, ChJTable.DEFAULT_RENDERER).s(selectIdx);
                pcp(ChRenderer.KEY_EDITOR_COMPONENT,"",gui=monospc(choice));
            }
            else gui=vTmp.get(0);
            td[row][COL_REPORT]=gui;
        } else td[row][COL_REPORT]=p.isLoadedFromStructureFile() ? SKIPPED : NOT_F;
        ChDelay.runAfterMS(PAUSE, threadUpdateTable,999);
    }

    public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        if (cmd=="GO") {
            final Object[][] data=(Object[][]) gcp("D",jTable);
            final int ii[]=ChJTable.selectedRowsConverted(jTable);
            final int n=sze(data);
            final boolean all=ii.length==0 && (n<3 || ChMsg.yesNo("Infer Structures on all "+n+" proteins?"));
            final Protein pp[]=new Protein[n];
            final String pdbs[]=new String[n];
            int count=0;
            for(int i=n; --i>=0;) {
                if (!all && idxOf(i,ii,0,MAX_INT)<0) continue;
                pp[count]=get(COL_PROT,data[i],Protein.class);
                final String txt=toStrg(get(COL_REPORT,data[i]));
                if (txt==null || txt.startsWith(NOT_F) || txt.startsWith(SKIPPED)) continue;
                final int iPdb=strstr(STRSTR_AFTER,"PDB:",txt);
                final String id;
                if (iPdb>=0) {
                    final int  spc=nxtE(SPC,txt,iPdb,MAX_INT);
                    id=txt.substring(iPdb,spc);
                } else {
                    final String s=delPfx("pdb",txt);
                    id=s.substring(0,nxtE(SPC,s,0,MAX_INT));
                }
                if (id!=null) {
                    final Protein p=(Protein)deref(data[i][COL_PROT]);
                    wrte(file(DIR_ANNOTATIONS+"/"+p+EXT), id);
                    ((StrapProtein)p).setInfoAssociatedPdb(txt);
                    pdbs[count++]=id.replace(':','_');
                }
            }
            StrapProteinUtils.threadInferCoordinates(pp,pdbs,ChLog.NO_LOG,SimilarPdb.INFER_3D_LOG).start();
        }
    }

    public static void showTablePDB(Protein pp[], boolean all) {
        instance();
        int count=0;
         Object tableData[][]=new Object[pp.length][];
        for(int iP=0; iP<pp.length; iP++) {
            final Protein p=pp[iP];
            if (p==null) continue;
            String id_chain=null;
            if (p.isLoadedFromStructureFile()) { id_chain=SKIPPED; if (!all) continue;}
            else if (p.getPdbRef()!=null) {
                id_chain=p.getPdbRef();
                if (id_chain==null) id_chain=readStrg(DIR_ANNOTATIONS+"/"+p+"getCoordinatesFrom");
            } else if (all) {
                final ByteArray baPDB=readBytes(file(DIR_ANNOTATIONS+"/"+p+EXT),baTmpPdb);
                id_chain=baPDB!=null ? toStrg(baPDB) : "";
            }
            if (id_chain!=null) tableData[count++]=new Object[]{intObjct(iP),p,id_chain};
        }
        tableData=chSze(tableData,count,Object[].class);
        pcp("D",tableData,jTable);
        jTable.setModel(new ChTableModel("No","Protein", "Type the PDB-ID").setData(tableData).editable(false,false,true));
        ChJTable.setRowSorter(new Comparator[]{jTable},jTable);

        jTable.setColumnWidth(1,60*EM,true);
        jTable.setColumnWidth(2,16*EM,true);
        jTable.revalidate();
        jTable.repaint();

        frame().shw(ChFrame.AT_CLICK);
    }
}
/*
 http://www.ebi.ac.uk/pdbe-site/pdbemotif/
*/