package charite.christo.strap;
import charite.christo.*;
import charite.christo.ChFile;
import charite.christo.hotswap.ProxyClass;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import charite.christo.strap.extensions.*;
import charite.christo.strap.interfaces.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.io.File;
import java.util.*;
import java.util.List;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static java.awt.Color.*;

/**
   @author Christoph Gille
*/
public class StrapPlugins implements ActionListener, ChRunnable {
    private  static StrapPlugins inst;
    private static StrapPlugins instance() {if (inst==null) inst=new StrapPlugins(); return inst;}

    private StrapPlugins(){}
    private static void loadClassesIntoVectors(int i) {
        final Class interf=INTERFACES[i];
        UniqueList l=allClassesVV()[i];
        if (l==null) {
            allClassesVV()[i]=l=new UniqueList(Object.class);
            l.run(ChRunnable.SET_TREE_VALUE,interf);
        }
        builtInClassesV(builtInPlugins()[i],l);
        if (userPluginsActivated) {
            for(Class c: getUserPlugins()) {
                if (c!=null && interf.isAssignableFrom(c)) l.add(c);
            }
            getAllClasses_modificationCount++;
        }
        UIupdateList.update(UIupdateList.PLUGINS_ADDED_OR_REMOVED);
    }
 private static int getAllClasses_modificationCount;
    public static void setChanged() {getAllClasses_modificationCount++;}
    /**
       Changes each time the plugins change;
       If only the built in plugins are activated, it returns 0;
    */
    public static int getAllClasses_modificationCount() { return getAllClasses_modificationCount;}
    public static UniqueList allClassesV(Class clas) {
        return allClassesV(class2index(clas));
    }
    private final static UniqueList EMPTY=new UniqueList(Object.class);
    protected static UniqueList allClassesV(int interf) {
        if (interf<0 || interf>=INTERFACES.length) { EMPTY.clear(); return EMPTY;}
        if (allClassesVV()[interf]==null) loadClassesIntoVectors(interf);
        return allClassesVV()[interf];
    }
    public static void builtInClassesV(Class clas,List<Class> v) {
        for(Object plugins[]:builtInPlugins())
            if (plugins[0]==clas)
                builtInClassesV(plugins,v);
    }
    private static void builtInClassesV(Object plugins[],List<Class> v) {
        final Object[] cc=plugins;
        for(int j=1;j<cc.length;j++) {
            final Object o=cc[j];
            if (o instanceof Class) {  v.add((Class) o);continue;}
            try { v.add(Class.forName((String)o)); } catch(Exception e){}
        }
    }
    private static Object[][] _builtInPlugins;
    private static Object[][] builtInPlugins() {
        if (_builtInPlugins==null) {
            final Object

                class_jv=CLASS_ChJvPROXY,
                class_rasmol=isWindows()||isMacintosh() ? null : CLASS_Rasmol,
                class_vmd=isWindows()||isMacintosh() ? null : CLASS_Vmd;

            _builtInPlugins=new Object[][]{
                {ProteinParser.class,DSSP_Parser.class,PDB_Parser.class,SingleFastaParser.class,
                 NumberedSequence_Parser.class,XML_SequenceParser.class,SBD_Parser.class,StupidParser.class},
                {ProteinWriter.class,ProteinWriter1.class},
                {AlignmentWriter.class,ExportAlignment.class,CLASS_AlignmentWriterHSSP},
                {StrapExtension.class,CLASS_ConsensusSequence,CLASS_DemoAssociateObjectsToProteins},
                {ResidueSelection.class,CLASS_Distance3DToHetero,CLASS_Distance3DToCursor},
                {ProteinProteinValue.class,CLASS_SequenceDisSimilarityAsAligned,CLASS_IndividualSequenceDisSimilarity,CLASS_AlignmentScore,RmsdValue.class},
                {SequenceDisSimilarity.class,CLASS_CrudeSequenceDisSimilarity},
                {ValueOfAlignPosition.class,CLASS_AlignmentEntropy},
                {ValueOfResidue.class,CLASS_Hydrophobicity,CLASS_ResidueSolventAccessibility,CLASS_ResidueValues_from_textfile /*,CLASS_ResidueHydrophobicity_TOPPRED*/},
                {ValueOfProtein.class,CLASS_CountResidues},
                {SequenceAligner.class, ClustalW_3D.class,CLASS_MultipleAlignerClustalW,CLASS_MultipleAlignerAmap,CLASS_MultipleAlignerMafft,
                 CLASS_MultipleAlignerT_Coffee, CLASS_MultipleAlignerKalign, CLASS_MultipleAlignerMuscle,CLASS_MultipleAlignerAlign_m,CLASS_MultipleAlignerDialign,
                 CLASS_MultipleAlignerDialignT,
                 CLASS_MultipleAlignerProbcons2,CLASS_MultipleAligner_SPEM,
                 CLASS_PairAlignerNeoBioPROXY,
                 PairAlignerJAlignerPROXY.class,CLASS_Superimpose_TM_align,CLASS_Superimpose_native_CE, CLASS_Superimpose_CEPROXY },

                //equenceAligner3D.class,CLASS_Superimpose_TM_align,CLASS_Superimpose_native_CE, CLASS_Superimpose_CEPROXY},

                {ProteinsSorter.class,CLASS_ProteinsSorter_with3D_without3D},
                {ProteinViewer.class,CLASS_ChJmolPROXY,CLASS_Pymol,class_rasmol,class_vmd, CLASS_ChAstexPROXY, class_jv /*, CLASS_ProteinWorkshopPROXY*/ },
                {Superimpose3D.class,CLASS_Superimpose_TM_align,CLASS_Superimpose_native_CE,CLASS_SuperimposeGoedePROXY, CLASS_Superimpose_GangstaPlus, CLASS_Superimpose_LajollaProteinPROXY, CLASS_Superimpose_CEPROXY/*, CLASS_Superimpose_LajollaRNA */},
                {TransmembraneHelix_Predictor.class,CLASS_TransmembraneHelix_Phobius, CLASS_TransmembraneHelix_HMMTOP,CLASS_TransmembraneHelix_TOPPRED,
                 CLASS_TransmembraneHelix_ORIENTM,CLASS_TransmembraneHelix_SPLIT4,CLASS_TransmembraneHelix_SPLIT3,
                 CLASS_TransmembraneHelix_SOSUI,CLASS_TransmembraneHelix_DAS,
                 CLASS_TransmembraneHelix_PRED_TMR,CLASS_TransmembraneHelix_TMHMM2,CLASS_TransmembraneHelix_TMAP,
                 CLASS_TransmembraneHelix_TMPRED,CLASS_TransmembraneHelix_MEMSAT,CLASS_TransmembraneHelix_WaveTM,
                 CLASS_TransmembraneHelix_THUMBUP,CLASS_TransmembraneHelix_TMAP_MULTI},
                {SecondaryStructure_Predictor.class, CLASS_SecondaryStructure_Sopma, CLASS_SecondaryStructure_Jnet, CLASS_SecondaryStructure_NNPREDICT},
                {CoiledCoil_Predictor.class,CLASS_CoiledCoil_PredictorRobinson},
                {SequenceBlaster.class,CLASS_Blaster_ebi_ac_uk,CLASS_Blaster_local_Wu, CLASS_Blaster_local_NCBI /*, CLASS_Blaster_local_NCBI_gapped_PSI */},
                {PhylogeneticTree.class,CLASS_PhylogeneticTreeATVPROXY, CLASS_PhylogeneticTree_TreeTop},
                {PredictSubcellularLocation.class,SubcellularLocationHum_mPloc.class, SubcellularLocationSherLoc.class, SubcellularLocationPLOC.class}
            };

            for(Object[] oo : _builtInPlugins) {
                for(int i=sze(oo); --i>=0;) {
                    if (!(oo[i] instanceof String)) continue;
                    final String  s=(String)oo[i];
                    if (!Insecure.CLASSLOADING_ALLOWED && s.endsWith("PROXY")
                        || !Insecure.EXEC_ALLOWED && (s.indexOf("Superimpose")>=0||s.indexOf("Aligner")>0||s==CLASS_Pymol||s==CLASS_Rasmol||s==CLASS_Vmd)
                        ) {
                        if (
                            s!=CLASS_PairAlignerNeoBioPROXY &&
                            s!=CLASS_Superimpose_CEPROXY &&
                            s!=CLASS_ChAstexPROXY &&
                            s!=CLASS_PairAlignerJAlignerPROXY
                            ) oo[i]=null;
                    }
                }

            }
        }
        return _builtInPlugins;
    }
    public final static Class[] INTERFACES=/*Customize.pluginInterfaces= */_interfaces();
    private static UniqueList _allClassesVV[];
     private static UniqueList[] allClassesVV() {
         if (_allClassesVV==null) _allClassesVV=new UniqueList[builtInPlugins().length];
         return _allClassesVV;
     }
    static UniqueList[] getInterfacesAndClasses() {
        final UniqueList vv[]=allClassesVV();
        for(int i=INTERFACES.length; --i>=0;) if (vv[i]==null) loadClassesIntoVectors(i);
        return vv;
    }
    private static Class[] _interfaces() {
        Class cc[]=new Class[builtInPlugins().length];
        for(int i=cc.length;--i>=0;) cc[i]=(Class)builtInPlugins()[i][0];
        return cc;
    }
    public static boolean isBuiltIn(Class c) {return idxOf(c,builtInPlugins())>=0;}
    private static int class2index(Class c) {
        for(int i=builtInPlugins().length;--i>=0;) if (c==INTERFACES[i]) return i;
        return -1;
    }
    // ----------------------------------------
    private final static StringBuffer ERROR_LOAD_CLASSES=new StringBuffer();
    private static Map<Class,String> mapPluginsToMenu=new WeakHashMap();
    private final static UniqueList<Class> vPLUGINS=new UniqueList(Class.class);
    private static int getUserPlugins_mc, dl_mc;

    protected static void downloadRequiredJars() {
        final String[][] cdd=ChZip.classesDownloadDelete();
        final int mc=ChZip.jarFiles_modificationCount();
        if (dl_mc!=mc) {
            dl_mc=mc;
            final String[]  download=cdd['L'], delete=cdd['D'];
            if (download!=null) for(String s: download) InteractiveDownload.downloadFilesAndUnzip(urls(s),dirPlugins());
            if (delete!=null) for(String s:delete) delFile(file(dirPlugins()+"/"+s));
        }
    }
    private static Class[] getUserPlugins(){
        if (isLog) LOG.appnd("getUserPlugins()\n");
        final String classes[]=ChZip.classesDownloadDelete()['C'];
        final int mc=ChZip.jarFiles_modificationCount();
        if (getUserPlugins_mc!=mc) {
            getUserPlugins_mc=mc;
            downloadRequiredJars();
            vPLUGINS.clear();
            ChZip.unpack200(dirPlugins(),ChZip.PACK_IF_NEWER);
            ERROR_LOAD_CLASSES.setLength(0);
            ChClassLoader.addUrls();
            for(String line: classes) {
                if (line==null) continue;
                final int iSpace=line.indexOf(' ');
                final String cn=iSpace<0 ? line : line.substring(0,iSpace);
                final String menu=iSpace>0 ? line.substring(iSpace+1) : "menu not specified";
                Class c=MAP_LOADED_CLASSES.get(cn);
                if (c==null) MAP_LOADED_CLASSES.put(cn,c=Insecure.findClassByName(cn));
                if (c!=null) {
                    vPLUGINS .add(c);
                    mapPluginsToMenu.put(c,menu);
                } else ERROR_LOAD_CLASSES.append("warning: could not load class ").append(line).append("\n");
            }
            if (jList!=null) jList.clearSelection();
        }
        return vPLUGINS.asArray();
    }
    private static boolean userPluginsActivated;
    private static Class lastSelectedClass;
    private final  static ChTextArea LOG=null;
    private final static boolean isLog=false;
    private static ChJList jList,jListMenus;
    private static Vector<Class> vClasses;
    private static Vector<String> vMenus;
    private static JFrame frame;

    private static void update() {
        if (isLog) LOG.appnd("update()\n");
        final int idx=jListMenus.getSelectedIndex();
        String menu=idx<0 || idx>vMenus.size() ? "" : vMenus.get(idx);
        vMenus.clear();
        for(Class c : getUserPlugins()) {
            final String men=mapPluginsToMenu.get(c);
            if (isLog) LOG.appnd("c="+c," ",men,"\n");
            if (!vMenus.contains(men)) vMenus.add(men);
        }
        sortArry(vMenus);
        if (vMenus.size()>0) {
            int iMenu=vMenus.indexOf(menu);
            if (iMenu<0) menu=vMenus.get(iMenu=0);
            jListMenus.setSelI(iMenu);
        }
        if (menu!=null) {
            vClasses.clear();
            for(Class c : getUserPlugins()) {
                if (menu.equals(mapPluginsToMenu.get(c)))
                    vClasses.add(c);
            }
        }
        jList.clearSelection();
        ChDelay.afterMS(EDT,ChDelay.CONTENTS_CHANGED,jListMenus,1,null);
        ChDelay.afterMS(EDT,ChDelay.CONTENTS_CHANGED,jList,1,null);
        if (frame!=null) frame.toFront();
        for(int i=INTERFACES.length; --i>=0;)loadClassesIntoVectors(i);
    }
    protected static void dialogStaticPlugins() {
        userPluginsActivated=true;
        if (frame==null) {
            jList=new ChJList(vClasses=new Vector(),ChJTable.CLASS_RENDERER|ChJTable.SINGLE_SELECTION);
            jListMenus=new ChJList(vMenus=new Vector(), ChJTable.SINGLE_SELECTION);
            jList.li(instance());
            jListMenus.li(instance());
            final ChRenderer render=new ChRenderer();
            render.LABEL.setHorizontalAlignment(SwingConstants.RIGHT);
            jListMenus.setCellRenderer(render);
            final JComponent
                pCenter=new JScrollPane(jList),
                but=new ChButton("DOWNLOAD").t("Download/update plugins").li(instance()),
                butDetail=new ChButton("CP").t("Technical","details",dim(64,32)).li(instance()),
                pNorth=newCNSEW(pnl(but,dirPlugins()),null,null,smallHelpBut(StrapPlugins.class),null),
                pSouth=pnl(BUT_APPLY,BUT_DOCUMENTAION," for developers:",BUT_RELOAD,BUT_SOURCE,butDetail);
            frame=new ChFrame("Plugins").ad(newCNSEW(pCenter,pNorth,pSouth,null,jListMenus)).size(666,333);
            if (isLog) LOG.showInFrame("Log",0);
            addActLi(instance(),HtmlDoc.class);
        }
        update();
        frame.show();
    }
    private static ChFrame frameDownload;
    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        if (q==jListMenus) {lastSelectedClass=null; update();}
        if (q==jList) {
            final int i=jList.getSelectedIndex();
            lastSelectedClass= i>=0 && i<vClasses.size() ? vClasses.get(i) : null;
            BUT_DOCUMENTAION.enabled(hasHelp(lastSelectedClass));
            BUT_SOURCE.enabled(hasSource(lastSelectedClass));
            BUT_APPLY.enabled(lastSelectedClass!=null);
            BUT_RELOAD.enabled(lastSelectedClass!=null);
        }
        if (cmd==ACTION_DOWNLOAD_PLUGIN) update();
        if (cmd=="DOWNLOAD") {
            if (frameDownload==null) {
                ChDelay.newThread(this,"DOWNLOAD",null).start();
                frameDownload=new ChFrame("http://www.bioinformatics.nu/").size(666,333).ad(new JLabel("<html><body>contacting Wiki site.<br>Please wait</body></html>"));
            }
            frameDownload.shw(0);
        }
        if (cmd=="CP") {
            new ChTextViewer("This is the Java class-path:\n\n"+joinLines(ChClassLoader.getInstance().getURLs())).shw(0);
        }
        if (lastSelectedClass!=null) {
            if (q==BUT_APPLY) applyPlugin_(lastSelectedClass);
            if (q==BUT_RELOAD) {
                final String cn=lastSelectedClass.getName();
                try{
                    ChClassLoader.newInstance();
                    MAP_LOADED_CLASSES.put(cn,lastSelectedClass=Insecure.findClassByName(cn));
                    for(int i=INTERFACES.length; --i>=0;)loadClassesIntoVectors(i);
                    applyPlugin_(lastSelectedClass);
                }catch(Exception e){}
            }
            if (q==BUT_DOCUMENTAION) showHelp(lastSelectedClass);
            if (q==BUT_SOURCE) showJavaSource(lastSelectedClass.getName(),(String[])null);
        }
    }
    public Object run(String id, Object arg) {
        if (id=="DOWNLOAD") {
            final String WIKI="http://www.bioinformatics.org/strap/wiki/Strap/Plugins/";
            final ByteArray wiki=readBytes(inStreamT(99,WIKI, 0L));
            if (wiki==null)  error("Could not download "+WIKI+"<br>Are the Web proxy settings OK?");
            else {
                final int E=wiki.getEnd();
                final byte T[]=wiki.getBytes();
                int count=0;
                final ByteArray txt=new ByteArray(9999)
                    .append(
                            "This text contains URLs of WIKI:Jar-files with the suffix ?plugins\n"+
                            "To download/update a plugin, click the respective Jar-file.\n"+
                            "The clicked Jar-file will be downloaded to the directory ").append(filepathUnixStyle(dirJars()))
                    .append("\nThe plugin[s] provided by this Java archive appear in the list of installed plugins in the Plugins panel.\n\n");
                for(int pos=0; (pos=strstr("?plugins</a>", T, pos+1, E))>0;) {
                    txt.append(" * ")
                        .append(T, strchr(STRSTR_E|STRSTR_PREV, '>',T,pos,0)+1, pos+8)
                        .append("\n   ")
                        .append(T, pos+(4+8), strchr(STRSTR_E,'<', T, pos+(4+8),E));
                    count++;
                }
                txt.append("\nTo contribute your application please modify the wiki page\n")
                    .append(WIKI).append("?action=edit   Enter passwd \"strap\"\n\n")
                    .append("Security warning:\n"+
                            "Plugins are potentially dangerous and may contain WIKI:Malware.\n"+
                            "You must not download plugins that you do not trust.\n"+
                            "Trusted URLs highlighted green\n"+
                            "If in doubt, do not run the plugin, or test it in a safe environment. \nSee ")
                    .append(URL_STRAP).append("security.html.");
                if (count==0)
                    txt.setEnd(0)
                        .append("Error in ").append(WIKI)
                        .append("\nMissing entries of the form http://..pluginfile.jar?plugins\nPlease send report christophgil@googlemail.com.");
                ChDelay.invokeLater(this,"SETTEXT",txt);
            }
        }
        if (id=="SETTEXT") {
            frameDownload.getContentPane().removeAll();
            final ChTextViewer ta=new ChTextViewer((ByteArray)arg);
            final ChTextComponents tools=ChTextComponents.tools(ta);
            frameDownload.ad(tools.scrollPane());
            for(String url: TRUSTED_PLUGINS) tools.highlightOccurrence(url ,null,null,0, C(0xff00,99));
            tools.underlineRefs(0);
            addActLi(this,tools);
            ta.repaint();
        }
        return null;
    }
    private final static ChButton
        BUT_APPLY=new ChButton(null,ChButton.DISABLED).t(ChButton.GO).li(instance()),
        BUT_RELOAD=new ChButton(null,ChButton.DISABLED).t("Reload","class",dim(64,32)).li(instance()),
        BUT_DOCUMENTAION=new ChButton("Documentation",ChButton.DISABLED).li(instance()),
        BUT_SOURCE=new ChButton(null,ChButton.DISABLED).t("Source","code",dim(64,32)).li(instance()).li(instance());
    // ----------------------------------------
				private final static Object DIALOG_AND_INTERFACES[][]={
        {ProteinParser.class, new String[]{CLASS_MainDialogProjectFiles,CLASS_DialogFetchSRS,CLASS_DialogFetchPdb}},
        {ResidueSelection.class, new String[]{CLASS_DialogSelectionOfResiduesMain,CLASS_DialogCopyAnnotations}},
        {SequenceAligner.class, new String[]{CLASS_DialogAlign}},
        {Superimpose3D.class, new String[]{CLASS_DialogSuperimpose3D,CLASS_DialogCompareProteins}},
        {SequenceDisSimilarity.class, new String[]{CLASS_DialogCompareProteins}},
        {ProteinProteinValue.class, new String[]{CLASS_DialogCompareProteins,CLASS_DialogBarChart}},
        {ProteinViewer.class, new String[]{CLASS_Dialog3DViewer}},
        {AlignmentWriter.class, new String[]{CLASS_DialogExportAlignment}},
        {ProteinWriter.class, new String[]{CLASS_DialogExportProteins}},
        {SecondaryStructure_Predictor.class, new String[]{CLASS_DialogPredictSecondaryStructures}},
        {TransmembraneHelix_Predictor.class, new String[]{CLASS_DialogPredictTmHelices2}},
        {CoiledCoil_Predictor.class, new String[]{CLASS_DialogPredictCoiledCoil}},
        {ProteinsSorter.class, new String[]{CLASS_DialogSortProteins}},
        {ValueOfProtein.class, new String[]{CLASS_DialogBarChart}},
        {SequenceBlaster.class, new String[]{CLASS_DialogBlast}},
        {PhylogeneticTree.class,new String[]{CLASS_DialogPhylogeneticTree}}
    };
    protected static String[] getDialogForPlugin(Class c) {
        if (c!=null)
            for(Object[] oo:DIALOG_AND_INTERFACES)
                if (((Class)oo[0]).isAssignableFrom(c))
                    return (String[]) oo[1];
        return new String[0];
    }
    public static void applyPlugin(Class c) { applyPlugin_(c); }
    public static void applyPlugin(ProxyClass c) { applyPlugin_(c);}
    private static void applyPlugin_(Object proxyOrClass) {
        final StrapAlign align=StrapAlign.getInstance();

        final Class clas;
        if (proxyOrClass instanceof ProxyClass) clas=((ProxyClass)proxyOrClass).getClassInstance();
        else {

            clas=(Class)proxyOrClass;
        }
        if (clas==null) return;
        try {
            if (StrapExtension.class.isAssignableFrom(clas)) {
                final StrapExtension ext=HotswapUtil.instance(proxyOrClass,StrapExtension.class,true);
                if (ext!=null) {  align.addListener(ext);   ext.run(); }
            }
            for(String sDialog : getDialogForPlugin(clas)) {
                final Object dialog=align.addDialog(sDialog,false);
                for(Object[] oo:DIALOG_AND_INTERFACES){
                    final Class c=(Class)oo[0];
                    if (c.isAssignableFrom(clas)) allClassesVV()[class2index(c)].add(clas);
                }
                final ChCombo comboClass=gcp(KEY_CHOICE_CLASS, dialog, ChCombo.class);
                if (comboClass!=null)  comboClass.s(proxyOrClass);
            }
            if (ValueOfResidue.class.isAssignableFrom(clas))
                ((DialogPlot)align.addDialog(CLASS_DialogPlot,false)).setVoR(proxyOrClass);
            if (ValueOfAlignPosition.class.isAssignableFrom(clas))
                ((DialogPlot)align.addDialog(CLASS_DialogPlot,false)).setVoAP(proxyOrClass);
            if (SequenceDisSimilarity.class.isAssignableFrom(clas)) {
                final Object d=align.addDialog(CLASS_DialogCompareProteins,false);
                final ChCombo comboClass=gcp(KEY_CHOICE_CLASS, d, ChCombo.class);
                if (comboClass!=null) {
                    comboClass.s(proxyOrClass);
                    final Object saa=((NeedsSharedInstance)d).getSharedInstance();
                    if (saa instanceof SequenceDisSimilarityAsAligned) ((SequenceDisSimilarityAsAligned)saa).setOtherClass(clas);
                }
            }

        } catch(Exception e){ debug("StrapPlugins.applyPlugin "+proxyOrClass); stckTrc(e);}
    }

    public static SequenceAligner instanceSA(Object o,int numProteins, boolean error) {
        final SequenceAligner sa= HotswapUtil.instance(o,SequenceAligner.class,error);
        if (sa==null) return null;
        final boolean only2=(sa.getPropertyFlags()&SequenceAligner.PROPERTY_ONLY_TWO_SEQUENCES)!=0 || sa instanceof Superimpose3D;
        return numProteins>2 && only2 ? new MultiFromPairAligner(o) : sa;
    }
}
