package charite.christo.strap;
import java.lang.ref.*;
import java.util.*;
import java.io.File;
import java.net.URL;
import charite.christo.*;
import charite.christo.strap.extensions.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
public class StrapProteinUtils implements Comparator {
    private static StrapProteinUtils _inst;
    private static StrapProteinUtils instance() { return _inst==null ? _inst=new StrapProteinUtils() : _inst;}
    public Object run(String id, Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[])arg : null;
        return null;
    }
    /* <<< Threading <<< */
    /* ---------------------------------------- */
    /* >>> 3D >>> */
    static { if (isMyComputer()) threadInferCoordinates(Protein.NONE, new String[0], ChLog.NO_LOG,  0L ); }
    public final static String INFER_COORDINATES_AUTO[]=new String[0];
    public static Thread threadInferCoordinates(Protein pp[],String[] pdbID,  ChLog chLog, long options) {
        final Object[] para={pp,pdbID,chLog,longObjct(options)};
        return threadM("inferCoordinates_TS",StrapProteinUtils.class, para);
    }
    private final static Object SYNC_map=new Object();
    private static Reference<Map> refMapFileP3d;
    public static void inferCoordinates_TS(Protein pp[],String[] pdbID,  ChLog chLog0, long options) { /* thread save */
        final ChLog chLog;
        final AsciiView ta;
        if ( 0!=(options&SimilarPdb.INFER_3D_LOG) && !ChLog.isEnabled(chLog0)) {
            ta=new AsciiView(999);
            chLog=new ChLog(ta.byteArray());
        } else { chLog=chLog0; ta=null;}
        final boolean doLog=ChLog.isEnabled(chLog);
        final StrapAlign align=StrapAlign.getInstance();
        if (chLog!=null) chLog.a(
                    "INFERING 3D-COORIDINATES\n\n"+
                    "Downloading pdb files from the pdb-server.\n"+
                    "With the help of WIKI:ClustalW, the 3D-coordinates will be assigned to the corresponding residues.\n");
        Map<String,StrapProtein[]> mapF3d;
        synchronized(SYNC_map) {
            mapF3d=deref(refMapFileP3d,Map.class);
            if (mapF3d==null) refMapFileP3d=new SoftReference(mapF3d=new HashMap());
        }
        final Protein[][] pp3d_array=new Protein[pp.length][];
        for(int iP=pp.length; --iP>=0;) {
            final Protein p=pp[iP];
            if (p==null) continue;
            final String pdb_chain=pdbID==INFER_COORDINATES_AUTO ? SimilarPdb.getBestPdbID(p) : get(iP,pdbID);
            if (sze(pdb_chain)==0) continue;
            final char chain=pdbChain(pdb_chain);
            StrapProtein[] pp3d;
            synchronized(SYNC_map) { pp3d=mapF3d.get(pdb_chain);}
            if (pp3d==null) {
                pp3d=StrapAlign.downloadPdbAndCreateProteins(pdbID(pdb_chain),chain, align!=null?align.getProteins() : StrapProtein.NONE);
                if (doLog) chLog.a("downloaded:  ").a(pdb_chain).a(pp3d).a('\n');
                if (sze(pp3d)==0) continue;
                synchronized(SYNC_map) {mapF3d.put(pdb_chain,pp3d); }
            }
            pp3d_array[iP]=pp3d;
        }
        inEDT( thread_infc_Blocking(pp,pp3d_array,chLog,options));
        if (ta!=null) ta.showInFrame("Infer coordinates",ChFrame.CENTER);
    }
    private static Thread thread_infc_Blocking(Protein pp[],Protein[][] pp3d_array, ChLog chLog, long options) {
        final Object para[]={pp,pp3d_array,chLog,longObjct(options)};
        return threadM("infc_Blocking", StrapProteinUtils.class, para);
    }
    public static boolean infc_Blocking(Protein pp[],Protein[][] pp3d_array, ChLog chLog,  long options) {
        final boolean doLog=ChLog.isEnabled(chLog);
        int success=0;
        for(int iP=pp.length; --iP>=0;) {
            final Protein p=pp[iP]; // (StrapProtein)tableRow[0];
            if (p==null) continue;
            final Protein[] pp3d=pp3d_array[iP];
            if (pp3d==null) continue;
            final float score=p.inferCoordinates(pp3d, charite.christo.strap.extensions.PairAlignerNeoBioPROXY.class, chLog);
            if (ChLog.isEnabled(chLog)) chLog.a(p).a(' ').a(" Align-score=").a(score).a(Float.isNaN(score)?" failure\n":"\n");
            if (!Float.isNaN(score)) {
                //p.setOnlyChains(String.valueOf(chain));
                success++;
            }
        }
        if (doLog) chLog.a('\n');
        if (success>0) {
            ChMsg.sound(ChMsg.SOUND_SUCCESS);
            final String log1="\nNow you can view these proteins three-dimensionally.\n"+
                "Place the alignment cursor on the respective protein and press the \"3d-structure\" toolbar.\n"+
                "Use Drag'n Drop to move more proteins into the same view.\n";
            if (doLog) chLog.a(log1);
            if ( (options&SimilarPdb.INFER_3D_SHOW_BACKBONE)!=0) StrapAlign.newBackbone3D_View(pp,0L);
        } else {
            //if (doLog) new ChTextViewer(log1.append("\nNo 3D-structure loaded\n")).title("Loading 3D").shw(0);
        }
        new StrapEvent(StrapProteinUtils.class, StrapEvent.ATOM_COORDINATES_CHANGED).run();
        return success>0;
    }
    /* <<< Infer coordinates <<< */
    /* ---------------------------------------- */
    /* >>> Alignment  >>> */
    private final static void rm(boolean[] shouldStop, Object aligner) {
        if (shouldStop==CAN_BE_STOPPED) {
            CanBeStopped.vALIGNMENTS.remove(shouldStop);
            CanBeStopped.vALIGNMENTS.remove(aligner);
        }
    }
    private static boolean[] isInterrupted(boolean shouldStop[], Object aligner) {
        if (shouldStop==CAN_BE_STOPPED) {
            final boolean[] bb={false};
            CanBeStopped.vALIGNMENTS.add(bb);
            CanBeStopped.vALIGNMENTS.add(aligner);
            return bb;
        } else return shouldStop;
    }
    public final static long ALIGN_CHANGE_ORDER=1<<1;

    public static SequenceAligner alignerInstance() {
        return
            Insecure.EXEC_ALLOWED ? new ClustalW_3D() :
            new MultiFromPairAligner(PairAlignerNeoBioPROXY.class).setSuperpositionMethod(Superimpose_CEPROXY.class);
    }
    public static Thread threadAlignProteins(SequenceAligner aligner,List<StrapProtein> vP, boolean shouldStop[], long options) {
       final Object para[]={aligner, vP, shouldStop,longObjct(options)};
        return threadM("alignProteins",StrapProteinUtils.class,para);
    }
    public static SequenceAligner alignProteins(SequenceAligner aligner0, List<StrapProtein> vP, boolean shouldStop[], long options) {
        final StrapProtein pp[]=toArry(vP,StrapProtein.class);
        if (sze(pp)==0) return null;
        final SequenceAligner aligner=aligner0!=null ? aligner0 : alignerInstance();
        final boolean isInterrupted[]=isInterrupted(shouldStop,aligner);
        StrapProteinUtils.setSeqs(aligner, pp);
        if (aligner instanceof NeedsProteins) ((NeedsProteins)aligner).setProteins(pp);
        addActLi(StrapAlign.li(),aligner);
        aligner.setSequences(Protein.getResidueType(pp));
        aligner.compute();
        CanBeStopped.vALIGNMENTS.remove(aligner);
        final byte[][] aligned=aligner.getAlignedSequences();
        if (sze(aligned)==0) putln("Error: StrapProteinUtils.alignProteins(): aligned=null");
        else if (aligned.length<pp.length) putln("Error: StrapProteinUtils.alignProteins(): aligned.length="+aligned.length+" pp.length="+pp.length);
        else {
            final StrapAlign align=StrapAlign.getInstance();
            if (aligner instanceof MultiFromPairAligner) ((MultiFromPairAligner)aligner).setIsInterrupted(isInterrupted);
            for(int i=pp.length; --i>=0;)  pp[i].inferGapsFromGappedSequence(aligned[i]);
            if (align!=null) {
                final int ii[]=0!=(options&ALIGN_CHANGE_ORDER) && aligner instanceof SequenceAlignerSorting ? ((SequenceAlignerSorting)aligner).getIndicesOfSequences(): null;
                if (sze(ii)>=pp.length) {
                    for(int i=ii.length; --i>=0;) {
                        pp[i].setPreferedOrder(ii[i]);
                    }
                    align.inferOrderOfProteins(pp, StrapAlign.SORT_PREFORDER);
                }
                align.setChanged();
            }
            StrapAlign.dispatchLater(StrapEvent.ALIGNMENT_CHANGED,1);
        }
        rm(shouldStop,aligner);
        return aligner;
    }
    /* <<< Alignment <<< */
    /* ---------------------------------------- */
    /* >>> Superimpose >>> */
    private synchronized static Superimpose3D.Result superimposeTwoProteins(Protein pM,Protein pR) {
        if (pM==null || pR==null) return null;
        Superimpose3D.Result r=pM.hashTMalign().get(pR);
        if (r==null) {
            r=getSuperpositionResult( StrapAlign.getDefaultSuperposition(),pM,pR);
            pM.hashTMalign().put(pR,r);
        }
        return r;
    }
    private synchronized static Superimpose3D.Result getSuperpositionResult(Object clazz,Protein pM,Protein pR) {
        if (pM==null || pR==null) return null;
        final Superimpose3D s= (Superimpose3D)(clazz instanceof Superimpose3D ? clazz : HotswapUtil.instance(clazz,Superimpose3D.class,true));
        s.setMobileProtein(pM);
        s.setReferenceProtein(pR);
        CanBeStopped.vALIGNMENTS.add(s);
        s.compute();
        CanBeStopped.vALIGNMENTS.remove(s);
        final Superimpose3D.Result result=s.getResult();
        if (result==null) putln("Error in StrapProteinUtils#getSuperpositionResult result==null for pM=",pM," pR ",pR);
        return result;
    }

    public final static long SUPERIMP_EVENT=1<<1;
    public final static boolean IS_SUPERIMPOSING[]={false};
    public static Thread threadSuperimposeProteins(List<? extends Protein> vP,boolean isInterrupted[], long options) { return threadSuperimposeProteins(vP,isInterrupted,null,options); }
    public static Thread threadSuperimposeProteins(List<? extends Protein> vP,boolean isInterrupted[], Protein[] pBest, long options) {
        final Object para[]={vP,isInterrupted,longObjct(options)};
        return threadM("superimposeProteins",StrapProteinUtils.class, para, pBest);
    }
    private static int countSP;
    public static Protein superimposeProteins(List<? extends Protein> vP, boolean shouldStop[], long options) {
        final boolean[] isInterrupted=isInterrupted(shouldStop, null);
        final StrapProtein pp[]=toArry(vP,StrapProtein.class);
        IS_SUPERIMPOSING[0]=true;
        Protein pBest=null;
        float bestScore=Float.MIN_VALUE;
        long timePG=0;
        final int N=pp.length*pp.length;

        int count=0;
        for(Protein pi : pp) {
            float scoreSum=0;
            for(Protein pj : pp) {
                if (get(0,isInterrupted)) { putln("Interrupted superimposeProteins");break;}
                ++count;
                if (pi!=null && pj!=null && pi!=pj && pi.hasCalpha() && pj.hasCalpha()) {
                    Superimpose3D.Result r=superimposeTwoProteins(pi,pj);
                    if (r!=null) scoreSum+=r.getScore();
                    if (System.currentTimeMillis()-timePG>200) {
                        timePG=System.currentTimeMillis();
                        StrapAlign.showProgress("#"+ ++countSP+"  "+count+"/"+N+" "+pi+" ./. "+pj);
                    }
                }
            }
            if (scoreSum>bestScore) { bestScore=scoreSum; pBest=pi;}
        }

        if (pBest!=null) {
            StrapAlign.showProgress(count+" Superpositions done. Reference protein="+pBest);
            pBest.setRotationAndTranslation(null);
            for(Protein p : pp) {
                if (!get(0,isInterrupted) && p!=null && p!=pBest) {
                    final Superimpose3D.Result r=superimposeTwoProteins(p,pBest);
                    if (r!=null) p.setRotationAndTranslation(r.getMatrix());
                }
            }
            pBest.setRotationAndTranslation(null);
        }
        IS_SUPERIMPOSING[0]=false;
        rm(shouldStop,null);
        if (0!=(options&SUPERIMP_EVENT) && sze(vP)>0) new StrapEvent(StrapProteinUtils.class, StrapEvent.PROTEIN_3D_MOVED).setParameters(vP.toArray()).run();
        return pBest;
    }

    /* <<<  superimposeProteins <<< */
    /* ---------------------------------------- */
    /* >>> Continue superposition till interrupted >>> */

    public static void loopSuperimposeProteinsV(UniqueList<? extends Protein> vP, boolean[] shouldStop, boolean[] isSuperimposing, long options) {
        final boolean isInterrupted[]=isInterrupted(shouldStop,null);
        do {
            final int modi=vP.getModificationCount();
            superimposeProteins(vP,isInterrupted,options);
            while(modi==vP.getModificationCount() && !get(0,isInterrupted)) sleep(444);
        } while(!get(0,isInterrupted));
        if (sze(isSuperimposing)>0) isSuperimposing[0]=false;
        rm(shouldStop,null);
    }

    public static Thread threadLoopSuperimposeProteinsV(UniqueList<? extends Protein> vP, boolean[] shouldStop, boolean[] isSuperimposing, long options) {
        final Object[] para={vP,shouldStop,isSuperimposing, longObjct(options)};
        return threadM("loopSuperimposeProteinsV", StrapProteinUtils.class,para);
    }
    /* <<<  Continue superposition  <<< */
    /* ---------------------------------------- */
    /* >>> View Comparison >>> */
    public static void viewComparison(StrapProtein p0, StrapProtein p1, Object clazzOrInstance) {
        if (p0==null || p1==null || clazzOrInstance==null) return;
        Object cl=clazzOrInstance;
        while(cl!=null) {
            final Object instance= cl instanceof Class || cl instanceof charite.christo.hotswap.ProxyClass ? HotswapUtil.instance(cl) : cl;
            if (isAssignblFrm(Superimpose3D.class,instance) && p0.hasCalpha()  && p1.hasCalpha() ) {
                final Protein3d p3D=StrapAlign.newBackbone3D_View(new StrapProtein[]{p0,p1}, ProteinViewerUtils.OPEN_NOT_ADD_PANEL);
                Superimpose3D.Result result=StrapProteinUtils.getSuperpositionResult(instance,p0,p1);
                final ProteinViewer pv=p3D.getView(p0);
                if (pv!=null) {
                    pv.transform(result.getMatrix());
                    new ChFrame("3d "+p0+" "+p1).ad(p3D).shw(ChFrame.AT_CLICK);
                }
                break;
            }
            if (isAssignblFrm(SequenceAligner.class,instance)) {
                final SequenceAligner aligner=(SequenceAligner)instance;
                StrapProteinUtils.setSeqs(aligner, new StrapProtein[]{p0,p1});
                if (aligner instanceof NeedsProteins) ((NeedsProteins)aligner).setProteins(p0,p1);
                aligner.compute();
                final byte[] gapped[]=aligner.getAlignedSequences();
                if (gapped==null) {
                    final java.awt.Container ctrlP=getCtrlPanel(aligner);
                    final String MSG="The computation did not produce a result. "+(CacheResult.isEnabled() ? "<br>Try again with disabled cache." : "");
                    if (ctrlP!=null) {
                        pcp(KEY_NORTH_PANEL,MSG,ctrlP);
                        ChFrame.frame(shrtClasNam(aligner),ctrlP,ChFrame.AT_CLICK);
                    }
                    error(MSG);
                }
                final StrapProtein a0=new StrapProtein(), a1=new StrapProtein();
                a0.setGappedSequence(gapped[0], MAX_INT);
                a1.setGappedSequence(gapped[1],MAX_INT);
                a0.setName(p0.getName());
                a1.setName(p1.getName());
                final ExportAlignment w=new ExportAlignment();
                w.setProteins(a0,a1);
                final StringBuffer sb=new StringBuffer(999);
                w.getText(charite.christo.strap.interfaces.AlignmentWriter.HTML,sb);
                new ChTextViewer(sb).title(p0+" "+p1).shw(ChFrame.AT_CLICK);
                break;
            }
            cl= isAssignblFrm(AbstractAlignAndCompare.class, cl) ? ((AbstractAlignAndCompare)cl).getOtherClass() : null;
        }
    }
    /* <<< View Comparison <<< */
    /* ---------------------------------------- */
    /* >>> Species Icons >>> */
    public static String speciesName2Icon(final StrapProtein p) {
        if (p==null) return null;
        String s=p.getOrganismScientific();
        return speciesName2Icon(s!=null ? s : p.getOrganism());
    }
    public static String speciesName2Icon(final String name) {
        final int c0=chrAt(0,name)|32, c1=chrAt(1,name)|32,nL=sze(name);
        String baseURL=null;
        final String lines[]=getCustomSettings(Customize.speciesIcons);
        if (sze(lines)*nL==0) return null;
        nextL:
        for(String line : lines) {
            line=line.trim();
            if (line.startsWith("http://")) {
                baseURL=line;
                continue;
            }
            final int L=line.length(), spc=line.indexOf(' ');
            if (spc<0 || L-spc<5 || c0!=(line.charAt(0)|32) || c1!=(line.charAt(1)|32)) continue;
            for(int i=0; i<L && i<nL; i++) {
                final char c=line.charAt(i);
                if (i==nL) {
                    if (c==' ' || c=='*') break;
                    else continue nextL;
                }
                if (c==' ') {
                    if (is(LETTR_DIGT_US,name,i)) continue nextL;
                    break;
                }
                if ((c|32) != (name.charAt(i)|32)) continue nextL;
            }
            final int iconB=nxtE(NO_SPC, line, spc,L);
            if (equls("NONE",line,iconB)) return "0";
            if (equls("http://",line, iconB)) return line.substring(iconB);
            else return baseURL==null ? null : baseURL+(baseURL.endsWith("/")?"":"/")+line.substring(iconB);
        }
        return null;
    }
    public static void setProteinIcon(StrapProtein pp[], boolean overwrite,char type) {
        for(StrapProtein p : pp) {
            if (p==null || !overwrite && sze(p.iconLink())>0) continue;
												final String
                pdbID=p.getPdbID(),
																s=  (type=='3'||type=='A')  && pdbID!=null ?
                URL_RCSB_IMAGES+withoutLastComponent(delPfx("pdb:",pdbID.toLowerCase()),':')+"_bio_r_250.jpg" :
																type=='S' || type=='A' ? speciesName2Icon(p) :
																null;
            if (s!=null && s!="0") p.setIconImage(s);
        }
    }
    /* <<< Species Icons <<< */
    /* ---------------------------------------- */
    /* >>> MSA  >>> */

    public static Thread threadDownloadOriginalProteins(StrapProtein pp[],long webOptions) {
        return threadDownloadOriginalProteins(pp,null,null,webOptions);
    }
    public static Thread threadDownloadOriginalProteins(StrapProtein pp[], String key, File files[], long webOptions) {
        final Object[] para={pp,key,files,longObjct(webOptions)};
        return threadM("downloadOriginalProteins", StrapProteinUtils.class,para);
    }
    public static void downloadOriginalProteins(StrapProtein pp[], long webOptions) { downloadOriginalProteins(pp,(String)null,(File[])null,webOptions); }
    public static void downloadOriginalProteins(StrapProtein pp[], String key, File files[], long webOptions) {
        if (files==null) {
            if (key==null) {
                final String keys[]=Hyperrefs.getDatabases(Hyperrefs.PROTEIN_FILE)[0];
                for(int iK=-1; iK<keys.length; iK++) {
                    final String k=iK==-1 ? "PDB:" : keys[iK];
                    final char k0=k.charAt(0);
                    boolean keyExists=false;
                    for(Protein p : pp) {
                        final String a=p.getAccessionID();
                        if (a!=null  && a.charAt(0)==k0 && (p.isInMsfFile()||p.getFile()==null) && a.startsWith(k)) {
                            keyExists=true;
                            break;
                        }
                    }
                    if (keyExists) threadDownloadOriginalProteins(pp, k, (File[])null, webOptions).start();
                }
            } else {
                File ff[]=null;
                final String[] aa=new String[pp.length];
                for(int iP=pp.length; --iP>=0;) aa[iP]=pp[iP].getAccessionID();
                final File[] ffEquivalent=null;
                for(int iP=pp.length; --iP>=0;) {
                    final String a=aa[iP];
                    if (a==null || !a.startsWith(key)) continue;
                    final File f;
                    if (a.startsWith("PDB:")) {
                        f= downloadPdbChain(a, pdbChain(a));
                    } else {
                        f=download_TS(url(Hyperrefs.toUrlString(a,Hyperrefs.PROTEIN_FILE)), MAX_INT, ffEquivalent, (ChRunnable)StrapAlign.getInstance());
                    }
                    if (f!=null) (ff==null?ff=new File[iP+1] : ff)[iP]=f;
                }
                if (ff!=null) inEDT(threadDownloadOriginalProteins(pp,"",ff, webOptions));
            }
        } else {
            assertEDT();
            ByteArray sbError=null;
            for(int iP=mini(pp.length,files.length); --iP>=0;) {
                final StrapProtein p=pp[iP];
                final File f=files[iP];
                final ByteArray ba=readBytes(f);
                final String a=p!=null ? p.getAccessionID() : null;
                if (a==null || sze(ba)==0) continue;
                final byte[] seq=p.getResidueTypeExactLength();
                int firstAmino=-1;
                String matchingCDS=null;
                if (a.startsWith("PDB:")) {
                    p.parse(ba, new ProteinParser[]{new PDB_Parser()},  ProteinParser.SEQUENCE_FEATURES);
                    firstAmino=strstr(STRSTR_IC,seq, p.getResidueTypeExactLength());
                    if (firstAmino<0)  {
                        final AsciiView ta=new AsciiView(999);
                        final String title="Infering coordinates";
                        ta.showInFrame(title,0);
                        StrapProteinUtils.inferCoordinates_TS(new StrapProtein[]{p}, ss(a), new ChLog(ta.byteArray()), 0L);
                        ta.showInFrame(title,0);
                    }
                } else {
                    final StrapProtein pTmp=new StrapProtein();
                    pTmp.parse((File)null, ba, 0);
                    final String cds[]=pTmp.getCDS();
                    byte[] newSeq=pTmp.getResidueTypeExactLength();
                    if (sze(cds)>0) {
                        for(int iCDS=cds.length; --iCDS>=0;) {
                            pTmp.applyCDS_expression(cds[iCDS]);
                            firstAmino=strstr(STRSTR_IC,seq, newSeq=pTmp.getResidueTypeExactLength());
                            if ( firstAmino>=0) {
                                matchingCDS=cds[iCDS];
                                break;
                            }
                        }
                    } else firstAmino=strstr(STRSTR_IC,seq, newSeq);
                    if (firstAmino<0) {
                        if (sbError==null) sbError=new ByteArray(3333);
                        sbError.append("Warning for "+a+"  The sequence in downloaded file does not match \n").append(newSeq).append('\n').append(seq).append("\n\n\n");
                        if (strstr("Runtime Error",ba)>0) delFileOnExit(f);
                    }
                }
                p.setIsInMsfFile(false);
                if (firstAmino>=0) {
                    p.parse((File)null, ba, 0);
                    p.setResidueSubset((firstAmino+1)+"-"+(firstAmino+seq.length));
                    p.applyCDS_expression(matchingCDS);
                    p.setFile(f);
                    p.setAccessionID(a);
                    if ((webOptions&StrapAlign.WEB_OPTION_RENAME_SWISS)!=0) {
                        final String newName=DialogRenameProteins.newName(p,DialogRenameProteins.PRC3_DROME);
                        if (sze(newName)>0) p.setName(newName);
                    }
                }
            }
            if (sbError!=null) new AsciiView(sbError).showInFrame("Error infering sequence files",ChFrame.CENTER);
            StrapProteinUtils.setProteinIcon(pp, true ,'A');
        }
    }

    public static StrapProtein[] ppPeptide(StrapProtein pp0[]) {
        final StrapProtein pp[]=pp0.clone();
        for(int i=pp.length; --i>=0;) if (pp[i]!=null && pp[i].countResidues()==0) pp[i]=null;
        return rmNull(pp,StrapProtein.class);
    }

    static void structureNeighbors(Protein pp[]) {
        final StrapProtein ppN[]=new StrapProtein[pp.length];
        String pdbId=null;
        for(int iP=pp.length; --iP>=0;) {
            final Protein p=pp[iP];
            if (p==null || sze(p.getChainsAsString())!=1)  continue;
            final String id=p.getPdbID()!=null ? p.getPdbID() : p.getInferredPdbID();
            if (id==null || (!id.equals(pdbId)  && pdbId!=null)) continue;
            pdbId=id;
            ppN[iP]=(StrapProtein)p;
        }
        if (pdbId!=null) new DialogNeighbors(ppN, pdbId);
    }

  /* <<<  <<< */
  /* ---------------------------------------- */
  /* >>> Comparator >>> */

    public static void sortVisibleOrder(StrapProtein pp[]) {
        final StrapProtein ppVis[]=StrapAlign.getVisibleProteins();
        for(int iP=ppVis.length; --iP>=0;) ppVis[iP].setRow(iP);
        sortArry(pp, getComparator('R'));
    }

  final static StrapProteinUtils comparators[]=new StrapProteinUtils['z'];
  private char comparator_type;
  public static Comparator getComparator(char type) {
    if (comparators[type]==null) (comparators[type]=new StrapProteinUtils()).comparator_type=type;
    return comparators[type];
  }
  public  int compare(Object o1, Object o2) {
    final StrapProtein
      p1=o1 instanceof StrapProtein ? (StrapProtein)o1 : get(0,o1,StrapProtein.class),
      p2=o2 instanceof StrapProtein ? (StrapProtein)o2 : get(0,o2,StrapProtein.class);
    if (p1==p2)  return 0;
    if (p1==null) return -1;
    if (p2==null) return 1;

    if (comparator_type=='P' || comparator_type=='R') {
      int iP1=-1,iP2=-1;
      switch(comparator_type) {
      case 'P':
        iP1=p1.getPreferedOrder();
        iP2=p2.getPreferedOrder();
        break;
      case 'R':
        //final StrapProtein[] pp=StrapAlign.getVisibleProteins(); iP1=idxOf(p1,pp);iP2=idxOf(p2,pp);
        iP1=p1.getRow();
        iP2=p2.getRow();
      }
      if (iP1==-1 && iP2==-1) return p1.getName().compareTo(p2.getName());
      return iP1<0 && iP2<0 ? p1.getRow()-p2.getRow() : iP1-iP2;
    }
    return 0;
  }
     /* <<< Comparator  <<< */
    /* ---------------------------------------- */
    /* >>> SequenceAligner utils  >>> */
    public static boolean hasResidueSelection(Protein p, String names[]) {
        if (p!=null) {
            for(ResidueSelection s : p.getAllResidueSelections()) {
                final String name=getNam(s);
                if (name!=null && idxOfStrg(name, names)>=0 ) return true;
            }
        }
        return false;
    }

    public static void setSeqs(SequenceAligner aligner, Protein pp[]) {
        aligner.setSequences(Protein.getResidueType(pp));
        if (isAssignblFrm(NeedsProteins.class, aligner)) ((NeedsProteins)aligner).setProteins(pp);
        if (isAssignblFrm(Superimpose3D.class, aligner)) {
            if (sze(pp)==2) {
                ((Superimpose3D)aligner).setReferenceProtein(pp[0]);
                ((Superimpose3D)aligner).setMobileProtein(pp[1]);
            } else assert false;
        }
    }

    static StrapProtein newProteinNarrowToResidueSelection(Protein p, ResidueSelection[] ss) {
        if (p==null || ss==null) return null;
        int from=MAX_INT, to=-1;
        for(ResidueSelection s : ss) {
            if (s.getProtein()==p) {
                from=mini(from, firstTrue(s.getSelectedAminoacids())+s.getSelectedAminoacidsOffset());
                to=maxi(from, lastTrue(s.getSelectedAminoacids())+s.getSelectedAminoacidsOffset());
            }
        }
        if (from<to) return null;
        final StrapProtein pNew=p.cloneProtein(new StrapProtein());
        pNew.setResidueSubset(from+"-"+to);
        return pNew;
    }

    public static StrapProtein newProteinColumnFromTo(StrapProtein p, int west, int east) {
        if (p==null) return null;
        final int i0=p.getFirstResidueIndex();
        final StrapProtein pNew=p.cloneProtein(new StrapProtein());
        pNew.setResidueSubset((i0+p.column2nextIndex(west)+1)+"-"+(i0+p.column2thisOrPreviousIndex(east-1)+1));
        return pNew;
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Childs >>> */
    static String titleForChild(Class c) {
        return
            c==null ? null :
            c==HeteroCompound.class ? HeteroPopup.TITLE :
            c==ResidueSelection.class ? "Residue selections" :
            c==ResidueAnnotation.class ? "Annotated residue selections" :
            c==SequenceFeatures.class ? "Sequence features" :
            null;
    }

    //    static Object keyForJlist(Object o) {    }
    static void showChildObjects(Class c, Protein... pp) {
        if (pp.length==1) {
            final ChJTree jT=((StrapProtein)pp[0]).getJTree();
            ChFrame.frame("Child objects "+pp[0],jT, ChFrame.PACK).shw(ChFrame.AT_CLICK|ChFrame.TO_FRONT);
            ((StrapProtein)pp[0]).expandTreeNode(c);
            jT.updateKeepStateLater(0L, 99);
        } else {
            List v=null;
            Object[] childs=null;
            for(Protein p : pp) {
                childs=null;
                if (c==HeteroCompound.class) childs=p.getDnaAndRnaAndHeteros();
                if (c==ResidueSelection.class) childs=p.getAllResidueSelections();
                if (c==ResidueAnnotation.class || c==SequenceFeatures.class) {
                    final ResidueAnnotation aa[]=((StrapProtein)p).getResidueAnnotations().clone();
                    for(int i=aa.length; --i>=0;) {
                        if ((aa[i].getFeatureName()!=null) != (c==SequenceFeatures.class) || !aa[i].isEnabled()) aa[i]=null;
                    }
                    childs=rmNull(aa);
                }
                adAll(childs,v==null ? v=new ArrayList() : v);
            }
            final String title=titleForChild(c);
            assertNotNull(title);
            if (sze(v)>0) StrapAlign.showInJList(0L, null, v.toArray(), title,null).showInFrame(ChFrame.AT_CLICK|ChFrame.SIZE_REGARD_RESOLUTION,title);
        }
    }

}