package charite.christo.strap.extensions;
import static charite.christo.ChUtils.*;
import charite.christo.*;
import charite.christo.blast.*;
import charite.christo.interfaces.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;
import java.io.*;
import static java.lang.System.*;
import static charite.christo.ChConstants.*;
/**
   @author Christoph Gille InputStream
*/
public abstract class AbstractBlaster implements SequenceBlaster, HasControlPanel,Disposable,ActionListener, ChRunnable {
    /** This can be overridden */
    public void setAAQuerySequence(String s) {query=onlyLetters(s.toUpperCase()); isAAQuery=true;}
    public void setNTQuerySequence(String s) {query=onlyLetters(s.toUpperCase()); isAAQuery=false;}
    public String getQuerySequence() { return query;}
    private int numberOfAlignments=2000;
    public int getNumberOfAlignments() { return numberOfAlignments;}
    public void setNumberOfAlignments(int n) {numberOfAlignments=n;}
    public boolean isAAQuery=true;
    public boolean isAAQuery() {return isAAQuery;}
    public String getBlastProgram(){ return isAAQuery() ? "blastp" : "blastn";}
    public String urlResultHtml;
    private String query;
    public String getQuery() { return query;}
    private String fromCache="found in cache\n";
    public ByteArray getResultXml(){
        final String q=this.query;
        if (q==null) return null;
        ByteArray ba= CacheResult.getValue(getClass(),cacheSK('S'),cacheSK('K'),(ByteArray)null);
        if (ba!=null && strstr("FATAL: ",ba)>0) {
            putln("AbstractBlaster Error: found \"FATAL: \"");
            return null;
        }
        if (ba!=null && !this.hasResultXml) {
            this.hasResultXml=true;
            log(fromCache);
        }
        return ba;
    }
    public String urlResultHtml() {return urlResultHtml;}
    /* --- Matrix --- */
    public String matrix="blosum62";
    public void setMatrix(String mx) throws IllegalArgumentException {
        for(String m:getAvailableMatrices())
            if (mx.equals(m)) {matrix=m; return; }
        throw(new IllegalArgumentException(mx+": no suXbch blast similarity matrix"));
    }
    /**FIXME*/
    public String[] getAvailableMatrices() { return new String[]{"blosum64"}; }
    public String getMatrix() {return matrix;}
    /* --- Databases --- */
    public void setDatabase(String s) throws IllegalArgumentException {
        for(String d:getAvailableDatabases())
            if (s.equals(d) || s.equals(d.substring(d.lastIndexOf(' ')+1))) {
                database=s.substring(s.lastIndexOf(' ')+1);
                return;
            }
        log(ANSI_RESET+"Error: usuported database ",s,ANSI_RESET+"\n available databases: ");
        for(String d : getAvailableDatabases()) log(d,"\n");
    }
    public String getDatabase(){return database;}
    private AsciiView _log=new AsciiView(999);
    public AsciiView logStdout(CharSequence... oo) { putln(oo); return log(oo); }
    public AsciiView log(CharSequence... oo) {
        if (_log==null) _log=new AsciiView(999);
        return _log.appendSS(oo);

    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> GUI >>> */
    protected Container controlPanel;
    public abstract Container getControlPanel();

    private ChButton butViewXml,butViewTxt;
    public ChButton butViewXml() {
        if (butViewXml==null) butViewXml=new ChButton().li(this).t("resultXml");
        butViewXml.setEnabled(this.hasResultXml);
        return butViewXml;
    }
    public ChButton butViewTxt() {
        if (butViewTxt==null) butViewTxt=new ChButton().li(this).t("result text");
        butViewTxt.setEnabled(this.hasResultXml);
        return butViewTxt;
    }
    private String database=null;
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Cache >>> */
    private final PrgParas prgPara=new PrgParas();
    public final PrgParas getPrgParas() { return prgPara;}

    private String cacheS, cacheK;
    private final static StringBuffer sb=new StringBuffer();
public String cacheSK(char type) {
        final String q=this.query;
        if (q==null) return null;
        synchronized(sb) {
            sb.setLength(0);
            if (type=='K') {
                if (cacheK==null) {
                    sb.append(q);
                    join(getPrgParas().asStringArray(PrgParas.OPTION_EQUALS_SIGN), "_", sb);
                    cacheK=sb.toString();
                }
                return cacheK;
            } else {
                if (cacheS==null) cacheS="/"+getDatabase()+"/"+CacheResult.sectionForSequence(q);
                return cacheS;
            }
        }
    }
    /* <<< Cache <<< */
    /* ---------------------------------------- */
    /* >>> Disposable >>> */
    private boolean isDisposed;
    public void dispose(){isDisposed=true;}
    public boolean isDisposed(){return isDisposed;}
    /* <<<  Disposable <<< */
    /* ---------------------------------------- */
    /* >>> Compute >>> */
    private boolean hasResultXml;
    public abstract Result computeIt();
    public final void compute() {
        final String q=this.query;
        if (sze(q)<5) {
            log(RED_ERROR," in AbstractBlaster query sequence too short",q,"\n");
            return;
        }
        if (database==null) {
            log(RED_ERROR," in AbstractBlaster: database=null");
            return;
        }
        if (CacheResult.isEnabled()) {
            final ByteArray ba=getResultXml();
            if (ba!=null) {
                final long lastModified=CacheResult.getLastModified(getClass(), cacheSK('S'), cacheSK('K'));
                log(ANSI_GREEN+"Found in cache\n"+ANSI_RESET);
                if (lastModified>0) log("Computation was performed on ",day(lastModified),"\n");
            }
        }
        if (!this.hasResultXml) {
            final Result result=computeIt();
            fromCache=null;
            if (result!=null){
                this.hasResultXml=result.resultXml!=null;
                this.urlResultHtml=result.urlResultHtml;
                CacheResult.putValue(getClass(),cacheSK('S'),cacheSK('K'),result.resultXml,true);
                ChMsg.sound(ChMsg.SOUND_SUCCESS);
                this.hasResultXml=result.resultXml!=null;
            } else ChMsg.sound(ChMsg.SOUND_FAILURE);
        }
        ChDelay.runAfterMS(EDT,this,"ENABLE",444);
    }
    public Object run(String id, Object arg) {
        if (id=="ENABLE"){
            if (butViewTxt!=null) butViewTxt.setEnabled(this.hasResultXml);
            if (butViewXml!=null) butViewXml.setEnabled(this.hasResultXml);
        }
        return null;
    }
    public static class Result {   ByteArray resultXml;   String urlResultHtml; }

    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final int modi=modifiers(ev);
        if (q==butViewTxt) {
            final File fTmp=createTempFile(".txt");
            try{
                final OutputStream fOut=fileOutStream(fTmp);
                new BlastParser().parseXML(getResultXml(), 60, fOut);
                closeStrm(fOut);
                viewFile(fTmp, modi);
            } catch(IOException e){putln("Caught in AbstractBlaster ",e);}
        }
        if (q==butViewXml) {
            final File fTmp=createTempFile(".xml");
            final ByteArray xml=getResultXml();
            if (xml!=null) {wrte(fTmp,xml); viewFile(fTmp, modi);}
        }
    }
}
