package charite.christo.strap.extensions;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import charite.christo.strap.*;
import charite.christo.*;
import charite.christo.strap.interfaces.*;
import charite.christo.interfaces.*;
import static java.lang.System.*;
/**HELP
Example
<pre class="data">
## PROTEINS : EMBL/SWISSPROT identifier and alignment statistics
NR.   ID    STRID %IDE %WSIM IFIR ILAS JFIR JLAS LALI NGAP LGAP LSEQ2 ACCNUM
1:cram_craab 1CRN 1.00  1.00   1   46    1   46   46    0    0   46  P01542
2:thn_pyrpu       0.53  0.59   2   46    2   47   45    1    1   47  P07504
3:thn_dencl       0.53  0.69   2   44    2   44   43    0    0   46  P01541
4:thn3_visal      0.49  0.66   2   46   28   72   45    0    0  111  P01538
5:thn_pholi       0.47  0.61   2   46    2   46   45    0    0   46  P01540
6:thnl_horvu      0.44  0.61   2   46   30   74   45    0    0  137  P09617
7:thnb_visal      0.44  0.61   2   46    8   52   45    0    0  103  P08943
8:thn6_horvu      0.40  0.57   2   46    2   46   45    0    0   46  P09618
Part 3) the aligned sequences. Residues 10-39 have been deleted to save space.
## ALIGNMENTS    1 -    8
 SeqNo  PDBNo AA STRUCTURE BP1 BP2  ACC NOCC  VAR  ....:....1....:....2....:
     1    1   T              0   0   77    2    0  T
     2    2   T  E     -A   34   0A  21    9   23  TSSSSSSS
     3    3   a  E     -A   33   0A   0    9    0  CCCCCCCC
     4    4   b        -     0   0    0    9    0  CCCCCCCC
     5    5   P  S    S+     0   0   52    9   28  PRPPPKPK
     6    6   S  S  > S-     0   0   48    9   35  SNTNSNND
     7    7   I  H  > S+     0   0  123    9   25  ITTTTTTT
     8    8   V  H  > S+     0   0   98    9   47  VWATTTTL
     9    9   A  H  > S+     0   0    6    9   15  AAAGAGGA
...
Part 4) the profile. The profile values for 16 residues have been
removed to save space.
## SEQUENCE PROFILE AND ENTROPY
 SeqNo PDBNo   V   L   N   D  NOCC NDEL NINS ENTROPY RELENT WEIGHT
    1    1     0   0   0   0     2    0    0   0.000      0  1.00
    2    2     0   0   0   0     9    0    0   0.530     24  1.00
    3    3     0   0   0   0     9    0    0   0.000      0  1.33
    4    4     0   0   0   0     9    0    0   0.000      0  1.33
    5    5     0   0   0   0     9    0    0   0.849     39  0.79
    6    6     0   0  44  11     9    0    0   1.215     55  0.75
    7    7     0   0   0   0     9    0    0   0.530     24  0.98
    8    8    22  11   0   0     9    0    0   1.427     65  0.47
    9    9     0   0   0   0     9    0    0   0.637     29  1.01
...
</pre>
*/
public class AlignmentWriterHSSP implements AlignmentWriter {
    private String[] names;
    private StrapProtein proteins[];
    private int firstColumn,lastColumn=9999999;
    public void setProteins(StrapProtein... pp) { proteins=pp;}
    public void setNames(String[] names) { this.names=names;}
    public void setColumnRange(int first, int last){ firstColumn=first; lastColumn=last;}
    public void setResiduesPerLine(int resPerLine) {}
    public void setCharForGapAndTerminalBlanks(char m,char n, char c) {}

    public String getFileExtension() { return "hssp";}
    public void getText(long options, StringBuffer sb){
        sb.append("HSSP       HOMOLOGY DERIVED SECONDARY STRUCTURE OF PROTEINS \n");
        final StrapProtein withSecStr=(StrapProtein)StrapAlign.getInstance().getProteinAtCursor();
        if (withSecStr==null) {
            sb.append("\nERROR: "+
                      "The HSSP-format requires a reference protein which usually has 3D structure.\n"+
                      "Please indicate the reference protein by placing the alignment cursor on the protein\n");
            return;
        }

            if (withSecStr.getResidueSecStrType()==null) {
            ChMsg.msgDialog(
                                "The reference protein is the one where the alignment cursor is.<br>"+
                                "In HSSP this should usually be a protein with 3D structure given as a PDB file.");
            }
        int firstAA=withSecStr.column2thisOrPreviousIndex(firstColumn);
        final int lastAA=withSecStr.column2thisOrPreviousIndex(lastColumn);
        //append(sb,"# structure file=",withSecStr.getName()," fromIdx=").append(firstAA).append(" toIndex=").append(lastAA);
        sb.append("# the structure file is the one where the cursor is\n");
        sb.append("# structure file=").append(withSecStr.getName()).append(" fromIdx=").append(firstAA).append(" toIndex=").append(lastAA);
        if (proteins==null) return;
        if (idxOf(withSecStr,proteins)<0) {
            final int n=proteins.length;
            final StrapProtein pp[]=new StrapProtein[n+1];
            pp[n]=withSecStr;
            arraycopy(proteins,0,pp,0,n);
            proteins=pp;
            names=chSze(names,n+1);
            names[n]=withSecStr.getName();
        }
        sb.append("\n# NUMBER OF PROTEINS: ").append(proteins.length).append("\n");
        sb.append("# PROTEINS: ");
        for(String s:names) sb.append(s).append(' ');
        sb.append("\n");
        if (firstAA<0) firstAA=0;
        transpose();
        int iBlock=0;
        while(exportHSSP(sb,iBlock++,withSecStr,firstAA,lastAA));
        sb.append("## \n");
    }
    // ----------------------------------------
    private boolean exportHSSP(StringBuffer sb,int iBlock,StrapProtein withSecStr, int firstAA,int lastAA) {
        final int nProt=ali.length;
        if (nProt<1) return false;
        int i,j;
        boolean anOtherBlock=true;
        int colMin=iBlock*70,colMax=(iBlock+1)*70;
        if (nProt<=colMax) {colMax=nProt; anOtherBlock=false;}
        if (iBlock==0) {
            sb.append("SEQLENGTH   ").append(lastAA-firstAA)
                .append("\nNALIGN       ").append(nProt)
                .append("\n  NR.    ID         STRID   %IDE %WSIM IFIR ILAS JFIR JLAS LALI NGAP LGAP LSEQ2 ACCNUM     PROTEIN\n");
        }
        sb.append("## ALIGNMENTS    ").append(colMin+1).append("  -  ").append(colMax)
            .append("\n SeqNo  PDBNo AA STRUCTURE BP1 BP2  ACC NOCC  VAR  ");
        // ....:....1....:....2....:....3....:....4....:....5....
        for(i=0;i<nProt;i++) {
            char c='.';
            if (i%10==4) c=':';
            if (i%10==9) c= (char)(((i+1)/10)%10+'0');
            sb.append(c);
        }
        sb.append("\n");
        int pdbNo[]=withSecStr.getResidueNumber();
        if (pdbNo==null) {
            final int n=withSecStr.countResidues();
            pdbNo=new int[n];
            for(int c=0; c<n; c++) pdbNo[c]=c+1;
        }
        final byte resType[]=withSecStr.getResidueType();
        final byte secStr[]=withSecStr.getResidueSecStrType();
        final float accessibility[]=withSecStr.getResidueSolventAccessibility();
        for(j=firstAA;j<lastAA && j<pdbNo.length;j++) {
            int num=(j+1-firstAA);
            format10(num,6,false,sb);
            num=pdbNo.length>j && pdbNo[j]!=MIN_INT ? (pdbNo[j]) : (0);
            format10(num,5,false,sb);
            final byte ss=secStr!=null && secStr.length>j ? secStr[j] : 0;
            sb.append("   ")
                .append((char)resType[j])
                .append("  ").append(ss!=0 ? (char)ss : ' ').append("           0   0");
            final float fNum=accessibility!=null && accessibility.length>j ? accessibility[j] : 0;
            ftoa(fNum,5,0,sb);
            format10(1,10,false,sb).append("  ");
            appnd(aliTranspon[j],colMin,colMax,sb).append("\n");
        }
        return anOtherBlock;
    }
    private byte ali[][],aliTranspon[][];
    private void transpose() {
        final StrapProtein pp[]=proteins;
        final int n=pp.length;
        final int hMax=StrapProtein.getMaxColumn(pp);
        ali=new byte[n][hMax];aliTranspon=new byte[hMax][n];
        //initialize
        for(int j=0;j<hMax;j++)
            for(int i=0;i<n;i++)
                ali[i][j]=aliTranspon[j][i]=32;
        for(int i=0;i<n;i++) {
            final StrapProtein p=pp[i];
            final byte aa[]=p.getResidueType();
            final int resPos[]=p.getResidueColumn();
            final int nA=p.countResidues();
            for(int j=0;j<nA;j++) {
                final byte b=(byte)(aa[j]&~32);
                if (b<'A' || b>'Z') continue;
                final int pos=resPos[j];
                if (pos>=hMax) continue;
                ali[i][pos]=aliTranspon[pos][i]=b;
            }
        }
    }
}
