package charite.christo.strap.extensions;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import charite.christo.strap.StrapProteinUtils;
import charite.christo.strap.interfaces.*;
import java.util.*;
import java.awt.event.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP
   This is a combination  of ClustalW and TM_align.
   If there are two or more 3D structures, they are superimposed with TM_align (PUBMED:15849316).
   Subsequently, clustalW is called with all sequences taking the 3D alignment as
   a profile.
*/
public class ClustalW_3D implements CanBeStopped,NeedsProteins,SequenceAlignerSorting,HasControlPanel, ActionListener {
    public void setOptions(long flags){}
    public long getOptions(){ return 0L;}

    public long getPropertyFlags(){ return 0L;}
    public java.awt.Container getControlPanel() { return vB(getCtrlPanel(clustalW),getCtrlPanel(tm_align));}

    public float getScore() {
        return clustalW!=null ? clustalW.getScore() :
            tm_align instanceof Superimpose_TM_align ? ((Superimpose_TM_align)tm_align).getScore() :Float.NaN;
    }
    {vALIGNMENTS.add(weakRef(this));}
    private boolean stopped;
    public void stop() {
        stopped=true;
        if (clustalW!=null) clustalW.stop();
        else if (tm_align!=null) ((CanBeStopped)tm_align).stop();
    }
    public void compute() {
        final ArrayList<Protein> v3D=new ArrayList();
        for(Protein p : pp)  if (p.hasCalpha())  adUniq(p,v3D);
        final Protein pp3D[]=toArry(v3D, Protein.class);
        final byte[][] profile;
        if (pp3D.length>=2) {
            if (stopped) { putln("ClustalW_3D stopped"); return; }
            tm_align=  pp3D.length==2 ? new Superimpose_TM_align() : new MultiFromPairAligner(Superimpose_TM_align.class);
            addActLi(this,tm_align);
            StrapProteinUtils.setSeqs(tm_align, pp3D);
            if (stopped) return;
            tm_align.compute();
            profile=tm_align.getAlignedSequences();
            if (pp3D.length==pp.length) { aligned=profile;  return;}
        } else profile=null;
        if (stopped) return;
        clustalW=new MultipleAlignerClustalW();
        StrapProteinUtils.setSeqs(clustalW, pp);
        clustalW.addProfile(profile);
        clustalW.compute();
        this.aligned=clustalW.getAlignedSequences();
    }
    public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        if (cmd.startsWith(KEY_PROGRESS)) handleEvt(this,ev);
    }
    private byte[][] sequences,aligned;
    private MultipleAlignerClustalW clustalW;
    private SequenceAligner tm_align;
    private Protein pp[];
    public void setProteins(Protein... proteins) { pp=proteins;}

    public int[] getIndicesOfSequences() {
        return clustalW!=null ? clustalW.getIndicesOfSequences() :
            tm_align instanceof SequenceAlignerSorting ? ((SequenceAlignerSorting)tm_align).getIndicesOfSequences() : count01234(pp.length) ;
    }

    public byte[][] getSequences() { return sequences;}
    public byte[][] getAlignedSequences() { return aligned;}
    public void setSequences(byte[]...ss) { sequences=ss;}
}
