package charite.christo.strap.extensions;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import charite.christo.strap.*;
import charite.christo.strap.interfaces.*;
import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.*;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP
A white space separated list of hetero-groups is needed.

Hetero-groups can be specified:
<ol>
<li>The name E.G. FAD, FMN, HEM. </li>
<li>The number as found in the pdb file. </li>
<li>In case of DNA/RNA  a capital letter which designates the chain letter </li>
</ol>

<br><b>Limitations:</b> Currently only c-alpha atoms.
<i>SEE_CLASS:Distance3DToCursor</i>
*/
public class Distance3DToHetero extends AbstractResidueSelection implements StrapListener,IsEnabled,HasSharedControlPanel,ActionListener,NeedsSharedInstance {
    public final static Object KEY_DEFAULT_HETERO=new Object();
    /* scope for following fields is shared instance */
    private ChLabel labMax;
    private JSlider maxSlider_;
    private Container ctrlPanel;
    private ChTextField tfHeteroCode;
    private int modi;
    private final static ByteArray baTip=new ByteArray(0);;
    /* private fields */
    private boolean selected[];
    private ChButton togSideChain;
    private String toStrg=null;
    @Override public String toString() {
        final String s=sharedInstance!=null ? sharedInstance.toStrg : toStrg;
        return s!=null ? s : super.toString();
    }
    public boolean isEnabled(Object o){  return o instanceof Protein  && ((Protein)o).getResidueCalpha()!=null;}

    public Container getSharedControlPanel() {
        if (sharedInstance!=null) return getSharedCtrlPan(sharedInstance, false);
        if (ctrlPanel==null) {
            labMax=new ChLabel();
            pcp(KEY_IF_EMPTY,"Enter list of hetero group names or hetero residue numbers here",tfHeteroCode=new ChTextField(defaultHeteroCode).saveInFile("Distance3DToHetero_hetero").li(this));
            addActLi(this,maxSlider_=new ChSlider("",0,MAX_VALUE,maxValue));
            maxSlider_.setMajorTickSpacing(100);
            maxSlider_.setMajorTickSpacing(200);
            maxSlider_.setPaintLabels(false);
            maxSlider_.setPaintTicks(true);
            label(MAX_VALUE/2);
            final JComponent
                pan=vB(
                       hB("Enter a space separated list of heterogroups.<br>Either enter residue name like \"FMN\" or the residue number. <br>"+
                         "Optionally append colon and chain-letter such as FMN:A or 250:A<br>"+
                          "Specify DNA/RNA by a single capital letter denoting the chain or type \"DNA\" or \"RNA\".<br>"),

                       hB((togSideChain=toggl("Consider not only C-Alpha atoms but also side chain atoms").li(this)).cb()),
                       tfHeteroCode,
                       maxSlider_,

                       hB("Highlight residues closer than ",monospc(labMax),"to the selected heterogroup","#"),
                       hB("<br><sub>Note: Usually, heteros and DNA/RNA are compared only to the peptides, they are assigned to.<br>"+
                       "But if the protein they belong to is given in the form \"protein|hetero\", distance is measured  to all proteins.<br>"+
                       "For example peptide chain F of PDB:1gd2 has a RNA with chain letter \"A\".  <br>"+
                          "The resulting expression will be \"1gd2_F.pdb|A\" </sub>")
                       );

            ctrlPanel=pan;
        }
        return ctrlPanel;
    }
    // ----------------------------------------
    private final static int MAX_VALUE=3200;
    private static int maxValue=MAX_VALUE;
    public static void setDefaultHeteroCode(String s) { defaultHeteroCode=s;}
    private static String defaultHeteroCode="";
    private CharSequence label(int l) { CharSequence s=ftoa(l/100f,5,3,null).append("\u00c5 "); if (labMax!=null) labMax.t(s); return s;}
    public final void handleEvent(StrapEvent ev) {
        if ((ev.getType()&StrapEvent.FLAG_ATOM_COORDINATES_CHANGED)!=0) StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
    }
    @Override public void setSelectedAminoacids(boolean bb[],int offset) {}

    private Distance3DToHetero sharedInstance;
    public void setSharedInstance(Object shared) {  this.sharedInstance=(Distance3DToHetero)shared; }
    public Object getSharedInstance() { return sharedInstance;}
    private boolean alreadyTriedSideChain;
    private static ChTokenizer _tok;
    @Override public boolean[] getSelectedAminoacids() {
        final ChTokenizer TOK= _tok==null ? _tok=new ChTokenizer() : _tok;
        final Distance3DToHetero si=sharedInstance!=null ? sharedInstance : this;
        final ChTextField tf=si.tfHeteroCode;
        final ChButton togSS=si.togSideChain;
        final Protein p=getProtein();
        final float[] xyzCA=p!=null ? p.getResidueCalpha() : null;
        if (tf==null || xyzCA==null) return null;
        final ByteArray ba=tf.tools().byteArray();
        final byte T[]=ba.getBytes();
        final float max=si.maxValue*0.01f;
        float[] xyzSS=null;
        if (togSS!=null && togSS.s()) {
            xyzSS=p.getAtomCoordinates();
            if (xyzSS==null &&  !alreadyTriedSideChain) {
                alreadyTriedSideChain=true;
                p.loadSideChainAtoms(null);
                xyzSS=p.getAtomCoordinates();
            }
            if (xyzSS==null) StrapAlign.errorMsg(" No side chain for protein "+p);
        }
        final int mc=p.getModificationCount(ProteinModiCount.MC_ATOM_COORD)+si.modi;
        if (this.selected==null || this.modi!=mc) {
            this.modi=mc;
            Protein.incrementModificationCount(ProteinModiCount.MC_RESIDUE_SELECTIONS, p);
            final int nR=mini(xyzCA.length/3, p.countResidues());
            boolean[] selected=null;
            TOK.setText(T,0,ba.getEnd());
            while(TOK.nextToken()) {
                final int from0=TOK.getFrom(), to=TOK.getTo();
                final int pipe=strchr('|',T,from0,to);
                final int from=pipe>0?pipe+1:from0;
                final Protein pHet=pipe<0 ? p :StrapProtein.getProteinWithName(new String(T,from0,pipe-from0),null);
                if (pHet==null) continue;
                final int colon=strchr(':', T, from,to);
                final int end=colon>0?colon:to;
                final int L=end-from;

                final int name32=
                    L==3 ? charsTo32bit( T[from], T[from+1], T[from+2], ' ') :
                    L==2 ? charsTo32bit( ' ',  T[from], T[from+1], ' ') :
                    L==1 ? charsTo32bit( ' ',  ' ', T[from], ' ') :
                    0;

                char chain=colon+2==to ? (char)T[colon+1] : colon+1==to ? ' ' : 0;
                int no=cntainsOnly(DIGT, T, from, end) ? TOK.asInt() : MIN_INT;
                if (to-from==1 && T[from]=='!') {
                    final String s=(String)gcp(KEY_DEFAULT_HETERO,p);
                    if (s!=null) {
                        final int col=s.indexOf(':');
                        if (col>0 && col+1<s.length()) chain=s.charAt(col+1);
                        no=atoi(s);
                    }
                }
                final Matrix3D m3d=pHet.getRotationAndTranslation();
                final float max2=max*max;
                for(HeteroCompound h : pHet.getDnaAndRnaAndHeteros()) {
                    if (!h.isNucleotideChain()) {
                        if(no!=MIN_INT && no!=h.getCompoundNumber()) continue;
                        if (no==MIN_INT && name32!=h.getCompoundName32()) continue;
                        if (chain!=0 && h.getCompoundChain()!=chain) continue;
                    } else {
                        if ((L!=1 || T[from]!=h.getCompoundChain()) && !equls(h.isDNA() ? "DNA" : "RNA",T,from)  ) continue;
                    }
                    nextAmino:
                    for(int iA=0;iA<nR;iA++) {
                        final boolean ss=xyzSS!=null;
                        if (selected==null) selected=new boolean[nR];
                        selected[iA]=selected[iA]||h.squareDistanceToNucleotidechain(ss?xyzSS:xyzCA, ss?pHet.getResidueFirstAtomIdx(iA):iA, ss?pHet.getResidueLastAtomIdx(iA):iA+1,  m3d, 0)<max2;
                    }
                }
            }
            TabItemTipIcon.set(null,null,baTip.newString(),null,this);
            this.selected=selected;
        }
        return this.selected;
    }

    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        if (q==maxSlider_) {
            maxValue=maxSlider_.getValue();
            setName(addHtmlTagsAsStrg(" distance of C-alpha Atom to hetero group distance < "+label(maxValue)));
            TabItemTipIcon.set(null,null,"distance <"+maxValue,null,this);
        }

        if (q==maxSlider_ ||    q==tfHeteroCode && (cmd==ACTION_FOCUS_LOST || cmd==ACTION_ENTER ) ||q==togSideChain) {
            toStrg="Distance3DToHetero "+tfHeteroCode;
            modi++;
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
        }
    }
}
