package charite.christo.strap.extensions;
import static charite.christo.ChUtils.*;
import charite.christo.protein.*;
import charite.christo.*;
import charite.christo.strap.*;
import charite.christo.interfaces.*;
import java.io.File;
/**HELP
Home-page <i>HOME_PAGE:MultipleAlignerDialign</i>
<br>
Author: Burkhard Morgenstern and Said Abdeddaim
<br>
Publication: PUBMED:10222408
<br>
*/
public final class MultipleAlignerDialign extends AbstractSuperposition implements SequenceAligner {
        private final static String PARAMERS[][]={
            {"-afc","[x]Creates additional output file <i>*.afc</i> containing data of all fragments considered for alignment. WARNING: this file can be HUGE !"},
            {"-afc_v","[x]like <i>\"-afc\"</i> but verbose: fragments are explicitly printed. WARNING: this file can be EVEN BIGGER !"},
            {"-ff","[x]Creates file <i>*.frg</i> containing information about all fragments that are part of the respective optimal pairwise alignments plus information about consistency in the multiple alignment"},
            {"-fop","[x]Creates file <i>*.fop</i> containing coordinates of all fragments"},
            {"-fsm","[x]Creates file <i>*.fsm</i> containing coordinates of all fragments that are part of the final alignment"},
            {"-iw","[x]overlap weights switched off.<br> Higher quality, less speed"},
            {"-lo","[x](Long Output) Additional file <i>*.log</i> with more information"},
            {"-mat","[x]Creates file *mat with substitution counts."},
            {"-mat_thr","[x]Like <i>-mat</i> but only fragments with weight score"},
            {"-max_link","[x]<i>maximum linkage</i> clustering used to construct sequence tree instead of UPGMA"},
            {"-min_link","[x]minimum linkage"},
            {"-mot","[x]motif option"},
            {"-msf","[x]separate output file in MSF format"},
            {"-o","[x]fast version, resulting alignments may be slightly different"},
            {"-ow","[x]overlap weights enforced. <br>Less quality, high speed"},
            {"-thr","Threshold T = x"}
        };

    public MultipleAlignerDialign() {
        getPrgParas().defineGUI(PARAMERS, PrgParas.OPTION_FURTHER_PARAMETERS);
    }

    @Override public Superimpose3D.Result computeResult(byte[][] seqs, Protein[] proteinsNotNeeded) {
        final String n=
            "dialign2-2",
            instScript=
            "cd dialign_package/src\n"+
            "make\n"+
            "cp dialign2-2"+winDotExe()+" ../../"+n+".exe\n",
            srcUrls[]={"http://dialign.gobics.de/download/dialign_package.tgz"};
        final AbstractExecutable e=new AbstractExecutable(n);
        setExecutable(e, srcUrls, instScript, (String[])null);
        if (!Insecure.EXEC_ALLOWED) return null;
        final File dirTmp=dirTemp(), fileOut=file(dirTmp,"output.fa"), mfaInfile=mfaInFile(seqs), fileExe=e.fileExecutable();
        e.exec()
            .setCommandLineV(fileExe, getPrgParas().asStringArray(PrgParas.OPTION_EQUALS_SIGN), "-fa","-cw","-fn","output",mfaInfile)
            .dir(dirTmp).setOptions(ChExec.STDOUT | ChExec.STDERR | ChExec.CYGWIN_DLL | ChExec.LOG)
            .addToEnvironement(file(e.dirBinaries()+"/dialign_package/dialign2_dir/dialign2_dir/"))
            .run();
        return parse(readBytes(file(dirTmp,"output.fa")), null, Superimpose3D.PARSE_SCORE|Superimpose3D.PARSE_MSA_START_S);
    }
}
