package charite.christo.strap.extensions;
import charite.christo.protein.*;
import charite.christo.*;
import charite.christo.strap.*;
import charite.christo.interfaces.*;
import charite.christo.strap.interfaces.*;
import java.io.File;
import static charite.christo.ChUtils.*;
/**HELP

Home: <i>HOME_PAGE:MultipleAlignerDialignT</i>
<br>
Author: Amarendran R. Subramanian.
<br>
Publication PUBMED:15784139
@author Christoph Gille
*/
//  gcc -lm *.c
public final class MultipleAlignerDialignT extends AbstractSuperposition implements SequenceAligner {
         private final static String PARAMERS[][]={
            {"-d","Debug-Mode  [DEFAULT 0] 0 no , 1 current phase of the processing, 2 very loquacious, 5 hardcore debugging"},
            {"-s","max amount of input sequences [DEFAULT 5000]"},
            {"-a","max number of characters per line in a FASTA file [DEFAULT 100]"},
            {"-c","max amount of characters per line when printing a sequence [DEFAULT 80]"},
            {"-m","score matrix file name (in the config directory) [DEFAULT PROTEIN: BLOSUM.scr,  DNA: dna_matrix.scr]"},
            {"-w","    defines the min weight when the weight formula is changed to 1-pow(1-prob, factor) [DEFAULT 0.000000065]"},
            {"-p","probability distribution file name (in the config directory) [DEFAULT PROTEIN: BLOSUM.diag_prob_t10,  DNA: dna_diag_prob_100_exp_550000]"},
            {"-v","add to each score (to prevent negative values) [DEFAULT 0]"},
            {"-t","\"even\" threshold for low score for sequences alignment [DEFAULT PROTEIN: 4, DNA: 0]"},
            {"-n","max number of consecutive positions for window containing low scoring positions [DEFAULT PROTEIN: 4, DNA: 1]"},
            {"-g","global min fragment length for stop criterion [DEFAULT PROTEIN: 40, DNA: 1]"},
            {"-m","minimal allowed average score in frag window containing low scoring positions [DEFAULT PROTEIN: 4.0,  DNA: 0.9]"},
            {"-o","wether overlap weights are calculated or not [DEFAULT 0]"},
            {"-f","min fragment length [DEFAULT 1]"},
            {"-r","threshold weight to consider the fragment at all [DEFAULT 0.0]"},
            {"-u","[DEFAULT 0] 1: only use a sqrt(amount_of_seqs) stripe of neighbor<br> sequences to calculate pairwise alignments (increase performance)<br>0: all pairwise alignments will be calculated"},
            {"-D","input is DNA-sequence"},
            {"-T","translate DNA into aminoacids from begin to end (length will be cut to mod 3 = 0) WARNING: Do not use -D with this option"},
            {"-L","compare only longest ORF WARNING: Do not use -D with this option (Default values for PROTEIN input will be loaded)"},
            {"-O","translate DNA to aminoacids, reading frame for each sequence calculated due to its longest ORF WARNING: Do not use -D with this option "},
            {"-P","output in aminoacids, no retranslation of DNA sequences [DEFAULT: input = output]"},
            {"-C","generate probability table saved in <config_dir>/prob_table and exit"}
        };
    private static String conf;
    public MultipleAlignerDialignT() {
        getPrgParas().defineGUI(PARAMERS, PrgParas.OPTION_FURTHER_PARAMETERS);
    }
    @Override public Superimpose3D.Result computeResult(byte[][] seqs, Protein[] proteinsNotNeeded) {
        final String
            n="DIALIGN-TX_1.0.2",
            instScript="cd "+n+"/source\n"+
            "make\n"+
            "cp dialign-tx"+winDotExe()+" "+n+".exe\n",
            srcUrls[]={"http://dialign-tx.gobics.de/"+n+".tar.gz"}; /* WATCH_URL */
        final AbstractExecutable e=new AbstractExecutable(n);
        setExecutable(e, srcUrls, instScript, (String[])null);
        if (!Insecure.EXEC_ALLOWED) return null;
        final File dirTmp=dirTemp(), fileOut=file(dirTmp,"output.msf"), mfaInfile=mfaInFile(seqs), fileExe=e.fileExecutable(), dirBin=e.dirBinaries();

        if (conf==null) {
            final String ss[]=dirBin.list();
            if (ss!=null) {
                for(String s : ss) {
                    if (s.startsWith("DIALIGN") && !s.endsWith(".exe")) {
                        conf=(dirBin+"/"+s+"/conf").replace('\\','/');
                        break;
                    }
                }
            }
        }
        e.exec()
            .setCommandLineV(fileExe, getPrgParas().asStringArray(PrgParas.OPTION_EQUALS_SIGN),conf,mfaInfile,fileOut)
            .dir(dirTmp)
            .setOptions(ChExec.STDOUT | ChExec.STDERR | ChExec.CYGWIN_DLL | ChExec.LOG)
            .run();
        return parse(readBytes(fileOut), null, Superimpose3D.PARSE_SCORE|Superimpose3D.PARSE_MSA_START_S);

    }
}
