package charite.christo.strap.extensions;
import charite.christo.protein.*;
import charite.christo.strap.*;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.io.File;
import static charite.christo.ChUtils.*;
/**HELP

SPEM aligns multiple sequences using pre-processed sequence profiles
and predicted secondary structures for pairwise alignment,
consistency-based scoring for refinement of the pairwise alignment and
a progressive algorithm for final multiple alignment.

<br>
<b>Installation:</b> <i>HOME_PAGE:http://sparks.informatics.iupui.edu/</i>.
<br>
<b>Publication:</b> PUBMED:16020471
<br>
<b>Authors:</b>Hongyi Zhou  and Yaoqi Zhou
*/
public final class MultipleAligner_SPEM extends AbstractSuperposition implements SequenceAligner {
    @Override public Superimpose3D.Result computeResult(byte[][] seqs, Protein[] proteinsNotNeeded) {

        final String
            n="spem",
            instScript=null;
        final AbstractExecutable e=new AbstractExecutable(n);
        setExecutable(e, (String[])null, instScript,(String[])null);
        if (!Insecure.EXEC_ALLOWED) return null;
        final File mfaInfile=mfaInFile(seqs), fileOut=file(withoutDotSuffix(mfaInfile)+".aln");
        e.exec()
            .setCommandLineV("scan_spem_alone.job",mfaInfile)
            .dir(dirTemp())
            .setOptions(ChExec.STDOUT | ChExec.STDERR | ChExec.LOG | ChExec.CYGWIN_DLL)
            .run();

        if (sze(fileOut)==0) {
            final String error="Error could not run SPEM!\nPlease install the SPEM package manually.";
            putln(ChConstants.ANSI_RED+"Error: "+ChConstants.ANSI_RESET+error);
            error(error);
        }
        return parse(readBytes(fileOut), null, Superimpose3D.PARSE_SCORE|Superimpose3D.PARSE_MSA_START_S);

    }
}
