package charite.christo.strap.extensions;
import charite.christo.*;
import charite.christo.strap.*;
import charite.christo.interfaces.*;
import charite.christo.strap.interfaces.*;
import charite.christo.protein.*;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import static java.lang.System.*;
import static charite.christo.ChUtils.*;
/**HELP
The purpose of this class is to assign numeric values to residues in
proteins.  STRAP expects a plain text file containing the data.  These
data file can be typed by hand using a text program or may be
generated by other programs or scripts.  When the contents of the data
file[s] is changed the plot is updated automatically.  The numbers in
the file must be separated by space (space, line break, tab) or
colon. The decimal point is a dot and not a colon.  Undefined values
should be coded as NaN ("Not A Number").  The file name is specified in the
control panel.
*/
public final class ResidueValues_from_textfile  extends Thread implements  ValueOfResidue, HasControlPanel,StrapListener,Disposable,ActionListener, HasModiCount {
    { setSettings(this, Customize.fileBrowsers, Customize.textEditors); }
    private boolean isRunning=true;
    private File file_;
    private Protein file_lastProt;
    private int file_lastDataset;
    private File f() {
        final Protein p=getProtein();
        final int d=comboDataset.getSelectedIndex();
        if (p==null) return null;
        if (file_==null || file_lastDataset!=d || file_lastProt!=p) {
            file_=file("residueData/"+p+"."+d);
            file_lastProt=p;
            file_lastDataset=d;
            mkParentDrs(file_);
        }
        return file_;
    }
    // ----------------------------------------
    private void updateTextFields() {
        File f=f();
        tfFile.t(f + (f!=null && f.exists() ? " (exists) ":" (not existing) "));
    }
    private ChTextField tfFile=new ChTextField();
    private ChCombo comboDataset=new ChCombo("0","1","2","3","4","5","6","7","8","9","10","11","12").li(this);
    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        if (q==comboDataset){
            values=null;
            new StrapEvent(ResidueValues_from_textfile.this,StrapEvent.VALUE_OF_RESIDUE_CHANGED).run();
            updateTextFields();
        }
        if (cmd=="EDIT" && f()!=null) edFile(-1, f(), modifiers(ev));
    }

    private Protein protein;
    public boolean setProtein(Protein p) {  if (protein!=p) values=null;  protein=p; updateTextFields(); return true; }
    public Protein getProtein() { return protein;}
    private double[] values;
    private ProteinAlignment align;
    public void handleEvent(StrapEvent ev){
        if (align==null) {
            (align=ev.getProteinAlignment()).removeListener(this);updateTextFields();
            start();
        }
    }
    private Container controlPanel;
    public Container getControlPanel() {
        if (protein==null) return null;
        if (controlPanel==null) {
            controlPanel=vB
                (
                 pnl("dataset: ",comboDataset),
                 pnl("data file: ",tfFile, new ChButton("EDIT",ChButton.NO_MARGIN).t("edit").li(this)),
                 f()!=null ? file(f().getParent()) : null
                 );
        }
        return controlPanel;
    }

    public int getModificationCount() { return modiCount;}
    private int modiCount;

    public double[] getValues() {
        final Protein p=getProtein();
        if (p==null) return null;
        final File fData=f();
        if (fData==null || !fData.exists()) {
            error("file "+fData+" does not exist.\n<br>This file should contain a float value for each residue.\n");
            return null;
        }
        final long lastModi=fData.lastModified();
        if (values==null || lastModified!=lastModi) {
            modiCount++;
            lastModified=lastModi;
            values=readValues(fData,p.countResidues());
        }
        return values;
    }
    private double[] readValues(File fData,int nResidues) {
        double vv[]=new double[nResidues];
        java.util.Arrays.fill(vv,Double.NaN);
        try {
            InputStream is=new FileInputStream(fData);
            StreamTokenizer tokenizer=new StreamTokenizer(is);
            for( int type,iAa=0; (type=tokenizer.nextToken())!=StreamTokenizer.TT_EOF && iAa<nResidues;) {
                if (type==StreamTokenizer.TT_NUMBER) {
                    vv[iAa++]=tokenizer.nval;
                }
            }
            is.close();
        } catch(Exception e){
            return null;
        }
        return vv;
    }

    private long lastModified;
    public void myUpdate() {
        while(isRunning) {
            ChUtils.sleep(333);
            final File f=f();
            if (f==null) continue;
            final long lastModi=f.lastModified();
            if (lastModified!=lastModi) {
                lastModified=lastModi;
                values=null;
                new StrapEvent(ResidueValues_from_textfile.this,StrapEvent.VALUE_OF_RESIDUE_CHANGED).run();
            }
        }
    }
    public void dispose() {
        isRunning=false;
    }
}
