package charite.christo.strap.extensions;
import java.io.*;
import java.awt.*;
import java.util.*;
import java.util.List;
import charite.christo.*;
import charite.christo.protein.*;
import charite.christo.strap.*;
import charite.christo.strap.interfaces.*;
import charite.christo.interfaces.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import org.biojava.bio.structure.AtomImpl;
import org.biojava.bio.structure.Atom;
import org.biojava.bio.structure.AminoAcidImpl;
import org.biojava.bio.structure.ChainImpl;
import org.biojava.bio.structure.StructureImpl;
import org.biojava.bio.structure.StructureTools;

import org.biojava.bio.structure.StructureException;
import org.biojava.bio.structure.io.PDBParseException;

import org.biojava.bio.structure.align.StructureAlignment;
import org.biojava.bio.structure.align.ce.CeMain;
import org.biojava.bio.structure.align.ce.CeParameters;
import org.biojava.bio.structure.align.model.AFPChain;
import org.biojava.bio.structure.jama.Matrix;
/**HELP
<br><b>Publication:</b> PUBMED:11125099 , PUBMED:9796821

This is the Java based implementation of CE/CL superposition.
Ported from C++ to Java by Andreas Prlic.

@author Christoph Gille
java test.biojava3.TestStructure

http://biojava.org/wiki/CVS_to_SVN_Migration
svn co svn://code.open-bio.org/biojava/biojava-live/trunk biojava-live

export BIOJAVA_LIVE=/local/java/bioinf/biojava-live
export JAR=~/.StrapAlign/jars/Structure_biojava.jar
rm $JAR $JAR.pack.gz
rm -r tmpClasses
mkdir tmpClasses
CP=$(echo $BIOJAVA_LIVE/[a-z]*[a-z]/src/main/java| tr ' ' ':')
javac -cp $CP -O -source 1.5 -target 1.5  $BIOJAVA_LIVE/structure/src/main/java/demo/DemoCE.java -d tmpClasses

pushd tmpClasses
jar -cf $JAR org
popd

pack200 $JAR.pack.gz $JAR

scp  $JAR.pack.gz $STRAP_DIR_B/otherPackages
scp  $JAR.pack.gz $STRAP_DIR_B/lite

#java demo.DemoCE

zip -0 -r
*/
public class Superimpose_CE extends AbstractSuperposition implements HasScore, SequenceAligner, Superimpose3D  {
    /* >>> compute >>> */
    @Override public Result computeResult(byte[][] ssNotNeeded, Protein[] proteins) {
        final Protein pR=proteins[0], pM=proteins[1];
        //putln(ANSI_GREEN+"computeResult"+ANSI_RESET+" "+pR+" "+pM);

        final CeParameters params=new CeParameters();
        final StructureImpl strR=toBiojavaStructureCA(pR,(boolean[])null,true), strM=toBiojavaStructureCA(pM,(boolean[])null,true);
        if (strR==null||strM==null) return null;
        final Atom[] caR=StructureTools.getAtomCAArray(strR);
        final Atom[] caM=StructureTools.getAtomCAArray(strM);

        final	StructureAlignment algorithm =new CeMain();

        AFPChain afpChain=null;
        try {
            afpChain=algorithm.align(caR,caM,params);
        } catch(StructureException ex) { stckTrc(ex); }

        afpChain.setName1(pR.getName());
        afpChain.setName2(pM.getName());
        /*
        putln(afpChain.toFatcat(caR,caM));
        putln(afpChain.toRotMat());
        putln("========================================================== ");
        */
        final Matrix[] mm=afpChain.getBlockRotationMatrix();
        final Atom[] shifts=afpChain.getBlockShiftVector();
        final Superimpose3D.Result result=new Superimpose3D.Result();
        final double r[][]=mm[0].getArray(), translation[]=shifts[0].getCoords();
        final Matrix3D m3d=new Matrix3D(
                                        r[0][0],r[1][0],r[2][0],  translation[0],
                                        r[0][1],r[1][1],r[2][1],  translation[1],
                                        r[0][2],r[1][2],r[2][2],  translation[2]);
        result.setMatrix(m3d);
        result.setRmsd(afpChain.getChainRmsd());
        result.setScore(afpChain.getProbability());
        //putln("Alnseq1=",chars2bytes(afpChain.getAlnseq1()));
        //putln("Alnseq2=",chars2bytes(afpChain.getAlnseq2()));

        final byte[] ggR=chars2bytes(afpChain.getAlnseq1()), ggM=chars2bytes(afpChain.getAlnseq2());
        final byte[][] gapped= insertWhereNoCoordinates(afpChain.getAlnbeg1(),ggR, afpChain.getAlnbeg2(),ggM, pR, pM);
        result.setGappedSequences(gapped);
        //putln("result=\n"+result);
        return result;
    }

    public static StructureImpl toBiojavaStructureCA(Protein p, boolean selectedAminos[], boolean original) {
        final float[] xyz=p==null ? null : p.getResidueCalpha(original ? null : p.getRotationAndTranslation()), buf=new float[3];
        if (xyz==null) return null;
        final int nRes=p.countResidues(), iMax=mini(xyz.length/3, nRes);
        final byte[] resType=p.getResidueType();
        if (xyz==null) return null;

        StructureImpl structure=null;
        final int modi=p.getModificationCount(ProteinModiCount.MC_CALPHA);
        final String 
            SYNC="SCE$$KS_", 
            key=(SYNC+boolToTxt(selectedAminos)+"_"+modi).intern();

        synchronized(SYNC) {
            if (original) structure=gcp(key,p, StructureImpl.class);
        }

        if (structure==null) {
            final ChainImpl chain=new ChainImpl();
            for(int i=0;i<iMax; i++) {
                if (selectedAminos!=null && !(selectedAminos.length>i && selectedAminos[i])) continue;
                if (Float.isNaN(xyz[3*i]) || Float.isNaN(xyz[3*i+1]) || Float.isNaN(xyz[3*i+1])) continue;
                final AminoAcidImpl aminoAcid=new AminoAcidImpl();
                try {
                    aminoAcid.setPDBName(Protein.toThreeLetterCode(resType[i]));
                } catch(PDBParseException ex) {stckTrc(ex);}
                final Atom atomCA=new AtomImpl();
                atomCA.setCoords(new double[]{xyz[3*i], xyz[3*i+1], xyz[3*i+2]});
                atomCA.setPDBserial(i);
                atomCA.setName("CA");
                aminoAcid.addAtom(atomCA);

                aminoAcid.setAminoType(new Character((char)resType[i]));
                chain.addGroup(aminoAcid);
            }
            ArrayList model=new ArrayList();
            model.add(chain);
            structure=new StructureImpl();
            structure.addModel(model);
            synchronized(SYNC) {
                if (original) pcp(key,new java.lang.ref.SoftReference(structure),p);
            }
        }
        return structure;
    }
}
