package charite.christo.strap.extensions;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import charite.christo.strap.*;
import java.io.*;
import java.util.*;
import javax.swing.AbstractButton;
import static charite.christo.ChUtils.*;
/**HELP
<br><b>Publication:</b> PUBMED:15849316
<br><b>Home page:</b> <i>HOME_PAGE:Superimpose_TM_align</i>
<br><b>Authors:</b> Yang Zhang and Jeffrey Skolnick
   @author Christoph Gille

http://www.apple.com/downloads/macosx/development_tools/intelfortrancompiler,professionaledition,formacos.html

Versager: Superimposed s6stdout -load="PDB:1G51_B PDB:1J5W_A"
*/
public final class Superimpose_TM_align extends AbstractSuperposition implements HasScore, SequenceAligner, Superimpose3D {
    private final static AbstractButton cbCommutative=cbox("improve speed: reduce # of computations by commutativity"); static { cbCommutative.setSelected(true);}

    @Override public Result computeResult(byte[][] ssNotNeeded, Protein[] proteins) {
        final String n="TM_align2";
        final AbstractExecutable e=new AbstractExecutable(n);
        e.setBinaryPackageURLs();
        setExecutable(e,(String[])null, "FORTRAN_COMPILER  MNO_CYGWIN_FLAG "+n+".f -o "+n+".exe", AbstractExecutable.DEFAULT_BINARY);
        if (!Insecure.EXEC_ALLOWED) return null;
        final long fileOpt=ProteinWriter.PDB|ProteinWriter.ATOM_LINES|ProteinWriter.MET_INSTEAD_OF_MSE;
        final File dir=dirTemp(),  fR=pdbFile(proteins[0], 0, (boolean[])null,fileOpt), fM=pdbFile(proteins[1], 1, (boolean[])null, fileOpt);
        if (e.execNZ() || fM==null || fR==null) return null;
        e.exec().setCommandLineV(e.fileExecutable(),fM.getName(),fR.getName()).dir(dir).setOptions(ChExec.STDOUT | ChExec.STDERR | ChExec.LOG);
        e.exec().run();
        final ByteArray output=e.exec().getStdoutAsByteArray();
        final byte[] T=output!=null ? output.getBytes() : null;
        if (T==null) return null;
        final int l=output.getEnd();
        //putln("Superimpose_TM_align \n",output);
        final Result result=new Result();
        try {
            final ChTokenizer TOKEN=new ChTokenizer();
            final double r[][]=new double[3][3], t[]=new double[3];
            int pos=strstr("ROTATION M",T,0,l);
            if (pos>=0) {
                while(pos<l && T[pos++]!='\n');
                TOKEN.setText(T,pos,l);
                for(int i=0;i<3;i++) {
                    for(int j=0;j<3;j++) {
                        if (!TOKEN.nextToken()) return null;
                        r[i][j]=atof(T,TOKEN.getFrom());
                    }
                    if (!TOKEN.nextToken()) return null;
                    t[i]=atof(T,TOKEN.getFrom());
                }
            } else {
                 pos=strstr("rotation matrix to",T,0,999999);
                 if (pos>0) {
                     TOKEN.setText(T,strstr(" 1 ",T,pos,l)-1,l);
                     for(int i=0; i<3; i++) {
                         if (!TOKEN.nextToken()) return null;
                         if (!TOKEN.nextToken()) return null;
                         t[i]=atof(T,TOKEN.getFrom());
                         for(int j=0;j<3;j++) {
                             if (!TOKEN.nextToken()) return null;
                             r[i][j]=atof(T,TOKEN.getFrom());
                         }
                     }
                 }
            }

            result.setMatrix(new Matrix3D().setRotation(r).setTranslation(t));

            pos=strstr("Aligned length=",T,0,l)+1;
            if (pos<1||pos>l) { putln("TM_align1 not found \"Aligned length=\" "+output.newString());return null;}
            AbstractSuperposition.parse(output,result, Superimpose3D.PARSE_SCORE);
            {
                final Protein pR=proteins[0], pM=proteins[1];
                pos=strstr("(\":\" ",T,pos,l)+1;
                if (pos<1||pos>l) { putln("TM_align1 not found \"(\"...\" ");return null;}
                while(pos<l && T[pos++]!='\n');
                final int posMobile=pos;
                while(pos<l && T[pos++]!='\n');
                final byte[] ggM=new byte[pos-posMobile]; System.arraycopy(T,posMobile,ggM,0,ggM.length);
                while(pos<l && T[pos++]!='\n');
                final int posRefer=pos;
                while(pos<l && T[pos++]!='\n');
                final byte[] ggR=new byte[pos-posRefer]; System.arraycopy(T,posRefer,ggR,0,ggR.length);

                final byte[][] gapped=insertWhereNoCoordinates(0,ggR,0,ggM, pR, pM);
                result.setGappedSequences(gapped);

                //result.setGappedSequences(new byte[][]{ggR,ggM});
            }
        } catch(Exception ex) { debug(output); stckTrc(ex); }
        return result;
    }
   /*
      @Override public String getBinaryPackageURL() {
      String b=super.getBinaryPackageURL();
      if (IS_86 && isMacintosh()) { b=b.replaceAll(".zip",".x86.zip"); putln(b);}
      return  b;
      }
    */
}
