package charite.christo.strap.extensions;
import static charite.christo.ChUtils.*;
import charite.christo.*;
import java.util.*;
import java.io.File;
/**
   @author Christoph Gille
*/
public class Toppred  extends AbstractExecutable {
    private final String PARAMERS[][]={
        {"-c","<val>    Use <val> certain cut-off. ( Default value: 1 )"},
        {"-d","<val>    Use <val> as critical distance between 2 transmembrane segments. ( Default value: 2 )"},
        {"-e","         For use with Eucaryotes."},
        {"-H","{CYTEXT-scale,GES-scale,GVH-scale,KD-scale} <file>   Use Hydrophobycitie values from <file>. "},
        {"-n","<val>    Use <val> as core window length.  ( Default value: 11 )"},
        {"-p","<val>    Use <val> as putative cut-off. ( Default value: 0.6 )"},
        {"-q","<val>    Use <val> as wedge window length. ( Default value: 5 )"},
        {"-s","<val>    Use <val> as critical loop length.  ( Default value: 60 )"}
    };

    public Toppred() {
        setName("toppred");
        setSourceInstallationScript("sh configure \n make \n mv  src/toppred toppred.exe");
        getPrgParas().defineGUI(PARAMERS, PrgParas.OPTION_FURTHER_PARAMETERS);
    }

    private String sequence;
    private String seqName="test";
    public void setSequence(String name,String s) {seqName=name; sequence=onlyLetters(s); }
    private char[] prediction;
    private double[] hydrophobicity;
    public char[] getTM_Helices(){ return prediction;}
    public double[] getHydrophobicity(){  return hydrophobicity; }
    // java charite.christo.strap.extensions.TransmembraneHelix_TOPPRED
    /* --- Computation --- */

    private String fnPrediction,fnSequence;
    private static boolean isInitializing;
    public void compute() {
        if (isInitializing) return;
        isInitializing=true;
        iniT();
        isInitializing=false;
        if (sequence==null) return;
        final File
            dirTemp=dirTemp(),
            fSequence=file(dirTemp,fnSequence=seqName+".fa"),
            fGraph=file(dirTemp,seqName+".ps"),
            fPrediction=file(dirTemp,fnPrediction=seqName+".pred"),
            fHydro=file(dirTemp,seqName+".hydro"),
            dirData=file(dirBinaries(),"data");
        delFile(fSequence);
        delFile(fGraph);
        delFile(fPrediction);
        delFile(fHydro);
        wrte(fSequence,">"+seqName+"\n"+sequence+"\n");
        exec().setCommandLineV(fileExecutable(), getPrgParas().asStringArray(0L), "-g", "ps", "-o", fnPrediction, fnSequence)
            .dir(dirTemp).setOptions(ChExec.STDOUT | ChExec.STDERR | ChExec.CYGWIN_DLL | ChExec.LOG)
            .addToEnvironement(("TOPPREDDATA="+dirData).replace('\\','/'))
            .run();
        /* --- Transmembrane --- */
         String[] ss=readLines(fPrediction);
        final int L=sze(sequence);

        if (ss!=null) {
            boolean started=false;
            final char pred[]=new char[L];
            Arrays.fill(pred,' ');
            for(String s:ss){
                if (started) {
                    if ("".equals(s.trim())) break;
                    final StringTokenizer t=new StringTokenizer(s," \t-");
                    t.nextToken();
                    final int from=atoi(t.nextToken())-1;
                    final int to=atoi(t.nextToken())-1;
                    if (from<0|| from>=L || to<0|| to>=L) {
                        putln("error in Toppred: from="+from+" to="+to);
                    }
                    else Arrays.fill(pred,from,to,'H');
                }
                started=started||s.startsWith(" Helix Begin - End   Score Certainity");
            }
            prediction=pred;
        }
        /* --- Hydrophobicity  --- */
        ss=readLines(fHydro);
        if (ss!=null) {
            final double[] ff=new double[L];
            int iAa=0;
            for(int i=0; i<ss.length && iAa<L; i++) {
                final String s=ss[i];
                final int iSpace=s.indexOf('\t');
                if (s.startsWith("#")|| iSpace<0) continue;
                ff[iAa++]=atof(s,iSpace);
            }
            final int start=(L-iAa)/2;
            System.arraycopy(ff,0, ff, start,iAa);
            Arrays.fill(ff,0,start,Double.NaN);
            Arrays.fill(ff,start+iAa,L,Double.NaN);
            hydrophobicity=ff;
        }
    }
}
