package charite.christo.strap.extensions;
import charite.christo.*;
import charite.christo.protein.*;
import charite.christo.interfaces.*;
import java.awt.*;
import javax.swing.*;
import java.util.*;
import static charite.christo.ChUtils.*;
/**HELP
   Server: <i>HOME_PAGE:TransmembraneHelix_TMAP_MULTI</i>
   <br>
   It takes a multiple sequence alignment as input and predicts the transmembrane helices from the alignment.
   <br>
   Authors:By  B Persson and P Argos
   <br>
   Publication: PUBMED:9246628
   @author Christoph Gille
*/
public class TransmembraneHelix_TMAP_MULTI implements TransmembraneHelix_Predictor,HasControlPanel,NeedsInternet {

    public String sequences[], alignedSequences[],names[];
    public void setGappedSequences(String ss[]) {
        if (ss!=null) {
            sequences=new String[ss.length];
            for(int i=ss.length; --i>=0; ) sequences[i]=onlyLetters(ss[i]);
        }
    }
    private final ChTextArea LOG=new ChTextArea("");
    private char[][] prediction;
    public char[][] getPrediction(){ return prediction;}
    private Container controlPanel;
    public Container getControlPanel() {
        if (controlPanel==null) {
            controlPanel=newCNSEW(
                                  new JScrollPane(LOG),
                                  TabItemTipIcon.homePage(this),
                                  null,
                                  null,null
                                  );
        }
        return controlPanel;
    }

    public void compute(){
        final int n=sequences.length;
        if (n<2) { error("TransmembraneHelix_TMAP_MULTI requires  at least two aligned sequences");return;}
        prediction=new char[n][0];
        if (n>0) prediction[n-1]=TransmembraneHelix_TMAP.parse(getResults(sequences));
    }
    public ByteArray getResults(String sequences[]) {
        //StringBuffer sbAli=MakeAlignmentFileWithoutGaps.msf(new StringBuffer(),sequences,50);
        final StringBuffer sbAli=ChAlignUtils.toGappedMSF(strings2bytes(sequences),"s",0,false,null);
        final String  SERVER="http://bioinfo.limbo.ifm.liu.se/tmap/tmap1.pl";
        ByteArray data=Web.getServerResponse(SERVER,new Object[][]{{"sequence",sbAli} });
        LOG.appnd("contacting ",SERVER,data,"\n");
        if (data==null) {LOG.appnd("error: could not load\n"); return null; }
        return data;
    }
}
